//
// File:         Array.h
// Description:  Rubber arrays.
// Created:      Jan 30, 2000, 20:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef ARRAY_H
//=============================================================================
#define ARRAY_H 
#include "jargon.h"
#include "BLDebugFlags.h"
#include "ClassDesc.h"
#include "Exit.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ARRAY
 #define DEBUG_NAMESPACE "Array<class Alloc,class ElType,ulong Quantum,class InstanceId>"
#endif
#include "debugMacros.h"
//============================================================================

template <class Alloc,class ElType,ulong Quantum,class InstanceId>
class Array
{
 public:
  Array() : _currSize(Quantum) 
  {
   _memory = (ElType*)Alloc::allocate(sizeof(ElType)*Quantum);
   #ifdef DEBUG_ARRAY
    _name = "?";
   #endif
  }; 
  Array(const char* nm) : _currSize(Quantum) 
  {
   _memory = (ElType*)Alloc::allocate(sizeof(ElType)*Quantum);
   #ifdef DEBUG_ARRAY
    _name = nm;
   #endif
  }; 
  Array(const ElType& initElemVal,int /* dummy */) : _currSize(Quantum) 
  {
   _memory = (ElType*)Alloc::allocate(sizeof(ElType)*Quantum);
   _initialElementValue = initElemVal;
   for (ulong i = 0; i < Quantum; i++)
    _memory[i] = _initialElementValue;
   #ifdef DEBUG_ARRAY
    _name = "?";
   #endif
  }; 
  Array(const ElType& initElemVal,const char* nm) : _currSize(Quantum) 
  {
   _memory = (ElType*)Alloc::allocate(sizeof(ElType)*Quantum);
   _initialElementValue = initElemVal;
   for (ulong i = 0; i < Quantum; i++)
    _memory[i] = _initialElementValue;
   #ifdef DEBUG_ARRAY
    _name = nm;
   #endif
  }; 

  ~Array() { CALL("destructor ~Array()"); if (!Exit::fast) destroy(); };
  void destroy() { Alloc::deallocate((void*)_memory,sizeof(ElType)*_currSize); };
  void* operator new(size_t) 
  {
   CALL("operator new (size_t)");
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    ALWAYS(_classDesc.registerAllocated(sizeof(Array)));
    return Alloc::allocate(sizeof(Array),&_classDesc); 
   #else
    return Alloc::allocate(sizeof(Array)); 
   #endif
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)");
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    ALWAYS(_classDesc.registerDeallocated(sizeof(Array)));
    Alloc::deallocate(obj,sizeof(Array),&_classDesc); 
   #else
    Alloc::deallocate(obj,sizeof(Array)); 
   #endif
  }; 
  ElType& operator[](ulong i) 
  {
   CALLM(_name,"operator[](ulong i)");
   if (i >= _currSize) expand(i);
   return sub(i);
  };
  const ElType& sub(ulong i) const // unsafe but efficient
  {
   CALLM(_name,"const ElType& sub(ulong i) const");
   ASSERT(i < _currSize);
   return _memory[i];
  };
  ElType& sub(ulong i) // unsafe but efficient
  {
   CALLM(_name,"ElType& sub(ulong i)");
   ASSERT(i < _currSize);
   return _memory[i];
  };
  ulong size() const { return _currSize; };
  static ulong minSize() { return sizeof(Array); }; // has nothing to do with size()
  static ulong maxSize() { return sizeof(Array); }; // has nothing to do with size()
 private:
  void expand(ulong i)
  {
   CALLM(_name,"expand(ulong i)");
   ASSERT(i >= _currSize);
   ulong newSize = (i/Quantum)*Quantum;
   if (i >= newSize) newSize += Quantum; 
   ASSERT(i < newSize);
   ElType* newMemory = (ElType*)Alloc::allocate(sizeof(ElType)*newSize);
   for (ulong j = 0; j < _currSize; j++) newMemory[j] = _memory[j];
   for (ulong k = _currSize; k < newSize; k++) newMemory[k] = _initialElementValue;
   Alloc::deallocate((void*)_memory,sizeof(ElType)*_currSize);
   _currSize = newSize;  
   _memory = newMemory;
   ASSERT(i < _currSize);
  };
 private:
  ElType* _memory;
  ulong _currSize;   
  ElType _initialElementValue; 
  #ifdef DEBUG_ARRAY
   const char* _name;
  #endif
  #ifdef DEBUG_ALLOC_OBJ_TYPE 
   static ClassDesc _classDesc;
  #endif 
}; // template <class Alloc,class ElType,ulong Quantum,class InstanceId> class Array
//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
