//
// File:         CommandLine.cpp
// Description:  Parsing command line and printing usage.
// Created:      Apr 12, 2001, 18:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//
//======================================================================================
#include "CommandLine.h"
//======================================================================================
#ifdef DEBUG_ALLOC_OBJ_TYPE
 ClassDesc CommandLine::StrList::_classDesc("CommandLine::StrList",
                                            CommandLine::StrList::minSize(),
                                            CommandLine::StrList::maxSize()); 
 
 ClassDesc CommandLine::OptDescList::_classDesc("CommandLine::OptDescList",
                                                CommandLine::OptDescList::minSize(),
                                                CommandLine::OptDescList::maxSize());
                                           
#endif
//======================================================================================


bool CommandLine::parseOptions()
{
 int argn = 1;
 const char* arg;
 OptDescList* desc;

 next_arg: 
  if (argn >= argc) { nonoptArg = argn; return true; }; // all arguments are options
  arg = argv[argn];
  if (arg[0] != '-') { nonoptArg = argn; return true; };  
  // arg is an option
  if (arg[1] == '-') // arg is a long option or "--"
   {
    if (!(arg[2])) { nonoptArg = argn + 1; return true; }; // explicit end of options   
    desc = findDesc(arg + 2,false);    
    if (!desc) { nonoptArg = argn; return false; };  
    if (desc->hd().parName) // long option with parameter
     {    
      if (strlen(desc->hd().name) < strlen(arg + 2)) 
       {// treat the rest of arg as a parameter
	if (!checkType(arg + (2 + strlen(desc->hd().name)),desc->hd().ptype))     
	 { nonoptArg = argn; return false; };
        desc->hd().addVal(arg + (2 + strlen(desc->hd().name)));       
        argn++;
        goto next_arg;        
       }
      else // read parameter from the next argument 
       {
	argn++;
	if ((argn >= argc) || (!checkType(argv[argn],desc->hd().ptype)))
         { nonoptArg = argn; return false; };                   
        desc->hd().addVal(argv[argn]);
        argn++;
        goto next_arg;
       };
     }
    else // long option without parameter
     {
      if (strlen(desc->hd().name) != strlen(arg + 2)) 
       { 
        nonoptArg = argn; return false;     
       };
      desc->hd().addVal(arg);       
      argn++;
      goto next_arg;
     };
   } 
  else // arg is a collection of short options or the non-option "-"
   {
    if (!(arg[1])) { nonoptArg = argn; return true; };   
    for (unsigned c = 1; c < strlen(arg); c++)
     {     
      desc = findDesc(arg + c,true);

      if (!desc) { nonoptArg = argn; return false; };     
      if (desc->hd().parName)
       {// short option with parameter      
	if (c + 1 < strlen(arg))
	 {// treat the rest of arg as a parameter
	  if (!checkType(arg + (c + 1),desc->hd().ptype)) 
           { nonoptArg = argn; return false; };        
	  desc->hd().addVal(arg + (c + 1));           
          argn++;
          goto next_arg;
         }
        else// read parameter from the next argument
	 {
          argn++;
	  if ((argn >= argc) || (!checkType(argv[argn],desc->hd().ptype))) 
           { nonoptArg = argn; return false; };                                
          desc->hd().addVal(argv[argn]);
          argn++;
          goto next_arg;
         };
       }
      else // short option without parameter
       {
	desc->hd().addVal(arg);
       };
     };

    argn++;
    goto next_arg;
   };
}; // bool CommandLine::parseOptions() const 

CommandLine::OptDescList* CommandLine::findDesc(const char* optName,bool shortOpt) const
{
 OptDescList* lst = optDescs;
 while (lst)
  {
   if ((strpref(lst->hd().name,optName)) && (lst->hd().shortOpt == shortOpt)) return lst;
   lst = lst->tl(); 
  };
 return (OptDescList*)0;
}; // OptDescList* CommandLine::findDesc(const char* optName)

bool CommandLine::checkType(const char* str,ParamType tp)
{
 long dummyLong; 
 double dummyDouble;
 bool dummyBool;

 #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
  switch(tp)
  { 
    case Long: return Values::readInf(str,dummyBool) || StrToX::conv(dummyLong,str);
    case Double: return Values::readInf(str,dummyBool) || StrToX::conv(dummyDouble,str);
    case String: return true;
    default : return false;
  };
 #else
  switch(tp)
  { 
    case Long: return Values::readInf(str,dummyBool) || StrToX::conv(dummyLong,(char*)str);
    case Double: return Values::readInf(str,dummyBool) || StrToX::conv(dummyDouble,(char*)str);
    case String: return true;
    default : return false;
  }; 
  
 #endif

}; // bool CommandLine::checkType(const char* str,ParamType tp)

int CommandLine::findStr(const char* str,int arrSize,const char* const *arr)
{
 for (int i = 0; i < arrSize; i++) if (!strcmp(str,arr[i])) return i; 
 return -1;
}; // int CommandLine::findStr(const char* str,int arrSize,const char* const *arr)    

bool CommandLine::Values::next(const char*& str)
{
 if (bool(*this))
  {
   str = StrListIter::currEl();   
   StrListIter::next();
   return true;
  };
 return false;
}; // bool CommandLine::Values::next(const char*& str)

bool CommandLine::Values::next(long& n)
{
 if (bool(*this))
  {  
   bool posInf;
   if (readInf(StrListIter::currEl(),posInf))
    {
     n = (posInf) ? LONG_MAX : LONG_MIN;
     StrListIter::next();
     return true;
    };
   #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
    StrToX str(StrListIter::currEl());
   #else
    StrToX str((char*)StrListIter::currEl());   
   #endif
   StrListIter::next();
   str.read(n);
   return true;
  };
 return false; 
}; // bool CommandLine::Values::next(long& n)

bool CommandLine::Values::next(double& n)
{
 if (bool(*this))
  {  
   bool posInf;
   if (readInf(StrListIter::currEl(),posInf))
    {
     n = (posInf) ? DBL_MAX : DBL_MIN;
     StrListIter::next();
     return true;
    };

   #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
    StrToX str(StrListIter::currEl());
   #else
    StrToX str((char*)StrListIter::currEl());   
   #endif

   StrListIter::next();
   str.read(n);
   return true;
  };
 return false; 
}; // bool CommandLine::Values::next(double& n)

void CommandLine::printUsage(const char* command) const
{
 OptDescList* tmpOptDescs = OptDescList::inv(optDescs);
 cerr << "usage: " << command << " ";
 if (tmpOptDescs) cerr << "[options] [--] ";
 int i;
 for (i = 0; i < nonoptPrefSize; i++) cerr << nonoptPrefNames[i] << " ";
 if (nonoptMiddleName) cerr << nonoptMiddleName << " .. " << nonoptMiddleName << " ";
 for (i = 0; i < nonoptSuffSize; i++) cerr << nonoptSuffNames[i] << " ";  
 cerr << "\n";
 if (tmpOptDescs)
  {
   cerr << "options: "; 
   int currEnd = 9;
   int scount;
   for(OptDescList* desc = tmpOptDescs; desc; desc = desc->tl())
    {
     scount = strlen(desc->hd().name);
     if (desc->hd().shortOpt) { scount++; } else scount += 2; 
     if (desc->hd().parName) scount += (strlen(desc->hd().parName) + 1);  
     if (desc->tl()) scount += 2; 

     if (scount + currEnd > 90) { currEnd = 9 + scount; cerr << "\n         "; };  
     currEnd += scount;
     cerr << "-" << ((desc->hd().shortOpt) ? "" : "-");
     cerr << desc->hd().name;
     if (desc->hd().parName) cerr << " " << desc->hd().parName;  
     if (desc->tl()) cerr << ", "; 
    };
   cerr << "\n";
  };
 OptDescList::destroy(tmpOptDescs);
}; // void CommandLine::printUsage() const

void CommandLine::printHelp(const char* command) const
{
 printUsage(command);
 unsigned indent;
 const long stdLineLength = 80;

 // comments on nonoption arguments
 if (nonoptPrefNames)
  {
   for (int i = 0; i < nonoptPrefSize; i++)
    {
     cerr << "  " << nonoptPrefNames[i] << " ";
     indent = strlen(nonoptPrefNames[i]) + 3;
     FormatText fcomm(nonoptPrefComments[i]);
     const char* cmtLine = fcomm.nextLine(stdLineLength - indent);
     if (cmtLine) 
      { 
       cerr << cmtLine << "\n"; 
       while ((cmtLine = fcomm.nextLine(stdLineLength - 5))) 
        cerr << "     " << cmtLine << "\n";
      }
     else cerr << "\n";     
    };
  };

 if (nonoptMiddleName) 
  {
   cerr << "  " << nonoptMiddleName << ".." << nonoptMiddleName << " ";   
   indent = 2*strlen(nonoptMiddleName) + 5;
   FormatText fcomm(nonoptMiddleComment);
   const char* cmtLine = fcomm.nextLine(stdLineLength - indent);
   if (cmtLine) 
    { 
     cerr << cmtLine << "\n"; 
     while ((cmtLine = fcomm.nextLine(stdLineLength - 5))) 
      cerr << "     " << cmtLine << "\n";
    }
   else cerr << "\n";   
  };

 if (nonoptSuffNames)
  {
   for (int i = 0; i < nonoptSuffSize; i++)
    {
     cerr << "  " << nonoptSuffNames[i] << " ";
     indent = strlen(nonoptSuffNames[i]) + 3;
     FormatText fcomm(nonoptSuffComments[i]);
     const char* cmtLine = fcomm.nextLine(stdLineLength - indent);
     if (cmtLine) 
      { 
       cerr << cmtLine << "\n"; 
       while ((cmtLine = fcomm.nextLine(stdLineLength - 5))) 
        cerr << "     " << cmtLine << "\n";
      }
     else cerr << "\n";     
    };
  };
 
  // comments on options

 OptDescList* tmpOptDescs = OptDescList::inv(optDescs);
 if (tmpOptDescs)
  {
   cerr << "\n";   
   for (OptDescList* desc = tmpOptDescs; desc; desc = desc->tl())
    {   
     cerr << "  " << "-";
     indent = 3;
     if (!desc->hd().shortOpt) { cerr << "-"; indent++; };
     cerr << desc->hd().name;
     indent += strlen(desc->hd().name);
     if (desc->hd().parName) 
      { 
       cerr << " " << desc->hd().parName; 
       indent += (strlen(desc->hd().parName) + 1);
      };
     cerr << " ";
     indent++;
     FormatText fcomm(desc->hd().comment);
     const char* cmtLine = fcomm.nextLine(stdLineLength - indent);
     if (cmtLine) 
      { 
       cerr << cmtLine << "\n"; 
       while ((cmtLine = fcomm.nextLine(stdLineLength - 5))) cerr << "     " << cmtLine << "\n";
      }
     else cerr << "\n";
   };
  };
 OptDescList::destroy(tmpOptDescs);
}; // void CommandLine::printHelp(const char* command) const

void CommandLine::printError() const
{
 if (!optParsSuccess) 
  {
   cerr << "Error in " << argv[nonoptArg] << " : can not parse options.\n"; 
  }
 else
  {
   cerr << "Error after " << argv[nonoptArg - 1] << " : wrong number of non-option arguments.\n"; 
  };
}; // void CommandLine::printError() const

ostream& CommandLine::printCommand(ostream& str) const
{
 for (int i = 0; i < argc; i++) str << argv[i] << ' ';
 return str;
}; // ostream& CommandLine::printCommand(ostream& str) const







//======================================================================================



