//
// File:         Debug.cpp
// Description:  
// Created:      Oct 1, 1999, 20:25
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//
//==============================================================================
#ifndef NO_DEBUG
//==============================================================================
#include <iostream.h>
#include <signal.h>
#include "Debug.h"
//==============================================================================
const char* Debug::job = (const char*)0;  
ulong Debug::eventId = 0;
const char* Debug::controlPoint = (const char*)0;
const char* Debug::callName = (const char*)0;
long Debug::callDepth = 0;
const char* Debug::objectName = (const char*)0;
const char* Debug::nameSpace = (const char*)0;
const char* Debug::errorDescription = 0; 
Debug::StrList* Debug::controlPoints = (StrList*)0;
Debug::StrList* Debug::callNames = (StrList*)0;
Debug::StrList* Debug::objectNames = (StrList*)0;
Debug::StrList* Debug::nameSpaces = (StrList*)0;
Debug::StrList* Debug::recycledSL = (StrList*)0;
ostream* Debug::trace = (ostream*)0;
bool Debug::traceCP = false;
bool Debug::traceAfterError = false;
Debug Debug::_initStatic; 

Debug::Call::Call(const char* nameSp,const char* callNm)
{
 nextEvent();
 callDepth++;
 controlPoints = newStrList(controlPoint,controlPoints);  
 callNames = newStrList(callName,callNames);
 objectNames = newStrList(objectName,objectNames);
 nameSpaces = newStrList(nameSpace,nameSpaces);
 controlPoint = "start";
 callName = callNm;
 objectName = "?";
 nameSpace = nameSp;
 if (trace)
  {
   *trace << eventId << " -> [" << callDepth << "] " << nameSpace << "::" << callName 
          << " [" << objectName << "] from " << controlPoints->hd << '\n'; 
  };
}; // Debug::Call::Call(const char* nameSp,const char* callNm)

Debug::Call::Call(const char* nameSp,const char* objName,const char* callNm)
{ 
 nextEvent();
 callDepth++; 
 controlPoints = newStrList(controlPoint,controlPoints);  
 callNames = newStrList(callName,callNames);
 objectNames = newStrList(objectName,objectNames);
 nameSpaces = newStrList(nameSpace,nameSpaces);
 controlPoint = "start";
 callName = callNm;
 objectName = objName;
 nameSpace = nameSp;
 if (trace)
  {
   *trace << eventId << " -> [" << callDepth << "] " << nameSpace << "::" << callName 
          << " [" << objectName << "] from " << controlPoints->hd << '\n'; 
  };
}; // Debug::Call::Call(const char* nameSp,const char* objName,const char* callNm)

Debug::Call::~Call()
{
 nextEvent();
 StrList* tmp;
 if (trace)
  {
   *trace << eventId << " <- [" << callDepth << "] " << nameSpace << "::" << callName 
          << " [" << objectName << "] after " << controlPoint << '\n'; 
  };
 
 // popping control point
 controlPoint = controlPoints->hd;
 tmp = controlPoints->tl;
 recycle(controlPoints);
 controlPoints = tmp; 

 // popping call name
 callName = callNames->hd;
 tmp = callNames->tl;
 recycle(callNames);
 callNames = tmp; 

 // popping object names
 objectName = objectNames->hd;
 tmp = objectNames->tl;
 recycle(objectNames);
 objectNames = tmp; 

 // popping name space
 nameSpace = nameSpaces->hd;
 tmp = nameSpaces->tl;
 recycle(nameSpaces);
 nameSpaces = tmp;

 callDepth--;   
}; // Debug::Call::~Call() 


void Debug::initStatic()
{
 #ifdef DEBUG_ERROR_SIGNALS
  initSigHandlers();
 #endif
 callDepth = 0;
 job = "unknown job";
 eventId = 0;
 controlPoint = "start";
 callName = "main";
 objectName = "main";
 nameSpace = "main";
 errorDescription = "unknown";
 controlPoints = (StrList*)0;
 callNames = (StrList*)0;
 objectNames = (StrList*)0;
 nameSpaces = (StrList*)0;
 recycledSL = (StrList*)0;
 trace = (ostream*)0;
 traceCP = false;   
 traceAfterError = false;
};

void Debug::reportCP(ulong ev,const char* ns,const char* cn,const char* cp,const char* on)
{
 cout << ev << " " << ns << "::" << cn << '/' << cp << " [" << on << ']';
};

void Debug::error()
{
 cout << "Job " << job << " aborted.\n"; 
 cout << "Error : " << errorDescription << "\n in ";
 reportCP(eventId,nameSpace,callName,controlPoint,objectName);
 cout << '\n';
 history(); 
 cout << "The job was " << job << '\n';
 if (!traceAfterError) tracingOff();
 Exit::exit(1);
};

void Debug::error(const char* msg)
{
 errorDescription = msg;
 error();
};

void Debug::debugFeatureInternalFault(const char* msg)
{
 cout << "Job " << job << " aborted.\n"; 
 cout << "Error : debug feature internal fault : " << msg << "\n in ";
 reportCP(eventId,nameSpace,callName,controlPoint,objectName);
 cout << '\n';
 history(); 
 cout << "The job was " << job << '\n';
 if (!traceAfterError) tracingOff();
 Exit::exit(1);
};

void Debug::reportCP()
{
 reportCP(eventId,nameSpace,callName,controlPoint,objectName); 
 cout << '\n';
};

void Debug::history()
{
 StrList* cpl = controlPoints;
 StrList* cnl = callNames;
 StrList* onl = objectNames;
 StrList* nsl = nameSpaces;
 int indent = 0;
 while (cpl)
  {
   for (int i = 0; i < indent; i++) cout << ' ';
   cout << "|_ ";
   reportCP(0,nsl->hd,cnl->hd,cpl->hd,onl->hd);
   cout << "\n";
   indent++;
   cpl = cpl->tl; 
   cnl = cnl->tl;
   onl = onl->tl;
   nsl = nsl->tl;
  };                 
}; // void Debug::history()

void Debug::reportSlowProgress(const char* checkName,float maxInt,float realInt)
{
  cout << "Debug warning: slow progress in " << checkName 
       << ", max. interval = " << maxInt << ", real interval = " << realInt << '\n';  
}; // void Debug::reportSlowProgress(const char* checkName,float maxInt,float realInt)


void Debug::initSigHandlers()
{
 #ifdef TARGET_OS_TYPE 	 
  #if (TARGET_OS_TYPE == OS_TYPE_UNIX)
   signal(SIGFPE,handleSIGFPE);
   signal(SIGILL,handleSIGILL);
   signal(SIGSEGV,handleSIGSEGV);
   signal(SIGBUS,handleSIGBUS);  
   signal(SIGTRAP,handleSIGTRAP);
   signal(SIGABRT,handleSIGABRT);
  #elif (TARGET_OS_TYPE == OS_TYPE_WIN32)
   signal(SIGFPE,handleSIGFPE);
   signal(SIGILL,handleSIGILL);
   signal(SIGSEGV,handleSIGSEGV); 
  #else
   #error "Unsupported target OS type, check TARGET_OS_TYPE"
  #endif
 #else
  #error "Undefined TARGET_OS_TYPE"
 #endif
};
                
void Debug::handleSIGFPE(int)
{
 error("Program Error Signal SIGFPE (arithmetic error)");
};
void Debug::handleSIGILL(int)
{
 error("Program Error Signal SIGILL (illegal instruction)");
};
void Debug::handleSIGSEGV(int)
{
 error("Program Error Signal SIGSEGV (segmentation violation)");
};
void Debug::handleSIGBUS(int) 
{
 error("Program Error Signal SIGBUS (bus error)");
};
void Debug::handleSIGABRT(int)
{
 error("Program Error Signal SIGABRT");
};
/******
void Debug::handleSIGIOT(int)
{
 error("Program Error Signal SIGIOT");
};
*******/
void Debug::handleSIGTRAP(int)
{
 error("Program Error Signal SIGTRAP");
};

//=========================================================================
#endif
