//
// File:         FormatText.h
// Description:  Splits a text into lines and removes extra white space.  
// Created:      Feb 2, 2000, 16:45
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//=================================================================
#ifndef FORMAT_TEXT_H
//=================================================================
#define FORMAT_TEXT_H 
#include <string.h>
#include "jargon.h"
#include "BLDebugFlags.h"
#include "ClassDesc.h"
#include "Malloc.h"
#include "GList.h"
//=================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORMAT_TEXT
 #define DEBUG_NAMESPACE "FormatText"
#endif
#include "debugMacros.h" 

//=================================================================
class FormatText
{
 private:
  typedef GList<Malloc,char*,FormatText> StrList;
 public:
  FormatText(const char* str) 
   : words(0),
     currLine(new char[1])
  {
   readWords(str);
  };
  ~FormatText() { destroyWords(); delete [] currLine; };
  const char* nextLine(long lineLength);
 private: 
  FormatText() {};
  void readWords(const char* str);
  void destroyWords();
  bool spaceChar(char ch) { return (ch == ' ') || (ch == '\n') || (ch == '\t'); };
 private:
  StrList* words;
  char* currLine;
}; // class FormatText 

//================================================================

inline const char* FormatText::nextLine(long lineLength)
{
 if (!words) return (const char*)0; 
 StrList* tmp;
 if (long(strlen(words->hd())) + 2 >= lineLength)
  {
   delete [] currLine;
   currLine = new char[strlen(words->hd()) + 1];
   strcpy(currLine,words->hd());
   delete [] words->hd();
   tmp = words->tl();
   delete words; 
   words = tmp;
   return currLine;
  };

 // line of normal size 
 delete [] currLine;
 currLine = new char[lineLength + 1];
 currLine[lineLength] = (char)0;
 long currSym = 0;
 next_word:
  if ((!words) || (long(strlen(words->hd())) + 1 + currSym > lineLength)) 
   {
    while (currSym < lineLength) { currLine[currSym] = ' '; currSym++; };         
    return currLine;
   };
  currLine[currSym] = ' ';
  strcpy(currLine + currSym + 1,words->hd());
  currSym += 1 + strlen(words->hd());
  delete [] words->hd();
  tmp = words->tl();
  delete words; 
  words = tmp;
  goto next_word;
}; // const char* FormatText::nextLine(long lineLength)

inline void FormatText::readWords(const char* str)
{
 long n = strlen(str);
 bool spaces = true;
 char ch;
 long wlen = 0;
 const char* word = (const char*)0;
 char* tmpWord;

 for (long i = 0; i <= n; i++)
  {
   ch = str[i];
   if (spaces) // skipping spaces
    {
     if (!spaceChar(ch)) // begin new word
      {
       spaces = false;
       word = str + i; 
       wlen = 1;
      };
    }   
   else // reading a word
    {
     if (spaceChar(ch) || (!ch)) // end of the word   
      {
       spaces = true;        
       tmpWord = new char[wlen+1];
       tmpWord[wlen] = (char)0;
       strncpy(tmpWord,word,wlen);
       words = new StrList(tmpWord,words);
      }
     else // continue reading the word
      {
       wlen++;
      };
    };
  };
 
 StrList::inverse(words); 
}; // void FormatText::readWords(const char* str) 

inline void FormatText::destroyWords() 
{
 StrList* tmp;
 while (words) 
  { 
   delete [] words->hd();
   tmp = words->tl();
   delete words;
   words = tmp;
  };
}; // void FormatText::destroyWords()
//=================================================================
#undef DEBUG_NAMESPACE
//=================================================================
#endif
