//
// File:         OutPort.h
// Description:  Abstract class for output ports.
// Created:      Sep 15, 1999, 18:35
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//================================================================
#ifndef OUT_PORT_H
//================================================================
#define OUT_PORT_H
#include <iostream.h>
#include "jargon.h"
#include "TmpString.h"
//================================================================
class OutPort
{
 public:
  class Null;  
 public:
  OutPort() {};
  OutPort& operator<<(const char* str)
  {
   output(str);
   return *this;
  };
  OutPort& operator<<(const char& c)
  {
   char tmp[2];
   tmp[0] = c;
   tmp[1] = '\0'; 
   this->operator<<(tmp);
   return *this;
  }; 
  OutPort& operator<<(const int& n)
  {
   return *this << TmpString(n).str();
  };
  OutPort& operator<<(const long& n)
  {
   return *this << TmpString(n).str(); 
  };
  OutPort& operator<<(const unsigned int& n)
  {
   return *this << TmpString(n).str();  
  };
  OutPort& operator<<(const unsigned long& n)
  {
   return *this << TmpString(n).str();  
  };
  OutPort& operator<<(const bool& n)
  {
   return *this << TmpString(n).str();   
  }; 
  OutPort& operator<<(const float& n)
  {
   return *this << TmpString(n).str();   
  };  
  OutPort& operator<<(const double& n)
  {
   return *this << TmpString(n).str();  
  };
  virtual ~OutPort() {}; 
  virtual void output(const char* s) = 0;
  virtual void flush() = 0;
}; 

class OutPort::Null : public OutPort
{
 public: 
  ~Null() {};
  void output(const char*) {};
  void flush() {};
};

class OStreamPort : public OutPort
{
 public:
  OStreamPort() : theStream((ostream*)0) {};
  explicit OStreamPort(ostream* str) : theStream(str) {};
  void connect(ostream* str) { theStream = str; };
  ~OStreamPort() {};
  void output(const char* s) { if (theStream) (*theStream) << s; };  
  void flush() { theStream->flush(); };
 private:
  ostream* theStream; 
};
//==================================================================
#endif


