//
// File:         ULStack.h
// Description:  Stacks of unlimited size.
// Created:      Oct 07, 1999, 18:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk 
//=======================================================================
#ifndef U_L_STACK_H
//=======================================================================
#define U_L_STACK_H
#include "jargon.h"
#include "BLDebugFlags.h"
#include "GDLList.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_U_L_STACK
 #define DEBUG_NAMESPACE "ULStack<class Alloc,class ElementT>"
#endif
#include "debugMacros.h"
//======================================================================

template <class Alloc,class ElementT>
class ULStack
{
 public:
  ULStack() : _bottom((List*)0,(List*)0), _top(&_bottom), _size(0UL)
  {
   #ifdef DEBUG_U_L_STACK
    _name = "?";
   #endif  
  };
  ULStack(const char* nm) : _bottom((List*)0,(List*)0), _top(&_bottom), _size(0UL) 
  {
   #ifdef DEBUG_U_L_STACK
    _name = nm;
   #endif  
  };
  
  ~ULStack() { destroy(); CALLM(_name,"destructor ~ULStack()"); };
  void destroy() 
  {
   CALLM(_name,"destroy()");
   List* tmp1 = _bottom.before();
   while (tmp1)
    {
     List* tmp2 = tmp1;
     tmp1 = tmp1->before();
     delete tmp2;
    };
  };
  void init() 
  {
   #ifdef DEBUG_STACK
    _name = "?";
   #endif 
   _top = &_bottom; 
   _size = 0UL;
  };
 
  void init(const char* nm) 
  {
   #ifdef DEBUG_STACK
    _name = nm;
   #endif 
   _bottom.after() = (List*)0;
   _bottom.before() = (List*)0;
   _top = &_bottom; 
   _size = 0UL;
  };  

  void reset() 
  {
   _top = &_bottom;
   _size = 0UL;
  };

  operator bool() const { return _top != &_bottom; };
  bool empty() const { return _top == &_bottom; };
  bool nonempty() const { return _top != &_bottom; };

  ulong size() const { return _size; };

  ElementT& top() 
  {
   CALLM(_name,"(ElementT& top())"); 
   ASSERT(nonempty());
   return _top->val();
  };
  const ElementT& top() const 
  {
   CALLM(_name,"(const ElementT& top())"); 
   ASSERT(nonempty());
   return _top->val();
  };

  ElementT& next() 
  {
   CALLM(_name,"(ElementT& next())");
   if (!_top->before()) _top->before() = new List(_top);
   return _top->before()->val();
  };
  const ElementT& next() const 
  {
   CALLM(_name,"(const ElementT& next())");
   if (!_top->before()) _top->before() = new List(_top);
   return _top->before()->val();
  };

  void push(const ElementT& el) 
  {
   CALLM(_name,"push(const ElementT& el)");
   if (!_top->before()) _top->before() = new List(_top);
   _top = _top->before();
   _top->val() = el;
  };
  void push() 
  { 
   CALLM(_name,"push()");
   if (!_top->before()) _top->before() = new List(_top);
   _top = _top->before();
  };

  const ElementT& pop() 
  {
   CALLM(_name,"const ElementT& pop()");
   ASSERT(nonempty());
   _top = _top->after();
   return _top->before()->val();
  };

  void skipTop() 
  {
   CALLM(_name,"skipTop()");
   ASSERT(nonempty());
   _top = _top->after(); 
  }; 

 private:
  typedef GDLList<Alloc,ElementT> List; 
 
 private: 
  List _bottom;
  List* _top;
  ulong _size;
  #ifdef DEBUG_U_L_STACK
   const char* _name;
  #endif   
}; // template <class Alloc,class Alloc> class ULStack
//=========================================================================
#undef DEBUG_NAMESPACE
//=========================================================================
#endif
