//
// File:         GematogenCompilationTest.cpp
// Description:  File for testing compilation of the Gematogen library.
// Created:      Jan 28, 2001, 17:40
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include <iostream.h>
#include "jargon.h"
#include "ClassDesc.h"
#include "ObjDesc.h"
#include "ForwardMatchingIndex.h"
//====================================================

// Dummy classes and values to be used as parameters of the templates 

namespace GematogenCompilationTest
{
 const ulong MaxTermSize = 1024;
 const ulong MaxNumOfVariables = 1024;
 const ulong MaxNumOfFunctors = 1024;

 class Allocator
 {
  public:
   static void* allocate(size_t size); 
   static void deallocate(void* obj,size_t size); 
   #ifndef NO_DEBUG
    static void* allocate(size_t size,ClassDesc* classDesc); 
    static void deallocate(void* obj,size_t size,ClassDesc* classDesc); 
    static ObjDesc* hostObj(void* ptr); 
   #endif
   
 };

 class Symbol
 {
  public:
   typedef ulong Arity;
   typedef ulong Functor;
   typedef ulong Polarity;
  public:   
   Symbol() {};
   Symbol(Arity arity,Functor functor,Polarity polarity) {}; // creates a functor  
   ulong functor() const { return 0UL; };   // returns functor number
   ulong var() const { return 0UL; }; // returns variable number
   bool isComplex() const { return true; }; // symbol.isComplex() <=> symbol is a function 
   bool isVariable() const { return true; }; // symbol.isVariable() <=> symbol is a variable
   bool isNotVariable() const { return true; }; // symbol.isNotVariable() <=> symbol is not a variable, 
                                                // symbol.isNotVariable() is not necessarily the same as symbol.isComplex()
                                                // although symbol.isComplex() => symbol.isNotVariable()
   bool operator==(const Symbol& t) const { return true; };
   bool operator>(const Symbol& t) const { return true; }; 
   bool operator<(const Symbol& t) const { return true; }; 
  public:
   static const Polarity PositivePolarity = 0UL;
   static const Polarity NegativePolarity = 1UL;
 };
 class Flatterm
 {
  public:  
   const Symbol& symbolRef() const { return _sym; };
   bool equal(const Flatterm* t) const { return true; };
   const Flatterm* after() const { return this; };
   const Flatterm* next() const { return (this + 1); };
  private:
   Symbol _sym;
 };
 class IndexTreeLeafInfo
 {
 };
 class IndexInstanceId
 {
 };
 
 ForwardMatchingIndex<MaxTermSize,MaxNumOfVariables,MaxNumOfFunctors,Allocator,Symbol,Flatterm,IndexTreeLeafInfo,IndexInstanceId> index;
 
 void indexMethodsTest()
 {
  Symbol fun;
  Symbol var;
  
  // Integration 
  index.integrationReset(fun);
  index.integrationFunc(fun);
  index.integrationVar(var);
  ForwardMatchingIndexTreeLeaf<MaxTermSize,MaxNumOfVariables,Allocator,Symbol,Flatterm,IndexTreeLeafInfo,IndexInstanceId>* leaf;
  bool newLeaf;
  leaf = index.integrationGetLeaf(newLeaf);
  ulong num1 = index.integrationGetVariables().Size();
  const Flatterm** ptr1 = index.integrationMapVariable(var.var());

  // Retrieval
  Flatterm query[3];
  leaf = index.Match(query);
  leaf = index.MatchAgain();
  
  // Removal 
  index.removalReset(fun);
  index.removalFunc(fun);  
  index.removalVar(var);
  ForwardMatchingIndexTreeInstruction<MaxTermSize,MaxNumOfVariables,Allocator,Symbol,Flatterm,IndexTreeLeafInfo,IndexInstanceId>** branchToDestroy;
  index.removalFinish(leaf,branchToDestroy);
 
  // Destruction
    
  index.Destroy();
 }; 
   
}; // namespace GematogenCompilationTest
 

//====================================================
