//
// File:         BackwardSubsumption.cpp
// Description:  Interface for several BS related modules.
// Created:      Feb 26, 2000, 20:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#include "BackwardSubsumption.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUBSUMPTION
 #define DEBUG_NAMESPACE "BackwardSubsumption"
#endif
#include "debugMacros.h"
//============================================================================

bool BackwardSubsumption::query(const Clause* cl) 
{
 CALL("query(const Clause* cl)");
 INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_QUERY(cl);
 _queryClause = cl;
 bool res = preproQuery(cl); 
 #ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
  if (!res) INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_TOTAL;
 #endif
 return res;
}; // bool BackwardSubsumption::query(const Clause* cl)

bool BackwardSubsumption::subsumeNextSetMode() 
{ 
 CALL("subsumeNextSetMode()");
 ulong subsumedClNum;

 //COP("COP10");

 while (_machine.subsumeNextSetMode(subsumedClNum)) 
  {
    //COP("COP20");
   _recentlySubsumedCl = _clauseNumIndex->find(subsumedClNum);
   //COP("COP30");
   ASSERT(_recentlySubsumedCl);    
   //COP("COP40");
   #ifdef DEBUG_NAMESPACE
    _debugSubsumptionCheck.resetSubst();
    //COP("COP50");
    _machine.resetSubstitutionCollection();
    //COP("COP60");
    ulong debugVar;
    TERM debugTerm;
    while (_machine.substNextPair(debugVar,debugTerm)) 
     {
       //COP("COP70");
      _debugSubsumptionCheck.substPair(debugVar,debugTerm);
     };
    //COP("COP80");
//ASSERT(_debugSubsumptionCheck.allVarsAssigned(_queryClause));
    //COP("COP90");
//ASSERT(_debugSubsumptionCheck.subsumes(_queryClause,_recentlySubsumedCl));
    //COP("COP100");
   #endif
    //COP("COP110");
   INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_RESULT(_recentlySubsumedCl);
   //COP("COP120");
   if ((_recentlySubsumedCl->Active())
      && (subsumptionAllowedInSetMode(_queryClause,_recentlySubsumedCl)))
     // this check must be done earlier in _machine, right after clause equalizing
    return true;
  };
 return false; 
}; // bool BackwardSubsumption::subsumeNextSetMode()


bool BackwardSubsumption::subsumeNextMultisetMode() 
{ 
 CALL("subsumeNextMultisetMode()");
 ulong subsumedClNum;
 while (_machine.subsumeNextMultisetMode(subsumedClNum)) 
  {
   _recentlySubsumedCl = _clauseNumIndex->find(subsumedClNum);

   ASSERT(_recentlySubsumedCl);    
   #ifdef DEBUG_NAMESPACE
    _debugSubsumptionCheck.resetSubst();
    _machine.resetSubstitutionCollection();
    ulong debugVar;
    TERM debugTerm;
    while (_machine.substNextPair(debugVar,debugTerm)) 
     _debugSubsumptionCheck.substPair(debugVar,debugTerm);
    ASSERT(_debugSubsumptionCheck.allVarsAssigned(_queryClause));
    ASSERT(_debugSubsumptionCheck.subsumes(_queryClause,_recentlySubsumedCl));
   #endif
   INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_RESULT(_recentlySubsumedCl);
   if ((_recentlySubsumedCl->Active())
      && (subsumptionAllowedInMultisetMode(_queryClause,_recentlySubsumedCl)))
     // this check must be done earlier in _machine, right after clause equalizing
    return true;
  };
 
 return false; 
}; // bool BackwardSubsumption::subsumeNextMultisetMode()


 
//#define DEBUG_NAMESPACE "SHIT"
 
//DF;

void BackwardSubsumption::integrate(const Clause* cl)
{
 CALL("integrate(const Clause* cl)");
 ASSERT(_clauseNumIndex->find(cl->Number()) == cl); 
 //DF; cout << "            WRP-6   " << _termTraversal.wrapper[0].Content() << "\n";
 
 OLD_IMPL(_oldImplementation.integrate(cl));

 INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTEGRATE(cl);

 _integrator.clause(cl->Number());
 TERM header;

 for(LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
  { 
   header = *(iter.currentLiteral());
   if (header.arity())
    {
     if (header.IsSymLitHeader())
      {
       _integrator.symLitHeader(header);
       // integrating first argument

       _termTraversal.reset(*(iter.currentLiteral()->Args()));
       ASSERT(((bool)(_termTraversal)));

       while (_termTraversal)       
	{
	 _integrator.symLitSubterm(_termTraversal.term());  
         _termTraversal.next();
              
        };
       // integrating second argument
       _integrator.symLitSecondArg();
       _termTraversal.reset(*(iter.currentLiteral()->Args()->Next()));
       ASSERT(((bool)(_termTraversal)));
       while (_termTraversal)       
	{
	 _integrator.symLitSubterm(_termTraversal.term());         
         _termTraversal.next();
        };
      }
     else // ordinary literal
      {
       _integrator.litHeader(header);
       _termTraversal.reset(iter.currentElement()->literalRef());
       _termTraversal.next();
       ASSERT(((bool)(_termTraversal)));
       while (_termTraversal)       
	{
	 _integrator.subterm(_termTraversal.term());         
         _termTraversal.next();
        };
      };
    } 
   else // prop. lit.
    _integrator.propLit(header);
  };
 ASSERT(preproQuery(cl));
}; // void BackwardSubsumption::integrate(const Clause* cl)

//#undef DEBUG_NAMESPACE 

void BackwardSubsumption::remove(const Clause* cl)
{
 CALL("remove(const Clause* cl)"); 
 ASSERT(_clauseNumIndex->find(cl->Number()) == cl);   

 INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REMOVE(cl);
 
 _removal.clause(cl->Number());
 TERM header;
 for(LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
  { 
   header = *(iter.currentLiteral());   
   if (header.arity())
    {
     if (header.IsSymLitHeader())
      {
       if (_removal.symLitHeader(header))  
	{     
         // removing first argument
         _termTraversal.reset(*(iter.currentLiteral()->Args()));
         ASSERT(((bool)(_termTraversal)));
         while (_termTraversal)       
	  if (_removal.symLitSymbol(_termTraversal.symbol())) { _termTraversal.next(); }
	  else { _removal.symLitAfter(); _termTraversal.after(); };
         
         // removing second argument
         _termTraversal.reset(*(iter.currentLiteral()->Args()->Next()));
         _removal.symLitSecondArg();
         ASSERT(((bool)(_termTraversal)));
         while (_termTraversal)       
	  if (_removal.symLitSymbol(_termTraversal.symbol())) { _termTraversal.next(); }
	  else { _removal.symLitAfter(); _termTraversal.after(); }; 
         
         // recycle
         _removal.endOfSymLit();  
        };    
      }
     else // ordinary literal
      if (_removal.litHeader(header))
       { 
        _termTraversal.reset(iter.currentElement()->literalRef());
        _termTraversal.next();
        ASSERT(((bool)(_termTraversal)));
        while (_termTraversal)
	 if (_removal.symbol(_termTraversal.symbol())) { _termTraversal.next(); }
	 else { _removal.after(); _termTraversal.after(); };
        // recycle
        _removal.endOfLit();  
       };
    } 
   else // prop. lit.
    _removal.propLit(header);
  };
}; // void BackwardSubsumption::remove(const Clause* cl)




//============================================================================


