//
// File:         ClauseBackground.h
// Description:  Ancestors of a clause.
// Created:      Feb 12, 2000, 20:10
// Revised:      Feb 1, 2001, AR, 
//               representation changed from arrays to lists
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk 
//==================================================================
#ifndef  ClauseBackground_H
//==================================================================
#define ClauseBackground_H
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "ClassDesc.h"
#include "ObjDesc.h"
#include "GlobAlloc.h"
#include "Stack.h"
#include "BitWord.h"
#include "GList.h"
class Clause;
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_BACKGROUND
 #define DEBUG_NAMESPACE "ClauseBackground"
#endif
#include "debugMacros.h"
//===================================================================
class ClauseBackground
 {
  public:
   enum Rule 
   { 
    Input = 0, 
    Prepro = 1, 
    BinRes = 2,
    Hyper = 3,    
    ForwSup = 4,
    BackSup = 5,
    EqRes = 6,
    EqFact = 7, 
    ForwDemod = 8,
    BackDemod = 9,
    EqResSimp = 10,
    ForwSubsRes = 11,
    BackSubsRes = 12,     
    Split = 13,
    Reanimation = 14,
    Name = 15,
    NegSel = 16,
    VIP = 31
   };
   typedef BitWord Rules; 
   typedef GList<GlobAlloc,Clause*,ClauseBackground> ClauseList;
  public:
   ClauseBackground(ClauseList* clauses) 
    : _rules(), 
      //number(num)
      _ancestors(clauses) 
    {
     CALL("constructor ClauseBackground(ClauseList* clauses)");
    };
   ~ClauseBackground()
   {
    CALL("desctructor ~ClauseBackground()");
    ClauseList::destroy(_ancestors);
   }; 
   const Rules& TheRules() const { CALL("TheRules() const"); ASSERT(checkObj()); return _rules; };
   void SetRules(const Rules& r) { CALL("SetRules(Rules r"); ASSERT(checkObj()); _rules = r; };
   void MakeVIP() { CALL("MakeVIP()"); ASSERT(checkObj()); _rules.set(VIP); };
   bool IsVIP() const { CALL("MakeVIP()"); ASSERT(checkObj()); return _rules.bit(VIP); };
   void MarkAsName() { CALL("MarkAsName()"); ASSERT(checkObj()); _rules.set(Name); };
   bool NegativeSelection() 
   { 
    CALL("NegativeSelection()"); 
    ASSERT(checkObj()); 
    return _rules.bit(NegSel); 
   };
   void SetNegativeSelection(bool fl) 
   {
    CALL("SetNegativeSelection(bool fl)");
    ASSERT(checkObj());
    if (fl) { _rules.set(NegSel); } else _rules.clear(NegSel);
   }; 
   ClauseList* ancestors() { return _ancestors; };
     
   void* operator new(size_t)
    {
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_classDesc.registerAllocated(sizeof(ClauseBackground)));
      return GlobAlloc::allocate(sizeof(ClauseBackground),&_classDesc);
     #else 
      return GlobAlloc::allocate(sizeof(ClauseBackground));
     #endif
    };
   void operator delete(void* obj)
    {
     #ifdef DEBUG_ALLOC_OBJ_TYPE 
      ALWAYS(_classDesc.registerDeallocated(sizeof(ClauseBackground)));
      GlobAlloc::deallocate(obj,sizeof(ClauseBackground),&_classDesc);
     #else
      GlobAlloc::deallocate(obj,sizeof(ClauseBackground));
     #endif
    };

   ClauseBackground* copy() const 
   {
    CALL("copy() const");
    ClauseList* orgClauses = _ancestors;
    ClauseList* targClauses = (ClauseList*)0;
    ClauseList** targClausesAddr = &targClauses;      
    while (orgClauses)
    {
     *targClausesAddr = new ClauseList(orgClauses->hd(),(ClauseList*)0);
     targClausesAddr = &((*targClausesAddr)->tl());
     orgClauses = orgClauses->tl();
    };

    ClauseBackground* res = new ClauseBackground(targClauses);
    res->_rules = _rules;
    return res;
   };
   
   static ostream& outputRules(ostream& str,const BitWord& _rules) 
   {
    if (_rules.bit(ClauseBackground::Input)) str << "in ";
    if (_rules.bit(ClauseBackground::Prepro)) str << "pp ";
    if (_rules.bit(ClauseBackground::BinRes)) str << "br ";
    if (_rules.bit(ClauseBackground::Hyper)) str << "hr ";
    if (_rules.bit(ClauseBackground::ForwSup)) str << "fs ";
    if (_rules.bit(ClauseBackground::BackSup)) str << "bs ";
    if (_rules.bit(ClauseBackground::EqRes)) str << "er ";
    if (_rules.bit(ClauseBackground::EqFact)) str << "ef ";
    if (_rules.bit(ClauseBackground::ForwDemod)) str << "fd ";
    if (_rules.bit(ClauseBackground::BackDemod)) str << "bd ";
    if (_rules.bit(ClauseBackground::EqResSimp)) str << "ers ";
    if (_rules.bit(ClauseBackground::ForwSubsRes)) str << "fsr ";
    if (_rules.bit(ClauseBackground::BackSubsRes)) str << "bsr ";
    if (_rules.bit(ClauseBackground::Split)) str << "sp ";
    if (_rules.bit(ClauseBackground::Reanimation)) str << "rea ";
    if (_rules.bit(ClauseBackground::Name)) str << "nm ";
    if (_rules.bit(ClauseBackground::NegSel)) str << "ns ";
    return str;
   };    
   
  private:
   #ifndef NO_DEBUG
    bool checkObj() const
    {
     #ifdef DEBUG_ALLOC
      ObjDesc* objDesc = GlobAlloc::hostObj((void*)this);
      if (!objDesc) 
       {
        DMSG("ClauseBackground::checkObj()> Null pointer to object descriptor.\n");  
        return false;
       };   
      bool res = true;
      #ifdef DEBUG_ALLOC_OBJ_SIZE
       if (objDesc->size !=  sizeof(ClauseBackground)) 
        {
         DMSG("ClauseBackground::checkObj()> objDesc->size !=  sizeof(ClauseBackground)\n");
         res = false;
        };
      #endif
      #ifdef DEBUG_ALLOC_OBJ_STATUS
       if (objDesc->status != ObjDesc::Persistent) 
        { 
         DMSG("ClauseBackground::checkObj()> objDesc->status != ObjDesc::Persistent\n");
         res = false;
        };
      #endif
      #ifdef DEBUG_ALLOC_OBJ_TYPE
       if (objDesc->classDesc != &_classDesc) 
        {
         DMSG("ClauseBackground::checkObj()> objDesc->classDesc != &_classDesc\n"); 
         res = false;
        };
      #endif  
       if (!res) DOP(DMSG("ClauseBackground::checkObj()> host object ") << *objDesc << "\n"); 
       return res;
     #endif
     return true; 
    };
   #endif
  private:
   BitWord _rules;
   ClauseList* _ancestors; 
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    static ClassDesc _classDesc;
   #endif
  friend class OpenClauseBackground;
 }; // class ClauseBackground
//========================================
#undef DEBUG_NAMESPACE
//========================================
#ifdef DEBUG_CLAUSE_BACKGROUND
 #define DEBUG_NAMESPACE "OpenClauseBackground"
#else 
 #undef DEBUG_NAMESPACE
#endif
#include "debugMacros.h"
//========================================
class OpenClauseBackground
 {
  private:
   ClauseBackground::Rules rules;
   Stack<Clause*,HellConst::MaxNumOfClauseAncestors> ancestors;
  public:
   OpenClauseBackground() 
    : rules(),
      ancestors("OpenClauseBackground::ancestors")
     {};
   void Reset() 
    { 
     rules.clear();
     ancestors.reset(); 
    };
   void MakeVIP() { rules.set(ClauseBackground::VIP); };
   void UsedRule(ClauseBackground::Rule r) { rules.set(r); };
   void CancelRule(ClauseBackground::Rule r) { rules.clear(r); };
   void UsedRules(const ClauseBackground::Rules& rs) { rules |= rs; };
   void Push(Clause* cl) { ancestors.push(cl); };
   void InverseOrder() 
   {
    Clause** b = ancestors.begin();
    Clause** e = ancestors.last();    
    while (b < e) 
     {
      Clause* tmp = *b;
      *b = *e;
      *e = tmp;
      b++;
      e--;
     };
   };
   ClauseBackground* NewAncestors() 
    {
     ClauseBackground::ClauseList* clauses = (ClauseBackground::ClauseList*)0;

     for (Clause** anc = ancestors.begin(); anc < ancestors.end(); anc++)
      {
       clauses = new ClauseBackground::ClauseList(*anc,clauses);
      };
     
     ClauseBackground* res = new ClauseBackground(clauses);
     res->SetRules(rules); 
     return res;
    };  
   const BitWord& Rules() const { return rules; };
   ulong numberOfAncestors() const { return ancestors.size(); }; 
   Clause* const * begin() const { return ancestors.begin(); };
   Clause* const * end() const { return ancestors.end(); };
   Clause* const * last() const { return ancestors.last(); };
   bool full() const { return ancestors.full(); };
   bool roomFor(ulong n) { return ancestors.roomFor(n); };
   bool singletone() const { return ancestors.size() == 1; }; 

 }; // OpenClauseBackground
//=================================================================
#undef DEBUG_NAMESPACE   
//==========================================================================
#endif
