#ifndef COMPONENT_NAME_H
//===========================================================
#define COMPONENT_NAME_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "fs_code.h"
#include "fs_command.h"
#include "Stack.h"
#include "tmp_lit_list.h"
#include "Signature.h"
#include "GlobAlloc.h"
#include "MultisetOfVariables.h"
extern Signature TheSignature;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
 #define DEBUG_NAMESPACE "CN_QUERY"
#endif 
#include "debugMacros.h"
//=================================================
class CN_QUERY
 {
  public:
   class LIT
    {    
     private:
      const Flatterm* arg1;
      const Flatterm* arg2;
      bool captured;
      LIT* next;
     public:  
      const Flatterm* Args() const { return arg1; };
      void SetArgs(const Flatterm* a) { arg1 = a; };      
      const Flatterm* Arg1() const { return arg1; }; // for equations
      const Flatterm* Arg2() const { return arg2; }; // for equations
      void SetArgs(const Flatterm* a1,const Flatterm* a2)
       { arg1 = a1; arg2 = a2; };
      LIT* Next() const { return next; };
      void SetNext(LIT* nl) { next = nl; };
      void Set(const Flatterm* a,LIT* nl) { arg1 = a; next = nl; };
      void Set(const Flatterm* a1,const Flatterm* a2,LIT* nl) 
       { arg1 = a1; arg2 = a2; next = nl; };
      void Capture() { captured = true; };
      void Release() { captured = false; };
      bool Captured() const { return captured; };
     public:
      //#ifndef NO_DEBUG_VIS
       ostream& output(ostream& str) const { return (Args() - 1)->outputTerm(str); };
       //#endif
    }; // class LIT
  private: // structure 
   LIT literal[HellConst::MaxNumOfLiterals];
   LIT* next_literal;
   LIT* header_index[HellConst::MaxNumOfDifferentHeaders];
   LIT** const pos_ord_equations;
   LIT** const neg_ord_equations;
   LIT** present_header[HellConst::MaxNumOfLiterals];
   LIT*** next_present_header; 
   LIT* last_pos_unord_eq;
   LIT* last_neg_unord_eq;
   long num_of_pos_eq;
   bool collect_signature;
   FunSet* signature;
   ulong numOfDiffVars;
   ulong numOfLits;
  public:
   CN_QUERY() 
    : pos_ord_equations(header_index + TERM::positiveOrdEqHeader().HeaderNum()),
      neg_ord_equations(header_index + TERM::negativeOrdEqHeader().HeaderNum()),
      collect_signature(true) // default
    {
     CALL("constructor CN_QUERY()");
     for (ulong h = 0; h < HellConst::MaxNumOfDifferentHeaders; h++)
      {
       header_index[h] = (LIT*)0;
      };
     next_present_header = present_header;
     signature = (FunSet*)0; 
    };

   ~CN_QUERY() { CALL("destructor ~CN_QUERY()"); };
   void ResetQuery()
    {
     next_literal = literal;
     while (next_present_header != present_header)
      {
       next_present_header--;
       **next_present_header = (LIT*)0;
      };
     last_pos_unord_eq = (LIT*)0;
     last_neg_unord_eq = (LIT*)0;
     num_of_pos_eq = 0;
     if (collect_signature) 
      {
       if (signature) delete signature;
       signature = new FunSet();
      };
     numOfLits = 0;
    };
   void SetCollectSignature(bool fl) 
    { 
     collect_signature = fl; 
     //collect_signature = true; // debug
    };
   void LinkEquations()
    {
     if (last_pos_unord_eq) last_pos_unord_eq->SetNext(*pos_ord_equations);
     if (last_neg_unord_eq) last_neg_unord_eq->SetNext(*neg_ord_equations);
    }; 
   void EndOfQuery()
    {
     LinkEquations();
    }; 
   void PushLit(const Flatterm* lit)
    {
     CALL("PushLit(const Flatterm* lit)");
     ASSERT(!lit->Symbol().IsEquality());
     
     LIT** lit_lst_ptr = header_index + lit->Symbol().HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(lit + 1,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(lit + 1,(LIT*)0);
      };
     *lit_lst_ptr = next_literal; 
     next_literal->Release();
     next_literal++;  
     if (collect_signature) CollectFunctors(lit);
    };   
    
   void PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
    {
     CALL("PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
     ASSERT(header.IsUnorderedEq());
     
     if (header.Positive()) num_of_pos_eq++;
     LIT** lit_lst_ptr = header_index + header.HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(arg1,arg2,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(arg1,arg2,(LIT*)0);
       if (header.Negative()) { last_neg_unord_eq = next_literal; }
       else // positive
        last_pos_unord_eq = next_literal;
      };
     *lit_lst_ptr = next_literal; 
     next_literal->Release();
     next_literal++; 
     if (collect_signature) 
      {
       AddFunctor(header.HeaderNum());
       CollectFunctors(arg1); 
       CollectFunctors(arg2);
      };
    };
    
   void PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
    {
     CALL("PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
     ASSERT(header.IsEquality() && header.IsOrderedEq());
         
     if (header.Positive()) num_of_pos_eq++;
     LIT** lit_lst_ptr = header_index + header.HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(arg1,arg2,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(arg1,arg2,(LIT*)0);
      };
     *lit_lst_ptr = next_literal;
     next_literal->Release(); 
     next_literal++; 
     if (collect_signature) 
      {
       header.MakeUnordered();
       AddFunctor(header.HeaderNum());
       CollectFunctors(arg1); 
       CollectFunctors(arg2);
      };
    }; 
    
   LIT* FirstLit(ulong header_num)  
    {
     return header_index[header_num];
    }; 
   
   bool SubsumptionAllowed(Clause* subsuming_cl)
    {
     return subsuming_cl->NumOfPosEq() <= num_of_pos_eq;
    }; 

   const FunSet* Signature() const { return signature; }; 

   ulong NumOfDiffVars() const { return numOfDiffVars; };
   ulong NumOfLits() const { return numOfLits; }
   void Load(TMP_LIT_LIST& component,ulong numOfVars)
    {
     CALL("Load(TMP_LIT_LIST& component,ulong numOfVars)");
     ResetQuery(); 
     numOfDiffVars = numOfVars;
     numOfLits = component.Length();
     TMP_LITERAL* lit;
     for (ulong ln = 0; ln < component.Length(); ln++)
      { 
       lit = component.Nth(ln);
       if (lit->IsEquality())
	{
	 if (lit->IsOrderedEq())   
	  {
           PushOrdEq(lit->Header(),lit->Arg1(),lit->Arg2());
          }      
         else PushUnordEq(lit->Header(),lit->Arg1(),lit->Arg2());
        } 
       else PushLit(lit->FlatTerm());
      };
    };

  private:
   void AddFunctor(ulong fun_num)
    {
     if (!(signature->write((size_t)fun_num)))
      {
       FunSet* new_sig = signature->add((size_t)fun_num);
       delete signature;
       signature = new_sig; 
      };
    };
   void CollectFunctors(const Flatterm* t)
    {
     CALL("CollectFunctors(const Flatterm* t)");
     ASSERT(collect_signature);
     Flatterm::Iterator iter(t);
     TERM sym;
     while (iter.NotEndOfTerm())
      {
       sym = iter.CurrPos()->symbolRef();
       if (sym.isComplex()) AddFunctor(sym.HeaderNum());       
       iter.NextSym();
      };
    };
    
  public:
   //#ifndef NO_DEBUG_VIS 
    ostream& output(ostream& str) const
    {
     for (const LIT* const * const * p = present_header; p < next_present_header; p++)
      for (const LIT* lit = **p; lit; lit = lit->Next())
       lit->output(str) << '\n';
     return str;
    };
    //#endif
 }; // class CN_QUERY
 
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
 #define DEBUG_NAMESPACE "ComponentName"
#endif 
#include "debugMacros.h"
//=================================================
class ComponentName 
{
 public:
  ComponentName() : variantIndex(subst), query(), dummyTmpLiteral(), dummyComponent() 
   {
    variantIndex.SetOldStyle(false);
    variantIndex.SetOldStyle(true); // must be so, because the components are integrated in the old style
    variantIndex.SetUseSigFilters(false);
    query.SetCollectSignature(false); 
    dummyComponent.Push(&dummyTmpLiteral);
   };
  ~ComponentName() { CALL("destructor ~ComponentName()"); };
  void destroyIndex() 
   {
    variantIndex.Destroy(destroyName);
   };
  
  bool findName(TMP_LIT_LIST& component,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& name)
  {
   CALL("findName(TMP_LIT_LIST& component,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& name)");

   //cout << "vars.NumOfDifferent() = " << vars.NumOfDifferent() << "\n"; // debug

   query.Load(component,vars.NumOfDifferent());

   //query.output(cout << "QUERY ") << " VARS " << query.NumOfDiffVars() << "\n"; // debug  

   bool res = findVariant(name);

   //cout << "RES = " << res << "\n"; // debug

   return res;
  };

  bool findName(const Flatterm* term,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& name)
  {
   CALL("findName(const Flatterm* term,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& name)");
   makeDummyLit(term);
   return findName(dummyComponent,vars,name); 
  };
  

  void integrate(TMP_LIT_LIST& component,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong name)
  {

    //DF; component.output(cout << "INTEGRATE ") << "\n";  

   ulong* namep = (ulong*)(GlobAlloc::allocate(sizeof(ulong)));
   *namep = name;
   variantIndex.Integrate(component,vars,(void*)namep);
  };

  void integrate(const Flatterm* term,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong name)
  {
   CALL("integrate(const Flatterm* term,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong name)");
   makeDummyLit(term);
   integrate(dummyComponent,vars,name);
  };

  bool nameFor(TMP_LIT_LIST& component,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& name,bool& newName)
  {
   if (findName(component,vars.NumOfDifferent(),name))
    {
     newName = false;
     return true;
    }; 
   // new name  
   name = TheSignature.newSkolemPredicateNumber(0UL); 
   if (name >= HellConst::MaxNumOfFunctors) return false;
   newName = true;
   ulong* namep = (ulong*)(GlobAlloc::allocate(sizeof(ulong)));
   *namep = name; 
   integrate(component,vars,namep);
   return true; 
  }; 

  bool nameFor(const Flatterm* term,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& name,bool& newName)
  {
   CALL("nameFor(const Flatterm* term,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& name,bool& newName)");   
   makeDummyLit(term);
   if (findName(dummyComponent,vars.NumOfDifferent(),name))
    {
     newName = false;
     return true;
    }; 
   // new name  
   name = TheSignature.newSkolemPredicateNumber(1UL); 
   if (name >= HellConst::MaxNumOfFunctors) return false;
   newName = true;
   ulong* namep = (ulong*)(GlobAlloc::allocate(sizeof(ulong)));
   *namep = name; 
   integrate(dummyComponent,vars,namep);
   return true;    
  };

  #ifndef NO_DEBUG_VIS 
   ostream& outputQuery(ostream& str) { return query.output(str); };
  #endif

 private: 
  void integrate(TMP_LIT_LIST& component,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong* name)
  {
   variantIndex.Integrate(component,vars,(void*)name);
  };
  bool findName(TMP_LIT_LIST& component,ulong numOfVars,ulong& name)
  {
   query.Load(component,numOfVars);
   return  findVariant(name);
  };
   
  bool findVariant(ulong& name);

  static void destroyName(void* name)
  {
   GlobAlloc::deallocate(name,sizeof(ulong));
  };
  
 
  void makeDummyLit(const Flatterm* term)
  {
   CALL("makeDummyLit(const Flatterm* term)");
   dummyTmpLiteral.Reset();
   dummyTmpLiteral.PushNonpropHeader(::TERM::dummyPositiveHeader1());
   dummyTmpLiteral.PushTerm(term);
  };
 private:
  const Flatterm* subst[HellConst::MaxClauseSize]; 
  FS_CODE variantIndex;
  CN_QUERY query;
  TMP_LITERAL dummyTmpLiteral;
  TMP_LIT_LIST dummyComponent;
}; // class ComponentName 
//================================================
#undef DEBUG_NAMESPACE
//===============================================================
#endif
