//
// File:         ForwardDemodulation.cpp
// Description:  
// Created:      ???
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "ForwardDemodulation.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
 #define DEBUG_NAMESPACE "ForwardDemodulation"
#endif 
#include "debugMacros.h"
//=================================================

#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"
 ClassDesc ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>::_classDesc("ForwardMatchingIndexInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>");
 
 ClassDesc ForwardDemodulation::Index::StoredSubstitution::_classDesc("ForwardDemodulation::Index::StoredSubstitution",
                                                                      ForwardDemodulation::Index::StoredSubstitution::minSize(),
                                                                      ForwardDemodulation::Index::StoredSubstitution::maxSize());
 ClassDesc ForwardDemodulation::Index::ConstrRuleList::_classDesc("ForwardDemodulation::Index::ConstrRuleList",
                                                                   ForwardDemodulation::Index::ConstrRuleList::minSize(),
                                                                   ForwardDemodulation::Index::ConstrRuleList::maxSize());

#endif


void ForwardDemodulation::Integrate(Clause* cl) 
{ 
 if (CanBeRewritingRule(cl)) index.Integrate(cl); 
}; 

void ForwardDemodulation::Remove(Clause* cl) 
{ 
 if (CanBeRewritingRule(cl)) index.Remove(cl); 
};

bool ForwardDemodulation::RewriteByPreordered(const Flatterm* query,const SetOfSplittingBranches& splittingBranches)
 {
  CALL("RewriteByPreordered(const Flatterm* query,const SetOfSplittingBranches& splittingBranches)");
  ASSERT(query->Symbol().isComplex());

  ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* leaf = index.Match(query);
  check_leaf:
   if (leaf)
    {
     rule = leaf->leafInfo().clause();
     if (rule->Active())
      {
       
       if (CheckSplittingBranches(rule,splittingBranches))
        {
         TERM rule_rhs_term = *(rule->LitList()->firstLit()->NthArg(1));
         _substitutionCache.load(leaf->leafInfo().substitution());
         rule_rhs.Reset(rule_rhs_term,&_substitutionCache);    
         return true;
        };
      };
     
     leaf = index.MatchAgain();
     goto check_leaf;
    };
  return false;      

 }; // bool ForwardDemodulation::RewriteByPreordered(const Flatterm* query,const SetOfSplittingBranches& splittingBranches) 

bool ForwardDemodulation::RewriteByConstrained(const Flatterm* query,const SetOfSplittingBranches& splittingBranches)
 {
  CALL("RewriteByConstrained(const Flatterm* query,const SetOfSplittingBranches& splittingBranches)");

  ASSERT(query->Symbol().isComplex());


  for (ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* leaf = index.Match(query); leaf; leaf = index.MatchAgain())
   {
    ASSERT(leaf->isLeaf());  
    Index::ConstrRuleList* crl =  leaf->leafInfo().rules(); 
    ASSERT(crl);  
    do
     {
      rule = crl->hd().clause;
      if (rule->Active())
       {
	// check the constraint
        SharedKBConstr* constr = crl->hd().constr;
        //Index::StoredSubstitution* subst = crl->hd().subst;
        _substitutionCache.load(crl->hd().subst);

        if (!constr) // preordered 
	 {
	  ASSERT(rule->PreorderedPositiveUnitEquality() || fDOnSplittingBranchesFlag);
          ASSERT(crl->hd().lhs == rule->LitList()->firstLit()->Arg1()->First()); 

          if (CheckSplittingBranches(rule,splittingBranches))
           {
	    TERM rule_rhs_term = *(rule->LitList()->firstLit()->Arg2());

	    //DF; cout << "QUERY " << query << "\n";	  
	    //DF; cout << "RULE " << rule << "\n";
	    //DF; cout << "LHS " << crl->hd().lhs << "\n";
	    //DF; cout << "RHS " << rule_rhs_term << "\n";  

            ASSERT(Index::TermWithSubst::KBGreater(crl->hd().lhs,rule_rhs_term,&_substitutionCache));
            rule_rhs.Reset(rule_rhs_term,&_substitutionCache);
            return true;
           };
	 }
        else
	 {
          // constrained rule, check the constraint and splitting branches
	  if ((checkConstraint(constr,&_substitutionCache)) && (CheckSplittingBranches(rule,splittingBranches)))
	   {
	    const TERM* lhs = crl->hd().lhs;
            ASSERT(lhs);
            ASSERT(lhs->isComplex());
	    TERM rule_rhs_term;
            if (rule->LitList()->firstLit()->Arg1()->First() == lhs)
	     {
	      rule_rhs_term = *(rule->LitList()->firstLit()->Arg2());
             }          
            else
	     {
	      ASSERT(rule->LitList()->firstLit()->Arg2()->First() == lhs);
              rule_rhs_term = *(rule->LitList()->firstLit()->Arg1());
              ASSERT(rule_rhs_term.First()->isComplex());
             };

            ASSERT(Index::TermWithSubst::KBGreater(lhs,rule_rhs_term,&_substitutionCache));
            rule_rhs.Reset(rule_rhs_term,&_substitutionCache);   
            return true;      
           };              
	 };
       };
      crl = crl->tl();
     }
    while (crl);
   };
  return false;      
 }; // bool ForwardDemodulation::RewriteByConstrained(const Flatterm* query,const SetOfSplittingBranches& splittingBranches) 

ForwardDemodulation::Index::ConstrRuleList* ForwardDemodulation::removeConstrRule(Clause* cl,TERM* t,Index::ConstrRuleList* lst)
{
 CALL("removeConstrRule(Clause* cl,TERM* t,Index::ConstrRuleList* lst)");
 if (!lst) return (Index::ConstrRuleList*)0;
 if ((lst->hd().clause == cl) && (lst->hd().lhs == t))
  {
   Index::ConstrRuleList* res = lst->tl();
   if (lst->hd().constr) delete lst->hd().constr; 
   
   //delete lst->hd().subst;
   Index::StoredSubstitution::destroy(lst->hd().subst);

   delete lst;
   return res;
  };
 lst->tl() = removeConstrRule(cl,t,lst->tl());
 return lst;
}; // ForwardDemodulation::Index::ConstrRuleList* ForwardDemodulation::removeConstrRule(Clause* cl,TERM* t,Index::ConstrRuleList* lst)

ForwardDemodulation::Index::ConstrRuleList* ForwardDemodulation::insertConstrRule(Clause* cl,SharedKBConstr* constr,Index::StoredSubstitution* subst,TERM* t,Index::ConstrRuleList* lst)
{
 CALL("insertConstrRule(Clause* cl,SharedKBConstr* constr,Index::StoredSubstitution* subst,TERM* t,Index::ConstrRuleList* lst)");
 return new Index::ConstrRuleList(Index::ConstrRule(cl,constr,subst,t),lst); 
}; // ForwardDemodulation::Index::ConstrRuleList* ForwardDemodulation::insertConstrRule(Clause* cl,SharedKBConstr* constr,Index::StoredSubstitution* subst,TERM* t,Index::ConstrRuleList* lst)



bool ForwardDemodulation::checkConstraint(SharedKBConstr* constr,Index::SubstitutionCache* subst)
{
 CALL("checkConstraint(SharedKBConstr* constr,Index::SubstitutionCache* subst)");
 ASSERT(constr);
 switch (constr->category())
  {
   case KBConstraint::NearlyTautology: return checkNearlyTautology(constr,subst);            
   case KBConstraint::NearlyContradiction: return checkNearlyContradiction(constr,subst);
   case KBConstraint::Volatile: return checkVolatile(constr,subst);
   case KBConstraint::PurelyLexicographic: return checkPurelyLexicographic(constr,subst);
  };
 #ifdef DEBUG_NAMESPACE
  ICP("ICP0"); return false;
 #elif defined _SUPPRESS_WARNINGS_
   return false;
 #endif
}; // bool ForwardDemodulation::checkConstraint(SharedKBConstr* constr,SubstitutionCache* subst)


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
 #define DEBUG_NAMESPACE "ForwardDemodulation::Index"
#endif 
#include "debugMacros.h"
//=================================================

void ForwardDemodulation::Index::IntegratePreordered(Clause* cl)
 {
  CALL("IntegratePreordered(Clause* cl)");
  #ifdef DEBUG_NAMESPACE
   if (cl->Length() > 1)
    {
     LiteralList::Iterator iter(cl->LitList());
     iter.next(); 
     do
      {
       if (!iter.currentLiteral()->IsSplitting()) 
	{
	 ERROR("E1");
        };
       iter.next();
      }
     while (iter.notEnd());
    };  
  #endif          
  
  ASSERT(cl->LitList()->numOfSelectedLits() == 1); 

  TERM* lit = cl->LitList()->firstLit();

  ASSERT(lit->Positive());
  ASSERT(lit->IsEquality());

  if (lit->IsOrderedEq())
   {          
    ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* leaf = Integrate(lit->Args()->First());
    if (!leaf->leafInfo().clause())
     {
      leaf->leafInfo().clause() = cl;

        
      ulong num_of_vars = m_index.integrationGetVariables().NumOfDifferent();

      //StoredSubstitution* subst = new(num_of_vars) StoredSubstitution(num_of_vars);
      _openSubstitution.reset(num_of_vars);

      m_index.integrationGetVariables().ResetIterator();
      for (const ulong* v = m_index.integrationGetVariables().NextVar(); v; v = m_index.integrationGetVariables().NextVar())
       {
	 //*(subst->Subst(*v)) = m_index.integrationMapVariable(*v);
      
        _openSubstitution.map(*v,m_index.integrationMapVariable(*v));
       };
      //leaf->leafInfo().substitution() = subst;
      leaf->leafInfo().substitution() = _openSubstitution.getStoredSubstitution();

      INTERCEPT_FORWARD_MATCHING_INDEXING_INTEGRATE(lit->Args()->First());            
     }
    else // some clause is already in this leaf, replace it
     {	 
      ASSERT(!leaf->leafInfo().clause()->Active());
      DOP(TERM* debugTerm1 = leaf->leafInfo().clause()->LitList()->firstLit()->Arg1()->First());
      DOP(TERM* debugTerm2 = cl->LitList()->firstLit()->Arg1()->First());
      ASSERT(debugTerm1 == debugTerm2);
      INTERCEPT_FORWARD_MATCHING_INDEXING_REMOVE(leaf->leafInfo().clause()->LitList()->firstLit()->Arg1()->First());   
      leaf->leafInfo().clause() = cl;
      // the variable renaming must be OK since we are talking about
      // the first argument of the first literal    
      INTERCEPT_FORWARD_MATCHING_INDEXING_INTEGRATE(lit->Args()->First());
     };
   }; 
 }; // void ForwardDemodulation::Index::IntegratePreordered(Clause* cl) 

void ForwardDemodulation::Index::RemovePreordered(Clause* cl) 
 { 
  CALL("RemovePreordered(Clause* cl)"); 

  #ifdef DEBUG_NAMESPACE
   if (cl->Length() > 1)
    {
     LiteralList::Iterator iter(cl->LitList());
     iter.next(); 
     do
      {
       if (!iter.currentLiteral()->IsSplitting()) 
	{
	 ERROR("E1");
        };
       iter.next();
      }
     while (iter.notEnd());
    };  
  #endif          
  
  ASSERT(cl->LitList()->numOfSelectedLits() == 1); 

  TERM* lit = cl->LitList()->firstLit();

  ASSERT(lit->Positive());
  ASSERT(lit->IsEquality());
 
  if (lit->IsOrderedEq())
   {
    ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* leaf;
    ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>** branch_to_destroy;
    if (Remove(lit->Args()->First(),leaf,branch_to_destroy))
     { 
      ASSERT(leaf->isLeaf());

      ASSERT(leaf->leafInfo().clause()->LitList()->firstLit()->Args()->First() == lit->Args()->First());

      if (leaf->leafInfo().clause() == cl)
       {
	//delete leaf->leafInfo().substitution(); 
	StoredSubstitution::destroy(leaf->leafInfo().substitution());
 
        ForwardMatchingIndexTree<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>::DestroyBranch(branch_to_destroy);
        INTERCEPT_FORWARD_MATCHING_INDEXING_REMOVE(lit->Args()->First()); 
       };
     }
    else
     {
      ERROR("LHS of a removed rewriting rule is not found in the FD index."); 
     };
   };
 }; // void ForwardDemodulation::Index::RemovePreordered(Clause* cl) 


void ForwardDemodulation::Index::IntegrateConstrained(Clause* cl)
 {
  CALL("IntegrateConstrained(Clause* cl)");

  #ifdef DEBUG_NAMESPACE
   if (cl->Length() > 1)
    {
     LiteralList::Iterator iter(cl->LitList());
     iter.next(); 
     do
      {
       if (!iter.currentLiteral()->IsSplitting()) 
	{
	 ERROR("E1");
        };
       iter.next();
      }
     while (iter.notEnd());
    };  
  #endif          
  
  ASSERT(cl->LitList()->numOfSelectedLits() == 1); 

  TERM* lit = cl->LitList()->firstLit();

  ASSERT(lit->Positive());
  ASSERT(lit->IsEquality());


  if (lit->IsOrderedEq())
   {
    ASSERT(lit->Args()->IsReference());
    TERM* lhs = lit->Args()->First();
    ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* leaf = Integrate(lhs); 
    INTERCEPT_FORWARD_MATCHING_INDEXING_INTEGRATE(lhs);    
    ulong num_of_vars = m_index.integrationGetVariables().NumOfDifferent();

    //StoredSubstitution* subst = new(num_of_vars) StoredSubstitution(num_of_vars); 
    _openSubstitution.reset(num_of_vars);

    m_index.integrationGetVariables().ResetIterator();
    for (const ulong* v = m_index.integrationGetVariables().NextVar(); v; v = m_index.integrationGetVariables().NextVar())
     {
      //*(subst->Subst(*v)) = m_index.integrationMapVariable(*v);
      _openSubstitution.map(*v,m_index.integrationMapVariable(*v));
     };

    Index::ConstrRuleList* crl = leaf->leafInfo().rules();   
    //crl = new Index::ConstrRuleList(ConstrRule(cl,(SharedKBConstr*)0,subst,lhs),crl); 
    crl = new Index::ConstrRuleList(ConstrRule(cl,(SharedKBConstr*)0,_openSubstitution.getStoredSubstitution(),lhs),crl); 
    leaf->leafInfo().rules() = crl;
   }      
  else
   {
    TERM* arg1ref = lit->Args();
    TERM* arg2ref = arg1ref->Next();
    if ((arg1ref->isVariable()) || (arg2ref->isVariable())) return;        
    // both complex, collect variables and functors
    TERM* arg1 = arg1ref->First();
    TERM* arg2 = arg2ref->First();
    ulong funcs1;
    ulong funcs2;   
    collectVarsAndFuncs(arg1,_vars1,funcs1);     
    collectVarsAndFuncs(arg2,_vars2,funcs2);     
    IntegrateConstrained(cl,arg1,_vars1,funcs1,arg2,_vars2,funcs2);
    IntegrateConstrained(cl,arg2,_vars2,funcs2,arg1,_vars1,funcs1);
   };
 }; // void ForwardDemodulation::Index::IntegrateConstrained(Clause* cl)

void ForwardDemodulation::Index::IntegrateConstrained(Clause* cl,
                                                      TERM* t1,
                                                      MultisetOfVariables<HellConst::MaxNumOfVariables>& vms1,
                                                      ulong f1,
                                                      TERM* t2,
                                                      MultisetOfVariables<HellConst::MaxNumOfVariables>& vms2,
                                                      ulong f2)
{
 CALL("IntegrateConstrained(Clause*,TERM*,MultisetOfVariables<HellConst::MaxNumOfVariables>&,ulong f1,...)");
 if (!vms2.SubsetOf(vms1)) return; 
 SharedKBConstr::assumeGreater(t1,vms1,f1,t2,vms2,f2);
 if (!appropriateDegree(SharedKBConstr::collectedCategory())) return;

 SharedKBConstr* constr = SharedKBConstr::getShared();

 ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* leaf = Integrate(t1);
 INTERCEPT_FORWARD_MATCHING_INDEXING_INTEGRATE(t1);

 // collecting substitution
 ulong num_of_vars = m_index.integrationGetVariables().NumOfDifferent();

 //StoredSubstitution* subst = new(num_of_vars) StoredSubstitution(num_of_vars); 
 _openSubstitution.reset(num_of_vars);

 m_index.integrationGetVariables().ResetIterator();
 for (const ulong* v = m_index.integrationGetVariables().NextVar(); v; v = m_index.integrationGetVariables().NextVar())
  {
   //*(subst->Subst(*v)) = m_index.integrationMapVariable(*v);
   _openSubstitution.map(*v,m_index.integrationMapVariable(*v));
  };

 Index::ConstrRuleList* crl = leaf->leafInfo().rules();  
 
 //crl = insertConstrRule(cl,constr,subst,t1,crl);
 crl = insertConstrRule(cl,constr,_openSubstitution.getStoredSubstitution(),t1,crl);

 leaf->leafInfo().rules() = crl;
 
}; // void ForwardDemodulation::Index::IntegrateConstrained(Clause* cl,TERM* t1,...
 

void ForwardDemodulation::Index::RemoveConstrained(Clause* cl) 
 { 
  CALL("RemoveConstrained(Clause* cl)"); 
  
  #ifdef DEBUG_NAMESPACE
   if (cl->Length() > 1)
    {
     LiteralList::Iterator iter(cl->LitList());
     iter.next(); 
     do
      {
       if (!iter.currentLiteral()->IsSplitting()) 
	{
	 ERROR("E1");
        };
       iter.next();
      }
     while (iter.notEnd());
    };  
  #endif          
  
  ASSERT(cl->LitList()->numOfSelectedLits() == 1); 

  TERM* lit = cl->LitList()->firstLit();

  ASSERT(lit->Positive());
  ASSERT(lit->IsEquality());

  if (lit->IsOrderedEq())
   { 
    RemoveConstrained(cl,lit->Args()->First());
   }
  else
   {       
    TERM* arg1ref = lit->Arg1();
    TERM* arg2ref = lit->Arg2();
    if (arg1ref->isVariable() || arg2ref->isVariable()) return;
    TERM* arg1 = arg1ref->First();
    TERM* arg2 = arg2ref->First();
    RemoveConstrained(cl,arg1);
    RemoveConstrained(cl,arg2);    
   };  
 }; // void ForwardDemodulation::Index::RemoveConstrained(Clause* cl)
 
void ForwardDemodulation::Index::RemoveConstrained(Clause* cl,TERM* lhs) 
 { 
  CALL("RemoveConstrained(Clause* cl,TERM* lhs)");
  ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* leaf;
  ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>** branch_to_destroy;
  if (Remove(lhs,leaf,branch_to_destroy))
   { 
    ASSERT(leaf->isLeaf());
    INTERCEPT_FORWARD_MATCHING_INDEXING_REMOVE(lhs);
    Index::ConstrRuleList* crl = leaf->leafInfo().rules();      
    ASSERT(crl);   
    crl = removeConstrRule(cl,lhs,crl);
    if (crl) { leaf->leafInfo().rules() = crl; }
    else // the branch must be destroyed
     {
      ForwardMatchingIndexTree<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>::DestroyBranch(branch_to_destroy);

      /**********
      ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>* fork = (*branch_to_destroy)->Fork();
      (*branch_to_destroy)->SetFork((ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>*)0);     
      ForwardMatchingIndexTree<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,Index::LeafInfo,ForwardDemodulationIndexId>::Destroy(*branch_to_destroy);
      *branch_to_destroy = fork;      
      ***********/
     };
   };
 }; // void ForwardDemodulation::Index::RemoveConstrained(Clause* cl,TERM* lhs)

//=================================================














