//
// File:         ForwardDemodulation.h
// Description:  Implements forward demodulation. 
// Created:      ???
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef FORWARD_DEMODULATION_H
//==================================================================
#define FORWARD_DEMODULATION_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif 
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellInterceptMacros.h"
#include "HellConst.h"
#include "GlobAlloc.h"
#include "GList.h"
#include "Exit.h"
#include "ForwardMatchingIndexTree.h"
#include "ForwardMatchingIndex.h"
#include "SharedKBConstr.h"
#include "variables.h"
#include "Multiset.h"
#include "term.h"
#include "Flatterm.h"
#include "clause.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
 #define DEBUG_NAMESPACE "ForwardDemodulation"
#endif 
#include "debugMacros.h"
//=================================================

class ForwardDemodulationIndexId // only to identify the instance of ForwardMatchingIndex<..>
{
}; 

class ForwardDemodulation
 {
  public:
   class Index 
    {
     public:
      //class StoredSubstitution;

      typedef GList<GlobAlloc,const Flatterm* const *,ForwardDemodulation::Index> StoredSubstitution;
      class OpenSubstitution
      {
       public:
        OpenSubstitution() { CALL("constructor Index::OpenSubstitution()");  };
        void reset(ulong numOfVars) { _numOfVars = numOfVars; };
        StoredSubstitution* getStoredSubstitution() const 
	{
         StoredSubstitution* res = (StoredSubstitution*)0;
	 for (ulong v = _numOfVars; v; v--) 
	  res = new StoredSubstitution(_substitution[v-1],res); 
         return res;         
        };
        void map(ulong var,const Flatterm* const * ptrVar) 
	{
	 _substitution[var] = ptrVar;
        };
       private:
        //StoredSubstitution* _subst;
        const Flatterm* const * _substitution[HellConst::MaxNumOfVariables];
        ulong _numOfVars;   
      };
      class SubstitutionCache
      {
       public:
        SubstitutionCache() { CALL("constructor Index::SubstitutionCache()"); };
        void load(const StoredSubstitution* s);
        const Flatterm* const * const * subst() const { return _subst; };
        const Flatterm* const * const * subst(ulong v) { return _subst + v; };

        #ifdef DEBUG_NAMESPACE
	 ulong numOfVars() const { return _debugNumOfVars; };
        #endif

       private:
        const Flatterm* const * _subst[HellConst::MaxNumOfVariables];
        #ifdef DEBUG_NAMESPACE
	  ulong _debugNumOfVars; 
        #endif
      };

      class ConstrRule
       {
        public:
         ConstrRule(Clause* cl,SharedKBConstr* con,StoredSubstitution* sub,TERM* t) 
          : clause(cl), 
            constr(con),
            subst(sub),
            lhs(t)
         {
	  CALL("constructor Index::ConstrRule(Clause* cl,SharedKBConstr* con,StoredSubstitution* sub,TERM* t)");
         };
         Clause* clause;
         SharedKBConstr* constr;
         StoredSubstitution* subst;  
         TERM* lhs;  
      };
   
      typedef GList<GlobAlloc,ConstrRule,ForwardDemodulation::Index> ConstrRuleList; 
      

      class TermWithSubst
       {
        public:
         TermWithSubst() : term_iter(),subst_iter(),dummy_term((TERM::Arity)0,(TERM::Functor)0,TERM::PositivePolarity) 
         {
	   CALL("constructor Index::TermWithSubst()");
         };
         TermWithSubst(const TERM& t,SubstitutionCache* s)
          : term_iter(),
            subst_iter(),
            dummy_term((TERM::Arity)0,(TERM::Functor)0,TERM::PositivePolarity)
          {
	   CALL("constructor Index::TermWithSubst(const TERM& t,SubstitutionCache* s)");
           Reset(t,s);
          };
          
         TermWithSubst(const TERM* t,SubstitutionCache* s) // for terms that are known to be complex
          : term_iter(),
            subst_iter(),
            dummy_term((TERM::Arity)0,(TERM::Functor)0,TERM::PositivePolarity)
          {
	   CALL("constructor Index::TermWithSubst(const TERM* t,SubstitutionCache* s)");
           Reset(t,s);
          };

	 void Reset(const TERM& t,SubstitutionCache* s);
	 void Reset(const TERM* t,SubstitutionCache* s);
	 bool Next();
	 const TERM& Symbol();
         #ifndef NO_DEBUG
          static void CollectWeight(const TERM* t,SubstitutionCache* s,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs);
          static void CollectWeight(const TERM& t,SubstitutionCache* s,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs);
          static bool LexGreater(const TERM* t1,const TERM& t2,SubstitutionCache* s);
          static bool KBGreater(const TERM* t1,const TERM& t2,SubstitutionCache* s);
         #endif
         #ifndef NO_DEBUG_VIS
          static ostream& output(ostream& str,const TERM* t,SubstitutionCache* s);
         #endif

        private:
         TERM::Iterator term_iter;
         Flatterm::Iterator subst_iter;
         const Flatterm* const * const * subst;
         #ifdef DEBUG_NAMESPACE
          ulong max_var_num;
         #endif
         bool in_term;
         TERM dummy_term;
           
       }; // class TermWithSubst
      
     public:
      class LeafInfo 
      {
       public:
        void init() { _clauseOrRules = (void*)0; _substitution = (StoredSubstitution*)0; }; 
        Clause*& clause() { return (Clause*&)(_clauseOrRules); };
        ConstrRuleList*& rules() { return (ConstrRuleList*&)(_clauseOrRules); };
        StoredSubstitution*& substitution() { return _substitution; };
       private:
        void* _clauseOrRules;
        StoredSubstitution* _substitution;        
      };
     private:  
      ForwardMatchingIndex<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,HellConst::MaxNumOfFunctors,GlobAlloc,TERM,Flatterm,LeafInfo,ForwardDemodulationIndexId> m_index;

      long degree;
      MultisetOfVariables<HellConst::MaxNumOfVariables> _vars1;
      MultisetOfVariables<HellConst::MaxNumOfVariables> _vars2;
      TERM::Iterator _termIter;
      OpenSubstitution _openSubstitution;
     public:
      Index() : m_index(), degree(0L), _vars1(), _vars2(), _termIter(), _openSubstitution()
       {
	CALL("constructor Index::Index()");
       };
      ~Index() { CALL("ForwardDemodulation::destructor ~Index"); if (!Exit::fast) Destroy(); }; 
      void SetDegree(long d) { degree = d; }; 
      void Integrate(Clause* cl)
      {
       if (degree == 1) { IntegratePreordered(cl); }
       else IntegrateConstrained(cl);
      };
      void Remove(Clause* cl)
      {
       if (degree == 1) { RemovePreordered(cl); }
       else RemoveConstrained(cl);
      };    
      ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,LeafInfo,ForwardDemodulationIndexId>* Match(const Flatterm* query) 
      { 
       return INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_RESULT(m_index.Match(INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_QUERY(query))); 
      }; 
      ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,LeafInfo,ForwardDemodulationIndexId>* MatchAgain() 
      { 
       return m_index.MatchAgain();    
      }; 
      void Destroy() 
      { 
       m_index.Destroy(); 
      };
      #ifndef NO_DEBUG
       bool clean() const { return  m_index.clean(); };
      #endif
      #ifndef NO_DEBUG_VIS       
       ostream&  output(ostream& str) const { return m_index.output(str); }; 
      #endif     
     private:
      ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,LeafInfo,ForwardDemodulationIndexId>* Integrate(TERM* t);
      bool Remove(TERM* t,ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,LeafInfo,ForwardDemodulationIndexId>*& leaf,ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,LeafInfo,ForwardDemodulationIndexId>**& branch_to_destroy);
      void IntegratePreordered(Clause* cl);
      void IntegrateConstrained(Clause* cl);
      void IntegrateConstrained(Clause* cl,
                                TERM* t1,
                                MultisetOfVariables<HellConst::MaxNumOfVariables>& vms1,
                                ulong f1,
                                TERM* t2,
                                MultisetOfVariables<HellConst::MaxNumOfVariables>& vms2,
                                ulong f2);
      void RemovePreordered(Clause* cl);
      void RemoveConstrained(Clause* cl);
      void RemoveConstrained(Clause* cl,TERM* lhs);
      bool appropriateDegree(const KBConstraint::Category& cat) const;
      void collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& funcs);
    }; // class Index


   static Index::ConstrRuleList* removeConstrRule(Clause* cl,TERM* t,Index::ConstrRuleList* lst);
   static Index::ConstrRuleList* insertConstrRule(Clause* cl,SharedKBConstr* constr,Index::StoredSubstitution* subst,TERM* t,Index::ConstrRuleList* lst);

   typedef Multiset<ulong,HellConst::MaxNumOfDifferentHeaders> SetOfSplittingBranches;
   
  public:
   ForwardDemodulation() : index(), rule_rhs(), fDOnSplittingBranchesFlag(false), _constraintInstance(), _substitutionCache() 
   {
    CALL("constructor ForwardDemodulation()");
   };
   ~ForwardDemodulation() { if (!Exit::fast) Destroy(); };
   void SetDegree(long d) { degree = d; index.SetDegree(d); };
   void SetFDOnSplittingBranchesFlag(bool fl) 
   {
    if (fDOnSplittingBranchesFlag && (!fl)) RuntimeError::report("Switching fDOnSplittingBranchesFlag off.");
    fDOnSplittingBranchesFlag = fl; 
   }; 
   void Integrate(Clause* cl);
   void Remove(Clause* cl);

   bool Rewrite(const Flatterm* query,const SetOfSplittingBranches& splittingBranches)
   {
    CALL("Rewrite(const Flatterm* query,const SetOfSplittingBranches& splittingBranches)");
    ASSERT(degree > 0); 
    branchesUsed = false;
    if (degree == 1) return RewriteByPreordered(query,splittingBranches);
    return RewriteByConstrained(query,splittingBranches);
   };

   Clause* Rule() const { return rule; };  
   bool RHSNext() { return rule_rhs.Next(); };    
   TERM RHSSymbol() { return rule_rhs.Symbol(); };   
   bool SplittingBranchesUsed() const { return branchesUsed; }; 

   void Destroy() { index.Destroy(); };

   #ifndef NO_DEBUG
    bool clean() const { return index.clean(); };
   #endif

   #ifndef NO_DEBUG_VIS 
    ostream& outputIndex(ostream& str) { return index.output(str); };
   #endif

  private:
   bool CanBeRewritingRule(const Clause* cl);   
   bool CheckSplittingBranches(const Clause* cl,const SetOfSplittingBranches& splittingBranches);
   bool RewriteByPreordered(const Flatterm* query,const SetOfSplittingBranches& splittingBranches);
   bool RewriteByConstrained(const Flatterm* query,const SetOfSplittingBranches& splittingBranches);
   bool checkConstraint(SharedKBConstr* constr,Index::SubstitutionCache* subst);
   bool checkNearlyTautology(SharedKBConstr* constr,Index::SubstitutionCache* subst);
   bool checkNearlyContradiction(SharedKBConstr* constr,Index::SubstitutionCache* subst);
   bool checkVolatile(SharedKBConstr* constr,Index::SubstitutionCache* subst);  
   bool checkPurelyLexicographic(SharedKBConstr* constr,Index::SubstitutionCache* subst);    
   bool lexicographicallyGreater(const TERM* t1,const TERM* t2,Index::SubstitutionCache* subst);  

  private: // data
   Index index;
   Clause* rule;
   Index::TermWithSubst rule_rhs;
   long degree;
   bool fDOnSplittingBranchesFlag;
   bool branchesUsed;
   KBConstraint _constraintInstance; 
   Index::SubstitutionCache _substitutionCache;      
 }; // class ForwardDemodulation

inline bool ForwardDemodulation::CanBeRewritingRule(const Clause* cl)
{
 CALL("CanBeRewritingRule(const Clause* cl)");
 // cl is either unit positive equality or a clause with exactly one selected positive equality literal
 // and a number of literals that are splitting branches 
 TERM* lit; 
 if (cl->Unit())
  {
   lit = cl->LitList()->firstLit();
   return (lit->Positive() && lit->IsEquality());
  }
 else  
  {
   if (!fDOnSplittingBranchesFlag) return false;
   lit = cl->LitList()->firstLit();
   if (lit->Negative() || (!lit->IsEquality()) || (cl->LitList()->numOfSelectedLits() != 1)) return false;
   LiteralList::Iterator iter(cl->LitList());
   iter.next();
   ASSERT(iter.notEnd());
   do
    {
     if (!iter.currentLiteral()->IsSplitting()) return false;
     iter.next();
    }
   while (iter.notEnd());
   return true;
  };
}; // bool ForwardDemodulation::CanBeRewritingRule(const Clause* cl) 


inline bool ForwardDemodulation::CheckSplittingBranches(const Clause* cl,const SetOfSplittingBranches& splittingBranches)
{
 CALL("CheckSplittingBranches(const Clause* cl,const SetOfSplittingBranches& splittingBranches)");
 if ((!fDOnSplittingBranchesFlag) || cl->Unit())
  {
   ASSERT(cl->Unit());
   return true; 
  }; 
 LiteralList::Iterator iter(cl->LitList());
 iter.next(); 
 while (iter.notEnd()) 
  {
   if (!splittingBranches.scoreOf(iter.currentLiteral()->HeaderNum())) return false;
   iter.next();
  };
 branchesUsed = true;
 return true;
}; // bool ForwardDemodulation::CheckSplittingBranches(const Clause* cl,const SetOfSplittingBranches& splittingBranches)


 
inline bool ForwardDemodulation::lexicographicallyGreater(const TERM* t1,const TERM* t2,Index::SubstitutionCache* subst)
{
 CALL("lexicographicallyGreater(const TERM* t1,const TERM* t2,Index::SubstitutionCache* subst)");

 //DF; cout << "LEX " << t1 << " vs " << t2 << "\n"; 
//DF; Index::TermWithSubst::output(cout,t1,subst) << " vs ";
//DF; Index::TermWithSubst::output(cout,t2,subst) << "\n";


 static Index::TermWithSubst iter1;
 static Index::TermWithSubst iter2;
 iter1.Reset(t1,subst);
 iter2.Reset(t2,subst);
 check_symbols:
  if (iter1.Symbol().isVariable())
   {
    if (iter2.Symbol() != iter1.Symbol()) return false; // incomparable 
   }
  else
   if (iter2.Symbol().isVariable()) 
    {
     return false; // incomparable
    }
   else // both are functions
    { 
     if (iter1.Symbol().greaterPrecedence(iter2.Symbol())) return true; 
     if (iter2.Symbol().greaterPrecedence(iter1.Symbol())) return false; 
    };
  ASSERT(iter2.Symbol() == iter1.Symbol());
  if (iter1.Next())
   {
    ALWAYS(iter2.Next());
    goto check_symbols;
   }; 
  return false; // equal terms
}; // bool ForwardDemodulation::lexicographicallyGreater(const TERM* t1,const TERM* t2,Index::SubstitutionCache* subst)


inline bool ForwardDemodulation::checkNearlyTautology(SharedKBConstr* constr,Index::SubstitutionCache* subst)
{
 CALL("checkNearlyTautology(SharedKBConstr* constr,Index::SubstitutionCache* subst)");

 const SharedKBConstr::WeightForm* posWF = constr->posWeightForm();
 ASSERT(posWF);
 ulong negFuncs = constr->negFuncs(); 
 ulong posWeight = 0;
 SharedKBConstr::WeightForm::Member mem;
 do 
  {
   mem = posWF->hd();
   ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
   posWeight += mem.coeff() * ((*(*(subst->subst(mem.var()))))->size()); 
   if (posWeight > negFuncs) 
    {
     return true;
    };
   posWF = posWF->tl();
  } 
 while (posWF);

 if ((posWeight == negFuncs)) 
  {
   if (constr->lexComp() == Greater) return true; 
   if (constr->lexComp() == Less) return false;
   return lexicographicallyGreater(constr->lhsTerm(),constr->rhsTerm(),subst); 
  }; 

 return false;
}; // bool ForwardDemodulation::checkNearlyTautology(SharedKBConstr* constr,Index::SubstitutionCache* subst)

inline bool ForwardDemodulation::checkNearlyContradiction(SharedKBConstr* constr,Index::SubstitutionCache* subst)
{
 CALL("checkNearlyContradiction(SharedKBConstr* constr,Index::SubstitutionCache* subst)");

 const SharedKBConstr::WeightForm* negWF = constr->negWeightForm();
 ASSERT(negWF);
 ulong posFuncs = constr->posFuncs(); 
 ulong negWeight = 0;
 SharedKBConstr::WeightForm::Member mem;
 const SharedKBConstr::WeightForm* tmp = negWF;
 do 
  {
   mem = negWF->hd();
   ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
   negWeight += mem.coeff() * ((*(*(subst->subst(mem.var()))))->size());
   if (negWeight > posFuncs)
    { 
     return false;
    };
   negWF = negWF->tl();
  }
 while (negWF);

 negWF = tmp;
 // all the terms must be ground
 do 
  {
   mem = negWF->hd();
   ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
   if ((*(*(subst->subst(mem.var()))))->Nonground())
    {
     return false;
    };
   negWF = negWF->tl();
  }
 while (negWF);

 if (negWeight < posFuncs) 
  {
   return true;
  };

 ASSERT(negWeight == posFuncs);
 if (constr->lexComp() == Greater) return true;
 if (constr->lexComp() == Less) return false;
 return lexicographicallyGreater(constr->lhsTerm(),constr->rhsTerm(),subst); 
}; // bool ForwardDemodulation::checkNearlyContradiction(SharedKBConstr* constr,Index::SubstitutionCache* subst) 

inline bool ForwardDemodulation::checkVolatile(SharedKBConstr* constr,Index::SubstitutionCache* subst)
{
 CALL("checkVolatile(SharedKBConstr* constr,Index::SubstitutionCache* subst)");

 const SharedKBConstr::WeightForm* posWF = constr->posWeightForm();
 const SharedKBConstr::WeightForm* negWF = constr->negWeightForm();
 ASSERT(posWF);
 ASSERT(negWF);
 ulong posWeight = 0;
 ulong negWeight = 0;
 SharedKBConstr::WeightForm::Member mem;
 const SharedKBConstr::WeightForm* tmpPos = posWF;
 const SharedKBConstr::WeightForm* tmpNeg = negWF;

 // collecting weight of the positive part
 do 
  {
   mem = posWF->hd();
   if (mem.tag() == SharedKBConstr::WeightForm::PlusConst)
    {
     posWeight += mem.constant();
    }
   else
    {
     ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
     posWeight += mem.coeff() * ((*(*(subst->subst(mem.var()))))->size());
    };
   posWF = posWF->tl();
  }
 while (posWF);

 // collecting weight of the negative part
 do 
  {
   mem = negWF->hd();
   if (mem.tag() == SharedKBConstr::WeightForm::PlusConst)
    {
     negWeight += mem.constant();
    }
   else
    {
     ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
     negWeight += mem.coeff() * ((*(*(subst->subst(mem.var()))))->size());
    };
  
   if (posWeight < negWeight)
    {
     return false;
    };   

   negWF = negWF->tl();
  }
 while (negWF);
  
 ASSERT(posWeight >= negWeight);

 // more thorough test needed
 posWF = tmpPos;
 negWF = tmpNeg;
 const Flatterm* substTerm;
 Flatterm::Iterator substTermIter; 
 TERM substSym;
 ulong substCoeff;

 // collecting instance of the positive part
 _constraintInstance.initPosSubst(0UL);
 do 
  {
   mem = posWF->hd();
   if (mem.tag() == SharedKBConstr::WeightForm::PlusConst)
    {
     _constraintInstance.posSubstFunc(mem.constant());
    }
   else
    {
     ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
     substCoeff = mem.coeff();
     substTerm = *(*(subst->subst(mem.var())));
     substTermIter.Reset(substTerm);
     do 
      {
       substSym = substTermIter.CurrPos()->symbolRef(); 
       if (substSym.isVariable())
	{
	 _constraintInstance.posSubstVar(substSym.var(),substCoeff);
        }       
       else
	{
	 ASSERT(substSym.isComplex());
         _constraintInstance.posSubstFunc(substCoeff);
        };
       substTermIter.NextSym();
      }
     while (substTermIter.NotEndOfTerm()); 
    };
   posWF = posWF->tl();
  }
 while (posWF); 

 // collecting instance of the negative part

 _constraintInstance.initNegSubst(0UL);
 do 
  {
   mem = negWF->hd();
   if (mem.tag() == SharedKBConstr::WeightForm::PlusConst)
    {
     ALWAYS(_constraintInstance.negSubstFunc(mem.constant()));
    }
   else
    {
     ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
     substCoeff = mem.coeff();
     substTerm = *(*(subst->subst(mem.var())));
     substTermIter.Reset(substTerm);
     do 
      {
       substSym = substTermIter.CurrPos()->symbolRef(); 
       if (substSym.isVariable())
	{
	 if (!_constraintInstance.negSubstVar(substSym.var(),substCoeff))
	  { 
           return false;
          };
        }       
       else
	{
	 ASSERT(substSym.isComplex());
         ALWAYS(_constraintInstance.negSubstFunc(substCoeff));
        };
       substTermIter.NextSym();
      }
     while (substTermIter.NotEndOfTerm()); 
    };
   negWF = negWF->tl();
  }
 while (negWF);  


 ASSERT(_constraintInstance.posSubstWeight() >= _constraintInstance.negSubstWeight());

 if (_constraintInstance.posSubstWeight() == _constraintInstance.negSubstWeight())
  {     
   if (constr->lexComp() == Greater) return true;
   if (constr->lexComp() == Less) return false;
   return lexicographicallyGreater(constr->lhsTerm(),constr->rhsTerm(),subst); 
  };

 return true;  
}; // bool ForwardDemodulation::checkVolatile(SharedKBConstr* constr,Index::SubstitutionCache* subst) 

inline bool ForwardDemodulation::checkPurelyLexicographic(SharedKBConstr* constr,Index::SubstitutionCache* subst) 
{
 CALL("bool checkPurelyLexicographic(SharedKBConstr* constr,Index::SubstitutionCache* subst)");
 return lexicographicallyGreater(constr->lhsTerm(),constr->rhsTerm(),subst); 
}; // bool ForwardDemodulation::checkPurelyLexicographic(SharedKBConstr* constr,Index::SubstitutionCache* subst)


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
 #define DEBUG_NAMESPACE "ForwardDemodulation::Index::SubstitutionCache"
#endif 
#include "debugMacros.h"
//=================================================

inline void ForwardDemodulation::Index::SubstitutionCache::load(const StoredSubstitution* s)
{
 CALL("load(const StoredSubstitution* s)");
 ulong v = 0;
 while (s) 
 {
  _subst[v] = s->hd();
  s = s->tl();
  v++;
 };
 //for (ulong v = 0; v < s->NumOfVars(); v++) _subst[v] = s->Subst()[v];  

}; //  void ForwardDemodulation::Index::SubstitutionCache::load(const StoredSubstitution* s)


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
 #define DEBUG_NAMESPACE "ForwardDemodulation::Index"
#endif 
#include "debugMacros.h"
//=================================================


inline bool ForwardDemodulation::Index::appropriateDegree(const KBConstraint::Category& cat) const 
{
 CALL("appropriateDegree(const KBConstraint::Category& cat) const");
 switch (cat) 
  {
   case KBConstraint::NearlyTautology: return (degree >= 2);
   case KBConstraint::Volatile: return (degree >= 3);
   case KBConstraint::NearlyContradiction: return (degree >= 4);
   case KBConstraint::PurelyLexicographic: return (degree >= 4);
  };
 #ifdef DEBUG_NAMESPACE
  ICP("ICP1"); return false;
 #else
  #ifdef _SUPPRESS_WARNINGS_
   return false;
  #endif
 #endif
}; // bool ForwardDemodulation::Index::appropriateDegree(const KBConstraint::Category& cat) const 

inline void ForwardDemodulation::Index::collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& funcs)
{
 CALL("collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& funcs)");
 ASSERT(t->isComplex());
 vars.Reset();
 funcs = 0;
 _termIter.Reset(t);       
 do 
  {
   if (_termIter.CurrentSymIsVar())
    {
     vars.Register(_termIter.symbolRef().var());
    } 
   else funcs++;
  }
 while (_termIter.Next());  
}; // void ForwardDemodulation::Index::collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& funcs)


inline ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>* ForwardDemodulation::Index::Integrate(TERM* t)
{
 CALL("Integrate(TERM* t)");
 ASSERT(t->isComplex()); 
 m_index.integrationReset(*t);
 TERM::Iterator iter(t); 
 while (iter.Next()) 
  if (iter.CurrentSymIsVar()) 
   { m_index.integrationVar(iter.symbolRef()); }
  else m_index.integrationFunc(iter.symbolRef());

 bool newLeaf;
 ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>* res =  m_index.integrationGetLeaf(newLeaf);
 if (newLeaf) res->leafInfo().init();
 return res;
}; // ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>* ForwardDemodulation::Index::Integrate(TERM* t)

inline bool ForwardDemodulation::Index::Remove(TERM* t,ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>*& leaf,ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>**& branch_to_destroy)
{
 CALL("Remove(TERM* t,ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>*& leaf,ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>**& branch_to_destroy)");
 if (!m_index.removalReset(*t)) return false;
 TERM::Iterator iter(t);
 while (iter.Next())  
  {    
   if (iter.CurrentSymIsVar())
    {
     if (!m_index.removalVar(iter.symbolRef())) return false;
    }
   else if (!m_index.removalFunc(iter.symbolRef())) return false;
  };

 return m_index.removalFinish(leaf,branch_to_destroy);
}; // bool ForwardDemodulation::Index::Remove(TERM* t,ForwardMatchingIndexTreeLeaf<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>*& leaf,ForwardMatchingIndexTreeInstruction<HellConst::MaxTermSize,HellConst::MaxNumOfVariables,GlobAlloc,TERM,Flatterm,ForwardDemodulation::Index::LeafInfo,ForwardDemodulationIndexId>**& branch_to_destroy) 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
 #define DEBUG_NAMESPACE "ForwardDemodulation::Index::TermWithSubst"
#endif 
#include "debugMacros.h"
//=================================================

inline void ForwardDemodulation::Index::TermWithSubst::Reset(const TERM& t,SubstitutionCache* s)
{
 CALL("Reset(const TERM& t,SubstitutionCache* s)");
 DOP(max_var_num = s->numOfVars() - 1); 
 subst = s->subst();
 if (t.isVariable())
  {
   in_term = false;
   ASSERT(t.var() <= max_var_num);
   subst_iter.Reset(*(subst[t.var()]));
   term_iter.Reset(&dummy_term);
   term_iter.Next();
  }
 else // t is reference to a complex term 
  {
   in_term = true;
   term_iter.Reset(t.First());
  };  
}; // void ForwardDemodulation::Index::TermWithSubst::Reset(const TERM& t,SubstitutionCache* s)

        
inline void ForwardDemodulation::Index::TermWithSubst::Reset(const TERM* t,SubstitutionCache* s)
{
 // for terms that are known to be complex
 CALL("Reset(const TERM* t,SubstitutionCache* s)");
 ASSERT(t->isComplex());
 DOP(max_var_num = s->numOfVars() - 1); 
 subst = s->subst();          
 in_term = true;
 term_iter.Reset(t);
}; // void ForwardDemodulation::Index::TermWithSubst::Reset(const TERM* t,SubstitutionCache* s)

inline bool ForwardDemodulation::Index::TermWithSubst::Next()
{
 CALL("TermWithSubst::Next()");
 if (in_term)
  {
  next_top_sym:
   if (!term_iter.Next()) return false;
   if (term_iter.symbolRef().isVariable()) 
    {
     in_term = false;
     ASSERT(term_iter.symbolRef().var() <= max_var_num);
     subst_iter.Reset(*(subst[term_iter.symbolRef().var()]));
    };
   return true;
  }
 else
  {
  subst_iter.NextSym();
  if (!subst_iter.EndOfTerm()) return true;
  in_term = true;
  goto next_top_sym;
  };
}; // bool ForwardDemodulation::Index::TermWithSubst::Next()

inline const TERM& ForwardDemodulation::Index::TermWithSubst::Symbol()
{
 if (in_term) return term_iter.symbolRef();
 return subst_iter.CurrPos()->symbolRef();
}; 


#ifndef NO_DEBUG
 inline void ForwardDemodulation::Index::TermWithSubst::CollectWeight(const TERM* t,SubstitutionCache* s,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs)
 {
  CALL("CollectWeight(const TERM* t,SubstitutionCache* s,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs)");
  TermWithSubst iter(t,s);
  TERM sym;
  do
   {     
    sym = iter.Symbol();
    if (sym.isComplex()) { funcs++; }
    else vms.Register(sym.var());             
   }
  while (iter.Next());
 };
 inline void ForwardDemodulation::Index::TermWithSubst::CollectWeight(const TERM& t,SubstitutionCache* s,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs)
 {
  CALL("CollectWeight(const TERM& t,SubstitutionCache* s,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs)");
  TermWithSubst iter(t,s);
  TERM sym;
  do
   {     
    sym = iter.Symbol();
    if (sym.isComplex()) { funcs++; }
    else vms.Register(sym.var());             
   }
  while (iter.Next());
 };
    
 inline bool ForwardDemodulation::Index::TermWithSubst::LexGreater(const TERM* t1,const TERM& t2,SubstitutionCache* s)
 {
  CALL("LexGreater(const TERM* t1,const TERM& t2,SubstitutionCache* s)");
  TermWithSubst iter1(t1,s);
  TermWithSubst iter2(t2,s);
  TERM sym1;
  TERM sym2;
  do
   {
    sym1 = iter1.Symbol();
    sym2 = iter2.Symbol(); 
    if (sym1 != sym2)
     { 
      return (sym1.isComplex()
             && sym2.isComplex()
             && (sym1.CompareFunctor(sym2) == Greater));                       
     };
    iter2.Next();
   }
  while (iter1.Next());
  return false; // equal
 };


 inline bool ForwardDemodulation::Index::TermWithSubst::KBGreater(const TERM* t1,const TERM& t2,SubstitutionCache* s)
 { 
  CALL("KBGreater(const TERM* t1,const TERM& t2,SubstitutionCache* s)");
  MultisetOfVariables<HellConst::MaxNumOfVariables> vars1;
  MultisetOfVariables<HellConst::MaxNumOfVariables> vars2;        
  ulong funcs1 = 0;
  ulong funcs2 = 0;
  CollectWeight(t1,s,vars1,funcs1);
  CollectWeight(t2,s,vars2,funcs2);
  if (!vars1.GreaterOrEqual(vars2)) return false;
  ulong weight1 = funcs1 + vars1.Size();           
  ulong weight2 = funcs2 + vars2.Size(); 
  if (weight1 == weight2) return LexGreater(t1,t2,s); 
  return (weight1 > weight2);          
 };
#endif

#ifndef NO_DEBUG_VIS
 inline ostream& ForwardDemodulation::Index::TermWithSubst::output(ostream& str,const TERM* t,SubstitutionCache* s)
 {
  CALL("output(ostream& str,const TERM* t,SubstitutionCache* s)");
  TermWithSubst iter(t,s);
  do
   { 
    str << iter.Symbol() << ' ';
   }
  while (iter.Next());
  return str;
 }; 
#endif


//================================================
#undef DEBUG_NAMESPACE
//=================================================================
#endif
