//
// File:         HellConst.h
// Description:  Global constants for Hell.
// Created:      Feb 12, 2000, 17:50
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef HELL_CONST_H
//====================================================
#define HELL_CONST_H
#include <limits.h>
#include <stddef.h>
#include "jargon.h"
//====================================================
namespace HellConst
{
  // interface 
 const char* const SkolemConstantNamePrefix = "c__";
 const char* const SkolemFunctorNamePrefix = "f__";
 const char* const SkolemPredicateNamePrefix = "p__";
 const char* const CanonicalVariableNamePrefix = "X";
  // basic parameters
 const ulong MaxNumOfLiterals = 31; // can not be greater than 31!
 const ulong MaxNumOfVariables = 64;
 #ifdef LARGE_TERMS
  const ulong MaxTermDepth = 48;
  const ulong MaxTermSize = 256;  
 #else 
  #ifdef HUGE_TERMS
   const ulong MaxTermDepth = 700; //1024;
   const ulong MaxTermSize = 700; //1024; 
  #else
   const ulong MaxTermDepth = 24;  
   const ulong MaxTermSize = 128;  
  #endif
 #endif
 const ulong MaxNumOfFunctors = 1024;
 const ulong MaxSizeOfSmallObject = 1024;   
 const ulong MaxNumOfClauseAncestors = 1016; 
   // can not be less than MaxTermSize or MaxNumOfVariables
 const ulong MaxNumOfOptions = 256;
 const long MinMaxWeight = 1L;
 const ulong MaxClauseWeight = 1024; 

 // hard values: don't change this unless you know what you are doing!
 const ulong UnordEqNum = 0;
 const ulong OrdEqNum = 1;
 const ulong FirstNoneqFun = 2; 
 const ulong ReservedPred0 = 2;
 const ulong ReservedPred1 = 3;
 const ulong ReservedPred2 = 4;
 
  // derived values
 const ulong MaxNumOfTermArguments = MaxNumOfLiterals;
 const ulong MaxClauseSize = MaxNumOfLiterals*MaxTermSize;
 const ulong MaxNumOfPredicates = MaxNumOfFunctors;
 const ulong MaxNumOfDifferentHeaders = 2*MaxNumOfFunctors;
 const ulong MaxNumOfStandardIndices = MaxNumOfLiterals + 1;
 const ulong MaxNumOfIndices = MaxNumOfStandardIndices + 2;
  // 2 extra indices for unit conflict in hyperresolution
 const ulong MaxTmpTermSize = 8*MaxTermSize; // potentially unsafe
 const ulong MaxTmpTermDepth = MaxTmpTermSize; // potentially unsafe
  // checking integrity
 inline bool checkIntegrity() 
 {
  return (MaxNumOfLiterals <= 31)
         && (UnordEqNum == 0)
         && (OrdEqNum == 1)
         && (FirstNoneqFun == 2)
         && (MaxSizeOfSmallObject >= MaxTermSize)
         && (MaxSizeOfSmallObject >= MaxNumOfVariables)
         && (MaxNumOfClauseAncestors + 8 <= MaxSizeOfSmallObject);
 
 };
  // functions
 inline bool isEquality(const ulong& num) { return (num < FirstNoneqFun); }; 
 inline bool isOrderedEq(const ulong& num) { return (bool)num; }; 
 inline bool isUnorderedEq(const ulong& num) { return !num; }; 
 inline bool isReservedPred(const ulong& num) { return (num > 1) && (num < 5); };
}; // namespace HellConst
//====================================================
#endif



