//
// File:         HellInterceptMacros.h
// Description:  Macros for intercepting some operations from the library.
// Created:      Dec 05, 2000, 14:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef HELL_INTERCEPT_MACROS_H  
//====================================================
#define HELL_INTERCEPT_MACROS_H 
#include <iostream.h> 
#include "Exit.h"
//====================================================   

// Indexing operations for forward matching:     

#ifdef INTERCEPT_FORWARD_MATCHING_INDEXING
                   
 #include "Logging.h"

 extern Logging FMILog; // declared in HellInterceptMacros.cpp
 extern const Flatterm* interceptedForwardMatchingQuery;        
 extern long maxSizeOfFMILogFile; 
 inline void checkMaxSizeOfFMILogFile()
 {
  if (FMILog.counter() >= maxSizeOfFMILogFile) 
   { 
    cout << "%    Forced termination: file fmilog is full.\n"; 
    Exit::exit(0); 
   }; 
 };
 #define INTERCEPT_FORWARD_MATCHING_INDEXING_INTEGRATE(term) \
  { \
   FMILog.output("+"); \
   FMILog.output(term); \
   FMILog.output("\n"); \
   FMILog.flush(); \
   checkMaxSizeOfFMILogFile(); \
  }

 #define INTERCEPT_FORWARD_MATCHING_INDEXING_REMOVE(term) \
  { \
   FMILog.output("-"); \
   FMILog.output(term); \
   FMILog.output("\n"); \
   FMILog.flush(); \
   checkMaxSizeOfFMILogFile(); \
  }

 #define INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_QUERY(term) \
  ( \
   (interceptedForwardMatchingQuery=term), \
   term \
  )

 #define INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_RESULT(res)\
  ( \
     (\
      (interceptedForwardMatchingQuery) ? \
       ( \
        (res) ? \
	 ( \
          FMILog.output("!"), \
          FMILog.output(interceptedForwardMatchingQuery), \
          FMILog.output("\n"), \
          interceptedForwardMatchingQuery = (const Flatterm*)0, \
          FMILog.flush(), \
          checkMaxSizeOfFMILogFile(), \
          res \
         ) \
	 : \
	 ( \
          FMILog.output("?"), \
          FMILog.output(interceptedForwardMatchingQuery), \
          FMILog.output("\n"), \
          interceptedForwardMatchingQuery = (const Flatterm*)0, \
          FMILog.flush(), \
          checkMaxSizeOfFMILogFile(), \
          res \
         ) \
       ) \
       : \
       res \
     ) \
  )
#endif

// Undefining the macros in the case an efficient version of Vampire is built  

#ifndef INTERCEPT_FORWARD_MATCHING_INDEXING

 #undef INTERCEPT_FORWARD_MATCHING_INDEXING_INTEGRATE
 #define INTERCEPT_FORWARD_MATCHING_INDEXING_INTEGRATE(term) 
 
 #undef INTERCEPT_FORWARD_MATCHING_INDEXING_REMOVE
 #define INTERCEPT_FORWARD_MATCHING_INDEXING_REMOVE(term)

 #undef INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_QUERY  
 #define INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_QUERY(term) term

 #undef INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_RESULT
 #define INTERCEPT_FORWARD_MATCHING_INDEXING_RETRIEVAL_RESULT(res) res
 
#endif

// Indexing operations for backward subsumption:
    
#ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
                   
 #include "Logging.h"

 extern Logging BSILog; // declared in HellInterceptMacros.cpp
 extern const Clause* interceptedBackwardSubsumptionQuery;
 extern ulong interceptedBackwardSubsumptionResultTotal;       
 extern long maxSizeOfBSILogFile; 
 inline void checkMaxSizeOfBSILogFile()
 {
  if (BSILog.counter() >= maxSizeOfBSILogFile) 
   { 
    cout << "%    Forced termination: file bsiilog is full.\n"; 
    Exit::exit(0); 
   }; 
 };

 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTEGRATE(clause) \
  { \
   BSILog.output("+"); \
   BSILog.output(clause->Number()); \
   BSILog.output("\n"); \
   BSILog.flush(); \
   checkMaxSizeOfBSILogFile(); \
  }

 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REMOVE(clause) \
  { \
   BSILog.output("-"); \
   BSILog.output(clause->Number()); \
   BSILog.output("\n"); \
   BSILog.flush(); \
   checkMaxSizeOfBSILogFile(); \
  }

 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_QUERY(clause) \
  { \
   interceptedBackwardSubsumptionQuery=clause; \
   interceptedBackwardSubsumptionResultTotal = 0UL; \
   BSILog.output("?"); \
   BSILog.output(clause->Number()); \
   BSILog.output("\n"); \
   BSILog.flush(); \
   checkMaxSizeOfBSILogFile(); \
  }

 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_RESULT(res) \
  { \
   interceptedBackwardSubsumptionResultTotal++; \
   BSILog.output("!"); \
   BSILog.output(res->Number()); \
   BSILog.output("\n"); \
   BSILog.flush(); \
   checkMaxSizeOfBSILogFile(); \
  }

 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_TOTAL \
  { \
   BSILog.output("$"); \
   BSILog.output(interceptedBackwardSubsumptionResultTotal); \
   BSILog.output(" "); \
   BSILog.output("\n"); \
   BSILog.flush(); \
   checkMaxSizeOfBSILogFile(); \
  } 

 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE(clause) \
  { \
   BSILog.output("#"); \
   BSILog.output(clause->Number()); \
   BSILog.output(":"); \
   BSILog.output(clause); \
   BSILog.output("\n"); \
   BSILog.flush(); \
   checkMaxSizeOfBSILogFile(); \
  } 
#endif

// Undefining the macros in the case an efficient version of Vampire is built  

#ifndef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING

 #undef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTEGRATE
 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTEGRATE(clause) 
 
 #undef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REMOVE
 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REMOVE(clause)

 #undef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_QUERY  
 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_QUERY(clause)

 #undef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_RESULT
 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_RESULT(res) 

  
 #undef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_TOTAL 
 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_TOTAL(comment)

 #undef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE
 #define INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE(clause)
#endif

//====================================================    
#endif




