//
// File:         InferenceTree.h
// Description:  Collecting inference trees.
// Created:      Feb 13, 2000, 18:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef INFERENCE_TREE_H
//=============================================================================
#define INFERENCE_TREE_H 
#include "jargon.h"
#include "HellDebugFlags.h"
#include "GSortTree.h"
#include "clause.h"
#include "ClauseBackground.h"
#include "GlobAlloc.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INFERENCE_TREE
 #define DEBUG_NAMESPACE "InferenceTree"
#endif
#include "debugMacros.h"
//============================================================================
class InferenceTree
{
 public:
  InferenceTree() : collected(), iter() 
  {
   CALL("constructor InferenceTree()");
  };
  InferenceTree(const Clause* root) : collected(), iter()
  {
   CALL("constructor InferenceTree(const Clause* root)");
   collect(root);
  };
  ~InferenceTree() { CALL("destructor ~InferenceTree()"); };
  void collect(const Clause* root)
  {
   collected.destroy();
   collectInf(root);
  };
  void resetIter() { iter.reset(collected); };
  const Clause* getNextClause() 
  {
   if (iter)
    {
     const Clause* res = iter.value().clause();
     iter.next();
     return res; 
    }; 
   return (const Clause*)0; 
  };
 private:
  class ClWrapper
  {
   public:
    ClWrapper(const Clause* c) : cl(c) {};
    const Clause* clause() const { return cl; };
    bool operator==(const ClWrapper& cw) const
    {
     return cl->Number() == cw.clause()->Number();
    };
    bool operator<(const ClWrapper& cw) const
    {
     return cl->Number() < cw.clause()->Number();
    };
    bool operator>(const ClWrapper& cw) const
    {
     return cl->Number() > cw.clause()->Number();
    };
   private:
    ClWrapper() { ERROR("Default constructor called"); };
   private:  
    const Clause* cl; 
  };
  typedef GSortTree<GlobAlloc,ClWrapper,InferenceTree> ClBag;
 private:
  void collectInf(const Clause* root)
  {
   unsigned long oldScore;
   collected.insert(ClWrapper(root),oldScore);
   if (!oldScore)
    {
     for (ClauseBackground::ClauseList* anc = root->Ancestors()->ancestors(); anc; anc = anc->tl())
      collectInf(anc->hd());
    };
  };
 private:
  ClBag collected;
  ClBag::LeftRightIter iter;    
}; //class InferenceTree 

//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
