#include "NewClause.h"
#include "HellDebugFlags.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEW_CLAUSE
 #define DEBUG_NAMESPACE "NewClause"
#endif 
#include "debugMacros.h"
//=================================================

void NewClause::select1(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");

 /*
   If lits contains some inherently selected literals, only they are selected. 
   Otherwise, we select all maximal literals.
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif


 //DF; cout << "SELECT1\n"; 
 if (stat.inherently_selected) 
  {
   //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 negSelUsed = false; 
 ASSERT(lits.Length() > 1);
 TMP_LITERAL* cand;
 TERM max_header; 
 if (_literalComparisonMode == 0L) goto compare_atoms;

 // filter out literals with nonmaximal headers
 max_header = (*(lits.Last()))->Header();
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((_literalComparisonMode == 2L) 
      && ((*lit)->Header().functor() == max_header.functor()))
    {     
     // do nothing
    }
   else
    {
     switch (GlobalHeaderWeighting.Compare((*lit)->Header(),max_header))
      {
       case Greater :
        max_header = (*lit)->Header();
        do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);
        break; 
       case Less :
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
        break;
       case Equal :
        ASSERT(((*lit)->Header() == max_header) || ((*lit)->Header().IsEquality() &&  max_header.IsEquality() && ((*lit)->Header().polarity() == max_header.polarity())));
        break;
       case Incomparable : 
        #ifdef DEBUG_NAMESPACE
         ERROR("E0"); 
        #endif
        break; 
      };
    };
  };
      
 // all remaining literals have the maximal header  

 compare_atoms:
  ASSERT(lits.Nonempty());
  cand = lits.Pop();
  ASSERT((_literalComparisonMode != 1L) 
           || (cand->Header() == max_header) 
           || (cand->Header().IsEquality() 
              &&  max_header.IsEquality() 
	      && (cand->Header().polarity() == max_header.polarity()))); 
  ASSERT((_literalComparisonMode != 2L) 
         || (cand->Header().functor() == max_header.functor())
         || (cand->Header().IsEquality() 
              &&  max_header.IsEquality())); 
  if (lits.Empty()) { max_lits.Push(cand); return; };
  for (TMP_LITERAL** lit = lits.Last(); lit >= lits.First(); lit--)
   {
    ASSERT((_literalComparisonMode != 1L) 
           || ((*lit)->Header() == max_header) 
           || ((*lit)->Header().IsEquality() 
              &&  max_header.IsEquality() 
	      && ((*lit)->Header().polarity() == max_header.polarity()))); 
    ASSERT((_literalComparisonMode != 2L) 
           || (cand->Header().functor() == max_header.functor())
           || (cand->Header().IsEquality() 
               &&  max_header.IsEquality())); 

    switch ((*lit)->KBCompare(*cand))
     {
      case Greater:
       nonmax_lits.Push(cand); goto compare_atoms;
      case Less:
       nonmax_lits.Push(*lit); lits.ReplaceByTop(lit); break; 
      case Incomparable: break;

      #ifdef DEBUG_NAMESPACE
    
       case Equal:
        ICP("ICP0"); break;
       default: ICP("ICP1"); break;
      #else  
       #ifdef _SUPPRESS_WARNINGS_
        default: break;
       #endif
      #endif
     };
   };  
  max_lits.Push(cand);
  if (lits.Empty()) return;
  goto compare_atoms;
  ASSERT(max_lits.Nonempty());
}; // void NewClause::select1(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)


void NewClause::select2(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select2(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 
 /* 
   If lits contains some inherently selected literals, only they are selected. 
   Otherwise, if lits contains a negative literal, a negative literal is selected.
   Negative literals with greater weights are prefered.
   If there is no negative literal, we proceed with selection 1.  
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif
 ASSERT(lits.Length() > 1);

 if (stat.inherently_selected) 
  {
   //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 if (stat.negative)
  {
   negSelUsed = true;
   //lits.SelectNegative1(max_lits,nonmax_lits,_inversePolarityForSelection);
   selectNegativeWithMaximalWeight(lits,max_lits,nonmax_lits);
  }
 else // no negative literals
  { 
   select1(lits,max_lits,nonmax_lits,negSelUsed,stat);
  };  
}; // void NewClause::select2(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)

void NewClause::select3(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select3(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");

 /*
   If lits contains some inherently selected literals, only they are selected. 
   Otherwise, if lits contains a negative literal, a negative literal is selected.
   To choose between negative literals we heuristically estimate the number of possible
   inferences and a literal promising less inferences is selected.
   If there is no negative literal, we proceed with selection 1. 
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif ASSERT(lits.Length() > 1);

 if (stat.inherently_selected) 
  {
    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 if (stat.negative)
  {
   negSelUsed = true;
   //lits.SelectNegative2(max_lits,nonmax_lits,_inversePolarityForSelection);
   selectNegativeLeastProlific(lits,max_lits,nonmax_lits);
  }
 else // no negative literals
  { 
   select1(lits,max_lits,nonmax_lits,negSelUsed,stat);
  };    
}; // void NewClause::select3(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)

void NewClause::select4(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select4(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 
 /* 
   If lits contains some inherently selected literals, only they are selected. 
   Otherwise, we first select maximal literals. Then we try to improve selection 
   by selecting a negative literal. This is based on the assumption that 
   selecting one literal is better than selecting several literals and 
   selecting a negative literal with greater weight is better than selecting 
   a maximal literal.
 */ 
 
 ASSERT(lits.Length() > 1);
 
 if (stat.inherently_selected) 
  {

    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 select1(lits,max_lits,nonmax_lits,negSelUsed,stat); 
 if (stat.negative) // quality of selection can be questioned 
  {
   if (!satisfactorySelection1(max_lits,nonmax_lits))
    {
     lits.CollectFrom(nonmax_lits);
     lits.CollectFrom(max_lits);
     //lits.SelectNegative1(max_lits,nonmax_lits,_inversePolarityForSelection);
     selectNegativeWithMaximalWeight(lits,max_lits,nonmax_lits);
     negSelUsed = true;  
    };
  };
}; // void NewClause::select4(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)


void NewClause::select5(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select5(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /* 
   If lits contains some inherently selected literals, only they are selected.
   Otherwise, we first select maximal literals. Then we try to improve selection by selecting
   a negative literal. This is based on the assumption that selecting one literal
   is better than selecting several literals and selecting a negative literal 
   with less possible inferences (heuristicaly estimated) is better than 
   selecting a maximal literal.
 */ 
 
 ASSERT(lits.Length() > 1);

 if (stat.inherently_selected) 
  {
    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 select1(lits,max_lits,nonmax_lits,negSelUsed,stat);
 if (stat.negative) // quality of selection can be questioned 
  {
   if (!satisfactorySelection2(max_lits,nonmax_lits))
    {
     lits.CollectFrom(nonmax_lits);
     lits.CollectFrom(max_lits);
     //lits.SelectNegative2(max_lits,nonmax_lits,_inversePolarityForSelection);
     selectNegativeLeastProlific(lits,max_lits,nonmax_lits);
     negSelUsed = true;
    };
  };
}; // void NewClause::select5(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)



void NewClause::select6(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select6(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 
 /* 
   If lits contains some inherently selected literals, only they are selected. 
   Otherwise, if lits contains a negative literal, a negative literal is selected.
   Among the negative literals we always select a literal 
   with the maximal number of different variables. Among such literals 
   we prefer those having smaller depths of variable occurences.
   Negative literals with greater weights are prefered.
   If there is no negative literal, we proceed with selection 1.  
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif
 ASSERT(lits.Length() > 1);

 if (stat.inherently_selected) 
  {
    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 if (stat.negative)
  {
   negSelUsed = true;
   //lits.SelectNegative1(max_lits,nonmax_lits,_inversePolarityForSelection);
   selectNegativeMaximallyNongroundWithMaximalWeight(lits,max_lits,nonmax_lits);
  }
 else // no negative literals
  { 
   select1(lits,max_lits,nonmax_lits,negSelUsed,stat);
  };  
}; // void NewClause::select6(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)


void NewClause::select7(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select7(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 
 /* 
   If lits contains some inherently selected literals, only they are selected. 
   Otherwise, we first select maximal literals. Then we try to improve selection 
   by selecting a negative literal. This is based on the assumption that 
   selecting one literal is better than selecting several literals and 
   selecting a negative literal with more different variables is better than selecting 
   a maximal literal. If we decide to use negative selection, among the negative literals
   we chose those having more different variable. Among these literals we chose 
   those with smaller depths of variable occurences and then select those having 
   greater size.   
 */ 
 
 ASSERT(lits.Length() > 1);
 
 if (stat.inherently_selected) 
  {

    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 select1(lits,max_lits,nonmax_lits,negSelUsed,stat); 
 if (stat.negative) // quality of selection can be questioned 
  {
   if (!satisfactorySelection1(max_lits,nonmax_lits))
    {
     lits.CollectFrom(nonmax_lits);
     lits.CollectFrom(max_lits);
     selectNegativeMaximallyNongroundWithMaximalWeight(lits,max_lits,nonmax_lits);
     negSelUsed = true;  
    };
  };
}; // void NewClause::select7(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)



void NewClause::select1002(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1002(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
   Arbitrary selection (almost always makes everything incomplete). 
   If lits contains some inherently selected literals, only they are selected.
   Otherwise, if lits contains a literal which is not a positive equality,
   we never select a positive equality. 
   Among the literals available for selection we always select a literal 
   with the maximal number of different variables. Among such literals we prefer those having 
   smaller depths of variable occurences. Among such literals we select those with maximal size.
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif
 ASSERT(lits.Length() > 1);
 if (stat.inherently_selected) 
  {
    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 
 bool selectPositiveEq = (!stat.negative) && (!stat.noneq_lits);

 if (!selectPositiveEq)
  {
   while ((*(lits.Last()))->IsEquality() && (*(lits.Last()))->Positive())
    nonmax_lits.Push(lits.Pop());
  };

 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 
 ulong maxNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent();

 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     // compare the number of different variables 
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars > maxNumOfDiffVars)
      {
       maxNumOfDiffVars = numOfDiffVars;
       do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
      }
     else 
      if (numOfDiffVars < maxNumOfDiffVars)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       };
    };
  };
 // now all the literals have the maximal number of different variables 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };

 // additionally, compare the multisets of depths of variable occurences    
 
 static TMP_LITERAL::VarDepthProfile varDepthProfile1;
 static TMP_LITERAL::VarDepthProfile varDepthProfile2;
 TMP_LITERAL::VarDepthProfile* maxVarDepthProfile = &varDepthProfile1;
 TMP_LITERAL::VarDepthProfile* litVarDepthProfile = &varDepthProfile2; 
 (*(lits.Last()))->collectVarDepthProfile(*maxVarDepthProfile);
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   ASSERT(selectPositiveEq || (!(*lit)->IsEquality()) || (*lit)->Negative());
   (*lit)->collectVarDepthProfile(*litVarDepthProfile); 
   // compare the profiles
   switch (maxVarDepthProfile->compareByInverselyInducedOrder(*litVarDepthProfile))
    {
     case Less :
      maxVarDepthProfile = litVarDepthProfile;
      litVarDepthProfile = ((litVarDepthProfile == &varDepthProfile1) ? &varDepthProfile2 : &varDepthProfile1);
      do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);  
      break;   

     case Greater :  
      nonmax_lits.Push(*lit);
      lits.ReplaceByTop(lit);
      break;

     case Equal : break;
     
     #ifdef DEBUG_NAMESPACE
      default : ICP("ICP0"); return;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: RuntimeError::report("Illegal control point in  NewClause::select1002(..)");
       return;
      #endif
     #endif 
    };
  };

 // select literals with greater size 
 long maxSize = (*(lits.Last()))->size();
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     long litSize = (*lit)->size();
     if (litSize > maxSize)
      {
       maxSize = litSize;
       do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
      }
     else 
      if (litSize < maxSize)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       };
    };
  };
 // now all the literals in lits are of the greatest size 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
  
 // now select on "don't care" basis
 max_lits.Push(lits.Pop());
 while (lits.Nonempty()) nonmax_lits.Push(lits.Pop());
 
}; // void NewClause::select1002(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)


void NewClause::select1003(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1003(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
   Arbitrary selection (almost always makes everything incomplete). 
   If lits contains some inherently selected literals, only they are selected.
   Otherwise, if lits contains a literal which is not a positive equality,
   we never select a positive equality. 
   Among the literals available for selection we always select a literal 
   of the maximal size. Among several such literals we select by the function 1002.        
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif
 ASSERT(lits.Length() > 1);
 if (stat.inherently_selected) 
  {
    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 
 bool selectPositiveEq = (!stat.negative) && (!stat.noneq_lits);
 
 long maxSize = (*(lits.Last()))->size();
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     long litSize = (*lit)->size();
     if (litSize > maxSize)
      {
       maxSize = litSize;
       do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
      }
     else 
      if (litSize < maxSize)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       };
    };
  };
 // now all the literals in lits are of the greatest size 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 TMP_LIT_LIST::STAT newStat;
 lits.CollectStat(newStat,_inversePolarityForSelection);
 
 select1002(lits,max_lits,nonmax_lits,negSelUsed,newStat);

}; // void NewClause::select1003(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)




void NewClause::select1004(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1004(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
   Arbitrary selection (almost always makes everything incomplete). 
   If lits contains some inherently selected literals, only they are selected.
   Otherwise, if lits contains a literal which is not a positive equality,
   we never select a positive equality. 
   Among the literals available for selection we always select a literal 
   with the minimal number of different variables.
   Among such literals we select those having minimal total number of variable
   occurences. Among such literals we select one of the greatest size. 
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif
 ASSERT(lits.Length() > 1);
 if (stat.inherently_selected) 
  {
    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits); 
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits); 
   return;  
  };
 
 bool selectPositiveEq = (!stat.negative) && (!stat.noneq_lits);

 if (!selectPositiveEq)
  {
   while ((*(lits.Last()))->IsEquality() && (*(lits.Last()))->Positive())
    nonmax_lits.Push(lits.Pop());
  };

 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 
 ulong minNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent();

 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     // compare the number of different variables 
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars < minNumOfDiffVars)
      {
       minNumOfDiffVars = numOfDiffVars;
       do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
      }
     else 
      if (numOfDiffVars > minNumOfDiffVars)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       };
    };
  };
 // now all the literals have the minimal number of different variables 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };

 // compare the total number of variable occurences  
 ulong  minNumOfVarOccurences =  (*(lits.Last()))->Variables().Size();
   
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     // compare the number of different variables 
     ulong numOfVarOccurences = (*lit)->Variables().Size();
     if (numOfVarOccurences < minNumOfVarOccurences)
      {
       minNumOfVarOccurences = numOfVarOccurences;
       do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
      }
     else 
      if (numOfVarOccurences > minNumOfVarOccurences)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       };
    };
  };
 // now all the literals have the minimal number of variable occurences 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 
 // now select literals with maximal size
 long maxSize = (*(lits.Last()))->size();
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     long litSize = (*lit)->size();
     if (litSize > maxSize)
      {
       maxSize = litSize;
       do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
      }
     else 
      if (litSize < maxSize)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       };
    };
  };
 // now all the literals in lits are of the greatest size 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 
 // select on "don't care" basis 

 max_lits.Push(lits.Pop());
 while (lits.Nonempty()) nonmax_lits.Push(lits.Pop());
 
}; // void NewClause::select1004(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)



void NewClause::select1005(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1005(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
   Arbitrary selection (almost always makes everything incomplete). 
   If lits contains some inherently selected literals, only they are selected.
   Otherwise, if lits contains a literal which is not a positive equality,
   we never select a positive equality. 
   Among the literals available for selection we always select
   either a "maximally ground" negative literal, or a "maximally nonground" positive literal. 
   "Maximally ground" means having minimal total number of variable occurences 
   among literals having minimal number of different variables. "Maximally non ground" 
   means having maximal total number of variable occurences among the literals 
   having maximal number of different variables. 
   Among such literals we select one with the maximal size.
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif
 ASSERT(lits.Length() > 1);
 if (stat.inherently_selected) 
  {
    //lits.SelectInherentlySelected(stat.inherently_selected,max_lits,nonmax_lits);
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
 
 bool selectPositiveEq = (!stat.negative) && (!stat.noneq_lits);

 if (!selectPositiveEq)
  {
   while ((*(lits.Last()))->IsEquality() && (*(lits.Last()))->Positive())
    nonmax_lits.Push(lits.Pop());
  };

 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };

 ulong minNegLitNumOfDiffVars;
 ulong minNegLitNumOfVarOccurences;
 ulong maxPosLitNumOfDiffVars;
 ulong maxPosLitNumOfVarOccurences;
 if (NegativeForSelection((*(lits.Last()))->Header())) 
  {
   minNegLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent();
   minNegLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
   maxPosLitNumOfDiffVars = 0;
   maxPosLitNumOfVarOccurences = 0;
  }
 else
  {
   minNegLitNumOfDiffVars = HellConst::MaxNumOfVariables;
   minNegLitNumOfVarOccurences = HellConst::MaxTermSize;
   maxPosLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent(); 
   maxPosLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
  }; 
 
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     if (NegativeForSelection((*lit)->Header()))
      {
       ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
       if (numOfDiffVars < minNegLitNumOfDiffVars)
	{
	 minNegLitNumOfDiffVars = numOfDiffVars;
         minNegLitNumOfVarOccurences = (*lit)->Variables().Size();
         for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	  if (NegativeForSelection((*lt)->Header()))
	   {
            nonmax_lits.Push(*lt);
            lits.ReplaceByTop(lt);
           };
        }
       else
	if (numOfDiffVars > minNegLitNumOfDiffVars)
	 {
          nonmax_lits.Push(*lit);
          lits.ReplaceByTop(lit);
         }
        else // numOfDiffVars == minNegLitNumOfDiffVars 
	 {
          ulong numOfVarOccurences = (*lit)->Variables().Size();
          if (numOfVarOccurences < minNegLitNumOfVarOccurences)
	   {
	    minNegLitNumOfVarOccurences = numOfVarOccurences; 
            for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	     if (NegativeForSelection((*lt)->Header()))
	      {
               nonmax_lits.Push(*lt);
               lits.ReplaceByTop(lt);
              };
           }
          else
           if (numOfVarOccurences > minNegLitNumOfVarOccurences)
	    {
             nonmax_lits.Push(*lit);
             lits.ReplaceByTop(lit);             
            };
         };
      }
     else // !NegativeForSelection((*lit)->Header())
      {
       ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
       if (numOfDiffVars > maxPosLitNumOfDiffVars)
	{
	 maxPosLitNumOfDiffVars = numOfDiffVars;
         maxPosLitNumOfVarOccurences = (*lit)->Variables().Size();
         for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	  if (!NegativeForSelection((*lt)->Header()))
	   {
            nonmax_lits.Push(*lt);
            lits.ReplaceByTop(lt);
           };
        }
       else
	if (numOfDiffVars < maxPosLitNumOfDiffVars)
	 {
          nonmax_lits.Push(*lit);
          lits.ReplaceByTop(lit);
         }
        else // numOfDiffVars == maxPosLitNumOfDiffVars 
	 {
          ulong numOfVarOccurences = (*lit)->Variables().Size();
          if (numOfVarOccurences > maxPosLitNumOfVarOccurences)
	   {
	    maxPosLitNumOfVarOccurences = numOfVarOccurences; 
            for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	     if (!NegativeForSelection((*lt)->Header()))
	      {
               nonmax_lits.Push(*lt);
               lits.ReplaceByTop(lt);
              };
           }
          else
           if (numOfVarOccurences < maxPosLitNumOfVarOccurences)
	    {
             nonmax_lits.Push(*lit);
             lits.ReplaceByTop(lit);             
            };
         };
      };


    };
  };
 // now all negative literals are maximally ground
 // and all positive literals are maximally nonground
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; }; 

 // now select literals with maximal size
 long maxSize = (*(lits.Last()))->size();
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if ((!selectPositiveEq) && ((*lit)->IsEquality()) && ((*lit)->Positive()))
    {
     nonmax_lits.Push(*lit);
     lits.ReplaceByTop(lit);   
    }
   else
    {
     long litSize = (*lit)->size();
     if (litSize > maxSize)
      {
       maxSize = litSize;
       do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
      }
     else 
      if (litSize < maxSize)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       };
    };
  };
 // now all the literals in lits are of the greatest size 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 
 // select on "don't care" basis 

 max_lits.Push(lits.Pop());
 while (lits.Nonempty()) nonmax_lits.Push(lits.Pop());   

}; // void NewClause::select1005(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat) 


void NewClause::select1006(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1006(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
  Bidirectional selection. 
   If lits contains some inherently selected literals, only they are selected.
  In a rule, i.e. a clause containing both 
  positive and negative literals we always select one positive and one negative
  literal. If we have several literals of the same polarity, we select 
  those that are maximally nonground. In a fact, i.e. a clause with all literals 
  of the same polarity, we select one maximally nonground literal.     
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif  
 ASSERT(lits.Length() > 1);
   
 if (stat.inherently_selected) 
  {
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };

 //DF; lits.output(cout << "LITS\n") << "\n";

 ulong maxNegLitNumOfDiffVars;
 ulong maxNegLitNumOfVarOccurences;
 ulong maxNegLitSize;
 ulong maxPosLitNumOfDiffVars;
 ulong maxPosLitNumOfVarOccurences;
 ulong maxPosLitSize;
 if (NegativeForSelection((*(lits.Last()))->Header())) 
  {
   maxNegLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent();
   maxNegLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
   maxNegLitSize = (*(lits.Last()))->size(); 
   maxPosLitNumOfDiffVars = 0;
   maxPosLitNumOfVarOccurences = 0;
   maxPosLitSize = 0;
  }
 else
  {
   maxNegLitNumOfDiffVars = 0;
   maxNegLitNumOfVarOccurences = 0; 
   maxNegLitSize = 0;
   maxPosLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent(); 
   maxPosLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
   maxPosLitSize = (*(lits.Last()))->size(); 
  }; 
 
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if (NegativeForSelection((*lit)->Header()))
    {
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars > maxNegLitNumOfDiffVars)
      {
       maxNegLitNumOfDiffVars = numOfDiffVars;
       maxNegLitNumOfVarOccurences = (*lit)->Variables().Size();
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         };
      }
     else
      if (numOfDiffVars < maxNegLitNumOfDiffVars)
       { 
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       }
      else // numOfDiffVars == maxNegLitNumOfDiffVars 
       {
        ulong numOfVarOccurences = (*lit)->Variables().Size();
        if (numOfVarOccurences > maxNegLitNumOfVarOccurences)
	 {
	  maxNegLitNumOfVarOccurences = numOfVarOccurences; 
          for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	   if (NegativeForSelection((*lt)->Header()))
	    {
             nonmax_lits.Push(*lt);
             lits.ReplaceByTop(lt);
            };
         }
        else
         if (numOfVarOccurences < maxNegLitNumOfVarOccurences)
	  {
           nonmax_lits.Push(*lit);
           lits.ReplaceByTop(lit);             
          }
         else // numOfVarOccurence ==  maxNegLitNumOfVarOccurences
	  {
	   ulong litSize = (*lit)->size();
           if (litSize > maxNegLitSize)
	    {
	     maxNegLitSize = litSize;
             for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	      if (NegativeForSelection((*lt)->Header()))
	       {
                nonmax_lits.Push(*lt);
                lits.ReplaceByTop(lt);
               }; 
            }            
           else
	    if (litSize < maxNegLitSize)
	     {
              nonmax_lits.Push(*lit);
              lits.ReplaceByTop(lit); 
             };
          };
       };
    }
   else // !NegativeForSelection((*lit)->Header())
    {
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars > maxPosLitNumOfDiffVars)
      {
       maxPosLitNumOfDiffVars = numOfDiffVars;
       maxPosLitNumOfVarOccurences = (*lit)->Variables().Size();
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (!NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         };
      }
     else
      if (numOfDiffVars < maxPosLitNumOfDiffVars)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       }
      else // numOfDiffVars == maxPosLitNumOfDiffVars 
       {
        ulong numOfVarOccurences = (*lit)->Variables().Size();
        if (numOfVarOccurences > maxPosLitNumOfVarOccurences)
	 {
	  maxPosLitNumOfVarOccurences = numOfVarOccurences; 
          for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	   if (!NegativeForSelection((*lt)->Header()))
	    {
             nonmax_lits.Push(*lt);
             lits.ReplaceByTop(lt);
            };
         }
        else
         if (numOfVarOccurences < maxPosLitNumOfVarOccurences)
	  {
           nonmax_lits.Push(*lit);
           lits.ReplaceByTop(lit);             
          }
         else // numOfVarOccurence == maxPosLitNumOfVarOccurences 
	  {
	   ulong litSize = (*lit)->size();
           if (litSize > maxPosLitSize)
	    {
	     maxPosLitSize = litSize;
             for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	      if (!NegativeForSelection((*lt)->Header()))
	       {
                nonmax_lits.Push(*lt);
                lits.ReplaceByTop(lt);
               }; 
            }            
           else
	    if (litSize < maxPosLitSize)
	     {
              nonmax_lits.Push(*lit);
              lits.ReplaceByTop(lit); 
             };
          };
       };
    };
  }; // for (..
 // all negative literals are maximally nonground and have the same size
 // all positive literals are maximally nonground and have the same size
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 

 //DF; lits.output(cout << "LITS1\n") << "\n";  
 //DF; max_lits.output(cout << "MAX1\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX1\n") << "\n";


  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
 bool positiveHasBeenSelected = false;
 bool negativeHasBeenSelected = false;
 while (lits.Nonempty())
  {
   TMP_LITERAL* lit = lits.Pop();
   if (NegativeForSelection(lit->Header()))
    {
     if (negativeHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       negativeHasBeenSelected = true;
       max_lits.Push(lit);
      };
    }
   else 
    {
     if (positiveHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       positiveHasBeenSelected = true;
       max_lits.Push(lit);
      };
    };
  };

 //DF; max_lits.output(cout << "MAX2\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX2\n") << "\n";

 
}; // void NewClause::select1006(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)


void NewClause::select1007(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1007(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
  Bidirectional selection. 
   If lits contains some inherently selected literals, only they are selected.
  In a rule, i.e. a clause containing both 
  positive and negative literals we always select one positive and one negative
  literal. If we have several literals of the same polarity, we select 
  those with maximal size. In a fact, i.e. a clause with all literals 
  of the same polarity, we select one maximally nonground literal.     
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif  
 ASSERT(lits.Length() > 1);
    
 if (stat.inherently_selected) 
  {
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };  

 //DF; lits.output(cout << "LITS\n") << "\n";

 ulong maxNegLitSize;
 ulong maxPosLitSize;
 if (NegativeForSelection((*(lits.Last()))->Header())) 
  {
   maxNegLitSize = (*(lits.Last()))->size(); 
   maxPosLitSize = 0;
  }
 else
  {
   maxNegLitSize = 0;
   maxPosLitSize = (*(lits.Last()))->size(); 
  }; 
 
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if (NegativeForSelection((*lit)->Header()))
    {
     ulong litSize = (*lit)->size();
     if (litSize > maxNegLitSize)
      {
       maxNegLitSize = litSize;
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         }; 
      }            
     else
      if (litSize < maxNegLitSize)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit); 
       };
    }
   else // !NegativeForSelection((*lit)->Header())
    {
     ulong litSize = (*lit)->size();
     if (litSize > maxPosLitSize)
      {
       maxPosLitSize = litSize;
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (!NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         }; 
      }            
     else
      if (litSize < maxPosLitSize)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit); 
       };
    };
  }; // for (..
 // all negative literals have maximal size
 // all positive literals have maximal size
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 

 //DF; lits.output(cout << "LITS1\n") << "\n";  
 //DF; max_lits.output(cout << "MAX1\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX1\n") << "\n";


  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
 bool positiveHasBeenSelected = false;
 bool negativeHasBeenSelected = false;
 while (lits.Nonempty())
  {
   TMP_LITERAL* lit = lits.Pop();
   if (NegativeForSelection(lit->Header()))
    {
     if (negativeHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       negativeHasBeenSelected = true;
       max_lits.Push(lit);
      };
    }
   else 
    {
     if (positiveHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       positiveHasBeenSelected = true;
       max_lits.Push(lit);
      };
    };
  };

 //DF; max_lits.output(cout << "MAX2\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX2\n") << "\n";

}; // void NewClause::select1007(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)


void NewClause::select1008(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1008(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
  Bidirectional selection. 
   If lits contains some inherently selected literals, only they are selected.
  In a rule, i.e. a clause containing both 
  positive and negative literals we always select one positive and one negative
  literal. If we have several literals of the same polarity, we select 
  those that are maximally ground. In a fact, i.e. a clause with all literals 
  of the same polarity, we select one maximally nonground literal.     
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif  
 ASSERT(lits.Length() > 1);
   
 if (stat.inherently_selected) 
  {
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };   

 //DF; lits.output(cout << "LITS\n") << "\n";

 ulong minNegLitNumOfDiffVars;
 ulong minNegLitNumOfVarOccurences;
 ulong maxNegLitSize;
 ulong minPosLitNumOfDiffVars;
 ulong minPosLitNumOfVarOccurences;
 ulong maxPosLitSize;
 if (NegativeForSelection((*(lits.Last()))->Header())) 
  {
   minNegLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent();
   minNegLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
   maxNegLitSize = (*(lits.Last()))->size(); 
   minPosLitNumOfDiffVars = HellConst::MaxNumOfVariables;
   minPosLitNumOfVarOccurences = HellConst::MaxTermSize;
   maxPosLitSize = 0;
  }
 else
  {
   minNegLitNumOfDiffVars = HellConst::MaxNumOfVariables;
   minNegLitNumOfVarOccurences = HellConst::MaxTermSize; 
   maxNegLitSize = 0;
   minPosLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent(); 
   minPosLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
   maxPosLitSize = (*(lits.Last()))->size(); 
  }; 
 
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if (NegativeForSelection((*lit)->Header()))
    {
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars < minNegLitNumOfDiffVars)
      {
       minNegLitNumOfDiffVars = numOfDiffVars;
       minNegLitNumOfVarOccurences = (*lit)->Variables().Size();
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         };
      }
     else
      if (numOfDiffVars > minNegLitNumOfDiffVars)
       { 
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       }
      else // numOfDiffVars == minNegLitNumOfDiffVars 
       {
        ulong numOfVarOccurences = (*lit)->Variables().Size();
        if (numOfVarOccurences < minNegLitNumOfVarOccurences)
	 {
	  minNegLitNumOfVarOccurences = numOfVarOccurences; 
          for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	   if (NegativeForSelection((*lt)->Header()))
	    {
             nonmax_lits.Push(*lt);
             lits.ReplaceByTop(lt);
            };
         }
        else
         if (numOfVarOccurences > minNegLitNumOfVarOccurences)
	  {
           nonmax_lits.Push(*lit);
           lits.ReplaceByTop(lit);             
          }
         else // numOfVarOccurence ==  minNegLitNumOfVarOccurences
	  {
	   ulong litSize = (*lit)->size();
           if (litSize > maxNegLitSize)
	    {
	     maxNegLitSize = litSize;
             for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	      if (NegativeForSelection((*lt)->Header()))
	       {
                nonmax_lits.Push(*lt);
                lits.ReplaceByTop(lt);
               }; 
            }            
           else
	    if (litSize < maxNegLitSize)
	     {
              nonmax_lits.Push(*lit);
              lits.ReplaceByTop(lit); 
             };
          };
       };
    }
   else // !NegativeForSelection((*lit)->Header())
    {
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars < minPosLitNumOfDiffVars)
      {
       minPosLitNumOfDiffVars = numOfDiffVars;
       minPosLitNumOfVarOccurences = (*lit)->Variables().Size();
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (!NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         };
      }
     else
      if (numOfDiffVars > minPosLitNumOfDiffVars)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       }
      else // numOfDiffVars == minPosLitNumOfDiffVars 
       {
        ulong numOfVarOccurences = (*lit)->Variables().Size();
        if (numOfVarOccurences < minPosLitNumOfVarOccurences)
	 {
	  minPosLitNumOfVarOccurences = numOfVarOccurences; 
          for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	   if (!NegativeForSelection((*lt)->Header()))
	    {
             nonmax_lits.Push(*lt);
             lits.ReplaceByTop(lt);
            };
         }
        else
         if (numOfVarOccurences > minPosLitNumOfVarOccurences)
	  {
           nonmax_lits.Push(*lit);
           lits.ReplaceByTop(lit);             
          }
         else // numOfVarOccurence == minPosLitNumOfVarOccurences 
	  {
	   ulong litSize = (*lit)->size();
           if (litSize > maxPosLitSize)
	    {
	     maxPosLitSize = litSize;
             for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	      if (!NegativeForSelection((*lt)->Header()))
	       {
                nonmax_lits.Push(*lt);
                lits.ReplaceByTop(lt);
               }; 
            }            
           else
	    if (litSize < maxPosLitSize)
	     {
              nonmax_lits.Push(*lit);
              lits.ReplaceByTop(lit); 
             };
          };
       };
    };
  }; // for (..
 // all negative literals are maximally ground and have the same size
 // all positive literals are maximally ground and have the same size
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 

 //DF; lits.output(cout << "LITS1\n") << "\n";  
 //DF; max_lits.output(cout << "MAX1\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX1\n") << "\n";


  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
 bool positiveHasBeenSelected = false;
 bool negativeHasBeenSelected = false;
 while (lits.Nonempty())
  {
   TMP_LITERAL* lit = lits.Pop();
   if (NegativeForSelection(lit->Header()))
    {
     if (negativeHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       negativeHasBeenSelected = true;
       max_lits.Push(lit);
      };
    }
   else 
    {
     if (positiveHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       positiveHasBeenSelected = true;
       max_lits.Push(lit);
      };
    };
  };

 //DF; max_lits.output(cout << "MAX2\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX2\n") << "\n";

}; // void NewClause::select1008(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)


void NewClause::select1009(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)
{ 
 CALL("select1009(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)");
 /*
  Bidirectional selection. 
   If lits contains some inherently selected literals, only they are selected.
  In a rule, i.e. a clause containing both 
  positive and negative literals we always select one positive and one negative
  literal. If we have several positive literals, we select 
  those that are maximally nonground. If we have several negative literals, 
  we select those that are maximally ground. 
  In a positive fact, i.e. a clause with only positive literals, 
  we select one maximally nonground literal. 
  In a negative fact, i.e. a clause with only negative literals, 
  we select one maximally ground literal.           
 */

 #ifdef DEBUG_NAMESPACE
  for (TMP_LITERAL** debugLit = lits.Last(); debugLit >= lits.First(); debugLit--)
   {
    ASSERT(!(*debugLit)->Header().IsSplitting());
   };
 #endif  
 ASSERT(lits.Length() > 1);
    
 if (stat.inherently_selected) 
  {
   selectInherentlySelected(lits,stat.inherently_selected,max_lits,nonmax_lits);
   return;  
  };
  
 //DF; lits.output(cout << "LITS\n") << "\n";

 ulong minNegLitNumOfDiffVars;
 ulong minNegLitNumOfVarOccurences;
 ulong maxNegLitSize;
 ulong maxPosLitNumOfDiffVars;
 ulong maxPosLitNumOfVarOccurences;
 ulong maxPosLitSize;
 if (NegativeForSelection((*(lits.Last()))->Header())) 
  {
   minNegLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent();
   minNegLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
   maxNegLitSize = (*(lits.Last()))->size(); 
   maxPosLitNumOfDiffVars = 0;
   maxPosLitNumOfVarOccurences = 0;
   maxPosLitSize = 0;
  }
 else
  {
   minNegLitNumOfDiffVars = HellConst::MaxNumOfVariables;
   minNegLitNumOfVarOccurences = HellConst::MaxTermSize; 
   maxNegLitSize = 0;
   maxPosLitNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent(); 
   maxPosLitNumOfVarOccurences = (*(lits.Last()))->Variables().Size(); 
   maxPosLitSize = (*(lits.Last()))->size(); 
  }; 
 
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   if (NegativeForSelection((*lit)->Header()))
    {
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars < minNegLitNumOfDiffVars)
      {
       minNegLitNumOfDiffVars = numOfDiffVars;
       minNegLitNumOfVarOccurences = (*lit)->Variables().Size();
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         };
      }
     else
      if (numOfDiffVars > minNegLitNumOfDiffVars)
       { 
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       }
      else // numOfDiffVars == minNegLitNumOfDiffVars 
       {
        ulong numOfVarOccurences = (*lit)->Variables().Size();
        if (numOfVarOccurences < minNegLitNumOfVarOccurences)
	 {
	  minNegLitNumOfVarOccurences = numOfVarOccurences; 
          for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	   if (NegativeForSelection((*lt)->Header()))
	    {
             nonmax_lits.Push(*lt);
             lits.ReplaceByTop(lt);
            };
         }
        else
         if (numOfVarOccurences > minNegLitNumOfVarOccurences)
	  {
           nonmax_lits.Push(*lit);
           lits.ReplaceByTop(lit);             
          }
         else // numOfVarOccurence ==  minNegLitNumOfVarOccurences
	  {
	   ulong litSize = (*lit)->size();
           if (litSize > maxNegLitSize)
	    {
	     maxNegLitSize = litSize;
             for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	      if (NegativeForSelection((*lt)->Header()))
	       {
                nonmax_lits.Push(*lt);
                lits.ReplaceByTop(lt);
               }; 
            }            
           else
	    if (litSize < maxNegLitSize)
	     {
              nonmax_lits.Push(*lit);
              lits.ReplaceByTop(lit); 
             };
          };
       };
    }
   else // !NegativeForSelection((*lit)->Header())
    {
     ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
     if (numOfDiffVars > maxPosLitNumOfDiffVars)
      {
       maxPosLitNumOfDiffVars = numOfDiffVars;
       maxPosLitNumOfVarOccurences = (*lit)->Variables().Size();
       for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	if (!NegativeForSelection((*lt)->Header()))
	 {
          nonmax_lits.Push(*lt);
          lits.ReplaceByTop(lt);
         };
      }
     else
      if (numOfDiffVars < maxPosLitNumOfDiffVars)
       {
        nonmax_lits.Push(*lit);
        lits.ReplaceByTop(lit);
       }
      else // numOfDiffVars == maxPosLitNumOfDiffVars 
       {
        ulong numOfVarOccurences = (*lit)->Variables().Size();
        if (numOfVarOccurences > maxPosLitNumOfVarOccurences)
	 {
	  maxPosLitNumOfVarOccurences = numOfVarOccurences; 
          for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	   if (!NegativeForSelection((*lt)->Header()))
	    {
             nonmax_lits.Push(*lt);
             lits.ReplaceByTop(lt);
            };
         }
        else
         if (numOfVarOccurences < maxPosLitNumOfVarOccurences)
	  {
           nonmax_lits.Push(*lit);
           lits.ReplaceByTop(lit);             
          }
         else // numOfVarOccurence == maxPosLitNumOfVarOccurences 
	  {
	   ulong litSize = (*lit)->size();
           if (litSize > maxPosLitSize)
	    {
	     maxPosLitSize = litSize;
             for (TMP_LITERAL** lt = lits.Last(); lt > lit; lt--)
	      if (!NegativeForSelection((*lt)->Header()))
	       {
                nonmax_lits.Push(*lt);
                lits.ReplaceByTop(lt);
               }; 
            }            
           else
	    if (litSize < maxPosLitSize)
	     {
              nonmax_lits.Push(*lit);
              lits.ReplaceByTop(lit); 
             };
          };
       };
    };
  }; // for (..
 // all negative literals are maximally nonground and have the same size
 // all positive literals are maximally nonground and have the same size
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 

 //DF; lits.output(cout << "LITS1\n") << "\n";  
 //DF; max_lits.output(cout << "MAX1\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX1\n") << "\n";


  // select 1 positive, if exists, and 1 negative, if exists, on "don't care" basis 
 bool positiveHasBeenSelected = false;
 bool negativeHasBeenSelected = false;
 while (lits.Nonempty())
  {
   TMP_LITERAL* lit = lits.Pop();
   if (NegativeForSelection(lit->Header()))
    {
     if (negativeHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       negativeHasBeenSelected = true;
       max_lits.Push(lit);
      };
    }
   else 
    {
     if (positiveHasBeenSelected) { nonmax_lits.Push(lit); }
     else 
      {
       positiveHasBeenSelected = true;
       max_lits.Push(lit);
      };
    };
  };

 //DF; max_lits.output(cout << "MAX2\n") << "\n";
 //DF; nonmax_lits.output(cout << "NONMAX2\n") << "\n";
 
}; // void NewClause::select1009(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat)

//=================================================


