//
// File:         Logging.h
// Description:  Various utilities for logging data.
// Created:      Dec 05, 2000, 16:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef LOGGING_H
//====================================================
#define LOGGING_H
#include <iostream.h>
#include <fstream.h>
#include "jargon.h"
#include "HellDebugFlags.h"     
#include "term.h"
#include "clause.h"
#include "Flatterm.h"   
#include "variables.h"         
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LOGGING
 #define DEBUG_NAMESPACE "Logging"
#endif
#include "debugMacros.h"
//=================================================
class Logging
{
 public:
  Logging() : _blocked(true), _counter(0L), _varRenaming() {};
  ~Logging() {};
  void initStatic() {};
  bool open(const char* fileName) 
  { 
   _stream.open(fileName,ios::app);
   return (bool)_stream;
  };
  void flush() { _stream.flush(); };
  long counter() const { return _counter; };
  long& counter() { return _counter; }; 
  void unblock() { _blocked = false; };
  void output(const TERM* t){ output(_stream,t); };
  void output(const Flatterm* t) { output(_stream,t); }; 
  void output(const Clause* cl) { output(_stream,cl); }; 
  void output(const char* str) { if (!_blocked) { _stream << str; _counter += strlen(str); }; }; 
  void output(int num)
  {
   if (!_blocked) 
    {
     _stream << num;
     _counter++; // the last digit in num
     if (num < 0) 
      {
       _counter++; // the minus
       num = (-1)*num;
      };
     // now count how many digits were printed 
     skip_digit:  
      num = num/10;
      if (num) 
       {
	// one more digit 
	_counter++;      
        goto skip_digit;
       };
    }; 
  };
  void output(long num)
  {
   if (!_blocked) 
    {
     _stream << num;
     _counter++; // the last digit in num
     if (num < 0) 
      {
       _counter++; // the minus
       num = (-1)*num;
      };
     // now count how many digits were printed 
     skip_digit:  
      num = num/10;
      if (num) 
       {
	// one more digit 
	_counter++;      
        goto skip_digit;
       };
    }; 
  };
  void output(ulong num)
  {
   if (!_blocked) 
    {
     _stream << num;
     _counter++; // the last digit in num
     // now count how many digits were printed 
     skip_digit:  
      num = num/10;
      if (num) 
       {
	// one more digit 
	_counter++;      
        goto skip_digit;
       };
    }; 
  };

  void outputSigSym(ulong sym,ulong arity) 
  {
   if (!_blocked) 
    {
     outputFun(_stream,sym);    
     _stream << "/" << arity;
     _counter+= 2; // "/" and the last digit in arity   
     // count how many digits are needed to print arity
     skip_digit: 
      arity = arity/10;
      if (arity)
       {
	// one more digit
        _counter++;
        goto skip_digit;
       };
    };  
  };

  void outputSigSym1(ulong sym,ulong arity,bool commutative) 
  {
   if (!_blocked) 
    {
     outputFun(_stream,sym);    
     _stream << "/" << arity;
     _counter+= 2; // "/" and the last digit in arity   
     if (commutative) { _stream << " c";  _counter+= 2; };
     // count how many digits are needed to print arity
     skip_digit: 
      arity = arity/10;
      if (arity)
       {
	// one more digit
        _counter++;
        goto skip_digit;
       };
    };  
  };
 
  void outputSignature(SIGNATURE& sig) 
  {
   for (ulong sym = HellConst::FirstNoneqFun; sym <= sig.MaxFunNum(); sym++)
    { 
     outputSigSym(sym,sig.Arity(sym)); 
     output("\n"); 
    }; 
  };
  void outputSignature1(SIGNATURE& sig) // another format 
  {
   outputSigSym1(HellConst::UnordEqNum,sig.Arity(HellConst::UnordEqNum),true); // equality
   output("\n");  
   for (ulong sym = HellConst::FirstNoneqFun; sym <= sig.MaxFunNum(); sym++)
    { 
     outputSigSym1(sym,sig.Arity(sym),sig.IsCommutative(sym)); 
     output("\n"); 
    }; 
  };
 private:
  ostream& outputFun(ostream& str,ulong fun);
  ostream& outputLongFun(ostream& str,ulong fun);
  ostream& outputVar(ostream& str,ulong var);
  ostream& output(ostream& str,const TERM* t);
  ostream& output(ostream& str,const Flatterm* t); 
  ostream& output(ostream& str,const Clause* t); 
 private:
  ofstream _stream; 
  bool _blocked;
  long _counter;
  INT_VAR_RENAMING _varRenaming;
}; // class Logging

inline ostream& Logging::outputFun(ostream& str,ulong fun)
{
 if (fun < 159) { _counter++; return str << ((char)(97 + fun)); };
 _counter += 2; // for the two "&"
 return outputLongFun(str << "&",fun) << "&";  
};

inline ostream& Logging::outputLongFun(ostream& str,ulong fun)
{
 _counter++;
 if (fun < 159) return str << ((char)(97 + fun)); 
 char md = (char)(fun % 159);
 ulong dv = fun / 159;
 return outputLongFun(str,dv) << (char)(97 + md);    
};

inline ostream&  Logging::outputVar(ostream& str,ulong var)
{
 var = _varRenaming.Rename(var);
 _counter++; // for the last digit in var
 if (var < 10) { return str << var; }
 else 
   if (var < 36)  
    return str << (char)(65 + var - 10);
 _counter+= 2; // for the two "%"
 // count how many decimal digits are needed to print var
 ulong tmp = var;
 skip_digit:
  tmp = tmp/10;
  if (tmp) 
   {
    // one more digit
    _counter++;
    goto skip_digit;
   };
  return str << "%" << var << "%";
};

inline ostream& Logging::output(ostream& str,const TERM* t)
{
 _varRenaming.Reset();
 if (_blocked) return str;
 TERM::Iterator iter(t);
 do
  {
   if (iter.symbolRef().isVariable())     
    {
     outputVar(str,iter.symbolRef().var());
    }
   else
    outputFun(str,iter.symbolRef().functor());
  }
 while (iter.Next()); 
 return str; 
}; 

inline ostream& Logging::output(ostream& str,const Flatterm* t)
{
 _varRenaming.Reset();
 if (_blocked) return str;
 Flatterm::Iterator iter(t);
 do
  {
   if (iter.CurrPos()->isVariable())     
    {
     outputVar(str,iter.CurrPos()->Var());
    }
   else
    outputFun(str,iter.CurrPos()->functor());
   iter.NextSym();
  }
 while (iter.NotEndOfTerm()); 
 return str; 
}; 

inline ostream& Logging::output(ostream& str,const Clause* cl)
{
 CALL("output(ostream& str,const Clause* cl)");
 _varRenaming.Reset();
 if (_blocked) return str;

 Clause::Traversal traversal(cl);
 ulong func;
 check_state:
  switch (traversal.state())
   {
    case Clause::Traversal::EndOfClause: break;
    case Clause::Traversal::LitHeader: 
     if (traversal.symbol().Positive())
      {
       output("+");
      }
     else output("-");
     func = traversal.symbol().functor();
     if (traversal.symbol().IsEquality()) func = HellConst::UnordEqNum; 
     outputFun(str,func); 
     traversal.next();   
     goto check_state;

    case Clause::Traversal::Func: 
     outputFun(str,traversal.symbol().functor());
     traversal.next();   
     goto check_state;

    case Clause::Traversal::Var:
     outputVar(str,traversal.symbol().var());
     traversal.next();   
     goto check_state;
   };
 return str; 
}; 

//====================================================
#undef DEBUG_NAMESPACE  
//====================================================
#endif
