#include "NewClause.h"
#include "HellDebugFlags.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEW_CLAUSE
 #define DEBUG_NAMESPACE "NewClause"
#endif 
#include "debugMacros.h"
//=================================================

bool NewClause::_inversePolarityForSelection;
long NewClause::_literalComparisonMode;

bool NewClause::transferToPagedPassive(PagedPassiveClauseStorage& storage)
{
 CALL("NewClause::transferToPagedPassive(PagedPassiveClauseStorage& storage)");
 _openClause.EndOfClause();
 ASSERT(!_openClause.AllLiterals().Empty());
 ulong size = _openClause.Weight();
 BitWord rules = _ancestors.Rules();
 ulong numOfLits = _openClause.Length();
 ulong numOfAncestors = _ancestors.numberOfAncestors();
 if (!storage.openClause(_clauseNumber,size,rules,numOfLits,numOfAncestors))
  return false; 

 // now write the literals 
 writeIntoPagedPassiveClauseStorage(storage,_openClause.AllLiterals());
 
 // now write the ancestors
  
 for (Clause* const * anc = _ancestors.begin(); anc < _ancestors.end(); anc++)
  storage.writeAncestor(*anc); 

 storage.endOfClause();
 return true;
}; // bool NewClause::transferToPagedPassive(PagedPassiveClauseStorage& storage)

void NewClause::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits)
{
 CALL("writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits)"); 

 for (ulong ln = 0; ln < lits.Length(); ln++)
  { 
   TMP_LITERAL* lit = lits.Nth(ln);
   writeIntoPagedPassiveClauseStorage(storage,lit);      
  };

}; // void NewClause::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits)

void NewClause::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LITERAL* lit)
{
 CALL("writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LITERAL* lit)");
 storage.openLiteral(lit->isInherentlySelected());
 TERM header = lit->Header();
 Flatterm::Iterator iter;
 if (header.IsEquality()) 
  {
   storage.writeSymbol(header);
   const Flatterm* arg1;
   const Flatterm* arg2;
   if (lit->EqSwapArgs()) 
    {
     arg1 = lit->Arg2();
     arg2 = lit->Arg1();
    }
   else 
    {
     arg1 = lit->Arg1();
     arg2 = lit->Arg2();
    };
   iter.Reset(arg1);
   do
    {
     storage.writeSymbol(iter.CurrPos()->symbolRef());          
     iter.NextSym(); 
    }
   while (iter.NotEndOfTerm());
   iter.Reset(arg2);
   do
    {
     storage.writeSymbol(iter.CurrPos()->symbolRef());          
     iter.NextSym();
    }
   while (iter.NotEndOfTerm());
  }
 else 
  {
   iter.Reset(lit->FlatTerm());
   do
    {
     storage.writeSymbol(iter.CurrPos()->symbolRef());          
     iter.NextSym();
    }
   while (iter.NotEndOfTerm());
  };
 storage.closeLiteral(); 
};  // void NewClause::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LITERAL* lit)

Clause* NewClause::Reanimate()
{
 CALL("Reanimate()");
 _openClause.EndOfClause();

 ASSERT(!_openClause.AllLiterals().Empty());
  
 if (_forwardSubsumptionFlag) 
  if (FSubsume()) 
   {
    AnotherForwardSubsumed();
    return (Clause*)0;
   };
  
 // simplify the new clause here
  
 bool redundant;
 Simplify(redundant);
 if (redundant) return (Clause*)0;
 if (_openClause.AllLiterals().Empty()) return MakeEmptyClause();

 if (_openClause.AllLiterals().Length() == 1) return MakeUnitClause(_openClause.AllLiterals().Pop());
  
 return MakeSharedClause(_openClause.AllLiterals());
}; // Clause* NewClause::Reanimate()


Clause* NewClause::checkForUnitConflict()
{
 CALL("checkForUnitConflict()");
 _openClause.EndOfClause();

 if (_openClause.AllLiterals().Length() == 1)  
  {  
   if (_unitConflict)
    { 
     const TMP_LITERAL* lit = *(_openClause.AllLiterals().First());
     Clause* uc_rule = _unitConflict->checkForUnitConflict(lit->FlatTerm());
     if (uc_rule) 
      {
       UsedRule(ClauseBackground::BinRes);         
       PushAncestor(uc_rule);
       Clause* newClause = MakeEmptyClause();
       newClause->setMainSet(Clause::MainSetNew);  
       return newClause;
      };
    };
  };
 return (Clause*)0;
}; // Clause* NewClause::checkForUnitConflict()

void NewClause::ProcessCollected3(Clause*& emptyClause,
                                  PagedPassiveClauseStorage& storage,
                                  ulong& numOfNewClauses,
                                  ulong& numOfRejectedNewClauses,
                                  ulong (*nextClauseNumFunc)(),
                                  bool showNewClauses)
 {  
  CALL("ProcessCollected3(..)");
  _openClause.EndOfClause();
  emptyClause = (Clause*)0;
  numOfNewClauses = 0;
  numOfRejectedNewClauses = 0;
  if (_openClause.AllLiterals().Empty()) 
   {
    emptyClause = MakeEmptyClause();
    emptyClause->SetNumber(nextClauseNumFunc());
    emptyClause->setMainSet(Clause::MainSetNew);
    numOfNewClauses++;
    if (showNewClauses) cout << "% " << emptyClause << "\n";
    return; 
   };
  
  if (_forwardSubsumptionFlag) 
   if (FSubsume()) 
    {
     AnotherForwardSubsumed();
     return;
    };
  
  // simplify the new clause here
  
  bool redundant;
  Simplify(redundant);
  if (redundant) return;
  if (_openClause.AllLiterals().Empty()) 
   {    
    emptyClause = MakeEmptyClause();
    emptyClause->SetNumber(nextClauseNumFunc());
    emptyClause->setMainSet(Clause::MainSetNew);
    numOfNewClauses++;
    if (showNewClauses) cout << "% " << emptyClause << "\n";
    return; 
   };
  
  if (_openClause.AllLiterals().Length() == 1)  
   {
    // try unit conflict
    if (_unitConflict)
     { 
      const TMP_LITERAL* lit = *(_openClause.AllLiterals().First());
      Clause* uc_rule = _unitConflict->checkForUnitConflict(lit->FlatTerm());
      if (uc_rule) 
       {
        UsedRule(ClauseBackground::BinRes);         
        PushAncestor(uc_rule);
       
        emptyClause = MakeEmptyClause();
        emptyClause->SetNumber(nextClauseNumFunc());
        emptyClause->setMainSet(Clause::MainSetNew);
        numOfNewClauses++;
        if (showNewClauses) cout << "% " << emptyClause << "\n";
        return;    
       };
     };
   };

  if (_ancestors.singletone() && (*(_ancestors.begin()))->IsSplitName())
   {
    // inherited property
    UsedRule(ClauseBackground::Split);
    UsedRule(ClauseBackground::Name);
   } 
  else
   if (_splittingFlag && (Split3(_openClause.AllLiterals(),
                                 emptyClause,
                                 storage,
                                 numOfNewClauses,
                                 numOfRejectedNewClauses,
                                 nextClauseNumFunc,
                                 showNewClauses)))
    { 
     return;
    };  


  if (makeClauseInPagedPassiveClauseStorage(storage,_openClause.AllLiterals(),nextClauseNumFunc(),showNewClauses,false))
   {
    numOfNewClauses++;                  
   }
  else
   {
    numOfRejectedNewClauses++;
   };
 }; // void NewClause::ProcessCollected3(Clause*& emptyClause,...


void NewClause::ProcessCollected2(ClauseQueue& newClauses) 
 {  
  CALL("ProcessCollected2()");
  _openClause.EndOfClause();

  if (_openClause.AllLiterals().Empty()) 
   {
    Clause* newClause = MakeEmptyClause();
    newClause->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(newClause); 
    return; 
   };
  
  if (_forwardSubsumptionFlag) 
   if (FSubsume()) 
    {
     AnotherForwardSubsumed();
     return;
    };
  
  // simplify the new clause here
  
  bool redundant;
  Simplify(redundant);
  if (redundant) return;
  if (_openClause.AllLiterals().Empty()) 
   {
    Clause* newClause = MakeEmptyClause();
    newClause->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(newClause);  
    return;  
   };
  
  if (_openClause.AllLiterals().Length() == 1)  
   {
    // try unit conflict
    if (_unitConflict)
     { 
      const TMP_LITERAL* lit = *(_openClause.AllLiterals().First());
      Clause* uc_rule = _unitConflict->checkForUnitConflict(lit->FlatTerm());
      if (uc_rule) 
       {
        UsedRule(ClauseBackground::BinRes);         
        PushAncestor(uc_rule);
        Clause* newClause = MakeEmptyClause();
        newClause->setMainSet(Clause::MainSetNew);
        newClauses.enqueue(newClause);  
        return;    
       };
     };
   };

  if (_ancestors.singletone() && (*(_ancestors.begin()))->IsSplitName())
   {
    // inherited property
    UsedRule(ClauseBackground::Split);
    UsedRule(ClauseBackground::Name);
   } 
  else
   if (_splittingFlag && (Split2(_openClause.AllLiterals(),newClauses)))
    { 
     return;
    };  

  Clause* cl = MakeFastUnsharedClause(_openClause.AllLiterals());
  cl->setMainSet(Clause::MainSetNew);
  newClauses.enqueue(cl);
  
 }; // void NewClause::ProcessCollected2(ClauseQueue& newClauses)  


void NewClause::ProcessCollected1(ClauseQueue& newClauses) 
 {  
  CALL("ProcessCollected1()");
  _openClause.EndOfClause();

  if (_openClause.AllLiterals().Empty()) 
   {
    Clause* newClause = MakeEmptyClause();
    newClause->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(newClause);  
    return; 
   };
  
  if (_forwardSubsumptionFlag) 
   if (FSubsume()) 
    {
     AnotherForwardSubsumed();
     return;
    };
  
  // simplify the new clause here
  
  bool redundant;
  Simplify(redundant);
  if (redundant) return;
  if (_openClause.AllLiterals().Empty()) 
   {
    Clause* newClause = MakeEmptyClause();
    newClause->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(newClause);  
    return;  
   };
  
  if (_negEqSplitting) 
   { 
    NegEqSplitting(newClauses);
    if (!_openClause.Weed()) return;
    if (_forwardSubsumptionFlag && FSubsume())
     { 
      AnotherForwardSubsumed();
      return;
     }; 
   };

  if (_openClause.AllLiterals().Length() == 1)  
   {
    // first try unit conflict
    if (_unitConflict)
     { 
      const TMP_LITERAL* lit = *(_openClause.AllLiterals().First());
      Clause* uc_rule = _unitConflict->checkForUnitConflict(lit->FlatTerm());
      if (uc_rule) 
       {
        UsedRule(ClauseBackground::BinRes);         
        PushAncestor(uc_rule);
        Clause* newClause = MakeEmptyClause();
        newClause->setMainSet(Clause::MainSetNew);
        newClauses.enqueue(newClause); 
        return;    
       };
     };

    Clause* unit_cl = MakeUnitClause(_openClause.AllLiterals().Pop());
    unit_cl->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(unit_cl); 
    return; 
   };
  
  if (_ancestors.singletone() && (*(_ancestors.begin()))->IsSplitName())
   {
    // inherited property
    UsedRule(ClauseBackground::Split);
    UsedRule(ClauseBackground::Name);
   } 
  else
   if (_splittingFlag && (Split(_openClause.AllLiterals(),newClauses)))
    { 
     return;
    };  

  Clause* cl = MakeSharedClause1(_openClause.AllLiterals());
  cl->setMainSet(Clause::MainSetNew);
  newClauses.enqueue(cl);
  
 }; // void NewClause::ProcessCollected1(ClauseQueue& newClauses)  



void NewClause::ProcessCollected(ClauseQueue& newClauses) 
 {  
  CALL("ProcessCollected()");
  _openClause.EndOfClause();

  if (_openClause.AllLiterals().Empty()) 
   {
    Clause* newClause = MakeEmptyClause();
    newClause->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(newClause);  
    return; 
   };
  
  if (_forwardSubsumptionFlag) 
   if (FSubsume()) 
    {
     AnotherForwardSubsumed();
     return;
    };
  
  // simplify the new clause here
  
  bool redundant;
  Simplify(redundant);
  if (redundant) return;
  if (_openClause.AllLiterals().Empty()) 
   {
    Clause* newClause = MakeEmptyClause();
    newClause->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(newClause);
    return;  
   };
  
  if (_negEqSplitting) 
   { 
    NegEqSplitting(newClauses);
    if (!_openClause.Weed()) return;
    if (_forwardSubsumptionFlag && FSubsume())
     { 
      AnotherForwardSubsumed();
      return;
     }; 
   };

  if (_openClause.AllLiterals().Length() == 1)  
   {
    // first try unit conflict
    if (_unitConflict)
     { 
      const TMP_LITERAL* lit = *(_openClause.AllLiterals().First());
      Clause* uc_rule = _unitConflict->checkForUnitConflict(lit->FlatTerm());
      if (uc_rule) 
       {
        UsedRule(ClauseBackground::BinRes);         
        PushAncestor(uc_rule);
        Clause* newClause = MakeEmptyClause();
        newClause->setMainSet(Clause::MainSetNew);
        newClauses.enqueue(newClause);
        return;    
       };
     };
    Clause* unit_cl = MakeUnitClause(_openClause.AllLiterals().Pop());
    unit_cl->setMainSet(Clause::MainSetNew);
    newClauses.enqueue(unit_cl); 
    return; 
   };
  
  if (_ancestors.singletone() && (*(_ancestors.begin()))->IsSplitName())
   {
    // inherited property
    UsedRule(ClauseBackground::Split);
    UsedRule(ClauseBackground::Name);
   } 
  else
   if (_splittingFlag && (Split(_openClause.AllLiterals(),newClauses)))
    { 
     return;
    };  

  Clause* cl = MakeSharedClause(_openClause.AllLiterals());
  cl->setMainSet(Clause::MainSetNew);
  newClauses.enqueue(cl);
  
 }; // void NewClause::ProcessCollected(ClauseQueue& newClauses)  

void NewClause::Simplify(bool& redundant)
{
 CALL("Simplify(bool& redundant)");
 redundant = false;
 if ((_forwardSubsumptionResolution) && (SimplifyByFS()))
  {
   UsedRule(ClauseBackground::ForwSubsRes);
   AnotherSimplifiedByFSR();
   MakeVIP();
   if (_openClause.AllLiterals().Empty()) return;
  };

 simplify:     
  if ((_commutativityTable) && (NormalizeByCommutativity()))
   { 
    UsedRule(ClauseBackground::ForwDemod);
    AnotherSimplifiedByDemod();
    MakeVIP();
    if ((redundant = (!_openClause.Weed()))) return;

    if (_openClause.AllLiterals().Empty()) return;
    if (_forwardSubsumptionFlag && ((redundant = FSubsume())))
     { 
      AnotherForwardSubsumed();
      return;
     }; 
   };

  if (_forwardDemodulation > 0)
   {
    demodulate:
     if (SimplifyByUEq())
      { 
       UsedRule(ClauseBackground::ForwDemod);
       AnotherSimplifiedByDemod();
       MakeVIP();

       if ((redundant = (!_openClause.Weed()))) return;

       if (_openClause.AllLiterals().Empty()) return;
       if (_forwardSubsumptionFlag && ((redundant = FSubsume())))
        { 
         AnotherForwardSubsumed();
         return;
        };
 
        if ((_commutativityTable) && (NormalizeByCommutativity()))
         { 
          UsedRule(ClauseBackground::ForwDemod);
          AnotherSimplifiedByDemod();
          MakeVIP();
          if ((redundant = (!_openClause.Weed()))) return;

          if (_openClause.AllLiterals().Empty()) return;
          if (_forwardSubsumptionFlag && ((redundant = FSubsume())))
           { 
            AnotherForwardSubsumed();
            return;
           };
          goto demodulate;  
         };
                       
      };     
   }; 

 if (_simplificationByEqRes)
  {
   if (SimplifyByEqRes())
    {
     UsedRule(ClauseBackground::EqResSimp);
     AnotherSimplifiedByEqRes();
     MakeVIP();
     if ((redundant = (!_openClause.Weed()))) return;

     if (_openClause.AllLiterals().Empty()) return;
     if (_forwardSubsumptionFlag && ((redundant = FSubsume())))
      { 
       AnotherForwardSubsumed();
       return;
      };       
     goto simplify;
    };  
  }; 

 if ((_forwardSubsumptionResolution) && (SimplifyByFS()))
  {
   UsedRule(ClauseBackground::ForwSubsRes);
   AnotherSimplifiedByFSR();
   MakeVIP();
   if (_openClause.AllLiterals().Empty()) return;
  };
}; // void NewClause::Simplify(bool& redundant)


bool NewClause::SimplifyByEqRes()
 {
  CALL("SimplifyByEqRes()");

  ulong var;
  const Flatterm* subst_term;
  unsigned subst_depth;
  long subst_weight;
  TMP_LITERAL* simplif_lit;
  TMP_LITERAL** lit = _openClause.AllLiterals().Last() + 1;   
  try_next_lit: 
   if (lit <= _openClause.AllLiterals().First()) return false; 
   lit--; 
   simplif_lit = *lit; 
   if (simplif_lit->Negative() && simplif_lit->IsEquality() && simplif_lit->IsUnorderedEq())
    {
     if (simplif_lit->Arg1()->Symbol().isVariable())
      {
       var = simplif_lit->Arg1()->Symbol().var();
       if (simplif_lit->Arg2Vars().Occurences(var)) goto try_next_lit;      
       subst_term = simplif_lit->Arg2();
      }
     else 
      if (simplif_lit->Arg2()->Symbol().isVariable())
       {
        var = simplif_lit->Arg2()->Symbol().var();
        if (simplif_lit->Arg1Vars().Occurences(var)) goto try_next_lit;
        subst_term = simplif_lit->Arg1();
       }
      else goto try_next_lit;
   
     _openClause.AllLiterals().ReplaceByTop(lit);  
     subst_depth = ((Flatterm*)subst_term)->Depth();
     subst_weight = subst_term->size();

     if (_openClause.AllLiterals().SubstPossible(var,subst_depth,subst_weight))
      { 
       for (TMP_LITERAL** old_lit = _openClause.AllLiterals().First();old_lit <= _openClause.AllLiterals().Last();old_lit++)
	{        	
         if ((*old_lit)->Variables().Occurences(var))
	  {
           (*old_lit)->Subst(*(_openClause.SpareLiteral()),var,(Flatterm*)subst_term);
           TMP_LITERAL* tmp = *old_lit;
           *old_lit = _openClause.SpareLiteral();
           _openClause.SpareLiteral() = tmp;
          };    
        };             
       return true; 
      }
     else 
      { 
       _openClause.AllLiterals().Push(simplif_lit);
       goto try_next_lit;
      };
    } 
   else goto try_next_lit;
 }; // bool NewClause::SimplifyByEqRes()


Clause* NewClause::RewriteByUEq(TMP_LITERAL* lit,TMP_LITERAL* new_lit)
 {
  CALL("RewriteByUEq(TMP_LITERAL* lit,TMP_LITERAL* new_lit)");
  TERM header = lit->Header();
  if (!header.arity()) return (Clause*)0; // propositional literal
  Flatterm::Iterator iter;
  iter.Reset(lit->FlatTerm());
  iter.NextSym(); // skipping the header
  const Flatterm* rewritten_subterm; 

  while (iter.NotEndOfTerm())
   {
    rewritten_subterm = iter.CurrPos();
    if (rewritten_subterm->Symbol().isComplex())
     {
      if (_forwardSimplificationByUEq->Rewrite(rewritten_subterm,_splittingBranches))
       {
	
        if (_forwardSimplificationByUEq->SplittingBranchesUsed()) 
	  AnotherFDOnSplittingBranches();    

        new_lit->Reset();
        if (header.IsEquality() && header.IsOrderedEq()) header.MakeUnordered();
        new_lit->PushNonpropHeader(header);        

        iter.Reset(lit->FlatTerm());
        iter.NextSym(); // skipping the header
        TERM sym;
        while (iter.NotEndOfTerm())
	 {
          if (iter.CurrPos() == rewritten_subterm)
	   {
            do
	     {
	      sym = _forwardSimplificationByUEq->RHSSymbol();
              if (sym.isVariable())
	       {
                new_lit->PushVar(sym);
	       }
              else // functor
	       {
                if (sym.arity())
		 {
                  if (!new_lit->PushNonconstFun(sym)) goto try_next_position;
                 }
                else // constant
		 {
                  new_lit->PushConst(sym);
                 };
               };
             }
            while (_forwardSimplificationByUEq->RHSNext());
            iter.after();
	   }
          else
	   {
            sym = iter.CurrPos()->Symbol();
        
            if (sym.isVariable())
	     {
              new_lit->PushVar(sym);
	     }
            else // functor
             {
              if (sym.arity())
      	       {
                if (!new_lit->PushNonconstFun(sym)) goto try_next_position;
               }
              else // constant
	       {
                new_lit->PushConst(sym);
               };
             };
        
            iter.NextSym();
           };
         };        

        Clause* simplifier = _forwardSimplificationByUEq->Rule();  
        ASSERT(simplifier->Active());  
        PushAncestor(simplifier);
        //simplifier->MarkAsSimplifier();
        simplifier->setFlag(Clause::FlagHasChildren);
        return simplifier;
       };
     };

    try_next_position:   
     iter.NextSym();
   };
  return (Clause*)0;
 }; // Clause* NewClause::RewriteByUEq(TMP_LITERAL* lit,TMP_LITERAL* new_lit)


bool NewClause::SimplifyByUEq(TMP_LITERAL*& lit,TMP_LITERAL*& spare_lit)
 {
  CALL("SimplifyByUEq(TMP_LITERAL*& lit,TMP_LITERAL*& spare_lit)");

  ulong num_of_rew = 0;
  Clause* simplifier;
  while ((!_ancestors.full())  && ((simplifier = RewriteByUEq(lit,spare_lit)))) // assignment, not a comparison 
   {
    RARE(_ancestors.full());
    ASSERT(((Flatterm*)(spare_lit->FlatTerm()))->CheckIntegrity());

    TMP_LITERAL* new_spare_lit = lit;
    lit = spare_lit;
    spare_lit = new_spare_lit;
    num_of_rew++;
   };      
  
  return (bool)num_of_rew;
 }; // bool NewClause::SimplifyByUEq(TMP_LITERAL*& lit,TMP_LITERAL*& spare_lit) 

bool NewClause::SimplifyByUEq()
 {  
  CALL("SimplifyByUEq()");
 
  if (_fDOnSplittingBranchesFlag) 
   {
    // collect _splittingBranches 
    _splittingBranches.reset();
   for (TMP_LITERAL** lit = _openClause.AllLiterals().First(); lit <= _openClause.AllLiterals().Last(); lit++) 
    {      
     if ((*lit)->Header().IsSplitting()) _splittingBranches.add((*lit)->Header().HeaderNum());
    };
   };  

  ulong num_of_simpl_lits = 0;
  for (TMP_LITERAL** lit = _openClause.AllLiterals().First(); lit <= _openClause.AllLiterals().Last(); lit++) 
   {
    if (SimplifyByUEq(*lit,_openClause.SpareLiteral()))
     { 
      num_of_simpl_lits++;
     }; 
   };
  return (bool)num_of_simpl_lits; 
 }; // bool NewClause::SimplifyByUEq()


Clause* NewClause::MakeEmptyClause()
 {
  CALL("MakeEmptyClause()");
  Clause* empty_clause = Clause::MakeEmptyClause();
  empty_clause->setSize(0);
  empty_clause->SetNumOfPosEq(0);
  empty_clause->setMinChildSize(0);
  empty_clause->SetAncestors(_ancestors.NewAncestors()); 
  return empty_clause;
 }; // Clause* NewClause::MakeEmptyClause()


Clause* NewClause::MakeUnitClause(TMP_LITERAL* lit)
 {
  CALL("MakeUnitClause(TMP_LITERAL* lit)");
  _intVarRenaming.Reset();
  lit->RenameVariables(_intVarRenaming);
  LiteralList* lit_list = new (1) LiteralList(1,1);
  Clause* result = new Clause(lit_list);
  lit_list->firstElem()->setLiteral(TmpLitSharedCopy(lit,result));
  lit->setInherentlySelected(_inheritedNegativeSelection 
                             && NegativeForSelection(lit->Header()));
   
  if (lit->isInherentlySelected())
   {
    lit_list->firstElem()->setMark(LiteralList::InherentlySelected);
   };
  result->SetAncestors(_ancestors.NewAncestors()); 
  result->setSize(lit->size());
  if (lit->IsEquality() && lit->Positive()) { result->SetNumOfPosEq(1); }
  else result->SetNumOfPosEq(0); 
  if (lit->IsEquality() && lit->Positive())
   { result->setMinChildSize(lit->PosEqMinLHSSize() + 1); }
  else result->setMinChildSize(0);
  result->setFlag(Clause::FlagInClauseIndex);
  result->setFlag(Clause::FlagLiteralsShared);
  return result;
 }; // Clause* NewClause::MakeUnitClause(Flatterm* lit)


Clause* NewClause::MakeSharedClause1(TMP_LIT_LIST& lits)
{
 CALL("MakeSharedClause1(TMP_LIT_LIST& lits)");
 if (lits.Empty()) return MakeEmptyClause();
 long weight = lits.Weight();

 ulong num_of_max = lits.Length(); // all literals are "maximal"
 _intVarRenaming.Reset();
 lits.RenameVariables(_intVarRenaming);

 LiteralList* lit_list = new(num_of_max) LiteralList(num_of_max,num_of_max);
 Clause* result = new Clause(lit_list);

 
 LiteralList::Iterator iter(lit_list);
 TERM* lit;
 do 
  {
   ASSERT(iter.notEnd());
   const TMP_LITERAL* tmp_lit = lits.Pop();
   lit = TmpLitSharedCopy(tmp_lit,result);
   iter.currentElement()->setLiteral(lit);
   if (tmp_lit->isInherentlySelected()) iter.currentElement()->setMark(LiteralList::InherentlySelected); 
   iter.next();
  }
 while (lits.Nonempty()); 

 ASSERT(iter.end());
           
 result->SetAncestors(_ancestors.NewAncestors()); 
 result->setSize(weight);
 result->SetNumOfPosEq(0);
 result->setMinChildSize(0);  
 result->setFlag(Clause::FlagInClauseIndex);
 result->setFlag(Clause::FlagLiteralsShared);
 return result; 
}; // Clause* NewClause::MakeSharedClause1(TMP_LIT_LIST& lits)


Clause* NewClause::MakeSharedClause(TMP_LIT_LIST& lits)
 {
  CALL("MakeSharedClause(TMP_LIT_LIST& lits)");
  if (lits.Empty()) return MakeEmptyClause();
  long weight;
  long num_of_pos_eq;
  if (_equalityPossible)
   {
    lits.WeightAndNumOfPosEq(weight,num_of_pos_eq);
   }
  else
   { 
    weight = lits.Weight();
    num_of_pos_eq = 0;
   }; 
   
  TMP_LIT_LIST max_lits;
  TMP_LIT_LIST nonmax_lits; 
  TMP_LIT_LIST tmp_lits;
  TMP_LIT_LIST* ordered_max_lits;
  TMP_LIT_LIST* ordered_nonmax_lits; 

  DOP(ulong debugLitsLength = lits.Length());

  bool negSelUsed = false;
  if (_predElimTab && lits.Eliminate(max_lits,nonmax_lits,_predElimTab))
   {
     // nothing to do here 
   }
  else // order literals as usual
   {
    select(lits,max_lits,nonmax_lits,negSelUsed);
    /*******
    if (negSelUsed)
     {
      ASSERT(max_lits.Length() == 1);
      //ASSERT((*(max_lits.First()))->Negative()); 
      ASSERT(NegativeForSelection((*(max_lits.First()))->Header())); 
      (*(max_lits.First()))->setInherentlySelected(_inheritedNegativeSelection);
     };
    **********/
   };
  // don't use lits anymore! it must be empty here!

  ASSERT(lits.Empty());
  ASSERT(!max_lits.Empty());  

  ulong num_of_max = max_lits.Length();
  ulong num_of_nonmax = nonmax_lits.Length();

  ASSERT(debugLitsLength == num_of_max + num_of_nonmax);

  if (num_of_max > 1)
   {
    lits.CreateVarWeighting(_varWeighting);
    max_lits.Order(tmp_lits,_varWeighting);
    ordered_max_lits = &tmp_lits;
    if (num_of_nonmax > 1) 
     {
      max_lits.Reset(); 
      nonmax_lits.Order(max_lits,_varWeighting);
      ordered_nonmax_lits = &max_lits; // this is not an error!
     }
    else ordered_nonmax_lits = &nonmax_lits;  
   }
  else
   {
    ordered_max_lits = &max_lits;
    if (num_of_nonmax > 1)
     {
      lits.CreateVarWeighting(_varWeighting);
      nonmax_lits.Order(tmp_lits,_varWeighting);
      ordered_nonmax_lits = &tmp_lits;
     }
    else ordered_nonmax_lits = &nonmax_lits;  
   };  
  
  _intVarRenaming.Reset();
  // order of renaming is significant!
  ordered_max_lits->RenameVariables(_intVarRenaming);
  ordered_nonmax_lits->RenameVariables(_intVarRenaming);
  
  long weight_of_nonmax = ordered_nonmax_lits->MinFactorWeightUpperBound();
  
  long min_weight_of_max_pos_eq_lhs = 0;
  if (_equalityPossible)
   {
    min_weight_of_max_pos_eq_lhs = ordered_max_lits->MinPosEqLHSWeight();
    if (min_weight_of_max_pos_eq_lhs) min_weight_of_max_pos_eq_lhs++;
   };
  
  LiteralList* lit_list = new(num_of_max + num_of_nonmax) LiteralList(num_of_max + num_of_nonmax,num_of_max);
  Clause* result = new Clause(lit_list);
 
  LiteralList::Iterator iter(lit_list);
  TERM* lit;
  TMP_LITERAL* tmp_lit;
  do
   {
    tmp_lit = ordered_max_lits->Pop();

    lit = TmpLitSharedCopy(tmp_lit,result);
    ASSERT(iter.notEnd());  
    iter.currentElement()->setLiteral(lit);
    tmp_lit->setInherentlySelected(_inheritedNegativeSelection
                                   && NegativeForSelection(tmp_lit->Header()));
    if (tmp_lit->isInherentlySelected())  iter.currentElement()->setMark(LiteralList::InherentlySelected);  
    iter.next();
    num_of_max--;
   }
  while (num_of_max); 
           
  while (num_of_nonmax)
   {
    tmp_lit = ordered_nonmax_lits->Pop();
    ASSERT(!tmp_lit->isInherentlySelected());
    lit = TmpLitSharedCopy(tmp_lit,result);
    ASSERT(iter.notEnd());  
    iter.currentElement()->setLiteral(lit);        
    iter.next();
    num_of_nonmax--;
   };

  ASSERT(iter.end());  

  if (negSelUsed) 
   { 
    _ancestors.UsedRule(ClauseBackground::NegSel);
    result->SetAncestors(_ancestors.NewAncestors());
    _ancestors.CancelRule(ClauseBackground::NegSel); 
   }
  else 
   result->SetAncestors(_ancestors.NewAncestors());
  result->setSize(weight);
  result->SetNumOfPosEq(num_of_pos_eq);
  result->setMinChildSize(weight_of_nonmax + min_weight_of_max_pos_eq_lhs);
  result->setFlag(Clause::FlagInClauseIndex);
  result->setFlag(Clause::FlagLiteralsShared);
  return result; 
 }; // Clause* NewClause::MakeSharedClause(TMP_LIT_LIST& lits)




Clause* NewClause::MakeFastUnsharedClause(TMP_LIT_LIST& lits)
{
 CALL("MakeFastUnsharedClause(TMP_LIT_LIST& lits)");
 ulong num_of_lits = lits.Length();
 ulong weight = lits.Weight();;
 LiteralList* lit_list = new (num_of_lits) LiteralList(num_of_lits,num_of_lits);
 Clause* result = new Clause(lit_list);
          
 for(LiteralList::Iterator iter(lit_list);iter.notEnd();iter.next())
  { 
   const TMP_LITERAL* tmp_lit = lits.Pop(); 
   TERM* lit = FastUnsharedLit(*tmp_lit);
   iter.currentElement()->setLiteral(lit);
   if (tmp_lit->isInherentlySelected()) 
    { 
     iter.currentElement()->setMark(LiteralList::InherentlySelected);
    };
  };


 result->SetAncestors(_ancestors.NewAncestors()); 
 result->setSize(weight);
 result->SetNumOfPosEq(0);
 result->setMinChildSize(0);  
 return result;  
}; // Clause* NewClause::MakeFastUnsharedClause(TMP_LIT_LIST& lits) 




bool NewClause::makeClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits,long clauseNum,bool showNewClause,bool markAsName)
{
 CALL("makeClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits,long clauseNum,bool showNewClause,bool markAsName");
 ASSERT(!lits.Empty());

 ulong size = lits.Weight();
 BitWord rules = _ancestors.Rules();
 if (markAsName) rules.set(ClauseBackground::Name);
 ulong numOfLits = lits.Length();
 ulong numOfAncestors = _ancestors.numberOfAncestors();
 PagedPassiveClauseStorage::StoredClauseHandler newClauseHandler;
 try_to_open:   
  storage.openClause(newClauseHandler,clauseNum,size,rules,numOfLits,numOfAncestors);
  if (!newClauseHandler)
   {
    // try to cancel some bad clauses in storage
     PagedPassiveClauseStorage::ClauseProfile worstClauseProfile;
     if (storage.worstClause(worstClauseProfile))
      {
       if (worstClauseProfile.clauseSize() > lits.Weight())
	{
	 ALWAYS(storage.cancelWorstClause());
         goto try_to_open;
        };
      }; 
     return false;
   };
 

 // now write the literals 
 writeIntoPagedPassiveClauseStorage(storage,lits);
 
 // now write the ancestors
  
 for (Clause* const * anc = _ancestors.last(); anc >= _ancestors.begin(); anc--)
  storage.writeAncestor(*anc); 

 storage.endOfClause();

 if (showNewClause)
  {
   PagedPassiveClauseStorage::StoredClause newClause;
   PagedPassiveClauseStorage::loadClause(newClauseHandler,newClause);
   newClause.output(cout << "% ") << "\n";
  };
 return true;
}; // bool NewClause::makeClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits,long clauseNum,bool showNewClause,bool markAsName)


bool NewClause::PushLiteral(TERM* t,bool inherentlySelected)
 {
  CALL("PushLiteral(TERM* t,bool inherentlySelected)");
  TERM header = *t;
  if (!(header.arity())) 
   {
    return PushPropLit(header,inherentlySelected);
   };
  if (!(OpenNonpropLiteral(header,inherentlySelected))) return false;
  TERM::Iterator iter;
  iter.Reset(t);
  while (iter.Next()) 
   {
    if (iter.CurrentSymIsVar())
     {
      PushVar(iter.Symbol());
     }
    else
     {
      if (iter.Symbol().arity())
       {
        if (!(PushNonconstFun(iter.Symbol()))) return false;
       }
      else PushConst(iter.Symbol()); 
     }; 
   };
  return CloseLiteral();
 }; // bool NewClause::PushLiteral(TERM* t,bool inherentlySelected)

bool NewClause::LoadClause(const Clause* c)
 {
  CALL("LoadClause(const Clause* c)");

  Reset();

  for(LiteralList::Iterator iter(c->LitList());iter.notEnd();iter.next())
    if (!(PushLiteral(iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) return false;
       
  return true;
 }; // void NewClause::LoadClause(Clause* c)

bool NewClause::LoadClause(PagedPassiveClauseStorage::ClauseLiterals& literals)
{
 CALL("LoadClause(PagedPassiveClauseStorage::ClauseLiterals& literals)");
 literals.reset();
 Reset();

 while (literals.nextLiteral())
  {
   bool inherentlySelected = literals.currentLiteralIsInherentlySelected();
   TERM sym;
   ALWAYS(literals.nextSymbol(sym)); // header
   if (!sym.arity()) 
    {
     if (!PushPropLit(sym,inherentlySelected)) return false;
    }
   else
    {
     if (!(OpenNonpropLiteral(sym,inherentlySelected))) return false; 
     while (literals.nextSymbol(sym)) if (!PushSymbol(sym)) return false;
     if (!CloseLiteral()) return false;
    };
  };
 return true;
}; // bool NewClause::LoadClause(PagedPassiveClauseStorage::ClauseLiterals& literals)



bool NewClause::PushRewrittenLit(PREFIX_SYM* lit,
                                 PREFIX_SYM* redex,
                                 TERM replacing_term,
                                 ulong index,
                                 bool inherentlySelected)
 {
  CALL("PushRewrittenLit(PREFIX_SYM* lit,PREFIX_SYM* redex,TERM replacing_term,ulong index,bool inherentlySelected)");
  TERM header = lit->Head().Func();
  ASSERT(header.arity());
  if (header.IsEquality()) header.MakeUnordered(); 
  if (!(OpenNonpropLiteral(header,inherentlySelected))) return false;
  
  PREFIX::ITER iter;
  iter.Reset(lit);
  TERM sym;
  iter.SkipSym(); // skipping the header
  
  do
   {
    if (iter.CurrVar()) // free variable
     { 
      if (!(PushVar(_varRenaming.Rename(iter.CurrVar())))) { return false; }; 
      iter.SkipSym(); 
     }
    else // functor
     {
      if (iter.CurrSubterm() == redex)
       {
        if (!PushStandardTerm(index,replacing_term)) return false;
        iter.SkipTerm();
       }
      else
       {
        sym = iter.CurrSubterm()->Head().Func();
        if (sym.arity())
         {
          if (!(PushNonconstFun(sym))) return false;
         }
        else // constant
         PushConst(sym);
        iter.SkipSym();
       }; 
     };
   }
  while (!iter.Empty()); 
 
  return CloseLiteral();
 }; //  bool NewClause::PushRewrittenLit(PREFIX_SYM* lit,...  



void NewClause::NegEqSplitting(ClauseQueue& newClauses)
{
 CALL("NegEqSplitting(ClauseQueue& newClauses)");

 bool firstArgEligible;
 bool secondArgEligible;
 bool firstNameExists;
 bool secondNameExists;
 ulong firstName;
 ulong secondName;
 bool newName; 
 TMP_LITERAL* lit;
 for (TMP_LITERAL** litp = _openClause.AllLiterals().First(); litp <= _openClause.AllLiterals().Last(); litp++) 
  {
   lit = *litp;
   if (lit->IsEquality() && lit->Negative())
    {
     firstArgEligible = (!(lit->Arg1Vars())) && (lit->Arg1()->size() > _negEqSplitting);
     secondArgEligible = (!(lit->Arg2Vars())) && (lit->Arg2()->size() > _negEqSplitting);

     if (firstArgEligible)
      {
       if (secondArgEligible)
        {
	 // both ground and heavy enough        
	 firstNameExists = _compName->findName(lit->Arg1(),lit->Arg1Vars(),firstName);
         secondNameExists = _compName->findName(lit->Arg2(),lit->Arg2Vars(),secondName);
         if (firstNameExists)
	  {
           if (secondNameExists)
	    { 
	     // both arguments have names
	     ASSERT(firstName != secondName);
	     if (lit->Arg1()->size() > lit->Arg2()->size())
	      {            
               TERM negHeader((TERM::Arity)1,(TERM::Functor)firstName,TERM::NegativePolarity);               
	       _openClause.SpareLiteral()->Reset();
               _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
               _openClause.SpareLiteral()->PushTerm(lit->Arg2());
              }          
             else // lit->Arg1()->size() <= lit->Arg2()->size()
	      {
               TERM negHeader((TERM::Arity)1,(TERM::Functor)secondName,TERM::NegativePolarity);               
	       _openClause.SpareLiteral()->Reset();
               _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
               _openClause.SpareLiteral()->PushTerm(lit->Arg1());
              };
            }
           else // firstNameExists && !secondNameExits
	    {
             TERM negHeader((TERM::Arity)1,(TERM::Functor)firstName,TERM::NegativePolarity);               
	     _openClause.SpareLiteral()->Reset();
             _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
             _openClause.SpareLiteral()->PushTerm(lit->Arg2());
 
            };
           TMP_LITERAL* tmp = _openClause.SpareLiteral();
           _openClause.SpareLiteral() = lit;
           *litp = tmp;
           UsedRule(ClauseBackground::Split);
          }
         else // !firstNameExists 
	  if (secondNameExists)
	   { 
            TERM negHeader((TERM::Arity)1,(TERM::Functor)secondName,TERM::NegativePolarity);               
	    _openClause.SpareLiteral()->Reset();
            _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
            _openClause.SpareLiteral()->PushTerm(lit->Arg1());
            TMP_LITERAL* tmp = _openClause.SpareLiteral();
            _openClause.SpareLiteral() = lit;
            *litp = tmp;
            UsedRule(ClauseBackground::Split);
           }
          else // !firstNameExists && !secondNameExits
	   {
	    // a new name must be created            
            if (lit->Arg1()->size() > lit->Arg2()->size())
	     {
	      if (_compName->nameFor(lit->Arg1(),lit->Arg1Vars(),firstName,newName))
	       {
		ASSERT_IN(newName,"COP100");
                TERM posHeader((TERM::Arity)1,(TERM::Functor)firstName,TERM::PositivePolarity); 
                TERM negHeader((TERM::Arity)1,(TERM::Functor)firstName,TERM::NegativePolarity);
                GlobalHeaderWeighting.MakeMaximal(posHeader);
                GlobalHeaderWeighting.MakeMaximal(negHeader);

                _openClause.SpareLiteral()->Reset();
                _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
                _openClause.SpareLiteral()->PushTerm(lit->Arg2());
                TMP_LITERAL* tmp = _openClause.SpareLiteral();
                _openClause.SpareLiteral() = lit;
                *litp = tmp;    
                
                _positiveName.Reset();
                _positiveName.PushNonpropHeader(posHeader);
                _positiveName.PushTerm(lit->Arg1());
                _components[0].Reset();
                _components[0].Push(&_positiveName);
                UsedRule(ClauseBackground::Split);
                Clause* cl = MakeSharedClause(_components[0]);
                cl->setMainSet(Clause::MainSetNew);
                newClauses.enqueue(cl);                  
               };
             } 
            else // lit->Arg1()->size() <= lit->Arg2()->size() 
	     {
              if (_compName->nameFor(lit->Arg2(),lit->Arg2Vars(),secondName,newName))
	       {
		ASSERT_IN(newName,"COP200");
                TERM posHeader((TERM::Arity)1,(TERM::Functor)secondName,TERM::PositivePolarity); 
                TERM negHeader((TERM::Arity)1,(TERM::Functor)secondName,TERM::NegativePolarity);
                GlobalHeaderWeighting.MakeMaximal(posHeader);
                GlobalHeaderWeighting.MakeMaximal(negHeader);

                _openClause.SpareLiteral()->Reset();
                _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
                _openClause.SpareLiteral()->PushTerm(lit->Arg1());
                TMP_LITERAL* tmp = _openClause.SpareLiteral();
                _openClause.SpareLiteral() = lit;
                *litp = tmp;    
                
                _positiveName.Reset();
                _positiveName.PushNonpropHeader(posHeader);
                _positiveName.PushTerm(lit->Arg2());
                _components[0].Reset();
                _components[0].Push(&_positiveName);
                UsedRule(ClauseBackground::Split);
                Clause* cl = MakeSharedClause(_components[0]);
                cl->setMainSet(Clause::MainSetNew);
                newClauses.enqueue(cl);                   
               };
             };
           };
        }
       else // only firstArgEligible
        {
         firstNameExists = _compName->findName(lit->Arg1(),lit->Arg1Vars(),firstName);
         if (firstNameExists) // the old name can be used
	  {
           TERM negHeader((TERM::Arity)1,(TERM::Functor)firstName,TERM::NegativePolarity);               
	   _openClause.SpareLiteral()->Reset();
           _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
           _openClause.SpareLiteral()->PushTerm(lit->Arg2());
           TMP_LITERAL* tmp = _openClause.SpareLiteral();
           _openClause.SpareLiteral() = lit;
           *litp = tmp;
           UsedRule(ClauseBackground::Split);
          }
         else // a new name must be introduced
	  {
	   if ((lit->Arg2()->size() > 1) 
               && (_compName->nameFor(lit->Arg1(),lit->Arg1Vars(),firstName,newName)))
	    {
	     ASSERT_IN(newName,"COP300");
             TERM posHeader((TERM::Arity)1,(TERM::Functor)firstName,TERM::PositivePolarity); 
             TERM negHeader((TERM::Arity)1,(TERM::Functor)firstName,TERM::NegativePolarity);
             GlobalHeaderWeighting.MakeMaximal(posHeader);
             GlobalHeaderWeighting.MakeMaximal(negHeader);

             _openClause.SpareLiteral()->Reset();
             _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
             _openClause.SpareLiteral()->PushTerm(lit->Arg2());
             TMP_LITERAL* tmp = _openClause.SpareLiteral();
             _openClause.SpareLiteral() = lit;
             *litp = tmp;    
                
             _positiveName.Reset();
             _positiveName.PushNonpropHeader(posHeader);
             _positiveName.PushTerm(lit->Arg1());
             _components[0].Reset();
             _components[0].Push(&_positiveName);
             UsedRule(ClauseBackground::Split);  
             Clause* cl = MakeSharedClause(_components[0]);
             cl->setMainSet(Clause::MainSetNew);
             newClauses.enqueue(cl);                    
            };
          };
        };
      }
     else // !firstArgEligible
      if (secondArgEligible)
       {
	// only secondArgEligible            
        secondNameExists = _compName->findName(lit->Arg2(),lit->Arg2Vars(),secondName);     
        if (secondNameExists) // the old name can be used
	 {
          TERM negHeader((TERM::Arity)1,(TERM::Functor)secondName,TERM::NegativePolarity);               
	  _openClause.SpareLiteral()->Reset();
          _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
          _openClause.SpareLiteral()->PushTerm(lit->Arg1());
          TMP_LITERAL* tmp = _openClause.SpareLiteral();
          _openClause.SpareLiteral() = lit;
          *litp = tmp;
          UsedRule(ClauseBackground::Split);         
         }
        else // a new name must be introduced
	 {
          if ((lit->Arg1()->size() > 1)
              && (_compName->nameFor(lit->Arg2(),lit->Arg2Vars(),secondName,newName)))
	   {
            ASSERT_IN(newName,"COP400");
            TERM posHeader((TERM::Arity)1,(TERM::Functor)secondName,TERM::PositivePolarity); 
            TERM negHeader((TERM::Arity)1,(TERM::Functor)secondName,TERM::NegativePolarity);
            GlobalHeaderWeighting.MakeMaximal(posHeader);
            GlobalHeaderWeighting.MakeMaximal(negHeader);

            _openClause.SpareLiteral()->Reset();
            _openClause.SpareLiteral()->PushNonpropHeader(negHeader);
            _openClause.SpareLiteral()->PushTerm(lit->Arg1());
            TMP_LITERAL* tmp = _openClause.SpareLiteral();
            _openClause.SpareLiteral() = lit;
            *litp = tmp;    
                
            _positiveName.Reset();
            _positiveName.PushNonpropHeader(posHeader);
            _positiveName.PushTerm(lit->Arg2());
            _components[0].Reset();
            _components[0].Push(&_positiveName);
            UsedRule(ClauseBackground::Split);   
            Clause* cl = MakeSharedClause(_components[0]);
            cl->setMainSet(Clause::MainSetNew);
            newClauses.enqueue(cl); 
	   };            
         };
       };

   }; // if (lit->IsEquality() && lit->Negative()) ...
 }; // for ...
  
}; // void NewClause::NegEqSplitting(ClauseQueue& newClauses)


bool NewClause::NormalizeByCommutativity()
{
 CALL("NormalizeByCommutativity()");

 ulong num_of_norm_lits = 0;
 for (TMP_LITERAL** lit = _openClause.AllLiterals().First(); lit <= _openClause.AllLiterals().Last(); lit++) 
  {
   if ((*lit)->arity() && (NormalizeByCommutativity(*lit)))
    { 
     num_of_norm_lits++;
    }; 
  };
 return (bool)num_of_norm_lits; 
}; // bool NewClause::NormalizeByCommutativity()

bool NewClause::NormalizeByCommutativity(TMP_LITERAL* lit)
{
 CALL("NormalizeByCommutativity(TMP_LITERAL* lit)");
 Flatterm* arg1;
 Flatterm* arg2;
 Flatterm::Iterator iter(lit->FlatTerm());
 iter.NextSym();
 ASSERT(iter.NotEndOfTerm());
 bool swapped = false;
 const Clause* commutativityLaw;
 do 
  {
   if ((iter.CurrPos()->isComplex()) && ((commutativityLaw = _commutativityTable[iter.CurrPos()->functor()])) && (!_ancestors.full()))
    {
     // compare arguments lexicografically
     arg1 = (Flatterm*)(iter.CurrPos() + 1);
     arg2 = arg1->after();
     if (arg1->CompareLex(arg2) == Greater)
      {       
       ((Flatterm*)iter.CurrPos())->BinSwapArgs(); 
       swapped = true;        
       PushAncestor((Clause*)commutativityLaw);
      };
    };    
   iter.NextSym();
  }
 while (iter.NotEndOfTerm());  
 if (lit->IsEquality() && swapped) lit->EqMakeUnordered();
 return swapped;
}; // bool NewClause::NormalizeByCommutativity(TMP_LITERAL* lit)


ostream& NewClause::output(ostream& str) const 
{
 return outputAncestors(str << _openClause);
};
 
ostream& NewClause::outputAncestors(ostream& str) const
{
 ClauseBackground::outputRules(str << '[',_ancestors.Rules());  
 for (Clause* const * anc = _ancestors.last(); anc >= _ancestors.begin(); anc--)
  str << (*anc)->Number() << ' '; 
 return str << ']';              
};
   
#ifndef NO_DEBUG_VIS
ostream& NewClause::outputComponents(ostream& str) const 
{
 for (ulong n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   ulong compNum = _registeredCompNums.nth(n);
   str << _components[compNum];
   if (n + 1 < _registeredCompNums.numOfDiffElements()) str << " || ";
  };  
 return str; 
};
#endif



//===========================================================================

