#ifndef NEW_CLAUSE_H
//=========================================================================
#define NEW_CLAUSE_H
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "ClassDesc.h"
#include "RuntimeError.h"
#include "term.h"
#include "prefix.h"
#include "variables.h"
#include "Flatterm.h"
#include "tmp_literal.h"
#include "OpenNewClause.h"
#include "ClauseBackground.h"
#include "clause.h"                       
#include "forward_subsumption.h"
#include "ForwardDemodulation.h"
#include "DivIntoComp.h"
#include "ComponentName.h"
#include "FIFO.h"
#include "Signature.h"
#include "UnitConflict.h"
#include "ClauseQueue.h"
#include "PagedPassiveClauseStorage.h"
extern HEADER_WEIGHTING GlobalHeaderWeighting;
extern Signature TheSignature;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEW_CLAUSE
 #define DEBUG_NAMESPACE "NewClause"
#endif 
#include "debugMacros.h"
//=================================================

class NewClause
 {
  public:
   typedef Multiset<ulong,HellConst::MaxNumOfDifferentHeaders> SetOfSplittingBranches;
   class Statistics
   {
    public:
     Statistics() { reset(); };
     void reset() 
     {
      generated = 0;
      forwardSubsumed = 0;
      simplifiedByFSR = 0;
      simplifiedByDemod = 0;
      fDOnSplittingBranches = 0;
      simplifiedByEqRes = 0;
      differentComponents = 0;
      split = 0;
      components = 0;
     };
     ulong generated;
     ulong forwardSubsumed;
     ulong simplifiedByFSR;
     ulong simplifiedByDemod;
     ulong fDOnSplittingBranches;
     ulong simplifiedByEqRes;
     ulong differentComponents;
     ulong split;
     ulong components;
   };

  public:
   Statistics statistics;
 
  public:
   NewClause() 
     : statistics(),
      _openClause(), 
      _newClauses("NewClause::_newClauses"),
      _basicComponent(),
      _registeredCompNums(),
      _positiveName(),
      _predElimTab((const bool*)0),
      _defHeaderTab((const bool*)0),
      _fastOpenUnsharedTerm(),
      _clauseNumber(0L),
      _ancestors(), 
      _varRenaming(), 
      _intVarRenaming(),
      _varWeighting(),
      _divisionIntoComponents(),
      _clauseIndex((CLAUSE_INDEX*)0),
      _compName((ComponentName*)0),
      _forwardSubsumption((FS*)0),
      _forwardSimplificationByUEq((ForwardDemodulation*)0),
      _fDOnSplittingBranchesFlag(false), // default
      _splittingBranches(),
      _forwardSubsumptionFlag(false), // default
      _forwardSubsumptionResolution(0L), // default
      _unitConflict((UnitConflict*)0), // default 
      _forwardDemodulation(0), // default
      _commutativityTable((const Clause* const *)0), // default 
      _simplificationByEqRes(false), // default
      _splittingFlag(false), // default
      _splittingHasBeenUsed(false),
      _splittingStyle(0L), // default
      _splittingWithNamingFlag(false), // default
      _negEqSplitting(0L), // off, default

      _select(select1),          // default
      _inheritedNegativeSelection(false), // default

      _equalityPossible(false), // default
      _maxWeight(HellConst::MaxClauseWeight),
      _previousMaxWeight(HellConst::MaxClauseSize),
      _optimalGroundCompWeight(ULONG_MAX)
    {
    };

   ~NewClause() { CALL("destructor ~NewClause()"); };     
   void Reset()
    {               
     statistics.generated++;
     _openClause.Reset();
     _ancestors.Reset();
    };

      // setting options and parameters
   void SetClauseIndex(CLAUSE_INDEX* ci) { _clauseIndex = ci; }; 
   void SetComponentName(ComponentName* cn) { _compName = cn; };
   void SetForwardSubsumption(FS* fs) { _forwardSubsumption = fs; };
   void SetSimplificationByUEq(ForwardDemodulation* fsueq) { _forwardSimplificationByUEq = fsueq; };
   void SetFDOnSplittingBranchesFlag(bool fl) { _fDOnSplittingBranchesFlag = fl; };
   void SetNormalizationByCommutativity(const Clause* const * com_tab) { _commutativityTable = com_tab; };
   void SetPredElimTable(const bool* t) { _predElimTab = t; }; 
   void SetDefHeaderTable(const bool* t) { _defHeaderTab = t; }; 
   
   void SetForwardSubsumption(bool fl) { _forwardSubsumptionFlag = fl; };
   void SetSimplificationByFS(long fl) 
   {
    if (fl > 0)
     {
      _forwardSubsumptionResolution = (ulong)fl; 
     }
    else _forwardSubsumptionResolution = 0;
   };
   void SetUnitConflict(UnitConflict* uc) { _unitConflict = uc; };
   void SetSimplificationByUEq(long fl) { _forwardDemodulation = fl; };
   void SetSimplificationByEqRes(bool fl) { _simplificationByEqRes = fl; };
   void SetSplitting(bool fl) 
   { 
    _splittingFlag = fl;
    if (fl) _splittingHasBeenUsed = true; // clumsy! 
   };
   void SetSplittingStyle(long num) { _splittingStyle = num; };
   void SetSplittingWithNaming(bool fl) { _splittingWithNamingFlag = fl; };
   void SetNegEqSplitting(long w) { _negEqSplitting = w; };

   void SetSelection(long num) 
   {
    switch (num) 
    {
     case 1L : 
      _select = select1; 
      _inversePolarityForSelection = false;
      return;
     case 2L : 
      _select = select2; 
      _inversePolarityForSelection = false; 
      return;
     case 3L : 
      _select = select3; 
      _inversePolarityForSelection = false; 
      return;
     case 4L : 
      _select = select4; 
      _inversePolarityForSelection = false; 
      return;
     case 5L : 
      _select = select5; 
      _inversePolarityForSelection = false; 
      return;
     case 6L : 
      _select = select6; 
      _inversePolarityForSelection = false; 
      return;
     case 7L : 
      _select = select7; 
      _inversePolarityForSelection = false; 
      return;


     case 1002L :
      _select = select1002; 
      _inversePolarityForSelection = false; 
      return;
     case 1003L :
      _select = select1003; 
      _inversePolarityForSelection = false; 
      return;
     case 1004L :
      _select = select1004; 
      _inversePolarityForSelection = false; 
      return;
     case 1005L :
      _select = select1005; 
      _inversePolarityForSelection = false; 
      return;
     case 1006L :
      _select = select1006; 
      _inversePolarityForSelection = false; 
      return;
     case 1007L :
      _select = select1007; 
      _inversePolarityForSelection = false; 
      return;
     case 1008L :
      _select = select1008; 
      _inversePolarityForSelection = false; 
      return;    
     case 1009L :
      _select = select1009; 
      _inversePolarityForSelection = false; 
      return;          
     

     case -2L : 
      _select = select2;
      _inversePolarityForSelection = true;
      return;
     case -3L : _select = select3;
      _inversePolarityForSelection = true; 
      return;
     case -4L : 
      _select = select4;
      _inversePolarityForSelection = true; 
      return;
     case -5L : 
      _select = select5;
      _inversePolarityForSelection = true; 
      return;
     case -6L : 
      _select = select6;
      _inversePolarityForSelection = true; 
      return;
     case -7L : 
      _select = select7;
      _inversePolarityForSelection = true; 
      return;

     case -1005L :
      _select = select1005; 
      _inversePolarityForSelection = true; 
      return;
     case -1009L :
      _select = select1009; 
      _inversePolarityForSelection = true; 
      return;

     default: RuntimeError::report("Wrong selection number.");
    };
   };
   void SetLiteralComparisonMode(long mode) { _literalComparisonMode = mode; };
   void SetInheritedNegativeSelection(bool fl) { _inheritedNegativeSelection = fl; };   
   void SetEqualityPossible(bool fl) { _equalityPossible = fl; };
   void SetMaxWeight(ulong mw) 
    {
     _previousMaxWeight = _maxWeight;
     _maxWeight = (mw > HellConst::MaxClauseSize) ? HellConst::MaxClauseSize : mw;
     if (_maxWeight < HellConst::MinMaxWeight) _maxWeight = HellConst::MinMaxWeight;
     _openClause.SetMaxWeight(_maxWeight);
    };
   void SetMaxLength(long ml) { _openClause.SetMaxLength(ml); };
   void SetMaxDepth(long md) { _openClause.SetMaxDepth(md); };
   void SetMaxLitWeight(long mlw) { _openClause.SetMaxLitWeight(mlw); };
   void SetOptimalGroundCompWeight(ulong ogcw) { _optimalGroundCompWeight = ogcw; };

   // getting options and parameters 

   long MaxWeight() const { return _maxWeight; };
   long PreviousMaxWeight() const { return _previousMaxWeight; };


   // keeping statistics 

   void AnotherForwardSubsumed() { statistics.forwardSubsumed++; };

   void AnotherSimplifiedByFSR() { statistics.simplifiedByFSR++; };
   void AnotherSimplifiedByDemod() { statistics.simplifiedByDemod++; };  
   void AnotherFDOnSplittingBranches() { statistics.fDOnSplittingBranches++; };
   void AnotherSimplifiedByEqRes() { statistics.simplifiedByDemod++; };  

   void AnotherComponent() { statistics.differentComponents++; };
   void AnotherSplit(ulong numOfComponents) 
   { 
    statistics.split++;
    statistics.components += numOfComponents; 
   };
   void AnotherTooHeavyClause() { _openClause.AnotherTooHeavyClause(); };

   // getting statistics 


   ulong statTooHeavy() const { return _openClause.statistics.tooHeavy; };
   ulong statTooHeavyLiterals() const { return _openClause.statistics.tooHeavyLiterals; };
   ulong statTooManyVariables() const { return _openClause.statistics.tooManyVariables; };
   ulong statTooLong() const { return _openClause.statistics.tooLong; };
   ulong statTooDeep() const { return _openClause.statistics.tooDeep; };
   ulong statEqTautologies() const { return _openClause.statistics.eqTautologies; };
   ulong statPropTautologies() const { return _openClause.statistics.propTautologies; };
 
   const bool& WasTautology() const { return _openClause.WasTautology(); };   

  public: // collecting a clause 

   void MakeVIP() { _ancestors.MakeVIP(); };
   bool PushSymbol(const TERM& sym)
    {
     if (sym.isVariable()) { PushVar(sym); return true; };
     if (sym.arity()) return PushNonconstFun(sym);
     PushConst(sym);
     return true;
    };   
   bool PushFunc(const TERM& f)
    {
     CALL("PushFunc(const TERM& f)");
     ASSERT(f.isComplex());
     if (f.arity()) return PushNonconstFun(f);
     PushConst(f); 
     return true;
    };
 
   bool OpenNonpropLiteral(const TERM& header,bool inherentlySelected) 
   { 
    return _openClause.OpenNonpropLiteral(header,inherentlySelected); 
   };
   bool CloseLiteral() { return _openClause.CloseLiteral(); };
   bool PushPropLit(const TERM& lit,bool inherentlySelected) 
   { 
    return _openClause.PushPropLit(lit,inherentlySelected); 
   }; 

   bool PushRewrittenLit(ulong index,TERM* lit,TERM* redex,PREFIX_SYM* replacing_term,bool inherentlySelected);
   bool PushRewrittenLit(PREFIX_SYM* lit,
                         PREFIX_SYM* redex,
                         TERM replacing_term,
                         ulong index,
                         bool inherentlySelected);
   bool PushPrefixLit(PREFIX_SYM* word,bool inherentlySelected); 
   bool PushPrefixEqLit(TERM header,PREFIX_SYM* arg1,PREFIX_SYM* arg2,bool inherentlySelected);
   bool PushPrefixComplexTerm(PREFIX_SYM* word);
   bool PushPrefixTerm(PREFIX_SYM* t);
   bool PushStandardLit(ulong index,TERM* t,bool inherentlySelected); 
   bool PushStandardTerm(ulong index,TERM t);
   bool PushStandardComplexTerm(ulong index,TERM* t);
   bool PushLiteral(TERM* t,bool inherentlySelected);   
   bool LoadClause(const Clause* c);
   bool LoadClause(PagedPassiveClauseStorage::ClauseLiterals& literals);

   void setClauseNumber(long num) { _clauseNumber = num; }; 
   void UsedRule(ClauseBackground::Rule r) { _ancestors.UsedRule(r); }; 
   void CancelRule(ClauseBackground::Rule r) { _ancestors.CancelRule(r); }; 
   void UsedRules(const ClauseBackground::Rules& rs) { _ancestors.UsedRules(rs); }; 
   void PushAncestor(Clause* cl) 
   {
    CALL("PushAncestor(Clause* cl)");
    // ASSERT(cl->Active()); 
    _ancestors.Push(cl); 
   };
   void InverseAncestorOrder() { _ancestors.InverseOrder(); };
  public:  // processing the collected clause
   Clause* MakeEmptyClause();
   Clause* FSubsume(Clause* cl);
   Clause* checkForUnitConflict(); 
   void ProcessCollected(ClauseQueue& newClauses);
   void ProcessCollected1(ClauseQueue& newClauses);
   void ProcessCollected2(ClauseQueue& newClauses);
   void ProcessCollected3(Clause*& emptyClause,
                          PagedPassiveClauseStorage& storage,
                          ulong& numOfNewClauses,
                          ulong& numOfRejectedNewClauses,
                          ulong (*nextClauseNumFunc)(),
                          bool showNewClauses); 
   Clause* Reanimate();
   bool transferToPagedPassive(PagedPassiveClauseStorage& storage);
   void writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits);  
   void writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LITERAL* lit);

  public: // output
   ostream& output(ostream& str) const;
   ostream& outputAncestors(ostream& str) const;
   #ifndef NO_DEBUG_VIS
    ostream& outputComponents(ostream& str) const;
   #endif
   
  private:  // collecting a clause 
   bool PushVar(const ulong v) { return _openClause.PushVar(v); };
   void PushVar(const TERM& v) { _openClause.PushVar(v); };  
   void PushConst(const TERM& c) { _openClause.PushConst(c); };
   bool PushNonconstFun(const TERM& f) { return _openClause.PushNonconstFun(f); };

  private: // processing collected clause
   static bool PositiveForSelection(const TERM& header);
   static bool NegativeForSelection(const TERM& header);
   void select(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed);
   void selectWithPragmaticSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed);
   void selectWithNonblockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed);
   void selectWithBlockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed);
   

   static void select1(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select2(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select3(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select4(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select5(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat); 
   static void select6(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);    
   static void select7(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);

   static void select1002(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select1003(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select1004(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select1005(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select1006(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select1007(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select1008(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);
   static void select1009(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed,const TMP_LIT_LIST::STAT& stat);

   static bool satisfactorySelection1(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);
   static bool satisfactorySelection2(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);
   static bool satisfactorySelection3(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);
   

   static void selectNegativeWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);
   static void selectNegativeLeastProlific(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);  
  static void selectNegativeMaximallyNongroundWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);

   static void selectInherentlySelected(TMP_LIT_LIST& lits,ulong numOfSelected,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits);  

   void Simplify(bool& redundant);
   void FSQueryPushLit(TMP_LITERAL* lit);
   Clause* FSubsume();
   Clause* FSubsume(TMP_LIT_LIST& lits);
   ulong SimplifyByFS();
   ulong SimplifyByFS(TMP_LIT_LIST& lits);
   Clause* RewriteByUEq(TMP_LITERAL* lit,TMP_LITERAL* new_lit);
   bool SimplifyByUEq(TMP_LITERAL*& lit,TMP_LITERAL*& spare_lit);
   bool SimplifyByUEq();
   bool SimplifyByEqRes();
   
   bool NormalizeByCommutativity();  
   bool NormalizeByCommutativity(TMP_LITERAL* lit);

   Clause* MakeUnitClause(TMP_LITERAL* lit);

   Clause* MakeSharedClause(TMP_LIT_LIST& lits);
   Clause* MakeSharedClause1(TMP_LIT_LIST& lits);
   TERM* TmpLitSharedCopy(const TMP_LITERAL* lit,Clause* cl) const;
   TERM* FastUnsharedLit(const TMP_LITERAL& lit);
   
   Clause* MakeFastUnsharedClause(TMP_LIT_LIST& lits);
   bool makeClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,TMP_LIT_LIST& lits,long clauseNum,bool showNewClause,bool markAsName);
 
 
  private: // splitting
   bool Split(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool Split2(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool Split3(TMP_LIT_LIST& lits,
               Clause*& emptyClause,
               PagedPassiveClauseStorage& storage,
               ulong& numOfNewClauses,
               ulong& numOfRejectedNewClauses,
               ulong (*nextClauseNumFunc)(),
               bool showNewClauses);
   bool PragmaticSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool PragmaticSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool PragmaticSplitting3(TMP_LIT_LIST& lits,
                            Clause*& emptyClause,
                            PagedPassiveClauseStorage& storage,
                            ulong& numOfNewClauses,
                            ulong& numOfRejectedNewClauses,
                            ulong (*nextClauseNumFunc)(),
                            bool showNewClauses);


   bool NonblockingSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool NonblockingSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool NonblockingSplitting3(TMP_LIT_LIST& lits,  
                              Clause*& emptyClause, 
                              PagedPassiveClauseStorage& storage,
                              ulong& numOfNewClauses,
                              ulong& numOfRejectedNewClauses,
                              ulong (*nextClauseNumFunc)(),
                              bool showNewClauses);
   


   bool BlockingSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool BlockingSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses);
   bool BlockingSplitting3(TMP_LIT_LIST& lits,                            
                           Clause*& emptyClause,
                           PagedPassiveClauseStorage& storage,
                           ulong& numOfNewClauses,
                           ulong& numOfRejectedNewClauses,
                           ulong (*nextClauseNumFunc)(),
                           bool showNewClauses);



   void NegEqSplitting(ClauseQueue& newClauses);
 
  private: // structure

   OpenNewClause _openClause;
   Stack<TMP_LIT_LIST*,HellConst::MaxNumOfLiterals> _newClauses;  
   
   OpenNewClause _basicComponent;
   TMP_LIT_LIST _components[HellConst::MaxNumOfLiterals];
   Multiset<ulong,HellConst::MaxNumOfLiterals> _registeredCompNums;
   ulong _componentName[HellConst::MaxNumOfLiterals];    
   bool _newCompName[HellConst::MaxNumOfLiterals];
   TMP_LITERAL _compNameLit;
   TMP_LITERAL _positiveName;
   const bool* _predElimTab;
   const bool* _defHeaderTab;

   TERM::FastOpenUnshared _fastOpenUnsharedTerm;

   long _clauseNumber;
   OpenClauseBackground _ancestors;
   VAR_TO_INT_RENAMING _varRenaming;
   INT_VAR_RENAMING _intVarRenaming;
   VAR_WEIGHTING _varWeighting;
   DivIntoComp<HellConst::MaxNumOfLiterals,HellConst::MaxNumOfVariables> _divisionIntoComponents;
   ulong _litNumMap[HellConst::MaxNumOfLiterals];
   CLAUSE_INDEX* _clauseIndex;
   ComponentName* _compName;
   FS* _forwardSubsumption;
   ForwardDemodulation* _forwardSimplificationByUEq;
   bool _fDOnSplittingBranchesFlag;
   SetOfSplittingBranches _splittingBranches;

  private: // options and parameters
   bool _forwardSubsumptionFlag; 
   ulong _forwardSubsumptionResolution;
   UnitConflict* _unitConflict;
   long _forwardDemodulation;
   const Clause* const * _commutativityTable;
   bool _simplificationByEqRes;
   bool _splittingFlag;
   bool _splittingHasBeenUsed; // clumsy!
   long _splittingStyle;
   bool _splittingWithNamingFlag;
   long _negEqSplitting;
   void (*_select)(TMP_LIT_LIST&,TMP_LIT_LIST&,TMP_LIT_LIST&,bool&,const TMP_LIT_LIST::STAT&);
   static long _literalComparisonMode;
   bool _inheritedNegativeSelection;    
   static bool _inversePolarityForSelection;

   bool _equalityPossible;
   long _maxWeight; 
   long _previousMaxWeight;
   ulong _optimalGroundCompWeight; 
  
 }; // class NewClause 

//=================================================================

inline ostream& operator<<(ostream& str,const NewClause& cl) 
{
 return cl.output(str);
};

inline ostream& operator<<(ostream& str,const NewClause* cl) 
{
 return cl->output(str);
};


inline TERM* NewClause::FastUnsharedLit(const TMP_LITERAL& lit) 
{
 CALL("FastUnsharedLit(const TMP_LITERAL& lit)");

 _fastOpenUnsharedTerm.reset();
 TERM header(lit.Header());
 _fastOpenUnsharedTerm.litHeader(header);
 if (header.IsEquality() && lit.EqSwapArgs())
  {
   Flatterm::Iterator iter(lit.Arg2());
   while (iter.NotEndOfTerm())
    {
     _fastOpenUnsharedTerm.symbol(iter.CurrPos()->symbolRef()); 
     iter.NextSym();
    };     
   iter.Reset(lit.Arg1());
   while (iter.NotEndOfTerm())
    {
     _fastOpenUnsharedTerm.symbol(iter.CurrPos()->symbolRef()); 
     iter.NextSym();
    };  
  }
 else
  {
   Flatterm::Iterator iter(lit.FlatTerm());
   iter.NextSym();
   while (iter.NotEndOfTerm())
    {
     _fastOpenUnsharedTerm.symbol(iter.CurrPos()->symbolRef()); 
     iter.NextSym();
    };
  };
 return _fastOpenUnsharedTerm.result().First();
}; // TERM* NewClause::FastUnsharedLit(const TMP_LITERAL& lit) 

inline bool NewClause::PushRewrittenLit(ulong index,TERM* lit,TERM* redex,PREFIX_SYM* replacing_term,bool inherentlySelected)
 {
  CALL("PushRewrittenLit(ulong index,TERM* lit,TERM* redex,PREFIX_SYM* replacing_term,bool inherentlySelected)");
  TERM header = *lit;
  ASSERT(header.arity()); 
  if (header.IsEquality()) header.MakeUnordered();
  if (!(OpenNonpropLiteral(header,inherentlySelected))) return false;
  
  ST_TERM_WITH_SUBST iter;
  iter.Reset(index,lit);
  iter.SkipSym();
  PREF_SYM sym;
  
  while (!(iter.Empty()))
   {
    if (iter.CurrPosInBase() == redex)
     {
      if (!PushPrefixTerm(replacing_term)) return false;
      iter.SkipTerm();
     }
    else
     { 
      sym = iter.CurrSym();
      if (sym.isVariable()) // free variable, must be renamed
       {
        if (!(PushVar(_varRenaming.Rename(sym.var())))) { return false; };
       }
      else
       {
        if (sym.Func().arity())
         {
          if (!(PushNonconstFun(sym.Func()))) return false;
         }
        else PushConst(sym.Func()); 
       };
      iter.SkipSym();
     };  
   };
   
  return CloseLiteral();
 }; // bool NewClause::PushRewrittenLit(ulong index,TERM* lit,TERM* redex,PREFIX_SYM* replacing_term,bool inherentlySelected)


inline bool NewClause::PushPrefixLit(PREFIX_SYM* word,bool inherentlySelected)
 {
  CALL("PushPrefixLit(PREFIX_SYM* word,bool inherentlySelected)");
  ASSERT(PREFIX::CorrectComplexTerm(word));
  TERM header = word->Head().Func();
  if (!(header.arity())) { return PushPropLit(header,inherentlySelected); };
  
  if (!(OpenNonpropLiteral(header,inherentlySelected))) return false;
  
  PREFIX::ITER iter;
  iter.Reset(word);
  TERM sym;
  iter.SkipSym(); // skipping the header
  do
   {
    if (iter.CurrVar()) // free variable
     { 
      if (!(PushVar(_varRenaming.Rename(iter.CurrVar())))) { return false; };  
     }
    else // functor
     {
      sym = iter.CurrSubterm()->Head().Func();
      if (sym.arity())
       {
        if (!(PushNonconstFun(sym))) return false;
       }
      else // constant
       PushConst(sym);
     };
    iter.SkipSym();
   }
  while (!(iter.Empty()));
  return CloseLiteral();
 }; // bool NewClause::PushPrefixLit(PREFIX_SYM* word,bool inherentlySelected)



inline bool NewClause::PushPrefixEqLit(TERM header,PREFIX_SYM* arg1,PREFIX_SYM* arg2,bool inherentlySelected)
 {
  CALL("PushPrefixEqLit(TERM header,PREFIX_SYM* arg1,PREFIX_SYM* arg2,bool inherentlySelected)");
  ASSERT(header.IsEquality());   
  
  return OpenNonpropLiteral(header,inherentlySelected)
         && PushPrefixTerm(arg1)
         && PushPrefixTerm(arg2)
         && CloseLiteral();
 }; // bool NewClause::PushPrefixEqLit(TERM header,PREFIX_SYM* arg1,PREFIX_SYM* arg2,bool inherentlySelected)


inline bool NewClause::PushPrefixComplexTerm(PREFIX_SYM* word)
 {
  CALL("PushPrefixComplexTerm(PREFIX_SYM* word)");
  PREFIX::ITER iter;
  iter.Reset(word);
  TERM sym;
  do
   {
    if (iter.CurrVar()) // free variable
     { 
      if (!(PushVar(_varRenaming.Rename(iter.CurrVar())))) return false; 
     }
    else // functor
     {
      sym = iter.CurrSubterm()->Head().Func();
      if (sym.arity())
       {
        if (!(PushNonconstFun(sym))) return false;
       }
      else // constant
       PushConst(sym);
     };
    iter.SkipSym();
   }
  while (!(iter.Empty()));
  return true;
 }; // bool NewClause::PushPrefixComplexTerm(PREFIX_SYM* word)

inline bool NewClause::PushPrefixTerm(PREFIX_SYM* t)
 {
  CALL("PushPrefixTerm(PREFIX_SYM* t)");
  if (t->Head().isVariable())
   {
    bool free_var;
    void* bind = t->Head().var()->Unref(free_var);
    if (free_var) return PushVar(_varRenaming.Rename((VAR*)bind));
    return PushPrefixComplexTerm((PREFIX_SYM*)bind); 
   };
  return PushPrefixComplexTerm(t);
 }; // bool NewClause::PushPrefixTerm(PREFIX_SYM* t)

inline bool NewClause::PushStandardTerm(ulong index,TERM t)
 {
  CALL("PushStandardTerm(ulong index,TERM t)");
  // t is variable or reference 
  if (t.isVariable())
   {
    bool free_var;
    void* bind = VAR::Var(t.var(),index)->Unref(free_var);
    
    if (free_var) return PushVar(_varRenaming.Rename((VAR*)bind));
    return PushPrefixComplexTerm((PREFIX_SYM*)bind);
   };
  // t is reference to a complex term
  return PushStandardComplexTerm(index,t.First());
 }; // bool NewClause::PushStandardTerm(ulong index,TERM t)

inline bool NewClause::PushStandardComplexTerm(ulong index,TERM* t)
 {
  CALL("PushStandardComplexTerm(ulong index,TERM* t)");
  ST_TERM_WITH_SUBST term;
  term.Reset(index,t);
  PREF_SYM sym;
  do
   {
    sym = term.CurrSym();
    if (sym.isVariable()) // free variable, must be renamed
     {
      if (!(PushVar(_varRenaming.Rename(sym.var())))) return false; 
     }
    else
     {
      if (sym.Func().arity())
       {
        if (!(PushNonconstFun(sym.Func()))) return false;
       }
      else PushConst(sym.Func()); 
     };
    term.SkipSym();
   }
  while (!(term.Empty()));
  return true; 
 }; // inline bool NewClause::PushStandardComplexTerm(ulong index,TERM* t)

inline bool NewClause::PushStandardLit(ulong index,TERM* t,bool inherentlySelected)
 {
  CALL("PushStandardLit(ulong index,TERM* t,bool inherentlySelected)");

  // Note: this function uses the global substitution to substitute the variables!
  TERM header = *t;

  if (!(header.arity())) { return PushPropLit(header,inherentlySelected); };

  if (!(OpenNonpropLiteral(header,inherentlySelected))) return false;
 
  ST_TERM_WITH_SUBST term;
  term.Reset(index,t);
  term.SkipSym();
  PREF_SYM sym;
  while (!(term.Empty()))
   {
    sym = term.CurrSym();
    if (sym.isVariable()) // free variable, must be renamed
     {
      if (!(PushVar(_varRenaming.Rename(sym.var())))) { return false; };
     }
    else
     {
      if (sym.Func().arity())
       {
        if (!(PushNonconstFun(sym.Func()))) return false;
       }
      else 
       {
        PushConst(sym.Func());
       }; 
     };
    term.SkipSym();
   };
   
  return CloseLiteral();  
 }; // bool NewClause::PushStandardLit(ulong index,TERM* t,bool inherentlySelected)


inline void NewClause::FSQueryPushLit(TMP_LITERAL* lit)
 {
  CALL("FSQueryPushLit(TMP_LITERAL* lit)");
  if (lit->IsEquality())
   {
       
    lit->LinkEqArgs();

    if (lit->IsOrderedEq())
     {
      if (lit->EqSwapArgs())
       {
        _forwardSubsumption->PushOrdEq(lit->Header(),lit->Arg2(),lit->Arg1());
       }
      else
       { 
        _forwardSubsumption->PushOrdEq(lit->Header(),lit->Arg1(),lit->Arg2());
       }; 
     }
    else _forwardSubsumption->PushUnordEq(lit->Header(),lit->Arg1(),lit->Arg2()); 
   }
  else _forwardSubsumption->PushLit(lit->FlatTerm()); 
  
 }; // void NewClause::FSQueryPushLit(TMP_LITERAL* lit)


inline Clause* NewClause::FSubsume() 
 {
  CALL("FSubsume()");
  Clause* subsuming_cl;
  
  _forwardSubsumption->ResetQuery(); 
  for (TMP_LITERAL** lit = _openClause.AllLiterals().First(); lit <= _openClause.AllLiterals().Last(); lit++)
   FSQueryPushLit(*lit);
  subsuming_cl = _forwardSubsumption->Subsume();
  if (subsuming_cl) subsuming_cl->MakeVIP();
  return subsuming_cl;  
 }; // Clause* NewClause::FSubsume()

inline Clause* NewClause::FSubsume(TMP_LIT_LIST& lits) 
 {
  CALL("FSubsume(TMP_LIT_LIST& lits)");
  Clause* subsuming_cl;
  
  _forwardSubsumption->ResetQuery(); 
  for (TMP_LITERAL** lit = lits.First(); lit <= lits.Last(); lit++)
   FSQueryPushLit(*lit); 
  subsuming_cl = _forwardSubsumption->Subsume();
  return subsuming_cl;  
 }; //Clause* NewClause::FSubsume(TMP_LIT_LIST& lits)  

inline Clause* NewClause::FSubsume(Clause* cl)
 {
  CALL("FSubsume(Clause* cl)");
  LoadClause(cl);
  _openClause.EndOfClause();
  return FSubsume();
 }; // Clause* NewClause::FSubsume(Clause* cl)


inline ulong NewClause::SimplifyByFS()
 {
  CALL("SimplifyByFS()");
  return SimplifyByFS(_openClause.AllLiterals());

 }; // ulong NewClause::SimplifyByFS()


inline ulong NewClause::SimplifyByFS(TMP_LIT_LIST& lits)
 {
  CALL("SimplifyByFS(TMP_LIT_LIST& lits)");
  if (lits.Length() < _forwardSubsumptionResolution) return 0;
  ulong res = 0;
  
  TMP_LITERAL** cand_lit = lits.Last(); 
  try_lit: 
   if ((*cand_lit)->IsEquality())   
    {
     if (cand_lit == lits.First()) return res;
     cand_lit--;
     goto try_lit;
    };

   // try to simplify by resolving cand_lit
   ((Flatterm*)((*cand_lit)->FlatTerm()))->symbolRef().Inversepolarity();
   
   _forwardSubsumption->ResetQuery(); 
   for (TMP_LITERAL** lit = lits.First(); lit <= lits.Last(); lit++)
    FSQueryPushLit(*lit);
            
   Clause* simplifier = _forwardSubsumption->Subsume();
   if (simplifier)
    {
     res++;
     PushAncestor(simplifier);
     //simplifier->MarkAsSimplifier();
     simplifier->setFlag(Clause::FlagHasChildren);
     lits.ReplaceByTop(cand_lit);
     if ((lits.Length() < _forwardSubsumptionResolution) 
         || (cand_lit == lits.First()))
      return res;
     cand_lit--;
     goto try_lit;  
    }
   else 
    {
     ((Flatterm*)((*cand_lit)->FlatTerm()))->symbolRef().Inversepolarity();
     if (cand_lit == lits.First()) return res;
     cand_lit--;
     goto try_lit;
    }; 
 }; // ulong NewClause::SimplifyByFS(TMP_LIT_LIST& lits)


inline bool NewClause::Split(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 switch (_splittingStyle)
  {
   case 0L: return PragmaticSplitting(lits,newClauses);
   case 1L: return NonblockingSplitting(lits,newClauses);
   case 2L: return BlockingSplitting(lits,newClauses);
   default: 
    RuntimeError::report("Wrong splitting style number."); 
    return false; 
  };
}; //  bool NewClause::Split(TMP_LIT_LIST& lits,ClauseQueue& newClauses)

inline bool NewClause::Split2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 switch (_splittingStyle)
  {
   case 0L: return PragmaticSplitting2(lits,newClauses);
   case 1L: return NonblockingSplitting2(lits,newClauses);
   case 2L: return BlockingSplitting2(lits,newClauses);
   default: 
    RuntimeError::report("Wrong splitting style number."); 
    return false; 
  };
}; //  bool NewClause::Split2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)

inline bool NewClause::Split3(TMP_LIT_LIST& lits,                            
                              Clause*& emptyClause,
                              PagedPassiveClauseStorage& storage,
                              ulong& numOfNewClauses,
                              ulong& numOfRejectedNewClauses,
                              ulong (*nextClauseNumFunc)(),
                              bool showNewClauses)
{
  switch (_splittingStyle)
  {
   case 0L: return PragmaticSplitting3(lits,
                                       emptyClause,
                                       storage,
                                       numOfNewClauses,
                                       numOfRejectedNewClauses,
                                       nextClauseNumFunc,
                                       showNewClauses);
   case 1L: return NonblockingSplitting3(lits,
                                         emptyClause,
                                         storage,
                                         numOfNewClauses,
                                         numOfRejectedNewClauses,
                                         nextClauseNumFunc,
                                         showNewClauses);
   case 2L: return BlockingSplitting3(lits,
                                      emptyClause,
                                      storage,
                                      numOfNewClauses,
                                      numOfRejectedNewClauses,
                                      nextClauseNumFunc,
                                      showNewClauses);
   default: 
    RuntimeError::report("Wrong splitting style number."); 
    return false; 
  };

}; // bool NewClause::Split3(TMP_LIT_LIST& lits,PagedPassiveClauseStorage& storage,...

inline TERM* NewClause::TmpLitSharedCopy(const TMP_LITERAL* lit,Clause* cl) const
{
 CALL("TmpLitSharedCopy(const TMP_LITERAL* lit,const Clause* cl) const");
 if (lit->IsEquality() && lit->EqSwapArgs())
  {
   return _clauseIndex->IntegrateEq(lit->Header(),lit->Arg2(),lit->Arg1(),cl);
  }
 else 
  {
   return _clauseIndex->Integrate(lit->FlatTerm(),cl); 
  };
}; // TERM* NewClause::TmpLitSharedCopy(const TMP_LITERAL* lit,Clause* cl) const



inline bool NewClause::satisfactorySelection1(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
{
 CALL("satisfactorySelection1(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)");
 if (max_lits.Length() > 1) return false;
 // max_lits.Length() == 1 
 TMP_LITERAL* maxLit = *(max_lits.First()); 
 ulong maxLitWeight = maxLit->size();
 for (TMP_LITERAL** lit = nonmax_lits.Last(); lit >= nonmax_lits.First(); lit--)
  {
   // if ((*lit)->Negative() && ((*lit)->size() > maxLitWeight)) return false; 
   if (NegativeForSelection((*lit)->Header()) && ((*lit)->size() > maxLitWeight)) return false;  
  };
 return true;
}; // bool NewClause::satisfactorySelection1(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)

inline bool NewClause::satisfactorySelection2(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
{
 CALL("satisfactorySelection2(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)");
 if (max_lits.Length() > 1) return false;
 // max_lits.Length() == 1 
 TMP_LITERAL* maxLit = *(max_lits.First());
 ulong maxLitWeight = maxLit->size();
 //if (maxLit->Positive())
 if (PositiveForSelection(maxLit->Header()))
  {
   if (maxLit->IsEquality()) return false;
   // maxLit is a positive and is not equality
   for (TMP_LITERAL** lit = nonmax_lits.Last(); lit >= nonmax_lits.First(); lit--)
    {
     //if ((*lit)->Negative())
     if (NegativeForSelection((*lit)->Header()))
      {
       if ((*lit)->IsEquality() || ((*lit)->size() < maxLitWeight)) return false;
      };
    };
  }
 else // !PositiveForSelection(maxLit->Header())
  return false;
 return true;
}; // bool NewClause::satisfactorySelection2(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)


inline bool NewClause::satisfactorySelection3(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
{
 CALL("satisfactorySelection3(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)");
 if (max_lits.Length() > 1) return false;
 // max_lits.Length() == 1 
 TMP_LITERAL* maxLit = *(max_lits.First());
 //if (maxLit->Positive())
 if (PositiveForSelection(maxLit->Header()))
  {
   if (maxLit->IsEquality()) return false;
   // maxLit is a positive and is not equality
   ulong numOfDiffVars = maxLit->Variables().NumOfDifferent();
   static TMP_LITERAL::VarDepthProfile maxVarDepthProfile;
   maxLit->collectVarDepthProfile(maxVarDepthProfile);
   ulong maxLitWeight = maxLit->size();
   for (TMP_LITERAL** lit = nonmax_lits.Last(); lit >= nonmax_lits.First(); lit--)
    {
     if (NegativeForSelection((*lit)->Header()))
      {
       if ((*lit)->Variables().NumOfDifferent() > numOfDiffVars) return false;
       if ((*lit)->Variables().NumOfDifferent() == numOfDiffVars)
	{
	 // compare the depths of variable occurences
         static TMP_LITERAL::VarDepthProfile varDepthProfile;
         (*lit)->collectVarDepthProfile(varDepthProfile);
         switch (maxVarDepthProfile.compareByInverselyInducedOrder(varDepthProfile))
          {
	   case Less: return false;
	   case Equal: 
	    // compare the sizes
	    if (maxLitWeight < (*lit)->size()) return false;
	    break;
           #ifdef DEBUG_NAMESPACE
	    case Greater: break; 
	    default: ICP("ICP0"); break;
           #else 
            #ifdef _SUPPRESS_WARNINGS_
	     default: break;
            #endif 
           #endif         
          };
        };
      };
    };
  }
 else // !PositiveForSelection(maxLit->Header())
  return false;
 return true;
}; // bool NewClause::satisfactorySelection3(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)



inline bool NewClause::PositiveForSelection(const TERM& header)
{
 if (header.IsEquality() || header.IsSplitting() || (!_inversePolarityForSelection)) return header.Positive();
 return header.Negative();
}; 

inline bool NewClause::NegativeForSelection(const TERM& header)
{
 if (header.IsEquality() || header.IsSplitting()  || (!_inversePolarityForSelection)) return header.Negative();
 return header.Positive();
}; 

inline void NewClause::select(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)
{
 CALL("select(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)");
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 if (_splittingHasBeenUsed)
  {
   switch (_splittingStyle)
    {
     case 0L: selectWithPragmaticSplitting(lits,max_lits,nonmax_lits,negSelUsed); return;
     case 1L: selectWithNonblockingSplitting(lits,max_lits,nonmax_lits,negSelUsed); return;
     case 2L: selectWithBlockingSplitting(lits,max_lits,nonmax_lits,negSelUsed); return;
     default: 
      RuntimeError::report("Wrong splitting style number."); 
      return;
    };
  };
 // no splitting
 TMP_LIT_LIST::STAT stat; 
 lits.CollectStat(stat,_inversePolarityForSelection);
 _select(lits,max_lits,nonmax_lits,negSelUsed,stat);
}; // void NewClause::select(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)


inline void NewClause::selectWithPragmaticSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)
{
 CALL("selectWithPragmaticSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)");
 selectWithNonblockingSplitting(lits,max_lits,nonmax_lits,negSelUsed);

 /**********************************
 // This is how it was before v1.38 
 TMP_LIT_LIST::STAT stat;
 lits.CollectStatWithSplitting(stat,_inversePolarityForSelection);
 _select(lits,max_lits,nonmax_lits,negSelUsed,stat); // temporary dummy implementation
 ************************************/
}; // void NewClause::selectWithPragmaticSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)

inline void NewClause::selectWithNonblockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)
{
 CALL("selectWithNonblockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)");
 ASSERT(lits.Length() > 1);
 TMP_LIT_LIST::STAT stat;
 lits.CollectStatWithSplitting(stat,_inversePolarityForSelection);
 if (lits.Length() > (stat.negative_splitting_branches + stat.positive_splitting_branches))
  {
   // a nonsplitting literal exist, all splitting branches are nonselected
   for (TMP_LITERAL** lref = lits.Last(); lref >= lits.First(); lref--)
    if ((*lref)->Header().IsSplitting())     
     {
      (*lref)->setInherentlySelected(false); 
      nonmax_lits.Push(*lref);
      lits.ReplaceByTop(lref);
     };
   ASSERT(lits.Nonempty());
   if (lits.Length() == 1) 
    {
     max_lits.Push(lits.Pop());
     //if (NegativeForSelection((*(max_lits.First()))->Header())) negSelUsed = true;
     negSelUsed = false;
    }
   else // lits.Length() > 1
    {
     stat.Reset(); 
     lits.CollectStat(stat,_inversePolarityForSelection);
     _select(lits,max_lits,nonmax_lits,negSelUsed,stat); 
    };
  }
 else // all literals are splitting branches, select the youngest among them  
  {
   TMP_LITERAL* youngestBranch = lits.Pop();
   ASSERT(youngestBranch->Header().IsSplitting());
   TMP_LITERAL* currLit;
   do 
    {
     currLit = lits.Pop();
     ASSERT(currLit->Header().IsSplitting());
     if (currLit->Header().functor() > youngestBranch->Header().functor())
      {
       youngestBranch->setInherentlySelected(false);
       nonmax_lits.Push(youngestBranch);
       youngestBranch = currLit;
      }
     else 
      {
       currLit->setInherentlySelected(false);
       nonmax_lits.Push(currLit);
      };
    }     
   while (lits.Nonempty());
   //if (NegativeForSelection(youngestBranch->Header())) negSelUsed = true;
   negSelUsed = false;
   max_lits.Push(youngestBranch);
 };
}; // void NewClause::selectWithNonblockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)


inline void NewClause::selectWithBlockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)
{
 CALL("selectWithBlockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)");
 ASSERT(lits.Length() > 1);
 TMP_LIT_LIST::STAT stat;
 lits.CollectStatWithSplitting(stat,_inversePolarityForSelection);
 if (stat.negative_splitting_branches) // a negative branch exists, select the youngest among the negative branches
  {
   TMP_LITERAL* youngestNegBranch = (TMP_LITERAL*)0;
   TMP_LITERAL* currLit = lits.Pop(); 
   if (currLit->Header().IsSplitting() && NegativeForSelection(currLit->Header())) 
    { 
     youngestNegBranch = currLit;
    }
   else 
    {
     currLit->setInherentlySelected(false);
     nonmax_lits.Push(currLit);
    };
    
   do
    {   
     currLit = lits.Pop();
     if (currLit->Header().IsSplitting() && NegativeForSelection(currLit->Header())) 
      {
       if (!youngestNegBranch) 
        { 
         youngestNegBranch = currLit; 
        }
       else 
        if (currLit->Header().functor() > youngestNegBranch->Header().functor())
	 {
          youngestNegBranch->setInherentlySelected(false);
          nonmax_lits.Push(youngestNegBranch);
          youngestNegBranch = currLit; 
         }
        else 
         {
          currLit->setInherentlySelected(false);
          nonmax_lits.Push(currLit);
         };
       }
      else 
       {
        currLit->setInherentlySelected(false);
        nonmax_lits.Push(currLit);
       };
     }
    while (lits.Nonempty());
    max_lits.Push(youngestNegBranch);
    //negSelUsed = true;
    negSelUsed = false;
    // that's it
   }
  else // no negative splitting branches 
   {
    if (lits.Length() > stat.positive_splitting_branches) // a literal with input predicate exists
     {
      // all splitting literals are nonselected 
      for (TMP_LITERAL** lref = lits.Last(); lref >= lits.First(); lref--)
       if ((*lref)->Header().IsSplitting())     
	{ 
         (*lref)->setInherentlySelected(false);
	 nonmax_lits.Push(*lref);
         lits.ReplaceByTop(lref);
        };

      ASSERT(lits.Nonempty());
      if (lits.Length() == 1) 
       {
  	max_lits.Push(lits.Pop());
        //if (NegativeForSelection((*(max_lits.First()))->Header())) negSelUsed = true;
        negSelUsed = false;
       }
      else // lits.Length() > 1
       { 
	stat.Reset();
        lits.CollectStat(stat,_inversePolarityForSelection);
        _select(lits,max_lits,nonmax_lits,negSelUsed,stat); 
       };

     }
    else // lits contains only positive splitting branches, select the youngest among them
     {
      TMP_LITERAL* youngestPosBranch = lits.Pop();
      ASSERT(PositiveForSelection(youngestPosBranch->Header()));
      ASSERT(youngestPosBranch->Header().IsSplitting());
      TMP_LITERAL* currLit;
      do 
       {
        currLit = lits.Pop();
        ASSERT(PositiveForSelection(currLit->Header()));
        ASSERT(currLit->Header().IsSplitting());
        if (currLit->Header().functor() > youngestPosBranch->Header().functor())
	 {
          youngestPosBranch->setInherentlySelected(false);
	  nonmax_lits.Push(youngestPosBranch);
          youngestPosBranch = currLit;
         }
        else 
	 {
          currLit->setInherentlySelected(false);
	  nonmax_lits.Push(currLit);
         };
       }     
      while (lits.Nonempty());
      negSelUsed = false;
      max_lits.Push(youngestPosBranch);
    };
  };
};  // void NewClause::selectWithBlockingSplitting(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,bool& negSelUsed)


inline void NewClause::selectNegativeWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
{
 CALL("selectNegativeWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)");

 // lits contains at least 1 negative literal
 // find first negative literal
 TMP_LITERAL* cand = lits.Pop();
 while (PositiveForSelection(cand->Header()))
  {         
   nonmax_lits.Push(cand);
   ASSERT(lits.Nonempty());
   cand = lits.Pop(); 
  };
 ulong cand_weight = cand->size();
 TMP_LITERAL* next_lit;
 while (lits.Nonempty())
  {
   next_lit = lits.Pop();
   if (NegativeForSelection(next_lit->Header()) && (next_lit->size() > cand_weight)) 
    {
     nonmax_lits.Push(cand);
     cand = next_lit;
     cand_weight = cand->size();
    }
   else nonmax_lits.Push(next_lit);
  };
 max_lits.Push(cand);    
}; // void NewClause::selectNegativeWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)

inline void NewClause::selectNegativeLeastProlific(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
{
 CALL("selectNegativeLeastProlific(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)");

 // lits contains at least 1 negative literal
 // find first negative literal
 TMP_LITERAL* cand = lits.Pop();
 while (PositiveForSelection(cand->Header()))
  {         
   nonmax_lits.Push(cand);
   ASSERT(lits.Nonempty());
   cand = lits.Pop(); 
  };

 TMP_LITERAL* next_lit;
 while (lits.Nonempty())
  {
   next_lit = lits.Pop();
   if (NegativeForSelection(next_lit->Header()) && (next_lit->PreferredNegative(*cand)))  
    {
     nonmax_lits.Push(cand);
     cand = next_lit;
    }
   else nonmax_lits.Push(next_lit);
  };
 max_lits.Push(cand);     
}; // void NewClause::selectNegativeLeastProlific(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)  


inline void NewClause::selectNegativeMaximallyNongroundWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
{
 CALL("selectNegativeMaximallyNongroundWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)");

 // first, get rid of all positive literals
 for (TMP_LITERAL** lit = lits.Last(); lit >= lits.First(); lit--)
  if (PositiveForSelection((*lit)->Header()))
   {
    nonmax_lits.Push(*lit);
    lits.ReplaceByTop(lit);   
   };
 
 // lits contains at least 1 negative literal
 ASSERT(lits.Length());
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
 ulong maxNumOfDiffVars = (*(lits.Last()))->Variables().NumOfDifferent();

 // get rid of the literals that do not have the maximal number of different variables
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   // compare the number of different variables 
   ulong numOfDiffVars = (*lit)->Variables().NumOfDifferent();
   if (numOfDiffVars > maxNumOfDiffVars)
    {
     maxNumOfDiffVars = numOfDiffVars;
     do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
    }
   else 
    if (numOfDiffVars < maxNumOfDiffVars)
     {
      nonmax_lits.Push(*lit);
      lits.ReplaceByTop(lit);
     };
  };
 // now all the literals have the maximal number of different variables 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; }; 

 // additionally, compare the multisets of depths of variable occurences    
 
 static TMP_LITERAL::VarDepthProfile varDepthProfile1;
 static TMP_LITERAL::VarDepthProfile varDepthProfile2;
 TMP_LITERAL::VarDepthProfile* maxVarDepthProfile = &varDepthProfile1;
 TMP_LITERAL::VarDepthProfile* litVarDepthProfile = &varDepthProfile2; 
 (*(lits.Last()))->collectVarDepthProfile(*maxVarDepthProfile);
 
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   (*lit)->collectVarDepthProfile(*litVarDepthProfile); 
   // compare the profiles
   switch (maxVarDepthProfile->compareByInverselyInducedOrder(*litVarDepthProfile))
    {
     case Less:
      maxVarDepthProfile = litVarDepthProfile;

      litVarDepthProfile = ((litVarDepthProfile == &varDepthProfile1) ? &varDepthProfile2 : &varDepthProfile1);
      do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);  
      break;   

     case Greater :  
      nonmax_lits.Push(*lit);
      lits.ReplaceByTop(lit);
      break;

     case Equal : break;
     
     #ifdef DEBUG_NAMESPACE
      default : ICP("ICP0"); return;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: RuntimeError::report("Illegal control point in  NewClause::selectNegativeMaximallyNongroundWithMaximalWeight(..)");
       return;
      #endif
     #endif 
    };
  };

 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; }; 

 // select literals with greater size

 long maxSize = (*(lits.Last()))->size();
 for (TMP_LITERAL** lit = lits.Last() - 1; lit >= lits.First(); lit--)
  {
   long litSize = (*lit)->size();
   if (litSize > maxSize)
    {
     maxSize = litSize;
     do { nonmax_lits.Push(lits.Pop()); } while (lits.Last() > lit);     
    }
   else 
    if (litSize < maxSize)
     {
      nonmax_lits.Push(*lit);
      lits.ReplaceByTop(lit);
     };
  };
 // now all the literals in lits are of the greatest size 
 if (lits.Length() == 1) { max_lits.Push(lits.Pop()); return; };
  
 // now select on "don't care" basis
 max_lits.Push(lits.Pop());
 while (lits.Nonempty()) nonmax_lits.Push(lits.Pop());
 
}; // void NewClause::selectNegativeMaximallyNongroundWithMaximalWeight(TMP_LIT_LIST& lits,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)




inline void NewClause::selectInherentlySelected(TMP_LIT_LIST& lits,ulong numOfSelected,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)
{
 CALL("selectInherentlySelected(TMP_LIT_LIST& lits,ulong numOfSelected,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)");
 ASSERT(numOfSelected); 
 ASSERT(lits.Length() >= numOfSelected);
 do
  {
   ASSERT(lits.Length() >= numOfSelected);
   TMP_LITERAL* lit = lits.Pop();
   if (lit->isInherentlySelected()) { max_lits.Push(lit); numOfSelected--; }
   else nonmax_lits.Push(lit);    
  }
 while (numOfSelected);
 
 while (lits.Nonempty()) 
  {
   #ifdef DEBUG_NAMESPACE
    TMP_LITERAL* lit = lits.Pop();
    ASSERT(!lit->isInherentlySelected());
    nonmax_lits.Push(lit); 
   #else  
    nonmax_lits.Push(lits.Pop()); 
   #endif
  };
 
}; // void NewClause::selectInherentlySelected(TMP_LIT_LIST& lits,ulong numOfSelected,TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits)






//===========================================================================
#undef DEBUG_NAMESPACE
//=========================================================================
#endif
