#ifndef OPEN_NEW_CLAUSE_H
//===============================================================
#define OPEN_NEW_CLAUSE_H
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"
#include "variables.h"
#include "weighting.h"
#include "Flatterm.h"
#include "tmp_literal.h"
#include "tmp_lit_list.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OPEN_NEW_CLAUSE
 #define DEBUG_NAMESPACE "OpenNewClause"
#endif 
#include "debugMacros.h"
//=================================================

class OpenNewClause
 {
  public:
   struct Statistics
   { 
    public: 
     Statistics() { reset(); };
     void reset() 
     {
      tooHeavy = 0;
      tooHeavyLiterals = 0;
      tooManyVariables = 0;
      tooLong = 0;
      tooDeep = 0;
      propTautologies = 0;
      eqTautologies = 0;  
     };
     ulong tooHeavy;
     ulong tooHeavyLiterals;
     ulong tooManyVariables;
     ulong tooLong;
     ulong tooDeep;
     ulong propTautologies; 
     ulong eqTautologies;       
   };
  public:
   Statistics statistics;
  private:
   TMP_LITERAL literal[HellConst::MaxNumOfLiterals + 1];
   TMP_LITERAL* next_literal;
   const TMP_LITERAL* last_possible_literal; 
   TMP_LIT_LIST all_literals;
   TMP_LITERAL* spare_literal;
   ulong weight;
   bool was_tautology;
   ulong num_of_unordered_eq_lits;
  private: // limits
   ulong max_weight;
   ulong max_lit_weight;
   ulong max_length;  
  public:
   OpenNewClause() 
    :
     statistics(),
     last_possible_literal(literal + (HellConst::MaxNumOfLiterals - 1)),
     all_literals(),
     max_weight(HellConst::MaxClauseSize),
     max_lit_weight(HellConst::MaxTermSize),
     max_length(HellConst::MaxNumOfLiterals)
    {
    };
   ~OpenNewClause() { CALL("destructor ~OpenNewClause()"); };
   void Reset()
    {
     next_literal = literal;
     spare_literal = literal + HellConst::MaxNumOfLiterals;
     weight = 0;
     was_tautology = false;
     num_of_unordered_eq_lits = 0;
    }; 
  
   void AnotherTooHeavyClause() { statistics.tooHeavy++;  };
   void AnotherTooHeavyTerm() { statistics.tooHeavyLiterals++; };
   void AnotherClauseWithTooManyVariables() { statistics.tooManyVariables++; };
   void AnotherTooLongClause() { statistics.tooLong++; };
   void AnotherTooDeep() { statistics.tooDeep++; };
   void AnotherPropTautology() { was_tautology = true; statistics.propTautologies++; };
   void AnotherEqTautology() { was_tautology = true; statistics.eqTautologies++; };
   void SetMaxWeight(ulong mw) 
   { 
    max_weight = (mw > HellConst::MaxClauseSize) ? HellConst::MaxClauseSize : mw; 
   };
   void SetMaxLitWeight(ulong mlw) 
   { 
    max_lit_weight = (mlw > HellConst::MaxTermSize) ? HellConst::MaxTermSize : mlw; 
   };
   void SetMaxLength(ulong ml) 
   { 
    max_length = (ml > HellConst::MaxNumOfLiterals) ? HellConst::MaxNumOfLiterals : ml; 
   };  
   void SetMaxDepth(ulong md) 
   {
    for (TMP_LITERAL* lit = literal; lit <= last_possible_literal; lit++) 
     lit->SetMaxDepth(md);
   }; 
    
   ulong Length() const { return next_literal - literal; };
   //bool Empty() const { return next_literal == literal; };
   bool Unit() const { return (Length() == 1); };
   ulong Weight() const { return weight; };
   const bool& WasTautology() const { return was_tautology; };
     
   TMP_LIT_LIST& AllLiterals() { return all_literals; };
   TMP_LITERAL*& SpareLiteral() { return spare_literal; }; 
     
   // collecting a new clause
   
   //void OpenLiteral() { next_literal->Reset(); };
   
   bool OpenNonpropLiteral(const TERM& header,bool inherentlySelected)
    {
     CALL("OpenNonpropLiteral(const TERM& header)");
     // header can't be a propositional literal
     ASSERT(header.arity());
     weight += (header.arity() + 1);
     if (weight > max_weight)  
      {
       AnotherTooHeavyClause();
       return false;
      };
     next_literal->Reset();  
     next_literal->PushNonpropHeader(header);
     next_literal->setInherentlySelected(inherentlySelected); 
     return true;
    };
   
   bool CloseEqLiteral()
    {
     CALL("CloseEqLiteral()");

     //cout << "CLOSE EQ LIT " << next_literal << "\n";

     ASSERT(next_literal->IsEquality());
     // try to reduce equality or detect a tautology
     if (next_literal->EqEqualArgs())
      {
       if (next_literal->Positive())
        {
         AnotherEqTautology();
         return false;
        }
       else 
        {
         weight -= next_literal->size();

	 //cout << "EQ RES SIMP\n";

         return true;
        }; 
      };
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++)
      {
       if (lit->IsEquality())
        {
         if (next_literal->EqualEqLit(*lit)) 
          {
           weight -= next_literal->size();


	   //cout << "FACTORED\n";

           return true;
          };
          
         if (next_literal->ComplementaryEqLit(*lit))
          {
           AnotherPropTautology();
           return false;
          };
        };  
      }; 
     if ((next_literal->size() > max_lit_weight) || (weight > max_weight))
      {
       AnotherTooHeavyClause();
       return false;
      };
     if (Length() >= max_length) 
      {
       AnotherTooLongClause();
       return false;
      }; 
     if (!next_literal->IsOrderedEq()) num_of_unordered_eq_lits++;
     next_literal++;

     //cout << "ORDINARY\n";

     return true; 
    }; // bool CloseEqLiteral()
   
   bool CloseOrdinaryLiteral()
    {
    // check for identical or complementary literals
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++)
      {
       if (next_literal->equal(*lit)) 
        {
         weight -= next_literal->size();
         return true;
        }; 
       if (next_literal->Complementary(*lit))
        {
         AnotherPropTautology();
         return false;
        };
      };
     if ((next_literal->size() > max_lit_weight) || (weight > max_weight))
      {
       AnotherTooHeavyClause();
       return false;
      };
     if (Length() >= max_length) 
      {
       AnotherTooLongClause();
       return false;
      }; 
     next_literal++;
     return true;
    }; // bool CloseOrdinaryLiteral()
   
   bool CloseLiteral()
    {
     if (next_literal->IsEquality()) 
      {
       return CloseEqLiteral();
      };
     
     return CloseOrdinaryLiteral();
    }; // bool CloseLiteral() 
   
   void CollectLiterals()
    {
     all_literals.Reset();
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++) 
      {
       all_literals.Push(lit);
      }; 
    };

   void EndOfClause()
    {
     all_literals.Reset();
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++) 
      {
       if (lit->IsEquality()) lit->EqOrderArgs();
       all_literals.Push(lit);
      }; 
    };


  bool Weed()
   {
    // can be used after simplifications
    bool eqTautology;
    if (all_literals.Weed(eqTautology)) return true;
    if (eqTautology) { AnotherEqTautology(); }
    else AnotherPropTautology(); 
    return false; 
   }; 
     
  bool Weed(const Clause* const * commutativity_table)
  {
   // can be used after simplifications
   bool eqTautology;
   if (!commutativity_table) return Weed();
   if (all_literals.Weed(commutativity_table,eqTautology)) return true;
   if (eqTautology) { AnotherEqTautology(); }
   else AnotherPropTautology(); 
   return false;    
  };

   bool PushPropLit(const TERM& header,bool inherentlySelected)
    {
     TERM compl_header = header;
     compl_header.Inversepolarity();
     for (TMP_LITERAL* lit = literal; lit < next_literal; lit++)
      {
       if (lit->FlatTerm()->Symbol() == header) return true;
       if (lit->FlatTerm()->Symbol() == compl_header)
        {
         AnotherPropTautology();
         return false;
        };
      };
    
     weight++;
     if (weight > max_weight)
      {
       AnotherTooHeavyClause();
       return false;
      };
     if (Length() >= max_length) 
      {
       AnotherTooLongClause();
       return false;
      };
     next_literal->Reset(); 
     next_literal->PushPropLit(header); 
     next_literal->setInherentlySelected(inherentlySelected);
     next_literal++;
     return true;
    }; 
   bool PushVar(ulong v) 
    {
     // this check is too expensive and must be eliminated in the future!
     if (v >= HellConst::MaxNumOfVariables) 
      {
       AnotherClauseWithTooManyVariables();
       return false; 
      };
     next_literal->PushVar(v);  
     return true;  
    };
   void PushVar(const TERM& v) { next_literal->PushVar(v); };  
   void PushConst(const TERM& c) { next_literal->PushConst(c); };
   bool PushNonconstFun(const TERM& f) 
    {
     weight += f.arity();
     if (weight > max_weight) 
      {
       AnotherTooHeavyClause();
       return false;
      };
     
     if (!next_literal->PushNonconstFun(f))
      {
       AnotherTooDeep();
       return false;
      };
      
     if (next_literal->size() > max_lit_weight)
      {
       AnotherTooHeavyTerm();
       return false;
      }; 
      
     return true;
    };
    
   public:
    ostream& output(ostream& str) const
    {
     str << "{ ";
     for (const TMP_LITERAL* lit = literal; lit < next_literal; lit++)
      {
       str << lit; if (lit + 1 < next_literal) str << " | ";
      };
     return str << " }";     
    };
 }; // class OpenNewClause

inline ostream& operator<<(ostream& str,const OpenNewClause& cl) 
{
 return cl.output(str);
};

inline ostream& operator<<(ostream& str,const OpenNewClause* cl) 
{
 return cl->output(str);
};


//================================================
#undef DEBUG_NAMESPACE 
//================================================
#endif
