 //
// File:         SharedKBConstr.cpp
// Description:  Representation of simple shared KB ordering constrains.
// Created:      May 8, 2000, 10:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "SharedKBConstr.h"
//====================================================
const TERM* SharedKBConstr::_lhs;
const TERM* SharedKBConstr::_rhs;
KBConstraint SharedKBConstr::_raw;
SharedKBConstr::WeightFormSharing SharedKBConstr::_weightFormSharing;
#ifdef DEBUG_ALLOC_OBJ_TYPE
 ClassDesc SharedKBConstr::_classDesc("SharedKBConstr",SharedKBConstr::minSize(),SharedKBConstr::maxSize());
 ClassDesc SharedKBConstr::WeightForm::_classDesc("SharedKBConstr::WeightForm",SharedKBConstr::WeightForm::minSize(),SharedKBConstr::WeightForm::maxSize());
 ClassDesc SharedKBConstr::WeightFormSharing::DiscTree::Node::_classDesc("SharedKBConstr::WeightFormSharing::DiscTree::Node",
                                                                         SharedKBConstr::WeightForm::minSize(),
                                                                         SharedKBConstr::WeightForm::maxSize());
#endif

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SHARED_KB_CONSTR
 #define DEBUG_NAMESPACE "SharedKBConstr"
#endif 
#include "debugMacros.h"
//================================================= 

void SharedKBConstr::assumeGreater(const TERM* lhs,
                                   MultisetOfVariables<HellConst::MaxNumOfVariables>& lhsVars,
                                   ulong lhsFuncs,
                                   const TERM* rhs,
                                   MultisetOfVariables<HellConst::MaxNumOfVariables>& rhsVars,
                                   ulong rhsFuncs)
{  
 _raw.assumeGreater(lhs,lhsVars,lhsFuncs,rhs,rhsVars,rhsFuncs);   
 _lhs = lhs;
 _rhs = rhs;
};

SharedKBConstr* SharedKBConstr::getShared()
{
 CALL("getShared()");
 switch (_raw.category())
 {
  case KBConstraint::NearlyTautology: return getSharedNearlyTautology();
  case KBConstraint::NearlyContradiction: return getSharedNearlyContradiction();
  case KBConstraint::Volatile: return getSharedVolatile();
  case KBConstraint::PurelyLexicographic: return getSharedPurelyLexicographic();
 };
 #ifdef DEBUG_NAMESPACE
  ICP("ICP0");
  return (SharedKBConstr*)0;
 #else 
  #ifdef _SUPPRESS_WARNINGS_
   return (SharedKBConstr*)0;
  #endif
 #endif
}; // SharedKBConstr* SharedKBConstr::getShared()

void SharedKBConstr::destroySharingIndex() 
{
 CALL("destroySharingIndex()");
 _weightFormSharing.destroy();  
}; // void SharedKBConstr::destroySharingIndex() 



//====================================================
