#include "NewClause.h"
#include "HellDebugFlags.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEW_CLAUSE
 #define DEBUG_NAMESPACE "NewClause"
#endif 
#include "debugMacros.h"
//=================================================

bool NewClause::BlockingSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 CALL("BlockingSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses)");

 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
 _divisionIntoComponents.reset();
 _basicComponent.Reset();
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong litNum;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   if (lit->Header().IsSplitting())
    {
     // this literal represents a branch for previously made splitting 
     ASSERT(TheSignature.isSkolem(lit->Header().functor()));
     ASSERT(!(lit->Variables().NumOfDifferent()));
     ASSERT(!lit->Header().arity());
     if (lit->Header().polarity() == basicCompPol) return false; // the clause is blocked
     ALWAYS(_basicComponent.PushPropLit(lit->Header(),false));
    }
   else
    {
     groundLit = !(lit->Variables().NumOfDifferent());
     _litNumMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = litNum;
     lit->Variables().ResetIterator();
     while ((varp = lit->Variables().NextVar())) 
      {
       _divisionIntoComponents.nextVar(*varp);     
      };
    };        
  }; 
 
 if (_divisionIntoComponents.numOfComponents() <= 1) 
   // splitting makes no sense 
  {
   return false; 
  };  

 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();

 ulong compNum;

 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(_litNumMap[litNum]));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 TERM header;
 ulong numOfNewComps = 0;
 ulong n;
 ulong numOfNamedComponents = 0;

 // identify components that have already been given names

 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // this component has already been given name   
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     
     if (!_basicComponent.PushPropLit(header,false))
      {
       // discarded (tautalogy or too heavy)  
       return true;
      };

     numOfNamedComponents++;
    };
  };

 if (numOfNamedComponents) 
  {
   _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

   // collect other components in _basicComponent  
   for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
    {
     compn = _registeredCompNums.nth(n);  
     if (_newCompName[compn])
      while (_components[compn].Nonempty())
       { 
        _basicComponent.AllLiterals().Push(_components[compn].Pop());
       };     
    };

   // make a clause of _basicComponent.AllLiterals() 
   if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
    { 
     AnotherForwardSubsumed();
     return true; 
    };

   if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
    {
     MakeVIP();
     AnotherSimplifiedByFSR();
    };

   UsedRule(ClauseBackground::Split);
   Clause* cl = MakeSharedClause(_basicComponent.AllLiterals());
   cl->setMainSet(Clause::MainSetNew);
   newClauses.enqueue(cl); 

   AnotherSplit(numOfNamedComponents); 

   return true;   
  };

 // !namedComponentExists here

 if ((TheSignature.nextFunctorNumber() + 1) > HellConst::MaxNumOfFunctors) // can not create a new name
  return false;

 
 UsedRule(ClauseBackground::Split);

 // now, using heuristics, select a component to become the open one 
 
 ulong openComponentNum = _registeredCompNums.nth(0); // temporary simple solution, most certainly not the best one
 
 // give a name to the open component

 _componentName[openComponentNum] = TheSignature.newSkolemPredicateNumber(0UL);
 ASSERT(_componentName[openComponentNum] < HellConst::MaxNumOfFunctors);
 compVars.Reset();
 _components[openComponentNum].CollectVariables(compVars);
 
 if (_splittingWithNamingFlag)
  _compName->integrate(_components[openComponentNum],compVars,_componentName[openComponentNum]);

 header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[openComponentNum]),basicCompPol);
 header.MakeSplitting();
 GlobalHeaderWeighting.MakeMinimal(header);


 ALWAYS(_basicComponent.PushPropLit(header,false));


 header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[openComponentNum]),nonBasicCompPol);
 header.MakeSplitting();
 GlobalHeaderWeighting.MakeMinimal(header);
 _compNameLit.Reset();
 
 _compNameLit.PushPropLit(header);

 _components[openComponentNum].Push(&_compNameLit);
 Clause* cl = MakeSharedClause(_components[openComponentNum]); // the definition clause
 cl->MarkAsName();
 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl);    
 AnotherComponent(); 
 

 _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

 // collect other (blocked)  components in _basicComponent  
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);  
   ASSERT(_newCompName[compn]);
   if (compn != openComponentNum)
    while (_components[compn].Nonempty())
     { 
      _basicComponent.AllLiterals().Push(_components[compn].Pop());
     };     
  };

 // make a clause of _basicComponent.AllLiterals() 

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };

 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
  };

 cl = MakeSharedClause(_basicComponent.AllLiterals());
 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl); 
 AnotherSplit(2);
 return true;   

}; // bool NewClause::BlockingSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses)


bool NewClause::NonblockingSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 CALL("NonblockingSplitting(TMP_LIT_LIST& lits)");

 //DF; cout << "NAMING " << _splittingWithNamingFlag << "\n";

 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
 _divisionIntoComponents.reset();
 _basicComponent.Reset();
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong litNum;
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   if (lit->Header().IsSplitting())
    {
     // this literal represents a previously made splitting branch
     ASSERT(TheSignature.isSkolem(lit->Header().functor()));
     ASSERT(!(lit->Variables().NumOfDifferent()));
     ASSERT(!lit->Header().arity());
     ALWAYS(_basicComponent.PushPropLit(lit->Header(),false));
    }
   else
    {
     groundLit = !(lit->Variables().NumOfDifferent());
     _litNumMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = litNum;
     lit->Variables().ResetIterator();
     while ((varp = lit->Variables().NextVar())) 
      {
       _divisionIntoComponents.nextVar(*varp);     
      };
    };        
  }; 
 
 if (_divisionIntoComponents.numOfComponents() <= 1) 
   // splitting makes no sense 
  return false; 

  
 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();

 ulong compNum;

 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(_litNumMap[litNum]));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 TERM header;
 ulong numOfNewComps = 0;
 ulong n;

 // identify components that have already been given names
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);

   //DF; _components[compn].output(cout << "COMP ") << "\n";
 
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));

   //DF; cout << "NEW " << _newCompName[compn] << "\n";

   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // this component has already been given name   
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     if (!_basicComponent.PushPropLit(header,false))
      {  
       // tautology
       return true;
      };
    };
  };

 _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };
 
 // process new components giving new names to them
 bool basicComponentChosen = false;
 ulong chosenBasicComponent;
 #ifdef _SUPPRESS_WARNINGS_ 
  chosenBasicComponent = 0;
 #endif 
  
 DOP(chosenBasicComponent = ULONG_MAX);
 UsedRule(ClauseBackground::Split);
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   if (_newCompName[compn]) 
    {
      // this could change since we could have several components that are variants of each other, 
      // simply check this condition again to avoid duplicate names 
     compVars.Reset();
     _components[compn].CollectVariables(compVars);
     _newCompName[compn] = 
      (!_splittingWithNamingFlag) 
      || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
    };

   if (_newCompName[compn])
    {
     if (basicComponentChosen) 
      {
       _componentName[compn] = TheSignature.newSkolemPredicateNumber(0UL);
       ASSERT(_componentName[compn] < HellConst::MaxNumOfFunctors);
       compVars.Reset();
       _components[compn].CollectVariables(compVars);

       if (_splittingWithNamingFlag)
        _compName->integrate(_components[compn],compVars,_componentName[compn]);

       header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
       header.MakeSplitting();
       GlobalHeaderWeighting.MakeMinimal(header);

       ALWAYS(_basicComponent.PushPropLit(header,false));

       header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),nonBasicCompPol);
       header.MakeSplitting();
       GlobalHeaderWeighting.MakeMinimal(header);
       _compNameLit.Reset();
       _compNameLit.PushPropLit(header);
       _components[compn].Push(&_compNameLit);
       Clause* cl = MakeSharedClause(_components[compn]); // the definition clause
       cl->MarkAsName();
       cl->setMainSet(Clause::MainSetNew);
       newClauses.enqueue(cl);    
       AnotherComponent();
      }
     else // this component is chosen to be the basic one
      { 
       basicComponentChosen = true;
       chosenBasicComponent = compn;
      };
    };
  };

 _basicComponent.CollectLiterals(); // this is no problem as far as nothing has been pushed into _basicComponent.AllLiterals()

 if (basicComponentChosen)
  {
   // add all the literals from _components[chosenBasicComponent] to the basic component

   while (_components[chosenBasicComponent].Nonempty()) 
    _basicComponent.AllLiterals().Push(_components[chosenBasicComponent].Pop());
  };

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };

 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
  };

 Clause* cl = MakeSharedClause(_basicComponent.AllLiterals());
 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl); 

 AnotherSplit(_divisionIntoComponents.numOfComponents());

 return true;
}; // bool NewClause::NonblockingSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses)


bool NewClause::PragmaticSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 CALL("PragmaticSplitting(TMP_LIT_LIST& lits)");
 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(_optimalGroundCompWeight);
 _divisionIntoComponents.reset();
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong groundWeight = 0; 
 ulong litNum;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   groundLit = !(lit->Variables().NumOfDifferent());
   if (groundLit) groundWeight = lit->size();
   _divisionIntoComponents.nextLit(groundLit,groundWeight);
   lit->Variables().ResetIterator();
   while ((varp = lit->Variables().NextVar())) 
    {
     _divisionIntoComponents.nextVar(*varp);     
    };        
  }; 
 
 ASSERT(_divisionIntoComponents.numOfComponents());

 if (_divisionIntoComponents.numOfComponents() == 1) 
   // splitting makes no sense 
  return false; 
  
 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();
 ulong compNum;
 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(litNum));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 _basicComponent.Reset();
 TERM header;
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;
 ulong numOfNewComps = 0;
 ulong n;

 // identify already known components
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      {
	

	//DF; cout << "AGAAAAAAAAAAAAAA\n";

       return false;
      };
    }
   else // known component
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol); 
     header.MakeSplitting();
     if (!_basicComponent.PushPropLit(header,false)) 
      {
	// discarded 
       return true;
      };
    };
  };
 _basicComponent.CollectLiterals();
 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };
 
 // process new components
 UsedRule(ClauseBackground::Split);
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   if (_newCompName[compn]) 
    {
      // this could change, check it again 
     compVars.Reset();
     _components[compn].CollectVariables(compVars);     

     _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
    };
   if (_newCompName[compn])
    {
     _componentName[compn] = TheSignature.newSkolemPredicateNumber(0UL);
     //RegisterAsSplitting(_componentName[compn]);
     ASSERT(_componentName[compn] < HellConst::MaxNumOfFunctors);
     compVars.Reset();
     _components[compn].CollectVariables(compVars);

     if (_splittingWithNamingFlag)
      _compName->integrate(_components[compn],compVars,_componentName[compn]);

     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     GlobalHeaderWeighting.MakeMinimal(header);
     ALWAYS(_basicComponent.PushPropLit(header,false));
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),nonBasicCompPol);
     header.MakeSplitting();
     GlobalHeaderWeighting.MakeMinimal(header);
     _compNameLit.Reset();
     _compNameLit.PushPropLit(header);
     _components[compn].Push(&_compNameLit);

     Clause* cl = MakeSharedClause(_components[compn]);

     cl->MarkAsName();
     cl->setMainSet(Clause::MainSetNew);
     newClauses.enqueue(cl);    
     AnotherComponent();
    };
  };

 _basicComponent.CollectLiterals();
 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
  };

 Clause* cl = MakeSharedClause(_basicComponent.AllLiterals());

 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl); 

 AnotherSplit(_divisionIntoComponents.numOfComponents());

 return true;
}; // bool NewClause::PragmaticSplitting(TMP_LIT_LIST& lits,ClauseQueue& newClauses)



bool NewClause::BlockingSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 CALL("BlockingSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)");

 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
 _divisionIntoComponents.reset();
 _basicComponent.Reset();
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong litNum;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   if (lit->Header().IsSplitting())
    {
     // this literal represents a branch for previously made splitting 
     ASSERT(TheSignature.isSkolem(lit->Header().functor()));
     ASSERT(!(lit->Variables().NumOfDifferent()));
     ASSERT(!lit->Header().arity());
     if (lit->Header().polarity() == basicCompPol) return false; // the clause is blocked
     ALWAYS(_basicComponent.PushPropLit(lit->Header(),false));
    }
   else
    {
     groundLit = !(lit->Variables().NumOfDifferent());
     _litNumMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = litNum;
     lit->Variables().ResetIterator();
     while ((varp = lit->Variables().NextVar())) 
      {
       _divisionIntoComponents.nextVar(*varp);     
      };
    };        
  }; 
 
 if (_divisionIntoComponents.numOfComponents() <= 1) 
   // splitting makes no sense 
  {
   return false; 
  };  

 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();

 ulong compNum;

 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(_litNumMap[litNum]));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 TERM header;
 ulong numOfNewComps = 0;
 ulong n;
 ulong numOfNamedComponents = 0;

 // identify components that have already been given names

 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // this component has already been given name   
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     
     if (!_basicComponent.PushPropLit(header,false))
      {
       // tautology
       return true;
      };
     numOfNamedComponents++;
    };
  };

 if (numOfNamedComponents) 
  {
   _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

   // collect other components in _basicComponent  
   for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
    {
     compn = _registeredCompNums.nth(n);  
     if (_newCompName[compn])
      while (_components[compn].Nonempty()) 
       _basicComponent.AllLiterals().Push(_components[compn].Pop());     
    };

   // make a clause of _basicComponent.AllLiterals() 
   if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
    { 
     AnotherForwardSubsumed();
     return true; 
    };

   if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
    {
     MakeVIP();
     AnotherSimplifiedByFSR();
    };

   UsedRule(ClauseBackground::Split);
   Clause* cl = MakeFastUnsharedClause(_basicComponent.AllLiterals());
   cl->setMainSet(Clause::MainSetNew);
   newClauses.enqueue(cl); 

   AnotherSplit(numOfNamedComponents); 

   return true;   
  };

 // !namedComponentExists here

 if ((TheSignature.nextFunctorNumber() + 1) > HellConst::MaxNumOfFunctors) // can not create a new name
  return false;

 
 UsedRule(ClauseBackground::Split);

 // now, using heuristics, select a component to become the open one 
 
 ulong openComponentNum = _registeredCompNums.nth(0); // temporary simple solution, most certainly not the best one
 
 // give a name to the open component

 _componentName[openComponentNum] = TheSignature.newSkolemPredicateNumber(0UL);
 ASSERT(_componentName[openComponentNum] < HellConst::MaxNumOfFunctors);
 compVars.Reset();
 _components[openComponentNum].CollectVariables(compVars);
 
 if (_splittingWithNamingFlag)
  _compName->integrate(_components[openComponentNum],compVars,_componentName[openComponentNum]);

 header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[openComponentNum]),basicCompPol);
 header.MakeSplitting();
 GlobalHeaderWeighting.MakeMinimal(header);
 ALWAYS(_basicComponent.PushPropLit(header,false));
 header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[openComponentNum]),nonBasicCompPol);
 header.MakeSplitting();
 GlobalHeaderWeighting.MakeMinimal(header);
 _compNameLit.Reset();
 _compNameLit.PushPropLit(header);
 _components[openComponentNum].Push(&_compNameLit);
 Clause* cl = MakeFastUnsharedClause(_components[openComponentNum]); // the definition clause
 cl->MarkAsName();
 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl);    
 AnotherComponent(); 
 

 _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

 // collect other (blocked)  components in _basicComponent  
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);  
   ASSERT(_newCompName[compn]);
   if (compn != openComponentNum)
    while (_components[compn].Nonempty()) 
     _basicComponent.AllLiterals().Push(_components[compn].Pop());     
  };

 // make a clause of _basicComponent.AllLiterals() 

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };

 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
  };

 cl = MakeFastUnsharedClause(_basicComponent.AllLiterals());
 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl); 
 AnotherSplit(2);
 return true;   

}; // bool NewClause::BlockingSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)


bool NewClause::NonblockingSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 CALL("NonblockingSplitting2(TMP_LIT_LIST& lits)");
 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
 _divisionIntoComponents.reset();
 _basicComponent.Reset();
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong litNum;
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   if (lit->Header().IsSplitting())
    {
     // this literal represents a previously made splitting branch
     ASSERT(TheSignature.isSkolem(lit->Header().functor()));
     ASSERT(!(lit->Variables().NumOfDifferent()));
     ASSERT(!lit->Header().arity());

     ALWAYS(_basicComponent.PushPropLit(lit->Header(),false));

    }
   else
    {
     groundLit = !(lit->Variables().NumOfDifferent());
     _litNumMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = litNum;
     lit->Variables().ResetIterator();
     while ((varp = lit->Variables().NextVar())) 
      {
       _divisionIntoComponents.nextVar(*varp);     
      };
    };        
  }; 
 
 if (_divisionIntoComponents.numOfComponents() <= 1) 
   // splitting makes no sense 
  return false; 

  
 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();

 ulong compNum;

 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(_litNumMap[litNum]));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 TERM header;
 ulong numOfNewComps = 0;
 ulong n;

 // identify components that have already been given names
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // this component has already been given name   
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();

     if (!_basicComponent.PushPropLit(header,false))
      {
       // tautology 
       return true;
      };
    };
  };

 _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };
 
 // process new components giving new names to them
 bool basicComponentChosen = false;
 ulong chosenBasicComponent;
 #ifdef _SUPPRESS_WARNINGS_ 
  chosenBasicComponent = 0;
 #endif 
  
 DOP(chosenBasicComponent = ULONG_MAX);
 UsedRule(ClauseBackground::Split);
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   if (_newCompName[compn]) 
    {
      // this could change since we could have several components that are variants of each other, 
      // simply check this condition again to avoid duplicate names 
     compVars.Reset();
     _components[compn].CollectVariables(compVars);
     _newCompName[compn] = 
      (!_splittingWithNamingFlag) 
      || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
    };

   if (_newCompName[compn])
    {
     if (basicComponentChosen) 
      {
       _componentName[compn] = TheSignature.newSkolemPredicateNumber(0UL);
       ASSERT(_componentName[compn] < HellConst::MaxNumOfFunctors);
       compVars.Reset();
       _components[compn].CollectVariables(compVars);

       if (_splittingWithNamingFlag)
        _compName->integrate(_components[compn],compVars,_componentName[compn]);

       header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
       header.MakeSplitting();
       GlobalHeaderWeighting.MakeMinimal(header);
       ALWAYS(_basicComponent.PushPropLit(header,false));
       header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),nonBasicCompPol);
       header.MakeSplitting();
       GlobalHeaderWeighting.MakeMinimal(header);
       _compNameLit.Reset();
       _compNameLit.PushPropLit(header);
       _components[compn].Push(&_compNameLit);
       Clause* cl = MakeFastUnsharedClause(_components[compn]); // the definition clause
       cl->MarkAsName();
       cl->setMainSet(Clause::MainSetNew);
       newClauses.enqueue(cl);    
       AnotherComponent();
      }
     else // this component is chosen to be the basic one
      { 
       basicComponentChosen = true;
       chosenBasicComponent = compn;
      };
    };
  };

 _basicComponent.CollectLiterals(); // this is no problem as far as nothing has been pushed into _basicComponent.AllLiterals()

 if (basicComponentChosen)
  {
   // add all the literals from _components[chosenBasicComponent] to the basic component

   while (_components[chosenBasicComponent].Nonempty()) 
    _basicComponent.AllLiterals().Push(_components[chosenBasicComponent].Pop());
  };

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };

 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
  };

 Clause* cl = MakeFastUnsharedClause(_basicComponent.AllLiterals());
 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl); 

 AnotherSplit(_divisionIntoComponents.numOfComponents());

 return true;
}; // bool NewClause::NonblockingSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)


bool NewClause::PragmaticSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)
{
 CALL("PragmaticSplitting2(TMP_LIT_LIST& lits)");
 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(_optimalGroundCompWeight);
 _divisionIntoComponents.reset();
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong groundWeight = 0; 
 ulong litNum;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   groundLit = !(lit->Variables().NumOfDifferent());
   if (groundLit) groundWeight = lit->size();
   _divisionIntoComponents.nextLit(groundLit,groundWeight);
   lit->Variables().ResetIterator();
   while ((varp = lit->Variables().NextVar())) 
    {
     _divisionIntoComponents.nextVar(*varp);     
    };        
  }; 
 
 ASSERT(_divisionIntoComponents.numOfComponents());

 if (_divisionIntoComponents.numOfComponents() == 1) 
   // splitting makes no sense 
  return false; 
  
 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();
 ulong compNum;
 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(litNum));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 _basicComponent.Reset();
 TERM header;
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;
 ulong numOfNewComps = 0;
 ulong n;

 // identify already known components
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // known component
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol); 
     header.MakeSplitting();
     if (!_basicComponent.PushPropLit(header,false))
      {
       // tautology
       return true;
      };
    };
  };
 _basicComponent.CollectLiterals();
 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };
 
 // process new components
 UsedRule(ClauseBackground::Split);
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   if (_newCompName[compn]) 
    {
      // this could change, check it again 
     compVars.Reset();
     _components[compn].CollectVariables(compVars);
     _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
    };
   if (_newCompName[compn])
    {
     _componentName[compn] = TheSignature.newSkolemPredicateNumber(0UL);
     //RegisterAsSplitting(_componentName[compn]);
     ASSERT(_componentName[compn] < HellConst::MaxNumOfFunctors);
     compVars.Reset();
     _components[compn].CollectVariables(compVars);

     if (_splittingWithNamingFlag)
      _compName->integrate(_components[compn],compVars,_componentName[compn]);

     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     GlobalHeaderWeighting.MakeMinimal(header);
     ALWAYS(_basicComponent.PushPropLit(header,false));
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),nonBasicCompPol);
     header.MakeSplitting();
     GlobalHeaderWeighting.MakeMinimal(header);
     _compNameLit.Reset();
     _compNameLit.PushPropLit(header);
     _components[compn].Push(&_compNameLit);
     Clause* cl = MakeFastUnsharedClause(_components[compn]);
     cl->MarkAsName();
     cl->setMainSet(Clause::MainSetNew);
     newClauses.enqueue(cl);    
     AnotherComponent();
    };
  };

 _basicComponent.CollectLiterals();
 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
  };
 Clause* cl = MakeFastUnsharedClause(_basicComponent.AllLiterals());
 cl->setMainSet(Clause::MainSetNew);
 newClauses.enqueue(cl); 

 AnotherSplit(_divisionIntoComponents.numOfComponents());

 return true;
}; // bool NewClause::PragmaticSplitting2(TMP_LIT_LIST& lits,ClauseQueue& newClauses)



bool NewClause::BlockingSplitting3(TMP_LIT_LIST& lits,                            
                                   Clause*& emptyClause,
                                   PagedPassiveClauseStorage& storage,
                                   ulong& numOfNewClauses,
                                   ulong& numOfRejectedNewClauses,
                                   ulong (*nextClauseNumFunc)(),
                                   bool showNewClauses)
{
 CALL("BlockingSplitting3(TMP_LIT_LIST& lits,Clause*& emptyClause,PagedPassiveClauseStorage& storage,..)");

 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
 _divisionIntoComponents.reset();
 _basicComponent.Reset();
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong litNum;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   if (lit->Header().IsSplitting())
    {
     // this literal represents a branch for previously made splitting 
     ASSERT(TheSignature.isSkolem(lit->Header().functor()));
     ASSERT(!(lit->Variables().NumOfDifferent()));
     ASSERT(!lit->Header().arity());
     if (lit->Header().polarity() == basicCompPol) return false; // the clause is blocked
     ALWAYS(_basicComponent.PushPropLit(lit->Header(),false));
    }
   else
    {
     groundLit = !(lit->Variables().NumOfDifferent());
     _litNumMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = litNum;
     lit->Variables().ResetIterator();
     while ((varp = lit->Variables().NextVar())) 
      {
       _divisionIntoComponents.nextVar(*varp);     
      };
    };        
  }; 
 
 if (_divisionIntoComponents.numOfComponents() <= 1) 
   // splitting makes no sense 
  {
   return false; 
  };  

 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();

 ulong compNum;

 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(_litNumMap[litNum]));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 TERM header;
 ulong numOfNewComps = 0;
 ulong n;
 ulong numOfNamedComponents = 0;

 // identify components that have already been given names

 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // this component has already been given name   
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     if (!_basicComponent.PushPropLit(header,false))
      {
       // tautology
       return true;
      };
     numOfNamedComponents++;
    };
  };

 if (numOfNamedComponents) 
  {
   _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

   //DF; _basicComponent.AllLiterals().output(cout << "BASICC1 ") << "\n"; 

   // collect other components in _basicComponent  
   for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
    {
     compn = _registeredCompNums.nth(n);  
     if (_newCompName[compn])
      while (_components[compn].Nonempty()) 
       _basicComponent.AllLiterals().Push(_components[compn].Pop());     
    };

   //DF; _basicComponent.AllLiterals().output(cout << "BASICC2 ") << "\n";

   // make a clause of _basicComponent.AllLiterals() 
   if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
    { 
     AnotherForwardSubsumed();
     return true; 
    };

   //DF; _basicComponent.AllLiterals().output(cout << "BASICC3 ") << "\n";

   if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
    {
     MakeVIP();
     AnotherSimplifiedByFSR();
     if (_basicComponent.AllLiterals().Empty()) // happens!
      {
       emptyClause = MakeEmptyClause();
       emptyClause->SetNumber(nextClauseNumFunc());
       emptyClause->setMainSet(Clause::MainSetNew);
       numOfNewClauses++;
       if (showNewClauses) cout << "% " << emptyClause << "\n";
       return true;  
      };
    };

   UsedRule(ClauseBackground::Split);
  
   //COP("COP100"); 
   if (makeClauseInPagedPassiveClauseStorage(storage,_basicComponent.AllLiterals(),nextClauseNumFunc(),showNewClauses,false))
    {
     numOfNewClauses++;                  
    }
   else
    {
     numOfRejectedNewClauses++;
    };
   
   AnotherSplit(numOfNamedComponents); 

   return true;   
  };

 // !numOfNamedComponents    here

 if ((TheSignature.nextFunctorNumber() + 1) > HellConst::MaxNumOfFunctors) // can not create a new name
  return false;

 
 UsedRule(ClauseBackground::Split);

 // now, using heuristics, select a component to become the open one 
 
 ulong openComponentNum = _registeredCompNums.nth(0); // temporary simple solution, most certainly not the best one
 
 // give a name to the open component

 _componentName[openComponentNum] = TheSignature.newSkolemPredicateNumber(0UL);
 ASSERT(_componentName[openComponentNum] < HellConst::MaxNumOfFunctors);
 compVars.Reset();
 _components[openComponentNum].CollectVariables(compVars);
 
 if (_splittingWithNamingFlag)
  _compName->integrate(_components[openComponentNum],compVars,_componentName[openComponentNum]);

 header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[openComponentNum]),basicCompPol);
 header.MakeSplitting();
 GlobalHeaderWeighting.MakeMinimal(header);
 ALWAYS(_basicComponent.PushPropLit(header,false));
 header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[openComponentNum]),nonBasicCompPol);
 header.MakeSplitting();
 GlobalHeaderWeighting.MakeMinimal(header);
 _compNameLit.Reset();
 _compNameLit.PushPropLit(header);
 _components[openComponentNum].Push(&_compNameLit);

 //COP("COP200"); 
 if (makeClauseInPagedPassiveClauseStorage(storage,_components[openComponentNum],nextClauseNumFunc(),showNewClauses,true))
  {
   numOfNewClauses++;                  
  }
 else
  {
   numOfRejectedNewClauses++;
  };


 AnotherComponent(); 
 

 _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

 // collect other (blocked)  components in _basicComponent  
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);  
   ASSERT(_newCompName[compn]);
   if (compn != openComponentNum)
    while (_components[compn].Nonempty()) 
     _basicComponent.AllLiterals().Push(_components[compn].Pop());     
  };

 // make a clause of _basicComponent.AllLiterals() 

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };

 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
   ASSERT(!_basicComponent.AllLiterals().Empty());
  };

 //COP("COP300"); 
 if (makeClauseInPagedPassiveClauseStorage(storage,_basicComponent.AllLiterals(),nextClauseNumFunc(),showNewClauses,false))
  {
   numOfNewClauses++;                  
  }
 else
  {
   numOfRejectedNewClauses++;
  };

 AnotherSplit(2);
 return true;   

}; // bool NewClause::BlockingSplitting3(TMP_LIT_LIST& lits,Clause*& emptyClause,PagedPassiveClauseStorage& storage,..)


bool NewClause::NonblockingSplitting3(TMP_LIT_LIST& lits,
                                      Clause*& emptyClause,
                                      PagedPassiveClauseStorage& storage,
                                      ulong& numOfNewClauses,
                                      ulong& numOfRejectedNewClauses,
                                      ulong (*nextClauseNumFunc)(),
                                      bool showNewClauses)
{
 CALL("NonblockingSplitting3(TMP_LIT_LIST& lits,Clause*& emptyClause,PagedPassiveClauseStorage& storage,..)");
 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
 _divisionIntoComponents.reset();
 _basicComponent.Reset();
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong litNum;
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   if (lit->Header().IsSplitting())
    {
     // this literal represents a previously made splitting branch
     ASSERT(TheSignature.isSkolem(lit->Header().functor()));
     ASSERT(!(lit->Variables().NumOfDifferent()));
     ASSERT(!lit->Header().arity());

     ALWAYS(_basicComponent.PushPropLit(lit->Header(),false));
 
    }
   else
    {
     groundLit = !(lit->Variables().NumOfDifferent());
     _litNumMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = litNum;
     lit->Variables().ResetIterator();
     while ((varp = lit->Variables().NextVar())) 
      {
       _divisionIntoComponents.nextVar(*varp);     
      };
    };        
  }; 
 
 if (_divisionIntoComponents.numOfComponents() <= 1) 
   // splitting makes no sense 
  return false; 

  
 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();

 ulong compNum;

 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(_litNumMap[litNum]));
  };

 //_basicComponent.CollectLiterals(); // debug
 //_basicComponent.AllLiterals().output(cout << "*******BC0 ") << "\n"; // debug

 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 TERM header;
 ulong numOfNewComps = 0;
 ulong n;

 // identify components that have already been given names
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {

      //_components[compn].output(cout << "NEW COMP " << compn << " ") << "\n"; 
 
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // this component has already been given name   
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     if (!_basicComponent.PushPropLit(header,false))
      {
       // tautology
       return true;
      };
     //DF; cout << "PUSH " << header << "\n";

    };
  };

 _basicComponent.CollectLiterals(); // to make _basicComponent.AllLiterals() available

 //_basicComponent.AllLiterals().output(cout << "*******BC1 ") << "\n"; // debug

 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };
 
 // process new components giving new names to them
 bool basicComponentChosen = false;
 ulong chosenBasicComponent;
 #ifdef _SUPPRESS_WARNINGS_ 
  chosenBasicComponent = 0;
 #endif 
  
 DOP(chosenBasicComponent = ULONG_MAX);
 UsedRule(ClauseBackground::Split);
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);

   //cout << "COMPNUM " << compn << "\n"; // debug


   if (_newCompName[compn]) 
    {

      //_components[compn].output(cout << "NEW COMP " << compn << " ") << "\n"; // debug
      //cout << "NAME = " <<  _componentName[compn] << "\n"; // debug


      // this could change since we could have several components that are variants of each other, 
      // simply check this condition again to avoid duplicate names 
     compVars.Reset();
     _components[compn].CollectVariables(compVars);     
     _newCompName[compn] = 
      (!_splittingWithNamingFlag) 
      || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
     
     //cout << "NOW new = " <<  _newCompName[compn] << "\n"; // debug
     //cout << "NAME = " <<  _componentName[compn] << "\n"; // debug

    };

   if (_newCompName[compn])
    {
     if (basicComponentChosen) 
      {
       _componentName[compn] = TheSignature.newSkolemPredicateNumber(0UL);
       ASSERT(_componentName[compn] < HellConst::MaxNumOfFunctors);
       compVars.Reset();
       _components[compn].CollectVariables(compVars);

       if (_splittingWithNamingFlag)
        _compName->integrate(_components[compn],compVars,_componentName[compn]);

       header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
       header.MakeSplitting();
       GlobalHeaderWeighting.MakeMinimal(header);
       ALWAYS(_basicComponent.PushPropLit(header,false));
       header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),nonBasicCompPol);
       header.MakeSplitting();
       GlobalHeaderWeighting.MakeMinimal(header);
       _compNameLit.Reset();
       _compNameLit.PushPropLit(header);
       _components[compn].Push(&_compNameLit);
   
   
       if (makeClauseInPagedPassiveClauseStorage(storage,_components[compn],nextClauseNumFunc(),showNewClauses,true))
        {
         numOfNewClauses++;                  
        }
       else
        {
         numOfRejectedNewClauses++;
        };

       AnotherComponent();
      }
     else // this component is chosen to be the basic one
      { 
       basicComponentChosen = true;
       chosenBasicComponent = compn;
      };
    };
  };

 _basicComponent.CollectLiterals(); // this is no problem as far as nothing has been pushed into _basicComponent.AllLiterals()


 //_basicComponent.AllLiterals().output(cout << "*******BC2 ") << "\n"; // debug


 if (basicComponentChosen)
  {
   // add all the literals from _components[chosenBasicComponent] to the basic component

   while (_components[chosenBasicComponent].Nonempty()) 
    {
     TMP_LITERAL* bclit = _components[chosenBasicComponent].Pop();

     //cout << "PUSH " << bclit << "\n";

     _basicComponent.AllLiterals().Push(bclit);
    };
  }
 else
  {

    //cout << "!basicComponentChosen\n"; // debug

  };

 //_basicComponent.AllLiterals().output(cout << "*******BC3 ") << "\n"; // debug


 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };

 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
   if (_basicComponent.AllLiterals().Empty()) // happens!
    {
     emptyClause = MakeEmptyClause();
     emptyClause->SetNumber(nextClauseNumFunc());
     emptyClause->setMainSet(Clause::MainSetNew);
     numOfNewClauses++;
     if (showNewClauses) cout << "% " << emptyClause << "\n";
     return true;  
    };
  };

 //_basicComponent.AllLiterals().output(cout << "*******BC4 ") << "\n"; // debug

 if (makeClauseInPagedPassiveClauseStorage(storage,_basicComponent.AllLiterals(),nextClauseNumFunc(),showNewClauses,false))
  {
   numOfNewClauses++;                  
  }
 else
  {
   numOfRejectedNewClauses++;
  };

 AnotherSplit(_divisionIntoComponents.numOfComponents());

 return true;
}; // bool NewClause::NonblockingSplitting3(TMP_LIT_LIST& lits,Clause*& emptyClause,PagedPassiveClauseStorage& storage,..)








bool NewClause::PragmaticSplitting3(TMP_LIT_LIST& lits,
                                    Clause*& emptyClause,
                                    PagedPassiveClauseStorage& storage,
                                    ulong& numOfNewClauses,
                                    ulong& numOfRejectedNewClauses,
                                    ulong (*nextClauseNumFunc)(),
                                    bool showNewClauses)
{
 CALL("PragmaticSplitting3(TMP_LIT_LIST& lits,Clause*& emptyClause,PagedPassiveClauseStorage& storage,..)");
 ASSERT(lits.Length()); 
 _divisionIntoComponents.setGroundCompOptimalWeight(_optimalGroundCompWeight);
 _divisionIntoComponents.reset();
 TMP_LITERAL* lit;
 ulong const * varp;      
 bool groundLit;
 ulong groundWeight = 0; 
 ulong litNum;

 for (litNum = 0; litNum < lits.Length(); litNum++)
  {
   lit = lits.Nth(litNum);
   groundLit = !(lit->Variables().NumOfDifferent());
   if (groundLit) groundWeight = lit->size();
   _divisionIntoComponents.nextLit(groundLit,groundWeight);
   lit->Variables().ResetIterator();
   while ((varp = lit->Variables().NextVar())) 
    {
     _divisionIntoComponents.nextVar(*varp);     
    };        
  }; 
 
 ASSERT(_divisionIntoComponents.numOfComponents());

 if (_divisionIntoComponents.numOfComponents() == 1) 
   // splitting makes no sense 
  return false; 
  
 // splitting makes sense
 _registeredCompNums.reset();    

 _divisionIntoComponents.startResultCollection();
 ulong compNum;
 while (_divisionIntoComponents.litCompPair(litNum,compNum))
  {
   if (!(_registeredCompNums.scoreOf(compNum)))
    {
     _registeredCompNums.add(compNum);
     _components[compNum].Reset();    
    };
   _components[compNum].Push(lits.Nth(litNum));
  };


 MultisetOfVariables<HellConst::MaxNumOfVariables> compVars;
 ulong compn;
 _basicComponent.Reset();
 TERM header;
 TERM::Polarity basicCompPol = TERM::NegativePolarity;
 TERM::Polarity nonBasicCompPol = TERM::PositivePolarity;
 ulong numOfNewComps = 0;
 ulong n;

 // identify already known components
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   compVars.Reset();
   _components[compn].CollectVariables(compVars);
   _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
   if (_newCompName[compn])
    {
     numOfNewComps++;
     if ((TheSignature.nextFunctorNumber() + numOfNewComps) > HellConst::MaxNumOfFunctors) 
      return false;
    }
   else // known component
    {
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol); 
     header.MakeSplitting();
     if (!_basicComponent.PushPropLit(header,false))
      {
       // tautology
       return true;
      };
    };
  };
 _basicComponent.CollectLiterals();
 if (_forwardSubsumptionFlag && (FSubsume(_basicComponent.AllLiterals())))
  { 
   AnotherForwardSubsumed();
   return true; 
  };
 
 // process new components
 UsedRule(ClauseBackground::Split);
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
  {
   compn = _registeredCompNums.nth(n);
   if (_newCompName[compn]) 
    {
      // this could change, check it again 
     compVars.Reset();
     _components[compn].CollectVariables(compVars);
     _newCompName[compn] = (!_splittingWithNamingFlag) || (!_compName->findName(_components[compn],compVars,_componentName[compn]));
    };
   if (_newCompName[compn])
    {
     _componentName[compn] = TheSignature.newSkolemPredicateNumber(0UL);
     //RegisterAsSplitting(_componentName[compn]);
     ASSERT(_componentName[compn] < HellConst::MaxNumOfFunctors);
     compVars.Reset();
     _components[compn].CollectVariables(compVars);

     if (_splittingWithNamingFlag)
      _compName->integrate(_components[compn],compVars,_componentName[compn]);

     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),basicCompPol);
     header.MakeSplitting();
     GlobalHeaderWeighting.MakeMinimal(header);
     ALWAYS(_basicComponent.PushPropLit(header,false));
     header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[compn]),nonBasicCompPol);
     header.MakeSplitting();
     GlobalHeaderWeighting.MakeMinimal(header);
     _compNameLit.Reset();
     _compNameLit.PushPropLit(header);
     _components[compn].Push(&_compNameLit);
  
     if (makeClauseInPagedPassiveClauseStorage(storage,_components[compn],nextClauseNumFunc(),showNewClauses,true))
      {
       numOfNewClauses++;                  
      }
     else
      {
       numOfRejectedNewClauses++;
      };

     AnotherComponent();
    };
  };

 _basicComponent.CollectLiterals();
 if ((_forwardSubsumptionResolution) && (SimplifyByFS(_basicComponent.AllLiterals())))
  {
   MakeVIP();
   AnotherSimplifiedByFSR();
   if (_basicComponent.AllLiterals().Empty()) // happens!
    {
     emptyClause = MakeEmptyClause();
     emptyClause->SetNumber(nextClauseNumFunc());
     emptyClause->setMainSet(Clause::MainSetNew);
     numOfNewClauses++;
     if (showNewClauses) cout << "% " << emptyClause << "\n";
     return true;  
    };
  };


 if (makeClauseInPagedPassiveClauseStorage(storage,_basicComponent.AllLiterals(),nextClauseNumFunc(),showNewClauses,false))
  {
   numOfNewClauses++;                  
  }
 else
  {
   numOfRejectedNewClauses++;
  };


 AnotherSplit(_divisionIntoComponents.numOfComponents());

 return true;
}; // bool NewClause::PragmaticSplitting3(TMP_LIT_LIST& lits,Clause*& emptyClause,PagedPassiveClauseStorage& storage,..)


//=================================================
