//
// File:         TPTPParser.cpp
// Description:  
// Created:      Mar 6, 2000, 14:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "TPTPParser.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TPTP_PARSER
 #define DEBUG_NAMESPACE "TPTPParser"
#endif
#include "debugMacros.h"
//====================================================
#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"

 ClassDesc TPTPParser::HeaderList::_classDesc("TPTPParser::HeaderList",
                                              TPTPParser::HeaderList::minSize(),
                                              TPTPParser::HeaderList::maxSize());
 ClassDesc TPTPParser::SymbolPrecedenceList::_classDesc("TPTPParser::SymbolPrecedenceList",
                                                        TPTPParser::SymbolPrecedenceList::minSize(),
						        TPTPParser::SymbolPrecedenceList::maxSize());
#endif
ulong TPTPParser::_nextClauseNumber = 1;
Clause::OpenUnshared TPTPParser::_tmpClause;
ClauseQueue TPTPParser::_inputClauses;
TPTPParser::HeaderList* TPTPParser::_definedHeaders = (TPTPParser::HeaderList*)0; 
TPTPParser::SymbolPrecedenceList* TPTPParser::_symbolPrecedenceList = (TPTPParser::SymbolPrecedenceList*)0;

bool TPTPParser::_someInputClausesRejected = false;

bool TPTPParser::readFile(const char* TPTPDir, const char* problemName)
{
 CALL("readFile(const char* TPTPDir, const char* problemName)");
 PR_FILE_PARS_RESULT* file_pars_res = new PR_FILE_PARS_RESULT();  
 ParseFile(problemName,file_pars_res);
 if (file_pars_res->Error())
  {
   cout << "Error in " << problemName << " ";
   cout << "[" << file_pars_res->ErrorLineNum() << "/" << file_pars_res->ErrorCharNum() << "] : ";
   cout << file_pars_res->ErrorMessage() << "\n";
   return false;
  };
  
 _someInputClausesRejected = false;
 for (PR_LITERAL_LIST* lst = file_pars_res->Literals(); lst; lst = lst->Next())
  if (!(processInputItem(lst->Literal(),TPTPDir,problemName))) return false;
  
 delete file_pars_res;

 return true;   
}; // bool TPTPParser::readFile(const char* TPTPDir, const char* problemName)

/*******************
HEADER_WEIGHT* TPTPParser::createHW(PR_LITERAL* item)
{
 CALL("TPTPParser::createHW(PR_LITERAL* item)");
 if (PR_TERM_LIST::Length(item->Arguments()) == 2)
  {
   PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),1);
   PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),0);
   TERM header;
   if (createHeader(header,arg1)) 
    {
     if (arg2->Tag() == PR_TERM::ATOM_NUMBER)
      {
       long num;
       #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
        StrToX strToLong(arg2->Symbol()->Symbol());
       #else
        StrToX strToLong((char*)(arg2->Symbol()->Symbol()));
       #endif
       strToLong.read(num); 
       return new HEADER_WEIGHT(header,num);
      }
     else
      {
       if (arg2->Tag() == PR_TERM::ATOM_FUNCTOR)
        {
         if (!(strcmp(arg2->Symbol()->Symbol(),"min"))) 
          return new HEADER_WEIGHT(header,HEADER_WEIGHT::MIN);          
         if (!(strcmp(arg2->Symbol()->Symbol(),"max"))) 
          return new HEADER_WEIGHT(header,HEADER_WEIGHT::MAX);
        };
      };
    };
  };
 return (HEADER_WEIGHT*)0;
}; // HEADER_WEIGHT* TPTPParser::createHW(PR_LITERAL* item)
**********************/

bool TPTPParser::readSymbolPrecedence(PR_LITERAL* item,ulong& symbolNum,long& precedence)
{
 CALL("readSymbolPrecedence(PR_LITERAL* item,ulong& symbolNum,long& precedence)");
 if (PR_TERM_LIST::Length(item->Arguments()) == 2)
  {
   PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),1);
   PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),0);
    
   if (arg1->Tag() != PR_TERM::ATOM_FUNCTOR) return false;      
   symbolNum = functorNum(arg1->Symbol()->Symbol());

   if (arg2->Tag() == PR_TERM::ATOM_NUMBER)
    {
     #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
      StrToX strToLong(arg2->Symbol()->Symbol());
     #else
      StrToX strToLong((char*)(arg2->Symbol()->Symbol()));
     #endif
     strToLong.read(precedence); 
     return true;
    };
  };
 return false; 
}; // bool TPTPParser::readSymbolPrecedence(PR_LITERAL* item,ulong& symbolNum,long& precedence)


bool TPTPParser::addDefinedHeaders(PR_LITERAL* item)
{
 CALL("addDefinedHeaders(PR_LITERAL* item)");
 if (PR_TERM_LIST::Length(item->Arguments()) != 1) return false;
 if (PR_TERM_LIST::Nth(item->Arguments(),0)->Tag() != PR_TERM::LIST) return false; 
 PR_TERM_LIST* hd_list = PR_TERM_LIST::Nth(item->Arguments(),0)->Arguments();
 TERM header;
 while (hd_list)
  {
   if (!createHeader(header,hd_list->Term())) return false;
   if (!(HeaderList::member(header,_definedHeaders)))
    _definedHeaders = new HeaderList(header,_definedHeaders);
   hd_list = hd_list->Next();
  };
 return true;
}; // bool TPTPParser::addDefinedHeaders(PR_LITERAL* item)



bool TPTPParser::processInputItem(PR_LITERAL* item,const char* TPTPDir,const char* problemName)
{
 CALL("processInputItem(PR_LITERAL* item,const char* TPTPDir,const char* problemName)");
 bool reallyCreated;
 switch (TPTPClassify(item))
  {
   case TPTP_INCLUDE_DIRECTIVE :
    if (PR_TERM_LIST::Length(item->Arguments()) == 1)
     {
      if (item->Arguments()->Term()->Tag() == PR_TERM::ATOM_STRING)
       {
        const char* incl_file_name = item->Arguments()->Term()->Symbol()->Symbol();

        char* file_name = new char [strlen(TPTPDir) + strlen(incl_file_name) + 2];
        
		strcpy(file_name,TPTPDir);
        file_name[strlen(TPTPDir)] = '/';
        strcpy(file_name + strlen(TPTPDir) + 1,incl_file_name);
        file_name[strlen(TPTPDir)  + strlen(incl_file_name) + 1] = (char)0;
        return readFile(TPTPDir,file_name);
       }
      else
       {
        cout << "Error in " << problemName << " ";
        cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
        cout << "Wrong argument. " << "\n";
        return false;
       };
     }
    else 
     {
      cout << "Error in " << problemName << " ";
      cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
      cout << "Wrong arity. " << "\n";
      return false; 
     };

   case TPTP_INPUT_CLAUSE : 
           
    if (createClause(item,reallyCreated))
     {
      if (!reallyCreated)
       {
	_someInputClausesRejected = true;
	cout << "WARNING: input clause rejected.\n";
       };       
      return true;
     }
    else
     {
             
      cout << "Error in " << problemName << "\n";
      cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
      cout << "Can not read an input clause. " << "\n";
      return false;
     };
     
   case TPTP_INPUT_FORMULA :
    cout << "************* FOF syntax is not imlemented yet! **************\n";
    return false;

   case EXT_TPTP_SET_OPTION_DIRECTIVE : 
    if (PR_TERM_LIST::Length(item->Arguments()) == 2)
     {
      if (item->Arguments()->Next()->Term()->Tag() == PR_TERM::ATOM_FUNCTOR)
       {
	 /****
        char* option_name = (char*)(item->Arguments()->Next()->Term()->Symbol()->Symbol());
        char* option_value = (char*)(item->Arguments()->Term()->Symbol()->Symbol());
        char* expl = (char*)0;
                
        if (!OPTION::Set(option_name,option_value,expl))
         {
          cout << "Warning: can not set option \"" << option_name << "\",";
          cout << expl << ".\n";
          cout << "in " << problemName;
          cout << " [" << item->Arguments()->Next()->Term()->Symbol()->LineNum();
          cout << "/" << item->Arguments()->Next()->Term()->Symbol()->CharNum() << "]\n";
         };
	****/
        return true;
       }
      else
       {
        cout << "Error in " << problemName << " ";
        cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
        cout << "Wrong option name. " << "\n";
        return false;
       };
     }
    else 
     {
      cout << "Error in " << problemName << " ";
      cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
      cout << "Wrong arity. " << "\n";
      return false; 
     };

   case EXT_TPTP_HEADER_WEIGHT : 
    {
     ulong symbolNum;
     long precedence;
     if (readSymbolPrecedence(item,symbolNum,precedence))
      {
       _symbolPrecedenceList = new SymbolPrecedenceList(SymbolPrecedencePair(symbolNum,precedence),_symbolPrecedenceList);
       return true;
      }
     else
      {
       cout << "Error in " << problemName << " ";
       cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
       cout << "Can not read symbol precedence. " << "\n";
       return false;
      };
    };

   case EXT_TPTP_DEFINED_HEADER_LIST : 

    if (!addDefinedHeaders(item))  
     {
      cout << "Error in " << problemName << " ";
      cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
      cout << "Can not read a list of defined headers. " << "\n";
      return false;
     };   
    return true;

   case TPTP_UNCKNOWN_ITEM_SORT : 
    cout << "Error in " << problemName << " ";
    cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
    cout << "Uncknown input item type. " << "\n";
    return false;
     
  };
 return false; // to surpress warnings;
}; // bool TPTPParser::processInputItem(PR_LITERAL* item,const char* TPTPDir,const char* problemName)

bool TPTPParser::createHeader(TERM& header,PR_TERM* t)
{
 CALL("createHeader(TERM& header,PR_TERM* t)");
 if (t->Tag() != PR_TERM::COMPLEX) { return false; };
 if (PR_TERM_LIST::Length(t->Arguments()) != 1) { return false; };
 if (strcmp(t->Symbol()->Symbol(),"++"))
  {
   if (strcmp(t->Symbol()->Symbol(),"--")) { return false; }
   else header = TERM((TERM::Arity)0,(TERM::Functor)0,TERM::NegativePolarity);
  }
 else header = TERM((TERM::Arity)0,(TERM::Functor)0,TERM::PositivePolarity);
 PR_TERM* fun = t->Arguments()->Term();
 if (fun->Tag() != PR_TERM::ATOM_FUNCTOR) { return false; };
 ulong funNum = functorNum(fun->Symbol()->Symbol());
 if (funNum > HellConst::MaxNumOfFunctors - 4) return false;
 header.SetFunctor(funNum);  
 return true; 
}; //bool TPTPParser::createHeader(TERM& header,PR_TERM* t) 

TPTPParser::TPTP_ITEM_SORT TPTPParser::TPTPClassify(const char* pref)
{
 if (!(strcmp(pref,"include"))) { return TPTP_INCLUDE_DIRECTIVE; };
 if (!(strcmp(pref,"input_clause"))) { return TPTP_INPUT_CLAUSE; };
 if (!(strcmp(pref,"input_formula"))) { return TPTP_INPUT_FORMULA; };
 if (!(strcmp(pref,"set"))) { return EXT_TPTP_SET_OPTION_DIRECTIVE; };
 if (!(strcmp(pref,"symbol_precedence"))) { return EXT_TPTP_HEADER_WEIGHT; };
 if (!(strcmp(pref,"defined_headers"))) { return EXT_TPTP_DEFINED_HEADER_LIST; };
 return TPTP_UNCKNOWN_ITEM_SORT;
}; // TPTP_ITEM_SORT TPTPParser::TPTPClassify(const char* pref)

TPTPParser::TPTP_ITEM_SORT TPTPParser::TPTPClassify(const PR_LITERAL* lit)
{
 return TPTPClassify(lit->Header()->Symbol());
}; // TPTP_ITEM_SORT TPTPParser::TPTPClassify(const PR_LITERAL* lit)

TPTPParser::TPTP_ITEM_TYPE TPTPParser::TPTPItemType(PR_TERM* it)
{
 if (it->Tag() != PR_TERM::ATOM_FUNCTOR) { return TPTP_WRONG_ITEM_TYPE; };
 const char* str = it->Symbol()->Symbol(); 
 if (!(strcmp(str,"axiom"))) { return TPTP_AXIOM; };
 if (!(strcmp(str,"conjecture"))) { return TPTP_CONJECTURE; };
 if (!(strcmp(str,"hypothesis"))) { return TPTP_HYPOTHESIS; };
 return TPTP_WRONG_ITEM_TYPE;
}; // TPTP_ITEM_TYPE TPTPParser::TPTPItemType(PR_TERM* it)

bool TPTPParser::createClause(PR_LITERAL* input_cl,bool& reallyCreated)
{
 CALL("createClause(PR_LITERAL* input_cl,bool& reallyCreated)");
 if (PR_TERM_LIST::Length(input_cl->Arguments()) != 3) return false; 
 if (PR_TERM_LIST::Nth(input_cl->Arguments(),0)->Tag() != PR_TERM::LIST) return false;
 TPTP_ITEM_TYPE cl_type = TPTPItemType(PR_TERM_LIST::Nth(input_cl->Arguments(),1));
 if (cl_type == TPTP_WRONG_ITEM_TYPE) return false; 
 return createClause(PR_TERM_LIST::Nth(input_cl->Arguments(),0)->Arguments(),cl_type,reallyCreated);
}; // bool TPTPParser::createClause(PR_LITERAL* input_cl,bool& reallyCreated)

bool TPTPParser::createClause(PR_TERM_LIST* lit_list,TPTP_ITEM_TYPE cl_type,bool& reallyCreated)
{
 CALL("createClause(PR_TERM_LIST* lit_list,TPTP_ITEM_TYPE cl_type,bool& reallyCreated)");
 _tmpClause.reset();
 _tmpClause.usedRule(ClauseBackground::Input);
 TheSignature.resetInputVariables();
 PR_TERM_LIST* tmp_lit_list = lit_list;
 while (tmp_lit_list)
  {
   if (!checkLiteralSyntax(tmp_lit_list->Term())) return false; 
   tmp_lit_list = tmp_lit_list->Next();   
  };
 // syntax is OK
 while (lit_list)
  { 
   if (!collectLiteral(lit_list->Term())) { reallyCreated = false; return true; }; 
   lit_list = lit_list->Next();  
  };

 Clause* newClause = _tmpClause.result();
 newClause->SetNumber(_nextClauseNumber);
 _nextClauseNumber++; 
 _inputClauses.enqueue(newClause);
 reallyCreated = true;
 return true; 
}; // bool TPTPParser::createClause(PR_TERM_LIST* lit_list,TPTP_ITEM_TYPE cl_type,bool& reallyCreated)

bool TPTPParser::checkLiteralSyntax(PR_TERM* lit)
{
 if ((lit->Tag() != PR_TERM::COMPLEX) || (PR_TERM_LIST::Length(lit->Arguments()) != 1))
  return false;
 if ((strcmp(lit->Symbol()->Symbol(),"++")) && (strcmp(lit->Symbol()->Symbol(),"--")))
  return false;
 return checkComplexTermSyntax(lit->Arguments()->Term());
}; // bool TPTPParser::checkLiteralSyntax(PR_TERM* lit)

bool TPTPParser::checkComplexTermSyntax(PR_TERM* t)
{
 PR_TERM_LIST* args;
 //ulong arity;
 switch (t->Tag())
  {
   case PR_TERM::ATOM_FUNCTOR: return true; // constant
   case PR_TERM::ATOM_NUMBER: return true; // constant
   case PR_TERM::COMPLEX:
    args = t->Arguments();
    //arity = PR_TERM_LIST::Length(args);
    //if (arity > HellConst::MaxNumOfTermArguments) return false;   
    while (args)
     {
      if ((args->Term()->Tag() != PR_TERM::VAR) 
          && (!checkComplexTermSyntax(args->Term()))) return false;
      args = args->Next();
     };
    return true;
   default: return false;
  };  
 return false;
}; // bool TPTPParser::checkComplexTermSyntax(PR_TERM* t)

bool TPTPParser::collectLiteral(PR_TERM* lit)
{
 CALL("collectLiteral(PR_TERM* lit)");
 TERM::Polarity polarity = strcmp(lit->Symbol()->Symbol(),"++") ? TERM::NegativePolarity : TERM::PositivePolarity;
 PR_TERM* body = lit->Arguments()->Term();
 PR_TERM_LIST* args = body->Arguments();
 ulong arity = PR_TERM_LIST::Length(args);
 if (arity > HellConst::MaxNumOfTermArguments) { _tmpClause.recycle(); return false; };
 ulong func = functorNum(body->Symbol()->Symbol(),arity);

 if (func > HellConst::MaxNumOfFunctors - 4) return false;

 MakePredicate(func); 
 //if (arity > HellConst::MaxNumOfTermArguments) { _tmpClause.recycle(); return false; };
 TERM header((TERM::Arity)arity,(TERM::Functor)func,polarity);

 if (!_tmpClause.litHeader(header)) return false;

 while (args)
  { 
   if (!collectTerm(args->Term())) return false; 
   args = args->Next(); 
  };
 return true;
}; // bool TPTPParser::collectLiteral(PR_TERM* lit) 

bool TPTPParser::collectTerm(PR_TERM* t)
{
 CALL("collectTerm(PR_TERM* t)");
 ulong func;
 ulong var;
 ulong arity;
 PR_TERM_LIST* args;
 switch (t->Tag())
  { 
   case PR_TERM::ATOM_FUNCTOR: // constant, no break here
   case PR_TERM::ATOM_NUMBER: // constant  
    func = functorNum(t->Symbol()->Symbol(),0UL);
    if (func > HellConst::MaxNumOfFunctors - 4) return false;
    MakeNonpredicate(func); 
    return _tmpClause.functor(TERM((TERM::Arity)0,(TERM::Functor)func,TERM::PositivePolarity));
   case PR_TERM::VAR:
    var = variableNum(t->Symbol()->Symbol()); 
    _tmpClause.variable(TERM(TermVar,var));
    return true;
   case PR_TERM::COMPLEX:
    args = t->Arguments();
    arity = PR_TERM_LIST::Length(args);   
    func = functorNum(t->Symbol()->Symbol(),arity);
    if (func > HellConst::MaxNumOfFunctors - 4) return false;
    if (arity > HellConst::MaxNumOfTermArguments) { _tmpClause.recycle(); return false; };
    MakeNonpredicate(func);
    //if (arity > HellConst::MaxNumOfTermArguments) { _tmpClause.recycle(); return false; };
    if (!_tmpClause.functor(TERM((TERM::Arity)arity,(TERM::Functor)func,TERM::PositivePolarity))) return false;
    // collect arguments
    while (args)
     { 
      if (!collectTerm(args->Term())) return false; 
      args = args->Next(); 
     };    
    return true;
   default: return false; // must be error
  };
 return false; 
}; // bool TPTPParser::collectTerm(PR_TERM* t) 


//====================================================

