//
// File:         TPTPParser.h
// Description:  
// Created:      Mar 6, 2000, 13:50
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef TPTP_PARSER_H
//====================================================
#define TPTP_PARSER_H
#include <string.h>
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "GlobAlloc.h"
#include "StrToX.h"
#include "pr_abstract_syntax.h"
#include "pr_parser.h"
#include "weighting.h"
#include "Signature.h"
#include "clause.h"
#include "term.h"
#include "ClauseBackground.h"
#include "FIFO.h"
#include "GList.h" 
#include "ClauseQueue.h"
#include "Tuple.h"
extern HEADER_WEIGHT_LIST* HWList;
extern Signature TheSignature;
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TPTP_PARSER
 #define DEBUG_NAMESPACE "TPTPParser"
#endif
#include "debugMacros.h"
//====================================================
class TPTPParser
{
 public: 
  typedef GList<GlobAlloc,TERM,TPTPParser> HeaderList;
  typedef Tuple2<ulong,long> SymbolPrecedencePair;
  typedef GList<GlobAlloc,SymbolPrecedencePair,TPTPParser> SymbolPrecedenceList;
 public:
  static bool readFile(const char* TPTPDir, const char* problemName);
  static const ClauseQueue& inputClauses() 
  { 
   return _inputClauses; 
  };
  static ulong nextClauseNumber() { return _nextClauseNumber; };
  static const HeaderList* definedHeaders() { return _definedHeaders; };
  static const SymbolPrecedenceList* symbolPrecedenceList() { return _symbolPrecedenceList; };

  static bool someInputClausesRejected() { return _someInputClausesRejected; };
 //private:
 public: // made public to please VC++
  enum TPTP_ITEM_SORT
  {
   TPTP_INCLUDE_DIRECTIVE,
   TPTP_INPUT_CLAUSE,
   TPTP_INPUT_FORMULA,
   EXT_TPTP_SET_OPTION_DIRECTIVE,
   EXT_TPTP_HEADER_WEIGHT,
   EXT_TPTP_DEFINED_HEADER_LIST,
   TPTP_UNCKNOWN_ITEM_SORT
  }; // enum TPTP_ITEM_SORT 

  enum TPTP_ITEM_TYPE
  {
   TPTP_AXIOM,
   TPTP_CONJECTURE,
   TPTP_HYPOTHESIS,
   TPTP_WRONG_ITEM_TYPE
  }; // enum TPTP_ITEM_TYPE

 private:
  //static HEADER_WEIGHT* createHW(PR_LITERAL* item);
  static bool readSymbolPrecedence(PR_LITERAL* item,ulong& symbolNum,long& precedence);
  static bool addDefinedHeaders(PR_LITERAL* item);
  static bool processInputItem(PR_LITERAL* item,const char* TPTPDir,const char* problemName);
  static bool createHeader(TERM& header,PR_TERM* t);
  static ulong functorNum(const char* f,const ulong& arity) 
  { 
   return TheSignature.inputFunctorNumber(f,arity); 
  };
  static ulong functorNum(const char* f) { return TheSignature.inputFunctorNumber(f); };

  static void MakePredicate(const ulong& fnum) { TheSignature.makePredicate(fnum); };
  static void MakeNonpredicate(const ulong& fnum) { TheSignature.makeNonpredicate(fnum); };  
  static ulong variableNum(const char* v) { return TheSignature.inputVariableNumber(v); };
  static TPTP_ITEM_SORT TPTPClassify(const char* pref);
  static TPTP_ITEM_SORT TPTPClassify(const PR_LITERAL* lit);
  static TPTP_ITEM_TYPE TPTPItemType(PR_TERM* it);
  static bool createClause(PR_LITERAL* input_cl,bool& reallyCreated);
  static bool createClause(PR_TERM_LIST* lit_list,TPTP_ITEM_TYPE cl_type,bool& reallyCreated);
  static bool checkLiteralSyntax(PR_TERM* lit);
  static bool checkComplexTermSyntax(PR_TERM* lit);
  static bool collectLiteral(PR_TERM* lit);
  static bool collectTerm(PR_TERM* t);
 private:
  static ulong _nextClauseNumber;
  static Clause::OpenUnshared _tmpClause;
  static ClauseQueue _inputClauses;
  static HeaderList* _definedHeaders;
  static SymbolPrecedenceList* _symbolPrecedenceList;
  static bool _someInputClausesRejected;
}; // class TPTPParser
//====================================================
#undef DEBUG_NAMESPACE
//====================================================
#endif
