//
// File:         Tabulation.h
// Description:  
// Created:      Mar 21, 2000, 16:35
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//=======================================================
#ifndef TABULATION_H
//=======================================================
#define TABULATION_H
#include <iostream.h>
#include "jargon.h"
#include "term.h"
#include "clause.h"
//=======================================================
class Tabulation   
{
 public:
  Tabulation() : _str((ostream*)0), _jobId("uncknown_job_id") {};
  void assignStream(ostream* str) { _str = str; };
  void assignJob(const char* jobId) { _jobId = jobId; };

  // basic operations  
  void flush() { _str->flush(); };
  void open_unit() { *_str << "vout("; quoted_atom(_jobId); *_str << ','; };    
  void close_unit() { *_str << ").\n"; flush(); };
  
  void open_name() { *_str << '['; }; 
  void close_name() { *_str << ']'; };

  void comma() { *_str << ','; };
  void lbr() { *_str << '['; };
  void rbr() { *_str << ']'; };
  void quoted_atom(const char* s) { *_str << '\'' << s << '\''; };  
  void string(const char* s) { *_str << '\"' << s << '\"'; };
  void number(int n) { *_str << n; }; 
  void number(unsigned n) { *_str << n; }; 
  void number(long n) { *_str << n; }; 
  void number(unsigned long n) { *_str << n; }; 
  void number(float n) { *_str << n; };
  void number(double n) { *_str << n; };
  void boolean(bool b) { *_str << b; };
  void day(int dd) { number(dd); };
  void month(int mm) { number(mm); };
  void year(int yy) { number(yy); };
  void date(int dd,int mm,int yy) 
  { 
   lbr(); day(dd); comma(); month(mm); comma(); year(yy); rbr(); 
  };
  void hour(int hh) { number(hh); }; 
  void minute(int mm) { number(mm); };
  void second(int ss) { number(ss); }; 
  void time(int hh,int mm,int ss) 
  {
   lbr(); hour(hh); comma(); minute(mm); comma(); second(ss); rbr(); 
  };
  void simple_name(const char* n) { open_name(); quoted_atom(n); close_name(); };
  void simple_name(const char* n1,const char* n2)
  { 
   open_name(); quoted_atom(n1); comma(); quoted_atom(n2); close_name(); 
  };

  //============ derived shorthands: =======================================

  void header(const char* testId,const char* problemName)
  { 
   *_str << "%========  " << testId << " : " << problemName << "  ==========\n";
   
  };
  void problemName(const char* pn) 
  { 
   open_unit(); simple_name("problem_name"); comma(); quoted_atom(pn); close_unit();
    
  };
  void testId(const char* tid)
  {
   open_unit(); simple_name("test_id"); comma(); quoted_atom(tid); close_unit();
   
  };
  void hostName(const char* hn)
  {
   open_unit(); simple_name("host_name"); comma(); quoted_atom(hn); close_unit();   
  };
  void jobDate(int dd,int mm,int yy)
  {
   open_unit(); simple_name("job_date"); comma(); date(dd,mm,yy); close_unit();
   
  };
  void jobTime(int hh,int mm,int ss)
  {
   open_unit(); simple_name("job_time"); comma(); time(hh,mm,ss); close_unit();
   
  };
  void version(float num,const char* desc) 
  {
   open_unit(); simple_name("version"); comma();
   lbr(); number(num); comma(); quoted_atom(desc); rbr(); close_unit(); 
   
  };
  void versionDate()
  {
   open_unit(); simple_name("version_date"); comma(); quoted_atom(__DATE__); close_unit();
   
  };
  void versionTime()
  {
   open_unit(); simple_name("version_time"); comma(); quoted_atom(__TIME__); close_unit();
   
  };
  void option(const char* optName,long optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  };
  void option(const char* optName,unsigned long optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  };
  void option(const char* optName,bool optVal)
  {
   open_unit(); simple_name("option",optName); comma(); boolean(optVal); close_unit();
      
  };
  void option(const char* optName,const char* optVal) 
  {
   open_unit(); simple_name("option",optName); comma(); quoted_atom(optVal); close_unit();
   
  }; 
  void option(const char* optName,float optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  }; 
  void option(const char* optName,double optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  };
  
  void stat(const char* paramName,bool paramVal)
  {   
   open_unit(); simple_name("stat",paramName); comma(); boolean(paramVal); close_unit(); 
  };
  void stat(const char* paramName,int paramVal)
  {   
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit(); 
  };
  void stat(const char* paramName,unsigned paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,long paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,unsigned long paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,float paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,double paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,const char* paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); quoted_atom(paramVal); close_unit();
  };
  void result(char r)
  {
   open_unit();
   simple_name("result"); comma();
   if (r == '+') { quoted_atom("+"); }
   else 
    if (r == '-') { quoted_atom("-"); }
    else quoted_atom("0");       
   close_unit();
  };
  void terminationReason(const char* tr)
  {
   open_unit(); simple_name("termination_reason"); comma(); quoted_atom(tr); close_unit();
  };  
 
  void headerRank(TERM header,const long& rank)
  {
   open_unit(); simple_name("header_rank"); comma(); lbr(); 
   if (header.Positive()) { *_str << "++"; } else *_str << "--";
   if (header.IsEquality()) { *_str << "equal"; }
   else { header.makePositive(); *_str << header; };
   comma(); 
   number(header.arity());
   comma();
   number(rank);     
   rbr(); close_unit();
  };

  /*****
  void proofClause(const Clause* cl) 
  { 
   open_unit(); simple_name("proof"); comma(); cl->outputAsPrologTerm(*_str); close_unit();
  };
  *****/

  void proofClause(const Clause* cl) 
  {
   *_str << "vproof("; quoted_atom(_jobId); comma(); cl->outputAsPrologTerm(*_str); *_str << ").\n"; flush(); 
  };  
  
 private:
  ostream* _str;
  const char* _jobId;  
}; // class Tabulation


//=======================================================
#endif
