//
// File:         UnitConflict.h
// Description:  Unit conflict on flatterms.
// Created:      Jun 6, 2000, 17:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef UNIT_CONFLICT_H
//====================================================
#define UNIT_CONFLICT_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "GlobAlloc.h"
#include "GList.h"
#include "Tuple.h"
#include "term.h" 
#include "Flatterm.h" 
#include "clause.h"
#include "FTUnification.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIT_CONFLICT
 #define DEBUG_NAMESPACE "UnitConflict"
#endif 
#include "debugMacros.h"
//================================================= 

class UnitConflict
{
 public:
  UnitConflict()
   : _unification(),
     _rules((CLPairList*)0),
     _openFT()
  {
   CALL("constructor UnitConflict()");
  };
  void reset() 
  {
   CALL("reset()");
   for (CLPairList* clp = _rules; clp; clp = clp->tl())
    GlobAlloc::deallocate(clp->hd().el2,HellConst::MaxTermSize*sizeof(Flatterm));
   CLPairList::destroy(_rules);
   _rules = (CLPairList*)0; 
  };
  bool addRule(Clause* cl);
  Clause* checkForUnitConflict(const Flatterm* lit); 
  #ifndef NO_DEBUG_VIS
   ostream& outputRules(ostream& str) const;
  #endif
 private: 
  typedef Tuple2<Clause*,Flatterm*> CLPair;
  typedef GList<GlobAlloc,CLPair,UnitConflict> CLPairList;
 private:
  void addStandardRule(Clause* cl,TERM* lit);
 private:
  FTUnification _unification;
  CLPairList* _rules;
  Flatterm::Open _openFT;
  enum
  {
   _varOffset = HellConst::MaxNumOfVariables
  };
}; // class UnitConflict 

inline bool UnitConflict::addRule(Clause* cl)
{
 CALL("addRule(Clause* cl)");
 if (!cl->Unit()) return false;  
 addStandardRule(cl,cl->LitList()->firstElem()->literal()); 
 return true;
}; // bool UnitConflict::addRule(Clause* cl)

inline void UnitConflict::addStandardRule(Clause* cl,TERM* lit)
{
 CALL("addStandardRule(Clause* cl,TERM* lit)");
 Flatterm* ft = (Flatterm*)(GlobAlloc::allocate(HellConst::MaxTermSize*sizeof(Flatterm)));
 _rules = new CLPairList(CLPair(cl,ft),_rules);
 _openFT.Reset(ft);
 _openFT.PushStandardComplexTerm(lit,_varOffset);
 ft->symbolRef().Inversepolarity();    
}; // void UnitConflict::addStandardRule(Clause* cl,TERM* lit)


inline Clause* UnitConflict::checkForUnitConflict(const Flatterm* lit)
{
 CALL("checkForUnitConflict(const Flatterm* lit)");
 for (const CLPairList* rule = _rules; rule; rule = rule->tl())   
  {
   if (_unification.unify(lit,rule->hd().el2)) return rule->hd().el1; 
  }; 
 return (Clause*)0;
}; // Clause* UnitConflict::checkForUnitConflict(const Flatterm* lit)

#ifndef NO_DEBUG_VIS
 inline ostream& UnitConflict::outputRules(ostream& str) const
 {
  CALL("outputRules(ostream& str) const");
  str << "%================= Unit conflict rules: ============\n";
  for (const CLPairList* rule = _rules; rule; rule = rule->tl())
   str << "% UCR " << rule->hd().el1 << " (" << rule->hd().el2 << ")\n";   
  return str << "%=========== End of unit conflict rules. ============\n";
 }; // ostream& UnitConflict::outputRules(ostream& str) const
#endif

//================================================
#undef DEBUG_NAMESPACE
//================================================= 
#endif
