//
// File:         Vout.h
// Description:  
// Created:      Oct 20, 1999, 14:35
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//

#ifndef VOUT_H
#define VOUT_H

//===================================================================
#include "jargon.h"
#include "OutPort.h"
#include "PrologTerm.h"

//===================================================================


class Vout
{
 public:
  Vout() { init(); };
  void init() { jobId = "unknown_job_id"; port = new OutPort::Null(); };
  void assignPort(OutPort* p) { port = p; };
  void assignJob(const char* jid) { jobId = jid; }; 

  //============ basic output operations: ======================================

  void flush() { port->flush(); };
  void open_unit() { *port << "vout("; quoted_atom(jobId); *port << ','; };    
  void close_unit() { *port << ").\n"; flush(); };
  
  void open_name() { *port << '['; }; 
  void close_name() { *port << ']'; };

  void comma() { *port << ','; };
  void lbr() { *port << '['; };
  void rbr() { *port << ']'; };
  void quoted_atom(const char* s) { *port << '\'' << s << '\''; };  
  void string(const char* s) { *port << '\"' << s << '\"'; };
  void number(int n) { *port << n; }; 
  void number(unsigned n) { *port << n; }; 
  void number(long n) { *port << n; }; 
  void number(unsigned long n) { *port << n; }; 
  void number(float n) { *port << n; };
  void number(double n) { *port << n; };
  void boolean(bool b) { *port << b; };
  void day(int dd) { number(dd); };
  void month(int mm) { number(mm); };
  void year(int yy) { number(yy); };
  void date(int dd,int mm,int yy) 
  { 
   lbr(); day(dd); comma(); month(mm); comma(); year(yy); rbr(); 
  };
  void hour(int hh) { number(hh); }; 
  void minute(int mm) { number(mm); };
  void second(int ss) { number(ss); }; 
  void time(int hh,int mm,int ss) 
  {
   lbr(); hour(hh); comma(); minute(mm); comma(); second(ss); rbr(); 
  };
  void prologTerm(const PrologTerm& pt)
  {
   *port << pt;
  };  

  void simple_name(const char* n) { open_name(); quoted_atom(n); close_name(); };
  void simple_name(const char* n1,const char* n2)
  { 
   open_name(); quoted_atom(n1); comma(); quoted_atom(n2); close_name(); 
  };

  //============ derived shorthands: =======================================

  void header(const char* testId,const char* problemName)
  { 
   *port << "%========  " << testId << " : " << problemName << "  ==========\n";
   
  };
  void problemName(const char* pn) 
  { 
   open_unit(); simple_name("problem_name"); comma(); quoted_atom(pn); close_unit();
    
  };
  void testId(const char* tid)
  {
   open_unit(); simple_name("test_id"); comma(); quoted_atom(tid); close_unit();
   
  };
  void jobDate(int dd,int mm,int yy)
  {
   open_unit(); simple_name("job_date"); comma(); date(dd,mm,yy); close_unit();
   
  };
  void jobTime(int hh,int mm,int ss)
  {
   open_unit(); simple_name("job_time"); comma(); time(hh,mm,ss); close_unit();
   
  };
  void version(float num,const char* desc) 
  {
   open_unit(); simple_name("version"); comma();
   lbr(); number(num); comma(); quoted_atom(desc); rbr(); close_unit(); 
   
  };
  void versionDate()
  {
   open_unit(); simple_name("version_date"); comma(); quoted_atom(__DATE__); close_unit();
   
  };
  void versionTime()
  {
   open_unit(); simple_name("version_time"); comma(); quoted_atom(__TIME__); close_unit();
   
  };
  void option(const char* optName,long optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  };
  void option(const char* optName,unsigned long optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  };
  void option(const char* optName,bool optVal)
  {
   open_unit(); simple_name("option",optName); comma(); boolean(optVal); close_unit();
      
  };
  void option(const char* optName,const char* optVal) 
  {
   open_unit(); simple_name("option",optName); comma(); quoted_atom(optVal); close_unit();
   
  }; 
  void option(const char* optName,float optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  }; 
  void option(const char* optName,double optVal)
  {
   open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
  };
  void stat(const char* paramName,int paramVal)
  {   
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit(); 
  };
  void stat(const char* paramName,unsigned paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,long paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,unsigned long paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,float paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
  void stat(const char* paramName,double paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
  };
 void stat(const char* paramName,const char* paramVal)
  {
   open_unit(); simple_name("stat",paramName); comma(); quoted_atom(paramVal); close_unit();
  };
  void result(char r)
  {
   open_unit();
   simple_name("result"); comma();
   if (r == '+') { quoted_atom("+"); }
   else 
    if (r == '-') { quoted_atom("-"); }
    else quoted_atom("0");       
   close_unit();
  };
  void terminationReason(const char* tr)
  {
   open_unit(); simple_name("termination_reason"); comma(); quoted_atom(tr); close_unit();
  };  

  void proofClause(const PrologTerm& cl) 
  { 
   open_unit(); simple_name("proof"); comma(); prologTerm(cl); close_unit();
  };

 private:
  const char* jobId; 
  OutPort* port;  
}; // class Vout


//===================================================================
#endif
