#ifndef B_MATCHING_H
//=============================================================
#define B_MATCHING_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"
#include "sharing.h"
#include "MultisetOfVariables.h"
#include "NewClause.h"
#include "KBConstraint.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_B_MATCHING
 #define DEBUG_NAMESPACE "ST_PATH"
#endif
#include "debugMacros.h"
//====================================================
typedef SHARING_TREE_NODE ST_NODE; 

class ST_PATH
 {
  public:
   typedef TERM::Traversal::State TraversalState;
   class Iterator
   {
    public:
     Iterator() {};
     Iterator(const ST_PATH* path) { reset(path); }; 
     void reset(const ST_PATH* path) { _iter = path->nodes; _last = path->last; }
     const TERM& symbol() const { return (*_iter)->Tag(); };         
     bool next() { _iter++; return (_iter <= _last); };
    private:
     ST_NODE** _iter;
     ST_NODE** _last;
   }; // class Iterator
   class Traversal
   {
    public:
     Traversal() {};      
     Traversal(const ST_PATH* path) { reset(path); }; 
     void reset(const ST_PATH* path) { _iter = path->nodes; _last = path->last; };
     operator bool() const { return (_iter <= _last); };
     TraversalState state() const 
     {
      CALL("state() const");
      if (_iter > _last) return TERM::Traversal::End;
      if ((*_iter)->Tag().isVariable()) return TERM::Traversal::Var; 
      ASSERT((*_iter)->Tag().isComplex());
      return TERM::Traversal::Func; 
     };
     const TERM& symbol() const { return (*_iter)->Tag(); };   
     void next() { _iter++; };      
    private:
     ST_NODE** _iter;
     ST_NODE** _last; 
   };
  public:
   ST_PATH() : _variables(), _resolvent((NewClause*)0) {};
   void setResolvent(NewClause* res) { _resolvent = res; };
   void SetMemory(ST_NODE** nmem) { nodes = nmem; };
   void ExtractLeftMost(ST_NODE* node);
   void ExtractLeftMost(ST_NODE**& memory,ST_NODE*& node)
    {
     SetMemory(memory);
     ExtractLeftMost(node);
     memory = FreeMem();
     node = LastNode()->Below();
    };
   bool Alter();
   bool Alter(ST_NODE**& memory,ST_NODE*& node)
    {
     if (Alter())
      {
       memory = FreeMem();
       node = LastNode()->Below();
       return true;
      }
     else return false; 
    };
   bool SkipEqual(ST_NODE*& node);
   ST_NODE* LastNode() const { return *last; }; 
   ST_NODE** FreeMem() const { return last + 1; };

   /*****
   void ResetIter() { iter = nodes; };
   const TERM& IterCurrSym() const { return (*iter)->Tag(); };
   bool IterNext() { iter++; return (iter <= last); };
   *****/

   /*********
   void traversalReset() { iter = nodes; };
   bool traversal() const { return (iter <= last); };
   void traversalNext() { iter++; };
   const TERM& traversalSymbol() const { return (*iter)->Tag(); }; 
   TraversalState traversalState() const
   {
    CALL("traversalState()");
    if (iter > last) return TERM::Traversal::End;
    if ((*iter)->Tag().isVariable()) return TERM::Traversal::Var; 
    ASSERT((*iter)->Tag().isComplex());
    return TERM::Traversal::Func; 
   };     
   **********/ 

   MultisetOfVariables<HellConst::MaxNumOfVariables>& Variables() { return _variables; };    
   const MultisetOfVariables<HellConst::MaxNumOfVariables>& Variables() const { return _variables; };
   const ulong& NumOfFuncs() const { return _numOfFuncs; };

   void CountVarsAndFuncs()
   {
    CALL("CountVarsAndFuncs()");
    _numOfFuncs = 0;
    _variables.Reset();
    TERM currSym;
    for (const ST_NODE* const * currNode = nodes; currNode <= last; currNode++)
     {
      currSym = (*currNode)->Tag();
      if (currSym.isVariable())
       {
	_variables.Register(currSym.var());
       } 
      else _numOfFuncs++; 
     };
   }; // void CountVarsAndFuncs() 

   bool ground() const
   {
    for (const ST_NODE* const * currNode = nodes; currNode <= last; currNode++)
     if ((*currNode)->Tag().isVariable()) return false;
    return true;
   };

   bool nonground() const
   {
    for (const ST_NODE* const * currNode = nodes; currNode <= last; currNode++)
     if ((*currNode)->Tag().isVariable()) return true;
    return false;
   };

   ulong Weight() const
   {
    return (last - nodes) + 1;
   };

   bool Collect()
    {
     static Iterator iter;
     iter.reset(this);
     do
      {
       if (!_resolvent->PushSymbol(iter.symbol())) return false;
      }
     while (iter.next());
     return true;
    };      
   
   #ifdef DEBUG_NAMESPACE
    void debugSetEndOfMemory(ST_NODE** e)
    {
     debugEndOfMemory = e;
    };
   #endif  

  private:
   ST_NODE** nodes;
   ST_NODE** last;  
   long holes;
   ST_NODE** iter;
   MultisetOfVariables<HellConst::MaxNumOfVariables> _variables;
   ulong _numOfFuncs;
   NewClause* _resolvent;
   #ifdef DEBUG_NAMESPACE
    ST_NODE** debugEndOfMemory;  
   #endif
  friend class Iterator;
  friend class Traversal;   
 }; // class ST_PATH      

//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_B_MATCHING
 #define DEBUG_NAMESPACE "BM_INSTR"
#endif
#include "debugMacros.h"
//====================================================      
class BM_INSTR
 {
  public:
   enum TAG { START, FUNC, VAR_FIRST_OCC, VAR, SUCCESS };
  private:
   TAG tag;
   TERM symbol;
   BM_INSTR* backtrack_instr;
  public:
   BM_INSTR() {};
   TAG& Tag() { return tag; };
   const TAG& Tag() const { return tag; };         
   TERM& Symbol() { return symbol; };
   const TERM& Symbol() const { return symbol; };      
   BM_INSTR*& BacktrackInstr() { return backtrack_instr; };
   const BM_INSTR* const & BacktrackInstr() const { return backtrack_instr; };

   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str,const BM_INSTR* code_start) const
    {
     str << '[' << (int)(this - code_start) << ']';
     switch (Tag())
      {
       case START:         str << "START         " << Symbol(); break;         
       case FUNC:          str << "FUNC          " << Symbol(); break;   
       case VAR_FIRST_OCC: str << "VAR_FIRST_OCC " << Symbol(); break; 
       case VAR:           str << "VAR           " << Symbol(); break; 
       case SUCCESS:       str << "SUCCESS"; break;
      };          
     return str << " backtr[" << (int)(BacktrackInstr() - code_start) << "] ";
    }; 

    static ostream& outputCode(ostream& str,const BM_INSTR* code)
    {
     const BM_INSTR* start = code;
     while (code->Tag() != SUCCESS) { code->output(str,start) << '\n'; code++; };
     return code->output(str,start) << '\n';
    };
   #endif    

 }; // class BM_INSTR 

//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_B_MATCHING
 #define DEBUG_NAMESPACE "B_MATCHING"
#endif
#include "debugMacros.h"
//====================================================

class B_MATCHING
 {
  public:
   class BMConstr : public KBConstraint
   {
    public:
     BMConstr(ST_PATH* subst) : KBConstraint(), _subst(subst) {};
     void makeAlwaysValid(const bool& fl) { _alwaysValid = fl; };
     void compile() { _endOfCode = _code; KBConstraint::compile(_endOfCode); };
     bool check() 
     { 
      CALL("BMConstr::check()"); 

      //DF; cout << "******* CHECK\n";

      if (_alwaysValid) 
       {
   
	 //DF; cout << " ALWAYS VALID\n"; 

        return true;
       };
      const KBCommand* com = _code;     
      
      ulong w1;
      ulong w2;
      ulong weightRegister = 0;
      //ST_PATH* substTerm;
      TERM substSym;
      bool substCanBeEqual;
      bool equalTerms;
      static ST_PATH::Iterator pathIter;

      //DF; outputCode(cout << "******* CODE\n") << "\n";
 
      exec_command: 
       ASSERT(com < _endOfCode);

       //DF; com->output(cout << "        COMMAND ") << "\n";

       switch (com->tag())
	{
	 case KBCommand::Success: return true;          
	 case KBCommand::Failure: return false;

         case KBCommand::SuccIfWeightGreater:
	  w1 = (_subst[com->var()].Weight()) * com->coeff();
	  w2 = com->weightConst();

	  //DF; cout << "w1 = " << w1 << "   w2 = " << "\n";

          if (w1 != w2) return (w1 > w2);
          break;

         case KBCommand::FailIfWeightGreater:
	  if (((_subst[com->var()].Weight()) * (com->coeff())) > com->weightConst()) return false;
          break;
          
         case KBCommand::FailIfNonground:
	  if (_subst[com->var()].nonground()) return false;        
          break; 

         case KBCommand::SuccIfWeightLess:
          w1 = (_subst[com->var()].Weight()) * com->coeff();
	  w2 = com->weightConst();
          if (w1 != w2) return (w1 < w2); 
          break;

	 case KBCommand::InitWR: weightRegister = 0; break;
                   
         case KBCommand::AddWR:
	  weightRegister += ((_subst[com->var()].Weight()) * (com->coeff()));
          break;

         case KBCommand::SuccIfWRGreater:
	  if (weightRegister > com->weightConst()) return true;
          break;

         case KBCommand::FailIfWRLess:
	  if (weightRegister < com->weightConst()) return false;      
          break;

         case KBCommand::FailIfWRGreater:
          if (weightRegister > com->weightConst()) return false;
          break;
 
         case KBCommand::SuccIfWRLess:
	  w2 = com->weightConst();
          if (weightRegister != w2) return (weightRegister < w2);
          break;

         case KBCommand::InitPosSubst:
	  KBConstraint::initPosSubst(com->coeff());
          break;

         case KBCommand::InitNegSubst:
          KBConstraint::initNegSubst(com->coeff());
          break;
         
         case KBCommand::PosSubst:
	  w1 = com->coeff();

	  //substTerm = _subst + com->var();
          //substTerm->ResetIter();

          pathIter.reset(_subst + com->var());
          do
	   {
	     //substSym = substTerm->IterCurrSym();
            substSym = pathIter.symbol();
            if (substSym.isVariable())
	     {
	      KBConstraint::posSubstVar(substSym.var(),w1);
             }
            else 
	     {
              KBConstraint::posSubstFunc(w1);       
             };
           }
          while (pathIter.next());
          //while (substTerm->IterNext());
          break;

	 case KBCommand::NegSubst:
	  w1 = com->coeff();
	  //substTerm = _subst + com->var();
          //substTerm->ResetIter();
          pathIter.reset(_subst + com->var());
          do
	   {
	    //substSym = substTerm->IterCurrSym();
	    substSym = pathIter.symbol();

            if (substSym.isVariable())
	     {
	      if (!KBConstraint::negSubstVar(substSym.var(),w1)) return false;
             }
            else 
	     {
	      if (!KBConstraint::negSubstFunc(w1)) return false;    
             };
           }
          while (pathIter.next());
          //while (substTerm->IterNext());
          break;

	 case KBCommand::CompareSubst:
	  if (!KBConstraint::substGreaterOrEq(substCanBeEqual)) return false;
          if (!substCanBeEqual) return true;
          break;  
         
	 case KBCommand::LexCompareComplexComplex:

	   //DF; cout << "LEX " << com->term1() << " vs " << com->term2() << "\n"; 

          if (lexicographicallyGreaterOrEqual(com->term1(),com->term2(),equalTerms)) 
           {
	     //DF; cout << "^^^^^^^^^^^^^^^^^^^^^^^^^^^^ SUCCESS!!!!!!!!!!!\n";          
            return true;
           };
          if (!equalTerms) return false;

         #ifdef DEBUG_NAMESPACE
	  default: ICP("ICP0"); return false;
         #else
          #ifdef _SUPPRESS_WARNINGS_
           default: return false;
          #endif
         #endif
	}; // switch (com->tag()) 
        
       com++;
       goto exec_command;
     }; // bool check()

     #ifndef NO_DEBUG_VIS
      ostream& outputCode(ostream& str) 
      {
       return KBCommand::outputCode(str,_code,_endOfCode);
      }; 
     #endif   

    private: 
     BMConstr() : KBConstraint(), _subst(0) {};
     bool lexicographicallyGreaterOrEqual(const TERM* t1,const TERM* t2,bool& equal);     
      
    private:
     bool _alwaysValid;
     ST_PATH* const _subst;
     enum
     {
      _maxCodeSize = 4*HellConst::MaxNumOfVariables + 3 + 1
     };
     KBCommand _code[_maxCodeSize];
     KBCommand* _endOfCode;
   };
     

   class TermTraversal
   {
    public:
     typedef TERM::Traversal::State State;
    public:
     TermTraversal(ST_PATH* subst) : _top(DOP("B_MATCHING::TermTraversal::_top")), _subst(subst), _pathTraversal() {};
     TermTraversal(B_MATCHING& m) : _top(), _subst(m.subst), _pathTraversal() {};
     void reset(const TERM& t) 
     {
      _onTop = true;
      _top.reset(t);
     };    
     void reset(const TERM* t) 
     {
      _onTop = true;
      _top.reset(t);
     };    
     State state() const 
     { 
      if (_onTop) return _top.state();
      //return _currSubstTerm->traversalState(); 
      return _pathTraversal.state();
     };
     operator bool() const { return (bool)_top; };
     const TERM& symbol() const 
     {
      CALL("TermTraversal::symbol() const"); 
      ASSERT((bool)_top);     
      if (_onTop) { ASSERT(_top.symbol().isComplex()); return _top.symbol(); };
      ASSERT(_top.symbol().isVariable());
      //return _currSubstTerm->traversalSymbol();
      return _pathTraversal.symbol();
     }; 
     
     const TERM& symbolOnTop() const 
     {
      CALL("symbolOnTop() const");
      ASSERT((bool)_top);
      return _top.symbol(); 
     };

     void next()
     {
      CALL("TermTraversal::next()");
      if (_onTop)
       {
        ASSERT((bool)_top); 
        ASSERT(_top.symbol().isComplex());
        top_next:
         ASSERT((bool)_top);
         _top.next();
         if (_top.state() == TERM::Traversal::Var)
	  {
	   _onTop = false;
           //_currSubstTerm = _subst + _top.symbol().var();
           //_currSubstTerm->traversalReset();
           _pathTraversal.reset(_subst + _top.symbol().var());
          };
       }       
      else // inside a substitution term 
       {
	ASSERT(_top.state() == TERM::Traversal::Var);
        //_currSubstTerm->traversalNext();
	_pathTraversal.next();
        //if (!_currSubstTerm->traversal()) // end of current term
        if (!_pathTraversal)  // end of current term
	 {
	  _onTop = true;
          goto top_next;
         };
       };
     };

    private:
     TermTraversal() {};
    private:
     bool _onTop; 
     TERM::Traversal _top;          
     ST_PATH* _subst; 
     //ST_PATH* _currSubstTerm;
     ST_PATH::Traversal _pathTraversal;
   }; // class TermTraversal 
  public:
   B_MATCHING(SHARING_INDEX* ind,NewClause* res) 
    : index(ind), 
      var_counter(),
      _ordConstraint(subst)
      #ifdef DEBUG_NAMESPACE
       , debugTermWithSubst(subst),
         debugTermTraversal("B_MATCHING::debugTermTraversal")
      #endif 
   {
    for (ulong v = 0; v < HellConst::MaxNumOfVariables; v++)
     { 
      subst[v].setResolvent(res);
      DOP(subst[v].debugSetEndOfMemory(subst_memory + HellConst::MaxTermSize));
     };
   };

   BMConstr& Constraint() { return _ordConstraint; };
   void SetQuery(const TERM* q)
   {
    CALL("SetQuery(TERM* q)");
    ASSERT(q->isComplex());
    query = q;
    TERM::Iterator iter;
    iter.Reset(q); 
    PushStart(*q);
    while (iter.Next()) PushSymbol(iter.symbolRef());
    PushSuccess();           
    curr_instr = code; 
   }; 
   bool NextMatch();
   const TERM* Query() const { return query; };
   TERM* Match() const 
   {
    CALL("Match() const");
    ASSERT(match->isComplex());
    ASSERT(((B_MATCHING*)this)->SoundResult()); 
    return match; 
   };
   bool CollectSubst(unsigned long var) { return subst[var].Collect(); };
   #ifndef NO_DEBUG_VIS
    ostream& outputCode(ostream& str) const
    {
     return BM_INSTR::outputCode(str,code);
    };
   #endif
  
   #ifdef DEBUG_NAMESPACE
    bool SoundResult() 
    {
     CALL("SoundResult()");
     TERM dummyQ;
     dummyQ.SetFirst((TERM*)query);       
     TERM dummyR;
     dummyR.SetFirst((TERM*)match); 
     debugTermWithSubst.reset(dummyQ);
     debugTermTraversal.reset(dummyR);
     while (debugTermTraversal)
      {
       if (!debugTermWithSubst) return false;
       if (debugTermWithSubst.symbol() != debugTermTraversal.symbol()) return false;
       debugTermTraversal.next();
       debugTermWithSubst.next();
      }; 
     return !debugTermWithSubst;
    };
   #endif


  private:
   B_MATCHING()
    : _ordConstraint(subst)    
    #ifdef DEBUG_NAMESPACE
     , debugTermWithSubst((ST_PATH*)0)
    #endif
   {
   };
   void PushStart(const TERM& fun)
    {
     curr_instr = code;
     var_counter.Reset(); 
     curr_instr->Tag() = BM_INSTR::START;
     curr_instr->Symbol() = fun;
     curr_instr->BacktrackInstr() = curr_instr;
     last_backtrack_instr = curr_instr;
     curr_instr++;
     subst_free_mem = subst_memory;
    };
   void PushSymbol(const TERM& sym)
    {
     CALL("PushSymbol(const TERM& sym)");
     ASSERT(curr_instr < code + (HellConst::MaxTermSize + 1));
     curr_instr->Symbol() = sym;
     curr_instr->BacktrackInstr() = last_backtrack_instr;
     if (sym.isVariable())
      {
       unsigned long var = sym.var();
       if (var_counter.Occurences(var))
        {
         curr_instr->Tag() = BM_INSTR::VAR;
        }
       else // first occurence of the variable
        {
         curr_instr->Tag() = BM_INSTR::VAR_FIRST_OCC;
         last_backtrack_instr = curr_instr;
        };
       var_counter.Register(var);
      }
     else // functor
      {
       curr_instr->Tag() = BM_INSTR::FUNC;
      };
     curr_instr++;
    };
  void PushSuccess() 
   {
    CALL("PushSuccess()");
    ASSERT(curr_instr < code + (HellConst::MaxTermSize + 2));
    curr_instr->Tag() = BM_INSTR::SUCCESS;
    curr_instr->BacktrackInstr() = last_backtrack_instr;
    curr_instr++;
   };
   
  bool checkConstraint()
   {
    CALL("checkConstraint()");
    return _ordConstraint.check();
   };

  
   #ifdef DEBUG_NAMESPACE
    bool checkSFMRange() const 
    {  
     return (subst_free_mem >= subst_memory) && (subst_free_mem < subst_memory + HellConst::MaxTermSize);
    };    
   #endif

  private:
   SHARING_INDEX* index;
   BM_INSTR code[HellConst::MaxTermSize + 2];
   BM_INSTR* curr_instr; 
   MultisetOfVariables<HellConst::MaxNumOfVariables> var_counter;
   BM_INSTR* last_backtrack_instr;
   const TERM* query;
   TERM* match;
   ST_PATH subst[HellConst::MaxNumOfVariables];
   ST_NODE* subst_memory[HellConst::MaxTermSize];
   ST_NODE** subst_free_mem;  
   BMConstr _ordConstraint;
   #ifdef DEBUG_NAMESPACE
    TermTraversal debugTermWithSubst;
    TERM::Traversal debugTermTraversal;
   #endif
  friend class TermTraversal;
 }; // class B_MATCHING


//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_B_MATCHING
 #define DEBUG_NAMESPACE "B_MATCHING"
#endif
#include "debugMacros.h"
//====================================================


inline bool B_MATCHING::NextMatch()
 {
  CALL("NextMatch()");
  ST_NODE* curr_node;
  TERM sym;
  unsigned long var;     
  switch (curr_instr->Tag())
   {
    case BM_INSTR::START:
     if (!(curr_instr->Symbol().arity()))
      {
       match = index->Const(curr_instr->Symbol());
       curr_instr++;       
       return true;
      };
     curr_node = *(index->Tree(curr_instr->Symbol()));

     if (!curr_node) return false;
     curr_instr++;
     goto perform_instr;

    case BM_INSTR::SUCCESS: goto backtrack;
    #ifdef DEBUG_NAMESPACE
     default: ICP("ICP0"); return false;
    #else
     #ifdef _SUPPRESS_WARNINGS_
      default: return false;
     #endif
    #endif
   };  
  
  perform_instr:
   switch (curr_instr->Tag())
    {
     case BM_INSTR::FUNC:
      ASSERT(!curr_node->IsLeaf());
      sym = curr_instr->Symbol();
      while (curr_node->Tag() != sym)
       {
        curr_node = curr_node->Alternative();
        if (!curr_node) goto backtrack;
       };      
      curr_node = curr_node->Below();
      curr_instr++;
      goto perform_instr;

     case BM_INSTR::VAR_FIRST_OCC:
      var = curr_instr->Symbol().var();
      subst[var].ExtractLeftMost(subst_free_mem,curr_node);
      ASSERT(checkSFMRange());
      curr_instr++;       
      goto perform_instr;

     case BM_INSTR::VAR:
      var = curr_instr->Symbol().var();
      if (subst[var].SkipEqual(curr_node))
       {
        curr_instr++;       
        goto perform_instr;
       }       
      else goto backtrack; 
 
     case BM_INSTR::SUCCESS: 
      // curr_node must be a leaf here
      ASSERT(curr_node->IsLeaf());
      if (!checkConstraint()) goto backtrack;
      match = curr_node->Term(); 
      return true;      

     #ifdef DEBUG_NAMESPACE
      default: ICP("ICP1"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif    
    };  
  
  backtrack:
   curr_instr = curr_instr->BacktrackInstr();
   switch (curr_instr->Tag())
    {
     case BM_INSTR::START: return false;

     case BM_INSTR::VAR_FIRST_OCC:
      var = curr_instr->Symbol().var();
      if (subst[var].Alter(subst_free_mem,curr_node))
       {
        ASSERT(checkSFMRange());
        curr_instr++;       
        goto perform_instr;
       }       
      else goto backtrack;

     #ifdef DEBUG_NAMESPACE
      default:  ICP("ICP2"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif   
    };

 }; // bool B_MATCHING::NextMatch()

//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_B_MATCHING
 #define DEBUG_NAMESPACE "B_MATCHING::BMConstr"
#endif
#include "debugMacros.h"
//====================================================


inline bool B_MATCHING::BMConstr::lexicographicallyGreaterOrEqual(const TERM* t1,const TERM* t2,bool& equal)
{
 CALL("lexicographicallyGreaterOrEqual(const TERM* t1,const TERM* t2,bool& equal)");
 B_MATCHING::TermTraversal iter1(_subst);
 B_MATCHING::TermTraversal iter2(_subst);
 iter1.reset(t1);
 iter2.reset(t2);
 while ((bool)iter1)
  {
   ASSERT((bool)iter2);
   if (iter1.symbol() != iter2.symbol())
    {
     equal = false;
     return (iter1.symbol().isComplex()) 
            && (iter2.symbol().isComplex()) 
            && (iter1.symbol().greaterPrecedence(iter2.symbol())); 
    };

   //DF; cout << "[" << iter1.symbolOnTop() << "]" << iter1.symbol() 
   //      << " <> " << "[" << iter2.symbolOnTop() << "]" << iter2.symbol() << "\n";

   iter1.next();
   iter2.next();
  };
 equal = true;  
 return true;
}; // bool B_MATCHING::BMConstr::lexicographicallyGreaterOrEqual(const TERM* t1,const TERM* t2,bool& equal)


//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_B_MATCHING
 #define DEBUG_NAMESPACE "ST_PATH"
#endif
#include "debugMacros.h"
//====================================================
inline bool ST_PATH::Alter()
 {
  CALL("Alter()");
  ASSERT(!holes);
  ST_NODE* node;
  try_last:
   if (last < nodes) return false;
   node = *last;   
   last--;
   holes = holes - node->Tag().arity() + 1;
   if (node->Alternative())
    {
     ASSERT(holes > 0);  
     // now extract leftmost continuation    
     node = node->Alternative();
     do
      {
       last++;
       ASSERT(last < debugEndOfMemory);
       *last = node;
       holes = holes + node->Tag().arity() - 1;  
       node = node->Below();
      }
     while (holes);
     return true;
    }  
   else goto try_last; 
 }; // bool ST_PATH::Alter() 

inline void ST_PATH::ExtractLeftMost(ST_NODE* node)
 {
  CALL("ExtractLeftMost(ST_NODE* node)"); 
  *nodes = node;
  last = nodes;
  holes = node->Tag().arity();
  while (holes)
   {
    last++;
    ASSERT(last < debugEndOfMemory);
    node = node->Below();
    *last = node;
    holes = holes + node->Tag().arity() - 1;  
   };
 }; // void ST_PATH::ExtractLeftMost(ST_NODE* node) 

inline bool ST_PATH::SkipEqual(ST_NODE*& node)
 {
  for (ST_NODE** n = nodes; n <= last; n++)
   {
    TERM sym = (*n)->Tag();    
    try_curr_node:
     if (node->Tag() == sym)
      {
       node = node->Below();
      }
     else
      {
       node = node->Alternative();
       if (node) goto try_curr_node;
       return false; 
      }; 
   }; 
  return true; 
 }; // bool ST_PATH::SkipEqual(ST_NODE*& node) 
//====================================================
#undef DEBUG_NAMESPACE
//=============================================================
#endif
