//
// File:         backward_superposition.h
// Description:  Classes dealing with backward superposition.
// Created:      Mar 12, 2000, 19:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef BACKWARD_SUPERPOSITION_H
//==================================================================
#define BACKWARD_SUPERPOSITION_H
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"   
#include "para_machine.h"
#include "NewClause.h"
#include "Stack.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "BPARA_MACH"
#endif
#include "debugMacros.h"
//=================================================
class BPARA_MACH
 {
  private:
   PARA_MACH para_mach;
   OR_INDEX* _unifIndex;
   bool init_state;
   ORD_CONSTR* ord_constr; 
   Clause* query; // mostly for debugging
   bool unit_query;

   BackParaCandidates::Iterator _candidates;


  public:
   BPARA_MACH(ORD_CONSTR* constr,OR_INDEX* unifIndex) 
    : para_mach(), _unifIndex(unifIndex), ord_constr(constr), _candidates() {};
   ~BPARA_MACH() { CALL("destructor ~BPARA_MACH()"); };
   void SetIndex(ulong ind) { para_mach.SetIndex(ind); };
   ulong GetIndex() const { return para_mach.GetIndex(); };
   void SetUnitQuery(bool fl) { unit_query = fl; };
   void SetQuery(Clause* cl) { query = cl; };
   bool VariablesFromIndexAffected() { return para_mach.VariablesFromIndexAffected(); };
   
   void LoadLHS(PREFIX_SYM* t)
    {
     para_mach.CompileComplexTerm(t);     
     init_state = true;
    };
            
   bool NextCandidate(Clause*& candCl,BackParaCandidates::LitRedexPair*& candLitRedexPair)
   {
    CALL("NextCandidate(Clause*& candCl,BackParaCandidates::LitRedexPair*& candLitRedexPair)"); 

    if (init_state)
     {
      init_state = false;
      if (para_mach.FindFirstLeaf(*(_unifIndex->BParaTree())))
       {	         
        if (ord_constr->Violated()) goto try_next_leaf;
        _candidates.reset(*(para_mach.BPCandidatesFromLeaf()));            
        ALWAYS(_candidates.next(candCl,candLitRedexPair));
	return true;
       }
      else return false;
     }
    else // !init_state
     {
      if (_candidates.next(candCl,candLitRedexPair)) 
       {
        return true;
       };
      try_next_leaf: 

       if (para_mach.FindNextLeaf())
        {	             
         if (ord_constr->Violated()) goto try_next_leaf;
         _candidates.reset(*(para_mach.BPCandidatesFromLeaf()));            
         ALWAYS(_candidates.next(candCl,candLitRedexPair));
	 return true;
        }
       else 
        {
         return false;  
	};
     };     
   }; // bool NextCandidate(Clause*& candCl,BackParaCandidates::LitRedexPair*& candLitRedexPair)      
 }; // class BPARA_MACH

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "BPARA_MACHINE"
#endif
#include "debugMacros.h"
//================================================= 
class BPARA_MACHINE 
 {
  private:
   BPARA_MACH bpara_mach;
   PREFIX_SYM fst_arg_mem[HellConst::MaxTermSize];
   PREFIX_SYM snd_arg_mem[HellConst::MaxTermSize];
   PREFIX_SYM* rhs;
   ulong index;
   ulong eq_index;
   bool unordered_eq;
   bool wrap_args;

   //RES_PAIR_LIST* candidate;
   
   Clause* _candClause;
   BackParaCandidates::LitRedexPair* _candLitRedexPair;

   ORD_CONSTR ordering_constraint;  

   Clause* query;
   NewClause* _resolvent;
   long _simultaneousness;

   UnifKBConstr _unifConstr;
    
   bool _noRulesFlag;  

  public: 
   BPARA_MACHINE(NewClause* res,OR_INDEX* unifIndex) 
    : bpara_mach(&ordering_constraint,unifIndex),
     
      ordering_constraint(), 
      _resolvent(res),
      _unifConstr()
   {
    fst_arg_mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity));
    snd_arg_mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity));
   };
   ~BPARA_MACHINE() { CALL("destructor ~BPARA_MACHINE()"); };
   void setNoRulesFlag(bool fl) { _noRulesFlag = fl; };
   void setSimultaneous(long degree) { _simultaneousness = degree; };
   void SetIndex(ulong ind) { index = ind; bpara_mach.SetIndex(ind); };
   ulong GetIndex() const { return index; };    
   bool VariablesFromIndexAffected() { return bpara_mach.VariablesFromIndexAffected(); };
   void SetQuery(Clause* cl) 
    {
     query = cl;
     bpara_mach.SetQuery(cl);
     bpara_mach.SetUnitQuery(cl->Unit());
    };

   //   RES_PAIR_LIST* Candidate() const { return candidate; };
 
   Clause* candClause() const { return _candClause; };

   void LoadEq(const TERM* eq_lit,ulong eq_ind)
    {
     CALL("LoadEq(const TERM* eq_lit,ulong eq_ind)");
     ASSERT(eq_lit->IsEquality());
     ASSERT(eq_lit->Positive());
     eq_index = eq_ind; 
     const TERM* arg_ref = eq_lit->Args();
     PREFIX_SYM* mem = fst_arg_mem + 1;
     StandardTermToPrefix(eq_index,*arg_ref,mem);
     fst_arg_mem[0].SetBracket(mem);
     
     // another argument
     arg_ref = arg_ref->Next();
     mem = snd_arg_mem + 1;
     StandardTermToPrefix(eq_index,*arg_ref,mem);
     snd_arg_mem[0].SetBracket(mem);
     
     bpara_mach.LoadLHS(fst_arg_mem);
     rhs = snd_arg_mem + 1;
     unordered_eq = eq_lit->IsUnorderedEq();
     wrap_args = unordered_eq;
   
     if (unordered_eq)
      {
       ordering_constraint.Compile(fst_arg_mem,snd_arg_mem);
      }
     else ordering_constraint.MakeAlwaysValid();

    };
   
   bool Paramodulate()
    {
     CALL("Paramodulate()");
     try_next_cand:
      if (bpara_mach.NextCandidate(_candClause,_candLitRedexPair)) 
       {
	if (_noRulesFlag && _candClause->IsRule()) goto try_next_cand;
        return true;
       };
      if (wrap_args)
       {
        wrap_args = false;
        bpara_mach.LoadLHS(snd_arg_mem);
        rhs = fst_arg_mem + 1;    
        ordering_constraint.Inverse();
        goto try_next_cand;          
       };
      return false; 
    };  
  
  bool CollectLiterals() 
   {
    CALL("CollectLiterals()");

    //DF; _candClause->output(cout) << "\n"; 
    //DF; _candLitRedexPair->constraint()->output(cout) << "\n";

    const SharedKBConstr* constr = _candLitRedexPair->constraint();
    if (constr && _unifConstr.checkConstr(constr,index)) 
     {
       
       //DF; cout << "REJECTED\n";

      return false;
     };
    
    ulong rewritten_lit_num = _candLitRedexPair->litNum();

    LiteralList::Element* rewritten_lit = _candClause->LitList()->nth(rewritten_lit_num);
    for (LiteralList::Iterator iter(_candClause->LitList());iter.notEnd();iter.next())
     {
      if (iter.currentElement() == rewritten_lit)
       {        
        if (!_resolvent->PushRewrittenLit(index,iter.currentLiteral(),_candLitRedexPair->redex(),rhs,iter.currentElement()->isInherentlySelected()))
         return false;
       }
      else
       {
	if ((_simultaneousness > 0) && (iter.currentElement()->arity()))
	 {
          if (!_resolvent->PushRewrittenLit(index,iter.currentLiteral(),_candLitRedexPair->redex(),rhs,iter.currentElement()->isInherentlySelected()))
           return false;   
         }
        else
         if (!_resolvent->PushStandardLit(index,iter.currentElement()->literal(),iter.currentElement()->isInherentlySelected())) return false;  
       };  
     };    

    RARE((query->PreorderedPositiveUnitEquality() 
         && (!bpara_mach.VariablesFromIndexAffected())
	 && (query != _candClause)));      

    return true;
   };    
 }; // class BPARA_MACHINE 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "BPARA_PREMISE"
#endif
#include "debugMacros.h"
//================================================= 

class BPARA_PREMISE
 { 
  private:
   ulong index;
   Clause* clause;

   LiteralList* allLiterals;
   LiteralList::Element* firstNonmaxLit;
   LiteralList::Iterator maxLitIter;
 
   WEIGHT_FORMULA nonmax_weight_form;
   WEIGHT_FORMULA max_weight_form;
   NewClause* _resolvent;
  public:
   BPARA_PREMISE(NewClause* res) 
     : maxLitIter(), nonmax_weight_form(), max_weight_form(), _resolvent(res)
   {}; 
   ~BPARA_PREMISE() { CALL("destructor ~BPARA_PREMISE()"); };
   void SetIndex(ulong ind) 
    {
     index = ind; 
     nonmax_weight_form.SetIndex(ind);
     max_weight_form.SetIndex(ind);
    };
   ulong GetIndex() const { return index; };
   Clause* TheClause() const { return clause; };
   void CollectNonmaxWeightForm()
    {
     CALL("CollectNonmaxWeightForm()");
     nonmax_weight_form.Reset();

     for (LiteralList::Iterator iter(allLiterals,firstNonmaxLit);iter.notEnd();iter.next())
      nonmax_weight_form.Compile(iter.currentLiteral());
    };

   void RecollectMaxWeightForm()
    {
     CALL("RecollectMaxWeightForm()");
     max_weight_form.Reset();

     for (LiteralList::Iterator iter(allLiterals); iter.currentElement() != firstNonmaxLit;iter.next())
      {
       if (iter.currentElement() == maxLitIter.currentElement())
        {              
         ASSERT(iter.currentElement()->isEquality());
         ASSERT(iter.currentElement()->positive());
         if (iter.currentElement()->isOrderedEq())
          max_weight_form.Compile(*(iter.currentElement()->arg2()));
        }
       else max_weight_form.Compile(iter.currentElement()->literal());
      }; 
    }; 

   
   void Load(Clause* cl) 
    {
     CALL("Load(Clause* cl)");
     clause = cl;
     allLiterals = cl->LitList();
     firstNonmaxLit = allLiterals->nth(allLiterals->numOfSelectedLits());
     maxLitIter.reset(allLiterals);
    };


   const TERM* NextEq()
    {
     CALL("NextEq()");
     const TERM* lit;
     check_lit:
      if ( maxLitIter.currentElement() == firstNonmaxLit) return (TERM*)0;     
      lit = maxLitIter.currentLiteral();
      maxLitIter.next();
      if (lit->Positive() && lit->IsEquality()) 
       {
        RecollectMaxWeightForm();
        return lit;
       }; 
      goto check_lit;
    }; 
    
   long MinResultWeight()
    {
     return nonmax_weight_form.Compute() + max_weight_form.Compute();
    }; 

   bool CollectLiterals() 
    {
     CALL("CollectLiterals()");
     // collect max. literals except the used equality

     //cout << "REWR RULE " << maxLitIter.currentLiteral() << "\n";

     for (LiteralList::Iterator iter(allLiterals); iter.currentElement() != firstNonmaxLit;iter.next())
      if (iter.currentElement()->next() != maxLitIter.currentElement())
       {      

	 //cout << "PUSH MAX " << iter.currentLiteral() << "\n";

        if (!_resolvent->PushStandardLit(index,iter.currentLiteral(),iter.currentElement()->isInherentlySelected())) return false; 

	//_resolvent->output(cout << "Success, RES3 = ") << "\n";

       };      

     // collect nonmax. literals
     for (LiteralList::Iterator iter(allLiterals,firstNonmaxLit);iter.notEnd();iter.next())
      {
       ASSERT(!(iter.currentElement()->isInherentlySelected()));
       if (!_resolvent->PushStandardLit(index,iter.currentLiteral(),false)) return false;   
      };  
     return true;
    }; 
 }; // class BPARA_PREMISE

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "BACKWARD_SUPERPOSITION"
#endif
#include "debugMacros.h"
//================================================= 

class BACKWARD_SUPERPOSITION
 {
  private:
   BPARA_PREMISE bpara_premise;
   BPARA_MACHINE bpara_machine;
   NewClause* _resolvent; 
   const TERM* _currentPosEq;
   bool _noInfBetweenRulesFlag;
  public:
   BACKWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex)
    : bpara_premise(res),
      bpara_machine(res,unifIndex),
      _resolvent(res),
      _noInfBetweenRulesFlag(false)
    {
     bpara_premise.SetIndex(0);
     bpara_machine.SetIndex(1);
    }; 
   ~BACKWARD_SUPERPOSITION() { CALL("destructor ~BACKWARD_SUPERPOSITION()"); }; 
 
   void setSimultaneous(long degree) { bpara_machine.setSimultaneous(degree); };
   void setNoInfBetweenRulesFlag(bool fl)
   {
    _noInfBetweenRulesFlag = fl;     
   };
   
   bool SetQuery(Clause* cl) 
    { 
     bpara_machine.setNoRulesFlag(_noInfBetweenRulesFlag && cl->IsRule());
     bpara_machine.SetQuery(cl);
     bpara_premise.Load(cl);
     _currentPosEq = bpara_premise.NextEq();
     if (!_currentPosEq) return false;
     bpara_machine.LoadEq(_currentPosEq,bpara_premise.GetIndex());
     return true; 
    };
   bool CollectResult()
    {
     CALL("CollectResult()");
     if ((!bpara_premise.TheClause()->Active()) 
         || (!bpara_machine.candClause()->Active()))
       return false;
      
     if (bpara_premise.MinResultWeight()
         + bpara_machine.candClause()->MinResWeight()
         > _resolvent->MaxWeight())
      {
       _resolvent->AnotherTooHeavyClause();
       return false;
      };            
     
     _resolvent->Reset();
     if (bpara_machine.CollectLiterals() && bpara_premise.CollectLiterals())
      {
       _resolvent->UsedRule(ClauseBackground::BackSup);
       _resolvent->PushAncestor(bpara_premise.TheClause());
       _resolvent->PushAncestor(bpara_machine.candClause());
       return true;
      }
     else return false;
    }; 

   bool NextInference()
   {
    process_literal:
     while (bpara_machine.Paramodulate())
      { 
       if (CollectResult()) return true;
      };
     // try next literal
     _currentPosEq = bpara_premise.NextEq();       
     if (!_currentPosEq) return false;
     bpara_machine.LoadEq(_currentPosEq,bpara_premise.GetIndex());
     goto process_literal;
   };


   //void Process(Clause* cl);     
 }; // class BACKWARD_SUPERPOSITION

//================================================
#undef DEBUG_NAMESPACE
//==============================================================
#endif 







 
