#include "binary_resolution.h"

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
 #define DEBUG_NAMESPACE "BIN_ORD_RES"
#endif
#include "debugMacros.h"
//=================================================

bool BIN_ORD_RES::NextInf()
 {
  CALL("NextInf()");
  switch (_continuation)
   {
    case ContFirstLit: goto next_lit;
    case ContFirstFactor: goto first_factor;
    case ContNextFactor: goto next_factor; 
   };
  
  ICP("ICP1");

  next_lit:
   if (currentLit > premise1.LastMax()) { return false; };
  
   ORMachine.CompileLiteral(premise1.MaxLiterals()[currentLit]);
   ORMachine.Inversepolarity();
   premise1.Cut(currentLit);
  
   currentLit++;
  
   GlobalSubst.UnbindAll();
   RES_PAIR_LIST* candidate;
  
   if (ORMachine.FindFirstLeaf())  
    {
     process_leaf: 
      ORMachine.ResetCandidatesFromLeaf();
      try_next_cand:
       candidate  = ORMachine.NextCandidate();
       if (!candidate) goto try_next_leaf;       
       if (_sideClauseCanNotBeRule && candidate->TheClause()->IsRule()) 
        {

	  //DF; cout << "GAME WITH NO RULES\n";
	  //DF; cout << premise1.TheClause() << "\n" << candidate->TheClause() << "\n";

         goto try_next_cand;
        };
       if (candidate->TheClause()->MinResWeight() 
           + FirstPremise()->MinResWeight() 
           > _resolvent->MaxWeight())
        {
         _resolvent->AnotherTooHeavyClause();
         goto try_next_cand;
        };   
            
       
       premise2.LoadOptimized(index2,candidate);

       // first resolvent without factoring
       
       if (!MakeNewResolvent()) 
        {
         goto first_factor;
         //goto try_next_cand;  // bug! caused incompleteness
        };
       _continuation = ContFirstFactor;
       return true;
       
       first_factor:
        ResetFactoringState();
       
        next_factor:
         if (NextFactor())
	  {
	   if (!MakeNewResolvent()) goto next_factor;            
           _continuation = ContNextFactor;
           return true;          
          };
  
         // no_more_factors:
         premise2.Unload();
         goto try_next_cand;
     
         // no more candidates in this leaf, try to find another one
         try_next_leaf:
          if (ORMachine.FindNextLeaf()) { goto process_leaf; };
     };
   
   // no more resolvents for this literal
   premise1.UnmarkCut();
   goto next_lit;
  
 }; // bool BIN_ORD_RES::NextInf()


 
bool BIN_ORD_RES::NextFactor()
 {
  CALL("NextFactor()");
  if (premise2.WasFactored())
   {
    if (premise2.GetNextFactor())
     {
      return true;
     }
    else 
     {
      premise2.ResetFactoring();
      if (premise1.WasFactored())
       {
        return premise1.GetNextFactor();
       }
      else
       {
        return premise1.GetFirstFactor(premise1.CutLit()+1);
       };
     };
   }
  else 
   {
    if (premise2.GetFirstFactor(0))
     {
      return true;
     }
    else
     {
      if (premise1.WasFactored())
       {
        return premise1.GetNextFactor();
       }
      else
       {
        return premise1.GetFirstFactor(premise1.CutLit()+1);
       };
     };
   };
 }; // bool BIN_ORD_RES::NextFactor()
 

bool BIN_ORD_RES::MakeNewResolvent()
 {
  CALL("MakeNewResolvent()");
  if (!CollectResult()) return false;

  // resolvent collected in Resolvent
  _resolvent->UsedRule(ClauseBackground::BinRes);
  _resolvent->PushAncestor((Clause*)FirstPremise());
  _resolvent->PushAncestor((Clause*)SecondPremise());
  
  return true;
  //_resolvent->ProcessCollected(); 
 }; // bool BIN_ORD_RES::MakeNewResolvent() 
//=============================================================
 
 
 
 
 
 
 
 
