#ifndef BINARY_RESOLUTION_H
//======================================================
#define BINARY_RESOLUTION_H
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "clause.h"
#include "premise_factoring.h"
#include "iu_command.h"
#include "or_machine.h"
#include "NewClause.h"
#include "weight_formula.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
 #define DEBUG_NAMESPACE "BIN_ORD_RES"
#endif
#include "debugMacros.h"
//=================================================

class BIN_ORD_RES
 {
  private:
   enum Continuation { ContFirstLit, ContFirstFactor, ContNextFactor }; 
  private:
   OR_PREMISE_WITH_FACTORING premise1,premise2;
   const unsigned long index1,index2;
   unsigned long currentLit;
   OR_MACH ORMachine;
   WEIGHT_FORMULA weight_form;
   NewClause* _resolvent;
   Continuation _continuation;
   bool _noInfBetweenRulesFlag;
   bool _sideClauseCanNotBeRule;
  private: 
   void ResetFactoringState();
   bool NextFactor();
   bool MakeNewResolvent();
   void CollectWeightForm(const Clause* c);
   long MinResWeight() { return weight_form.Compute(); };
   const Clause* FirstPremise() const { return premise1.TheClause(); };
   const Clause* SecondPremise() const { return premise2.TheClause(); };
   bool CollectResult() 
    { 
     CALL("CollectResult()");
     if ((!premise1.TheClause()->Active())
         || (!premise2.TheClause()->Active()))
      return false;
     if (MinResWeight() + premise2.TheClause()->MinResWeight()
         > _resolvent->MaxWeight())
      {
       _resolvent->AnotherTooHeavyClause();
       return false;
      }; 
     _resolvent->Reset();
     return premise1.CollectNonmarkedLiterals() && premise2.CollectNonmarkedLiterals();
    }; 

   bool NextInf();

  public:
   BIN_ORD_RES(NewClause* resolvent,OR_INDEX* resIndex) 
    : premise1(resolvent), 
      premise2(resolvent),
      index1(0),
      index2(1),
      ORMachine(resIndex),
      weight_form(),
      _resolvent(resolvent),
      _noInfBetweenRulesFlag(false)
    {
     weight_form.SetIndex(index1);
    };
   ~BIN_ORD_RES() { CALL("destructor ~BIN_ORD_RES()"); };
   void setNoInfBetweenRulesFlag(bool fl)
   {
    _noInfBetweenRulesFlag = fl;
   };
   void Load(const Clause* c);
   bool NextInference() { return NextInf(); };  

   #ifndef NO_DEBUG 
    bool clean() const { return true; };
   #endif

   //bool NextLiteral();
 }; // class BIN_ORD_RES 
  
inline void BIN_ORD_RES::CollectWeightForm(const Clause* c)
 {
  weight_form.Reset();
  LiteralList* lits = c->LitList();
  for (LiteralList::Iterator iter(lits,lits->nth(lits->numOfSelectedLits()));iter.notEnd();iter.next())
   weight_form.Compile(iter.currentElement()->literal()); 
 }; // inline void CollectWeightForm() 
 
 
inline void BIN_ORD_RES::Load(const Clause* c)
 {
  // reset everithing here
  ORMachine.SetIndex(index2);
  premise1.Load(index1,c);
  CollectWeightForm(c);
  _sideClauseCanNotBeRule = _noInfBetweenRulesFlag && c->IsRule();
  currentLit = 0;
  _continuation = ContFirstLit;
 }; // inline void BIN_ORD_RES::Load(const Clause* c)   

inline void BIN_ORD_RES::ResetFactoringState()
 {
  premise1.ResetFactoring();
  premise2.ResetFactoring();
 }; // void BIN_ORD_RES::ResetFactoringState() 
 

//================================================
#undef DEBUG_NAMESPACE 
//======================================================
#endif
