//
// File:         clause.cpp
// Description:  Representation of stored clauses.
// Created:      Feb 13, 2000, 20:30
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "clause.h"
//====================================================
#ifdef DEBUG_ALLOC_OBJ_TYPE
 ClassDesc Clause::_classDesc("Clause",sizeof(Clause),sizeof(Clause));
 ClassDesc Clause::List::_classDesc("Clause::List",Clause::List::minSize(),Clause::List::maxSize());
#endif
float Clause::_selectedPosEqPenaltyCoefficient;
bool Clause::_useSelectedPosEqPenaltyCoefficient;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
 #define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.h"
//=================================================       


bool Clause::MakeDefinitionOf(const TERM& definedHeader)
{
 CALL("MakeDefinitionOf(const TERM& definedHeader)");
 // true iff there is at least one literal with this header
 Stack<LiteralList::Element*,HellConst::MaxNumOfLiterals> headRef("headRef");
       
 for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())
  if ((*(iter.currentLiteral())) == definedHeader)
   headRef.push(iter.currentElement()); 

 if (headRef.empty()) return false;
 
 LitList()->setNumOfSelected(headRef.size());

 LiteralList::Iterator iter(LitList());
 TERM* tmp; 
 LiteralList::Element* currHeadRef;
 for(LiteralList::Element** defHd = headRef.begin();defHd < headRef.end();defHd++)
  {
   currHeadRef = *defHd;
   ASSERT(iter.notEnd());
   tmp = iter.currentLiteral();
   iter.currentElement()->setLiteral(currHeadRef->literal());
   currHeadRef->setLiteral(tmp);
   iter.next();
  };
 return true;  
}; // bool Clause::MakeDefinitionOf(const TERM& definedHeader)

bool Clause::Definition(const bool* definedHeaderTable,TERM& definedHeader) const
{
 CALL("Definition(const bool* definedHeaderTable) const");
 for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())
  if (definedHeaderTable[iter.currentLiteral()->HeaderNum()]) 
   {
    definedHeader = *(iter.currentLiteral());
    return true;
   };
 return false; 
}; // bool Clause::Definition(const bool* definedHeaderTable,TERM& definedHeader) const

bool Clause::HornClause() const
{
 CALL("HornClause() const");
 if (Unit()) { return true; };
 ulong num_of_pos = 0;
 for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())
  {
    //if (iter.currentLiteral()->polarity() == TERM::PositivePolarity) 
   if (iter.currentLiteral()->Positive()) 
    {
     num_of_pos++;
     if (num_of_pos > 1) { return false; };
    };
  };
 return true;
}; // bool Clause::HornClause() const

bool Clause::SymmetryLaw() const 
{
 if (Length() != 2) return false;
 const TERM* lit1 = LitList()->firstLit();
 const TERM* lit2 = LitList()->nth(1)->literal();
 if (!TERM::Symmetry(lit1,lit2)) return false;
 return lit1->polarity() != lit2->polarity();
}; // bool Clause::SymmetryLaw() const


bool Clause::TotalityLaw() const 
{
 if (Length() != 2) return false;
 const TERM* lit1 = LitList()->firstLit();
 const TERM* lit2 = LitList()->nth(1)->literal();
 if (!TERM::Symmetry(lit1,lit2)) return false;
 return lit1->polarity() == lit2->polarity();
}; // bool Clause::TotalityLaw() const


bool Clause::CommutativityLaw(ulong& func) const
{
 CALL("CommutativityLaw(ulong& func) const");
 if (Unit())
  {
   const TERM* lit = LitList()->firstLit();
   if  (lit->Positive() 
        && lit->IsEquality()
        && lit->Arg1()->IsReference()
        && lit->Arg2()->IsReference()
        && (TERM::Symmetry(lit->Arg1()->First(),lit->Arg2()->First())))
     {
      func = lit->Arg1()->First()->functor();
      return true;
     };
  }; 
 return false;     
}; // bool Clause::CommutativityLaw(ulong& func) const

ostream& Clause::outputInTPTPFormat(ostream& str) const
{
 CALL("outputInTPTPFormat(ostream& str) const");

 str << "input_clause(";
 str << "cl" << Number() << "_anc";

 for (ClauseBackground::ClauseList* cl = Ancestors()->ancestors(); cl; cl = cl->tl())
   str << '_' << cl->hd()->Number();  

 str << "_org_";
 ClauseBackground::Rules rules = Ancestors()->TheRules();
 size_t numOfRules = rules.count();
 if (rules.bit(ClauseBackground::VIP)) numOfRules--; 
 if (rules.bit(ClauseBackground::NegSel)) numOfRules--; 
 if (rules.bit(ClauseBackground::Input))
  { 
   str << "in";
   numOfRules--;
   if (numOfRules) str << '_';   
  };

 if (rules.bit(ClauseBackground::Prepro)) 
  {    
   str << "pp";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::BinRes))
  {
   str << "br";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::Hyper)) 
  {
   str << "hr";
   numOfRules--;
   if (numOfRules) str << '_';  
  };

 if (rules.bit(ClauseBackground::ForwSup)) 
  {
   str << "fs";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::BackSup)) 
  {
   str << "bs";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::EqRes)) 
  {
   str << "er";
   numOfRules--;
   if (numOfRules) str << '_';
  };  

 if (rules.bit(ClauseBackground::EqFact)) 
  {
   str << "ef";
   numOfRules--;
   if (numOfRules) str << '_';
  };     

 if (rules.bit(ClauseBackground::ForwDemod)) 
  {
   str << "fd";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::BackDemod)) 
  {
   str << "bd";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::EqResSimp)) 
  {
   str << "ers";
   numOfRules--;
   if (numOfRules) str << '_';
  };  

 if (rules.bit(ClauseBackground::ForwSubsRes)) 
  {
   str << "fsr";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::BackSubsRes)) 
  {
   str << "bsr";
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::Split)) 
  {
   str << "sp"; 
   numOfRules--;
   if (numOfRules) str << '_';
  };

 if (rules.bit(ClauseBackground::Reanimation)) 
  {
   str << "rea"; 
   numOfRules--;
   if (numOfRules) str << '_';
  }; 

 if (rules.bit(ClauseBackground::Name)) str << "nm";

 str << ',';
 str << "conjecture";
 str << ',';
 str << "\n [";


 for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())  
  { 
   const TERM* lit = iter.currentLiteral();  
   str << "\n  " << (lit->Positive() ? "++" : "--");
   lit->outputInTPTPFormat(str);
   if (!iter.last()) str << ',';
  };
 str << "\n ]).";
 return str;
}; // ostream& Clause::outputInTPTPFormat(ostream& str) const


ostream& Clause::outputAsPrologTerm(ostream& str) const
{
 CALL("outputAsPrologTerm(ostream& str) const");
 str << '[';
 // clause body
 str << '[';

 ulong maxLits = LitList()->numOfSelectedLits();
 for (LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())
  {
   if (maxLits) 
    {
     str << (iter.currentLiteral()->Positive() ? "+++" : "---");
     maxLits--;
    }
   else str << (iter.currentLiteral()->Positive() ? "++" : "--");
   iter.currentLiteral()->outputAsPrologTerm(str);
   if (!iter.last()) str << ','; 
  };

 str << ']';

 // number
 str << ',' << Number();
 
 // ancestors
 str << ",[";

 for (ClauseBackground::ClauseList* cl = Ancestors()->ancestors(); cl; cl = cl->tl())
  {
   str << cl->hd()->Number();
   if (cl->tl()) str << ',';
  };
 str << ']';
 
 // rules
 str << ",[";
 ClauseBackground::Rules rules = Ancestors()->TheRules();
 size_t numOfRules = rules.count();
 if (rules.bit(ClauseBackground::VIP)) numOfRules--; 
 if (rules.bit(ClauseBackground::NegSel)) numOfRules--; 
   if (rules.bit(ClauseBackground::Input))
    { 
     str << "in";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::Prepro)) 
    {
     str << "pp";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::BinRes))
    {
     str << "br";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::Hyper)) 
    {
     str << "hr";
     numOfRules--;
     if (numOfRules) str << ',';  
    };

   if (rules.bit(ClauseBackground::ForwSup)) 
    {
     str << "fs";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::BackSup)) 
    {
     str << "bs";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::EqRes)) 
    {
     str << "er";
     numOfRules--;
     if (numOfRules) str << ',';
    };  

   if (rules.bit(ClauseBackground::EqFact)) 
    {
     str << "ef";
     numOfRules--;
     if (numOfRules) str << ',';
    };     

   if (rules.bit(ClauseBackground::ForwDemod)) 
    {
     str << "fd";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::BackDemod)) 
    {
     str << "bd";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::EqResSimp)) 
    {
     str << "ers";
     numOfRules--;
     if (numOfRules) str << ',';
    };  

   if (rules.bit(ClauseBackground::ForwSubsRes)) 
    {
     str << "fsr";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::BackSubsRes)) 
    {
     str << "bsr";
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::Split)) 
    {
     str << "sp"; 
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::Reanimation)) 
    {
     str << "rea"; 
     numOfRules--;
     if (numOfRules) str << ',';
    };

   if (rules.bit(ClauseBackground::Name)) str << "nm";

 str << ']'; 


  return str << ']';
}; // ostream& Clause::outputAsPrologTerm(ostream& str) const


//=================================================  
