#ifndef CLAUSE_INDEX_H
//===================================================================
#define CLAUSE_INDEX_H
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"
#include "Flatterm.h"
#include "clause.h"
#include "dtree.h"
#include "sharing.h"
extern SHARING_INDEX GlobalSharingIndex;
//==================================================
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_CLAUSE_INDEX
 #define DEBUG_NAMESPACE "LITERAL_INDEX"
#endif
#include "debugMacros.h"
//=================================================

class LITERAL_INDEX
 {
  private:
   SHARING_INDEX* _sharingIndex;
   TERM prop_pos[2*HellConst::MaxNumOfFunctors];
   TERM prop_neg[2*HellConst::MaxNumOfFunctors];
   DTREE_NODE* positive[HellConst::MaxNumOfFunctors];
   DTREE_NODE* negative[HellConst::MaxNumOfFunctors];
  public:
   LITERAL_INDEX(SHARING_INDEX* si) : _sharingIndex(si)
    {
     for (unsigned long pred = 0; pred < HellConst::MaxNumOfFunctors; pred++)
      {
       positive[pred] = (DTREE_NODE*)0;
       negative[pred] = (DTREE_NODE*)0;
      };

     TERM prop_lit;
     for (unsigned long prop_var = 0; prop_var < HellConst::MaxNumOfFunctors; prop_var++)
      {
       prop_lit.MakeComplex((TERM::Arity)0,(TERM::Functor)prop_var,TERM::TERM::PositivePolarity);
       *(Share(prop_lit)) = prop_lit;
       *(Share(prop_lit)->InfoField()) = (void*)0;
       
       prop_lit.Inversepolarity();
       *(Share(prop_lit)) = prop_lit;
       *(Share(prop_lit)->InfoField()) = (void*)0;
      };  
    }; 

   ~LITERAL_INDEX() { CALL("destructor ~LITERAL_INDEX()"); };
   void Destroy(); 
   
   #ifndef NO_DEBUG   
    bool clean() const 
    {
     for (unsigned long pred = 0; pred < HellConst::MaxNumOfFunctors; pred++)
      if (positive[pred] || negative[pred]) return false;
     TERM prop_lit;
     for (unsigned long prop_var = 0; prop_var < HellConst::MaxNumOfFunctors; prop_var++)
       {
        prop_lit.MakeComplex((TERM::Arity)0,(TERM::Functor)prop_var,TERM::TERM::PositivePolarity);
        if (*(Share(prop_lit)->InfoField())) return false;        

        prop_lit.Inversepolarity();
        if (*(Share(prop_lit)->InfoField())) return false;
       };  
     return true; 
    };
   #endif
 
  private:
   static void DestroyLeafContent(DTREE_NODE* leaf);
   inline TERM* ShareSplittingBranchName(const TERM& prop)
    {
     CALL("ShareSplittingBranchName(const TERM& prop)");
     ASSERT(!(prop.arity()));
     ASSERT(prop.IsSplitting());
     //TERM* res = (((prop.polarity() == TERM::PositivePolarity) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1);
     TERM* res = (((prop.Positive()) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1); 
     res->MakeSplitting(); // ugly
     return res;
    };

   inline const TERM* ShareSplittingBranchName(const TERM& prop) const
    {
     CALL("ShareSplittingBranchName(const TERM& prop) const");
     ASSERT(!(prop.arity()));
     ASSERT(prop.IsSplitting());
     //const TERM* res = (((prop.polarity() == TERM::PositivePolarity) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1);
     const TERM* res = (((prop.Positive()) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1);
     ((TERM*)res)->MakeSplitting(); // ugly
     return res;
    };


   inline TERM* Share(const TERM& prop)
    {
     CALL("Share(const TERM& prop)");
     ASSERT(!(prop.arity()));
     if (prop.IsSplitting()) return ShareSplittingBranchName(prop);
     //return (((prop.polarity() == TERM::PositivePolarity) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1);
     return (((prop.Positive()) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1); 
    };
   inline const TERM* Share(const TERM& prop) const
    {
     CALL("Share(const TERM& prop) const");
     ASSERT(!(prop.arity()));
     if (prop.IsSplitting()) return ShareSplittingBranchName(prop);
     //return (((prop.polarity() == TERM::PositivePolarity) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1);
     return (((prop.Positive()) ? prop_pos : prop_neg) + 2*(prop.functor()) + 1);
    };
  public:   
   inline DTREE_NODE** Tree(const TERM& header)
    {
      //if (header.polarity() == TERM::PositivePolarity) { return &(positive[header.functor()]); }
     if (header.Positive()) { return &(positive[header.functor()]); }
     else { return &(negative[header.functor()]); };
    };
   inline DTREE_NODE* Integrate(TERM* lit); // returns the leaf
  
   inline DTREE_NODE* Integrate(const Flatterm* lit); // returns the leaf 
   inline DTREE_NODE* IntegrateEq(const TERM& header,
                                  const Flatterm* arg1,
                                  const Flatterm* arg2); // returns the leaf  
   
   inline DTREE_NODE* FindLeaf(TERM* lit) // returns the leaf
    {
     // temporary version
     return Integrate(lit); 
    };
   inline void InfoForDeletion(TERM* lit,
                               DTREE_NODE**& branch_to_destroy,
                               DTREE_NODE*& leaf,
                               DTREE_NODE*& last_fork); // for nonpropositional literals
 }; // class LITERAL_INDEX
//==================================================
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_CLAUSE_INDEX
 #define DEBUG_NAMESPACE "CLAUSE_INDEX"
#endif
#include "debugMacros.h"
//=================================================
class CLAUSE_INDEX 
 {
  private:
   SHARING_INDEX* _sharingIndex;
   LITERAL_INDEX lit_index;
  private:
   inline void Remove(TERM* lit,Clause* cl,bool remove_from_sharing);
  public:
   CLAUSE_INDEX(SHARING_INDEX* si) : _sharingIndex(si), lit_index(si) {};
   ~CLAUSE_INDEX() { CALL("destructor ~CLAUSE_INDEX()"); };
   inline void Integrate(Clause* cl);
   inline void Remove(Clause* cl,bool remove_from_sharing);
   inline DTREE_NODE** Tree(const TERM& header) { return lit_index.Tree(header); };
  
   inline TERM* Integrate(const Flatterm* lit,Clause* cl)
    {
     CALL("Integrate(const Flatterm* lit,Clause* cl)");
     #ifdef _SUPPRESS_WARNINGS_
      cl = cl;
     #endif
    
     DTREE_NODE* leaf = lit_index.Integrate(lit);
     leaf->NumOfClauses()++;
     return leaf->Term();
    };
   inline TERM* IntegrateEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2,Clause* cl)
    {
     #ifdef _SUPPRESS_WARNINGS_
      cl = cl;
     #endif
    
     DTREE_NODE* leaf = lit_index.IntegrateEq(header,arg1,arg2);
     leaf->NumOfClauses()++;
     return leaf->Term();
    }; 
    
   inline void Destroy() { lit_index.Destroy(); };
   #ifndef NO_DEBUG
    bool clean() const { return lit_index.clean(); };
   #endif
 }; // class CLAUSE_INDEX

//******************* Definitions for CLAUSE_INDEX: ***********************

inline void CLAUSE_INDEX::Remove(Clause* cl,bool remove_from_sharing)
 {  
  for (LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
   Remove(iter.currentLiteral(),cl,remove_from_sharing); 
 }; // void CLAUSE_INDEX::Remove(Clause* cl,bool remove_from_sharing)

inline void CLAUSE_INDEX::Remove(TERM* lit,Clause* cl,bool remove_from_sharing)
 {
  CALL("Remove(TERM* lit,Clause* cl,bool remove_from_sharing)");
  #ifdef _SUPPRESS_WARNINGS_
    cl = cl;
  #endif
  if (lit->arity())
   {
    DTREE_NODE** branch_to_destroy;
    DTREE_NODE* leaf;
    DTREE_NODE* last_fork;
    lit_index.InfoForDeletion(lit,branch_to_destroy,leaf,last_fork);

    ASSERT(leaf->IsLeaf());
    
    if (remove_from_sharing) 
     {
      leaf->NumOfClauses()--;
    
      if (!(leaf->NumOfClauses())) // no more clauses with this literal
       {
        TERM* lit_copy = leaf->Term();

        DTREE::DestroyBranchWithoutForks(*branch_to_destroy);
        if (last_fork)
         {
          DTREE::ReduceFork(last_fork);
         }
        else
         {
          *branch_to_destroy = (DTREE_NODE*)0;
         }; 

     
        _sharingIndex->DestroyLiteralCopy(lit_copy);
        
       }; 
     };
   }
  else // propositional
   {
    if (remove_from_sharing) 
     {
      ((*(lit_index.Tree(*lit)))->NumOfClauses())--;
      if (!((*(lit_index.Tree(*lit)))->NumOfClauses())) // no more clauses with this literal
       {
        delete *(lit_index.Tree(*lit));
      
        *(lit_index.Tree(*lit)) = (DTREE_NODE*)0;
       };
     };
   };
 }; // void CLAUSE_INDEX::Remove(TERM* lit,Clause* cl,bool remove_from_sharing)

inline void CLAUSE_INDEX::Integrate(Clause* cl)
 {
  CALL("Integrate(Clause* cl)");
  DTREE_NODE* leaf;
  for (LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
   {
    leaf = lit_index.Integrate(iter.currentLiteral());   
    ASSERT(leaf->IsLeaf());
    ASSERT(TERM::EqualComplexTerms(iter.currentLiteral(),leaf->Term()));
    
    iter.currentElement()->setLiteral(leaf->Term());     

    leaf->NumOfClauses()++;
   };   

 }; // void CLAUSE_INDEX::Integrate(Clause* cl)

//==================================================
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_CLAUSE_INDEX
 #define DEBUG_NAMESPACE "LITERAL_INDEX"
#endif
#include "debugMacros.h"
//=================================================

inline DTREE_NODE* LITERAL_INDEX::Integrate(TERM* lit) 
 // returns the leaf corresponding to the literal
 {
  CALL("Integrate(TERM* lit)");
  TERM header = *lit;
  DTREE_NODE** tree = Tree(header);
  
  if (header.arity())
   {
    TERM::Iterator iter;
    iter.Reset(lit);
    DTREE::SURFER surf;
    surf.Reset(tree,1);
   
    if (*tree)
     {
      read_mode_next_sym:
       if (iter.Next())
        {
         if (surf.Read(iter.Symbol()))
          {
           goto read_mode_next_sym;
          }
         else
          {
           surf.Write(iter.Symbol());
           goto write_mode_next_sym;
          }; 
        }
       else // the literal is in the index already
        {
         return surf.Leaf();
        };
     }
    else
     {
      write_mode_next_sym:
       if (iter.Next())
        {
         surf.Write(iter.Symbol());
         goto write_mode_next_sym;
        }
       else // end of the literal, leaf must be created
        {
         surf.CreateLeaf(_sharingIndex->CopyLiteral(lit));
         
         *(surf.Leaf()->Term()->InfoField()) = (void*)0;
         
         return surf.Leaf();
        };
     }; 
   }
  else // propositional
   {
    if (!(*tree)) 
     {
      *tree = new DTREE_NODE(Share(header));
     };
    
    ASSERT((*tree)->IsLeaf());
    
    return *tree;
   };
 }; // DTREE_NODE* LITERAL_INDEX::Integrate(TERM* lit)

inline DTREE_NODE* LITERAL_INDEX::Integrate(const Flatterm* lit) 
 // returns the leaf corresponding to the literal
 {
  CALL("Integrate(const Flatterm* lit)");
  TERM header = lit->Symbol();
  DTREE_NODE** tree = Tree(header);
  
  if (header.arity())
   {
    Flatterm::Iterator iter;
    iter.Reset(lit);
    DTREE::SURFER surf;
    surf.Reset(tree,1);
    if (*tree)
     {
      read_mode_next_sym:
       iter.NextSym();
       if (iter.EndOfTerm()) return surf.Leaf(); // the literal is in the index already
       if (surf.Read(iter.CurrPos()->Symbol())) { goto read_mode_next_sym; };
       surf.Write(iter.CurrPos()->Symbol()); 
       goto write_mode_next_sym;
     }
    else
     {
      write_mode_next_sym:
       iter.NextSym();
       if (iter.EndOfTerm()) 
        {
         surf.CreateLeaf(_sharingIndex->CopyFlatLiteral(lit));
         
         *(surf.Leaf()->Term()->InfoField()) = (void*)0;
         
         return surf.Leaf();
        };       
       surf.Write(iter.CurrPos()->Symbol());
       goto write_mode_next_sym;
     }; 
   }
  else // propositional
   {
    if (!(*tree)) 
     {
      *tree = new DTREE_NODE(Share(header));
     };
    

    ASSERT((*tree)->IsLeaf());
    
    return *tree;
   };
 }; // DTREE_NODE* LITERAL_INDEX::Integrate(const Flatterm* lit) 

inline DTREE_NODE* LITERAL_INDEX::IntegrateEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2) 
 // returns the leaf corresponding to the literal
 {
  CALL("IntegrateEq(const TERM&,const Flatterm*,const Flatterm*)");
  DTREE_NODE** tree = Tree(header);
  Flatterm::Iterator iter;
  iter.Reset(arg1);
  bool first_arg = true;
  DTREE::SURFER surf;
  surf.Reset(tree,1);
  if (*tree)
   {
    read_mode_next_sym:
     if (iter.EndOfTerm()) 
      {
       if (first_arg)
        {// the second argument must be processed
         iter.Reset(arg2);
         first_arg = false;
         goto read_mode_next_sym;
        };
       return surf.Leaf(); // the literal is in the index already
      };
     if (surf.Read(iter.CurrPos()->Symbol())) 
      {
       iter.NextSym();
       goto read_mode_next_sym;
      };
     surf.Write(iter.CurrPos()->Symbol()); 
     iter.NextSym();
     goto write_mode_next_sym;
   }
  else
   {
    write_mode_next_sym:
     if (iter.EndOfTerm()) 
      {
       if (first_arg)
        {// the second argument must be processed
         iter.Reset(arg2);
         first_arg = false;
         goto write_mode_next_sym;
        };
       surf.CreateLeaf(_sharingIndex->CopyFlatEq(header,arg1,arg2));
       *(surf.Leaf()->Term()->InfoField()) = (void*)0;
       return surf.Leaf();
      };
       
     surf.Write(iter.CurrPos()->Symbol());
     iter.NextSym();
     goto write_mode_next_sym;
   }; 
 }; // DTREE_NODE* LITERAL_INDEX::IntegrateEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2) 

inline void LITERAL_INDEX::InfoForDeletion(TERM* lit,
                                           DTREE_NODE**& branch_to_destroy,
                                           DTREE_NODE*& leaf,
                                           DTREE_NODE*& last_fork)
 {
  CALL("InfoForDeletion(TERM*,DTREE_NODE**&,DTREE_NODE*&,DTREE_NODE*&)");
  // lit can't be propositional here
  ASSERT(lit->arity());
  
  DTREE_NODE** tree = Tree(*lit); // *tree must be nonnull and nonleaf here
    
  ASSERT(*tree);
  ASSERT(!((*tree)->IsLeaf()));
  
  TERM::Iterator iter;
  iter.Reset(lit);
  iter.Next(); // skipping the header
  
  DTREE::InfoForDeletion(iter,tree,branch_to_destroy,leaf,last_fork);
 }; // void LITERAL_INDEX::InfoForDeletion(TERM* lit,...                         
//================================================
#undef DEBUG_NAMESPACE   
//===================================================================
#endif                     























