#ifndef FORWARD_SUBSUMPTION_H
//===========================================================
#define FORWARD_SUBSUMPTION_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "fs_code.h"
#include "fs_command.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION
 #define DEBUG_NAMESPACE "FS_QUERY"
#endif 
#include "debugMacros.h"
//=================================================
class FS_QUERY
 {
  public:
   class LIT
    {    
     private:
      const Flatterm* arg1;
      const Flatterm* arg2;
      long captured;
      LIT* next;
     public:  
      inline const Flatterm* Args() const { return arg1; };
      inline void SetArgs(const Flatterm* a) { arg1 = a; };      
      inline const Flatterm* Arg1() const { return arg1; }; // for equations
      inline const Flatterm* Arg2() const { return arg2; }; // for equations
      inline void SetArgs(const Flatterm* a1,const Flatterm* a2)
       { arg1 = a1; arg2 = a2; };
      inline LIT* Next() const { return next; };
      inline void SetNext(LIT* nl) { next = nl; };
      inline void Set(const Flatterm* a,LIT* nl) { arg1 = a; captured = 0L; next = nl; };
      inline void Set(const Flatterm* a1,const Flatterm* a2,LIT* nl) 
       { arg1 = a1; arg2 = a2; captured = 0L; next = nl; };
      inline void Capture() { captured++; };
      inline void Release() { captured--; };
      inline long Captured() { return captured; };
     public:
      #ifndef NO_DEBUG_VIS
       ostream& output(ostream& str) const { return (Args() - 1)->outputTerm(str); };
      #endif
    }; // class LIT
  private: // structure 
   LIT literal[HellConst::MaxNumOfLiterals];
   LIT* next_literal;
   LIT* header_index[HellConst::MaxNumOfDifferentHeaders];
   LIT** const pos_ord_equations;
   LIT** const neg_ord_equations;
   LIT** present_header[HellConst::MaxNumOfLiterals];
   LIT*** next_present_header; 
   LIT* last_pos_unord_eq;
   LIT* last_neg_unord_eq;
   long num_of_pos_eq;
   bool collect_signature;
   FunSet* signature;
  public:
   FS_QUERY() 
    : pos_ord_equations(header_index + TERM::positiveOrdEqHeader().HeaderNum()),
      neg_ord_equations(header_index + TERM::negativeOrdEqHeader().HeaderNum()),
      collect_signature(true) // default
    {
     CALL("constructor FS_QUERY()");
     for (unsigned long h = 0; h < HellConst::MaxNumOfDifferentHeaders; h++)
      {
       header_index[h] = (LIT*)0;
      };
     next_present_header = present_header;
     signature = (FunSet*)0; 
     //signature = new FunSet(); 
    };
   ~FS_QUERY() { CALL("destructor ~FS_QUERY()"); };
   inline void ResetQuery()
    {
     CALL("ResetQuery()");
     next_literal = literal;
     while (next_present_header != present_header)
      {
       next_present_header--;
       **next_present_header = (LIT*)0;
      };
     last_pos_unord_eq = (LIT*)0;
     last_neg_unord_eq = (LIT*)0;
     num_of_pos_eq = 0;
     if (collect_signature) 
      {
       if (signature) delete signature;
       signature = new FunSet();
      };
    };
   inline void SetCollectSignature(bool fl) 
    { 
     collect_signature = fl; 
    };
   inline void LinkEquations()
    {
     if (last_pos_unord_eq) last_pos_unord_eq->SetNext(*pos_ord_equations);
     if (last_neg_unord_eq) last_neg_unord_eq->SetNext(*neg_ord_equations);
    }; 
   inline void EndOfQuery()
    {
     LinkEquations(); 
    }; 
   inline void PushLit(const Flatterm* lit)
    {
     CALL("PushLit(const Flatterm* lit)");
     ASSERT(!lit->Symbol().IsEquality());
     LIT** lit_lst_ptr = header_index + lit->Symbol().HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(lit + 1,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(lit + 1,(LIT*)0);
      };
     *lit_lst_ptr = next_literal; 
     next_literal++;  
     if (collect_signature) CollectFunctors(lit);
    };   
    
   inline void PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
    {
     CALL("PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
     ASSERT(header.IsUnorderedEq());
     if (header.Positive()) num_of_pos_eq++;
     LIT** lit_lst_ptr = header_index + header.HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(arg1,arg2,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(arg1,arg2,(LIT*)0);
       if (header.Negative()) { last_neg_unord_eq = next_literal; }
       else // positive
        last_pos_unord_eq = next_literal;
      };
     *lit_lst_ptr = next_literal; 
     next_literal++; 
     if (collect_signature) 
      {
       AddFunctor(header.HeaderNum());
       CollectFunctors(arg1); 
       CollectFunctors(arg2);
      };
    };
    
   inline void PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
    {
     CALL("PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
     ASSERT(header.IsEquality() && header.IsOrderedEq());  
     if (header.Positive()) num_of_pos_eq++;
     LIT** lit_lst_ptr = header_index + header.HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(arg1,arg2,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(arg1,arg2,(LIT*)0);
      };
     *lit_lst_ptr = next_literal; 
     next_literal++; 
     if (collect_signature) 
      {
       header.MakeUnordered();
       AddFunctor(header.HeaderNum());
       CollectFunctors(arg1); 
       CollectFunctors(arg2);
      };
    }; 
    
   inline LIT* FirstLit(unsigned long header_num)  
    {
     return header_index[header_num];
    }; 
   
   inline bool SubsumptionAllowedInSetMode(Clause* subsuming_cl)
    {
      // This doesn't guarantee completeness
     return subsuming_cl->NumOfPosEq() <= num_of_pos_eq;
    }; 

    inline bool SubsumptionAllowedInMultisetMode(Clause* subsuming_cl)
    {
     return true;
    };  
    inline bool SubsumptionAllowedInOptimisedMultisetMode(Clause* subsuming_cl,long numOfFactorings)
    {
      return (!numOfFactorings) 
             || (subsuming_cl->allMaximalSelected());
    };  
   inline const FunSet* Signature() const { return signature; }; 

  private:
   void AddFunctor(unsigned long fun_num)
    {
     if (!(signature->write((size_t)fun_num)))
      {
       FunSet* new_sig = signature->add((size_t)fun_num);
       delete signature;
       signature = new_sig; 
      };
    };
   void CollectFunctors(const Flatterm* t)
    {
     CALL("CollectFunctors(const Flatterm* t)");
     ASSERT(collect_signature);
     Flatterm::Iterator iter(t);
     TERM sym;
     while (iter.NotEndOfTerm())
      {
       sym = iter.CurrPos()->symbolRef();
       if (sym.isComplex()) AddFunctor(sym.HeaderNum());       
       iter.NextSym();
      };
    };
    
  public:
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
     for (const LIT* const * const * p = present_header; p < next_present_header; p++)
      for (const LIT* lit = **p; lit; lit = lit->Next())
       lit->output(str) << '\n';
     return str;
    };
   #endif
 }; // class FS_QUERY
 
#ifdef FS_STAT
class FS_STATISTICS
{
 public:
  FS_STATISTICS() 
   : totInstr(0UL),
     compInstr(0UL) 
  {   
   CALL("constructor FS_STATISTICS()");
  };
  void anotherCall() { calls++; };
  void anotherInstr() { totInstr++; };
  void anotherCompInstr() { compInstr++; };
  void anotherCompSucc() { compSucc++; };
  unsigned long statCalls() const { return calls; };
  unsigned long statTotInstr() const { return totInstr; };
  unsigned long statNonCompInstr() const { return totInstr - compInstr; };
  unsigned long statCompInstr() const { return compInstr; };
  unsigned long statCompSucc() const { return compSucc; };
  unsigned long statCompFail() const { return compInstr - compSucc; };  
 private: 
  unsigned long calls;
  unsigned long totInstr;
  unsigned long compInstr;
  unsigned long compSucc;
};
#endif

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION
 #define DEBUG_NAMESPACE "FS"
#endif 
#include "debugMacros.h"
//=================================================
class FS : public FS_CODE, public FS_QUERY 

#ifdef FS_STAT
 , public FS_STATISTICS
#endif

 {
  public:
   FS() 
    : FS_CODE(subst), 
      FS_QUERY(),
      #ifdef FS_STAT
       FS_STATISTICS(),
      #endif
      setMode(false) // default
    {
     CALL("constructor FS()");
    };
   void Destroy() { FS_CODE::Destroy(); };
   #ifndef NO_DEBUG
    bool clean() const { return FS_CODE::clean(); };
   #endif
   Clause* Subsume() 
   {
    CALL("Subsume()");    
    if (setMode) return SubsumeInSetMode();
    if (optimisedMultisetMode) return SubsumeInOptimisedMultisetMode();
    return SubsumeInMultisetMode();
   };
   Clause* SubsumeInMultisetMode();
   Clause* SubsumeInOptimisedMultisetMode();
   Clause* SubsumeInSetMode();
   void SetUseSigFilters(bool fl) 
    { 
     FS_QUERY::SetCollectSignature(fl);
     FS_CODE::SetUseSigFilters(fl);
    };
   void SetSetMode(bool fl) 
   { 
    setMode = fl; 
   };
   void SetOptimisedMultisetMode(bool fl)
   {
    optimisedMultisetMode = fl;
   }; 
   #ifndef NO_DEBUG_VIS 
    ostream& outputQuery(ostream& str) { return FS_QUERY::output(str); };
   #endif
  private:
   const Flatterm* subst[HellConst::MaxClauseSize];
   bool setMode;
   bool optimisedMultisetMode;
 }; // class FS  : public FS_CODE 
//================================================
#undef DEBUG_NAMESPACE
//================================================
#endif
