//
// File:         fs_code.cpp
// Description:  Maintainence of index for FS.
// Created:      Oct 15, 1999, 15:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "fs_code.h"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE"
#endif
#include "debugMacros.h"
//===================================================
void FS_CODE::Remove(Clause* cl) 
 {
  CALL("Remove(Clause* cl)"); 
  
  //DF;  cout << "REMOVE " << cl << "\n"; DF;

  if (oldStyle) 
   {     
    #ifdef DEBUG_NAMESPACE
     bool removed = oldStyleCode.remove(cl); 
    #else 
     oldStyleCode.remove(cl); 
    #endif      

    ASSERT(removed);   
    return; 
   };

  if (!tree) return;
  removal.Reset(&tree,cl);
   
  LiteralList::Iterator iter(cl->LitList());
  TERM* lit;
  next_lit:
   if (iter.end()) 
    {
     // end of clause
     if (removal.FindSuccess())
      {
       removal.DestroyBranch();
      }
     else
      { 
       COP("COP1");
       DMSG("Removed clause is not found in the FS index:\n"); 
       DMSG(cl->Number());
       DMSG("\n");
       ERROR(""); 
      };
     return;
    };
   lit = iter.currentLiteral();
   iter.next();
   if (removal.SkipLiteral(lit)) goto next_lit;

   COP("COP2");
   DMSG("Removed clause is not found in the FS index:\n"); 
   DMSG(cl->Number());
   DMSG("\n");
   ERROR(""); 
 }; // void FS_CODE::Remove(Clause* cl) 


void FS_CODE::Integrate(Clause* cl) 
 {
  CALL("Integrate(Clause* cl)");

  //DF; cout << "INTEGRATE " << cl << "\n"; DF;

  if (oldStyle) { oldStyleCode.integrate(cl); return; };  

  integrator.Reset(&tree,cl);
  LiteralList::Iterator iter(cl->LitList());
  TERM* lit;
  if (!tree) goto write_mode_next_lit;
  
  read_mode_next_lit:
   if (iter.end()) 
    {
     // end of clause
     integrator.InsertSuccess(cl);
     return;
    }; 
   lit = iter.currentLiteral();
   iter.next();
   if (integrator.ReadLiteral(lit)) 
    {
     goto read_mode_next_lit;
    }
   else
    {
     ASSERT(!integrator.CurrInstr());
     goto write_mode_next_lit;
    }; 
    
  write_mode_next_lit:
   if (iter.end()) 
    {
     // end of clause
     integrator.WriteSuccess(cl);

     return;
    }; 
   lit = iter.currentLiteral();
   iter.next();
   integrator.WriteLiteral(lit);
   goto write_mode_next_lit;
 }; // void FS_CODE::Integrate(Clause* cl)  

//====================================================
#include "fs_code.h"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE_INTEGRATOR"
#endif
#include "debugMacros.h"
//===================================================

void FS_CODE_INTEGRATOR::WriteEqLiteral(TERM* lit)
 {
  CALL("WriteEqLiteral(TERM* lit)");
  ASSERT(lit->IsEquality());
  TERM* args = lit->Args();
  TERM arg1 = *args;
  TERM arg2 = *(args->Next());
  
  if (lit->IsOrderedEq())
   {
    WriteOrdEq(*lit); 
   }
  else
   {
    WriteUnordEq(*lit); 
   }; 
  
  WriteTerm(arg1);
  WriteSecondEqArg();
  WriteTerm(arg2);
 }; // void FS_CODE_INTEGRATOR::WriteEqLiteral(TERM* lit)

void FS_CODE_INTEGRATOR::WriteLiteral(TERM* lit)
 { 
  
  if (lit->IsEquality()) { WriteEqLiteral(lit); return; };
  TERM::Iterator iter;
  TERM sym;

  if (lit->arity())
   {
    WriteHeader(*lit);
    
    iter.Reset(lit);
    write_mode_next_sym:
     if (!(iter.Next())) return;
      
     sym = iter.Symbol();
     if (iter.CurrentSymIsVar())
      {
       WriteVar(sym.var());
      }
     else // functor
      {
       WriteFunc(sym);
      }; 
     goto write_mode_next_sym; 
   }
  else // propositional
   {
    WritePropLit(*lit);
    return;
   }; 
 }; // void FS_CODE_INTEGRATOR::WriteLiteral(TERM* lit)

bool FS_CODE_INTEGRATOR::ReadVar(const ulong& var)
 {
  CALL("ReadVar(ulong var)");
  if (!(ReadVar())) 
   { 
    InsertVar(fs_var);
    if (vars.Occurences(var))
     {
      COP("was before");
      EnterWriteMode(); 
      WriteCompare(first_occ[var],fs_var);
     }
    else // first occurence
     {
      COP("first occ");
      vars.Register(var);
      first_occ[var] = fs_var;
      EnterWriteMode(); 
     }; 
    fs_var++;
    return false; 
   };
          
  if (vars.Occurences(var))
   {
    ReadCompare(first_occ[var],fs_var);
   } 
  else // first occurence of the variable
   {
    vars.Register(var);
    first_occ[var] = fs_var;
   }; 
  fs_var++;
  return true;
 }; // bool FS_CODE_INTEGRATOR::ReadVar(ulong var)

bool FS_CODE_INTEGRATOR::ReadComplexTerm(TERM* t)
 {
  CALL("ReadComplexTerm(TERM* t)");
  TERM::Iterator iter;
  TERM sym;
  iter.Reset(t);
  if (ReadFunc(*t)) 
   {
    read_mode_next_sym:
     if (!iter.Next()) return true; 
     sym = iter.Symbol();
     if (iter.CurrentSymIsVar())
      {
       if (!(ReadVar(sym.var()))) goto write_mode_next_sym; 
      }
     else // functor
      {
       if (!(ReadFunc(sym))) 
        { 
         InsertFunc(sym);
         COP("func inserted 1");
         EnterWriteMode(); 
         goto write_mode_next_sym; 
        };
      }; 
     goto read_mode_next_sym;
   };
  InsertFunc(*t);
  COP("func inserted 2");
  EnterWriteMode();  
  
  write_mode_next_sym:
   if (!iter.Next()) return false;
   sym = iter.Symbol();
   if (iter.CurrentSymIsVar())
    {
     WriteVar(sym.var());
    }
   else // functor
    {
     WriteFunc(sym);
    }; 
   goto write_mode_next_sym; 
 }; // bool FS_CODE_INTEGRATOR::ReadComplexTerm(TERM* t)

bool FS_CODE_INTEGRATOR::ReadEqLiteral(TERM* lit)
 {
  CALL("ReadEqLiteral(TERM* lit)");
  ASSERT(lit->IsEquality());
  TERM* args = lit->Args();
  TERM arg1 = *args;
  TERM arg2 = *(args->Next());
  
  if (lit->IsOrderedEq())
   {
    if (!ReadOrdEq(*lit)) { WriteOrdEq(*lit); goto write_mode_first_arg; };
   }
  else
   {
    if (!ReadUnordEq(*lit)) { WriteUnordEq(*lit); goto write_mode_first_arg; };
   }; 
  
  // read mode, first argument 
  if (!ReadTerm(arg1)) goto write_mode_second_arg; 
  
  // read mode, second argument
  ReadSecondEqArg(); 
  return (ReadTerm(arg2)) && (ReadCollectedComps()); 
  
  write_mode_first_arg:
   WriteTerm(arg1);
  write_mode_second_arg:
   WriteSecondEqArg();
   WriteTerm(arg2);
   return false;
 }; // bool FS_CODE_INTEGRATOR::ReadEqLiteral(TERM* lit) 

bool FS_CODE_INTEGRATOR::ReadLiteral(TERM* lit)
 {
  CALL("ReadLiteral(TERM* lit)"); 
  ASSERT((!debugLitNum) || constr_integrated.Implies(constr_old));
  ASSERT((!debugLitNum) || constr_old.Implies(constr_passed));
  ASSERT((!debugLitNum) || constr_passed.Implies(constr_integrated));
  DOP(debugLitNum++);

  if (lit->IsEquality()) return ReadEqLiteral(lit);
  
  TERM::Iterator iter;
  TERM sym;
  
  if (lit->arity())
   {
    if (ReadHeader(*lit))
     {      
      iter.Reset(lit);
      read_mode_next_sym:
       if (!(iter.Next())) return ReadCollectedComps();
       sym = iter.Symbol();
       if (iter.CurrentSymIsVar())
        {
         if (!(ReadVar(sym.var()))) goto write_mode_next_sym; 
        }
       else // functor
        {
         if (!(ReadFunc(sym))) 
          { 
           InsertFunc(sym);
           EnterWriteMode(); 
           ASSERT_IN(!CurrInstr(),"A30");
           goto write_mode_next_sym; 
          };
        }; 
       goto read_mode_next_sym; 
     } 
    else // the header was not read
     {
      ASSERT_IN(!CurrInstr(),"A40");
      WriteHeader(*lit);
      goto write_mode_lit_args;
     };  
   }
  else // propositional literal
   {
    if (ReadPropLit(*lit)) return true;
    // the literal was not read
    ASSERT_IN(!CurrInstr(),"A50");
    WritePropLit(*lit);
    return false;
   }; 
  
  // write mode: 
  write_mode_lit_args: 
   iter.Reset(lit);
   write_mode_next_sym:       
    if (!(iter.Next())) return false;
       
    sym = iter.Symbol();
    if (iter.CurrentSymIsVar())
     {
      WriteVar(sym.var());
     }
    else // functor
     {
      WriteFunc(sym);
     }; 
    goto write_mode_next_sym; 
   
 }; // bool FS_CODE_INTEGRATOR::ReadLiteral(TERM* lit)

bool FS_CODE_INTEGRATOR::ReadHeader(const TERM& header)
 {
  CALL("ReadHeader(const TERM& header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  ASSERT(!header.IsEquality()); 

  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();

  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::LIT_HEADER :
      if (curr_instr->HeaderNum() == header_num) 
       {
        ShiftSigFilterRight();
        MoveDown(); 
        UnregFunToInsert(header_num);
        UnregFunToShift(header_num);
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::PROP_LIT :  // no break here! go to the next case 
     case FS_COMMAND::UNORD_EQ :  // no break here! go to the next case 
     case FS_COMMAND::ORD_EQ :  
      
      // mismatch occured, try fork
      ShiftSigFilterDown();
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      UnregFunToInsert(header_num); 
      return false;

     case FS_COMMAND::SIG_FILTER :
      PassSigFilter(); 
      goto check_tag;      
   
     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      DWARN("Clause being integrated into FS index is subsumed by already integrated");
      DMSG("Clause\n ");
      DMSG(new_clause->Number());
      DMSG("\n is subsumed by\n ");
      DMSG(curr_instr->TheClause()->Number());  
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef DEBUG_NAMESPACE 
      default : ERROR("X0"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadHeader(const TERM& header)
 

bool FS_CODE_INTEGRATOR::ReadPropLit(const TERM& header)
 {
  CALL("ReadPropLit(const TERM& header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  
  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::PROP_LIT :
      if (curr_instr->HeaderNum() == header_num) 
       {
        ShiftSigFilterRight();
        MoveDown(); 
        UnregFunToInsert(header_num);
        UnregFunToShift(header_num);
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::LIT_HEADER : // no break here! go to the next case 
     case FS_COMMAND::UNORD_EQ : // no break here! go to the next case 
     case FS_COMMAND::ORD_EQ :
      ShiftSigFilterDown();   
      // mismatch occured, try fork
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      UnregFunToInsert(header_num);
      return false;
  
     case FS_COMMAND::SIG_FILTER :   
      PassSigFilter();      
      goto check_tag; 

     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      DWARN("Clause being integrated into FS index is subsumed by already integrated");
      DMSG("Clause\n ");
      DMSG(new_clause->Number());
      DMSG("\n is subsumed by\n ");
      DMSG(curr_instr->TheClause()->Number());  

      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef DEBUG_NAMESPACE 
      default : ERROR("X0"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadPropLit(const TERM& header)
 

bool FS_CODE_INTEGRATOR::ReadUnordEq(const TERM& header)
 {
  CALL("ReadUnordEq(const TERM& header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  
  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::UNORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        ShiftSigFilterRight();
        MoveDown(); 
        UnregFunToInsert(header_num);
        UnregFunToShift(header_num);
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::LIT_HEADER : // no break here! go to the next case 
     case FS_COMMAND::PROP_LIT : // no break here! go to the next case 
     case FS_COMMAND::ORD_EQ :   
      // mismatch occured, try fork
      ShiftSigFilterDown();
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      UnregFunToInsert(header_num);
      return false;
     
     case FS_COMMAND::SIG_FILTER :
      PassSigFilter();
      goto check_tag; 

     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      DWARN("Clause being integrated into FS index is subsumed by already integrated");
      DMSG("Clause\n ");
      DMSG(new_clause->Number());
      DMSG("\n is subsumed by\n ");
      DMSG(curr_instr->TheClause()->Number());  

      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef DEBUG_NAMESPACE 
      default : ERROR("X0"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadUnordEq(const TERM& header)

bool FS_CODE_INTEGRATOR::ReadOrdEq(TERM header)
 {
  CALL("ReadOrdEq(TERM header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  
  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::ORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        ShiftSigFilterRight();
        MoveDown(); 
        header.MakeUnordered();
        UnregFunToInsert(header.HeaderNum());
        UnregFunToShift(header.HeaderNum());
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::LIT_HEADER : // no break here! go to the next case 
     case FS_COMMAND::PROP_LIT : // no break here! go to the next case 
     case FS_COMMAND::UNORD_EQ :   
      // mismatch occured, try fork
      ShiftSigFilterDown();
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      header.MakeUnordered();
      UnregFunToInsert(header.HeaderNum());
      return false;

     case FS_COMMAND::SIG_FILTER :
      PassSigFilter();
      goto check_tag;

     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      DWARN("Clause being integrated into FS index is subsumed by already integrated");
      DMSG("Clause\n ");
      DMSG(new_clause->Number());
      DMSG("\n is subsumed by\n ");
      DMSG(curr_instr->TheClause()->Number());  

      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef DEBUG_NAMESPACE
      default : ERROR("X0"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadOrdEq(TERM header)

void FS_CODE_INTEGRATOR::ReadSecondEqArg()
 {
  CALL("ReadSecondEqArg()");

  FS_COMMAND* curr_instr;
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::SECOND_EQ_ARG : MoveDown(); return;
     case FS_COMMAND::COMPARE_NF :
      PassCompNF(curr_instr);
      goto check_tag; 

     case FS_COMMAND::SIG_FILTER :
      PassSigFilter();
      goto check_tag;

     #ifdef DEBUG_NAMESPACE 
      default : ERROR("X0"); return;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return;
      #endif 
     #endif 
    }; 
 }; // void FS_CODE_INTEGRATOR::ReadSecondEqArg()

bool FS_CODE_INTEGRATOR::ReadFunc(const TERM& func)
 {
  CALL("ReadFunc(const TERM& func)");
  FS_COMMAND* curr_instr;
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::FUNC :
      if (curr_instr->Func() == func) 
       {
        DOP(CheckConstrIntegrity());
        if (curr_instr->Fork()) 
         {
	  if (!constr_to_shift.Identity())  
	   {       
	    constr_passed_in_shift = constr_passed; 
            CopyCompsToShift(curr_instr->ForkAddr()); 
           };
         };
        ShiftSigFilterRight();       
        MoveDown(); 
        UnregFunToInsert(func.HeaderNum()); 
        UnregFunToShift(func.HeaderNum()); 
        return true; 
       };
      // different symbol
      if (curr_instr->Func() > func)
       {
        if (!constr_to_shift.Identity())  
	 {
          constr_passed_in_shift = constr_passed;
          InsertCompsToShift(node_addr);
         };

        curr_instr = *node_addr;
        if ((use_sig_filters) && (sig_to_shift->nonempty()))
	 { 
	  if (!sig_to_shift->containsAtLeast(2)) // singletone
	   {
            ulong hd_num;
            if ((curr_instr->SkeletonHeaderNum(hd_num))
                && (sig_to_shift->contains(hd_num)))
             goto terminate_with_failure;  
           };
          FunSet* new_filter = sig_to_shift->copy();
          FS_COMMAND* new_comm = FS_COMMAND::NewSIG_FILTER(new_filter);
          new_comm->SetNext(*node_addr);
          *node_addr = new_comm;
         }; 
	
        terminate_with_failure:
         UnregFunToInsert(func.HeaderNum());
         return false;
       };  
      // curr_instr->Func() < func here
      // no break here! go to the next case 
      
     case FS_COMMAND::VAR :
      // mismatch occured, try fork
      DOP(CheckConstrIntegrity());
      ShiftSigFilterDown();
      if (curr_instr->Fork())
       {
        if (!constr_to_shift.Identity())
	 {
          constr_passed_in_shift = constr_passed;
          CopyCompsToShift(curr_instr->NextAddr());
         };
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      if (!constr_to_shift.Identity())
       {
        constr_passed_in_shift = constr_passed;
        InsertCompsToShift(curr_instr->NextAddr());
       };    

      MoveRight();
      UnregFunToInsert(func.HeaderNum());
      return false;
     
     case FS_COMMAND::COMPARE_NF :
      PassCompNF(curr_instr);
      goto check_tag;

     case FS_COMMAND::SIG_FILTER :
      PassSigFilter();
      goto check_tag;

 
     #ifdef DEBUG_NAMESPACE 
      default : ERROR("X0"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif 
    };
 }; // bool FS_CODE_INTEGRATOR::ReadFunc(const TERM& func)


bool FS_CODE_INTEGRATOR::ReadVar()
 {
  CALL("ReadVar()");
  FS_COMMAND* curr_instr;
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::VAR : 
      DOP(CheckConstrIntegrity());
      if (curr_instr->Fork()) 
       { 
        if (!constr_to_shift.Identity())
	 {
          constr_passed_in_shift = constr_passed; 
          CopyCompsToShift(curr_instr->ForkAddr()); 
         };
       };
      ShiftSigFilterRight();
      MoveDown(); 
      return true;
      
     case FS_COMMAND::FUNC :
      // a FUNC can not precede a VAR in a tree

      if (!constr_to_shift.Identity())
       {
        constr_passed_in_shift = constr_passed;
        InsertCompsToShift(node_addr);
       };

      curr_instr = *node_addr;
      if ((use_sig_filters) && (sig_to_shift->nonempty()))
       {
        if (!sig_to_shift->containsAtLeast(2)) // singletone
	 {
          ulong hd_num;
          if ((curr_instr->SkeletonHeaderNum(hd_num))
              && (sig_to_shift->contains(hd_num)))
           return false;
         }; 
        FunSet* new_filter = sig_to_shift->copy();
        FS_COMMAND* new_comm = FS_COMMAND::NewSIG_FILTER(new_filter);
        new_comm->SetNext(*node_addr);
        *node_addr = new_comm;
       };   
      return false;
       
     case FS_COMMAND::COMPARE_NF :
      PassCompNF(curr_instr);
      goto check_tag;
     
     case FS_COMMAND::SIG_FILTER :
      PassSigFilter();
      goto check_tag;
  
     #ifdef DEBUG_NAMESPACE 
       default : ERROR("X0"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif 
    };
 }; // bool FS_CODE_INTEGRATOR::ReadVar()

void FS_CODE_INTEGRATOR::ReadCompare(const Flatterm** v1,const Flatterm** v2)
 {
  CALL("ReadCompare(const Flatterm** v1,const Flatterm** v2)");
  ASSERT(v1 < v2);
  ulong var_num1 = VarNum(v1);
  ulong var_num2 = VarNum(v2);
  MakeIntegrated(var_num1,var_num2);
  constr_to_insert.MakeEquivalent(var_num1,var_num2); 
  DOP(CheckConstrIntegrity());
 }; // void FS_CODE_INTEGRATOR::ReadCompare(const Flatterm** v1,const Flatterm** v2)


bool FS_CODE_INTEGRATOR::ReadCollectedComps()
{
 CALL("ReadCollectedComps()");
 DOP(CheckConstrIntegrity());


 FS_COMMAND* curr_instr;
 check_instr:
  curr_instr = *node_addr;
  if (curr_instr->IsCompare() || curr_instr->IsCompareNF())
   {
    if (PassComp(curr_instr) || (*node_addr)) goto check_instr;
    DOP(CheckConstrIntegrity());
    MoveCompsToInsert(node_addr);
    constr_to_shift.Reset();
    ASSERT(!(*node_addr));
    return false;
   };
  
  // ! (curr_instr->IsCompare() || curr_instr->IsCompareNF()) 

  if (curr_instr->IsSigFilter()) { PassSigFilter(); goto check_instr; };
   
  // !(curr_instr->IsCompare() || curr_instr->IsCompareNF()) and !curr_instr->IsSigFilter() 
  
  if (constr_to_shift.Identity())
   {      
    FS_COMMAND** node_addr_backup = node_addr;
    *node_addr = (FS_COMMAND*)0;
    DOP(CheckConstrIntegrity());
    if (MoveCompsToInsert(node_addr))
     {
      ASSERT(!(*node_addr));
      ASSERT((*node_addr_backup)->IsCompareNF());
      (*node_addr_backup)->SetTag(FS_COMMAND::COMPARE);
      (*node_addr_backup)->SetFork(curr_instr);
      if ((use_sig_filters) && (sig_to_shift->nonempty()))
       { 
	if (!sig_to_shift->containsAtLeast(2)) // singletone 
	 { 
          ulong hd_num;
          if ((curr_instr->SkeletonHeaderNum(hd_num))
               && (sig_to_shift->contains(hd_num))) 
	   // no need to shift the sig. filter, the next command cancels it
	   return false;         
         };
        // the sig. filter must be shifted
        FunSet* new_filter = sig_to_shift->copy();
        FS_COMMAND* new_comm = FS_COMMAND::NewSIG_FILTER(new_filter);           
        new_comm->SetNext(curr_instr);
        (*node_addr_backup)->SetFork(new_comm);
       };
      return false;
     }
    else // !MoveCompsToInsert(node_addr) (did not insert anything)
     {
      ASSERT(node_addr == node_addr_backup);
      ASSERT(!(*node_addr));
      *node_addr = curr_instr;          
      return true; 
     };
   }
  else // !constr_to_shift.Identity(), simple case
   {    
    constr_passed_in_shift = constr_passed;
    InsertCompsToShift(node_addr);
    ShiftSigFilterDown();
    ASSERT((*node_addr)->IsCompareNF());    
    (*node_addr)->SetTag(FS_COMMAND::COMPARE);       
    ASSERT((!constr_integrated.Equivalent(VarNum((*node_addr)->Var1()),VarNum((*node_addr)->Var2()))));
    MoveRight();
    MoveCompsToInsert(node_addr);
    ASSERT(!(*node_addr));
    return false; 
   };

}; // bool FS_CODE_INTEGRATOR::ReadCollectedComps()


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE_REMOVAL"
#endif
#include "debugMacros.h"
//=================================================

bool FS_CODE_REMOVAL::SkipComplexTerm(TERM* t)
 {
  TERM::Iterator iter;
  iter.Reset(t);
  skip_sym:
   if (iter.CurrentSymIsVar())
    {
     if (!SkipVar(iter.Symbol().var())) return false;
    }
   else // functor
    {
     if (!(SkipFunc(iter.Symbol()))) return false;
    };
   if (iter.Next()) goto skip_sym; 
   return true; 
 }; // bool FS_CODE_REMOVAL::SkipComplexTerm(TERM* t)

bool FS_CODE_REMOVAL::SkipEqLiteral(TERM* lit)
 {
  ASSERT(lit->IsEquality());               
  if (lit->IsOrderedEq())
   {
    if (!(SkipOrdEq(*lit))) return false;
   }
  else
   {
    if (!(SkipUnordEq(*lit))) return false;
   }; 
  TERM* args = lit->Args();
  TERM arg1 = *args;
  TERM arg2 = *(args->Next());
  
  return (SkipTerm(arg1)) && (SkipSecondEqArg()) && (SkipTerm(arg2));
  
 }; // bool FS_CODE_REMOVAL::SkipEqLiteral(TERM* lit)


bool FS_CODE_REMOVAL::SkipLiteral(TERM* lit)
 {
  if (lit->IsEquality()) return SkipEqLiteral(lit);
 
  TERM::Iterator iter;
  TERM sym;
  
  if (lit->arity())
   {
    if (SkipHeader(*lit))
     {
      iter.Reset(lit);
      next_sym:
       if (!(iter.Next())) return true;
       sym = iter.Symbol();
       if (iter.CurrentSymIsVar())
        {
         if (!SkipVar(sym.var())) return false;
        }
       else // functor
        {
         if (!(SkipFunc(sym))) return false;
        };
       goto next_sym;   
      }
     else return false;
    }
   else // propositional literal
    {
     return SkipPropLit(*lit);
    };
 }; // bool FS_CODE_REMOVAL::SkipLiteral(TERM* lit)
//===========================================================================
