//
// File:         fs_code.h
// Description:  Maintainence of index for FS.
// Created:      Oct 15, 1999, 15:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef FS_CODE_H
//===================================================================
#define FS_CODE_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "GlobAlloc.h"
#include "term.h"
#include "clause.h"
#include "MultisetOfVariables.h"
#include "Flatterm.h"
#include "EquivalenceRelation.h"
#include "fs_command.h"
#include "old_fs_code.h"
//===================================================================
typedef EquivalenceRelation<HellConst::MaxClauseSize> FS_EQ_CONSTR; 
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE_INTEGRATOR"
#endif
#include "debugMacros.h"
//=================================================
class FS_CODE_INTEGRATOR // must be static - too costly constructor
 {  
  private:
   Clause* new_clause;
   FS_COMMAND** node_addr;
   
   FS_EQ_CONSTR constr_to_insert;
   FS_EQ_CONSTR constr_to_shift;
   FS_EQ_CONSTR constr_passed;
   FS_EQ_CONSTR constr_passed_in_shift; 
   FS_EQ_CONSTR constr_integrated;
   FS_EQ_CONSTR constr_old;  

   
   DOP(ulong debugLitNum);
   
   const Flatterm** fs_subst;
  
   const Flatterm** first_occ[HellConst::MaxNumOfVariables];
   MultisetOfVariables<HellConst::MaxNumOfVariables> vars; 
   const Flatterm** fs_var;
   bool use_sig_filters;
   FunSet* sig_to_insert;
   FunSet* sig_to_shift;
  public:
   FS_CODE_INTEGRATOR(const Flatterm** subst) 
    : constr_to_insert("constr_to_insert"),
      constr_to_shift("constr_to_shift"),
      constr_passed("constr_passed"),
      constr_passed_in_shift("constr_passed_in_shift"),
      constr_integrated("constr_integrated"),
      constr_old("constr_old"),
      fs_subst(subst),
      vars(),
      use_sig_filters(true) // default
    { 
     CALL("constructor FS_CODE_INTEGRATOR(const Flatterm** subst)");
     /****
     sig_to_insert = new FunSet();
     sig_to_shift = new FunSet(); 
     *****/
     sig_to_insert = (FunSet*)0;
     sig_to_shift = (FunSet*)0;
    };
   ~FS_CODE_INTEGRATOR() { CALL("destructor ~FS_CODE_INTEGRATOR()"); };
   void SetUseSigFilters(bool fl) { use_sig_filters = fl; };
   void Reset(FS_COMMAND** n,Clause* cl) 
    {
     new_clause = cl; 
     node_addr = n; 
     constr_to_insert.Reset();
     constr_to_shift.Reset();
     constr_passed.Reset();
     //constr_passed_in_shift.Reset();
     constr_integrated.Reset();
     constr_old.Reset();
     DOP(debugLitNum = 0);
     vars.Reset();
     fs_var = fs_subst;
     if (use_sig_filters) 
      {
       if (sig_to_insert) delete sig_to_insert;
       sig_to_insert = new FunSet();
       if (sig_to_shift) delete sig_to_shift;
       sig_to_shift = new FunSet();       
       CollectSignature(cl);       
      };
    };

   void CollectSignature(const Clause* cl);
   void CollectSignature(const TERM* lit);
   void RegFunToInsert(ulong fun_num);
   void UnregFunToInsert(ulong fun_num);
   void UnregFunToShift(ulong fun_num);
   void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
   void MoveRight() { node_addr = (*node_addr)->ForkAddr(); };
   ulong VarNum(const Flatterm** v) const { return v - fs_subst; };
   const Flatterm** Var(const ulong& var_num) const { return fs_subst + var_num; };

   void MakeShifted(const ulong& v1,const ulong& v2) 
   {
    CALL("MakeShifted(const ulong& v1,const ulong& v2)");
    ASSERT(v1 < v2);
    ASSERT(!constr_integrated.Equivalent(v1,v2));
    constr_to_shift.MakeEquivalent(v1,v2);
   };

   void MakeIntegrated(const ulong& v1,const ulong& v2) 
   {
    CALL("MakeIntegrated(const ulong& v1,const ulong& v2)");
    ASSERT(v1 < v2);
    ASSERT(!constr_integrated.Equivalent(v1,v2));
    ASSERT(!constr_passed.Equivalent(v1,v2));
    ASSERT(!constr_old.Equivalent(v1,v2));
    constr_integrated.MakeEquivalent(v1,v2);
   };

   #ifdef DEBUG_NAMESPACE
    void CheckConstrIntegrity()
    { 
     CALL("CheckConstrIntegrity()");
     ASSERT(constr_to_insert.Implies(constr_integrated));      
     ASSERT(constr_passed.Implies(constr_integrated));   
     ASSERT(constr_to_shift.Implies(constr_old));        
     ASSERT(constr_passed.Implies(constr_old));
     static FS_EQ_CONSTR debugER("debugER");  
     debugER.MakeComposition(constr_to_shift,constr_passed);
     ASSERT(constr_old.Implies(debugER));
     debugER.MakeComposition(constr_to_insert,constr_passed);
     ASSERT(constr_integrated.Implies(debugER));
    }; //void CheckConstrIntegrity() 
   #endif 

   void PassCompNF(FS_COMMAND* instr)
   {
    CALL("PassCompNF(FS_COMMAND* instr)");
    DOP(CheckConstrIntegrity());
    ASSERT(instr->IsCompareNF());
    const Flatterm** var1 = instr->Var1();
    const Flatterm** var2 = instr->Var2();
    const ulong var_num1 = VarNum(var1);
    const ulong var_num2 = VarNum(var2);

    ASSERT(!(instr->Fork()));
    ASSERT(!constr_to_shift.Equivalent(var_num1,var_num2));
    ASSERT(!constr_old.Equivalent(var_num1,var_num2));
    ASSERT(!constr_passed.Equivalent(var_num1,var_num2));
   
    bool shared = constr_integrated.Equivalent(var_num1,var_num2);
    
    if (shared)
     {
      // the current instruction is shared
      constr_passed.MakeEquivalent(var_num1,var_num2);
      node_addr = instr->NextAddr();
     }
    else // ! constr_integrated.Equivalent(var_num1,var_num2)
     {
      // the instruction itself must be shifted unless it can be replaced by an equivalent one 
      *node_addr = instr->Next();
      delete instr;
     };

    instr = *node_addr;
    FS_EQ_CONSTR::MergedPairs mergedPairs(constr_old,var_num1,var_num2);
    ulong merged1;
    ulong merged2;

    while (mergedPairs.nextPair(merged1,merged2))
     {
      ASSERT(merged1 < merged2);     
      ASSERT(!constr_old.Equivalent(merged1,merged2));
      if (constr_integrated.Equivalent(merged1,merged2) && (!constr_passed.Equivalent(merged1,merged2)))
       {
	// another COMPARE can be shared	 
	shared = true;
        constr_passed.MakeEquivalent(merged1,merged2);
        *node_addr = FS_COMMAND::NewCOMPARE_NF(Var(merged1),Var(merged2));
        (*node_addr)->SetNext(instr);
        MoveDown();
       };
     };    
  
    constr_old.MakeEquivalent(var_num1,var_num2);
    if (!shared) MakeShifted(var_num1,var_num2);
    ASSERT((*node_addr) == instr);
    DOP(CheckConstrIntegrity());
   }; // void PassCompNF(FS_COMMAND* instr) 

   bool PassComp(FS_COMMAND* instr)
   {
    CALL("PassComp(FS_COMMAND* instr)");
    DOP(CheckConstrIntegrity()); 
    ASSERT(instr->IsCompareNF() ||instr->IsCompare() );
    const Flatterm** var1 = instr->Var1();
    const Flatterm** var2 = instr->Var2();
    ulong var_num1 = VarNum(var1);
    ulong var_num2 = VarNum(var2);
    ASSERT(!constr_to_shift.Equivalent(var_num1,var_num2));
    ASSERT(!constr_old.Equivalent(var_num1,var_num2)); 
    ASSERT(!constr_passed.Equivalent(var_num1,var_num2));

    ASSERT((*node_addr) == instr);

    if (constr_integrated.Equivalent(var_num1,var_num2))
     {
      goto shift_and_insert_shared;      
     }
    else // try to replace this COMPARE by an equivalent one which is also in constr_integrated
     {
      FS_EQ_CONSTR::MergedPairs mergedPairs(constr_old,var_num1,var_num2);
      ulong merged1;
      ulong merged2; 
      while (mergedPairs.nextPair(merged1,merged2))
       {
        ASSERT(merged1 < merged2);     
        ASSERT(!constr_old.Equivalent(merged1,merged2));
        if (constr_integrated.Equivalent(merged1,merged2) && (!constr_passed.Equivalent(merged1,merged2)))
         {
	  // appropriate replacement found
	  var_num1 = merged1;
          var_num2 = merged2;
          // modify the arguments of instr
          instr->SetVar1(Var(var_num1));
          instr->SetVar2(Var(var_num2));         
          goto shift_and_insert_shared;
         };
       };
      
      // no appropriate replacement found, must follow the fork

      if (instr->IsCompareNF()) instr->SetTag(FS_COMMAND::COMPARE); 
      if (!constr_to_shift.Identity())
       {
        constr_passed_in_shift = constr_passed;
        constr_passed_in_shift.MakeEquivalent(var_num1,var_num2);
        CopyCompsToShift(instr->NextAddr());
       };
      ShiftSigFilterDown();
      MoveRight();  
      DOP(CheckConstrIntegrity());
      return false;
     };

    shift_and_insert_shared:
     COP("shift_and_insert_shared");
     // make necessary shifts into the fork
     if (instr->Fork()) 
      {
       if (!constr_to_shift.Identity())
        {
         constr_passed_in_shift = constr_passed; 
         CopyCompsToShift(instr->ForkAddr());  
        };      
       ShiftSigFilterRight();
      };
     constr_passed.MakeEquivalent(var_num1,var_num2);
     MoveDown();
     instr = *node_addr; 
     // now find all COMPARES that can be shared due to (var_num1 ~ var_num2) added to constr_old
 
    
     FS_EQ_CONSTR::MergedPairs mergedPairs(constr_old,var_num1,var_num2);
     ulong merged1;
     ulong merged2; 

     while (mergedPairs.nextPair(merged1,merged2))
      {
       ASSERT(merged1 < merged2);     
       ASSERT(!constr_old.Equivalent(merged1,merged2));

       if (constr_integrated.Equivalent(merged1,merged2) && (!constr_passed.Equivalent(merged1,merged2)))
        {
 	 // another COMPARE can be shared	
         constr_passed.MakeEquivalent(merged1,merged2);
         *node_addr = FS_COMMAND::NewCOMPARE_NF(Var(merged1),Var(merged2));
         (*node_addr)->SetNext(instr);
         MoveDown();
        };
      };    

     ASSERT((*node_addr) == instr);

     constr_old.MakeEquivalent(var_num1,var_num2); 

     COP("end");
     DOP(CheckConstrIntegrity());     
     return true;
   }; // bool PassComp(FS_COMMAND* instr)

   void CopyCompsToShift(FS_COMMAND** addr)
    {
     CALL("CopyCompsToShift(FS_COMMAND** addr)");
     DOP(bool debugAtLeastOneShifted = constr_to_shift.Identity());  
     ASSERT(constr_passed.Implies(constr_passed_in_shift));
     ulong vn1;
     ulong vn2;
     FS_EQ_CONSTR::Iterator iter(constr_to_shift);
     FS_COMMAND* old_instr = *addr;
     while (iter.Next(vn2,vn1)) 
      {
       ASSERT(vn1 < vn2); 
       ASSERT(!constr_integrated.Equivalent(vn1,vn2));
       ASSERT(constr_old.Equivalent(vn1,vn2));
      
       if (!constr_passed_in_shift.Equivalent(vn1,vn2))
	{
         DOP(debugAtLeastOneShifted = true);  
         constr_passed_in_shift.MakeEquivalent(vn1,vn2);
         FS_COMMAND* new_instr = FS_COMMAND::NewCOMPARE_NF(Var(vn1),Var(vn2));
         *addr = new_instr;
         addr = new_instr->NextAddr();
        }; 
      };
     *addr = old_instr;
     ASSERT(debugAtLeastOneShifted);
    }; // void CopyCompsToShift(FS_COMMAND** addr)


   void InsertCompsToShift(FS_COMMAND** addr)
    {
     CALL("InsertCompsToShift(FS_COMMAND** addr)");
     DOP(bool debugAtLeastOneShifted = constr_to_shift.Identity());  
     DOP(CheckConstrIntegrity());     
     ASSERT(constr_passed.Implies(constr_passed_in_shift));
     ulong vn1;
     ulong vn2;
     FS_COMMAND* old_instr = *addr;
     while (constr_to_shift.Split(vn2,vn1)) 
      { 
       ASSERT(vn1 < vn2);
       ASSERT(!constr_integrated.Equivalent(vn1,vn2));
       ASSERT(constr_old.Equivalent(vn1,vn2));

       if (!constr_passed_in_shift.Equivalent(vn1,vn2))
	{
	 DOP(debugAtLeastOneShifted = true);  
         constr_passed_in_shift.MakeEquivalent(vn1,vn2);
         FS_COMMAND* new_instr = FS_COMMAND::NewCOMPARE_NF(Var(vn1),Var(vn2));
         *addr = new_instr;
         addr = new_instr->NextAddr(); 
        };
      };
     ASSERT(constr_to_shift.Identity());
     *addr = old_instr;
     ASSERT(debugAtLeastOneShifted);
    }; // void InsertCompsToShift(FS_COMMAND** addr)


   bool MoveCompsToInsert(FS_COMMAND**& addr)
    {
     CALL("MoveCompsToInsert(FS_COMMAND**& addr)");
     ulong vn1;
     ulong vn2;
     FS_COMMAND* old_instr = *addr;
     bool res = false;
     while (constr_to_insert.Split(vn2,vn1)) 
      { 
       ASSERT(constr_integrated.Equivalent(vn1,vn2));
       ASSERT(vn1 < vn2);  
       if (!constr_passed.Equivalent(vn1,vn2))
	{
	 ASSERT(!constr_old.Equivalent(vn1,vn2));
	 constr_passed.MakeEquivalent(vn1,vn2);
         FS_COMMAND* new_instr = FS_COMMAND::NewCOMPARE_NF(Var(vn1),Var(vn2));
         *addr = new_instr;
         addr = new_instr->NextAddr();
         res = true;
        };
      };
     ASSERT(constr_to_insert.Identity());
     *addr = old_instr;

     COP("end");
     ASSERT(constr_to_insert.Identity());
     ASSERT(constr_passed.Implies(constr_integrated));
     ASSERT(constr_integrated.Implies(constr_passed));
     return res;
    }; // bool MoveCompsToInsert(FS_COMMAND**& addr)
   


   void PassSigFilter()
    {
     CALL("PassSigFilter()");
     ASSERT(CurrInstr()->IsSigFilter());
     ASSERT(!CurrInstr()->Fork()); 

     FS_COMMAND* curr_comm = *node_addr;
     FunSet* curr_filter = curr_comm->Functors();
     
     FunSet* intersection = sig_to_insert->intersection(*curr_filter);
     if (intersection->nonempty()) 
      {         
       curr_comm->SetFunctors(intersection);
       node_addr = curr_comm->NextAddr();
     
       FunSet* diff_to_shift = curr_filter->difference(*intersection);
       delete curr_filter;
       FunSet* new_sig_to_shift = sig_to_shift->sum(*diff_to_shift);
       delete diff_to_shift;
       delete sig_to_shift;
       sig_to_shift = new_sig_to_shift;
       
       FunSet* new_sig_to_insert = sig_to_insert->difference(*intersection);
       delete sig_to_insert;
       sig_to_insert = new_sig_to_insert;
      }
     else // intersection->empty()
      {
       delete intersection;
       *node_addr = curr_comm->Next();               
       FunSet* new_sig_to_shift = sig_to_shift->sum(*curr_filter);
       delete sig_to_shift;
       sig_to_shift = new_sig_to_shift;
       delete curr_comm;  
      };
    }; // void PassSigFilter()    

   void ShiftSigFilterRight()
    {
     FS_COMMAND* fork = (*node_addr)->Fork();
     if ((fork) 
         && (use_sig_filters) 
         && (sig_to_shift->nonempty())) 
      {
       if (!sig_to_shift->containsAtLeast(2)) // singletone
	{
         ulong hd_num;
         if ((fork->SkeletonHeaderNum(hd_num)) 
             && (sig_to_shift->contains(hd_num)))
          return; // there is no sense to insert the filter  
        };
       FS_COMMAND* filter = FS_COMMAND::NewSIG_FILTER(sig_to_shift->copy());
       filter->SetNext(fork);
       (*node_addr)->SetFork(filter);
      };
    };   
   
   void ShiftSigFilterDown()
    {
     if (use_sig_filters)
      {
       FS_COMMAND* curr_comm = *node_addr;
       ulong hd_num;
       FunSet* new_filter;
       if (curr_comm->SkeletonHeaderNum(hd_num)) { new_filter = sig_to_shift->remove(hd_num); }
       else new_filter = sig_to_shift->copy();
       if (new_filter->nonempty())
        {
         if (!new_filter->containsAtLeast(2)) // singletone
	  {
           if ((curr_comm->Next()->SkeletonHeaderNum(hd_num)) 
               && (new_filter->contains(hd_num)))
	     { // there is no sense to insert the filter  
              delete new_filter;
              return;
             };
          };
         FS_COMMAND* new_comm = FS_COMMAND::NewSIG_FILTER(new_filter);
         new_comm->SetNext(curr_comm->Next());
         curr_comm->SetNext(new_comm);
        }
       else delete new_filter;
      };
    };  


    // Read mode
    
   bool ReadHeader(const TERM& header);
   bool ReadPropLit(const TERM& header);
   bool ReadUnordEq(const TERM& header);
   bool ReadOrdEq(TERM header);
   void ReadSecondEqArg();
   bool ReadFunc(const TERM& func);
   bool ReadVar();
   bool ReadVar(const ulong& var);
   void ReadCompare(const Flatterm** v1,const Flatterm** v2);
   bool ReadCollectedComps();
   bool ReadLiteral(TERM* lit);
   bool ReadEqLiteral(TERM* lit); 
   bool ReadComplexTerm(TERM* t);
   bool ReadTerm(const TERM& t)
    {
     CALL("ReadTerm(const TERM& t)");
     return (t.isVariable()) ? ReadVar(t.var()) : ReadComplexTerm(t.First());
    };
   
    // Write mode
    
   void WriteLiteral(TERM* lit);
   void WriteEqLiteral(TERM* lit);
   
   void EnterWriteMode();
   
   void WriteHeader(TERM header) 
    { 
     *node_addr = FS_COMMAND::NewLIT_HEADER(header);
     MoveDown();
    }; 
   void WritePropLit(TERM header)
    {
     *node_addr = FS_COMMAND::NewPROP_LIT(header);
     MoveDown();
    }; 
   void WriteUnordEq(TERM header)
    {
     *node_addr = FS_COMMAND::NewUNORD_EQ(header);
     MoveDown();
    };
   void WriteOrdEq(TERM header)
    {
     *node_addr = FS_COMMAND::NewORD_EQ(header);
     MoveDown();
    };
   void WriteSecondEqArg()
    {
     *node_addr = FS_COMMAND::NewSECOND_EQ_ARG();
     MoveDown();
    }; 
    
   void WriteFunc(TERM func)
    {
     *node_addr = FS_COMMAND::NewFUNC(func);
     MoveDown();
    }; 
   
   void InsertFunc(TERM func)
    {
     FS_COMMAND* tmp = *node_addr;
     *node_addr = FS_COMMAND::NewFUNC(func);
     (*node_addr)->SetFork(tmp);
     MoveDown();
    };
    
   void WriteVar(const Flatterm** v)
    {
     *node_addr = FS_COMMAND::NewVAR(v);
     MoveDown();
    };  
   
   void WriteVar(ulong var)
    {
     WriteVar(fs_var);
     if (vars.Occurences(var))
      {
       WriteCompare(first_occ[var],fs_var);
      }
     else // first occurence
      {
       vars.Register(var);
       first_occ[var] = fs_var;
      }; 
     fs_var++;
    };
   
   void WriteComplexTerm(TERM* t)
    {
     TERM::Iterator iter;
     iter.Reset(t);
     
     write_curr_sym:
      if (iter.CurrentSymIsVar())
       {
        WriteVar(iter.Symbol().var());
       }
      else // functor
       {
        WriteFunc(iter.Symbol());
       };
      if (iter.Next()) goto write_curr_sym;
    }; 
    
   void WriteTerm(TERM t)
    {
     if (t.isVariable()) { WriteVar(t.var()); }
     else WriteComplexTerm(t.First());
    };
    
   void InsertVar(const Flatterm** v)
    {
     FS_COMMAND* tmp = *node_addr;
     *node_addr = FS_COMMAND::NewVAR(v);
     (*node_addr)->SetFork(tmp);
     MoveDown();
    }; 
   void WriteCompare(const Flatterm** v1,const Flatterm** v2)
    {
     CALL("WriteCompare(const Flatterm** v1,const Flatterm** v2)");
     DOP(MakeIntegrated(VarNum(v1),VarNum(v2)));
     *node_addr = FS_COMMAND::NewCOMPARE_NF(v1,v2);
     MoveDown(); 
    };
   void WriteSuccess(Clause* cl)
    {
     *node_addr = FS_COMMAND::NewSUCCESS(cl);
    };
   void InsertSuccess(Clause* cl)
    {
     FS_COMMAND* tmp = *node_addr;
     *node_addr = FS_COMMAND::NewSUCCESS(cl);
     (*node_addr)->SetFork(tmp);
    };
  public: // for debug
   #ifndef NO_DEBUG
    const FS_COMMAND* CurrInstr() const { return *node_addr; };  
   #endif

   #ifndef NO_DEBUG_VIS
    ostream& outputComps(ostream& str) 
    {
     constr_integrated.Normalize();
     constr_old.Normalize();
     constr_passed.Normalize();
     constr_to_shift.Normalize();
     constr_to_insert.Normalize();
     return str << constr_integrated << '\n'
                << constr_old << '\n'
                << constr_passed << '\n'
                << constr_to_shift << '\n'
                << constr_to_insert << '\n';      

    };
   #endif

  private: 
 
     // Processing clauses that are accidentally forward subsumed
   void AccidForwardSubsumed(Clause* subsumed,Clause* subsuming)
    {
     // temporary
     #ifdef DEBUG_NAMESPACE     
      DWARN("  Clause ");       
      DWARN(subsumed);
      DWARN("  being integrated for forward subsumption");
      DWARN("  was subsumed by already integrated one:");
      DWARN(subsuming);
     #else 
      #ifdef _SUPPRESS_WARNINGS_
       Clause* dummy = subsumed; dummy = subsuming; 
      #endif   
     #endif     
    };
     
 }; // class FS_CODE_INTEGRATOR

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE_REMOVAL"
#endif
#include "debugMacros.h"
//================================================= 
class FS_CODE_REMOVAL
 {
  private:
   Clause* removed_clause;
   FS_COMMAND** node_addr;
   FS_COMMAND** branch_to_destroy;
   
   FS_COMMAND* comps; // list of instructions, can be changed to a Stack
   
   FS_EQ_CONSTR constr;
   
   const Flatterm** fs_subst;
  
   const Flatterm** first_occ[HellConst::MaxNumOfVariables];
   MultisetOfVariables<HellConst::MaxNumOfVariables> vars; 
   const Flatterm** fs_var; 
   bool use_sig_filters;
   
  private:
   void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
   void MoveRight() { node_addr = (*node_addr)->ForkAddr(); };
   ulong VarNum(const Flatterm** v) const { return v - fs_subst; };
   const Flatterm** Var(ulong var_num) const { return fs_subst + var_num; }; 
  public:
   FS_CODE_REMOVAL(const Flatterm** subst)
    : constr(), 
      fs_subst(subst),
      vars(),
      use_sig_filters(true) // default
    {
     CALL("constructor FS_CODE_REMOVAL(const Flatterm** subst)");
    };
   ~FS_CODE_REMOVAL() { CALL("destructor ~FS_CODE_REMOVAL()"); };
   void Reset(FS_COMMAND** n,Clause* cl) 
    {
     removed_clause = cl;
     node_addr = n; 
     branch_to_destroy = n; 
     comps = (FS_COMMAND*)0;
     constr.Reset();
     vars.Reset();
     fs_var = fs_subst;
    };
   
   void SetUseSigFilters(bool fl) { use_sig_filters = fl; };

   bool SkipComplexTerm(TERM* t);
   bool SkipLiteral(TERM* lit); 
   bool SkipEqLiteral(TERM* lit); 
   
   bool SkipTerm(TERM t) 
    {
     return (t.isVariable()) ? SkipVar(t.var()) : SkipComplexTerm(t.First());
    };
   
    
   bool SkipHeader(TERM header);
   bool SkipPropLit(TERM header);
   bool SkipUnordEq(TERM header);
   bool SkipOrdEq(TERM header);
   bool SkipSecondEqArg();
   
   bool SkipFunc(TERM func);
   bool SkipVar();
   bool SkipVar(ulong var)
    {
     if (SkipVar())
      {
       if (vars.Occurences(var))
        {
         SkipCompare(first_occ[var],fs_var);
        } 
       else // first occurence of the variable
        {
         vars.Register(var);
         first_occ[var] = fs_var;
        }; 
       fs_var++;
       return true;
      }
     else return false;
    }; 
   void SkipCompare(const Flatterm** v1,const Flatterm** v2); 
   bool FindSuccess();
   void DestroyBranch() 
    {
     FS_COMMAND* tmp = *branch_to_destroy;
     *branch_to_destroy = tmp->Fork();
     tmp->SetFork((FS_COMMAND*)0);
     FS_COMMAND::DestroyTree(tmp);
    };
  public: // for debugging
   #ifndef NO_DEBUG
    FS_COMMAND* Node() const { return *node_addr; }; 
   #endif
 }; // class FS_CODE_REMOVAL 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE"
#endif
#include "debugMacros.h"
//=================================================
class FS_CODE 
 {
  private:
   bool oldStyle;
   OLD_FS_CODE oldStyleCode;
   FS_COMMAND* tree;
   const Flatterm** subst;
   FS_CODE_INTEGRATOR integrator; // quite heavy object (slow constructor), can not be local
   FS_CODE_REMOVAL removal; // quite heavy object (slow constructor), can not be local
  public:
   FS_CODE(const Flatterm** sbst) 
    : oldStyle(false),
      oldStyleCode(sbst),
      tree((FS_COMMAND*)0),
      subst(sbst),
      integrator(subst), 
      removal(subst)  
    {
     CALL("constructor FS_CODE(const Flatterm** sbst)");
    };
   ~FS_CODE() { CALL("destructor ~FS_CODE()"); };
   #ifndef NO_DEBUG      
    bool clean() const
    {
     return oldStyleCode.clean()
            && (!tree);  
    };
   #endif
   void SetOldStyle(bool fl) { oldStyle = fl; };
   void SetUseSigFilters(bool fl)
    {
     integrator.SetUseSigFilters(fl);
     removal.SetUseSigFilters(fl);
    };
   FS_COMMAND* Tree() const 
    {
      //DF; cout << "OLD STYLE " << oldStyle << "\n";
     if (oldStyle) return oldStyleCode.theTree();
     return tree; 
    }; 
   const Flatterm** Subst() { return subst; };
   void Integrate(Clause* cl);
   void Integrate(TMP_LIT_LIST& component,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,void* name)
    {
     oldStyleCode.integrate(component,vars,name);
    };
   void Remove(Clause* cl);
   void Destroy()
    {
     if (oldStyle) { oldStyleCode.destroy(); return; };
     if (tree) FS_COMMAND::DestroyTree(tree);
     tree = (FS_COMMAND*)0;
    };
   void Destroy(void (*destroyIndexedObj)(void*))
    {
     if (oldStyle) { oldStyleCode.destroy(destroyIndexedObj); return; };
     if (tree) FS_COMMAND::DestroyTree(tree,destroyIndexedObj);
     tree = (FS_COMMAND*)0;
    };
  public: // for debugging
   #ifndef NO_DEBUG_VIS
    ostream& outputCode(ostream& str) const
    {
     if (oldStyle) return oldStyleCode.outputCode(str);    
     if (!tree) return str << "<<< Empty FS_CODE >>>\n"; 
     return FS_COMMAND::outputTree(str,tree,subst)
             << "<<< End of FS_CODE >>>\n";  
    }; 
   #endif
 }; // class FS_CODE

 
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE_INTEGRATOR"
#endif
#include "debugMacros.h"
//=================================================
inline void FS_CODE_INTEGRATOR::CollectSignature(const Clause* cl)
{
 for(LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
   CollectSignature(iter.currentElement()->literal());
}; // void FS_CODE_INTEGRATOR::CollectSignature(const Clause* cl)


inline void FS_CODE_INTEGRATOR::CollectSignature(const TERM* lit)
{
 TERM header = *lit;
 if (header.IsEquality()) header.MakeUnordered();
 RegFunToInsert(header.HeaderNum()); 
 TERM::Iterator iter;
 iter.Reset(lit); 
 while (iter.Next()) 
  {
   header = iter.symbolRef();
   if (header.isComplex()) RegFunToInsert(header.HeaderNum());        
  };
}; // void FS_CODE_INTEGRATOR::CollectSignature(const TERM* lit)

inline void FS_CODE_INTEGRATOR::RegFunToInsert(ulong fun_num)
{
 if (!sig_to_insert->write(fun_num))
  {
   FunSet* new_sig = sig_to_insert->add(fun_num);
   delete sig_to_insert;
   sig_to_insert = new_sig;
  };
}; // void FS_CODE_INTEGRATOR::RegFunToInsert(ulong fun_num)

inline void FS_CODE_INTEGRATOR::UnregFunToInsert(ulong fun_num)
{
 if (use_sig_filters)
  {
   if (!sig_to_insert->erase(fun_num))
    {
     FunSet* new_sig = sig_to_insert->remove(fun_num);
     delete sig_to_insert;
     sig_to_insert = new_sig;
    };
  };
}; // void FS_CODE_INTEGRATOR::UnregFunToInsert(ulong fun_num)

inline void FS_CODE_INTEGRATOR::UnregFunToShift(ulong fun_num)
{
 if (use_sig_filters)
  {
   if (!sig_to_shift->erase(fun_num))
    {
     FunSet* new_sig = sig_to_shift->remove(fun_num);
     delete sig_to_shift;
     sig_to_shift = new_sig;
    };
  };
}; // void FS_CODE_INTEGRATOR::UnregFunToShift(ulong fun_num)


inline void FS_CODE_INTEGRATOR::EnterWriteMode()
 {
  CALL("EnterWriteMode()");
  // must be used if ReadVar or ReadFunc failed
  ASSERT(constr_to_shift.Identity());

  ASSERT(constr_to_insert.Implies(constr_integrated));
  ASSERT(constr_passed.Implies(constr_integrated));
  ASSERT(constr_passed.Implies(constr_old));

  //constr_to_insert.Subtract(constr_passed);
    

  if ((use_sig_filters) && (sig_to_insert->nonempty()))
   {
    if ((!sig_to_insert->containsAtLeast(2)) && (constr_to_insert.Identity()))
     {
      COP("END0");
      ASSERT(!(*node_addr)); 
      return; // nothing to insert
     };
    FS_COMMAND* filter = FS_COMMAND::NewSIG_FILTER(sig_to_insert->copy());
    filter->SetNext(*node_addr);   
    *node_addr = filter;
    node_addr = filter->NextAddr();
   };   

  MoveCompsToInsert(node_addr);
  COP("END1");
  ASSERT(!(*node_addr));
 }; // void FS_CODE_INTEGRATOR::EnterWriteMode()

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FS_CODE
 #define DEBUG_NAMESPACE "FS_CODE_REMOVAL"
#endif
#include "debugMacros.h"
//=================================================
inline bool FS_CODE_REMOVAL::SkipHeader(TERM header)
 {
  CALL("SkipHeader(TERM header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::LIT_HEADER :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 
     case FS_COMMAND::COMPARE_NF : // no break here
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };
    
     case FS_COMMAND::SIG_FILTER :
      node_addr = curr_instr->NextAddr();
      goto check_instr;   

     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipHeader(TERM header)

inline bool FS_CODE_REMOVAL::SkipPropLit(TERM header)
 {
  CALL("SkipPropLit(TERM header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
 
  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::PROP_LIT :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 

     case FS_COMMAND::COMPARE_NF : // no break here
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };
     
     case FS_COMMAND::SIG_FILTER :
      node_addr = curr_instr->NextAddr();
      goto check_instr;   

     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipPropLit(TERM header)

inline bool FS_CODE_REMOVAL::SkipUnordEq(TERM header)
 {
  CALL("SkipUnordEq(TERM header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::UNORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 
  
     case FS_COMMAND::COMPARE_NF : // no break here
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2); 
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };
    
     case FS_COMMAND::SIG_FILTER :
      node_addr = curr_instr->NextAddr();
      goto check_instr;    

     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipUnordEq(TERM header)


inline bool FS_CODE_REMOVAL::SkipOrdEq(TERM header)
 {
  CALL("SkipOrdEq(TERM header)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  FS_COMMAND* curr_instr;
  ulong header_num = header.HeaderNum();
  ASSERT(*node_addr);
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::ORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 

     case FS_COMMAND::COMPARE_NF : // no break here
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };

     case FS_COMMAND::SIG_FILTER :
      node_addr = curr_instr->NextAddr();
      goto check_instr;   

     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipOrdEq(TERM header)

inline bool FS_CODE_REMOVAL::SkipSecondEqArg()
 {
  CALL("SkipSecondEqArg()");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  
  FS_COMMAND* curr_instr;
  check_instr:
   curr_instr = *node_addr;
   ASSERT(!(curr_instr->Fork()));
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::SECOND_EQ_ARG : 
      node_addr = curr_instr->NextAddr();
      return true;
     case FS_COMMAND::COMPARE_NF : // no break here
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else return false; 

     case FS_COMMAND::SIG_FILTER :
      node_addr = curr_instr->NextAddr();
      goto check_instr;   

     #ifdef DEBUG_NAMESPACE
      default: ERROR("X0"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif
     #endif 
    };
 }; // bool FS_CODE_REMOVAL::SkipSecondEqArg()


inline bool FS_CODE_REMOVAL::SkipFunc(TERM func)
 {
  CALL("SkipFunc(TERM func)");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  
  FS_COMMAND* curr_instr;
  ASSERT(*node_addr);
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::FUNC :
      if (curr_instr->Func() == func)
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false;
     case FS_COMMAND::COMPARE_NF : // no break here
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
	ASSERT(!(curr_instr->Fork()));
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else return false;
 
     case FS_COMMAND::SIG_FILTER :
      node_addr = curr_instr->NextAddr();
      goto check_instr;   

     default : goto try_fork;  
    };
 }; // bool FS_CODE_REMOVAL::SkipFunc(TERM func)

inline bool FS_CODE_REMOVAL::SkipVar()
 {
  CALL("SkipVar()");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
 
  FS_COMMAND* curr_instr;
  ASSERT(*node_addr);
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::VAR :
      if (curr_instr->Fork()) branch_to_destroy = node_addr;
      node_addr = curr_instr->NextAddr();
      return true;
     case FS_COMMAND::COMPARE_NF : // no break here
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
	ASSERT(!(curr_instr->Fork()));
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else return false;

     case FS_COMMAND::SIG_FILTER :
      node_addr = curr_instr->NextAddr();
      goto check_instr;   

     default : 
      // try fork 
      if (curr_instr->Fork()) 
       {
        node_addr = curr_instr->ForkAddr();
        branch_to_destroy = node_addr;
        goto check_instr;
       }; 
      return false;
    };
 }; // bool FS_CODE_REMOVAL::SkipVar()

inline void FS_CODE_REMOVAL::SkipCompare(const Flatterm** v1,const Flatterm** v2)
 {
  CALL("SkipCompare(const Flatterm** v1,const Flatterm** v2)"); 
  ASSERT(v1 < v2);
  constr.MakeEquivalent(VarNum(v1),VarNum(v2));
 }; // void FS_CODE_REMOVAL::SkipCompare(const Flatterm** v1,const Flatterm** v2)


inline bool FS_CODE_REMOVAL::FindSuccess()
 {
  CALL("FindSuccess()");
  const Flatterm** var1;
  const Flatterm** var2;
  ulong var_num1;
  ulong var_num2;
  
  FS_COMMAND* curr_instr;
  ASSERT(*node_addr);
  check_tag:
   curr_instr = *node_addr;
    if (curr_instr->IsCompare() || curr_instr->IsCompareNF())
     { 
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        MoveDown();
        goto check_tag;
       };
      try_fork: 
       if (curr_instr->Fork())
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_tag;
        };
       return false;  
     };
    // !(curr_instr->IsCompare() || curr_instr->IsCompareNF())
    if (curr_instr->IsSigFilter())
     { 
      node_addr = curr_instr->NextAddr();
      goto check_tag; 
     };
    
    // !(curr_instr->IsCompare() || curr_instr->IsCompareNF()) and !curr_instr->IsSigFilter()
    if (curr_instr->IsSuccess())
     {
      if (curr_instr->TheClause() == removed_clause)
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        return true;
       };
     };
    goto try_fork;
 }; // bool FS_CODE_REMOVAL::FindSuccess() 
//================================================
#undef DEBUG_NAMESPACE
//===================================================================
#endif
