#ifndef HYPER_CASCADE_H
//=======================================================
#define HYPER_CASCADE_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "iu_command.h"
#include "or_machine.h"
#include "premise_factoring.h"
#include "NewClause.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPERRESOLUTION
 #define DEBUG_NAMESPACE "SINGLE_HYPER_MACHINE"
#endif
#include "debugMacros.h"
//=================================================

class SINGLE_HYPER_MACHINE : public OR_MACH, public OR_PREMISE_WITH_FACTORING
 {
  private: // "best candidate" strategy
   bool variablesInitiallyAffected;
   bool bestCandidate;
  public:
   SINGLE_HYPER_MACHINE() 
    : OR_MACH((OR_INDEX*)0), 
      OR_PREMISE_WITH_FACTORING() 
   {
   };
   
   ~SINGLE_HYPER_MACHINE() { CALL("destructor ~SINGLE_HYPER_MACHINE()"); };

   void Destroy() 
   {
    CALL("Destroy()");
    OR_MACH::Destroy();
    OR_PREMISE_WITH_FACTORING::Destroy();
   };
   #ifndef NO_DEBUG
    bool clean() const 
    {
     return OR_MACH::clean() && OR_PREMISE_WITH_FACTORING::clean();
    };
   #endif   

   static void ResetStatic() { OR_MACH::ResetStatic(); };
   
   void SetBestCandidate(bool bc) { bestCandidate = bc; };
   bool BestCandidate() { return bestCandidate; };
   
   void assignResIndex(OR_INDEX* resIndex) { OR_MACH::assignResIndex(resIndex); };
   void assignResolvent(NewClause* res) { OR_PREMISE_WITH_FACTORING::assignResolvent(res); };

   bool Start(PREFIX_SYM* lit)
    {
     bool res = OR_MACH::CompileLiteral(lit);
     OR_MACH::MemorizeState();
     SetBestCandidate(false);
     return res; 
    };
   
   bool FindFirstLeaf()
    {
     CALL("FindFirstLeaf()");
     if (OR_MACH::FindFirstLeaf()) 
      {
       variablesInitiallyAffected = OR_MACH::VariablesFromQueryAffected();
       return true;
      };
     return false; 
    };
    
   bool FindNextLeaf()
    {
     CALL("FindNextLeaf()");
     if (OR_MACH::FindNextLeaf()) 
      {
       variablesInitiallyAffected = OR_MACH::VariablesFromQueryAffected();
       return true;
      };
     return false; 
    };
    
   bool UnitBestCandidate()
    {
     return ((!(variablesInitiallyAffected))
             && (OR_PREMISE_WITH_FACTORING::TheClause()->Unit()));
    };
  public:
   void LoadFirstCandidate()
    {
     CALL("LoadFirstCandidate()");
     OR_MACH::ResetCandidatesFromLeaf();
     OR_PREMISE_WITH_FACTORING::LoadOptimized(OR_MACH::GetIndex(),OR_MACH::NextCandidate());   
    };
   bool ChangeCandidate()
    {
     CALL("ChangeCandidate()");
     OR_PREMISE_WITH_FACTORING::Unload();
     RES_PAIR_LIST* new_candidate = OR_MACH::NextCandidate();
     if (new_candidate)
      {
       OR_PREMISE_WITH_FACTORING::LoadOptimized(OR_MACH::GetIndex(),new_candidate);
       return true;
      }
     else 
      {
       return false; 
      };
    }; // bool ChangeCandidate()
   #ifndef NO_DEBUG_VIS 
    ostream& output(ostream& str) const { return OR_MACH::output(str); };
   #endif
 }; // class SINGLE_HYPER_MACHINE

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPERRESOLUTION
 #define DEBUG_NAMESPACE "HYPER_CASCADE<ulong size>"
#endif
#include "debugMacros.h"
//=================================================
template <ulong size>
 class HYPER_CASCADE  
  {
   private:
    SINGLE_HYPER_MACHINE ORMachine[size]; 
    ulong next_machine;
    NewClause* _resolvent;
   public:
    HYPER_CASCADE(NewClause* resolvent,OR_INDEX* unifIndex) 
     : next_machine(0), _resolvent(resolvent)
     {
      ResetStatic();
      for (ulong i = 0; i < size; i++)
       {
        ORMachine[i].SetIndex(i+2);
        ORMachine[i].assignResIndex(unifIndex);
        ORMachine[i].assignResolvent(resolvent);
       };
     };
    ~HYPER_CASCADE() { CALL("destructor ~HYPER_CASCADE()"); };
    void Destroy() 
    {
     CALL("Destroy()");
     for (ulong i = 0; i < size; i++) ORMachine[i].Destroy();
     next_machine = 0;
    };
    #ifndef NO_DEBUG
     bool clean() const 
     {
      for (ulong i = 0; i < size; i++) if (!ORMachine[i].clean()) return false;
      return !next_machine;
     };
    #endif   

    static void ResetStatic() { SINGLE_HYPER_MACHINE::ResetStatic(); }; 
    bool LoadLHSLiteral(PREFIX_SYM* lhs_literal,bool wasUsed)
     {
      CALL("LoadLHSLiteral(PREFIX_SYM* lhs_literal,bool wasUsed)");
      ASSERT(next_machine < size);
      #ifdef _SUPPRESS_WARNINGS_
       wasUsed = false;
      #endif
      SINGLE_HYPER_MACHINE* machine = &(ORMachine[next_machine]);
      if ((machine->Start(lhs_literal)))
       {
        if ((machine->FindFirstLeaf()))
         {
          next_machine++;
          return true;
         };
       };
      machine->RestoreState();
      return false;
     }; // bool LoadLHSLiteral(PREFIX_SYM* lhs_literal,bool wasUsed)

   bool Backtrack(ulong& unloaded)
    {
     CALL("Backtrack(ulong& unloaded)");
     ASSERT(next_machine && (next_machine <= size));
     unloaded = 0;
     SINGLE_HYPER_MACHINE* machine = &(ORMachine[next_machine - 1]);
     
     try_next_leaf:
      
      if ((!(machine->BestCandidate())) && (machine->FindNextLeaf()))
       {        
        next_machine = (ulong)(machine - ORMachine) + 1;
        return true;
       }
      else 
       {
        machine->RestoreState();
        unloaded++;
        if (machine == ORMachine) // no more resolvents
         {
          next_machine = 0;
          return false;
         }
        else
         {
          machine--;          
          goto try_next_leaf;
         };
       };
    }; // bool Backtrack(ulong& unloaded) 
    
   void LoadFirstCandidates()
    {
     CALL("LoadFirstCandidates()");
     ASSERT(next_machine <= size);
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       machine->LoadFirstCandidate();
       machine->SetBestCandidate(machine->UnitBestCandidate());
      };
    }; // void LoadFirstCandidates()
    
   bool ChangeCandidates()
    {
     CALL("ChangeCandidates()");
     ASSERT(next_machine <= size);
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     SINGLE_HYPER_MACHINE* machine = lastMachine;
     while (machine >= ORMachine)
      {
       if ((!(machine->BestCandidate())) && (machine->ChangeCandidate()))
        {
         machine->SetBestCandidate(machine->UnitBestCandidate());
         machine++;
         while (machine <= lastMachine)
          {
           if (!(machine->BestCandidate()))
            {
             machine->LoadFirstCandidate();
            };
           machine++;
          };
         return true;
        };
       machine--;
      };
     return false;
    }; // bool ChangeCandidates()
    
   void ResetFactoring() 
    {
     CALL("ResetFactoring()");
     ASSERT(next_machine <= size);
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       if (!(machine->BestCandidate()))
        { 
         machine->ResetFactoring();
        };
      };
    }; // void ResetFactoring()  
    
   bool NextFactor() 
    {
     CALL("NextFactor()");
     ASSERT(next_machine <= size);
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       if (!(machine->BestCandidate()))
        {
         if (machine->WasFactored())
          {
           if (machine->GetNextFactor())
            {
             return true;
            }
           else
            {
             machine->ResetFactoring();
            };
          }
         else 
          {
           if (machine->GetFirstFactor(0))
            {
             return true;
            };
          };
        };
      };
     return false; // no more factors 
    }; // bool NextFactor() 
    
   bool CollectLiterals() 
    {
     CALL("CollectLiterals()");
     ASSERT(next_machine <= size);
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       if (!machine->TheClause()->Active()) return false;      
       if (!(machine->CollectNonmarkedLiterals()))
        {
         return false;
        };
      };
     return true;
    }; // bool CollectLiterals()
   void CollectAncestors() 
    {
     CALL("CollectAncestors()");
     ASSERT(next_machine <= size);
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       _resolvent->PushAncestor((Clause*)machine->TheClause());
      };
    }; // void CollectAncestors() 
  }; // class HYPER_CASCADE<ulong size>
//================================================
#undef DEBUG_NAMESPACE
//=======================================================
#endif
