#include "hyperresolution.h"

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPERRESOLUTION
 #define DEBUG_NAMESPACE "HYPERRESOLUTION"
#endif
#include "debugMacros.h"
//=================================================
bool HYPERRESOLUTION::NextInference()
 {
  CALL("NextInference()");
  switch (_continuation) 
   {
    case ContNewRule: goto process_new_rule;
    case ContShortLHSNextFactor: goto short_lhs_next_factor; 
    case ContLongLHSNextFactor: goto long_lhs_next_factor;     
   };

  process_new_rule:
   COP("process_new_rule");
   if (CurrentRule()->LongLHS()) 
    {
     COP("process_new_rule : long LHS");
     _continuation = ContLongLHSNextFactor; 
     find_leaves:
      if (FindLeaves())
       {          
        process_leaves:
	 // process current leaves
         COP("process_new_rule : long LHS : cascade.LoadFirstCandidates()");
         cascade.LoadFirstCandidates();
         long_lhs_process_candidates:
	  // process current candidates
          // first resolvent without factoring
	  if (NewResolvent())
	   {
            ResetFactoring(); 
            ASSERT(_continuation == ContLongLHSNextFactor);
            return true; 
           };
          ResetFactoring(); 
          // now build all factors 
 	  long_lhs_next_factor:
           if (NextFactor()) 
            {
             if (NewResolvent())
	      {
               ASSERT(_continuation == ContLongLHSNextFactor);
               return true; 
              };
             goto long_lhs_next_factor;
            }; 
           // try other candidates from the same leaves 
           COP("process_new_rule : long LHS : cascade.ChangeCandidates()");
           if (cascade.ChangeCandidates()) goto long_lhs_process_candidates; 
           // try to change leaves
           if (ChangeLeaves()) goto process_leaves;
        };
       if (CurrentRule()->NextApplication()) goto find_leaves;         
    }
   else // rule with 1 lhs literal
    {
     COP("process_new_rule : short LHS");
     _continuation = ContShortLHSNextFactor; 
     short_lhs_process_candidates:
      // process current candidates
      // first resolvent without factoring
      if (NewResolvent())
       {
        ResetFactoring();
        ASSERT(_continuation == ContShortLHSNextFactor);
        return true;
       };
      ResetFactoring(); 
      // now build all factors
      short_lhs_next_factor:      
       if (NextFactor()) 
        {
         if (NewResolvent())
          {
           ASSERT(_continuation == ContShortLHSNextFactor);
           return true;
          };
         goto short_lhs_next_factor;
        };       
       if (CurrentRule()->NextApplication()) goto short_lhs_process_candidates;  
    };
   // try another rule 
   COP("try another rule");
   if (FindNextRule()) goto process_new_rule; 
   return false;       
 }; // bool HYPERRESOLUTION::NextInference()


bool HYPERRESOLUTION::FindLeaves()
 {
  CALL("FindLeaves()");
  HYPER_ACTIVE_RULE::ResetIterOverLHS();
  
  bool wasUsed;
  PREFIX_SYM* lhs_lit = CurrentRule()->NextLHSLiteral(wasUsed);
    
  unsigned long unloaded;
  if (cascade.LoadLHSLiteral(lhs_lit,wasUsed))
   {
    try_next_lhs_literal:
     if ((lhs_lit = CurrentRule()->NextLHSLiteral(wasUsed)))
      {
       if (cascade.LoadLHSLiteral(lhs_lit,wasUsed))
        {
         goto try_next_lhs_literal;
        }
       else // try to backtrack
        {
         HYPER_ACTIVE_RULE::Rollback();
         
         if (cascade.Backtrack(unloaded))
          {          
           HYPER_ACTIVE_RULE::Rollback(unloaded);
           goto try_next_lhs_literal;
          }
         else // no resolvents
          {
           HYPER_ACTIVE_RULE::Rollback(unloaded);
           return false;
          };
        };
      }
     else // all lhs literals resolved
      {
       return true;
      };
   }
  else
   {
    HYPER_ACTIVE_RULE::Rollback();
    return false; // no resolvents
   };
 }; // bool HYPERRESOLUTION::FindLeaves()

bool HYPERRESOLUTION::ChangeLeaves() 
 {
  unsigned long unloaded;
  PREFIX_SYM* lhs_lit;
  bool wasUsed;
  backtrack:
   if (cascade.Backtrack(unloaded))
    {
     HYPER_ACTIVE_RULE::Rollback(unloaded);
     try_next_lhs_literal:
      if ((lhs_lit = CurrentRule()->NextLHSLiteral(wasUsed)))
       {
        if (cascade.LoadLHSLiteral(lhs_lit,wasUsed))
         {
          goto try_next_lhs_literal;
         }
        else
         {
          HYPER_ACTIVE_RULE::Rollback();
          goto backtrack;
         };
       }
      else // all lhs literals resolved
       {
        return true;
       };
    }
   else // no more leaves
    {
     HYPER_ACTIVE_RULE::Rollback(unloaded);
     return false;
    };
 }; // bool HYPERRESOLUTION::ChangeLeaves()
 
void HYPERRESOLUTION::CollectAncestors()
 {
   CALL("CollectAncestors()");
  _resolvent->UsedRule(ClauseBackground::Hyper);
  COP("CURR FACT");
  _resolvent->PushAncestor(HYPER_ACTIVE_RULE::FactClause());
  COP("RULE");
  _resolvent->PushAncestor(CurrentRule()->TheClause());
  COP("AUX FACTS");
  cascade.CollectAncestors();
 }; // void HYPERRESOLUTION::CollectAncestors() 

void HYPERRESOLUTION::ResetFactoring()
 {
  CurrentRule()->ResetFactoring();
  cascade.ResetFactoring();
 }; // void HYPERRESOLUTION::ResetFactoring()
 
bool HYPERRESOLUTION::NextFactor()
 {
  if (CurrentRule()->NextFactor()) // factoring the current fact
   {
    return true;
   }
  else
   {
    CurrentRule()->ResetFactoring();   
    return cascade.NextFactor();
   };
 }; // bool HYPERRESOLUTION::NextFactor()


bool HYPERRESOLUTION::NewResolvent()
 {
  CALL("NewResolvent()");
  _resolvent->Reset();

  if ((CurrentRule()->CollectLiterals(_resolvent)) && (cascade.CollectLiterals()))
   {
    // resolvent collected       
    CollectAncestors();            
    //_resolvent->ProcessCollected(); 
    return true;     
   };
  return false; 
 }; // void HYPERRESOLUTION::NewResolvent()

//=================================================================



