#ifndef HYPERRESOLUTION_H
//==============================================================
#define HYPERRESOLUTION_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "hyper_rule.h"
#include "clause.h" 
#include "hyper_cascade.h"
#include "NewClause.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPERRESOLUTION
 #define DEBUG_NAMESPACE "HYPER_ALL_RULES"
#endif
#include "debugMacros.h"
//=================================================

class HYPER_ALL_RULES
 {
  private:
   HYPER_ACTIVE_RULE* newRule;
   Clause::List* rules;
   Clause::List* unitRules;
   bool unit_conflict;
  private:
   Clause::List* restOfRules;
   Clause::List* moreRules;
   HYPER_ACTIVE_RULE* currentRule;
   HYPER_ACTIVE_RULE* unitConflictRule;
  public:
   HYPER_ALL_RULES() 
   : newRule(new HYPER_ACTIVE_RULE()),
     rules((Clause::List*)0),
     unitRules((Clause::List*)0),
     unit_conflict(true) // default
    {
     HYPER_RULE::SetUnitConflict(unit_conflict);
    };
   ~HYPER_ALL_RULES() { CALL("destructor ~HYPER_ALL_RULES()"); };
   #ifndef NO_DEBUG 
    bool clean() const { return (!rules) && (!unitRules); };
   #endif
   void ResetRules()
    {
     Clause::List* tmp = rules;
     while (tmp)
      {
       rules = rules->tl();
       
       delete ((HYPER_ACTIVE_RULE*)(tmp->hd()));
       delete tmp;
       tmp = rules;
      };
     tmp = unitRules;
     while (tmp)
      {
       unitRules = unitRules->tl();
       delete ((HYPER_ACTIVE_RULE*)(tmp->hd()));
       delete tmp;
       tmp = unitRules;
      };
    };
   void SetUnitConflict(bool uc) 
    {
     unit_conflict = uc; 
     HYPER_RULE::SetUnitConflict(uc);
    };
   bool UnitConflict() { return unit_conflict; };

   bool IsARule(const Clause* c) const { return HYPER_RULE::IsARule(c); };  
 
   void NewRule(Clause* c,bool compile)
    {  
     CALL("NewRule(Clause* c,bool compile)");
     ASSERT(IsARule(c));
     newRule->MakeRule(c,compile);
       
     if ((unit_conflict) && (newRule->Unit()))
      {
       unitRules = new Clause::List((Clause*)newRule,unitRules);
      }
     else
      {
       rules = new Clause::List((Clause*)newRule,rules);
      };
     
     newRule = new HYPER_ACTIVE_RULE();
    }; // void NewRule(Clause* c,bool compile)

   HYPER_ACTIVE_RULE* CurrentRule() const { return currentRule; };
   HYPER_ACTIVE_RULE* UnitConflictRule() { return unitConflictRule; };
   bool FindNextRule()
    {
     continue_search:
      while (restOfRules)
       {
        currentRule = (HYPER_ACTIVE_RULE*)(restOfRules->hd());
        restOfRules = restOfRules->tl();
        if (currentRule->TryToApply())
         {
          return true;
         };
       };
      if (moreRules)
       {
        restOfRules = moreRules;
        moreRules = (Clause::List*)0;
        goto continue_search;
       };
      currentRule = (HYPER_ACTIVE_RULE*)0; // for debugging
      return false;
    };
   bool ResetSearch(bool units_eligible)
    {
     restOfRules = rules; 
     if (units_eligible)
      {
       moreRules = unitRules;
      }
     else
      {
       moreRules = (Clause::List*)0;
      };
     currentRule = (HYPER_ACTIVE_RULE*)0;
     return FindNextRule();    
    };
   Clause* CheckForUnitConflict(Clause* suspect)
    {
     HYPER_ACTIVE_RULE::SetUnitConflictSuspect(suspect);
     Clause::List* restOfUnitRules = unitRules;
     while (restOfUnitRules)
      {
       unitConflictRule = (HYPER_ACTIVE_RULE*)(restOfUnitRules->hd());
       restOfUnitRules = restOfUnitRules->tl();
       if (unitConflictRule->Conflict())
        {        
         HYPER_ACTIVE_RULE::UnloadUnitConflictSuspect(); 
         return (unitConflictRule->TheClause());
        };              
      };
     HYPER_ACTIVE_RULE::UnloadUnitConflictSuspect();
     return (Clause*)0;
    };
  public: // for debugging
   #ifndef NO_DEBUG_VIS
    ostream& outputRules(ostream& str) const
    {
     if (unit_conflict) { str << "%======================== Nonunit rules: ======================\n"; }     
     else str << "%========================== All rules: ========================\n";
     Clause::List* rs = rules;
     while (rs)
      { 
       ((HYPER_ACTIVE_RULE*)(rs->hd()))->output(str) << '\n';
       rs = rs->tl();
      };
     if (unit_conflict)
      {
       str << "%=========================== Unit rules: ========================\n";
       rs = unitRules;
       while (rs)
        { 
         ((HYPER_ACTIVE_RULE*)(rs->hd()))->output(str) << '\n';
         rs = rs->tl();
        };
      };   
     return str <<  "%======================== End of rules: =====================\n";
    };
   #endif
 }; // class HYPER_ALL_RULES

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPERRESOLUTION
 #define DEBUG_NAMESPACE "HYPERRESOLUTION"
#endif
#include "debugMacros.h"
//=================================================

class HYPERRESOLUTION : public HYPER_ALL_RULES
 {  
  private: 
   enum Continuation 
    { 
     ContNewRule,
     ContShortLHSNextFactor,
     ContLongLHSNextFactor 
    };
  private:
   HYPER_CASCADE<HellConst::MaxNumOfLiterals-1> cascade;
   Clause* the_current_fact;
   NewClause* _resolvent;
   OR_INDEX* _unifIndex; 
   Continuation _continuation;
  private:
   bool FindLeaves();
   bool ChangeLeaves();
   bool NewResolvent();
   void ResetFactoring();
   bool NextFactor();
  public:
   HYPERRESOLUTION(NewClause* resolvent,OR_INDEX* unifIndex) 
    : HYPER_ALL_RULES(),
      cascade(resolvent,unifIndex),
      _resolvent(resolvent),
      _unifIndex(unifIndex)
    {
     HYPER_ACTIVE_RULE::Initialise(resolvent);
    };
  
   ~HYPERRESOLUTION() { CALL("destructor ~HYPERRESOLUTION()"); };      

   void Destroy() 
   {
    CALL("Destroy()");
    HYPER_ALL_RULES::ResetRules();
    cascade.Destroy(); 
   };
   #ifndef NO_DEBUG
    bool clean() const { return HYPER_ALL_RULES::clean() && cascade.clean(); };
   #endif

   static void SetDinamicLHSOrdering(bool flag) 
    { HYPER_ACTIVE_RULE::SetDinamicLHSOrdering(flag); };
   void SetUnitConfict(bool uc) 
    { HYPER_ALL_RULES::SetUnitConflict(uc); };
   static void EmptyRHSOnly() { HYPER_ACTIVE_RULE::EmptyRHSOnly(); };
   void Orient(bool hyper_positive) 
    { HYPER_RULE::PositiveHyper(hyper_positive); };
   bool Load(Clause* fact) // "= false" means that no rule applicable
    {
     // reset everything here
     the_current_fact = fact;
     HYPER_ACTIVE_RULE::SetCurrentFact(fact);
     _continuation = ContNewRule;
     return ResetSearch( (!(UnitConflict())) || (!(fact->Unit()))); 
    }; 
 
   bool NextInference();
   // void ProcessRule();    

   void CollectAncestors(); 
    
   bool FindNextRule()
    {
     if (HYPER_ALL_RULES::FindNextRule())
      {
       return true;
      }
     else
      {
       HYPER_ACTIVE_RULE::UnloadCurrentFact();
       return false;
      };
    }; // bool FindNextRule()

   Clause* UnitConflictTest(Clause* const c)
    {
     if ((c) 
         && (UnitConflict()) 
         && (c->Unit()))
       { return CheckForUnitConflict(c); }
     else { return (Clause*)0; };
    };

   #ifndef NO_DEBUG_VIS
    ostream& outputState(ostream& str) const
    { 
     str << "%===================== Hyperresolution state: =================\n";
     str << ((HYPER_RULE::isPositiveHyper()) ? "Positive" : "Negative") << " hyperresolution,\n";
     str << "Current rule: " << CurrentRule() << '\n';
     str << "Current fact: " << HYPER_ACTIVE_RULE::FactClause() << '\n';
     str << "Current substitution:\n";
     GlobalSubst.output(str,PREFIX::outputTerm) << '\n';
     return str << "%============== End of hyperresolution state: ===============\n";
    };
   #endif
 }; // class HYPERRESOLUTION 
//================================================
#undef DEBUG_NAMESPACE
//==============================================================
#endif 
 
 
 
 
 
 
 
 
