#ifndef IU_COMMAND_H
//===========================================================
#define IU_COMMAND_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "prefix.h"
#include "variables.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_IU_COMMAND
 #define DEBUG_NAMESPACE "IU_COMMAND"
#endif
#include "debugMacros.h"
//=================================================
class IU_COMMAND
 {
  public: 
   enum TAG
    {
     START,
     VAR,
     VAR_FIRST_OCC,
     FUNC,
     FUNC_PLAIN,
     FUNC_GROUND,
     CONST,
     END
    };
  private: // structure of a command
   TAG tag;
   PREF_SYM sym;
   PREFIX_SYM* term;
   IU_COMMAND* nextCommand;
  public:
   IU_COMMAND() {};
   TAG& Tag() { return tag; };
   const TAG& Tag() const { return tag; };
   const PREF_SYM& Symbol() const { return sym; };
   PREF_SYM& Symbol() { return sym; };
   PREFIX_SYM*& Term() { return term; };
   const PREFIX_SYM* const & Term() const { return term; };
   IU_COMMAND*& NextCommand() { return nextCommand; };
   const IU_COMMAND* const & NextCommand() const { return nextCommand; };
   IU_COMMAND*& AllVars() 
    {
     CALL("AllVars()");
     ASSERT(Tag() == START);
     return nextCommand; 
    };
   const IU_COMMAND* const & AllVars() const 
    {
     CALL("AllVars()");
     ASSERT(Tag() == START);
     return nextCommand; 
    };
   IU_COMMAND*& NextVar() 
    {
     CALL("NextVar()");
     ASSERT(Tag() == VAR_FIRST_OCC);
     return nextCommand; 
    };
   const IU_COMMAND* const & NextVar() const
    {
     CALL("NextVar()");
     ASSERT(Tag() == VAR_FIRST_OCC);
     return nextCommand; 
    }; 
   void Inversepolarity() // applicable to a START command only
    {
     CALL("Inversepolarity()");
     ASSERT(Tag() == START);
     sym.Inversepolarity();
    };
  public: // for debugging 
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str,const IU_COMMAND* start) const;
    ostream& outputCode(ostream& str) const; 
   #endif
 }; // class IU_COMMAND

#ifndef NO_DEBUG_VIS
 inline ostream& operator<<(ostream& str,IU_COMMAND::TAG tag)
 {
  switch (tag)
   {
    case IU_COMMAND::START: return str << "START";
    case IU_COMMAND::VAR: return str << "VAR"; 
    case IU_COMMAND::VAR_FIRST_OCC: return str << "VAR_FIRST_OCC"; 
    case IU_COMMAND::FUNC: return str << "FUNC"; 
    case IU_COMMAND::FUNC_PLAIN: return str << "FUNC_PLAIN"; 
    case IU_COMMAND::FUNC_GROUND: return str << "FUNC_GROUND"; 
    case IU_COMMAND::CONST: return str << "CONST";
    case IU_COMMAND::END: return str << "END"; 
    default: return str << "UNKNOWN IU_COMMAND"; 
   };
 };

 inline ostream& IU_COMMAND::output(ostream& str,const IU_COMMAND* start) const
 {
  str << '[' << this - start << "] " << Tag();
  if (Tag() == END) return str;
  str << ' ';
  Symbol().output(str); 

  if ((Tag() == FUNC) || (Tag() == FUNC_GROUND) || (Tag() == FUNC_PLAIN))
   {
    str << " next = [" << NextCommand() - start << "] ";
    PREFIX::outputTerm(str,Term());
   };
  if (Tag() == START) 
   {
    str << " variables = [" << (AllVars() ? (AllVars() - start) : 0) << "] ";
    PREFIX::outputTerm(str,Term());
   }; 
  if (Tag() == VAR_FIRST_OCC) 
   str << " next variable = [" << (NextVar() ? (NextVar() - start) : 0) << ']';
  return str;
 }; // ostream& IU_COMMAND::output(ostream& str,const IU_COMMAND* start) const

 inline ostream& IU_COMMAND::outputCode(ostream& str) const
 {
  const IU_COMMAND* com;
  for (com = this; com->Tag() != END; com++)
   com->output(str,this) << '\n';
  return com->output(str,this) << '\n';
 }; // ostream& IU_COMMAND::outputCode(ostream& str) const 
#endif

//================================================
#undef DEBUG_NAMESPACE
//===========================================================
#endif
