#ifndef IU_MACHINE_H
//==============================================================
#define IU_MACHINE_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "iu_command.h"
#include "iu_translator.h"
#include "unification.h"
#include "prefix.h"
#include "variables.h"
#include "or_index.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "IU_MACH"
#endif
#include "debugMacros.h"
//=================================================
class IU_MACH
 {
  private:
   enum BACKTRACK_ACTION
    {
     FAILURE,
     ALTER_BINDING_WITH_OC,
     ALTER_BINDING_WITH_OC_SAME_VAR,
     ALTER_BINDING,
     ALTER_UNIFIABLE,
     TRY_ALTERNATIVE_NODE,
     UNBIND_LAST_TRY_ALTERNATIVE_NODE,
     UNBIND_SECTION_TRY_ALTERNATIVE_NODE
    };
  private:
   IU_TRANSLATOR compiler;
   OR_PROCESSOR proc;
   IU_COMMAND code[HellConst::MaxTmpTermSize];
   IU_COMMAND* end_of_code;
   PREFIX_SYM data[HellConst::MaxTmpTermSize];
   IU_COMMAND* backtrack_point[HellConst::MaxTmpTermSize];
   BACKTRACK_ACTION backtrack_action[HellConst::MaxTmpTermSize];
   IU_COMMAND** next_backtrack_point;
   BACKTRACK_ACTION* next_backtrack_action;
   IU_COMMAND* command;
  private: // "sugar"
   VAR*** subst_state;
   RES_PAIR_LIST* candidates;
  public:
   IU_MACH() : compiler(), proc() 
    {
     ResetStatic();
     next_backtrack_point = backtrack_point;
    };
   ~IU_MACH() { CALL("destructor ~IU_MACH()"); };
   void Destroy() 
   {
    CALL("Destroy()");
    candidates = (RES_PAIR_LIST*)0;
   };
   #ifndef NO_DEBUG
    bool clean() const 
    {
     return !candidates;
    };
   #endif       


   static void ResetStatic() { OR_PROCESSOR::ResetStatic(); };
   void SetIndex(ulong index) { proc.SetIndex(index); };
   ulong GetIndex() const { return proc.GetIndex(); };
   IU_COMMAND* Code() { return code; };
   bool CompileLiteral(PREFIX_SYM* lit)
    {
     CALL("CompileLiteral(PREFIX_SYM* lit)");
     end_of_code = code;
     PREFIX_SYM* dmem = data;
     bool res = compiler.CompileLiteral(lit,end_of_code,dmem,HellConst::MaxTmpTermSize);
     
     ResetBacktracking();
     return res; 
     //candidates = (RES_PAIR_LIST*)0;
    };
   
   void CompilePseudoLiteral(TERM header,PREFIX_SYM* arg)
    {
     end_of_code = code;
     PREFIX_SYM* dmem = data;
     compiler.CompilePseudoLiteral(header,arg,end_of_code,dmem);
     ResetBacktracking(); 
     //candidates = (RES_PAIR_LIST*)0;
    };
    
   void Inversepolarity()
    {
     code->Inversepolarity();
    };
  private:
   void ResetBacktracking()
    {
     next_backtrack_point = backtrack_point;
     next_backtrack_action = backtrack_action;
    };
   void PushBacktrackPoint(IU_COMMAND* const com)
    {
     CALL("PushBacktrackPoint(IU_COMMAND* const com)");
     *next_backtrack_point = com; 
     next_backtrack_point++;
     ASSERT(next_backtrack_point - backtrack_point <= (long)HellConst::MaxTmpTermSize);
    };
   IU_COMMAND* PopBacktrackPoint() 
    {
     CALL("PopBacktrackPoint()");
     next_backtrack_point--;
     ASSERT(next_backtrack_point >= backtrack_point);
     return *next_backtrack_point;
    }; 
   void PushBacktrackAction(const BACKTRACK_ACTION act)
    {
     *next_backtrack_action = act;
     next_backtrack_action++;
    };
   BACKTRACK_ACTION PopBacktrackAction()
    {
     CALL("PopBacktrackAction()");
     ASSERT(next_backtrack_action > backtrack_action);
     next_backtrack_action--;
     return *next_backtrack_action;
    };
  private:
   bool Backtrack();
   bool CompleteSearch();
  public:
   bool FindFirstLeaf(OR_INDEX_TREE_NODE* tree);
   bool FindNextLeaf();
   
  public: // "sugar"
  
   bool VariablesFromQueryAffected() 
    {
     IU_COMMAND* vars = code->AllVars();
     while (vars)
      {
       if (vars->Symbol().var()->Binding()) { return true; };
       vars = vars->NextVar();
      };
     return false;
    };
    
 
   bool VariablesFromIndexAffected()
    {
     return GlobalSubst.VariablesAffected(GetIndex());
    };

   void MemorizeState() { subst_state = GlobalSubst.MarkState(); };
   void RestoreState() { GlobalSubst.UnbindDownto(subst_state); };
   OR_INDEX_TREE_NODE* CurrentLeaf() const { return proc.CurrNode(); };
   void ResetCandidatesFromLeaf() 
    {      
     candidates = CurrentLeaf()->Clauses();
    };
   RES_PAIR_LIST* NextCandidate() 
    {
     RES_PAIR_LIST* res = candidates;
     if (candidates) { candidates = candidates->Next(); };
     return res;
    };
  public: // methods for debugging
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
     str << "CODE:\n";
     return outputCode(str) << "INDEX = " << GetIndex() << '\n';
    };
    ostream& outputCode(ostream& str) const
    {
     const IU_COMMAND* com;
     for (com = code; com->Tag() != IU_COMMAND::END; com++) com->output(str,code) << '\n';
     return com->output(str,code) << '\n';
    };
   #endif
 }; // class IU_MACH


//****************** Definitions: *************************************

inline bool IU_MACH::CompleteSearch()
 {
  CALL("CompleteSearch()"); 
 
  IU_COMMAND* com = command;
  bool same_var;
  bool free_var;
  void* bind;
  
  
  check_tag:
  
   ASSERT((com > code) && (com < end_of_code));
   ASSERT((com->Tag() == IU_COMMAND::END) || (!(proc.CurrNode()->IsLeaf())));
  
   
   switch (com->Tag())
    {
     case IU_COMMAND::START : 
      ICP("R0");
      return false;
     case IU_COMMAND::VAR :
      bind = com->Symbol().var()->Unref(free_var);
      if (free_var)
       {
        if (proc.BindWithOccCheck((VAR*)bind,same_var))
         {
          PushBacktrackPoint(com);
          if (same_var) { PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR); }
          else { PushBacktrackAction(ALTER_BINDING_WITH_OC); };
          com++;
          
          goto check_tag;  
         }
        else { goto backtrack; };
       }
      else
       {
       
        ASSERT(PREFIX::CorrectComplexTerm((PREFIX_SYM*)bind));
       
        if (proc.ExtractUnifiable((PREFIX_SYM*)bind))
         {
          PushBacktrackPoint(com);
          PushBacktrackAction(ALTER_UNIFIABLE);
          com++;
          
          goto check_tag;
         }
        else 
         {
          goto backtrack; 
         };
       };
      
     case IU_COMMAND::VAR_FIRST_OCC :
      proc.Bind(com->Symbol().var());
      PushBacktrackPoint(com);
      PushBacktrackAction(ALTER_BINDING);
      com++;
      
      goto check_tag;  
   
     case IU_COMMAND::FUNC :
      func_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            
            goto check_tag;
           }
          else
           {
            func_try_alternative:
             proc.Alternative();
             if (proc.CurrNode()) { goto func_check_node;}
             else { goto backtrack; };
           };
         case OR_INDEX_TREE_NODE::CONST : goto func_try_alternative;
         case OR_INDEX_TREE_NODE::VAR :
          bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
          if (free_var)
           {
            if (OCCUR_CHECK::Occurs((VAR*)bind,com->Term()))
             {
              goto func_try_alternative;
             }
            else // assignment possible
             {
              GlobalSubst.BindAlone((VAR*)bind,com->Term());
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              
              goto check_tag;
             };
           }
          else // complex term, unification needed
           {
            if (UNIFICATION::UnifyComplex((PREFIX_SYM*)bind,com->Term()))
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              
              goto check_tag;
             }
            else { goto func_try_alternative; };
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC :
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com = com->NextCommand();
          proc.Below();
          
          goto check_tag;
         case OR_INDEX_TREE_NODE::LEAF :
	  ICP("R1");
          return false;
          
         #ifdef DEBUG_NAMESPACE
         default : 
          ICP("A1");
          return false;    
         #endif
        };
     
     case IU_COMMAND::FUNC_PLAIN :
      func_plain_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            
            goto check_tag;
           }
          else
           {
            func_plain_try_alternative:
             proc.Alternative();
             if (proc.CurrNode()) { goto func_plain_check_node;}
             else { goto backtrack; };
           };
         case OR_INDEX_TREE_NODE::CONST : goto func_plain_try_alternative;
         case OR_INDEX_TREE_NODE::VAR :
          bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
          if (free_var)
           {
            GlobalSubst.BindAlone((VAR*)bind,com->Term());
            PushBacktrackPoint(com);
            PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
            com = com->NextCommand();
            proc.Below();
            
            goto check_tag;
           }
          else // complex term
           {
            if (UNIFICATION::UnifyPlain((PREFIX_SYM*)bind,com->Term()))
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              goto check_tag;
             }
            else { goto func_plain_try_alternative; }; 
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com = com->NextCommand();
          proc.Below();
          
          goto check_tag;
         case OR_INDEX_TREE_NODE::LEAF :
          ICP("S1");
          return false;
         #ifdef DEBUG_NAMESPACE
         default : 
          ICP("L1");
          return false;    
         #endif
        }; 
      
      /************/
      
     case IU_COMMAND::FUNC_GROUND :
      func_ground_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            goto check_tag;
           }
          else
           {
            func_ground_try_alternative:
             proc.Alternative();
             if (proc.CurrNode()) { goto func_ground_check_node;}
             else { goto backtrack; };
           };
         case OR_INDEX_TREE_NODE::CONST : goto func_ground_try_alternative;
         case OR_INDEX_TREE_NODE::VAR :
          bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
          if (free_var)
           {
            GlobalSubst.BindAlone((VAR*)bind,com->Term());
            PushBacktrackPoint(com);
            PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
            com = com->NextCommand();
            proc.Below();
            goto check_tag;
           }
          else // complex term
           {
            if (UNIFICATION::MatchGroundPlain((PREFIX_SYM*)bind,com->Term()))
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              goto check_tag;
             }
            else { goto func_ground_try_alternative; }; 
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com = com->NextCommand();
          proc.Below();
          goto check_tag;
         case OR_INDEX_TREE_NODE::LEAF :
          ICP("R2");
          return false;
         #ifdef DEBUG_NAMESPACE
         default : 
          ICP("A2");
          return false; 
         #endif
        };
     
     case IU_COMMAND::CONST :
      const_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC :
          const_try_alternative:
           proc.Alternative();
           if (proc.CurrNode()) { goto const_check_node;}
           else { goto backtrack; };
         case OR_INDEX_TREE_NODE::CONST : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            goto check_tag;
           }
          else { goto const_try_alternative; };
         case OR_INDEX_TREE_NODE::VAR : 
          bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
          if (free_var)
           {
            GlobalSubst.BindAlone((VAR*)bind,com->Term());
            PushBacktrackPoint(com);
            PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            goto check_tag;
           }
          else // complex term
           {
            if (com->Symbol() == ((PREFIX_SYM*)bind)->Head())
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(TRY_ALTERNATIVE_NODE);
              com++;
              proc.Below();
              goto check_tag;
             }
            else { goto const_try_alternative; };
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com++;
          proc.Below();
          goto check_tag;
         case OR_INDEX_TREE_NODE::LEAF :
          ICP("R3"); 
          return false;
         #ifdef DEBUG_NAMESPACE
         default : 
          ICP("A3"); 
          return false;
         #endif
        };
     case IU_COMMAND::END :
      ASSERT(proc.CurrNode()->IsLeaf());      
      return true;
     #ifdef DEBUG_NAMESPACE
     default :
      ICP("A3"); 
      return false; 
     #endif  
    };
  backtrack:
   command = com;
   if (Backtrack()) { com = command; goto check_tag; }
   else { return false; };
 }; // bool IU_MACH::CompleteSearch() 


inline bool IU_MACH::FindFirstLeaf(OR_INDEX_TREE_NODE* tree)
 {
  command = code+1;
  if (tree)
   {
    PushBacktrackPoint(code);
    PushBacktrackAction(FAILURE);
    proc.Start(tree);
    return CompleteSearch();
   }
  else 
   {
    return false; 
   };
 }; // bool IU_MACH::FindFirstLeaf()

inline bool IU_MACH::FindNextLeaf()
 {
  return ((Backtrack()) && (CompleteSearch()));
 }; // inline bool IU_MACH::FindNextLeaf()


inline bool IU_MACH::Backtrack()
 {
  CALL("Backtrack()");
 
  IU_COMMAND* com;
  bool same_var;
  VAR* var;
  
  rollback:
   com = PopBacktrackPoint();
   
   BACKTRACK_ACTION act = PopBacktrackAction(); // debug
   
  // switch (PopBacktrackAction())
   switch (act) // debug
    {
     case FAILURE : return false;
     case ALTER_BINDING_WITH_OC : 
     
      ASSERT(com->Symbol().var()->Binding());
      if (proc.AlterBindingWithOccCheck(GlobalSubst.LastBound(),same_var))
       {
        PushBacktrackPoint(com);
        if (same_var)
         {
          PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR);
         }
        else
         {
          ASSERT(com->Symbol().var()->Binding());
          PushBacktrackAction(ALTER_BINDING_WITH_OC);
         };
        command = com + 1;
        return true;
       }
      else 
       {
        goto rollback; 
       };
     case ALTER_BINDING_WITH_OC_SAME_VAR :
      
      var = (VAR*)(com->Symbol().var()->Unref());
      ASSERT(VAR::IsVar((void*)var));
      if (proc.AlterBindingWithOccCheck(var,same_var))
       {
        PushBacktrackPoint(com);
        if (same_var)
         {
          PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR);
         }
        else
         {
          PushBacktrackAction(ALTER_BINDING_WITH_OC);
         };
        command = com + 1;
        return true;
       }
      else 
       {
        goto rollback; 
       };
      
     case ALTER_BINDING : 
     
      ASSERT(com->Tag() == IU_COMMAND::VAR_FIRST_OCC);
      ASSERT(com->Symbol().var()->Binding());
      
      if (proc.AlterBinding())
       {
        ASSERT(com->Symbol().var()->Binding());
        PushBacktrackPoint(com);
        PushBacktrackAction(ALTER_BINDING);
        command = com + 1;
        return true;
       }
      else 
       {
        goto rollback; 
       };
      
     case ALTER_UNIFIABLE :
     
      if (proc.AlterUnifiable())
       {
        
        PushBacktrackPoint(com);
        PushBacktrackAction(ALTER_UNIFIABLE);
        command = com + 1;
        return true;
       }
      else 
       {
        goto rollback; 
       };
      
     case TRY_ALTERNATIVE_NODE :
      
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
       {
        goto rollback; 
       };
      
     case UNBIND_LAST_TRY_ALTERNATIVE_NODE :
     
      GlobalSubst.UnbindLast();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
       {
        goto rollback; 
       };
      
     case UNBIND_SECTION_TRY_ALTERNATIVE_NODE :
     
      GlobalSubst.UnbindLastSection();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
       {
        goto rollback; 
       };
    };   
  ICP("R1");
  return false;
 };
//================================================
#undef DEBUG_NAMESPACE 
//==============================================================
#endif
