#ifndef IU_TRANSLATOR_H
//===========================================================
#define IU_TRANSLATOR_H
#include "jargon.h"
#include "HellDebugFlags.h"
#include "iu_command.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "IU_TRANSLATOR"
#endif
#include "debugMacros.h"
//=================================================
class IU_TRANSLATOR
 {
  private:
   LARGE_VAR_COUNTER var_counter;
  public: 
   IU_TRANSLATOR() : var_counter() 
   {
    CALL("constructor IU_TRANSLATOR()");
   };
   ~IU_TRANSLATOR() { CALL("destructor ~IU_TRANSLATOR()"); };
   inline bool CompileLiteral(PREFIX_SYM* word,
                              IU_COMMAND*& code_mem,
                              PREFIX_SYM*& data_mem,
                              unsigned long max_code_size);
   inline void CompilePseudoLiteral(TERM header,
                                    PREFIX_SYM* word,
                                    IU_COMMAND*& code_mem,
                                    PREFIX_SYM*& data_mem);                           
 }; // class IU_TRANSLATOR

//******************* Definitions for IU_TRANSLATOR: *********************

inline bool IU_TRANSLATOR::CompileLiteral(PREFIX_SYM* word,
                                          IU_COMMAND*& code_mem,
                                          PREFIX_SYM*& data_mem,
                                          unsigned long max_code_size)
 {
  CALL("CompileLiteral(PREFIX_SYM*,IU_COMMAND*&,PREFIX_SYM*&,unsigned long)");

  IU_COMMAND* open_bracket[HellConst::MaxTmpTermSize]; // stack
  IU_COMMAND** next_open_bracket = open_bracket+1;
  PREFIX::ITER iter;
  BracketCounter<HellConst::MaxTmpTermDepth> bracket_counter;
  iter.Reset(word);
     
  IU_COMMAND* var_list = (IU_COMMAND*)0; 
  IU_COMMAND* start = code_mem;
  PREFIX_SYM* data = data_mem;
  if (iter.CurrSubterm()->Head().Func().arity())
   bracket_counter.Open(iter.CurrSubterm()->Head().Func().arity());
  open_bracket[0] = code_mem;   
  code_mem->Tag() = IU_COMMAND::START;
  code_mem->Symbol() = word->Head();
  data_mem->Head() = word->Head();
  //code_mem->Symbol().Func().Inversepolarity();
  code_mem->Term() = data_mem;
  code_mem++;
  data_mem++;
  max_code_size--;  
     
  iter.SkipSym(); // skipping header
     
  var_counter.Reset();
     
  VAR* var;
  IU_COMMAND* last_var = code_mem-1;
  IU_COMMAND* last_var_not_fo = code_mem-1;
  unsigned closed = 0;
     
  check_iter:
   if (iter.Empty()) 
    {
     code_mem->Tag() = IU_COMMAND::END;
     code_mem++;
     max_code_size--;
     if (max_code_size <= 0) return false;
     data->SetBracket(data_mem); // needed only when the literal is propositional
     start->AllVars() = var_list;
     return true;
    };

   var = iter.CurrVar();
   if (var)
    {
     //cout << "VAR\n"; var->Output(); cout << "\n"; // debug
      
     if (var_counter.Register(var))
      {
       code_mem->Tag() = IU_COMMAND::VAR;
       last_var_not_fo = code_mem;
      }
     else // first occurence of the variable
      {
       code_mem->Tag() = IU_COMMAND::VAR_FIRST_OCC;
       code_mem->NextVar() = var_list;
       var_list = code_mem;
      };
     code_mem->Symbol().MkVar(var);
        
     last_var = code_mem;
        
     data_mem->Head().MkVar(var);
     data_mem->SetBracket(data_mem + 1);
     closed = bracket_counter.Close();
    }
   else // functor
    {
     //cout << "FUNC\n"; iter.CurrSubterm()->Head().Output(); cout << "\n"; // debug
        
     unsigned long arity = iter.CurrSubterm()->Head().Func().arity();
     if (arity)
      {
       code_mem->Tag() = IU_COMMAND::FUNC;
       *next_open_bracket = code_mem;
       next_open_bracket++;
       bracket_counter.Open(arity);
      }
     else // constant
      {
       code_mem->Tag() = IU_COMMAND::CONST;
       closed = bracket_counter.Close();
       data_mem->SetBracket(data_mem + 1);
      };
     code_mem->Symbol() = iter.CurrSubterm()->Head();
     code_mem->Term() = data_mem;
     data_mem->Head() = iter.CurrSubterm()->Head();
    };
       
   iter.SkipSym();
   code_mem++; 
   data_mem++;
   max_code_size--;
   if (max_code_size <= 0) return false;    

   while (closed)
    {
     next_open_bracket--;
     (*next_open_bracket)->NextCommand() = code_mem;
     unsigned length = code_mem - (*next_open_bracket); 
     (*next_open_bracket)->Term()->SetBracket((*next_open_bracket)->Term() + length);
     if (last_var < (*next_open_bracket)) // the term is ground
      {
       (*next_open_bracket)->Tag() = IU_COMMAND::FUNC_GROUND;
      }
     else
      {
       if (last_var_not_fo < (*next_open_bracket))
        {
         (*next_open_bracket)->Tag() = IU_COMMAND::FUNC_PLAIN;
        };
      };
     closed--;
    };
      
   goto check_iter;
 }; // bool IU_TRANSLATOR::CompileLiteral(PREFIX_SYM* word,IU_COMMAND*& code_mem,PREFIX_SYM*& data_mem,unsigned long max_code_size)


inline void IU_TRANSLATOR::CompilePseudoLiteral(TERM header,
                                                PREFIX_SYM* word,
                                                IU_COMMAND*& code_mem,
                                                PREFIX_SYM*& data_mem)
 {
   CALL("IU_TRANSLATOR::CompilePseudoLiteral(TERM,PREFIX_SYM*,IU_COMMAND*&,PREFIX_SYM*&)");
  // word must represent a complex term
  IU_COMMAND* open_bracket[HellConst::MaxTmpTermSize]; // stack
  IU_COMMAND** next_open_bracket = open_bracket+1;
  PREFIX::ITER iter;
  BracketCounter<HellConst::MaxTmpTermDepth> bracket_counter;
  iter.Reset(word);
     
  IU_COMMAND* var_list = (IU_COMMAND*)0; 
  IU_COMMAND* start = code_mem;
  PREFIX_SYM* data = data_mem;
  if (iter.CurrSubterm()->Head().Func().arity())
   bracket_counter.Open(iter.CurrSubterm()->Head().Func().arity());
  open_bracket[0] = code_mem;   
  code_mem->Tag() = IU_COMMAND::START;
  
  
  PREF_SYM pr_header(header);     // most probably unnecessary  
  code_mem->Symbol() = pr_header; // most probably unnecessary  
  data_mem->Head() = pr_header;   // most probably unnecessary        
  code_mem->Symbol().Func().Inversepolarity(); // most probably unnecessary 
 
  code_mem->Term() = data_mem;
  code_mem++;
  data_mem++;
       
  //iter.SkipSym(); // skipping header
     
  var_counter.Reset();
     
  VAR* var;
  IU_COMMAND* last_var = code_mem-1;
  IU_COMMAND* last_var_not_fo = code_mem-1;
  unsigned closed = 0;
     
  check_iter:
   if (iter.Empty()) 
    {
     code_mem->Tag() = IU_COMMAND::END;
     code_mem++;
     data->SetBracket(data_mem); // needed only when the literal is propositional
     start->AllVars() = var_list;
     return;
    };
   var = iter.CurrVar();
   if (var)
    {
     //cout << "VAR\n"; var->Output(); cout << "\n"; // debug
      
     if (var_counter.Register(var))
      {
       code_mem->Tag() = IU_COMMAND::VAR;
       last_var_not_fo = code_mem;
      }
     else // first occurence of the variable
      {
       code_mem->Tag() = IU_COMMAND::VAR_FIRST_OCC;
       code_mem->NextVar() = var_list;
       var_list = code_mem;
      };
     code_mem->Symbol().MkVar(var);
        
     last_var = code_mem;
        
     data_mem->Head().MkVar(var);
     data_mem->SetBracket(data_mem + 1);
     closed = bracket_counter.Close();
    }
   else // functor
    {
     //cout << "FUNC\n"; iter.CurrSubterm()->Head().Output(); cout << "\n"; // debug
        
     unsigned long arity = iter.CurrSubterm()->Head().Func().arity();
     if (arity)
      {
       code_mem->Tag() = IU_COMMAND::FUNC;
       *next_open_bracket = code_mem;
       next_open_bracket++;
       bracket_counter.Open(arity);
      }
     else // constant
      {
       code_mem->Tag() = IU_COMMAND::CONST;
       closed = bracket_counter.Close();
       data_mem->SetBracket(data_mem + 1);
      };
     code_mem->Symbol() = iter.CurrSubterm()->Head();
     code_mem->Term() = data_mem;
     data_mem->Head() = iter.CurrSubterm()->Head();
    };
       
   iter.SkipSym();
   code_mem++; 
   data_mem++;
      
   while (closed)
    {
     next_open_bracket--;
     (*next_open_bracket)->NextCommand() = code_mem;
     unsigned length = code_mem - (*next_open_bracket); 
     (*next_open_bracket)->Term()->SetBracket((*next_open_bracket)->Term() + length);
     if (last_var < (*next_open_bracket)) // the term is ground
      {
       (*next_open_bracket)->Tag() = IU_COMMAND::FUNC_GROUND;
      }
     else
      {
       if (last_var_not_fo < (*next_open_bracket))
        {
         (*next_open_bracket)->Tag() = IU_COMMAND::FUNC_PLAIN;
        };
      };
     closed--;
    };
      
   goto check_iter;
 }; // void IU_TRANSLATOR::CompilePseudoLiteral(TERM header,PREFIX_SYM* word,IU_COMMAND*& code_mem,PREFIX_SYM*& data_mem)
//================================================
#undef DEBUG_NAMESPACE 
//===========================================================
#endif
