#include "or_index.h"

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_INDEX"
#endif 
#include "debugMacros.h"
//================================================= 
void OR_INDEX::Remove(Clause* cl)
 {   
  LiteralList* allLiterals = cl->LitList();
  LiteralList::Element* firstNonselectedLit = allLiterals->nth(allLiterals->numOfSelectedLits());
  for(LiteralList::Iterator iter(allLiterals);iter.currentElement() != firstNonselectedLit;iter.next())
    Remove(iter.currentLiteral(),cl); 
 }; // void OR_INDEX::Remove(Clause* c)

void OR_INDEX::RemovePropLit(TERM* lit,Clause* cl)
 {
  CALL("RemovePropLit(TERM* lit,Clause* cl)");
  ASSERT(!lit->arity());
  OR_INDEX_TREE_NODE** tree = Tree(*lit);
  ASSERT(*tree);
  ASSERT((*tree)->IsLeaf());
  (*tree)->Remove(cl);
  if (!(*tree)->Clauses()) 
   {
    (*tree)->DeleteLeaf();
    *tree = (OR_INDEX_TREE_NODE*)0;
   };
 }; // void OR_INDEX::RemovePropLit(TERM* lit,Clause* cl)
 
OR_INDEX_TREE_NODE* OR_INDEX::LeafAndBranch(TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch)
 {
  CALL("LeafAndBranch(TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch)");
  ASSERT(tree);
  ASSERT(t->arity());
  branch = tree;
  OR_INDEX_TREE_NODE* node = *tree;
  
  if (!node) return (OR_INDEX_TREE_NODE*)0; 
  
  OR_INDEX_TREE_NODE** node_addr = tree; 
  
  TERM::Iterator iter;  

  iter.Reset(t);

  next_sym:
   if (!iter.Next())
    {
     ASSERT(node->IsLeaf());
     return node;
    };
   check_node:
    if (node->SameSymbol(iter.Symbol()))
     {
      if (node->Alternative()) branch = node_addr;
      node_addr = node->BelowAddr();
      node = node->Below();
      ASSERT(node);
      goto next_sym;     
     }; 
    // try alternative
    if (node->Alternative())
     {
      node_addr = node->AlternativeAddr();
      branch = node_addr;
      node = node->Alternative();
      goto check_node;
     };
    // no appropriate alternative, search fails
    return (OR_INDEX_TREE_NODE*)0;
 }; // OR_INDEX_TREE_NODE* OR_INDEX::LeafAndBranch(TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch)



void OR_INDEX::Remove(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)
 {  
  CALL("Remove(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)");
  
  ASSERT(t->arity());
  OR_INDEX_TREE_NODE** branch;
  OR_INDEX_TREE_NODE* branch_to_delete;
  OR_INDEX_TREE_NODE* node_to_delete;
  
  OR_INDEX_TREE_NODE* leaf = LeafAndBranch(t,tree,branch);

  if (!leaf) return;

  leaf->Remove(cl);

  if (!(leaf->Clauses())) 
   {
    // no more clauses in the leaf, the whole branch must be deleted
    // *branch can't be a leaf here
    branch_to_delete = *branch;
    *branch = (*branch)->Alternative(); 
    while (!(branch_to_delete->IsLeaf()))
     {
      node_to_delete = branch_to_delete;
      branch_to_delete = branch_to_delete->Below();
      node_to_delete->DeleteNode();
     };
    branch_to_delete->DeleteLeaf(); // leaf
   };  
 }; // void OR_INDEX::Remove(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)

void OR_INDEX::RemoveBPCandidates(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)
 {  
  CALL("RemoveBPCandidates(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)");
  
  ASSERT(t->arity());
  OR_INDEX_TREE_NODE** branch;
  OR_INDEX_TREE_NODE* branch_to_delete;
  OR_INDEX_TREE_NODE* node_to_delete;
  
  OR_INDEX_TREE_NODE* leaf = LeafAndBranch(t,tree,branch);

  if (!leaf) return;

  leaf->RemoveBPCandidates(cl);

  if (!(leaf->BPCandidates())) 
   {
    // no more clauses in the leaf, the whole branch must be deleted
    // *branch can't be a leaf here
    branch_to_delete = *branch;
    *branch = (*branch)->Alternative(); 
    while (!(branch_to_delete->IsLeaf()))
     {
      node_to_delete = branch_to_delete;
      branch_to_delete = branch_to_delete->Below();
      node_to_delete->DeleteNode();
     };
    branch_to_delete->DeleteLeaf(); // leaf
   };  
 }; // void OR_INDEX::RemoveBPCandidates(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)

void OR_INDEX::RemoveFromBPI(TERM* lit,Clause* cl)
 {
  TERM::Iterator iter;
  if (lit->IsEquality() && lit->IsOrderedEq())
   {
    TERM* arg = lit->Args()->First();
    if (!arg->isVariable())
     {
      iter.Reset(arg);
      do
       {
        if (!iter.CurrentSymIsVar()) RemoveBPRedex(iter.CurrPos(),cl);
       }
      while (iter.Next());
     };
   }
  else // ordinary literal or unordered equality
   {
    iter.Reset(lit);
    while (iter.Next())
     if (!iter.CurrentSymIsVar()) RemoveBPRedex(iter.CurrPos(),cl);
   };
 }; // void OR_INDEX::RemoveFromBPI(TERM* lit,Clause* cl)

void OR_INDEX::RemoveBPRedex(TERM* t,Clause* cl)
 {
  CALL("RemoveBPRedex(TERM* t,Clause* cl)");
  pseudo_literal->Args()->SetFirst(t);
  RemoveBPCandidates(pseudo_literal,cl,&backward_para_tree);
 }; // void OR_INDEX::RemoveBPRedex(TERM* t,Clause* cl)

void OR_INDEX::RemoveFromFPI(TERM* lit,Clause* cl)
 {
  bool stop = false;
  TERM* arg = lit->Args();
  remove_arg:
   if (arg->isVariable())
    {
     pseudo_literal[1] = *arg;
     Remove(pseudo_literal,cl,&forward_para_from_var_tree);
    }
   else // complex  
    if (arg->First()->arity())
     {
      Remove(arg->First(),cl,FParaTree(*(arg->First()))); 
     }
    else // constant 
     {
      OR_INDEX_TREE_NODE** tree = FParaTree(*(arg->First()));
       
      if (*tree)
       {
        (*tree)->Remove(cl);
        if (!(*tree)->Clauses()) 
         {
          (*tree)->DeleteLeaf();
          *tree = (OR_INDEX_TREE_NODE*)0;
         };
       };
     };
  if (lit->IsOrderedEq() || stop) 
   { 
    return;
   }; 
  stop = true;
  arg = arg->Next();
  goto remove_arg;          
 }; // void OR_INDEX::RemoveFromFPI(TERM* lit,Clause* cl)


void OR_INDEX::Remove(TERM* lit,Clause* cl)
 {
  if (!lit->arity()) { RemovePropLit(lit,cl); return; };
  if (paramodulation) RemoveFromBPI(lit,cl);
  if (lit->IsEquality())
   {     
    if (lit->Positive()) 
     {
      RemoveFromFPI(lit,cl);
     };
   }
  else 
   {
    Remove(lit,cl,Tree(*lit));
   };
 }; // void OR_INDEX::Remove(TERM* lit,Clause* cl)


void OR_INDEX::Destroy()
 {  
  for (ulong i = 0; i < HellConst::MaxNumOfFunctors * 2; i++)
   {
    OR_INDEX_TREE_NODE::Destroy(header_index[i]);
    header_index[i] = (OR_INDEX_TREE_NODE*)0;
   }; 
  OR_INDEX_TREE_NODE::Destroy(forward_para_from_var_tree); 
  forward_para_from_var_tree = (OR_INDEX_TREE_NODE*)0;
  OR_INDEX_TREE_NODE::DestroyBPI(backward_para_tree);
  backward_para_tree = (OR_INDEX_TREE_NODE*)0;
 }; // void OR_INDEX::Destroy()

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_INDEX_TREE_NODE"
#endif 
#include "debugMacros.h"
//================================================= 

void OR_INDEX_TREE_NODE::Destroy(OR_INDEX_TREE_NODE* tree)
 {
  CALL("Destroy(OR_INDEX_TREE_NODE* tree)");
  if (!tree) return;
  if (tree->IsLeaf())
   {
    tree->DeleteLeaf();
   }
  else
   {
    OR_INDEX_TREE_NODE* tree_below = tree->Below();
    OR_INDEX_TREE_NODE* tree_alt = tree->Alternative(); 
    tree->DeleteNode();
    Destroy(tree_below);
    if (tree_alt)
     {
      Destroy(tree_alt);
     };   
   };
 }; // void OR_INDEX_TREE_NODE::Destroy(OR_INDEX_TREE_NODE* tree)

void OR_INDEX_TREE_NODE::DestroyBPI(OR_INDEX_TREE_NODE* tree)
 {
  CALL("DestroyBPI(OR_INDEX_TREE_NODE* tree)");
  if (!tree) return;
  if (tree->IsLeaf())
   {
    tree->DeleteBPILeaf();
   }
  else
   {
    OR_INDEX_TREE_NODE* tree_below = tree->Below();
    OR_INDEX_TREE_NODE* tree_alt = tree->Alternative(); 
    tree->DeleteNode();
    DestroyBPI(tree_below);
    if (tree_alt)
     {
      DestroyBPI(tree_alt);
     };   
   };
 }; // void OR_INDEX_TREE_NODE::DestroyBPI(OR_INDEX_TREE_NODE* tree)

//=================================================
