#ifndef OR_INDEX_H
//==================================================================
#define OR_INDEX_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "GlobAlloc.h"
#include "HellConst.h"
#include "term.h"
#include "clause.h"
#include "prefix.h"
#include "variables.h"
#include "MultisetOfVariables.h"
#include "unification.h"
#include "SharedKBConstr.h"
#include "BackParaCandidates.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_INDEX"
#endif 
#include "debugMacros.h"
//=================================================

class OR_INDEX_TREE_NODE;

class OR_INDEX
 {
  public:
   OR_INDEX() : 
    l_iter(), 
    var_counter(), 
    _vars1(),
    _vars2(),
    forward_para_from_var_tree((OR_INDEX_TREE_NODE*)0),
    backward_para_tree((OR_INDEX_TREE_NODE*)0)
     {
      CALL("constructor OR_INDEX()");
      for (ulong i = 0; i < HellConst::MaxNumOfFunctors * 2; i++)
       header_index[i] = (OR_INDEX_TREE_NODE*)0;
      pseudo_literal[0] = TERM((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity);
     };
   ~OR_INDEX() { CALL("destructor ~OR_INDEX()"); };

   #ifndef NO_DEBUG
   bool clean() const 
   {
    for (ulong i = 0; i < HellConst::MaxNumOfFunctors * 2; i++)
     if (header_index[i]) return false;    
    if (forward_para_from_var_tree || backward_para_tree) return false;
    return true;
   };
   #endif

   void SetParamodulation(bool fl) { paramodulation = fl; }; 
   void Destroy();
   OR_INDEX_TREE_NODE** Tree(TERM hd) { return header_index + (hd.functor()*2) + hd.polarity(); };
   OR_INDEX_TREE_NODE** FParaTree(TERM fun) { return header_index + (2*fun.functor()); }; 
   OR_INDEX_TREE_NODE** FParaFromVarTree() { return &forward_para_from_var_tree; }; 
   OR_INDEX_TREE_NODE** BParaTree() { return &backward_para_tree; }; 
   void Integrate(Clause* c);
   void Remove(Clause* c);
   
  private:
   OR_INDEX_TREE_NODE* CreateBranch(OR_INDEX_TREE_NODE*& leaf);
   void Integrate(TERM* lit,Clause* cl,unsigned lit_num);
   OR_INDEX_TREE_NODE* IntegratePropLit(TERM lit); // returns the leaf
   OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree); // returns the leaf
   void IntegrateForBackwardPara(TERM* lit,Clause* cl,unsigned lit_num);
   void IntegrateBPRedex(TERM* redex,Clause* cl,unsigned lit_num,SharedKBConstr* cns);
   void IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num);
   void IntegratePreorderedForForwardPara(TERM* lit,Clause* cl,unsigned lit_num);   
   void IntegrateConstrainedForForwardPara(const TERM& lhs,const TERM& rhs,Clause* cl,unsigned lit_num);
   OR_INDEX_TREE_NODE* IntegrateNonpropLit(TERM* lit); // returns the leaf
   
   OR_INDEX_TREE_NODE* LeafAndBranch(TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch);
   void RemovePropLit(TERM* lit,Clause* cl);
   void Remove(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree);
   void RemoveBPCandidates(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree);
   void RemoveBPRedex(TERM* t,Clause* cl);
   void RemoveFromBPI(TERM* lit,Clause* cl);
   void RemoveFromFPI(TERM* lit,Clause* cl);
   void Remove(TERM* lit,Clause* cl);
   
   static void collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& funcs)
   {
    CALL("collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& funcs)");
    ASSERT(t->isComplex());
    vars.Reset();
    funcs = 0;
    TERM::Iterator termIter(t);       
    do 
     {
      if (termIter.CurrentSymIsVar())
       {
        vars.Register(termIter.symbolRef().var());
       } 
      else funcs++;
     }
    while (termIter.Next());  
   }; // void collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vars,ulong& funcs)

  private: // auxilliary data for maintainence operations
   TERM::IteratorWithDepth l_iter;
   MultisetOfVariables<HellConst::MaxNumOfVariables> var_counter;
   TERM pseudo_literal[2];
   MultisetOfVariables<HellConst::MaxNumOfVariables> _vars1;
   MultisetOfVariables<HellConst::MaxNumOfVariables> _vars2;
  private: // structure
   OR_INDEX_TREE_NODE* header_index[HellConst::MaxNumOfFunctors * 2];
   OR_INDEX_TREE_NODE* forward_para_from_var_tree;  
   OR_INDEX_TREE_NODE* backward_para_tree; 
   bool paramodulation;
 }; // class OR_INDEX 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "RES_PAIR_LIST"
#endif 
#include "debugMacros.h"
//=================================================
class RES_PAIR_LIST
 {
  private: // sructure
   unsigned lit_num;
   Clause* clause;
   TERM term; // for forward and backward paramodulation
   SharedKBConstr* constr; // for forward superposition   
   RES_PAIR_LIST* next_pair;
  public: // constructors and memory management
   RES_PAIR_LIST(unsigned ln,Clause* cl,RES_PAIR_LIST* np)
    : lit_num(ln), clause(cl), constr((SharedKBConstr*)0), next_pair(np) 
    {};
   RES_PAIR_LIST(unsigned ln,Clause* cl,TERM t,SharedKBConstr* cns,RES_PAIR_LIST* np) 
    // for forward paramodulation
    : lit_num(ln), 
      clause(cl), 
      term(t),
      constr(cns),
      next_pair(np) 
    {
     if (cns) cns->refCounter()++;      
    }; 
   RES_PAIR_LIST(unsigned ln,Clause* cl,TERM* t,SharedKBConstr* cns,RES_PAIR_LIST* np) 
    // for backward paramodulation
    : lit_num(ln), 
      clause(cl), 
      constr(cns),
      next_pair(np) 
    {
     term.SetFirst(t);
     if (cns) cns->refCounter()++;
    };
   ~RES_PAIR_LIST() 
   { 
    CALL("destructor ~RES_PAIR_LIST()"); 
    if (constr) 
     {
      ASSERT(constr->refCounter() > 0);
      constr->refCounter()--;
      if (!constr->refCounter()) delete constr;
     }; 
   };
   void* operator new(size_t)
    {
     return GlobAlloc::allocate(sizeof(RES_PAIR_LIST));
    };
   void operator delete(void* obj)
    {
     GlobAlloc::deallocate(obj,sizeof(RES_PAIR_LIST));
    };
  public:
   unsigned LitNum() const { return lit_num; };
   Clause* TheClause() const { return clause; };
   const SharedKBConstr* Constraint() const { return constr; };
   TERM ReplacingTerm() const { return term; }; // for forward paramodulation
   TERM* Redex() const { return term.First(); }; // for backward paramodulation
   RES_PAIR_LIST* Next() { return next_pair; };
   RES_PAIR_LIST** Tail() { return &next_pair; };
   
   
   static void Remove(Clause* c,RES_PAIR_LIST*& lst)
    {
     CALL("Remove(Clause* c,RES_PAIR_LIST*& lst)");
     RES_PAIR_LIST* pair_to_delete;
     while ((lst) && (lst->TheClause() == c))
      {
       pair_to_delete = lst;
       lst = lst->Next();
       delete pair_to_delete;
      };
     if (!lst) return;
     RES_PAIR_LIST** tail = lst->Tail();
     while (*tail) 
      {
       if ((*tail)->TheClause() == c)
        {
         pair_to_delete = *tail;
         *tail = (*tail)->Next();
         delete pair_to_delete;  
        }
       else tail = (*tail)->Tail(); 
      }; 
    }; // void Remove(Clause* const c,RES_PAIR_LIST* lst)

   static void Destroy(RES_PAIR_LIST* lst)
   {
    CALL("Destroy(RES_PAIR_LIST* lst)");
    RES_PAIR_LIST* tmp;
    while (lst) 
     {
      tmp = lst->Next();    
      delete lst;      
      lst = tmp;
     };
   }; // void Destroy(RES_PAIR_LIST* lst) 
 }; // class RES_PAIR_LIST

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_INDEX_TREE_NODE"
#endif 
#include "debugMacros.h"
//================================================= 
class OR_INDEX_TREE_NODE
 {
  public:
   enum TAG
    {
     FUNC,
     CONST,
     VAR,
     VAR_FIRST_OCC,
     LEAF
    };
  private:
   struct INTERMEDIATE
    {
     TAG tag;
     unsigned depth;
     PREF_SYM symbol;
     OR_INDEX_TREE_NODE* below;
     OR_INDEX_TREE_NODE* alternative;
    };
   struct LEAF_NODE
    {
     TAG tag;
     unsigned depth;
     TERM term;
     union
      {
       RES_PAIR_LIST* clauses;
       BackParaCandidates* bp_candidates;
      };
    };
  public: 
   static OR_INDEX_TREE_NODE* NewNode(TAG t,PREF_SYM s,unsigned d)
    {
     INTERMEDIATE* res = (INTERMEDIATE*)(GlobAlloc::allocate(sizeof(INTERMEDIATE)));
     res->tag = t;
     res->depth = d;
     res->symbol = s;
     res->below = 0;
     res->alternative = 0;
     return (OR_INDEX_TREE_NODE*)res;
    };
   void DeleteNode()
    {
     GlobAlloc::deallocate((void*)this,sizeof(INTERMEDIATE));
    };
   static OR_INDEX_TREE_NODE* NewLeaf()
    {
     LEAF_NODE* res = (LEAF_NODE*)(GlobAlloc::allocate(sizeof(LEAF_NODE)));
     res->tag = LEAF;
     res->depth = 0;
     res->term.SetFirst((TERM*)0);
     res->clauses = (RES_PAIR_LIST*)0; 
     return (OR_INDEX_TREE_NODE*)res;
    }; 
   void DeleteLeaf()
    {
     CALL("DeleteLeaf()");
     if (((LEAF_NODE*)this)->clauses) RES_PAIR_LIST::Destroy(((LEAF_NODE*)this)->clauses);
     GlobAlloc::deallocate((void*)this,sizeof(LEAF_NODE));
    };

   void DeleteBPILeaf()
    {
     CALL("DeleteBPILeaf()");
     if (((LEAF_NODE*)this)->bp_candidates) delete ((LEAF_NODE*)this)->bp_candidates;
     GlobAlloc::deallocate((void*)this,sizeof(LEAF_NODE));
    };  
   
   const TAG& Tag() const { return ((INTERMEDIATE*)this)->tag; }; // applicable to leaves
   bool IsLeaf() const { return (Tag() == LEAF); };
   bool IsVar() const { return ((Tag() == VAR) || (Tag() == VAR_FIRST_OCC)); };
   const PREF_SYM& Symbol() const { return ((INTERMEDIATE*)this)->symbol; };
   OR_INDEX_TREE_NODE* Below() const { return ((INTERMEDIATE*)this)->below; };
   OR_INDEX_TREE_NODE** BelowAddr() { return &(((INTERMEDIATE*)this)->below); };
   void SetBelow(OR_INDEX_TREE_NODE* b) { ((INTERMEDIATE*)this)->below = b; };
   OR_INDEX_TREE_NODE* Alternative() const { return ((INTERMEDIATE*)this)->alternative; };
   OR_INDEX_TREE_NODE** AlternativeAddr() { return &(((INTERMEDIATE*)this)->alternative); };
   void SetAlternative(OR_INDEX_TREE_NODE* a) { ((INTERMEDIATE*)this)->alternative = a; };
   unsigned Depth() const { return ((INTERMEDIATE*)this)->depth; }; // applicable to leaves
   
   bool SameSymbol(const TERM s) const
    {
     if (s.isVariable())
      {
       return (Symbol().var() == VAR::Base(s.var()));
      }
     else // here s must be a functor
      { 
       return (Symbol().Func() == s);
      };
    }; // bool SameSymbol(const TERM s) const

   // for leaves:
   const TERM& Term() const { return ((LEAF_NODE*)this)->term; };
   void SetTerm(TERM* t) { ((LEAF_NODE*)this)->term.SetFirst(t); };
   void SetTerm(TERM t) { ((LEAF_NODE*)this)->term = t; };
   RES_PAIR_LIST* Clauses() const { return ((LEAF_NODE*)this)->clauses; };
   BackParaCandidates* BPCandidates() const { return ((LEAF_NODE*)this)->bp_candidates; };
   void AddPair(unsigned ln,Clause* cl)
    {
     ((LEAF_NODE*)this)->clauses = new RES_PAIR_LIST(ln,cl,((LEAF_NODE*)this)->clauses);
    };
   void AddFParaTripple(unsigned ln,Clause* cl,TERM t,SharedKBConstr* cns)
    {
     ((LEAF_NODE*)this)->clauses = new RES_PAIR_LIST(ln,cl,t,cns,((LEAF_NODE*)this)->clauses);
    }; 
   void AddBParaTripple(unsigned ln,Clause* cl,TERM* t,SharedKBConstr* cns)
    {
     CALL("AddBParaTripple(unsigned ln,Clause* cl,TERM* t,SharedKBConstr* cns)");
     BackParaCandidates* candidates = ((LEAF_NODE*)this)->bp_candidates;
     if (!candidates)         
      {
       candidates = new BackParaCandidates();
       ((LEAF_NODE*)this)->bp_candidates = candidates;
      };    
     candidates->add(cl,ln,t,cns);

     /************
     RES_PAIR_LIST* old_clauses = ((LEAF_NODE*)this)->clauses;
     if (old_clauses 
         && (old_clauses->TheClause() == cl) 
         && (old_clauses->LitNum() == ln) 
         && (old_clauses->Redex() == t))
      return; 
     ((LEAF_NODE*)this)->clauses = new RES_PAIR_LIST(ln,cl,t,cns,old_clauses);
     ************/
    }; 

   void Remove(Clause* c)
    {
     RES_PAIR_LIST::Remove(c,((LEAF_NODE*)this)->clauses);
    };
  
      
   void RemoveBPCandidates(Clause* c)      
   {
    CALL("RemoveBPCandidates(Clause* c)");
    ASSERT(((LEAF_NODE*)this)->bp_candidates);
    BackParaCandidates* candidates = ((LEAF_NODE*)this)->bp_candidates;
    candidates->remove(c); 
    if (candidates->empty()) 
     {
      delete candidates;
      ((LEAF_NODE*)this)->bp_candidates = (BackParaCandidates*)0;
     };     
    //RES_PAIR_LIST::Remove(c,((LEAF_NODE*)this)->clauses);     
   };

  public: 
   static void Destroy(OR_INDEX_TREE_NODE* tree);
   static void DestroyBPI(OR_INDEX_TREE_NODE* tree);
  public: // methods for debugging
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const { return output(str,0); };
    ostream& output(ostream& str,unsigned var_offset) const;
   #endif
 }; // class OR_INDEX_TREE_NODE

#ifndef NO_DEBUG_VIS
 inline ostream& operator<<(ostream& str,const OR_INDEX_TREE_NODE::TAG& tag)
 {
  switch (tag)
   {
    case OR_INDEX_TREE_NODE::LEAF: return str << "LEAF";
    case OR_INDEX_TREE_NODE::FUNC: return str << "FUNC";
    case OR_INDEX_TREE_NODE::CONST: return str << "CONST"; 
    case OR_INDEX_TREE_NODE::VAR: return str << "VAR";
    case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: return str << "VAR_FIRST_OCC"; 
    default: return str << "UNCKNOWN TAG"; 
   };
 }; // ostream& operator<<(ostream& str,const OR_INDEX_TREE_NODE::TAG& tag)
 
 inline ostream& OR_INDEX_TREE_NODE::output(ostream& str,unsigned var_offset) const
 {
  str << '[' << (ulong)this << "] " << Tag() << ' ';
  switch (Tag())
   {
    case LEAF: str << ' ' << Term(); break;
    case FUNC: 
    case CONST:
    case VAR: 
    case VAR_FIRST_OCC:    
      Symbol().output(str,var_offset) << " [" << (ulong)Below() << "] ["
                                      << (ulong)Alternative() << "] "
                                      << Depth();
      break;
   };
  return str;
}; // ostream& OR_INDEX_TREE_NODE::output(ostream& str,unsigned var_offset) const
#endif

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_INDEX"
#endif 
#include "debugMacros.h"
//================================================= 
inline void OR_INDEX::Integrate(Clause* cl)
 {
  CALL("Integrate(Clause* cl)");
  LiteralList* lits = cl->LitList();
  LiteralList::Element* firstNonselectedLit = lits->nth(lits->numOfSelectedLits());
  ulong litNum = 0;


  for(LiteralList::Iterator iter(lits);iter.currentElement() != firstNonselectedLit;iter.next()) 
   {
    Integrate(iter.currentLiteral(),cl,litNum);
    litNum++;
   };
 }; // void OR_INDEX::Integrate(Clause* const cl)

inline void OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)
 {
  CALL("Integrate(TERM* lit,Clause* cl,unsigned lit_num)");
  if (lit->arity())
   {
    if (paramodulation) IntegrateForBackwardPara(lit,cl,lit_num);
    if (lit->IsEquality())
     {      
      if (lit->Positive()) IntegrateForForwardPara(lit,cl,lit_num);
     }
    else
     {
      OR_INDEX_TREE_NODE* leaf = IntegrateNonpropLit(lit);
      ASSERT(leaf->Term().IsReference());
      ASSERT((!leaf->Term().First()) || (leaf->Term().First() == lit));
      leaf->SetTerm(lit);
      leaf->AddPair(lit_num,cl); 
     }; 
   }
  else
   {
    OR_INDEX_TREE_NODE* leaf = IntegratePropLit(*lit);    
    ASSERT(leaf->Term().IsReference());
    ASSERT((!leaf->Term().First()) || (leaf->Term().First() == lit));
    leaf->SetTerm(lit);
    leaf->AddPair(lit_num,cl); 
   };  
 }; // void OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)

inline void OR_INDEX::IntegrateBPRedex(TERM* redex,Clause* cl,unsigned lit_num,SharedKBConstr* cns)
 {
  CALL("IntegrateBPRedex(TERM* redex,Clause* cl,unsigned lit_num,SharedKBConstr* cns)");
  ASSERT(redex->isComplex());
  OR_INDEX_TREE_NODE** tree = &backward_para_tree;
  OR_INDEX_TREE_NODE* leaf;
  pseudo_literal->Args()->SetFirst(redex);
  l_iter.Reset(pseudo_literal,0);
  l_iter.Next();
  leaf = IntegrateArgs(tree);
  leaf->SetTerm(redex);
  leaf->AddBParaTripple(lit_num,cl,redex,cns);  
 }; // void OR_INDEX::IntegrateBPRedex(TERM* redex,Clause* cl,unsigned lit_num,SharedKBConstr* cns)

inline void OR_INDEX::IntegrateForBackwardPara(TERM* lit,Clause* cl,unsigned lit_num)
 {
  CALL("IntegrateForBackwardPara(TERM* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->arity());
  SharedKBConstr* constr = (SharedKBConstr*)0;
  TERM::Iterator iter;
  if (lit->IsEquality())
   {
    if (lit->IsOrderedEq())
     {
      // only the first argument can be rewritten
      ASSERT(lit->Arg1()->IsReference());
      TERM* arg = lit->Arg1()->First();
      iter.Reset(arg);
      do
       {
        if (!iter.CurrentSymIsVar()) IntegrateBPRedex(iter.CurrPos(),cl,lit_num,constr); 
       }
      while (iter.Next());
     }
    else // unordered equality
     {

       //DF; cout << "EQU " << lit << "\n";

      TERM* arg1ref = lit->Arg1();
      TERM* arg2ref = arg1ref->Next();
      if (arg1ref->isVariable() || arg2ref->isVariable()) goto integrate_ordinary;
      TERM* arg1 = arg1ref->First();      
      TERM* arg2 = arg2ref->First();

      ulong funcs1;
      ulong funcs2; 
      collectVarsAndFuncs(arg1,_vars1,funcs1);
      collectVarsAndFuncs(arg2,_vars2,funcs2);
      
      // intergrate first argument
      if (_vars1.SubsetOf(_vars2))
       {
        // create a constraint
        SharedKBConstr::assumeGreater(arg2,_vars2,funcs2,arg1,_vars1,funcs1);
        constr = SharedKBConstr::getShared();
       };

      //DF; constr->output(cout << "ARG1 < ARG2 :") << "\n";

      iter.Reset(arg1);
      do
       {
        if (!iter.CurrentSymIsVar()) IntegrateBPRedex(iter.CurrPos(),cl,lit_num,constr); 
       }
      while (iter.Next());

      // intergrate second argument
      if (_vars2.SubsetOf(_vars1))
       {
        // create a constraint
        SharedKBConstr::assumeGreater(arg1,_vars1,funcs1,arg2,_vars2,funcs2);
        constr = SharedKBConstr::getShared();
       }
      else 
       constr = (SharedKBConstr*)0;

      //DF; constr->output(cout << "ARG2 < ARG1 :") << "\n";

      iter.Reset(arg2);
      do
       {
        if (!iter.CurrentSymIsVar()) IntegrateBPRedex(iter.CurrPos(),cl,lit_num,constr); 
       }
      while (iter.Next());
     };
   }
  else // ordinary literal
   {
    integrate_ordinary:
     ASSERT(!constr);
     iter.Reset(lit);
     while (iter.Next()) 
      if (!iter.CurrentSymIsVar()) 
       {
        IntegrateBPRedex(iter.CurrPos(),cl,lit_num,constr); 
       };    
   };
 }; // void OR_INDEX::IntegrateForBackwardPara(TERM* lit,Clause* cl,unsigned lit_num)

inline void OR_INDEX::IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)
 {
  CALL("IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->IsEquality());
  ASSERT(lit->Positive());

  if (lit->IsOrderedEq()) { IntegratePreorderedForForwardPara(lit,cl,lit_num); return; };

  TERM* arg1ref = lit->Args();
  TERM* arg2ref = arg1ref->Next();
  if (arg1ref->IsReference() && arg2ref->IsReference())
   {
    IntegrateConstrainedForForwardPara(*arg1ref,*arg2ref,cl,lit_num);   
    IntegrateConstrainedForForwardPara(*arg2ref,*arg1ref,cl,lit_num);
    return;
   };

  OR_INDEX_TREE_NODE* leaf;
  // integrate the first argument
  TERM* arg = lit->Args();
  TERM replacing_term = *(arg->Next());
  bool stop = false;
  integrate_arg: 
   if (arg->isVariable())
    {
     pseudo_literal[1] = *arg;
     l_iter.Reset(pseudo_literal,0);
     l_iter.Next();
     leaf = IntegrateArgs(&forward_para_from_var_tree);
     leaf->SetTerm(*arg);
     leaf->AddFParaTripple(lit_num,cl,replacing_term,(SharedKBConstr*)0); 
    }
   else // complex term, integrate as ordinary literal
    {
     if (arg->First()->arity())
      {
       l_iter.Reset(arg->First(),0);
       l_iter.Next();
       leaf = IntegrateArgs(FParaTree(*(arg->First())));
      }
     else
      {
       leaf = *(FParaTree(*(arg->First())));
       if (!leaf) 
        {
         leaf = OR_INDEX_TREE_NODE::NewLeaf();
         *(FParaTree(*(arg->First()))) = leaf;
        };
      }; 
     leaf->SetTerm(arg->First());
     leaf->AddFParaTripple(lit_num,cl,replacing_term,(SharedKBConstr*)0);
    };    
   
  if (lit->IsOrderedEq() || stop) return; 
  // integrate the second argument 
  replacing_term = *arg;
  arg = arg->Next();
  stop = true;
  goto integrate_arg;
 }; // void OR_INDEX::IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)

inline void OR_INDEX::IntegrateConstrainedForForwardPara(const TERM& lhs,const TERM& rhs,Clause* cl,unsigned lit_num)
{
 CALL("IntegrateConstrainedForForwardPara(const TERM& lhs,const TERM& rhs,Clause* cl,unsigned lit_num)");
 ASSERT(lhs.IsReference());
 ASSERT(rhs.IsReference());  
 // try to create a constraint 
 SharedKBConstr* constr = (SharedKBConstr*)0;
 ulong lhs_funcs;
 ulong rhs_funcs; 
 collectVarsAndFuncs(lhs.First(),_vars1,lhs_funcs);
 collectVarsAndFuncs(rhs.First(),_vars2,rhs_funcs);
 if (_vars1.SubsetOf(_vars2))
  {
   SharedKBConstr::assumeGreater(rhs.First(),_vars2,rhs_funcs,lhs.First(),_vars1,lhs_funcs);
   constr = SharedKBConstr::getShared();  
  };

 OR_INDEX_TREE_NODE* leaf;
 if (lhs.First()->arity())
  {
   l_iter.Reset(lhs.First(),0);
   l_iter.Next();
   leaf = IntegrateArgs(FParaTree(*(lhs.First())));
  }
 else
  {
   leaf = *(FParaTree(*(lhs.First())));
   if (!leaf) 
    {
     leaf = OR_INDEX_TREE_NODE::NewLeaf();
     *(FParaTree(*(lhs.First()))) = leaf;
    };
  }; 
 leaf->SetTerm(lhs.First());
 leaf->AddFParaTripple(lit_num,cl,rhs,constr); 
}; // void OR_INDEX::IntegrateConstrainedForForwardPara(const TERM& lhs,const TERM& rhs,Clause* cl,unsigned lit_num)


inline void OR_INDEX::IntegratePreorderedForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)
{
 CALL("IntegratePreorderedForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)");
 ASSERT(lit->IsOrderedEq());
 OR_INDEX_TREE_NODE* leaf;
 // integrate only the first argument
 TERM* arg = lit->Args();
 ASSERT(arg->IsReference());
 TERM replacing_term = *(arg->Next());   
 if (arg->First()->arity())
  {
   l_iter.Reset(arg->First(),0);
   l_iter.Next();
   leaf = IntegrateArgs(FParaTree(*(arg->First())));
  }
 else
  {
   leaf = *(FParaTree(*(arg->First())));
   if (!leaf) 
    {
     leaf = OR_INDEX_TREE_NODE::NewLeaf();
     *(FParaTree(*(arg->First()))) = leaf;
    };
  }; 
 leaf->SetTerm(arg->First());
 leaf->AddFParaTripple(lit_num,cl,replacing_term,(SharedKBConstr*)0);  
}; // void OR_INDEX::IntegratePreorderedForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)





inline OR_INDEX_TREE_NODE* OR_INDEX::IntegratePropLit(TERM lit)
 {
  OR_INDEX_TREE_NODE** tree = Tree(lit);
  if (!(*tree))
   {
    // no tree for this header yet, create it here
    *tree = OR_INDEX_TREE_NODE::NewLeaf();
   };
  return *tree;
 }; // OR_INDEX_TREE_NODE* OR_INDEX::IntegratePropLit(TERM lit)

inline OR_INDEX_TREE_NODE* OR_INDEX::IntegrateNonpropLit(TERM* lit)
 {
  OR_INDEX_TREE_NODE** tree = Tree(*lit);
  l_iter.Reset(lit,0);
  l_iter.Next();
  return IntegrateArgs(tree);
 }; // OR_INDEX_TREE_NODE* OR_INDEX::IntegrateNonpropLit(TERM* lit)
 
inline OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree)
 {
  CALL("IntegrateArgs(OR_INDEX_TREE_NODE** tree)");
  var_counter.Reset();
  if (*tree)
   {
    OR_INDEX_TREE_NODE* node = *tree;
    PREF_SYM sym;
    take_next_sym:
     if (l_iter.CurrentSymIsVar())
      {
       ulong var_num = l_iter.Symbol().var();
       sym.MkVar(VAR::Base(var_num));
       var_counter.Register(var_num);
      }
     else
      {
       sym.MkFunc(l_iter.Symbol());
      };
     check_node:
      if (node->Symbol() == sym)
       {
        node = node->Below();
        if (l_iter.Next())
         {
          goto take_next_sym;
         }
        else // node must be a leaf
         {
          ASSERT(node->IsLeaf());
          return node;
         };
       }
      else 
       {
        if (node->Alternative())
         {
          node = node->Alternative();
          goto check_node;
         }
        else // no alternative, new branch must be constructed
         {
          OR_INDEX_TREE_NODE* leaf;
          node->SetAlternative(CreateBranch(leaf));
          return leaf;
         };
       };      
   }
  else // empty tree, create branch
   {
    OR_INDEX_TREE_NODE* leaf;
    *tree = CreateBranch(leaf);
    return leaf;
   };
 }; // OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree)





//**************************************************************************

inline OR_INDEX_TREE_NODE* OR_INDEX::CreateBranch(OR_INDEX_TREE_NODE*& leaf)
 {
  PREF_SYM sym;
  OR_INDEX_TREE_NODE::TAG tag;
  if (l_iter.CurrentSymIsVar())
   {
    if (var_counter.Occurences(l_iter.Symbol().var()))
     {
      tag = OR_INDEX_TREE_NODE::VAR;
     }
    else // first occurence of the variable
     {
      var_counter.Register(l_iter.Symbol().var());
      tag = OR_INDEX_TREE_NODE::VAR_FIRST_OCC;
     };
    sym.MkVar(VAR::Base(l_iter.Symbol().var()));
   }
  else
   {
    if (l_iter.Symbol().arity())
     {
      tag = OR_INDEX_TREE_NODE::FUNC;
     }
    else
     {
      tag = OR_INDEX_TREE_NODE::CONST;
     };
    sym.MkFunc(l_iter.Symbol());
   };
  OR_INDEX_TREE_NODE* result = OR_INDEX_TREE_NODE::NewNode(tag,sym,l_iter.Depth());
  OR_INDEX_TREE_NODE* node = result;
  while (l_iter.Next())
   {
    if (l_iter.CurrentSymIsVar())
     {
      if (var_counter.Occurences(l_iter.Symbol().var()))
       {
        tag = OR_INDEX_TREE_NODE::VAR;
       }
      else // first occurence of the variable
       {
        var_counter.Register(l_iter.Symbol().var());
        tag = OR_INDEX_TREE_NODE::VAR_FIRST_OCC;
       };
      sym.MkVar(VAR::Base(l_iter.Symbol().var()));
     }
    else
     {
      if (l_iter.Symbol().arity())
       {
        tag = OR_INDEX_TREE_NODE::FUNC;
       }
      else
       {
        tag = OR_INDEX_TREE_NODE::CONST;
       };
      sym.MkFunc(l_iter.Symbol());
     };
    node->SetBelow(OR_INDEX_TREE_NODE::NewNode(tag,sym,l_iter.Depth()));
    node = node->Below();
   };
  
  leaf = OR_INDEX_TREE_NODE::NewLeaf();
  node->SetBelow(leaf);
  return result;
 }; // OR_INDEX_TREE_NODE* OR_INDEX::CreateBranch(OR_INDEX_TREE_NODE*& leaf)


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_INDEX_PATH"
#endif 
#include "debugMacros.h"
//================================================= 

class OR_INDEX_PATH
 {
  private:
   OR_INDEX_TREE_NODE** bookmark[HellConst::MaxTermSize];
   OR_INDEX_TREE_NODE*** next_bookmark;
   OR_INDEX_TREE_NODE* path[HellConst::MaxTermSize];
   OR_INDEX_TREE_NODE** next;
   OR_INDEX_TREE_NODE** changed_suffix;
   OR_INDEX_TREE_NODE* curr_node;
   unsigned var_offset;
   VAR* var_extracted;
  public:
   OR_INDEX_PATH() {};
   void SetIndex(ulong index) 
    {
     var_offset = VAR::Offset(index); 
    };
   ulong GetIndex() const { return VAR::Index(var_offset); };
   VAR* Shift(VAR* const v) { return (v + var_offset); }; 
   void Reset(OR_INDEX_TREE_NODE* const tree)
    {
     CALL("Reset(OR_INDEX_TREE_NODE* const tree)");
     ASSERT(tree);
     next = path;
     curr_node = tree;
     next_bookmark = bookmark;
    };
   static void ResetStatic() { VAR::Reset(); };
   OR_INDEX_TREE_NODE* CurrNode() const { return curr_node; };
   OR_INDEX_TREE_NODE** Position() const { return next; };
   OR_INDEX_TREE_NODE** ChangedSuffix() const { return changed_suffix; };
   void MakeBookmark() { *next_bookmark = Position(); next_bookmark++; };
   void ForgetBookmark() { next_bookmark--; };
   OR_INDEX_TREE_NODE** LastBookmark() const { return *(next_bookmark-1); };
   VAR* VarExtracted() const { return var_extracted; };
   void MemorizeCurrent() { *next = curr_node; next++; };
   void RestoreCurrent() { next--; curr_node = *next; };
   void Below() 
    {
     CALL("Below()");
     ASSERT(!(curr_node->IsLeaf()));
     MemorizeCurrent();
     curr_node = curr_node->Below();
    };
   void Alternative() 
    { 
     CALL("!(curr_node->IsLeaf())");
     ASSERT(!(curr_node->IsLeaf())); 
     curr_node = curr_node->Alternative();
    };
   void ExtractTerm()
    {
     MakeBookmark();
     if (curr_node->IsVar())
      {
       var_extracted = Shift(curr_node->Symbol().var());
       Below();
      }
     else // complex term to be extracted
      {
       var_extracted = (VAR*)0;
       unsigned init_depth = curr_node->Depth();
       do 
        {
         Below();
        }
       while (init_depth < curr_node->Depth());
      };
    }; // void ExtractTerm()
   bool ModifyTerm();
   bool ExtractTermWithout(VAR* const var,bool& same_var);
   bool ModifyTermWithout(VAR* const var,
                                 bool& same_var);
   void ToPrefix(OR_INDEX_TREE_NODE** pos,PREFIX_SYM*& memory);
  public: // methods for debugging
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
     for (const OR_INDEX_TREE_NODE* const * p = path; p<next; p++)    
      (*p)->output(str,var_offset) << '\n'; 
     str << "CURRENT NODE = ";
     if (curr_node) { curr_node->output(str,var_offset); }
     else str << "[0]";  
     return str << '\n'; 
    };
   #endif
 }; // class OR_INDEX_PATH 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_PROCESSOR"
#endif 
#include "debugMacros.h"
//================================================= 
class OR_PROCESSOR : private OR_INDEX_PATH
 {  
  private:
   PREFIX_SYM memory[HellConst::MaxTermSize];
   PREFIX_SYM* free_memory;
   PREFIX::ITER_WITH_BACKTRACK* term[HellConst::MaxTermSize];
   PREFIX::ITER_WITH_BACKTRACK** next_term;
   PREFIX::ITER_WITH_BACKTRACK* curr_iter;
  public: // made public because VC++ doesn't like friends,
	      // see ostream& operator<<(ostream& str,const OR_PROCESSOR::SITUATION& sit) 
   enum SITUATION
    {
     START_OF_EXTRACTION,
     VAR_SAME_VAR,
     VAR_ANYTHING,
     FUNCTOR_FUNCTOR,
     FUNCTOR_FREE_VAR,
     FUNCTOR_BOUND_VAR
    };
  private:
   SITUATION backtrack_situation[HellConst::MaxTermSize*2];
   SITUATION* curr_sit; 
  private:
   void NextIter() 
    {
     CALL("NextIter()");
     next_term++;
     curr_iter = *next_term;
     if (!curr_iter) 
      {
       curr_iter = new PREFIX::ITER_WITH_BACKTRACK(); 
       *next_term = curr_iter;
      };
     ASSERT(next_term - term <= (long)HellConst::MaxTermSize);
    };
   void ForgetIter() 
    {
      CALL("ForgetIter()");
     next_term--; 
     ASSERT(next_term >= term);
     curr_iter = *next_term;
    };
   PREFIX::ITER_WITH_BACKTRACK* LastIter() { return *(next_term-1); };
  public:
   OR_PROCESSOR() : OR_INDEX_PATH() 
    {
     for (ulong i = 0; i < HellConst::MaxTermSize; i++)
      {
       term[i] = (PREFIX::ITER_WITH_BACKTRACK*)0;
      };
    };
   static void ResetStatic() { OR_INDEX_PATH::ResetStatic(); };
   void SetIndex(ulong index) { OR_INDEX_PATH::SetIndex(index); };
   ulong GetIndex() const { return OR_INDEX_PATH::GetIndex(); };
   OR_INDEX_TREE_NODE* CurrNode() const { return OR_INDEX_PATH::CurrNode(); };
   void Below() { OR_INDEX_PATH::Below(); };
   void Alternative() { OR_INDEX_PATH::Alternative(); };
   void RestoreCurrent() { OR_INDEX_PATH::RestoreCurrent(); };
   VAR* Shift(VAR* v) { return OR_INDEX_PATH::Shift(v); };
  public: // commands
   void Start(OR_INDEX_TREE_NODE* tree) 
    {
     free_memory = memory;
     curr_sit = backtrack_situation;
     next_term = term;
     OR_INDEX_PATH::Reset(tree); 
    };
   void Bind(VAR* v);
   bool AlterBinding();
   bool BindWithOccCheck(VAR* v,bool& same_var);
   bool AlterBindingWithOccCheck(VAR* v,bool& same_var);
  private:
   bool BacktrackExtraction();
   bool CompleteExtraction();
  public:
   bool ExtractUnifiable(PREFIX_SYM* word);
   bool AlterUnifiable();
  public: // methods for debugging
  #ifndef NO_DEBUG_VIS
    ostream& outputPath(ostream& str) const { return OR_INDEX_PATH::output(str); };  
   friend ostream& operator<<(ostream& str,const OR_PROCESSOR::SITUATION& sit);
  #endif  
 }; // class OR_PROCESSOR : private OR_INDEX_PATH

#ifndef NO_DEBUG_VIS
inline ostream& operator<<(ostream& str,const OR_PROCESSOR::SITUATION& sit) 
{
 switch (sit)
  {
   case OR_PROCESSOR::START_OF_EXTRACTION : str << "START_OF_EXTRACTION"; break;
   case OR_PROCESSOR::VAR_SAME_VAR : str << "VAR_SAME_VAR"; break;
   case OR_PROCESSOR::VAR_ANYTHING : str << "VAR_ANYTHING"; break;
   case OR_PROCESSOR::FUNCTOR_FUNCTOR : str << "FUNCTOR_FUNCTOR"; break;
   case OR_PROCESSOR::FUNCTOR_FREE_VAR : str << "FUNCTOR_FREE_VAR"; break;
   case OR_PROCESSOR::FUNCTOR_BOUND_VAR : str << "FUNCTOR_BOUND_VAR"; break;
  };
 return str;
}; //ostream& operator<<(ostream& str,const OR_PROCESSOR::SITUATION& sit) 
#endif

//****************** Definitions: ********************************************

inline bool OR_PROCESSOR::CompleteExtraction()
 {
  CALL("CompleteExtraction()");
  check_iter:
  
   if (curr_iter->Empty()) 
    {
     return true; 
    };
   
   check_sym:
   
   
    VAR* var = curr_iter->CurrVar();
    if (var) // free variable
     {
     
      ASSERT(!(var->Binding()));
      bool same_var;
      
      if (BindWithOccCheck(var,same_var))
       {
       
        ASSERT(!(same_var && var->Binding()));
        if (same_var) { *curr_sit = VAR_SAME_VAR; }
        else { *curr_sit = VAR_ANYTHING; };
        curr_sit++;
        curr_iter->SkipSym();
        goto check_iter;
       }
      else
       {
        if (BacktrackExtraction()) { goto check_sym; }
        else { return false; };
       };
     }
    else // functor
     {
     
      check_node:
       switch (OR_INDEX_PATH::CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          // currently the same as for OR_INDEX_TREE_NODE::CONST,
          // but can be optimized by jumping to check_sym instead of check_iter
          
         case OR_INDEX_TREE_NODE::CONST : 
         
          if (OR_INDEX_PATH::CurrNode()->Symbol() == curr_iter->CurrSubterm()->Head())
           { // same functor
            
            curr_iter->SkipSym();
            OR_INDEX_PATH::Below();
            *curr_sit = FUNCTOR_FUNCTOR;
            curr_sit++;
            goto check_iter;
           }
          else // different functors, try alternative
           {
            OR_INDEX_PATH::Alternative();
            if (OR_INDEX_PATH::CurrNode())
             {
              goto check_node;
             }
            else // backtrack needed
             {
              if (BacktrackExtraction()) { goto check_sym; }
              else { return false; };
             };
           };
         
         case OR_INDEX_TREE_NODE::VAR :
          {
           bool free_var;
           void* bind = OR_INDEX_PATH::Shift(OR_INDEX_PATH::CurrNode()->Symbol().var())->Unref(free_var);
           PREFIX::CorrectBracket(curr_iter->CurrSubterm());
           
           ASSERT(PREFIX::CorrectComplexTerm(curr_iter->CurrSubterm()));
           
           if (free_var)
            {
             if (OCCUR_CHECK::Occurs((VAR*)bind,curr_iter->CurrSubterm()))
              {
               // try alternative
               OR_INDEX_PATH::Alternative();
               if (OR_INDEX_PATH::CurrNode())
                {
                 goto check_node;
                }
               else // backtrack needed
                {
                 if (BacktrackExtraction()) { goto check_sym; }
                 else { return false; };
                };
              }
             else // the variable doesn't occur in the subterm, make binding
              {
               GlobalSubst.BindAlone((VAR*)bind,curr_iter->CurrSubterm());
               curr_iter->SkipTerm();
               OR_INDEX_PATH::Below();
               *curr_sit = FUNCTOR_FREE_VAR;
               curr_sit++;
               goto check_iter;
              };
            }
           else // variable bound by a complex term, try to unify
            {
             if (UNIFICATION::UnifyComplex((PREFIX_SYM*)bind,curr_iter->CurrSubterm()))
              {
               curr_iter->SkipTerm();
               OR_INDEX_PATH::Below();
               *curr_sit = FUNCTOR_BOUND_VAR;
               curr_sit++;
               goto check_iter;
              }
             else // try alternative
              {
               OR_INDEX_PATH::Alternative();
               if (OR_INDEX_PATH::CurrNode())
                {
                 goto check_node;
                }
               else // backtrack needed
                {
                 if (BacktrackExtraction()) { goto check_sym; }
                 else { return false; };
                };
              };
            };
          };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC :
          PREFIX::CorrectBracket(curr_iter->CurrSubterm());
          
          ASSERT(PREFIX::CorrectComplexTerm(curr_iter->CurrSubterm()));
          
          GlobalSubst.BindAlone(OR_INDEX_PATH::Shift(OR_INDEX_PATH::CurrNode()->Symbol().var()),
                           curr_iter->CurrSubterm());
          
          curr_iter->SkipTerm();
          OR_INDEX_PATH::Below();
          *curr_sit = FUNCTOR_FREE_VAR;
          curr_sit++;
          goto check_iter;         
         
         #ifdef DEBUG_NAMESPACE
	  default : ICP("A2"); return false;
         #elif defined _SUPPRESS_WARNINGS_
          default : return false;
         #endif 
        }; // switch
       ICP("A3");
       return false;
     };  
 }; // bool OR_PROCESSOR::CompleteExtraction()

inline bool OR_PROCESSOR::BacktrackExtraction()
 {
  CALL("BacktrackExtraction()");
  rollback:
   ASSERT((!(OR_INDEX_PATH::CurrNode())) || OR_INDEX_PATH::CurrNode()->IsLeaf());
  
   curr_sit--;
   
   switch (*curr_sit)
    {
     case START_OF_EXTRACTION : 
      ForgetIter(); 
      return false;
     case VAR_SAME_VAR :
      {
       bool same_var;
       curr_iter->Backtrack();

       ASSERT(curr_iter->CurrVar());
       ASSERT(!(curr_iter->CurrVar()->Binding()));
       
       if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
        {
         if (!(same_var)) { *curr_sit = VAR_ANYTHING; };
         curr_sit++;
         curr_iter->SkipSym();
         return true; 
        }
       else 
        {
         goto rollback; 
        };
      };
     case VAR_ANYTHING : 
      {
       bool same_var;
       curr_iter->Backtrack();
       
       if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
        {
         if (same_var) 
          {
           *curr_sit = VAR_SAME_VAR; 
          };
         curr_sit++; 
         curr_iter->SkipSym();
         return true; 
        }
       else 
        {
         goto rollback; 
        };
      };

     case FUNCTOR_FUNCTOR : 
      curr_iter->Backtrack();
      
      
      ASSERT(curr_iter->CurrSubterm()->Head().IsFunctor());

      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) { return true; }
      else 
       {
        goto rollback; 
       };

     case FUNCTOR_FREE_VAR :
      GlobalSubst.UnbindLast();
      curr_iter->Backtrack();

      ASSERT(curr_iter->CurrSubterm()->Head().IsFunctor());
      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) { return true; }
      else 
       {
        goto rollback; 
       };

     case FUNCTOR_BOUND_VAR :
      
      GlobalSubst.UnbindLastSection();
      curr_iter->Backtrack();
      ASSERT(curr_iter->CurrSubterm()->Head().IsFunctor());
      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) 
       {
        return true; 
       }
      else 
       {
        goto rollback; 
       };
     #ifdef DEBUG_NAMESPACE
      default : ICP("ICP1"); return false;
     #endif
    };  
  ICP("ICP2"); 
  return false; 
 }; // bool OR_PROCESSOR::BacktrackExtraction()

inline bool OR_PROCESSOR::ExtractUnifiable(PREFIX_SYM* word) 
  // word must represent a complex term
 {
  NextIter();
  curr_iter->Reset(word);
  *curr_sit = START_OF_EXTRACTION;
  curr_sit++;
  return CompleteExtraction();
 }; // bool OR_PROCESSOR::ExtractUnifiable(PREFIX_SYM* word)
 
inline bool OR_PROCESSOR::AlterUnifiable()
 {
  if (BacktrackExtraction()) 
   {
    return CompleteExtraction(); 
   }
  else 
   {
    return false; 
   };
 }; // bool OR_PROCESSOR::AlterUnifiable()

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_INDEX_PATH"
#endif 
#include "debugMacros.h"
//================================================= 
inline bool OR_INDEX_PATH::ModifyTerm()
 {
  CALL("ModifyTerm()");
  OR_INDEX_TREE_NODE** begin = LastBookmark();
   
  ASSERT(begin < Position());
  
  rollback:
   RestoreCurrent();
   changed_suffix = Position();
   if (Position() == begin)  
    {
     Alternative();
     if (curr_node) { ForgetBookmark(); ExtractTerm(); return true; }
     else { ForgetBookmark(); return false; };
    };
   Alternative();
   if (curr_node)
    {
     unsigned init_depth = (*begin)->Depth();
     do 
      {
       Below();
      }
     while (init_depth < curr_node->Depth());
     var_extracted = (VAR*)0;
     return true;
    }
   else { goto rollback; };
 }; // bool OR_INDEX_PATH::ModifyTerm()

inline bool OR_INDEX_PATH::ExtractTermWithout(VAR* const var,bool& same_var)
 {
  CALL("ExtractTermWithout(VAR* const var,bool& same_var)");
  MakeBookmark();
  same_var = false;
  check_node:
   switch (curr_node->Tag())
    {
     case OR_INDEX_TREE_NODE::FUNC:
      {
       var_extracted = (VAR*)0;
       unsigned init_depth = curr_node->Depth();
       OR_INDEX_TREE_NODE** init_pos = Position();
       Below();
       check_depth:
        if (curr_node->Depth() <= init_depth) { return true; };
        intern_check_node:
         switch (curr_node->Tag())
          {
           case OR_INDEX_TREE_NODE::FUNC: Below(); goto intern_check_node;
           case OR_INDEX_TREE_NODE::CONST: Below(); goto check_depth;
           case OR_INDEX_TREE_NODE::VAR: 
            {
             bool free_var;
             void* bind = Shift(curr_node->Symbol().var())->Unref(free_var);
             if (free_var)
              {
               if ((VAR*)bind == var) { goto try_alternative; } 
               else { Below(); goto check_depth; };
              }
             else // variable bound by a complex term in the node
              {
               if (OCCUR_CHECK::Occurs(var,(PREFIX_SYM*)bind))
                { goto try_alternative; }
               else { Below(); goto check_depth; };
              };
            };
           case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: Below(); goto check_depth;
                      
           
           #ifdef DEBUG_NAMESPACE
            default: ICP("ICP1"); return false;
           #elif defined _SUPPRESS_WARNINGS_
	    default: return false;
           #endif
          };
       try_alternative:
        Alternative();
        if (curr_node) { goto intern_check_node; }
        else // backtrack needed
         {
          RestoreCurrent();
          if (Position() == init_pos) 
           {
            Alternative();
            if (curr_node) { goto check_node; }
            else { ForgetBookmark(); return false; };
           }
          else { goto try_alternative; };
         };
      }; // case OR_INDEX_TREE_NODE::FUNC
     case OR_INDEX_TREE_NODE::CONST: 
      Below(); 
      var_extracted = (VAR*)0;
      return true;
     case OR_INDEX_TREE_NODE::VAR: 
      {
       bool free_var;
       void* bind = Shift(curr_node->Symbol().var())->Unref(free_var);
       if (free_var)
        {
         var_extracted = Shift(curr_node->Symbol().var());
         Below();
         same_var = (((VAR*)bind) == var);
         return true;
        }
       else
        {
         if (OCCUR_CHECK::Occurs(var,(PREFIX_SYM*)bind))
          {
           Alternative();
           if (curr_node) { goto check_node; }
           else { ForgetBookmark(); return false; };
          }
         else
          {
           var_extracted = Shift(curr_node->Symbol().var());
           Below();
           return true;
          };
        };
      };
     case OR_INDEX_TREE_NODE::VAR_FIRST_OCC:
      var_extracted = Shift(curr_node->Symbol().var());
      Below(); 
      return true;
      
     #ifdef DEBUG_NAMESPACE
      default: ICP("ICP2"); return false;
     #elif defined _SUPPRESS_WARNINGS_
      default: return false;
     #endif
    };
  ICP("ICP3");
  return false;  
 }; // bool OR_INDEX_PATH::ExtractTermWithout(VAR* var,bool& same_var)

inline bool OR_INDEX_PATH::ModifyTermWithout(VAR* const var,bool& same_var)
 {
  CALL("ModifyTermWithout(VAR* const var,bool& same_var)");
  OR_INDEX_TREE_NODE** begin = LastBookmark();
  
  ASSERT(begin < Position());
  
  unsigned init_depth = (*begin)->Depth();
  
  same_var = false; 
  changed_suffix = Position()-1;
  rollback:
   RestoreCurrent();
   if (Position() < changed_suffix) { changed_suffix = Position(); };
   if (Position() == begin)  
    {
     Alternative();
     ForgetBookmark();
     return ((curr_node) && (ExtractTermWithout(var,same_var)));
    };
   try_alternative:
    Alternative();
    if (curr_node)
     {
      check_node:
       switch (curr_node->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC: Below(); goto check_node;
         case OR_INDEX_TREE_NODE::CONST: Below(); goto check_depth;
         case OR_INDEX_TREE_NODE::VAR:
          {
           bool free_var;
           void* bind = Shift(curr_node->Symbol().var())->Unref(free_var);
           if (free_var)
            {
             if ((VAR*)bind == var) { goto try_alternative; } 
             else { Below(); goto check_depth; };
            }
           else // variable bound by a complex term in the node
            {
             if (OCCUR_CHECK::Occurs(var,(PREFIX_SYM*)bind))
              { goto try_alternative; }
             else { Below(); goto check_depth; };
            };
          };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: Below(); goto check_depth;
             
         #ifdef DEBUG_NAMESPACE
          default: ICP("ICP1"); return false;
         #elif defined _SUPPRESS_WARNINGS_
          default: return false;
         #endif
        };
       ICP("ICP2");
       return false;
      check_depth:
       if (curr_node->Depth() <= init_depth) { var_extracted = (VAR*)0; return true; }
       else 
        {
         goto check_node; 
        };
     }
    else { goto rollback; };
 }; // bool OR_INDEX_PATH::ModifyTermWithout(VAR* var,.......

inline void OR_INDEX_PATH::ToPrefix(OR_INDEX_TREE_NODE** pos,PREFIX_SYM*& memory)
 {
  CALL("ToPrefix(OR_INDEX_TREE_NODE** pos,PREFIX_SYM*& memory)");
  OR_INDEX_TREE_NODE* node;
  while (pos < Position())
   {
    node = *pos;
    switch (node->Tag())
     {
      case OR_INDEX_TREE_NODE::FUNC:
       memory->Head() = node->Symbol();
       memory->ResetBracket();
       break;
      case OR_INDEX_TREE_NODE::CONST:
       memory->Head() = node->Symbol();
       memory->SetBracket(memory+1);
       break; 
      case OR_INDEX_TREE_NODE::VAR: 
      case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: 
       memory->Head().MkVar(Shift(node->Symbol().var()));
       memory->SetBracket(memory+1); 
       break;
      #ifdef DEBUG_NAMESPACE
       default: ICP("ICP1"); return; 
      #elif defined _SUPPRESS_WARNINGS_
       default: return; 
      #endif
     };
    pos++;
    memory++;   
   }; 
 }; // void OR_INDEX_PATH::ToPrefix(OR_INDEX_TREE_NODE** pos,PREFIX_SYM*& memory)

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
 #define DEBUG_NAMESPACE "OR_PROCESSOR"
#endif 
#include "debugMacros.h"
//================================================= 
inline void OR_PROCESSOR::Bind(VAR* v)
 {
  OR_INDEX_PATH::ExtractTerm();
  if (OR_INDEX_PATH::VarExtracted())
   {
    GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted()));
   }
  else // complex term extracted
   {
    PREFIX_SYM* word = free_memory;
    OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
    word->SetBracket(free_memory);
    GlobalSubst.BindAlone(v,(void*)word);
   };
 }; // void OR_PROCESSOR::Bind(VAR* v)
    
inline bool OR_PROCESSOR::AlterBinding()
 {
  CALL("AlterBinding()");
  VAR* v = GlobalSubst.LastBound();
  void* bind = v->Binding();
  
  ASSERT(bind);

  GlobalSubst.UnbindLast();
  
  if (OR_INDEX_PATH::ModifyTerm())
   {
    if (OR_INDEX_PATH::VarExtracted())
     {
      GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted()));
      if (!(VAR::IsVar(bind))) 
       {
        free_memory = (PREFIX_SYM*)bind; 
       };
     }
    else // complex term extracted
     {
      if (VAR::IsVar(bind))
       {
        PREFIX_SYM* word = free_memory;
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
        word->SetBracket(free_memory);
        GlobalSubst.BindAlone(v,(void*)word);
       }
      else
       {
        PREFIX_SYM* changed_suffix = (PREFIX_SYM*)bind + (OR_INDEX_PATH::ChangedSuffix() - OR_INDEX_PATH::LastBookmark());
        free_memory = changed_suffix;
        ASSERT(free_memory >= ((PREFIX_SYM*)bind));
        
        PREFIX_SYM* old_end_of_word = ((PREFIX_SYM*)bind)->Bracket();
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::ChangedSuffix(),free_memory);
        
        ASSERT(free_memory > ((PREFIX_SYM*)bind));
        
        ((PREFIX_SYM*)bind)->SetBracket(free_memory);
        PREFIX::ResetCorruptedBrackets(((PREFIX_SYM*)bind)+1,changed_suffix,old_end_of_word,free_memory);
        GlobalSubst.BindAlone(v,bind);
       };
     }
    return true;
   }
  else 
   {
    if (!(VAR::IsVar(bind))) { free_memory = (PREFIX_SYM*)bind; };
    return false; 
   };
 }; // bool OR_PROCESSOR::AlterBinding()
    
inline bool OR_PROCESSOR::BindWithOccCheck(VAR* v,bool& same_var)
 {
  CALL("BindWithOccCheck(VAR* v,bool& same_var)");
  if (OR_INDEX_PATH::ExtractTermWithout(v,same_var))
   {
    if (same_var) { return true; };
    if (OR_INDEX_PATH::VarExtracted())
     {
      GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted())); 
     }
    else // complex term extracted
     {
      PREFIX_SYM* word = free_memory;
      ASSERT(free_memory);
      
      OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
      word->SetBracket(free_memory);
      
      ASSERT(PREFIX::CorrectComplexTerm(word));
      
      /*********** WAS NOT COMMENTED OUT IN THE CASC-16 VERSION!
      if (OCCUR_CHECK::Occurs(v,word)) // debug
       {
        v->Output(); cout << " ----- ";
        PREFIX::OutputTerm(word); cout << "\n"; 
        OutputPath(); 
       };
      **********/      

      ASSERT(!(OCCUR_CHECK::Occurs(v,word)));
      
      GlobalSubst.BindAlone(v,(void*)word);
     };
    return true;
   }
  else 
   {
    return false;
   };
 }; // bool OR_PROCESSOR::BindWithOccCheck(VAR* v,bool& same_var)
 
inline bool OR_PROCESSOR::AlterBindingWithOccCheck(VAR* v,bool& same_var)
 {
  CALL("AlterBindingWithOccCheck(VAR* v,bool& same_var)");
  void* bind = v->Binding();
  bool was_var = true;
  if (bind) 
   {
    // here we suppose that v is the last bound variable
    ASSERT(v == GlobalSubst.LastBound());
    GlobalSubst.UnbindLast(); 
    if (!(VAR::IsVar(bind))) { was_var = false; free_memory = (PREFIX_SYM*)bind; };
   };
  if (OR_INDEX_PATH::ModifyTermWithout(v,same_var))
   {
    if (OR_INDEX_PATH::VarExtracted())
     {
      if (!(same_var)) { GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted())); };
     }
    else // complex term extracted
     {
      if (was_var)
       {
        PREFIX_SYM* word = free_memory;
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
        ASSERT(free_memory);
        
        word->SetBracket(free_memory);

        ASSERT(PREFIX::CorrectComplexTerm(word));
        ASSERT(!(OCCUR_CHECK::Occurs(v,word)));
        
        GlobalSubst.BindAlone(v,(void*)word);
       }
      else
       {
        PREFIX_SYM* changed_suffix = (PREFIX_SYM*)bind + (OR_INDEX_PATH::ChangedSuffix() - OR_INDEX_PATH::LastBookmark());
        free_memory = changed_suffix;
        
        ASSERT(free_memory);
        
        PREFIX_SYM* old_end_of_word = ((PREFIX_SYM*)bind)->Bracket();
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::ChangedSuffix(),free_memory);
        ASSERT(free_memory);
        
        ((PREFIX_SYM*)bind)->SetBracket(free_memory);
        PREFIX::ResetCorruptedBrackets(((PREFIX_SYM*)bind)+1,changed_suffix,old_end_of_word,free_memory);
       
        ASSERT(PREFIX::CorrectComplexTerm((PREFIX_SYM*)bind));
        ASSERT(!(OCCUR_CHECK::Occurs(v,(PREFIX_SYM*)bind)));

        GlobalSubst.BindAlone(v,bind);
       };
     }
    return true;
   }
  else 
   {
    return false; 
   };
 }; // bool OR_PROCESSOR::AlterBindingWithOccCheck(VAR* v,bool& same_var)

//================================================
#undef DEBUG_NAMESPACE
//==================================================================
#endif
