#ifndef OR_PREMISE_H
//=================================================================
#define OR_PREMISE_H
#ifndef NO_DEBUG_VIS 
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "clause.h"
#include "prefix.h" 
#include "or_index.h"
#include "NewClause.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_PREMISE
 #define DEBUG_NAMESPACE "OR_PREMISE"
#endif
#include "debugMacros.h"
//=================================================
class OR_PREMISE
 {
  public: // methods
   OR_PREMISE() : _resolvent((NewClause*)0) 
   {
    CALL("constructor OR_PREMISE()");
   };
   OR_PREMISE(NewClause* res) : _resolvent(res) 
   {
    CALL("constructor OR_PREMISE(NewClause* res)"); 
   };
   ~OR_PREMISE() { CALL("destructor ~OR_PREMISE()"); };
   void setResolvent(NewClause* res) { _resolvent = res; };
   PREFIX_SYM** MaxLiterals() { return max_literal; };
   const Clause* TheClause() const { return clause; };
   bool* Marks() { return mark; };
   ulong LastMax() { return last_max; };
   ulong Index() { return index; };
   ulong CutLit() { return cut_lit; };
   void Load(ulong ind,const Clause* c);
   void LoadOptimized(ulong ind,const Clause* c);
   void Unload() { free_memory = memory; };
   void Cut(ulong lit) 
    {
     cut_lit = lit;
     mark[cut_lit] = true;
    };
   void Load(ulong ind,RES_PAIR_LIST* res_pair)
    {
     CALL("Load(ulong ind,RES_PAIR_LIST* res_pair)");
     Load(ind,res_pair->TheClause());
     Cut((ulong)(res_pair->LitNum()));
    };
   void LoadOptimized(ulong ind,RES_PAIR_LIST* res_pair)
    {
     CALL("LoadOptimized(ulong ind,RES_PAIR_LIST* res_pair)");
     LoadOptimized(ind,res_pair->TheClause());
     Cut((ulong)(res_pair->LitNum()));
    };
   void Cut(TERM* lit) { Cut(clause->NumberOfLiteral(lit)); };
   void UnmarkCut() { mark[cut_lit] = false; };
   bool CollectNonmarkedLiterals();
   bool NonmarkedExist() { return nonmarkedExist; };
  public: // output for debugging
   #ifndef NO_DEBUG_VIS  
    ostream& output(ostream& str) const;
   #endif

  protected: // structure
   NewClause* _resolvent;
   const Clause* clause;
   ulong index;
   
   PREFIX_SYM memory[HellConst::MaxTermSize*HellConst::MaxNumOfLiterals];
   PREFIX_SYM* free_memory;
   
   PREFIX_SYM* max_literal[HellConst::MaxNumOfLiterals];
   bool inherentlySelected[HellConst::MaxNumOfLiterals];
   ulong last_max;
   bool mark[HellConst::MaxNumOfLiterals];

   //TERM* nonmax_literals;
   
   LiteralList* _allLiterals;
   LiteralList::Element* _firstNonselectedLit;

   ulong cut_lit;
   bool nonmarkedExist;

 }; // class OR_PREMISE
 
//******************* Definitions: *************************************

inline void OR_PREMISE::Load(ulong ind,const Clause* c)
 {
  CALL("Load(ulong ind,const Clause* c)");
  free_memory = memory;
  clause = c;
  index = ind;

  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  last_max = _allLiterals->numOfSelectedLits() - 1;
  ulong max_lit = 0; 
  for(LiteralList::Iterator iter(_allLiterals);iter.currentElement() != _firstNonselectedLit;iter.next())
   {
    max_literal[max_lit] = free_memory; 
    mark[max_lit] = false;
    StandardTermToPrefix(ind,iter.currentLiteral(),free_memory); 
    inherentlySelected[max_lit] = iter.currentElement()->isInherentlySelected();
    max_lit++;
   };   
 }; // void OR_PREMISE::Load(ulong ind,const Clause* c)

inline void OR_PREMISE::LoadOptimized(ulong ind,const Clause* c)
 {
  CALL("LoadOptimized(ulong ind,const Clause* c)");
  // if there is only one maximal literal we do not transform it
  // into prefix representation
  free_memory = memory;
  clause = c;
  index = ind;

  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  last_max = _allLiterals->numOfSelectedLits() - 1;

  if (last_max)
   {
    ulong max_lit = 0; 
    for(LiteralList::Iterator iter(_allLiterals);iter.currentElement() != _firstNonselectedLit;iter.next())
     {
      max_literal[max_lit] = free_memory; 
      mark[max_lit] = false;
      StandardTermToPrefix(ind,iter.currentLiteral(),free_memory); 
      inherentlySelected[max_lit] = iter.currentElement()->isInherentlySelected();
      max_lit++;
     };   
   }
  else
   {
    max_literal[0] = (PREFIX_SYM*)0; // just in case
    mark[0] = false; // just in case
   };
 }; // void OR_PREMISE::LoadOptimized(ulong ind,const Clause* c) 


inline bool OR_PREMISE::CollectNonmarkedLiterals()
 {
  CALL("CollectNonmarkedLiterals()");
  ulong currMax = 0;

  //nonmarkedExist = _firstNonselectedLit->isNotEndOf(_allLiterals);
  nonmarkedExist = (bool)_firstNonselectedLit;   

  while (currMax <= last_max)
   {    
    if (!(mark[currMax]))
     {
      nonmarkedExist = true;
      if (!(_resolvent->PushPrefixLit(max_literal[currMax],inherentlySelected[currMax]))) 
       {
        return false; 
       };
     };
    currMax++;
   };
  
   
  for(LiteralList::Iterator iter(_allLiterals,_firstNonselectedLit);iter.notEnd();iter.next())     
   if (!(_resolvent->PushStandardLit(index,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) return false;
  
  return true;
 }; // inline bool OR_PREMISE::CollectNonmarkedLiterals()

#ifndef NO_DEBUG_VIS  
 inline ostream& OR_PREMISE::output(ostream& str) const
 {
  str <<  "<<<<<< PREMISE >>>>>>>\n"
      << " Clause : " << clause << "\n Index = " << index
      << "\n Maximal literals: \n";
  for (ulong ml = 0; ml <= last_max; ml++)
   {
    str << "  [" << mark[ml] << "] ";
    PREFIX::outputTerm(str,max_literal[ml]);
    str << '\n';
   };
  return str; 
 }; // ostream& OR_PREMISE::output(ostream& str) const
#endif

//================================================
#undef DEBUG_NAMESPACE
//=================================================================
#endif
