//
// File:         premise_factoring.cpp
// Description:  OR premises with factoring.
// Created:      Feb 15, 2000, 16:50
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "premise_factoring.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PREMISE_FACTORING
 #define DEBUG_NAMESPACE "OR_PREMISE_WITH_FACTORING"
#endif
#include "debugMacros.h"
//=================================================
bool OR_PREMISE_WITH_FACTORING::GetFirstFactor(ulong first_candidate)
 {
  if (FactoringPossible())
   {
    factored.reset();
    if (FindFirstFactor(first_candidate))
     {
      last_factor = first_candidate;
      was_factored = true;
      return true;
     };
   };
  return false;
 }; // bool OR_PREMISE_WITH_FACTORING::GetFirstFactor(ulong first_candidate)
 
bool OR_PREMISE_WITH_FACTORING::GetNextFactor()
 {
  ulong candidate = last_factor + 1;
  next_factor:
   if (FindFirstFactor(candidate))
    {
     factored.push(last_factor);
     last_factor = candidate;
     return true;
    }
   else // backtrack needed
    {
     GlobalSubst.UnbindLastSection();
     mark[last_factor] = false;
     candidate = last_factor + 1;
     if (factored.empty())
      {
       if (FindFirstFactor(candidate))
        {
         last_factor = candidate;
         return true;
        }
       else // no more factors at all
        {
         return false;
        };
      }
     else 
      {
       last_factor = factored.pop();
       goto next_factor;        
      };
    };
 }; // bool OR_PREMISE_WITH_FACTORING::GetNextFactor()
 
bool OR_PREMISE_WITH_FACTORING::FindFirstFactor(ulong& candidate)
 {
  while (candidate <= last_max)
   {
    if ((candidate != cut_lit) 
       && (UNIFICATION::UnifyComplex(max_literal[candidate],cut_literal)))   
     {
      // factor found
      mark[candidate] = true;
      return true;
     };
    // continue search
    candidate++;
   };
  // no factor found 
  return false;
 }; // bool OR_PREMISE_WITH_FACTORING::FindFirstFactor(ulong& candidate)
//======================================================================
