//
// File:         premise_factoring.h
// Description:  OR premises with factoring.
// Created:      Feb 15, 2000, 16:35
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef PREMISE_FACTORING_H
//======================================================================
#define PREMISE_FACTORING_H
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"
#include "Stack.h"
#include "or_premise.h"
#include "prefix.h"
#include "NewClause.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PREMISE_FACTORING
 #define DEBUG_NAMESPACE "OR_PREMISE_WITH_FACTORING"
#endif
#include "debugMacros.h"
//=================================================

class OR_PREMISE_WITH_FACTORING : public OR_PREMISE
 {
  private:
   Stack<ulong,HellConst::MaxNumOfLiterals> factored;
   
   PREFIX_SYM* cut_literal; // the literal itself in prefix representation
   
   ulong first_factor;
   ulong last_factor;
   bool was_factored;
  public:
   OR_PREMISE_WITH_FACTORING() : OR_PREMISE(),factored("factored") 
   {
    CALL("constructor OR_PREMISE_WITH_FACTORING()");
   };
   OR_PREMISE_WITH_FACTORING(NewClause* res) : OR_PREMISE(res),factored("factored") 
   {
    CALL("constructor OR_PREMISE_WITH_FACTORING(NewClause* res)"); 
   };
   ~OR_PREMISE_WITH_FACTORING() { CALL("destructor ~OR_PREMISE_WITH_FACTORING()"); };
   void Destroy() 
   {
    CALL("Destroy()");
   };
   #ifndef NO_DEBUG
    bool clean() const 
    {
     return true;
    };
   #endif   


   void assignResolvent(NewClause* res) { OR_PREMISE::setResolvent(res); };
   void ResetFactoring()
    {
     was_factored = false;
     if (FactoringPossible())
      {
       factored.reset();
       cut_literal = max_literal[cut_lit];
      };
    };
   bool WasFactored() { return was_factored; };
   bool FactoringPossible() { return (bool)LastMax(); };
   bool GetFirstFactor(ulong first_candidate);
   bool GetNextFactor();
  private: // aux. methods
   bool FindFirstFactor(ulong& candidate);
 }; // class OR_PREMISE_WITH_FACTORING

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PREMISE_FACTORING
 #define DEBUG_NAMESPACE "OR_MULTIPREMISE<ulong numOfPremises>"
#endif
#include "debugMacros.h"
//================================================= 
template <ulong numOfPremises> 
 class OR_MULTIPREMISE  // for hyperresolution
  {
   private:
    OR_PREMISE_WITH_FACTORING premise[numOfPremises];
    Stack<ulong,numOfPremises> loaded;
   private: // might be changed in the future
    ulong next;
   private:
    OR_MULTIPREMISE() {};
   public:
    OR_MULTIPREMISE(NewClause* res) : loaded("loaded") 
    {
     next = 0;
     for (int i = 0; i < numOfPremises; i++) premise[i].setResolvent(res); 
    };
    ~OR_MULTIPREMISE() { CALL("destructor ~OR_MULTIPREMISE()"); };
    void Load(ulong ind,Clause* cl,TERM* cut_liter)
     {
      CALL("Load(ulong ind,Clause* cl,TERM* cut_liter)");
      premise[next].LoadOptimized(ind,cl);
      premise[next].Cut(cut_liter);
      loaded.push(next);
      next++;
     }; 
    void Load(ulong ind,RES_PAIR_LIST* res_pair)
     {
      CALL("Load(ulong ind,RES_PAIR_LIST* res_pair)");
      premise[next].LoadOptimized(ind,res_pair);
      loaded.push(next);
      next++;
     }; 
    void UnloadLast() 
     {
      next = loaded.pop();
      premise[next].Unload();
     };
    void Reset() 
     {
      while (!loaded.empty()) 
       {
        premise[loaded.pop()].Unload();
       };
      next = 0;
     };
    
    bool CollectNonmarkedLiterals()
     {
      ulong* end = loaded.end(); 
      for (ulong* i = loaded.begin(); i < end; i++)
       {
        if (!(premise[*i].CollectNonmarkedLiterals())) return false;
       };
      return true; // all literals successfully collected
     };
    void ResetFactoring()
     {
      ulong* end = loaded.end(); 
      for (ulong* i = loaded.begin(); i < end; i++)
       premise[*i].ResetFactoring();
     };
    bool NextFactor()
     {
      CALL("NextFactor()");
      OR_PREMISE_WITH_FACTORING* prem;
      ulong* end = loaded.end(); 
      for (ulong* i = loaded.begin(); i < end; i++)
       {
        prem = &(premise[*i]);
        if (prem->WasFactored())
         {
          if (prem->GetNextFactor())
           {
            return true;
           }
          else
           {
            prem->ResetFactoring();
           };
         }
        else 
         {
          if (prem->GetFirstFactor(0))
           {
            return true;
           };
         };
       };
      return false; // no more factors 
     };

   void WriteAllClauses(Clause** dest)
    {
     ulong* end = loaded.end(); 
     for (ulong* i = loaded.begin(); i < end; i++)
      {
       *dest = premise[*i].TheClause();
       dest++;
      };
    };
  }; // class OR_MULTIPREMISE<ulong numOfPremises>
//================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
