//
// File:         sharing.h
// Description:  Agressive sharing.
// Created:      Feb 15, 2000, 17:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef SHARING_H
//===========================================================
#define SHARING_H
#include <limits.h>
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "GlobAlloc.h"                 
#include "term.h"
#include "Flatterm.h"
#include "Stack.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"
#endif
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SHARING
 #define DEBUG_NAMESPACE "SHARING_INDEX"
#endif
#include "debugMacros.h"
//=================================================
class SHARING_TREE_NODE;
class SHARING_INDEX
 {  
  private:
   typedef Stack<TERM*,HellConst::MaxTermSize> PosStack;
  private: // auxilliary data
   TERM::Iterator t_iter;
   PosStack subterms;
  private: // structure
   SHARING_TREE_NODE* nonconst_index[HellConst::MaxNumOfFunctors];
   TERM const_index[HellConst::MaxNumOfFunctors*2];
  public:
   class SURFER; 
  public:
   SHARING_INDEX() : t_iter(), subterms("subterms")
    {
     CALL("constructor SHARING_INDEX()");

     for (unsigned long f = 0; f < HellConst::MaxNumOfFunctors; f++)
      {
       nonconst_index[f] = (SHARING_TREE_NODE*)0;
      };
     for (unsigned long c = 0; c < HellConst::MaxNumOfFunctors; c++)
      {
       //const_index[c].MakeComplex((TERM::Arity)0,(TERM::Functor)c,TERM::PositivePolarity);
       const_index[2*c].MakeComplex((TERM::Arity)0,(TERM::Functor)c,TERM::PositivePolarity);
       //const_index[2*c+1].SetFirst((TERM*)0);
       const_index[2*c].InitOccurences();
      };
    };
   ~SHARING_INDEX() { CALL("destructor ~SHARING_INDEX()"); };
   void Destroy() 
    {
     CALL("Destroy()");    
     for (unsigned long f = 0; f < HellConst::MaxNumOfFunctors; f++)
      {
       DestroyTree(nonconst_index[f]);
       nonconst_index[f] = (SHARING_TREE_NODE*)0;
      };
     for (unsigned long c = 0; c < HellConst::MaxNumOfFunctors; c++)
      {
       const_index[2*c].DestroyOccurences();
      }; 
    }; 

   #ifndef NO_DEBUG
    bool clean() const 
    {
     for (unsigned long f = 0; f < HellConst::MaxNumOfFunctors; f++)
      if (nonconst_index[f]) return false;    
     for (unsigned long c = 0; c < HellConst::MaxNumOfFunctors; c++)
      if (const_index[2*c].Occurences()) return false;
     return true;
    }; 
   #endif

  public:
   SHARING_TREE_NODE** Tree(const TERM& fun)
    {
     CALL("Tree(const TERM& fun)");
     ASSERT(fun.arity());
     return nonconst_index + fun.functor();
    };
  private:
   SHARING_TREE_NODE* CreateBranch(TERM* whole);
   TERM* IntegrateComplexNonconst(TERM* t);
   TERM* IntegrateFlatComplexNonconst(const Flatterm* t);
   static void DestroyTree(SHARING_TREE_NODE* tree);
  public: // main methods
   TERM* IntegrateConst(TERM* c)
    {
     CALL("IntegrateConst(TERM* c)");
     return &(const_index[2*(c->functor())]);
    }; // TERM* IntegrateConst(TERM* c)
   TERM* Const(const TERM& c)
    { 
     CALL("Const(const TERM& c)");
     return &(const_index[2*(c.functor())]);
    };

   TERM* IntegrateComplex(TERM* t);
   TERM* IntegrateFlatComplex(const Flatterm* t);
   TERM* CopyLiteral(TERM* source);
   TERM* CopyFlatLiteral(const Flatterm* source);
   TERM* CopyFlatEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2);
   void DestroyLiteralCopy(TERM* t);
   TERM* CopyComplex(TERM* source);
   void DestroyCopy(TERM* t); 
   void RemoveComplex(TERM* t);
   void RemoveComplexNonconst(TERM* t);
   
   void ScanIndex(ostream& str); // for debugging    
   void ScanTree(ostream& str,SHARING_TREE_NODE* tree,ulong& numOfNodes); // for debugging    
 }; // class SHARING_INDEX

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SHARING
 #define DEBUG_NAMESPACE "SHARING_TREE_NODE"
#endif
#include "debugMacros.h"
//================================================= 
class SHARING_TREE_NODE
 {
  
  private: // structure
   TERM tag;
   union
    {
     SHARING_TREE_NODE* below;
     long num;
    };
   union
    {
     SHARING_TREE_NODE* alternative;
     long dummy;
    };
   
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    static ClassDesc _classDesc;
   #endif  

    /**************
   public:
    static ulong _allocated; // for debugging
    static ulong _deallocated; // for debugging
  ************/
    
  public: // construction and destruction
   SHARING_TREE_NODE(const TERM& t) : tag(t) // for making intermediate nodes
    {
     CALL("SHARING_TREE_NODE(const TERM& t)");
     below = (SHARING_TREE_NODE*)0;
     alternative = (SHARING_TREE_NODE*)0;
    };
   SHARING_TREE_NODE(TERM* t) : tag(t) // for making leaves
    {
     CALL("SHARING_TREE_NODE(TERM* t)");
     num = 1;
    };
   void* operator new(size_t)
    {
     CALL("operator new(size_t)");
     //_allocated++;

     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_classDesc.registerAllocated(sizeof(SHARING_TREE_NODE)));
      return GlobAlloc::allocate(sizeof(SHARING_TREE_NODE),&_classDesc);
     #else
      return GlobAlloc::allocate(sizeof(SHARING_TREE_NODE));
     #endif
    };
   void operator delete(void* obj)
    {
     CALL("operator delete(void* obj)");
     //_deallocated++;

     #ifdef DEBUG_ALLOC_OBJ_TYPE 
      ALWAYS(_classDesc.registerDeallocated(sizeof(SHARING_TREE_NODE)));
      GlobAlloc::deallocate(obj,sizeof(SHARING_TREE_NODE),&_classDesc);
     #else
      GlobAlloc::deallocate(obj,sizeof(SHARING_TREE_NODE));
     #endif
    };
  public:
   static ulong minSize() { return sizeof(SHARING_TREE_NODE); };
   static ulong maxSize() { return sizeof(SHARING_TREE_NODE); };   
   bool IsLeaf() { return tag.IsReference(); };
   TERM& Tag() { return tag; };
   const TERM& Tag() const { return tag; };
    // methods for leaves
   TERM* Term() { return tag.First(); }; 
   long Num() { return num; };
   void IncNum() { if (num < LONG_MAX) { num++; }; };
   void DecNum() { if (num < LONG_MAX) { num--; }; };
    // methods for intermediate nodes
   SHARING_TREE_NODE* Below() { return below; };
   SHARING_TREE_NODE** BelowAddr() { return &below; };
   SHARING_TREE_NODE* Alternative() { return alternative; };
   SHARING_TREE_NODE** AlternativeAddr() { return &alternative; };
   void SetBelow(SHARING_TREE_NODE* n) { below = n; };  
   void SetAlternative(SHARING_TREE_NODE* n) { alternative = n; }; 
 }; // class SHARING_TREE_NODE  

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SHARING
 #define DEBUG_NAMESPACE "SHARING_INDEX::SURFER"
#endif
#include "debugMacros.h"
//================================================= 
class SHARING_INDEX::SURFER
 {
  private:
   SHARING_INDEX* index;
   SHARING_TREE_NODE** node_addr;
  public:
   SURFER(SHARING_INDEX* si) : index(si) {};
   bool Reset(const TERM& fun)
    {
     CALL("Reset(const TERM& fun)");
     node_addr = index->Tree(fun);
     return (bool)(*node_addr);
    };
   bool Read(const TERM& sym)
    {
     check_tag:
      if ((*node_addr)->Tag() == sym)
       {
        node_addr = (*node_addr)->BelowAddr();
        return true;
       }
      else 
       {
        node_addr = (*node_addr)->AlternativeAddr();
        if (*node_addr) { goto check_tag; }
        else return false;
       };
    }; // bool Read(const TERM& sym) 
   void Write(const TERM& sym)
    {
     *node_addr = new SHARING_TREE_NODE(sym);
     node_addr = (*node_addr)->BelowAddr();
    }; // void Write(const TERM& sym)
   void CreateLeaf(TERM* t) { *node_addr = new SHARING_TREE_NODE(t); };
   SHARING_TREE_NODE* Leaf() { return (*node_addr); }; 
 }; // class SHARING_INDEX::SURFER

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SHARING
 #define DEBUG_NAMESPACE "SHARING_INDEX"
#endif
#include "debugMacros.h"
//================================================= 
 
inline SHARING_TREE_NODE* SHARING_INDEX::CreateBranch(TERM* whole)
 {
  CALL("CreateBranch(TERM* whole)");
  SHARING_TREE_NODE* result = new SHARING_TREE_NODE(t_iter.Symbol());
  SHARING_TREE_NODE* node = result;
  while (t_iter.Next())
   {
    node->SetBelow(new SHARING_TREE_NODE(t_iter.Symbol()));
    node = node->Below();
   };
  node->SetBelow(new SHARING_TREE_NODE(whole));
  return result; 
 }; // SHARING_TREE_NODE* SHARING_INDEX::CreateBranch(TERM* whole)
 
 

inline TERM* SHARING_INDEX::IntegrateFlatComplexNonconst(const Flatterm* t)
 {
  CALL("IntegrateFlatComplexNonconst(const Flatterm* t)");
  ASSERT(t->Symbol().arity());
  
  Flatterm::Iterator iter;
  iter.Reset(t);
  SURFER surf(this);
  SHARING_TREE_NODE* leaf;
  TERM sym;
  
  if (surf.Reset(t->Symbol()))
   {
    read_mode_next_sym:
     iter.NextSym();
     if (iter.EndOfTerm()) // the term is in the index already
      {
       leaf = surf.Leaf();
       
       ASSERT(leaf);
       ASSERT(leaf->IsLeaf());
  
       leaf->IncNum();
       return leaf->Term();
      }; 
     sym = iter.CurrPos()->Symbol();
     if (surf.Read(sym)) { goto read_mode_next_sym; };
     surf.Write(sym);
     goto write_mode_next_sym;
   }
  else
   {
    write_mode_next_sym:
     iter.NextSym();
     if (iter.EndOfTerm()) // branch created, leaf with a copy of the term must be created also
      {
       TERM* copy = new (t->Symbol().arity()) TERM(t->Symbol());

       // copy->Info()->SetFirst((TERM*)0);
       copy->InitOccurences();

       surf.CreateLeaf(copy);
       TERM* targ = copy->Args();
       iter.Reset(t);
       iter.NextSym();
       copy_argument:
        if (iter.CurrPos()->Symbol().isVariable())
         {
          *targ = iter.CurrPos()->Symbol();
         }
        else // complex term
         {
          if (iter.CurrPos()->Symbol().arity())
           {
            targ->SetFirst(IntegrateFlatComplexNonconst(iter.CurrPos()));
           }
          else // constant
           {
            //targ->SetFirst(&(const_index[iter.CurrPos()->Symbol().functor()]));
            targ->SetFirst(Const(iter.CurrPos()->Symbol())); 
           }; 
         }; 
        iter.after();
        if (iter.EndOfTerm())
         {
          return copy;
         };
        targ = targ->Next(); 
        goto copy_argument;
       
      };
     surf.Write(iter.CurrPos()->Symbol());
     goto write_mode_next_sym;
   };
 }; // TERM* SHARING_INDEX::IntegrateFlatComplexNonconst(const Flatterm* t)

inline TERM* SHARING_INDEX::IntegrateComplexNonconst(TERM* t)
 {
  CALL("IntegrateComplexNonconst(TERM* t)");
  ASSERT(t->arity());
  TERM::Iterator iter;
  iter.Reset(t);
  SURFER surf(this);
  SHARING_TREE_NODE* leaf;
  TERM sym; 
  
  if (surf.Reset(*t))
   {
    read_mode_next_sym:
     if (iter.Next())
      {
       sym = iter.Symbol();
       if (surf.Read(sym)) { goto read_mode_next_sym; }
       else
        {
         surf.Write(sym);
         goto write_mode_next_sym;
        };
      }
     else // the term is in the index already
      {
       leaf = surf.Leaf();
       
       ASSERT(leaf);
       ASSERT(leaf->IsLeaf());
  
       leaf->IncNum();
       return leaf->Term();
      }; 
   }
  else
   {
    write_mode_next_sym:
     if (iter.Next())
      {
       surf.Write(iter.Symbol());
       goto write_mode_next_sym;
      }
     else // branch created, leaf with a copy of the term must be created also          
      {
       TERM* copy = t->CopyTop();
     

       //copy->Info()->SetFirst((TERM*)0);
       copy->InitOccurences();


       TERM* src = t->Args();
       TERM* targ = copy->Args();
       check_arg:
        switch (src->Flag())
         {
          case TermVar:
           *targ = *src;
           src = src->Next();
           targ = targ->Next();
           goto check_arg;
          case TermRef:
           if (src->First()->arity())
            {
             *targ = *src;
             subterms.push(targ);
            }
           else // constant
            {
             //targ->SetFirst(&(const_index[src->First()->functor()]));
             targ->SetFirst(Const(*(src->First())));
            };
           src = src->Next();
           targ = targ->Next();
           goto check_arg;
          case TermComplex:
           surf.CreateLeaf(copy); 
           return copy;
         };
        return (TERM*)0; // to surpress warnings 
      };
   };
 }; // TERM* SHARING_INDEX::IntegrateComplexNonconst(TERM* t) 


inline TERM* SHARING_INDEX::IntegrateComplex(TERM* t)
 {
  if (t->arity())
   {
    subterms.reset();
    TERM* result = IntegrateComplexNonconst(t);
    TERM* subt;
    while (!(subterms.empty()))
     {
      subt = subterms.pop();
      subt->SetFirst(IntegrateComplexNonconst(subt->First()));
     };
    return result;
   }
  else // constant
   {
    //return &(const_index[t->functor()]);
    return Const(*t);
   };
 }; // TERM* SHARING_INDEX::IntegrateComplex(TERM* t)


inline TERM* SHARING_INDEX::IntegrateFlatComplex(const Flatterm* t)
 {
  CALL("IntegrateFlatComplex(const Flatterm* t)");
  if (t->Symbol().arity())
   {
    return IntegrateFlatComplexNonconst(t);
   }
  else // constant
   {
    //return &(const_index[t->Symbol().functor()]);
    return Const(t->Symbol());
   };
 }; // TERM* SHARING_INDEX::IntegrateFlatComplex(const Flatterm* t)


inline TERM* SHARING_INDEX::CopyLiteral(TERM* source)
 {
  CALL("CopyLiteral(TERM* source)");
  TERM* target = source->CopyLitTop();
  source = source->Args();
  TERM* targ = target->Args();
  copy_arg:
   switch (source->Flag())
    {
     case TermVar:
      *targ = *source;
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TermRef:
      targ->SetFirst(IntegrateComplex(source->First()));
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TermComplex:
      return target;
    };
  return (TERM*)0; // to surpress warnings
 }; // TERM* SHARING_INDEX::CopyLiteral(TERM* source)
 
inline TERM* SHARING_INDEX::CopyFlatLiteral(const Flatterm* source)
 {
  CALL("CopyFlatLiteral(const Flatterm* source)");
  TERM* target = TERM::AllocLit(source->Symbol());
  TERM* targ = target->Args();
  Flatterm::Iterator iter;
  iter.Reset(source);
  iter.NextSym();
  copy_arg:
   if (iter.EndOfTerm()) { return target; };
   if (iter.CurrPos()->Symbol().isVariable())
    {
     *targ = iter.CurrPos()->Symbol(); 
    }
   else // complex
    {
     targ->SetFirst(IntegrateFlatComplex(iter.CurrPos()));
    }; 
   iter.after();
   targ = targ->Next(); 
   goto copy_arg;
 }; // TERM* SHARING_INDEX::CopyFlatLiteral(const Flatterm* source)

inline TERM* SHARING_INDEX::CopyFlatEq(const TERM& header,
                                       const Flatterm* arg1,
                                       const Flatterm* arg2)
 {
  CALL("CopyFlatEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2)");
  TERM* result = TERM::AllocLit(header);
  TERM* rarg = result->Args();
  if (arg1->Symbol().isVariable()) { *rarg = arg1->Symbol(); }
  else rarg->SetFirst(IntegrateFlatComplex(arg1));
  rarg = rarg->Next();
  if (arg2->Symbol().isVariable()) { *rarg = arg2->Symbol(); }
  else rarg->SetFirst(IntegrateFlatComplex(arg2));
  return result;
 }; // TERM* SHARING_INDEX::CopyFlatEq(const TERM& header,..  
 
inline void SHARING_INDEX::DestroyLiteralCopy(TERM* t)
 { 
  CALL("DestroyLiteralCopy(TERM* t)");
  for (TERM* arg = t->Args(); ; arg = arg->Next())
   {
    switch (arg->Flag())
     {
      case TermVar : break;
      case TermRef : 
       RemoveComplex(arg->First());
       break;
      case TermComplex : 
       t->DeleteLit();
       //delete t;
       return;
     };
   };
 }; // void SHARING_INDEX::DestroyLiteralCopy(TERM* t)     
 
 
inline TERM* SHARING_INDEX::CopyComplex(TERM* source)
 {
  CALL("CopyComplex(TERM* source)");
  TERM* target = source->CopyTop();


  //target->Info()->SetFirst((TERM*)0);
  target->InitOccurences();

  source = source->Args();
  TERM* targ = target->Args();
  copy_arg:
   switch (source->Flag())
    {
     case TermVar:
      *targ = *source;
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TermRef:
      targ->SetFirst(IntegrateComplex(source->First()));
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TermComplex:
      return target;
    };
  return (TERM*)0; // to surpress warnings
 }; // TERM* SHARING_INDEX::CopyComplex(TERM* source)
  
inline void SHARING_INDEX::DestroyCopy(TERM* t)
 {
  CALL("DestroyCopy(TERM* t)");
  for (TERM* arg = t->Args(); ; arg = arg->Next())
   {
    switch (arg->Flag())
     {
      case TermVar : break;
      case TermRef : 
       RemoveComplex(arg->First());
       break;
      case TermComplex : 
       delete t;
       return;
     };
   };
 }; // void SHARING_INDEX::DestroyCopy(TERM* t)     
 
  
inline void SHARING_INDEX::RemoveComplex(TERM* t)
 {
  CALL("RemoveComplex(TERM* t)");
  if (t->arity())
   {
    RemoveComplexNonconst(t);
   };
 }; // void SHARING_INDEX::Remove(TERM* t)
  
inline void SHARING_INDEX::RemoveComplexNonconst(TERM* t)
 {
  CALL("RemoveComplexNonconst(TERM* t)");
  ASSERT((t->Flag() == TermComplex) && (t->arity()));
  
  Stack<TERM*,HellConst::MaxTermSize> subterms("subterms(local)");
  TERM* subterm;
  SHARING_TREE_NODE* node;     
  SHARING_TREE_NODE** branch;
  SHARING_TREE_NODE** below_addr; 
  TERM::Iterator iter;
  SHARING_TREE_NODE* branch_to_delete;
  SHARING_TREE_NODE* node_to_delete;
 
  process_next_term: 
  
   ASSERT_IN((t->Flag() == TermComplex) && (t->arity()),"A2"); 
  
   node = nonconst_index[t->functor()];  
   
   ASSERT_IN(node,"A3"); 
      
   branch = &(nonconst_index[t->functor()]);
   below_addr = branch;
    iter.Reset(t);
    iter.Next();
    do
     {
           
      ASSERT_IN(!(node->IsLeaf()),"A4"); 
      if (node->Alternative())
       {
        branch = below_addr;
        check_node:
         if (node->Tag() == iter.Symbol())
          {
           below_addr = node->BelowAddr();
           node = node->Below();
          }
         else // try alternative
          {
           branch = node->AlternativeAddr();
           node = node->Alternative();
           ASSERT(node);
           goto check_node;
          };
       }
      else
       {
	ASSERT(node->Tag() == iter.Symbol());
        below_addr = node->BelowAddr();
        node = node->Below();
       };
     }
    while (iter.Next());
   
    // here node is a leaf
   
    node->DecNum();

    if (!(node->Num())) 
     {
      // last occurence of the term, top level of the term must be destroyed,
      // the branch must be deleted
     
      subterm = node->Term()->Args();
      while (!(subterm->IsNull()))
       {
        if ((subterm->IsReference()) && (subterm->First()->arity()))
         {
          subterms.push(subterm->First());
         };
        subterm = subterm->Next();
       }; 
     
      delete node->Term(); 
      
      branch_to_delete = *branch;
      *branch = (*branch)->Alternative(); 
      while (!(branch_to_delete->IsLeaf()))
       {
        node_to_delete = branch_to_delete;
        branch_to_delete = branch_to_delete->Below();
        delete node_to_delete;
       };
      delete branch_to_delete; // leaf

      /*********
      if (!(subterms.empty()))
       {
        t = subterms.pop();
        goto process_next_term;
       };
      *********/
     };

    if (!(subterms.empty()))    // was ommited in vampire 1.31 thus giving a serious memory leak
     {
      t = subterms.pop();
      goto process_next_term;
     };     
}; // void SHARING_INDEX::RemoveComplexNonconst(TERM* const t)

inline void SHARING_INDEX::DestroyTree(SHARING_TREE_NODE* tree)
 {
  CALL("DestroyTree(SHARING_TREE_NODE* tree)");
  if (tree)
   {
    if (tree->IsLeaf()) 
     {
      ASSERT(tree->Term()->isComplex());
      tree->Term()->DestroyOccurences(); 
      delete tree->Term();
      delete tree; 
     }
    else
     {
      if (tree->Alternative()) DestroyTree(tree->Alternative());
      if (tree->Below()) DestroyTree(tree->Below());
      delete tree;
     };
   };
 }; // void SHARING_INDEX::DestroyTree(SHARING_TREE_NODE* tree)

inline void SHARING_INDEX::ScanIndex(ostream& str)
{
 CALL("ScanIndex()");
 ulong numOfNodes = 0;
 for (unsigned long f = 0; f < HellConst::MaxNumOfFunctors; f++)
  {
   ScanTree(str,nonconst_index[f],numOfNodes); 
  };
 str << "NUMBER OF NODES: " << numOfNodes << "\n";
}; // void SHARING_INDEX::ScanIndex(ostream& str)

inline void SHARING_INDEX::ScanTree(ostream& str,SHARING_TREE_NODE* tree,ulong& numOfNodes)
{
 CALL("ScanTree(SHARING_TREE_NODE* tree,ulong& numOfNodes)");
 if (!tree) return;
 numOfNodes++;
 if (tree->IsLeaf()) 
  {
   str << tree->Num() << " x " << tree->Term() << "\n";
   
   /********
   str << "       OCC: ";
   if (tree->Term()->Occurences())
    {
     str << "{";
     for (TERM::OccList::Node* node = tree->Term()->Occurences()->first(); 
          node;
          node = node->next())
      {
       str << node->key() << " ";
      };
     str << "}";
    }
   else
    {
     str << "{}";
    };
   str << "\n";
   ********/

   return;
  }; 
 ScanTree(str,tree->Below(),numOfNodes);
 ScanTree(str,tree->Alternative(),numOfNodes);
}; // void SHARING_INDEX::ScanTree(ostream& str,SHARING_TREE_NODE* tree,ulong& numOfNodes)



//================================================
#undef DEBUG_NAMESPACE
//===========================================================
#endif








