#ifndef TMP_LIT_LIST_H
//======================================================================
#define TMP_LIT_LIST_H
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"
#include "variables.h"
#include "MultisetOfVariables.h"
#include "weighting.h"
#include "Flatterm.h"
#include "tmp_literal.h"
extern HEADER_WEIGHTING GlobalHeaderWeighting;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEW_CLAUSE
 #define DEBUG_NAMESPACE "TMP_LIT_LIST"
#endif 
#include "debugMacros.h"
//=================================================

class TMP_LIT_LIST
 {
  public:
   class STAT
    {
     public: 
      ulong negative;
      ulong equations;
      ulong noneq_lits;
      ulong inherently_selected;
      ulong negative_splitting_branches;
      ulong positive_splitting_branches;
      TERM max_noneq_header;
     public:
      STAT() { Reset(); };
      void Reset()
       {
        negative = 0;
        equations = 0;
        noneq_lits = 0;
        inherently_selected = 0;
        negative_splitting_branches = 0;
        positive_splitting_branches = 0;
       }; 
    };
  public: 
   class CONCAT;
  private:
   TMP_LITERAL* literal[HellConst::MaxNumOfLiterals];
   TMP_LITERAL** next_literal;
  public: 
   TMP_LIT_LIST() : next_literal(literal) {};
   TMP_LIT_LIST(const TMP_LIT_LIST& lst) : next_literal(literal)
    {
     for (TMP_LITERAL* const * lit = lst.literal; lit < lst.next_literal; lit++) Push(*lit);
    };
   ~TMP_LIT_LIST() { CALL("destructor ~TMP_LIT_LIST()"); };
   TMP_LIT_LIST& operator=(const TMP_LIT_LIST& lst)
    {
     next_literal = literal;
     for (TMP_LITERAL* const * lit = lst.literal; lit < lst.next_literal; lit++) Push(*lit);
     return *this;
    };
   
   void Reset() { next_literal = literal; };
   void Push(TMP_LITERAL* lit) { *next_literal = lit; next_literal++; };

   void CollectFrom(TMP_LIT_LIST& lst)
   {
    while (lst.Nonempty()) Push(lst.Pop());
   };

   TMP_LITERAL* Pop() { next_literal--; return *next_literal; };
   void ReplaceByTop(TMP_LITERAL** pos) { next_literal--; *pos = *next_literal; };
   bool Empty() const { return next_literal == literal; };
   bool Nonempty() const { return next_literal != literal; };
   ulong Length() const { return next_literal - literal; };
   TMP_LITERAL** First() { return literal; };
   TMP_LITERAL** Last()  { return next_literal - 1; };
   TMP_LITERAL* Nth(ulong n) const
    {
     return literal[n];
    };                

   long Weight() const 
    {
     long res = 0L;
     for (const TMP_LITERAL* const * lit = literal; lit < next_literal; lit++) res += (*lit)->size();
     return res;
    }; 

   bool SubstPossible(ulong var,unsigned subst_depth,long subst_weight)
    {
     CALL("SubstPossible(ulong var,unsigned subst_depth,long subst_weight)");
     for (TMP_LITERAL** lit = First(); lit <= Last(); lit++)
      {
       if (!(*lit)->SubstPossible(var,subst_depth,subst_weight)) return false; 
      };    
     return true;
    };

   void WeightAndNumOfPosEq(long& weight,long& num_of_pos_eq)
    {
     CALL("WeightAndNumOfPosEq(long& weight,long& num_of_pos_eq)");
     weight = 0;
     num_of_pos_eq = 0;
     for (const TMP_LITERAL* const * lit = literal; lit < next_literal; lit++) 
      {
       const TMP_LITERAL* curr_lit = *lit;
       weight += curr_lit->size();
       if (curr_lit->IsEquality() && curr_lit->Positive()) num_of_pos_eq++;
      };
    };
    
   long MinFactorWeightUpperBound() const { return Weight(); }; 

  
   long MinPosEqLHSWeight()
    { 
     long res = 0;
     long pemlw;
     for (const TMP_LITERAL* const * lit = literal; lit < next_literal; lit++) 
      if ((*lit)->IsEquality() && (*lit)->Positive()) 
       {
        res = (*lit)->PosEqMinLHSSize();
        lit++;
        for (; lit < next_literal; lit++) 
         if ((*lit)->IsEquality() && (*lit)->Positive())
	  {
	   pemlw = (*lit)->PosEqMinLHSSize();
           if (pemlw < res) res = pemlw;
	  };
	return res;
       };
     return res; 
    };


    
   void CollectStat(STAT& stat,bool inversePolarityForSelection);
   void CollectStatWithSplitting(STAT& stat,bool inversePolarityForSelection);

   bool Eliminate(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,const bool* eliminated);


   void Order(TMP_LIT_LIST& result,const VAR_WEIGHTING& var_weighting);
   void CreateVarWeighting(VAR_WEIGHTING& var_weighting);
   void RenameVariables(INT_VAR_RENAMING& var_renaming);
   void CollectGround(TMP_LIT_LIST& component);
   void CollectNextComponent(TMP_LIT_LIST& component);
 
   void CollectVariables(MultisetOfVariables<HellConst::MaxNumOfVariables>& vars) 
    {
     for (TMP_LITERAL** litp = literal; litp < next_literal; litp++)   
      {
       vars.Add((*litp)->Variables());
      };  
    };
  
   bool Weed(bool& eqTautology);
   bool Weed(const Clause* const * commutativity_table,bool& eqTautology);

  public:
   ostream& output(ostream& str) const 
   {
    for (const TMP_LITERAL* const* lit = literal; lit < next_literal; lit++)
     {
       str << *lit; if (lit + 1 < next_literal) str << " % ";
     };      
    return str;
   };
  private: 
   bool PositiveForSelection(const TERM& header,bool inversePolarityForSelection)
   {
    if (header.IsEquality() || header.IsSplitting() || (!inversePolarityForSelection)) return header.Positive();
    return header.Negative();
   }; 
   bool NegativeForSelection(const TERM& header,bool inversePolarityForSelection)
   {
    if (header.IsEquality() || header.IsSplitting() || (!inversePolarityForSelection)) return header.Negative();
    return header.Positive();
   }; 
 }; // class TMP_LIT_LIST


inline ostream& operator<<(ostream& str,const TMP_LIT_LIST& lst) 
{
 return lst.output(str);
};

inline ostream& operator<<(ostream& str,const TMP_LIT_LIST* lst) 
{
 return lst->output(str);
};


class TMP_LIT_LIST::CONCAT // only for nonempty lists
 {
  private:
   TMP_LITERAL** next_lit;
   TMP_LITERAL** last_lit;
   Flatterm::Iterator iter;
  public:
   CONCAT() : iter() {};
   CONCAT(TMP_LIT_LIST& lst) 
    : next_lit(lst.First() + 1),
      last_lit(lst.Last()),
      iter((*(lst.First()))->FlatTerm())
     {}; 
   inline void Reset(TMP_LIT_LIST& lst)
    {
     next_lit = lst.First() + 1;
     last_lit = lst.Last();
     iter.Reset((*(lst.First()))->FlatTerm());
    };
   inline bool Empty() const
    {
     return ((next_lit > last_lit) && (iter.EndOfTerm()));
    };
   inline bool Nonempty() const
    {
     return ((next_lit <= last_lit) || (iter.NotEndOfTerm()));
    };  
   inline void NextSym() 
    {
     iter.NextSym();
     if (iter.EndOfTerm())
      if (next_lit <= last_lit)
       {
        iter.Reset((*next_lit)->FlatTerm());
        next_lit++;
       }; 
    };
   inline TERM Symbol() const { return iter.CurrPos()->Symbol(); }; 
 }; // class TMP_LIT_LIST::CONCAT


//*************** Definitions for TMP_LIT_LIST: ****************************

inline bool TMP_LIT_LIST::Weed(bool& eqTautology)
 {
  CALL("Weed()");
  for (TMP_LITERAL** lit = Last(); lit >= First(); lit--)
   {
    TERM header = (*lit)->Header();
    if (header.IsEquality())
     {
      if ((*lit)->EqEqualArgs())
       {
	if (header.Positive()) { eqTautology = true; return false; };
        ReplaceByTop(lit); 
        goto next_lit;
       };   
      for (TMP_LITERAL** another_lit = First(); another_lit < lit; another_lit++)
       if ((*another_lit)->IsEquality())
        {
         if ((*another_lit)->EqualEqLit(**lit))
          {
           ReplaceByTop(lit); 
           goto next_lit; 
          };         
         if ((*another_lit)->ComplementaryEqLit(**lit)) 
          { 
           eqTautology = false;
           return false;     
          };    
        }; 

      if ((*lit)->IsUnorderedEq()) (*lit)->EqOrderArgs(); 
                      
     }
    else // not an equality
     {
      for (TMP_LITERAL** another_lit = First(); another_lit < lit; another_lit++)
       {
        if ((*another_lit)->equal(**lit))
         {
          ReplaceByTop(lit); 
          goto next_lit; 
         };         
        if ((*another_lit)->Complementary(**lit)) 
         { 
          eqTautology = false;
          return false;     
         };        
       };
     };
    next_lit: {}; // nothing to do here, just the end of the loop body
   };
  return true;
 }; // inline bool TMP_LIT_LIST::Weed()

inline bool TMP_LIT_LIST::Weed(const Clause* const * commutativity_table,bool& eqTautology)
{
 CALL("Weed(const Clause* const * commutativity_table)");
  for (TMP_LITERAL** lit = Last(); lit >= First(); lit--)
   {
    TERM header = (*lit)->Header();
    if (header.IsEquality())
     {
      if ((*lit)->EqEqualArgs(commutativity_table))
       {
        if (header.Positive()) 
         {
	  eqTautology = true;
          return false; 
         };
        ReplaceByTop(lit); 
        goto next_lit;
       };   
      for (TMP_LITERAL** another_lit = First(); another_lit < lit; another_lit++)
       if ((*another_lit)->IsEquality())
        {
         if ((*another_lit)->EqualEqLit(**lit,commutativity_table))
          {
           ReplaceByTop(lit); 
           goto next_lit; 
          };         
         if ((*another_lit)->ComplementaryEqLit(**lit,commutativity_table)) 
          {
	   eqTautology = false;
           return false; 
          };         
        }; 

      if ((*lit)->IsUnorderedEq()) (*lit)->EqOrderArgs(); 
                      
     }
    else // not an equality
     {
      for (TMP_LITERAL** another_lit = First(); another_lit < lit; another_lit++)
       {
        if ((*another_lit)->equal(**lit,commutativity_table))
          {
           ReplaceByTop(lit); 
           goto next_lit; 
          };         
         if ((*another_lit)->Complementary(**lit,commutativity_table))
          {
	   eqTautology = false;
           return false; 
          };    
       };
     };
    next_lit: {}; // nothing to do here, just the end of the loop body
   };
  return true;
}; // bool TMP_LIT_LIST::Weed(const Clause* const * commutativity_table)



inline void TMP_LIT_LIST::CollectStat(STAT& stat,bool inversePolarityForSelection)
 {
  CALL("CollectStat(STAT& stat,bool inversePolarityForSelection)");
  for (TMP_LITERAL** lit = literal; lit < next_literal; lit++)
   {
    if ((*lit)->isInherentlySelected()) stat.inherently_selected++;
    if ((*lit)->IsEquality()) { stat.equations++; }
    else
     {
      if (stat.noneq_lits)
       {
        if (GlobalHeaderWeighting.Compare((*lit)->Header(),stat.max_noneq_header) == Greater)
         stat.max_noneq_header = (*lit)->Header();
       }
      else stat.max_noneq_header = (*lit)->Header(); 
      stat.noneq_lits++;
     };
    //if ((*lit)->Negative()) stat.negative++;
    if (NegativeForSelection((*lit)->Header(),inversePolarityForSelection)) stat.negative++;
   };
 }; // void TMP_LIT_LIST::CollectStat(STAT& stat,bool inversePolarityForSelection)



inline void TMP_LIT_LIST::CollectStatWithSplitting(STAT& stat,bool inversePolarityForSelection)
 {
  CALL("CollectStatWithSplitting(STAT& stat,bool inversePolarityForSelection)");
  for (TMP_LITERAL** lit = literal; lit < next_literal; lit++)
   {
    if ((*lit)->isInherentlySelected()) stat.inherently_selected++;
    if ((*lit)->IsEquality()) { stat.equations++; }
    else
     {
      if (stat.noneq_lits)
       {
        if (GlobalHeaderWeighting.Compare((*lit)->Header(),stat.max_noneq_header) == Greater)
         stat.max_noneq_header = (*lit)->Header();
       }
      else stat.max_noneq_header = (*lit)->Header(); 
      stat.noneq_lits++;
      if ((*lit)->Header().IsSplitting())
       {
	 //if ((*lit)->Header().Negative())
	if (NegativeForSelection((*lit)->Header(),inversePolarityForSelection))
	 {
          stat.negative_splitting_branches++;
         }
        else stat.positive_splitting_branches++;
       };
     };
    //if ((*lit)->Negative()) stat.negative++;
    if (NegativeForSelection((*lit)->Header(),inversePolarityForSelection)) stat.negative++;
   };
 }; // void TMP_LIT_LIST::CollectStatWithSplitting(STAT& stat,bool inversePolarityForSelection)



inline bool TMP_LIT_LIST::Eliminate(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,const bool* eliminated)
 {
  CALL("Eliminate(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,const bool* eliminated)");  
  for (TMP_LITERAL** lit = Last(); lit >= literal; lit--) 
   if (eliminated[(*lit)->Header().functor()])
    {
     max_lits.Push(*lit);    
     ReplaceByTop(lit);
     while (Nonempty()) 
      {
       TMP_LITERAL* nonmax_lit = Pop(); 
       nonmax_lit->setInherentlySelected(false);
       ASSERT(!(eliminated[nonmax_lit->Header().functor()]));        

       nonmax_lits.Push(nonmax_lit);
       //nonmax_lits.Push(Pop()); 
      }
     return true;
    };        
  return false;
 }; //  bool TMP_LIT_LIST::Eliminate(TMP_LIT_LIST& max_lits,TMP_LIT_LIST& nonmax_lits,const bool* eliminated)


inline void TMP_LIT_LIST::Order(TMP_LIT_LIST& result,const VAR_WEIGHTING& var_weighting)
 {
  CALL("Order(TMP_LIT_LIST& result,const VAR_WEIGHTING& var_weighting)");
  
  // in the result greater literals will be closer to the top
  if (Length() == 1) 
   { 
    result.Push(Pop()); 
    return;
   };
  if (Length() == 2) 
   {
    TMP_LITERAL* fst = Pop();
    TMP_LITERAL* snd = Pop();
    
    if (fst->Compare(*snd,var_weighting) == Less) 
     {
      result.Push(fst); 
      result.Push(snd); 
      return;
     };
    result.Push(snd); 
    result.Push(fst);
    return; 
   };
  TMP_LITERAL* center = Pop();  
  TMP_LIT_LIST smaller;
  TMP_LIT_LIST greater; 
  do
   {
    TMP_LITERAL* lit = Pop();
    if (lit->Compare(*center,var_weighting) == Less) { smaller.Push(lit); } else { greater.Push(lit); };
   }
  while (Nonempty()); 
  if (smaller.Length()) smaller.Order(result,var_weighting);
  result.Push(center);
  if (greater.Length()) greater.Order(result,var_weighting);
 }; // void TMP_LIT_LIST::Order(TMP_LIT_LIST& result,const VAR_WEIGHTING& var_weighting)

inline void TMP_LIT_LIST::CreateVarWeighting(VAR_WEIGHTING& var_weighting)
 {
  var_weighting.Reset();
  for (TMP_LITERAL** lit = literal; lit < next_literal; lit++)
   var_weighting.AnotherTermVariables((*lit)->Variables()); 
 }; // void TMP_LIT_LIST::CreateVarWeighting(VAR_WEIGHTING& var_weighting)

inline void TMP_LIT_LIST::RenameVariables(INT_VAR_RENAMING& var_renaming)
 {
  for (TMP_LITERAL** lit = Last(); lit >= literal; lit--)
   // order is significant!
   (*lit)->RenameVariables(var_renaming);
 }; // void TMP_LIT_LIST::RenameVariables(INT_VAR_RENAMING& var_renaming)

inline void TMP_LIT_LIST::CollectGround(TMP_LIT_LIST& component)
 {
  component.Reset();
  for (TMP_LITERAL** lit = Last(); lit >= literal; lit--) // order is significant!
   if ((*lit)->Ground()) 
    {
     component.Push(*lit);    
     ReplaceByTop(lit); 
    }; 
 }; // void TMP_LIT_LIST::CollectGround(TMP_LIT_LIST& component)

inline void TMP_LIT_LIST::CollectNextComponent(TMP_LIT_LIST& component)
 {
  component.Reset();
  static MultisetOfVariables<HellConst::MaxNumOfVariables> variables;
  if (Length() == 1) { component.Push(Pop()); return; };  
  TMP_LITERAL* fst = Pop();
  component.Push(fst);
  variables = fst->Variables();
  
  ulong prev_num_of_different;
  new_search:
   prev_num_of_different = variables.NumOfDifferent();
   for (TMP_LITERAL** lit = Last(); lit >= literal; lit--) // order is significant
    {
     if ((*lit)->Variables().IntersectionNonempty(variables)) 
      {
       component.Push(*lit);
       variables.Add((*lit)->Variables());
       ReplaceByTop(lit);
      };
    };
   if ((Nonempty()) && (variables.NumOfDifferent() != prev_num_of_different))
    goto new_search;
 }; // void TMP_LIT_LIST::CollectNextComponent(TMP_LIT_LIST& component)
//===========================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
