#ifndef TMP_LITERAL_H
//===========================================================
#define TMP_LITERAL_H
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"
#include "variables.h"
#include "Multiset.h"
#include "weighting.h"
#include "Flatterm.h"
#include "clause_index.h"
extern HEADER_WEIGHTING GlobalHeaderWeighting;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEW_CLAUSE
 #define DEBUG_NAMESPACE "TMP_LITERAL"
#endif 
#include "debugMacros.h"
//=================================================
class TMP_LITERAL
 {
  public:
   typedef MultisetWithMinMax<ulong,HellConst::MaxTermDepth + 1> VarDepthProfile;
  private:
   TERM header;
   Flatterm::OpenWithMemory _openFlatterm;
   ulong _size;
   
   MultisetOfVariables<HellConst::MaxNumOfVariables> variables;
   MultisetOfVariables<HellConst::MaxNumOfVariables> eq_vars1;
   MultisetOfVariables<HellConst::MaxNumOfVariables> eq_vars2;
   bool eq_different_order;
   ulong max_depth;
   bool _inherentlySelected;
  public:
   TMP_LITERAL() 
    : _openFlatterm(), 
      variables(), 
      max_depth(HellConst::MaxTermDepth),
      _inherentlySelected(false)
     {};
   ~TMP_LITERAL() { CALL("destructor ~TMP_LITERAL()"); };
   void Reset()
    {
     _openFlatterm.Reset();
     _size = 0;
     variables.Reset();
     _inherentlySelected = false;
    };
   void SetMaxDepth(ulong md) 
    {
     if (md > HellConst::MaxTermDepth) { max_depth = HellConst::MaxTermDepth; }
     else max_depth = md;
    };
   void setInherentlySelected(bool fl) 
   {
    CALL("setInherentlySelected(bool fl)");
    _inherentlySelected = fl; 
   };
   bool isInherentlySelected() const { return _inherentlySelected; };
   const Flatterm* FlatTerm() const { return _openFlatterm.Term(); }; 
   ulong currentSize() const { return _openFlatterm.currentSize(); };
   const TERM& Header() const { return header; };
   ulong arity() const { return header.arity(); };
   unsigned Depth() { CALL("Depth()"); return ((Flatterm*)FlatTerm())->Depth(); };
   ulong size() const { return _size; };
   ulong NumOfRedexes(); 
   MultisetOfVariables<HellConst::MaxNumOfVariables>& Variables() { return variables; };
   bool Ground() const { return !(variables.Size()); };
   bool Nonground() const { return !(variables.Size()); };
   bool IsEquality() const { return header.IsEquality(); };
   bool IsOrderedEq() const { return header.IsOrderedEq(); };
   bool IsUnorderedEq() const { return header.IsUnorderedEq(); };
   bool Positive() const { return header.Positive(); };
   bool Negative() const { return header.Negative(); }; 
   TERM::Polarity polarity() const { return header.polarity(); };
   bool NotPropositional() const { return (bool)header.arity(); };
   bool Defined();  
     

   const Flatterm* Arg1() const { return FlatTerm() + 1; };
   const Flatterm* Arg2() const { return Arg1()->after(); };
   MultisetOfVariables<HellConst::MaxNumOfVariables>& Arg1Vars() { return eq_vars1; };
   MultisetOfVariables<HellConst::MaxNumOfVariables>& Arg2Vars() { return eq_vars2; };
   void LinkEqArgs() 
    { 
     CALL("LinkEqArgs()");
     ASSERT(FlatTerm() < (FlatTerm()->after()));
     ((Flatterm*)(FlatTerm()->after()))->SetBackJump(FlatTerm() + 1); 
    };
   bool EqSwapArgs() const { return eq_different_order; };
   const Flatterm* MaxArg() const 
   {
    CALL("MaxArg() const");
    ASSERT(IsEquality());
    ASSERT(IsOrderedEq());
    if (eq_different_order) return Arg2();
    return Arg1();
   };

   void EqCollectVars();    
   void EqMakeOrdered()
    {
     CALL("EqMakeOrdered()");
     ASSERT(IsEquality());
     header.MakeOrdered();
     ((Flatterm*)FlatTerm())->symbolRef() = header;
    };
   
   void EqMakeUnordered()
    {
     CALL("EqMakeUnordered()");
     header.MakeUnordered();
     ((Flatterm*)FlatTerm())->symbolRef() = header;
     eq_different_order = false;
    };

   ulong PosEqMinLHSSize() const
    {
     CALL("PosEqMinLHSSize() const");
     ASSERT(IsEquality() && Positive());
     if (IsOrderedEq())
      {
       if (EqSwapArgs()) return Arg1()->size();
       return Arg2()->size();
      }
     else // unordered 
      {
       long w1 = Arg1()->size();
       long w2 = Arg2()->size();
       return (w1 > w2) ? w2 : w1;
      }; 
    };
   
   void PushVar(const ulong& v)
    {
     variables.Register(v);
     _openFlatterm.PushVar(TERM(TermVar,v));
    };
   void PushVar(const TERM& v) 
    {
     variables.Register(v.var());
     _openFlatterm.PushVar(v); 
    };
   void PushConst(const TERM& c) { _openFlatterm.PushConst(c); };
   bool PushNonconstFun(const TERM& f) 
    {
     if (_openFlatterm.Depth() < max_depth)
      { 
       _openFlatterm.PushNonconstFun(f); 
       _size += f.arity();
       return true;
      };
     return false;   
    };
   
   bool PushSym(const TERM& sym)
    {
     if (sym.isVariable()) { PushVar(sym); return true; };
     if (sym.arity()) return PushNonconstFun(sym);
     PushConst(sym);
     return true;   
    };

   bool PushTerm(const Flatterm* term)
   {
    CALL("PushTerm(const Flatterm* term)");
    Flatterm::Iterator iter(term); 
    do
     {
      if (!PushSym(iter.CurrPos()->symbolRef())) return false;
      iter.NextSym();
     }
    while (iter.NotEndOfTerm());
    return true;
   };   

   void PushNonpropHeader(const TERM& hd)
    {
     _size = hd.arity() + 1;
     _openFlatterm.PushNonconstFun(hd);
     header = hd;
    };
   
   void PushPropLit(const TERM& lit)
    {
     _size = 1;
     _openFlatterm.PushConst(lit);
     header = lit;
    };
   
   bool EqEqualArgs()
    {
     CALL("EqEqualArgs()");
     ASSERT(IsEquality());
     return Arg1()->equal(Arg2());
    };
    
   bool EqEqualArgs(const Clause* const * commutativity_table)
    {
     CALL("EqEqualArgs(const Clause* const * commutativity_table)");
     ASSERT(IsEquality());
     return Arg1()->equal(Arg2(),commutativity_table);
    };

   bool equal(TMP_LITERAL& lit) const 
   {
    CALL("equal(TMP_LITERAL& lit) const"); 
    return FlatTerm()->equal(lit.FlatTerm()); 
   };
   bool equal(TMP_LITERAL& lit,const Clause* const * commutativity_table) const 
   { 
    CALL("equal(TMP_LITERAL& lit,const Clause* const * commutativity_table) const");
    return FlatTerm()->equal(lit.FlatTerm(),commutativity_table); 
   };
   bool Complementary(TMP_LITERAL& lit) const 
   {
    CALL("Complementary(TMP_LITERAL& lit) const"); 
    return FlatTerm()->Complementary(lit.FlatTerm()); 
   };
   bool Complementary(TMP_LITERAL& lit,const Clause* const * commutativity_table) const 
   { 
    CALL("Complementary(TMP_LITERAL& lit,const Clause* const * commutativity_table) const");
    return FlatTerm()->Complementary(lit.FlatTerm(),commutativity_table); 
   };

   bool EqualEqLit(TMP_LITERAL& lit) const;
   bool EqualEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const;
   bool ComplementaryEqLit(TMP_LITERAL& lit) const;
   bool ComplementaryEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const;  

   bool SubstPossible(ulong var,unsigned subst_depth,long subst_weight)
    {
     CALL("SubstPossible(ulong var,unsigned subst_depth,long subst_weight)");
     ulong occs = variables.Occurences(var);
     if (occs)
      {     
       ulong new_weight = size() + (subst_weight * occs) - occs;
       return (new_weight <= HellConst::MaxTermSize) && ((ulong)(Depth() + subst_depth) <= HellConst::MaxTermDepth);
      }
     else return true;
    };

   bool Subst(TMP_LITERAL& new_lit,ulong var,Flatterm* subst_term)
    {
     CALL("Subst(TMP_LITERAL& new_lit,ulong var,Flatterm* subst_term)");
     TERM new_header = Header();
     ASSERT(new_header.arity());
     if(new_header.IsEquality()) new_header.MakeUnordered();
     new_lit.Reset();
     new_lit.PushNonpropHeader(new_header);
     Flatterm::Iterator iter(FlatTerm());
     Flatterm::Iterator subst_iter;
     iter.NextSym(); // skipping the header
     while (iter.NotEndOfTerm())
      {
       TERM sym = iter.CurrPos()->Symbol();  
       if (sym.isVariable() && sym.var() == var) 
	{
         subst_iter.Reset(subst_term);             
         do
          {
           if (!new_lit.PushSym(((Flatterm*)(subst_iter.CurrPos()))->symbolRef())) return false;             
           subst_iter.NextSym();
          }
         while (subst_iter.NotEndOfTerm());
        }
       else if (!new_lit.PushSym(sym)) return false;
       iter.NextSym();
      };
     return true;
    };

   void RenameVariables(INT_VAR_RENAMING& var_renaming)
    {
     if ((IsEquality()) && eq_different_order)
      {
       ((Flatterm*)Arg2())->RenameVariables(var_renaming);
       ((Flatterm*)Arg1())->RenameVariables(var_renaming);
      }
     else ((Flatterm*)FlatTerm())->RenameVariables(var_renaming);
    };

   bool PreferredNegative(TMP_LITERAL& lit);
   bool PreferredNegativeEq(TMP_LITERAL& lit);
   bool PreferredNegativeNeq(TMP_LITERAL& lit);
   Comparison KBComparePos(TMP_LITERAL& lit);
   Comparison KBComparePosEq(TMP_LITERAL& lit);
   Comparison KBCompareEq(TMP_LITERAL& lit);

   
   Comparison KBCompareEqOO(TMP_LITERAL& lit);
   Comparison KBCompareEqOU(TMP_LITERAL& lit);
   Comparison KBCompareEqUU(TMP_LITERAL& lit);
   
   Comparison KBCompareEqArgT1S1(TMP_LITERAL& lit);
   Comparison KBCompareEqArgT1S2(TMP_LITERAL& lit);
   Comparison KBCompareEqArgT2S1(TMP_LITERAL& lit);
   Comparison KBCompareEqArgT2S2(TMP_LITERAL& lit);
   
   Comparison KBCompareArgs(TMP_LITERAL& lit);
   Comparison KBCompare(TMP_LITERAL& lit);
   Comparison Compare(TMP_LITERAL& lit,const VAR_WEIGHTING& var_weighting);
   void EqOrderArgs();
   TERM StandardRepCopy() // for input clauses
    {
     if ((IsEquality()) && eq_different_order)
      {
       TERM copy;
       copy.SetFirst(new(2) TERM(header));
       *(copy.First()->Args()) = Arg2()->StandardRepCopy(); 
       *(copy.First()->Args()->Next()) = Arg1()->StandardRepCopy(); 
       return copy;
      }
     else return FlatTerm()->StandardRepCopy(); 
    };
  
   void collectVarDepthProfile(VarDepthProfile& profile)
   {
    CALL("collectVarDepthProfile(VarDepthProfile& profile)");
    profile.reset();    
    if (variables.empty()) return; 
    Flatterm::Iterator iter(FlatTerm());
    SmartDepthCounter<HellConst::MaxTermDepth> depth_counter;
    depth_counter.Reset(1);   
    while (iter.NotEndOfTerm())
     {
      depth_counter.NextSym(iter.CurrPos()->symbolRef());  
      if (iter.CurrPos()->isVariable())
       { 
	profile.add(depth_counter.Depth());      
       };
   
      iter.NextSym();
     };

    //DF; output(cout) << "\n";
    //DF; profile.output(cout) << "\n";

 
   }; // void collectVarDepthProfile(VarDepthProfile& profile)
 

  public: 
   ostream& output(ostream& str) const
   {
    if (_inherentlySelected) str << "[[InhSel]] ";
    if (IsEquality())
     { 
      str << '(';
      if (IsOrderedEq() && EqSwapArgs())
       {
	Arg2()->outputTerm(str) << ')' << header << '(';
        Arg1()->outputTerm(str);
       }      
      else
       {
        Arg1()->outputTerm(str) << ')' << header << '(';
        Arg2()->outputTerm(str);
       };
      str << ')';
     }
    else FlatTerm()->outputTerm(str);
    return str; 
   };
 }; // class TMP_LITERAL 
 

inline ostream& operator<<(ostream& str,const TMP_LITERAL& lit) 
{
 return lit.output(str);
};

inline ostream& operator<<(ostream& str,const TMP_LITERAL* lit) 
{
 return lit->output(str);
};


//******************* Definitions for TMP_LITERAL: ************************** 

inline ulong TMP_LITERAL::NumOfRedexes() 
 {
  Flatterm* term_to_scan;
  if (IsEquality() && IsOrderedEq())
   {
    if (EqSwapArgs()) { term_to_scan = (Flatterm*)(Arg2()); }  
    else term_to_scan = (Flatterm*)(Arg1()); 
   }
  else term_to_scan = (Flatterm*)(FlatTerm());
  long res = 0;
  if (!term_to_scan->symbolRef().isVariable()) 
   {
    Flatterm::Iterator iter(term_to_scan);
    do
     {
      TERM sym = ((Flatterm*)(iter.CurrPos()))->symbolRef();     
      if (sym.isComplex() && sym.arity()) res++;
      iter.NextSym();
     }
    while (iter.NotEndOfTerm()); 
   };
  return res;
 }; // long TMP_LITERAL::NumOfRedexes()


inline bool TMP_LITERAL::Defined() 
 {
  

  bool res =  ((!IsEquality()) 
                && (header.arity() == variables.NumOfDifferent()));     
  return res;           
 }; // bool TMP_LITERAL::Defined()


inline void TMP_LITERAL::EqCollectVars()
 {
  eq_vars1.Reset();
  eq_vars2.Reset();
  if (Ground()) return;
  const Flatterm* end;
  if (Arg1()->size() < Arg2()->size())
   {
    end = Arg1()->after();
    for (const Flatterm* pos = Arg1(); pos < end; pos++)
     {
      TERM sym = pos->Symbol();
      if (sym.isVariable()) eq_vars1.Register(sym.var());
     };
    eq_vars2.CollectDifference(variables,eq_vars1); 
   }
  else
   {
    end = Arg2()->after();
    for (const Flatterm* pos = Arg2(); pos < end; pos++)
     {
      TERM sym = pos->Symbol();
      if (sym.isVariable()) eq_vars2.Register(sym.var());
     };
    eq_vars1.CollectDifference(variables,eq_vars2); 
   };
 }; // void TMP_LITERAL::EqCollectVars()

inline void TMP_LITERAL::EqOrderArgs()
 {
  CALL("EqOrderArgs()");
  // 1) don't apply this to already ordered equations!
  // 2) this must be an equational literal
  EqCollectVars();
  if (IsOrderedEq()) { eq_different_order = false; return; }; // already ordered  
  
  switch (Flatterm::KBCompare(Arg1(),&eq_vars1,Arg2(),&eq_vars2))
   {
    case Less:
     eq_different_order = true;
     EqMakeOrdered();
     return;
  
    case Equal: eq_different_order = false; return;
        
    case Greater:
     eq_different_order = false; 
     EqMakeOrdered();
     return;

    case Incomparable: eq_different_order = false; return;
   };
  
  #ifdef DEBUG_NAMESPACE
   ICP("ICP0");  
   return;          
  #else
   #ifdef _SUPPRESS_WARNINGS_
    return;
   #endif
  #endif
           
 }; // void TMP_LITERAL::EqOrderArgs()

inline bool TMP_LITERAL::EqualEqLit(TMP_LITERAL& lit) const 
 {
  CALL("EqualEqLit(TMP_LITERAL& lit) const");
  // both this and lit must be equational literals
  if (polarity() != lit.polarity()) return false;
  if (Arg1()->equal(lit.Arg1()))
   {
    
    return Arg2()->equal(lit.Arg2());
   }
  else
   {
    if ((IsUnorderedEq()) || (lit.IsUnorderedEq()))
     {
      return (Arg1()->equal(lit.Arg2())) && (Arg2()->equal(lit.Arg1()));
     }
    else return false; 
   }; 
 }; // bool TMP_LITERAL::EqualEqLit(TMP_LITERAL& lit) const 
    
inline bool TMP_LITERAL::EqualEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const 
 {
  CALL("EqualEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const");
  // both this and lit must be equational literals
  if (polarity() != lit.polarity()) return false;
  if (Arg1()->equal(lit.Arg1(),commutativity_table))
   {
    
    return Arg2()->equal(lit.Arg2(),commutativity_table);
   }
  else
   {
    if ((IsUnorderedEq()) || (lit.IsUnorderedEq()))
     {
      return (Arg1()->equal(lit.Arg2(),commutativity_table)) 
              && (Arg2()->equal(lit.Arg1(),commutativity_table));
     }
    else return false; 
   }; 
 }; // bool TMP_LITERAL::EqualEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const 


inline bool TMP_LITERAL::ComplementaryEqLit(TMP_LITERAL& lit) const 
 {
  CALL("ComplementaryEqLit(TMP_LITERAL& lit) const");
  // both this and lit must be equational literals
  if (polarity() == lit.polarity()) return false;
  if (Arg1()->equal(lit.Arg1()))
   {
    return Arg2()->equal(lit.Arg2());
   }
  else
   {
    if ((IsOrderedEq()) && (lit.IsOrderedEq()))
     {
      return false;
     }
    else return (Arg1()->equal(lit.Arg2())) && (Arg2()->equal(lit.Arg1()));
   };
 }; // bool TMP_LITERAL::ComplementaryEqLit(TMP_LITERAL& lit) const 
 

inline bool TMP_LITERAL::ComplementaryEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const 
 {
  CALL("ComplementaryEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const");
  // both this and lit must be equational literals
  if (polarity() == lit.polarity()) return false;
  if (Arg1()->equal(lit.Arg1(),commutativity_table))
   {
    return Arg2()->equal(lit.Arg2(),commutativity_table);
   }
  else
   {
    if ((IsOrderedEq()) && (lit.IsOrderedEq()))
     {
      return false;
     }
    else return (Arg1()->equal(lit.Arg2(),commutativity_table)) && (Arg2()->equal(lit.Arg1(),commutativity_table));
   };
 }; // bool TMP_LITERAL::ComplementaryEqLit(TMP_LITERAL& lit,const Clause* const * commutativity_table) const 



inline Comparison TMP_LITERAL::KBCompareArgs(TMP_LITERAL& lit)
 {
  CALL("KBCompareArgs(TMP_LITERAL& lit)");
  ASSERT(!IsEquality());
  ASSERT(Header() == lit.Header());

  return Flatterm::KBCompare(FlatTerm(),&variables,lit.FlatTerm(),&lit.variables);    
 }; // Comparison TMP_LITERAL::KBCompareArgs(TMP_LITERAL& lit) 
 
inline Comparison TMP_LITERAL::KBCompare(TMP_LITERAL& lit)
{
 CALL("KBCompare(TMP_LITERAL& lit)");
 if (IsEquality())
  {
   if ((lit.IsEquality())
       && (polarity() == lit.polarity()))
    return KBCompareEq(lit); 
   return (Positive()) ? Less : Greater;
  }
 else
  if (lit.IsEquality())
   {
    return (lit.Positive()) ? Greater : Less;
   };

 // both are not equations
 
 return Flatterm::KBCompareAtoms(FlatTerm(),&variables,lit.FlatTerm(),&lit.variables);
}; // Comparison TMP_LITERAL::KBCompare(TMP_LITERAL& lit)

inline bool TMP_LITERAL::PreferredNegativeEq(TMP_LITERAL& lit)
{
 CALL("PreferredNegativeEq(TMP_LITERAL& lit)");
 ASSERT(Negative());
 ASSERT(IsEquality());
 ASSERT(lit.Negative());
 ASSERT(lit.IsEquality());
 ulong size1;
 ulong size2;
 if (IsOrderedEq()) { size1 = MaxArg()->size(); }
 else size1 = size() - 1; 
 if (lit.IsOrderedEq()) { size2 = lit.MaxArg()->size(); }
 else size2 = lit.size() - 1;   
 if (size1 > size2) return false;
 if (size1 < size2) return true;
 // size1 == size2
 return true; // can be refined  
}; // bool TMP_LITERAL::PreferredNegativeEq(TMP_LITERAL& lit)

inline bool TMP_LITERAL::PreferredNegativeNeq(TMP_LITERAL& lit)
{
 CALL("PreferredNegativeNeq(TMP_LITERAL& lit)");
 ASSERT(!IsEquality());
 ASSERT(!lit.IsEquality());
 ulong size1;
 ulong size2;
 Comparison lexCmp;
 switch (GlobalHeaderWeighting.Compare(Header(),lit.Header()))
  {
   case Greater: return true;
   case Less: return false;
   case Incomparable: ICP("ICP0"); return false;
   case Equal: 
    size1 = size();
    size2 = lit.size();
    if (size1 > size2) return false;
    if (size1 < size2) return true;
    lexCmp = FlatTerm()->CompareLex(lit.FlatTerm());
    if (lexCmp == Less) return true;
    if (lexCmp == Greater) return false;
    return true; // can be refined
  };
 #ifdef DEBUG_NAMESPACE
  ICP("ICP1");
  return false;
 #else
  #ifdef _SUPPRESS_WARNINGS_
   return false;
  #endif
 #endif
}; // bool TMP_LITERAL::PreferredNegativeNeq(TMP_LITERAL& lit) 


inline bool TMP_LITERAL::PreferredNegative(TMP_LITERAL& lit)
{
 CALL("PreferredNegative(TMP_LITERAL& lit)"); 
 if (IsEquality())
  { 
   ASSERT(Negative());
   if (lit.IsEquality())
    {
     ASSERT(lit.Negative());
     // both are negative equalities
     return PreferredNegativeEq(lit);     
    }
   else return true;
  }
 else
  if (lit.IsEquality()) { return false; }
  else // both are not equalities
   {
    return PreferredNegativeNeq(lit);
   };
}; // bool TMP_LITERAL::PreferredNegative(TMP_LITERAL& lit)


inline Comparison TMP_LITERAL::KBComparePosEq(TMP_LITERAL& lit)
 {
  CALL("KBComparePosEq(TMP_LITERAL& lit)");
  ASSERT(Positive());
  ASSERT(IsEquality()); 
  ASSERT(lit.IsEquality());

  if (IsOrderedEq())
   {
    if (lit.IsOrderedEq()) return KBCompareEqOO(lit); 
    return KBCompareEqOU(lit);
   }; 
  // this is unordered  

  if (lit.IsOrderedEq()) return ::inverse(lit.KBCompareEqOU(*this)); 
  // both are unordered
  return KBCompareEqUU(lit);  
 }; // Comparison TMP_LITERAL::KBComparePosEq(TMP_LITERAL& lit)

inline Comparison TMP_LITERAL::KBCompareEq(TMP_LITERAL& lit)
 {
  CALL("KBCompareEq(TMP_LITERAL& lit)");
  ASSERT(IsEquality()); 
  ASSERT(lit.IsEquality());
  ASSERT(Header().polarity() == lit.Header().polarity());

  if (IsOrderedEq())
   {
    if (lit.IsOrderedEq()) return KBCompareEqOO(lit); 
    return KBCompareEqOU(lit);
   }; 
  // this is unordered  

  if (lit.IsOrderedEq()) return ::inverse(lit.KBCompareEqOU(*this)); 
  // both are unordered
  return KBCompareEqUU(lit);  
 }; // Comparison TMP_LITERAL::KBCompareEq(TMP_LITERAL& lit)

inline Comparison TMP_LITERAL::KBComparePos(TMP_LITERAL& lit)
 {
  CALL("KBComparePos(TMP_LITERAL& lit)");
  ASSERT(Positive());
  ASSERT(lit.Positive());
  
  if (IsEquality())
   {
    if (lit.IsEquality())  
     {
      // both are equalities
      return KBComparePosEq(lit);
     }
    else return (lit.arity()) ? Less : Greater; // only propositional literals are less than equalities 
   }
  else // this is not an equality
   {
    if (lit.IsEquality()) return (arity()) ? Greater : Less; // only propositional literals are less than equalities
    // both are not equalities
    Comparison cmp = GlobalHeaderWeighting.Compare(Header(),lit.Header());
    ASSERT(cmp != Incomparable);
    if (cmp == Equal) cmp = KBCompareArgs(lit);
    return cmp;
   };  
 }; // Comparison TMP_LITERAL::KBComparePos(TMP_LITERAL& lit) 


inline Comparison TMP_LITERAL::KBCompareEqOO(TMP_LITERAL& lit)
 {
  CALL("KBCompareEqOO(TMP_LITERAL& lit)");
  ASSERT(IsEquality());
  ASSERT(IsOrderedEq());
  ASSERT(Header() == lit.Header());
  
  const Flatterm* max;
  const Flatterm* lit_max;
  MultisetOfVariables<HellConst::MaxNumOfVariables>* max_vars;
  MultisetOfVariables<HellConst::MaxNumOfVariables>* lit_max_vars;
  
  if (EqSwapArgs()) { max = Arg2(); max_vars = &eq_vars2; }
  else { max = Arg1(); max_vars = &eq_vars1; };
  
  if (lit.EqSwapArgs()) { lit_max = lit.Arg2(); lit_max_vars = &(lit.eq_vars2); }
  else { lit_max = lit.Arg1(); lit_max_vars = &(lit.eq_vars1); };  
  
  Comparison cmp = Flatterm::KBCompare(max,max_vars,lit_max,lit_max_vars);
  if (cmp != Equal) return cmp;
  return Incomparable;
 }; // Comparison TMP_LITERAL::KBCompareEqOO(TMP_LITERAL& lit)

inline Comparison TMP_LITERAL::KBCompareEqOU(TMP_LITERAL& lit)
 {
  CALL("KBCompareEqOU(TMP_LITERAL& lit)");
  ASSERT(IsEquality());
  ASSERT(lit.IsEquality());
  ASSERT(IsOrderedEq());
  ASSERT(lit.IsUnorderedEq());
  ASSERT(Positive() == lit.Positive());
  const Flatterm* max;
  const Flatterm* lit_max;    
  MultisetOfVariables<HellConst::MaxNumOfVariables>* max_vars;
  MultisetOfVariables<HellConst::MaxNumOfVariables>* lit_max_vars; 
  
  if (EqSwapArgs()) { max = Arg2(); max_vars = &eq_vars2; }
  else { max = Arg1(); max_vars = &eq_vars1; };
  
  lit_max = lit.Arg1(); 
  lit_max_vars = &(lit.eq_vars1);  
  
  Comparison cmp = Flatterm::KBCompare(max,max_vars,lit_max,lit_max_vars);

  switch (cmp)
   {
    case Greater: 
     // t1 > s1
     lit_max = lit.Arg2(); 
     lit_max_vars = &(lit.eq_vars2);
     // compare t1 and s2
     cmp = Flatterm::KBCompare(max,max_vars,lit_max,lit_max_vars);
     if (cmp == Greater) return Greater;     
     return Incomparable;

    case Less:
     // t1 < s1
     return Less;

    case Incomparable:         
     return Incomparable; 

    case Equal: return Incomparable;
   };
  #ifdef DEBUG_NAMESPACE 
    ERROR("DE100");
    return Incomparable; 
  #else
   #ifdef _SUPPRESS_WARNINGS_
    return Incomparable; 
   #endif  
  #endif 
 }; // Comparison TMP_LITERAL::KBCompareEqOU(TMP_LITERAL& lit)

inline Comparison TMP_LITERAL::KBCompareEqUU(TMP_LITERAL& lit)  
 {
  CALL("KBCompareEqUU(TMP_LITERAL& lit)");
  // compare t1 = t2 with s1 = s2, below "*" means "incomparable"
  ASSERT(IsEquality());
  ASSERT(lit.IsEquality());
  ASSERT(IsUnorderedEq());
  ASSERT(lit.IsUnorderedEq());
  ASSERT(Positive() == lit.Positive());

  Comparison cmp = KBCompareEqArgT1S1(lit);
  switch (cmp)
   {
    case Greater:
     // t1 > s1
     cmp = KBCompareEqArgT1S2(lit); 
     switch (cmp)
      {
       case Greater:
        // t1 > s1, t1 > s2
        return Greater;
        
       case Less:
        // t1 > s1, t1 < s2  : error because s1 * s2
        #ifdef DEBUG_NAMESPACE 
         ERROR("DE10");
        #endif
        return Incomparable;
       
       case Incomparable:
        // t1 > s1, t1 * s2
        cmp = KBCompareEqArgT2S2(lit); 
        switch (cmp)
         {
          case Greater:
           // t1 > s1, t1 * s2, t2 > s2
           return Greater;
            
          case Less:
           // t1 > s1, t1 * s2, t2 < s2
           return Incomparable;
           
          case Incomparable:
           // t1 > s1, t1 * s2, t2 * s2
           return Incomparable;
           
          case Equal:
           // t1 > s1, t1 * s2, t2 = s2
	   return Greater;
         };
        
       case Equal:
        // t1 > s1, t1 = s2 : error because s1 * s2
        #ifdef DEBUG_NAMESPACE 
         ERROR("DE20"); 
        #endif
        return Incomparable;
      };

    case Less:
     // t1 < s1
     cmp = KBCompareEqArgT2S1(lit); 
     switch (cmp)
      {
       case Greater:
        // t1 < s1, t2 > s1 : error because t1 * t2
        #ifdef DEBUG_NAMESPACE 
         ERROR("DE30");
        #endif
        return Incomparable;
        
       case Less:
        // t1 < s1, t2 < s1
	return Less; // was Greater in the CASC-16 version (obvious incompleteness)
        
       case Incomparable:
        // t1 < s1, t2 * s1
        cmp = KBCompareEqArgT2S2(lit); 
        switch (cmp)
         {
          case Greater:
           // t1 < s1, t2 * s1, t2 > s2
           return Incomparable;
           
          case Less:
           // t1 < s1, t2 * s1, t2 < s2
           return Less;
           
          case Incomparable:
           // t1 < s1, t2 * s1, t2 * s2
           return Incomparable;
           
          case Equal:
           // t1 < s1, t2 * s1, t2 = s2
           return Less; 
         }; 
        
       case Equal:
        // t1 < s1, t2 = s1 : error because t1 * t2
        #ifdef DEBUG_NAMESPACE 
         ERROR("DE40");
        #endif
        return Incomparable;
      };
      
    case Incomparable:
     // t1 * s1
     cmp = KBCompareEqArgT1S2(lit); 
     switch (cmp)
      {
       case Greater:
        // t1 * s1, t1 > s2
        cmp = KBCompareEqArgT2S1(lit); 
        switch (cmp)
         {
          case Greater:
           // t1 * s1, t1 > s2, t2 > s1
           return Greater; 
           
          case Less:
           // t1 * s1, t1 > s2, t2 < s1
           return Incomparable;
           
          case Incomparable:
           // t1 * s1, t1 > s2, t2 * s1
           return Incomparable;
           
          case Equal:
           // t1 * s1, t1 > s2, t2 = s1
           return Greater;
         };
          
       case Less:
        // t1 * s1, t1 < s2
        cmp = KBCompareEqArgT2S1(lit); 
        switch (cmp)
         {
          case Greater:
           // t1 * s1, t1 < s2, t2 > s1
           return Incomparable;
          
          case Less:
           // t1 * s1, t1 < s2, t2 < s1
           return Less;
            
          case Incomparable:
           // t1 * s1, t1 < s2, t2 * s1
           cmp = KBCompareEqArgT2S2(lit); 
           switch (cmp)
            {
             case Greater:
              // t1 * s1, t1 < s2, t2 * s1, t2 > s2 : error because t1 * t2
              
              ICP("DE50");
                  
              return Incomparable; 
           
             case Less:
              // t1 * s1, t1 < s2, t2 * s1, t2 < s2
              return Less;
              
             case Incomparable:
              // t1 * s1, t1 < s2, t2 * s1, t2 * s2
              return Incomparable;
              
             case Equal:
              // t1 * s1, t1 < s2, t2 * s1, t2 = s2 : error because t1 * t2
	      ICP("DE60");
              return Incomparable; 
            };
            
          case Equal:
           // t1 * s1, t1 < s2, t2 = s1
           return Less; 
         }; 
         

       case Incomparable:
        // t1 * s1, t1 * s2
        cmp = KBCompareEqArgT2S1(lit); 
        switch (cmp)
         {
          case Greater:
           // t1 * s1, t1 * s2, t2 > s1
           cmp = KBCompareEqArgT2S2(lit); 
           switch (cmp)
            {
             case Greater:
              // t1 * s1, t1 * s2, t2 > s1, t2 > s2
              return Greater;
              
             case Less:
              // t1 * s1, t1 * s2, t2 > s1, t2 < s2 : error because s1 * s2
               
              ICP("DE70"); 
              
              return Incomparable; 
              
             case Incomparable:
              // t1 * s1, t1 * s2, t2 > s1, t2 * s2
              return Incomparable;
              
             case Equal:
              // t1 * s1, t1 * s2, t2 > s1, t2 = s2 : error because s1 * s2
               
              ICP("DE80");
              
              return Incomparable; 
            };
           
          case Less:
           // t1 * s1, t1 * s2, t2 < s1
           return Incomparable;
           
          case Incomparable:
           // t1 * s1, t1 * s2, t2 * s1
           return Incomparable;
           
          case Equal:
           // t1 * s1, t1 * s2, t2 = s1
           return Incomparable;
         }; 
        
       case Equal:
        // t1 * s1, t1 = s2
        cmp = KBCompareEqArgT2S1(lit); 
        return cmp;
      };
      
    case Equal:
     // t1 = s1
     cmp = KBCompareEqArgT2S2(lit); 
     return cmp;
   };
  #ifdef DEBUG_NAMESPACE 
    ICP("DE100");
    return Incomparable; 
  #else
   #ifdef _SUPPRESS_WARNINGS_
    return Incomparable; 
   #endif  
  #endif
   
 }; // Comparison TMP_LITERAL::KBCompareEqUU(TMP_LITERAL& lit)  


inline Comparison TMP_LITERAL::KBCompareEqArgT1S1(TMP_LITERAL& lit)
 {
  return Flatterm::KBCompare(Arg1(),&eq_vars1,lit.Arg1(),&lit.eq_vars1);
 }; // Comparison TMP_LITERAL::KBCompareEqT1S1(TMP_LITERAL& lit)
 
inline Comparison TMP_LITERAL::KBCompareEqArgT1S2(TMP_LITERAL& lit)
 {
  return Flatterm::KBCompare(Arg1(),&eq_vars1,lit.Arg2(),&lit.eq_vars2);
 }; // Comparison TMP_LITERAL::KBCompareEqT1S2(TMP_LITERAL& lit)

inline Comparison TMP_LITERAL::KBCompareEqArgT2S1(TMP_LITERAL& lit)
 {
  return Flatterm::KBCompare(Arg2(),&eq_vars2,lit.Arg1(),&lit.eq_vars1);
 }; // Comparison TMP_LITERAL::KBCompareEqT2S1(TMP_LITERAL& lit)   

inline Comparison TMP_LITERAL::KBCompareEqArgT2S2(TMP_LITERAL& lit)
 {
  return Flatterm::KBCompare(Arg2(),&eq_vars2,lit.Arg2(),&lit.eq_vars2);
 }; // Comparison TMP_LITERAL::KBCompareEqT2S2(TMP_LITERAL& lit)    
   
   
 
inline Comparison TMP_LITERAL::Compare(TMP_LITERAL& lit,const VAR_WEIGHTING& var_weighting)
 {
  CALL("Compare(TMP_LITERAL& lit,const VAR_WEIGHTING& var_weighting)");
  Comparison hd_cmp = GlobalHeaderWeighting.Compare(Header(),lit.Header());
  if (hd_cmp != Equal) { return hd_cmp; };
  
  // here the headers are equal
  if (size() > lit.size()) { return Greater; };
  if (size() < lit.size()) { return Less; };
  
  // here the weights are equal
  if (variables.Size() < lit.variables.Size()) { return Greater; };
  if (variables.Size() > lit.variables.Size()) { return Less; };
  
  // here both terms have same number of variable occurences
  if (variables.NumOfDifferent() < lit.variables.NumOfDifferent()) { return Greater; };
  if (variables.NumOfDifferent() > lit.variables.NumOfDifferent()) { return Less; };
  
  // compare lexicographically
  Comparison var_cmp;
  const Flatterm* fst = FlatTerm() + 1;
  const Flatterm* const end_of_fst = FlatTerm()->after();
  const Flatterm* snd = lit.FlatTerm();
  check_end:
   if (fst == end_of_fst) return Incomparable;
   snd++;
   if (fst->Symbol() != snd->Symbol())
    {
     if (fst->Symbol().isVariable()) 
      {
       if (snd->Symbol().isVariable()) 
        {
         // different variables
         var_cmp = var_weighting.Compare(fst->Symbol().var(),snd->Symbol().var());
         if (var_cmp != Equal) { return var_cmp; };
        }
       else return Less; // a variable is less than a functor
      }
     else 
      {
       if (snd->Symbol().isVariable()) return Greater;
       // both are functors
       return fst->Symbol().CompareFunctor(snd->Symbol()); 
      }; 
    };
   fst++;
   goto check_end;
 }; // Comparison TMP_LITERAL::Compare(TMP_LITERAL& lit,const VAR_WEIGHTING& var_weighting)

//===========================================================================
#undef DEBUG_NAMESPACE
//===========================================================
#endif
