//
// File:         Master.cpp
// Description:  Main loop.
// Created:      Mar 5, 2000, 20:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Master.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
 #define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.h"
//====================================================

void Master::endOfInput()
{
 CALL("endOfInput()");
 // destroy obsolete clauses
 ASSERT(_retainedNewClauses.empty());
 ASSERT(_queueForBackSimplification.empty());
 
 disposeObsolete(); 

 ASSERT(_queueForDisposal.empty());

 if (_parameters.unit_conflict)
  {
   // now scan the sets _passive and _active for unit facts and check them all
   // for unit confict
   Clause* cl;
   _passive.ResetIterator();
   while ((cl = _passive.IterNext()))
    if (cl->Unit()) 
     { 
      if (checkForUnitConflict(cl))
       {
	ASSERT(proof);
        terminate(Proof);  
        return;
       };         
     }; 

   _active.ResetIterator();
   while ((cl = _active.Next()))
    if (cl->Unit()) 
     { 
      if (checkForUnitConflict(cl))
       {
	ASSERT(proof);
        terminate(Proof);  
        return;
       };                    
     };   

  };
}; // void Master::endOfInput()

void Master::mainLoop0()
{
 CALL("mainLoop0()");

 timer.reset();
 timer.start();
 _startMemory = (float)GlobAlloc::occupiedByObjects();   
 _nextTick = _parameters.internal_tick_frequency;
 terminated = false;  
 proof = (Clause*)0;

 //endOfInput();
 //if (terminated) return;
 
 Clause* current;
 select_current: 

  PROGRESSCHECK("<select current> in Master::mainLoop0()",10.0);
  current = select0(); 
  if (!current) { terminate(EmptyPassive); return; };
  if (_active.Total() >= _parameters.maxActive) { terminate(MaxActive); return; }; 

  if ((_parameters.orphan_murder)
      && (!current->IsVIP())
      && (current->IsOrphan()))
   {      
    _statPassiveOrphansMurdered++;
    disposeClause(current);  
    estimateResources();
    goto select_current;
   }; 
  
  makeActive0(current); 

  eliminateAllocationBufferDeficit0();  

  _inferenceDispatcherSemaphore = true;
  _infDispatcher.newQuery(current);
  while (_infDispatcher.nextInference())
   {
    PROGRESSCHECK("<next inference> in Master::mainLoop0()",3.0);
    processCollected0();
    if (proof) { terminate(Proof); _inferenceDispatcherSemaphore = false; return; };
  
    if (_passive.Total() >= _parameters.maxPassive) { terminate(MaxPassive);  _inferenceDispatcherSemaphore = false; return; };
    if (terminated) { _inferenceDispatcherSemaphore = false; return; };

    anotherInference0();
 
    backSimplify();
    if (terminated) { _inferenceDispatcherSemaphore = false; return; };
   };
  // no more inferences with current
   _inferenceDispatcherSemaphore = false;
  
   //eliminateAllocationBufferDeficit0();

  anotherInference0(); // to force internal tick 
    

  if (_parameters.defined_header_table 
      && current->DefinitionUnfoldingSelected(_parameters.defined_header_table))
   {
    _statRewrittenByDefUnfolding++; 
    enqueueForDisposal(current);
   };

  // destroy obsolete clauses
  ASSERT(_retainedNewClauses.empty());
  ASSERT(_queueForBackSimplification.empty());
  //ASSERT(_queueForUnitConflict.empty());
  disposeObsolete(); 
  ASSERT(_queueForDisposal.empty());
  estimateResources();
  goto select_current; 
}; // void Master::mainLoop0()

void Master::backSimplify()
{
 CALL("backSimplify()");
 ASSERT((_parameters.backward_demodulation > 0) || _queueForBackSimplification.empty());
 _backwardDemodulationSemaphore = true;
 Clause* currentSimplifier;
 Clause* backSimplified;
 bool usedSimplified = false;
 if (_parameters.backward_demodulation > 0)
  {
   while (_queueForBackSimplification.dequeue(currentSimplifier))
    { 
     ASSERT(currentSimplifier->flag(Clause::FlagInQueueForBackSimplification));
     currentSimplifier->clearFlag(Clause::FlagInQueueForBackSimplification);
     if (backDemodulationQuery(currentSimplifier))
      { 
       while (nextBackDemodulated())
        {         
         _statBackSimplified++;
         
	 if (collectBackSimplified())
	  {   
	    //currentSimplifier->MarkAsSimplifier();
	   currentSimplifier->setFlag(Clause::FlagHasChildren);
	   backSimplified = _backDemodulation.simplified();
           usedSimplified = usedSimplified 
                            || (backSimplified->mainSet() == Clause::MainSetActive) 
                            || (backSimplified->mainSet() == Clause::MainSetDefinitions); 

           backSimplified->setFlag(Clause::FlagHasChildren);

           enqueueForDisposal(backSimplified);        
           processCollected0(); // not before enqueueForDisposal(backSimplified)!
           if (proof) { terminate(Proof); _backwardDemodulationSemaphore = false; return; };
           if (_passive.Total() >= _parameters.maxPassive) 
            { 
             terminate(MaxPassive); 
             _backwardDemodulationSemaphore = false; 
             return; 
            }; 
          }
         else
          if (_freshClause.WasTautology())
           {
            backSimplified = _backDemodulation.simplified();

            enqueueForDisposal(backSimplified);
           };
         anotherInference0();
        };  
      };
    };
   if (_parameters.orphan_murder && usedSimplified) killActiveOrphans0();
  };
 _backwardDemodulationSemaphore = false;
}; // void Master::backSimplify()


void Master::eliminateDefinitions()
{
 CALL("Master::eliminateDefinitions()");
 ASSERT(!_parameters.main_alg);
 Clause* cl;
 select:
  _passive.ResetIterator();
  try_next:
   cl = _passive.IterNext();
   if (cl) 
    {
     if (cl->ContainsEliminated(_parameters.pred_elim_table))
      {
       _passive.Remove(cl);
      
       // perform resolutions
       makeActive0(cl);
       _inferenceDispatcherSemaphore = true;
       _infDispatcher.newQuery(cl);     
       while (_infDispatcher.nextInference())
        {
         _freshClause.UsedRule(ClauseBackground::Prepro);
         processCollectedInput0();
         if (proof) 
          { 
           terminate(Proof); 
           _inferenceDispatcherSemaphore = false; 
           goto collect_results; 
          };
         backSimplify();
         if (terminated) { _inferenceDispatcherSemaphore = false; goto collect_results; };
        };
       _inferenceDispatcherSemaphore = false;
       // destroy obsolete clauses
       ASSERT(_retainedNewClauses.empty());
       ASSERT(_queueForBackSimplification.empty());
       //ASSERT(_queueForUnitConflict.empty());
       disposeObsolete(); 

       // use next
       goto select;     
      }
     else goto try_next;      
   };

   // all done, collect results
   collect_results: 
    ASSERT(_retainedNewClauses.empty());
    ASSERT(proof || _queueForBackSimplification.empty());
    _queueForBackSimplification.reset();


    // all _active are obsolete now, make them _history
    Clause* act;
    while (_active.Nonempty())
     {
      act = _active.Top();  
      enqueueForDisposal(act);
     };
    disposeObsolete();   
}; // void Master::eliminateDefinitions()

void Master::inputClause0(Clause* cl)
{
 CALL("inputClause0(Clause* cl)");
 terminated = false;  
 proof = (Clause*)0; 
 if (_freshClause.LoadClause(cl))
  {
   _freshClause.PushAncestor(cl); 
   _freshClause.UsedRule(ClauseBackground::Prepro); 
   _passive.MaxPriorityOn(); 
   processCollectedInput0(); 
   _passive.MaxPriorityOff(); 
   if (proof) 
    {  
     terminate(Proof); 
     ASSERT(_retainedNewClauses.empty());    
     disposeObsolete();  
     return; 
    }; 
   backSimplify(); 
  };
}; // void Master::inputClause0(Clause* cl)

void Master::processCollected0()
{
 CALL("processCollected0()");

 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';         
 _freshClause.ProcessCollected(_retainedNewClauses);


 Clause* newClause;  
 while (_retainedNewClauses.dequeue(newClause))
  { 

   ASSERT(newClause->mainSet() == Clause::MainSetNew);
   makePassive0(newClause);
   if (newClause->Empty()) 
    {
     proof = newClause; 
     return; 
    }; 

   if (newClause->weight() > _parameters.maxWeight)
    {        

     enqueueForDisposal(newClause);
     _statRecycled++;

    }
   else
    {
     if (_parameters.backward_subsumption) 
      if (backSubsumeBy(newClause))
       if (_parameters.orphan_murder) 
        killActiveOrphans0();

     if (_parameters.backward_demodulation > 0) 
      {
       ASSERT(!newClause->flag(Clause::FlagInQueueForBackSimplification));

       _queueForBackSimplification.enqueue(newClause);

       newClause->setFlag(Clause::FlagInQueueForBackSimplification);
      };

     integratePassive(newClause);

     eliminateAllocationBufferDeficit0();
    };
  };
}; // void Master::processCollected0()

void Master::processCollectedInput0()
{
 CALL("processCollectedInput0()");

ASSERT_IN(_backSubsumption._termTraversal.checkIntegrity(),"COP1");

 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
 _freshClause.ProcessCollected(_retainedNewClauses);
 Clause* newClause;  
 while (_retainedNewClauses.dequeue(newClause))
  { 
   ASSERT(newClause->mainSet() == Clause::MainSetNew);
   makePassive0(newClause); 
   if (newClause->Empty()) 
    {
     proof = newClause;      
     //return;  
    }
   else    
    if (_parameters.hyper_resolution && _infDispatcher.hyperRule(newClause))
     {
      _passive.Remove(newClause); 

      _unitConflict.addRule(newClause);

      _infDispatcher.integrateHyperRule(newClause);
      if (_parameters.showHyperRule) _parameters.showHyperRule(newClause);  
     }
    else
     {
      TERM definedHeader;
      if (_parameters.defined_header_table 
          && newClause->Definition(_parameters.defined_header_table,definedHeader))
       {
	ASSERT(newClause->mainSet() == Clause::MainSetPassive);

        _passive.Remove(newClause);
        //newClause->SetStore(CS_DEFINITION);
        newClause->setMainSet(Clause::MainSetDefinitions);
        _definitions.Insert(newClause);
        ALWAYS(newClause->MakeDefinitionOf(definedHeader));
        _infDispatcher.integrateDefinition(newClause,definedHeader);
        if (_parameters.showDefinition) _parameters.showDefinition(newClause);                     
       };

      if (_parameters.backward_subsumption) backSubsumeBy(newClause);
      if (_parameters.backward_demodulation > 0) 
       {
	ASSERT(!newClause->flag(Clause::FlagInQueueForBackSimplification));
        _queueForBackSimplification.enqueue(newClause);
        newClause->setFlag(Clause::FlagInQueueForBackSimplification);
       }; 
      integratePassive(newClause);
     };
  };
 ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollectedInput0()

Clause* Master::select0()
{
 CALL("select0()");
 Clause* selected;
 try_next_passive:
  selected = _passive.Next();
  if (!selected)
   {// _passive is empty, try _reservedPassive 
    _parameters.lrs = 0;
    _parameters.maxWeight = HellConst::MaxClauseWeight;  
    _freshClause.SetMaxWeight(_parameters.maxWeight);  
    if ((_parameters.showWeightChange) 
       && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
     _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight());    
    loadReservedClauses();    
    selected = _passive.Next();
    if (selected) 
     {
      ASSERT(selected->mainSet() == Clause::MainSetPassive);
      selected->setMainSet(Clause::MainSetFree);
     };  
    return selected;
   };
  if (selected->MinResWeight() > _parameters.maxWeight)
   { 
    makeReserved(selected); 
    goto try_next_passive;         
   };
  if (selected)     
   {
    ASSERT(selected->mainSet() == Clause::MainSetPassive);
    selected->setMainSet(Clause::MainSetFree);
   }; 
  return selected; 
}; // Clause* Master::select0()  

void Master::makeActive0(Clause* cl)
{
 CALL("makeActive0(Clause* cl)");
  //cl->SetStore(CS_USED);  
 cl->setMainSet(Clause::MainSetActive);
 cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed 
 _active.Insert(cl);
 _infDispatcher.makeActive(cl);
 if (_parameters.showActive) _parameters.showActive(cl); 
}; // void Master::makeActive0(Clause* cl)

void Master::makePassive0(Clause* cl)
{
 CALL("makePassive0(Clause* cl)");
 cl->SetNumber(_nextClauseNum());
 INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE(cl);
 cl->setMainSet(Clause::MainSetPassive);
 _passive.Insert(cl);
 if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(cl);
 if (_parameters.showPassive) _parameters.showPassive(cl);
}; // void Master::makePassive0(Clause* cl)

void Master::integratePassive(Clause* cl)
{
 CALL("integratePassive(Clause* cl)");
 // see also removeFromIndices(Clause* cl)

ASSERT_IN(_backSubsumption._termTraversal.checkIntegrity(),"COP1");

 if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);

 ASSERT_IN(_backSubsumption._termTraversal.checkIntegrity(),"COP2");

 if (_parameters.backward_subsumption) integrateIntoBackwardSubsumptionIndex(cl);
 if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
 if (_parameters.backward_demodulation > 0) integrateIntoBackDemodulationIndex(cl);
}; // void Master::integratePassive(Clause* cl) 

void Master::removeFromIndices(Clause* cl)
{
 CALL("removeFromIndices(Clause* cl)");
 // see also integratePassive(Clause* cl)
 if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl);
 if (cl->flag(Clause::FlagInForwardDemodulationIndex)) removeFromForwardDemodulationIndex(cl);
 if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) removeFromBackwardSubsumptionIndex(cl);
 if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) removeFromForwardSubsumptionIndex(cl);
}; // void Master::removeFromIndices(Clause* cl)

void Master::makeReserved(Clause* cl)
{
 CALL("makeReserved(Clause* cl)");
 cl->setMainSet(Clause::MainSetReservedPassive);
 _reservedPassive.Insert(cl);
}; // void Master::makeReserved(Clause* cl)

void Master::enqueueForDisposal(Clause* cl)
{
 CALL("enqueueForDisposal(Clause* cl)"); 
 if (cl->mainSet() == Clause::MainSetTrash) return;
 switch (cl->mainSet()) 
  {
   case Clause::MainSetFree: break;
   case Clause::MainSetPassive: _passive.Remove(cl); break;
   case Clause::MainSetReservedPassive: _reservedPassive.Remove(cl); break;    
   case Clause::MainSetActive: _active.Remove(cl); break;
   case Clause::MainSetDefinitions: _definitions.Remove(cl); break;
   #ifdef DEBUG_NAMESPACE
    default: ICP("ICP1"); return;
   #else
    default: return;
   #endif
  };

 bool canBeDestroyed = true;
 if (!_inferenceDispatcherSemaphore)
  {
   if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
   if (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)) _infDispatcher.removeDefinition(cl);
  }
 else
  {
   if ((cl->flag(Clause::FlagInInferenceDispatcherAsActive)) 
       || (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)))
    canBeDestroyed = false;
  };

 if (!_backwardDemodulationSemaphore)
  {
   if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl); 
  }
 else
  if (cl->flag(Clause::FlagInBackDemodulationIndex)) canBeDestroyed = false;

 if (cl->flag(Clause::FlagInForwardDemodulationIndex)) removeFromForwardDemodulationIndex(cl);

 if (!_backwardSubsumptionSemaphore)
  {
   if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) removeFromBackwardSubsumptionIndex(cl);
  }
 else if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) canBeDestroyed = false;

 if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) removeFromForwardSubsumptionIndex(cl);
 
 if (canBeDestroyed)
  {
   if (cl->flag(Clause::FlagHasChildren))
    {

     // must be in history
     if (cl->flag(Clause::FlagInClauseIndex)) 
      {
       _clauseIndex.Remove(cl,false);  
       cl->clearFlag(Clause::FlagInClauseIndex);
      };
     cl->setMainSet(Clause::MainSetHistory); 
     _history.Insert(cl); 
    }
   else // can be completely destroyed
    {
      

     bool shared = cl->flag(Clause::FlagLiteralsShared);
     if (cl->flag(Clause::FlagInClauseIndex)) 
      {
       _clauseIndex.Remove(cl,shared);
       cl->clearFlag(Clause::FlagInClauseIndex);
       cl->clearFlag(Clause::FlagLiteralsShared);
      };
     if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
     if (shared) { cl->Destroy(); } else cl->DestroyUnshared();
    };

  }
 else // must wait in the queue
  {
   cl->setMainSet(Clause::MainSetTrash);
   _queueForDisposal.enqueue(cl); 
  };
}; // void Master::enqueueForDisposal(Clause* cl)

void Master::disposeObsolete()
{
 CALL("disposeObsolete()"); 
 Clause* disposedClause;
 while (_queueForDisposal.dequeue(disposedClause))
  { 
   ASSERT(disposedClause->mainSet() == Clause::MainSetTrash);
   disposedClause->setMainSet(Clause::MainSetFree);
   disposeClause(disposedClause);
  };
}; // void Master::disposeObsolete()

void Master::disposeClause(Clause* cl)
{
 CALL("disposeClause(Clause* cl)");
 ASSERT(cl->mainSet() == Clause::MainSetFree);
 ASSERT((!_inferenceDispatcherSemaphore) || ((!cl->flag(Clause::FlagInInferenceDispatcherAsActive)) && (!cl->flag(Clause::FlagInInferenceDispatcherAsDefinition))));
 if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
 if (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)) _infDispatcher.removeDefinition(cl);
 removeFromIndices(cl);
 ASSERT((_parameters.main_alg == 2) || (_parameters.main_alg == 3) || (cl->flag(Clause::FlagInClauseIndex)));
 ASSERT((_parameters.main_alg == 2) || (_parameters.main_alg == 3) || (cl->flag(Clause::FlagLiteralsShared)));


 if (cl->flag(Clause::FlagHasChildren))
  {
   // must be in history
   if (cl->flag(Clause::FlagInClauseIndex)) 
    {
     _clauseIndex.Remove(cl,false);  
     cl->clearFlag(Clause::FlagInClauseIndex);
    };
   cl->setMainSet(Clause::MainSetHistory); 
   _history.Insert(cl); 

  }
 else // can be completely destroyed
  {
   bool shared = cl->flag(Clause::FlagLiteralsShared);
   if (cl->flag(Clause::FlagInClauseIndex)) 
    {
     _clauseIndex.Remove(cl,shared);
     cl->clearFlag(Clause::FlagInClauseIndex);
     cl->clearFlag(Clause::FlagLiteralsShared);
    };
   if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
   if (shared) { cl->Destroy(); } else cl->DestroyUnshared();

  };

}; // void Master::disposeClause(Clause* cl)


void Master::loadReservedClauses()
{
 CALL("loadReservedClauses()");
 Clause* cl;
 _reservedPassive.ResetIterator();
 while (!_reservedPassive.Empty())
  {
   cl = _reservedPassive.Pop();
   ASSERT(cl->mainSet() == Clause::MainSetReservedPassive); 
   cl->setMainSet(Clause::MainSetPassive);
   _passive.Insert(cl); 
  };
}; // void Master::loadReservedClauses()

bool Master::collectBackSimplified()
{
 CALL("collectBackSimplified()");
 return _backDemodulation.collectInference();  
}; // bool Master::collectBackSimplified()


bool Master::backSubsumeBy(Clause* cl)
{
 CALL("backSubsumeBy(Clause* cl)");
 _backwardSubsumptionSemaphore = true;
 Clause* subsumedClause;
 bool usedSubsumed = false;
 if (_backSubsumption.query(cl))
  {
   while (_backSubsumption.subsumeNext())
    { 
     _statBackSubsumed++;
     cl->MakeVIP();
     subsumedClause = (Clause*)_backSubsumption.subsumed();
     ASSERT(subsumedClause != cl);
     ASSERT(subsumedClause->Active()); 
     usedSubsumed = (usedSubsumed 
                     || (subsumedClause->mainSet() == Clause::MainSetActive) 
                     || (subsumedClause->mainSet() == Clause::MainSetDefinitions));     
     enqueueForDisposal(subsumedClause);     
    };
  };
 _backwardSubsumptionSemaphore = false;
 return (bool)usedSubsumed;
}; // bool Master::backSubsumeBy(Clause* cl)

void Master::estimateResources1()
{
 CALL("estimateResources1()");
 ASSERT(_parameters.lrs);
 float used_memory = usedMemory();
 float used_time = timer.elapsed();

 bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
 bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

 if (noTimeCheck && noMemoryCheck) return;

 long moved_from_passive = _passive.UsedAndRemoved();   
 long reachableByTime = (noTimeCheck)? 
                         LONG_MAX 
                          : 
                         (long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time);

 long reachableByMemory = (noMemoryCheck)? 
                           LONG_MAX
                            :
                           (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
 _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
 _passive.SetNumOfReachable(_parameters.reachable);
 _parameters.maxWeight = _passive.MaxAcceptedWeight();
 if (_parameters.maxWeight > _parameters.maxMaxWeight) 
  {
   _parameters.maxWeight = _parameters.maxMaxWeight;
  };
 _freshClause.SetMaxWeight(_parameters.maxWeight);
 if ((_parameters.showWeightChange) 
     && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
  _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 

 if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources1()

void Master::estimateResources2()
{
 CALL("estimateResources2()");
 ASSERT(_parameters.lrs == 2);
 float used_memory = usedMemory();
 float used_time = timer.elapsed();

 bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
 bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

 if (noTimeCheck && noMemoryCheck) return;

 long moved_from_passive = _passive.UsedAndRemoved();   
 long reachableByTime = (noTimeCheck)? 
                         (_halfLongMax) 
                          : 
                         ((long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time));

 long reachableByMemory = (noMemoryCheck)? 
                           (_halfLongMax)
                            :
                           ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
 _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
 long currentlyUnreachable = _passive.Size() - _parameters.reachable;

 if (currentlyUnreachable <= 0)
  { 
   if ((currentlyUnreachable <= _parameters.unreachable) && (_parameters.maxWeight < ((long)HellConst::MaxClauseWeight)))
    { 
     // weight limit increases 
     _parameters.maxWeight++;
     if (_parameters.maxWeight > _parameters.maxMaxWeight) 
      {
       _parameters.maxWeight = _parameters.maxMaxWeight;
      };
    };
  }
 else // currentlyUnreachable > 0
  { 
   if (currentlyUnreachable >= _parameters.unreachable)
    {
     // weight limit decreases 
     if (_parameters.maxWeight > _passive.MaxPresentWeight()) 
      {
       _parameters.maxWeight = _passive.MaxPresentWeight(); 
      }       
     else 
      if (_parameters.maxWeight > _passive.MinPresentWeight()) 
       { 
        _parameters.maxWeight--;
       };
    };
  };

 _parameters.unreachable = currentlyUnreachable;

 _freshClause.SetMaxWeight(_parameters.maxWeight);
 if ((_parameters.showWeightChange) 
     && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
  _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 
 if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources2()



void Master::estimateResources3()
{
 CALL("estimateResources3()");
 ASSERT(_parameters.lrs == 3);
 float used_memory = usedMemory();
 float used_time = timer.elapsed();

 bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
 bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

 if (noTimeCheck && noMemoryCheck) return;

 long moved_from_passive = _passive.UsedAndRemoved();   
 long reachableByTime = (noTimeCheck)? 
                         (_halfLongMax) 
                          : 
                         ((long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time));

 long reachableByMemory = (noMemoryCheck)? 
                           (_halfLongMax)
                            :
                           ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
 _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
 long currentlyUnreachable = _passive.Size() - _parameters.reachable;

 if (currentlyUnreachable <= 0)
  { 
   if ((currentlyUnreachable <= _parameters.unreachable) && (_parameters.maxWeight < ((long)HellConst::MaxClauseWeight)))
    { 
     // weight limit increases 
     _parameters.maxWeight++;
     if (_parameters.maxWeight > _parameters.maxMaxWeight) 
      {
       _parameters.maxWeight = _parameters.maxMaxWeight;
      };
    };
  }
 else // currentlyUnreachable > 0
  { 
   if (currentlyUnreachable >= _parameters.unreachable)
    {
     // weight limit decreases 
     if (_parameters.maxWeight > _passive.MaxPresentWeight()) 
      {
       _parameters.maxWeight = _passive.MaxPresentWeight(); 
      }       
     else 
      if (_parameters.maxWeight >= _passive.MinPresentWeight())
       { 
        _parameters.maxWeight--;
       };
    };
  };

 _parameters.unreachable = currentlyUnreachable;

 _freshClause.SetMaxWeight(_parameters.maxWeight);
 if ((_parameters.showWeightChange) 
     && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
  _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 
 if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources3()


bool Master::completeStrategyUsed()
{
 return !(_parameters.incompleteAPriori
          || statTooHeavy() 
          || statTooHeavyLiterals() 
          || statTooManyVariables()
          || statTooLong()
          || statTooDeep()
          || statRecycled()
          || (_parameters.selection >= 1000)
          || (_parameters.selection <= -1000));      
}; // bool Master::completeStrategyUsed()

void Master::killActiveOrphans0()
{
 CALL("killActiveOrphans0()");
 _active.ResetIterator();
 for (Clause* cl = _active.Next(); cl; cl = _active.Next())
  if ((cl->Active())
      && (!cl->IsVIP())
      && (cl->IsOrphan()))
   {
    enqueueForDisposal(cl);
    _statActiveOrphansMurdered++;
   };
}; // void Master::killActiveOrphans0()



bool Master::checkForUnitConflict(Clause* cl)
{
 CALL("checkForUnitConflict(Clause* cl)");
 if (_freshClause.LoadClause(cl))
  {
   if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
   _freshClause.PushAncestor(cl); 
   _freshClause.UsedRule(ClauseBackground::Prepro);
   return (bool)(proof = _freshClause.checkForUnitConflict());
  };
 return false;
}; // bool Master::checkForUnitConflict(Clause* cl)

//====================================================
