//
// File:         Master1.cpp
// Description:  Main loop implementing main_alg 1
// Created:      May 31, 2000, 20:45
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Master.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
 #define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.h"
//====================================================


void Master::mainLoop1()  
{
 CALL("mainLoop1()");
 
 timer.reset();
 timer.start();
 _startMemory = (float)GlobAlloc::occupiedByObjects();   
 _nextTick = _parameters.internal_tick_frequency;
 terminated = false;  
 proof = (Clause*)0;

 endOfInput();
 if (terminated) return;
 
 Clause* current;
 Clause* reanimated;
 select_current: 
  PROGRESSCHECK("<select current> in Master::mainLoop0()",10.0);
  current = select1(); 
  if (!current) { terminate(EmptyPassive); return; };
  if (_active.Total() >= _parameters.maxActive) { terminate(MaxActive); return; };

  reanimated = reanimate1(current); 
  disposeClause(current);
  //eliminateAllocationBufferDeficit1();

  if (!reanimated) goto select_current; 
  if (reanimated->Empty()) { proof = reanimated; terminate(Proof); return; };
  if (_parameters.backward_subsumption) backSubsumeBy(reanimated);   
  disposeObsolete();
  backSimplifyBy1(reanimated);
  if (proof) { terminate(Proof); return; }; 
  disposeObsolete();  
  makeActive1(reanimated);

  //eliminateAllocationBufferDeficit1();

  _inferenceDispatcherSemaphore = true; 
  _infDispatcher.newQuery(reanimated);
  while (_infDispatcher.nextInference())
   {
    PROGRESSCHECK("<next inference> in Master::mainLoop0()",3.0);
    processCollected1();
    if (proof) 
     { 
      terminate(Proof);
      _inferenceDispatcherSemaphore = false; 
      return; 
     };
  
    if (_passive.Total() >= _parameters.maxPassive) 
     { 
      terminate(MaxPassive);
      _inferenceDispatcherSemaphore = false;
      return; 
     };
    if (terminated) return;

    anotherInference1();
    if (terminated) 
     {
      _inferenceDispatcherSemaphore = false;
      return;
     };
    
   };
  _inferenceDispatcherSemaphore = false;

  anotherInference1(); // to force internal tick 

  // destroy obsolete clauses
  ASSERT(_retainedNewClauses.empty());
  ASSERT(_queueForBackSimplification.empty());
  //ASSERT(_queueForUnitConflict.empty());
  ASSERT(_queueForDisposal.empty());
  estimateResources();
  goto select_current; 
};// void Master::mainLoop1()

Clause* Master::reanimate1(Clause* cl)
{
 CALL("reanimate1(Clause* cl)");
 ASSERT(_parameters.main_alg);

 ulong max_weight = _freshClause.MaxWeight();
 _freshClause.SetMaxWeight(HellConst::MaxClauseWeight); 
 
 if (_freshClause.LoadClause(cl))
  {
   ClauseBackground* ancestors = cl->Ancestors();
   for (ClauseBackground::ClauseList* anc = ancestors->ancestors(); anc; anc = anc->tl())
    _freshClause.PushAncestor(anc->hd()); 

   _freshClause.InverseAncestorOrder();
      
   _freshClause.UsedRules(ancestors->TheRules());
   _freshClause.UsedRule(ClauseBackground::Reanimation);
   if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
   Clause* res = _freshClause.Reanimate();
   if (res)
    {
     res->SetNumber(_nextClauseNum());
     INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE(res);
     //res->SetStore(CS_KEPT);
     res->setMainSet(Clause::MainSetFree); 
     if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(res);
     if (_parameters.showPassive) _parameters.showPassive(res);
    }; 
   _freshClause.SetMaxWeight(max_weight);
   return res;  
  } 
 else 
  {
   _freshClause.SetMaxWeight(max_weight);
   return (Clause*)0;
  };
}; // Clause* Master::reanimate1(Clause* cl) 


void Master::backSimplifyBy1(Clause* simplifier)
{
 CALL("backSimplifyBy1(Clause* simplifier)");
 _backwardDemodulationSemaphore = true;
 Clause* backSimplified; 
 if (backDemodulationQuery(simplifier))
  { 
   while (nextBackDemodulated())
    {         
     _statBackSimplified++;
         
     if (collectBackSimplified())
      { 
       backSimplified = _backDemodulation.simplified();
       ASSERT(backSimplified->mainSet() == Clause::MainSetActive);
       enqueueForDisposal(backSimplified);        

       processCollected1(); // not before enqueueForDisposal1(backSimplified)!

       if (proof) { terminate(Proof); _backwardDemodulationSemaphore = false; return; };
       if (_passive.Total() >= _parameters.maxPassive) { terminate(MaxPassive); _backwardDemodulationSemaphore = false; return; }; 

      }
     else
      if (_freshClause.WasTautology())
       {
        backSimplified = _backDemodulation.simplified();
        enqueueForDisposal(backSimplified);
       };
      anotherInference1();
    };  
  };
  _backwardDemodulationSemaphore = false;
}; // void Master::backSimplifyBy1(Clause* simplifier)


void Master::inputClause1(Clause* cl)
{
 CALL("inputClause1(Clause* cl)");

 terminated = false;  
 proof = (Clause*)0; 

 if (_freshClause.LoadClause(cl))
  {
   _freshClause.PushAncestor(cl);    
   _freshClause.UsedRule(ClauseBackground::Prepro);
   _passive.MaxPriorityOn();
   processCollectedInput1();
   _passive.MaxPriorityOff();
   if (proof) 
    { 
     terminate(Proof);
     ASSERT(_retainedNewClauses.empty());    
     disposeObsolete();  
     return; 
    };
  };

}; // void Master::inputClause1(Clause* cl)

void Master::processCollected1()
{
 CALL("processCollected1()");
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';         
 _freshClause.ProcessCollected1(_retainedNewClauses);
 Clause* newClause;  
 while (_retainedNewClauses.dequeue(newClause))
  { 
   ASSERT(newClause->mainSet() == Clause::MainSetNew);
   makePassive1(newClause);

   if (newClause->Empty()) 
    {
     proof = newClause; 
     return; 
    }; 
   //eliminateAllocationBufferDeficit1();
  };
 ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollected1()


void Master::processCollectedInput1()
{
 CALL("processCollectedInput1()");
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
 _freshClause.ProcessCollected1(_retainedNewClauses);
 Clause* newClause;  

 while (_retainedNewClauses.dequeue(newClause))
  { 
   ASSERT(newClause->mainSet() == Clause::MainSetNew);
   makePassive1(newClause); 
   if (newClause->Empty()) 
    {
     proof = newClause;
    }
   else    
    if (_parameters.hyper_resolution && _infDispatcher.hyperRule(newClause))
     {
      _passive.Remove(newClause);
 
      _unitConflict.addRule(newClause);
 
      _infDispatcher.integrateHyperRule(newClause);
      if (_parameters.showHyperRule) _parameters.showHyperRule(newClause);
     }
    else
     {   
       //if (_parameters.unit_conflict) _queueForUnitConflict.enqueue(newClause); 
     };
  };
 ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollectedInput1()


void Master::makePassive1(Clause* cl)
{
 CALL("makePassive1(Clause* cl)");
 cl->SetNumber(_nextClauseNum());
 INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE(cl);
 //cl->SetStore(CS_KEPT);
 cl->setMainSet(Clause::MainSetPassive);
 _passive.Insert(cl);
 if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(cl);
 if (_parameters.showPassive) _parameters.showPassive(cl);
}; // void Master::makePassive1(Clause* cl)

void Master::makeActive1(Clause* cl)
{
  //cl->SetStore(CS_USED);
 cl->setMainSet(Clause::MainSetActive);
 cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed  
 _active.Insert(cl);
 _infDispatcher.makeActive(cl);
 if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
 if (_parameters.backward_subsumption) 
  { 
   integrateIntoBackwardSubsumptionIndex(cl);
  };
 if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
 if (_parameters.backward_demodulation > 0) integrateIntoBackDemodulationIndex(cl);
 if (_parameters.showActive) _parameters.showActive(cl); 
}; // void Master::makeActive1(Clause* cl)

//====================================================




