//
// File:         Master2.cpp
// Description:  Main loop implementing main_alg 2
// Created:      Jun 2, 2000, 21:15
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Master.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
 #define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.h"
//====================================================

void Master::mainLoop2()
{
 CALL("mainLoop2()");
 //GlobAlloc::setHardMemoryLimit(_parameters.memoryLimit);
 timer.reset();
 timer.start();
 _startMemory = (float)GlobAlloc::occupiedByObjects();   
 _nextTick = _parameters.internal_tick_frequency;
 terminated = false;  
 proof = (Clause*)0;

 endOfInput();
 if (terminated) return;
 
 Clause* current;
 Clause* reanimated;
 select_current: 
  PROGRESSCHECK("<select current> in Master::mainLoop0()",10.0);
  current = select2(); 
  if (!current) { terminate(EmptyPassive); return; };
  if (_active.Total() >= _parameters.maxActive) { terminate(MaxActive); return; };

  if ((_parameters.orphan_murder)
      && (!current->IsVIP())
      && (current->IsOrphan()))
   {      
    _statPassiveOrphansMurdered++;
    disposeClause(current);  
    estimateResources();
    goto select_current;
   };

  reanimated = reanimate2(current);   
  disposeClause(current);
  
  //eliminateAllocationBufferDeficit2();

  if (!reanimated) { estimateResources(); goto select_current; };
  if (reanimated->Empty()) { proof = reanimated; terminate(Proof); return; };
  if (_parameters.backward_subsumption) 
   if (backSubsumeBy(reanimated)) 
    if (_parameters.orphan_murder) killActiveOrphans2();      
    
  disposeObsolete();
  backSimplifyBy2(reanimated);
  if (proof) { terminate(Proof); return; };    
  disposeObsolete();  
  makeActive2(reanimated);

  //eliminateAllocationBufferDeficit2();

  _inferenceDispatcherSemaphore = true;  
  _infDispatcher.newQuery(reanimated);
  while (_infDispatcher.nextInference())
   {
    PROGRESSCHECK("<next inference> in Master::mainLoop0()",3.0);
    processCollected2();
    if (proof) 
     { 
      terminate(Proof); 
      _inferenceDispatcherSemaphore = false; 
      return; 
     };
  
    if (_passive.Total() >= _parameters.maxPassive) 
     { 
      terminate(MaxPassive); 
      _inferenceDispatcherSemaphore = false;
      return; 
     };
    if (terminated) { _inferenceDispatcherSemaphore = false; return; };

    anotherInference2();
    if (terminated) { _inferenceDispatcherSemaphore = false; return; };
   };
  _inferenceDispatcherSemaphore = false;
 

  anotherInference2(); // to force an extra internal tick 

  // destroy obsolete clauses
  ASSERT(_retainedNewClauses.empty());
  ASSERT(_queueForBackSimplification.empty());
  //ASSERT(_queueForUnitConflict.empty());
  ASSERT(_queueForDisposal.empty());
  estimateResources();
  goto select_current; 
};// void Master::mainLoop2()

void Master::backSimplifyBy2(Clause* simplifier)
{
 CALL("backSimplifyBy2(Clause* simplifier)");
 _backwardDemodulationSemaphore = true; 
 Clause* backSimplified; 
 bool somethingSimplified = false;
 if (backDemodulationQuery(simplifier))
  { 
   while (nextBackDemodulated())
    {         
     _statBackSimplified++;
         
     if (collectBackSimplified())
      { 
       backSimplified = _backDemodulation.simplified();
       ASSERT(backSimplified->mainSet() == Clause::MainSetActive);       
       enqueueForDisposal(backSimplified);        
       somethingSimplified = true;
       processCollected2(); // not before enqueueForDisposal2(backSimplified)!

       if (proof) 
        { 
         terminate(Proof); 
         _backwardDemodulationSemaphore = false; 
         return; 
        };
       if (_passive.Total() >= _parameters.maxPassive) 
        { 
         terminate(MaxPassive); 
         _backwardDemodulationSemaphore = false; 
         return; 
        }; 
      }
     else
      if (_freshClause.WasTautology())
       {
        backSimplified = _backDemodulation.simplified();
        enqueueForDisposal(backSimplified);
       };
      anotherInference2();
    };  
  }; 
 if (_parameters.orphan_murder && somethingSimplified) killActiveOrphans2();
 _backwardDemodulationSemaphore = false;
}; // void Master::backSimplifyBy2(Clause* simplifier)

Clause* Master::reanimate2(Clause* cl)
{
 CALL("reanimate2(Clause* cl)");
 ASSERT(_parameters.main_alg);

 ulong max_weight = _freshClause.MaxWeight();
 _freshClause.SetMaxWeight(HellConst::MaxClauseWeight); 
 
 _freshClause.LoadClause(cl);
 ClauseBackground* ancestors = cl->Ancestors();
      
 for (ClauseBackground::ClauseList* anc = ancestors->ancestors(); anc; anc = anc->tl())
  _freshClause.PushAncestor(anc->hd()); 

 _freshClause.InverseAncestorOrder();

 _freshClause.UsedRules(ancestors->TheRules());
 _freshClause.UsedRule(ClauseBackground::Reanimation);
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
 Clause* res = _freshClause.Reanimate();
 if (res)
  {
   res->SetNumber(_nextClauseNum());
   INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE(res);
   //res->SetStore(CS_KEPT);
   res->setMainSet(Clause::MainSetFree); 
   if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(res);
   if (_parameters.showPassive) _parameters.showPassive(res);
  }; 
 _freshClause.SetMaxWeight(max_weight);
 return res;   
}; // Clause* Master::reanimate2(Clause* cl) 


void Master::inputClause2(Clause* cl)
{
 CALL("inputClause2(Clause* cl)");

 terminated = false;  
 proof = (Clause*)0; 

 if (_freshClause.LoadClause(cl))
  {
   _freshClause.PushAncestor(cl);    
   _freshClause.UsedRule(ClauseBackground::Prepro);
   _passive.MaxPriorityOn();
   processCollectedInput2();
   _passive.MaxPriorityOff();
   if (proof) 
    { 
     terminate(Proof);
     ASSERT(_retainedNewClauses.empty());    
     disposeObsolete();  
     return; 
    };
  };

}; // void Master::inputClause2(Clause* cl)

void Master::processCollected2()
{
 CALL("processCollected2()");
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';         
 _freshClause.ProcessCollected2(_retainedNewClauses);
 Clause* newClause;  
 while (_retainedNewClauses.dequeue(newClause))
  { 
   ASSERT(newClause->mainSet() == Clause::MainSetNew);
   makePassive2(newClause);

   if (newClause->Empty()) 
    {
     proof = newClause; 
     return; 
    }; 
   //eliminateAllocationBufferDeficit2();
  };
 ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollected2()


void Master::processCollectedInput2()
{
 CALL("processCollectedInput2()");
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
 _freshClause.ProcessCollected2(_retainedNewClauses);
 Clause* newClause;  

 while (_retainedNewClauses.dequeue(newClause))
  { 
   ASSERT(newClause->mainSet() == Clause::MainSetNew);
   makePassive2(newClause); 
   if (newClause->Empty()) 
    {
     proof = newClause;
    }
   else    
    if (_parameters.hyper_resolution && _infDispatcher.hyperRule(newClause))
     {
      _passive.Remove(newClause);    
      newClause->setMainSet(Clause::MainSetFree);  
      Clause* reanimatedRule = reanimate2(newClause);
      disposeClause(newClause);
      if (reanimatedRule)
       {
        _unitConflict.addRule(reanimatedRule);
        _infDispatcher.integrateHyperRule(reanimatedRule);
        if (_parameters.showHyperRule) _parameters.showHyperRule(reanimatedRule);
       }
      else
       {
	ICP("ICP0");
       };
     }
    else
     {   
       //if (_parameters.unit_conflict) _queueForUnitConflict.enqueue(newClause); 
     };
  };
 ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollectedInput2()


void Master::makePassive2(Clause* cl)
{
 CALL("makePassive2(Clause* cl)");
 cl->SetNumber(_nextClauseNum());
 INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_REGISTER_CLAUSE(cl);
 //cl->SetStore(CS_KEPT);
 cl->setMainSet(Clause::MainSetPassive);
 _passive.Insert(cl);
 //if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(cl);
 if (_parameters.showPassive) _parameters.showPassive(cl);
}; // void Master::makePassive2(Clause* cl)

void Master::makeActive2(Clause* cl)
{
 CALL("makeActive2(Clause* cl)");
  //cl->SetStore(CS_USED);  
 cl->setMainSet(Clause::MainSetActive);
 cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed 
 _active.Insert(cl);
 _infDispatcher.makeActive(cl);
 if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
 if (_parameters.backward_subsumption) 
  { 
   integrateIntoBackwardSubsumptionIndex(cl);
  };
 if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
 if (_parameters.backward_demodulation > 0) integrateIntoBackDemodulationIndex(cl);
 if (_parameters.showActive) _parameters.showActive(cl); 
}; // void Master::makeActive2(Clause* cl)

void Master::destroyPassive2()
{
 CALL("destroyPassive2()");
 _passive.ResetIterator();
 for (Clause* cl = _passive.IterNext(); cl; cl = _passive.IterNext())
  {
   _passive.Remove(cl);
   cl->setMainSet(Clause::MainSetFree);
   cl->DestroyUnshared();
  };
 _passive.Reset();
 ASSERT(_passive.clean());
}; // void Master::destroyPassive2()

bool Master::releaseMemory2(ulong size)
{
 CALL("releaseMemory2(ulong size)");
 Clause* cl = _passive.TheWorstClause(); 
 if (!cl) return false;

 if (((cl->size()*sizeof(TERM)) < size)
     && (sizeof(Clause) < size))
   //&& (cl->Ancestors()->memorySize() < size))
  {      
   return false; 
  };
 ASSERT(cl->mainSet() == Clause::MainSetPassive);
 _passive.Remove(cl);
 cl->setMainSet(Clause::MainSetFree);
 disposeClause(cl);
 _statRecycled++;
 return true;
}; // bool Master::releaseMemory2(ulong size) 


void Master::killActiveOrphans2()
{
 CALL("killActiveOrphans2()");
 _active.ResetIterator();
 for (Clause* cl = _active.Next(); cl; cl = _active.Next())
  if ((cl->Active())
      && (!cl->IsVIP())
      && (cl->IsOrphan()))
   {
    enqueueForDisposal(cl);
    _statActiveOrphansMurdered++;
   };
}; // void Master::killActiveOrphans2()



//====================================================




