//
// File:         Master3.cpp
// Description:  Main loop implementing main_alg 3
// Created:      Apr 29, 2001, 16:50
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Master.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
 #define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.h"
//====================================================

void Master::inputClause3(Clause* cl)
{
 CALL("inputClause3(Clause* cl)");

 terminated = false;  
 proof = (Clause*)0; 

 if (_freshClause.LoadClause(cl))
  {
   _freshClause.PushAncestor(cl);    
   _freshClause.UsedRule(ClauseBackground::Prepro);
   _pagedPassive.maximalPriorityOn();
   processCollectedInput3();
   _pagedPassive.maximalPriorityOff();
   if (proof) 
    { 
     terminate(Proof);
     ASSERT(_retainedNewClauses.empty());    
     disposeObsolete();  
     return; 
    };
  };

}; // void Master::inputClause3(Clause* cl)

void Master::processCollectedInput3()
{
 CALL("processCollectedInput3()");
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';

 _freshClause.ProcessCollected2(_retainedNewClauses);
 Clause* newClause;  

 while (_retainedNewClauses.dequeue(newClause))
  { 
   ASSERT(newClause->mainSet() == Clause::MainSetNew);
   newClause->SetNumber(_nextClauseNum());

   if (_parameters.showPassive) _parameters.showPassive(newClause);    
 
   if (newClause->Empty()) 
    {
     proof = newClause;
    }
   else    
    if (_parameters.hyper_resolution && _infDispatcher.hyperRule(newClause))
     {
      newClause->setMainSet(Clause::MainSetFree);  
      Clause* reanimatedRule = reanimate2(newClause);
      disposeClause(newClause);
      if (reanimatedRule)
       {
        _unitConflict.addRule(reanimatedRule);
        _infDispatcher.integrateHyperRule(reanimatedRule);
        if (_parameters.showHyperRule) _parameters.showHyperRule(reanimatedRule);
       }
      else
       {
	ICP("ICP0");
	RuntimeError::report("An input clause (rule) was not successfully reanimated."); 
       };       
     }
    else // binary resolution is used or the clause is a fact
     {   
      newClause->setMainSet(Clause::MainSetFree);  
      ALWAYS(insertIntoPassive3(newClause)); 
      disposeClause(newClause);      
     };
  };
 ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollectedInput3()


bool Master::insertIntoPassive3(Clause* cl)
{
 CALL("Master::insertIntoPassive3(Clause* cl)");
 ASSERT(_parameters.main_alg == 3);
 ulong max_weight = _freshClause.MaxWeight();
 _freshClause.SetMaxWeight(HellConst::MaxClauseWeight); 
 
 ALWAYS(_freshClause.LoadClause(cl));
 _freshClause.setClauseNumber(cl->Number());
 ClauseBackground* ancestors = cl->Ancestors();
      
 for (ClauseBackground::ClauseList* anc = ancestors->ancestors(); anc; anc = anc->tl())
  _freshClause.PushAncestor(anc->hd()); 

 _freshClause.InverseAncestorOrder();

 _freshClause.UsedRules(ancestors->TheRules());

 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
  
 
 bool res = _freshClause.transferToPagedPassive(_pagedPassive);  
 
 _freshClause.SetMaxWeight(max_weight);

 return res;
}; // bool Master::insertIntoPassive3(Clause* cl)



void Master::mainLoop3()
{
 CALL("mainLoop3()");
 timer.reset();
 timer.start();
 _startMemory = (float)GlobAlloc::occupiedByObjects();   
 _nextTick = _parameters.internal_tick_frequency;
 terminated = false;  
 proof = (Clause*)0;

 endOfInput();
 if (terminated) return;
 
 PagedPassiveClauseStorage::StoredClause selectedClause;
 Clause* reanimated;
 select_current: 
  PROGRESSCHECK("<select current> in Master::mainLoop0()",10.0);
  if (!_pagedPassive.select(selectedClause)) { terminate(EmptyPassive); return; };
  if (_active.Total() >= _parameters.maxActive) { terminate(MaxActive); return; };

  if ((_parameters.orphan_murder)
      && (!selectedClause.isVIP())
      && (selectedClause.isOrphan()))
   {      
    _statPassiveOrphansMurdered++;
    _pagedPassive.popSelected();
    goto select_current;
   };


  // DF; selectedClause.output(cout << "REA ") << "\n"; 

  reanimated = reanimate3(selectedClause);
  
  // _pagedPassive.popSelected() is done in reanimate3(selectedClause)

  if (!reanimated) { estimateResources(); goto select_current; };
  if (reanimated->Empty()) { proof = reanimated; terminate(Proof); return; };
  if (_parameters.backward_subsumption) 
   if (backSubsumeBy(reanimated)) 
    if (_parameters.orphan_murder) killActiveOrphans3();      
    
  disposeObsolete();
  

  backSimplifyBy3(reanimated);
  if (proof) { terminate(Proof); return; };    
  disposeObsolete();  
  makeActive3(reanimated);

  _inferenceDispatcherSemaphore = true;  
  _infDispatcher.newQuery(reanimated);
  while (_infDispatcher.nextInference())
   {
    PROGRESSCHECK("<next inference> in Master::mainLoop0()",3.0);
    processCollected3();
    if (proof) 
     { 
      terminate(Proof); 
      _inferenceDispatcherSemaphore = false; 
      return; 
     };
  
    if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
     { 
      terminate(MaxPassive); 
      _inferenceDispatcherSemaphore = false;
      return; 
     };
    if (terminated) { _inferenceDispatcherSemaphore = false; return; };

    anotherInference3();
    if (terminated) { _inferenceDispatcherSemaphore = false; return; };
   };
  _inferenceDispatcherSemaphore = false;
 

  anotherInference3(); // to force an extra internal tick 

  // destroy obsolete clauses
  ASSERT(_retainedNewClauses.empty());
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForDisposal.empty());
  estimateResources();
  goto select_current; 

};// void Master::mainLoop3()


void Master::backSimplifyBy3(Clause* simplifier)
{
 CALL("backSimplifyBy3(Clause* simplifier)");
 _backwardDemodulationSemaphore = true; 
 Clause* backSimplified; 
 bool somethingSimplified = false;
 if (backDemodulationQuery(simplifier))
  { 
   while (nextBackDemodulated())
    {         
     _statBackSimplified++;
         
     if (collectBackSimplified())
      { 
       backSimplified = _backDemodulation.simplified();
       ASSERT(backSimplified->mainSet() == Clause::MainSetActive);       
       enqueueForDisposal(backSimplified);        
       somethingSimplified = true;
       processCollected3(); // not before enqueueForDisposal3(backSimplified)!

       if (proof) 
        { 
         terminate(Proof); 
         _backwardDemodulationSemaphore = false; 
         return; 
        };
       if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
        { 
         terminate(MaxPassive); 
         _backwardDemodulationSemaphore = false; 
         return; 
        }; 
      }
     else
      if (_freshClause.WasTautology())
       {
        backSimplified = _backDemodulation.simplified();
        enqueueForDisposal(backSimplified);
       };
      anotherInference3();
    };  
  }; 
 if (_parameters.orphan_murder && somethingSimplified) killActiveOrphans3();
 _backwardDemodulationSemaphore = false;
}; // void Master::backSimplifyBy3(Clause* simplifier)

Clause* Master::reanimate3(PagedPassiveClauseStorage::StoredClause& cl)
{
 CALL("reanimate3(PagedPassiveClauseStorage::StoredClause& cl)");

 //DF; cout << cl.output(cout << "REA ") << "\n"; 

 ASSERT(_parameters.main_alg == 3); 

 ulong max_weight = _freshClause.MaxWeight();
 _freshClause.SetMaxWeight(HellConst::MaxClauseSize); 
 
 _freshClause.LoadClause(cl.literals());
 
 //DF; _freshClause.output(cout << "LOADED " << statGenerated() << ' ') << '\n';

 cl.resetAncestors();
 Clause* ancestor;
 while (cl.nextAncestor(ancestor)) _freshClause.PushAncestor(ancestor);
 _freshClause.InverseAncestorOrder();

 _freshClause.UsedRules(cl.rules());
 _freshClause.UsedRule(ClauseBackground::Reanimation);
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';

 _pagedPassive.popSelected();

 Clause* res = _freshClause.Reanimate();
 if (res)
  {
   res->SetNumber(_nextClauseNum());
   res->setMainSet(Clause::MainSetFree); 
   if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(res);
   if (_parameters.showPassive) _parameters.showPassive(res);
  }; 
 _freshClause.SetMaxWeight(max_weight);
 return res;
}; // Clause* Master::reanimate3(PagedPassiveClauseStorage::StoredClause& cl) 

void Master::processCollected3()
{
 CALL("processCollected3()");
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';         
 Clause* emptyClause;
 ulong numOfNewClauses;
 ulong numOfRejectedNewClauses;
 _freshClause.ProcessCollected3(emptyClause,
                                _pagedPassive,
                                numOfNewClauses,
                                numOfRejectedNewClauses,
                                _nextClauseNum,
                                (bool)_parameters.showPassive);
 _statClausesRefusedAllocation += numOfRejectedNewClauses;
 if (emptyClause)  
  {
   proof = emptyClause; 
   return; 
  };
}; // void Master::processCollected3()



void Master::makeActive3(Clause* cl)
{
 CALL("makeActive3(Clause* cl)");
 cl->setMainSet(Clause::MainSetActive);
 cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed 
 _active.Insert(cl);
 _infDispatcher.makeActive(cl);
 if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
 if (_parameters.backward_subsumption) 
  { 
   integrateIntoBackwardSubsumptionIndex(cl);
  };
 if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
 if (_parameters.backward_demodulation > 0) integrateIntoBackDemodulationIndex(cl);
 if (_parameters.showActive) _parameters.showActive(cl);
}; // void Master::makeActive3(Clause* cl)


void Master::destroyPassive3()
{
 CALL("destroyPassive3()");
 destroyPassive2(); // _passive is used for input clauses
 _pagedPassive.destroy();
 _pagedPassive.init();  
}; // void Master::destroyPassive3()


bool Master::releaseMemory3(ulong size)
{
 CALL("releaseMemory3(ulong size)");
 if (size > PagedPassiveClauseStorage::pageSize()) return false;
 try_free_page:
  if (_pagedPassive.recyclePage()) return true;
  if (_pagedPassive.cancelWorstClause()) goto try_free_page; 
 return false;
}; // bool Master::releaseMemory3(ulong size) 


void Master::killActiveOrphans3()
{
 CALL("killActiveOrphans3()");
 _active.ResetIterator();
 for (Clause* cl = _active.Next(); cl; cl = _active.Next())
  if ((cl->Active())
      && (!cl->IsVIP())
      && (cl->IsOrphan()))
   {
    enqueueForDisposal(cl);
    _statActiveOrphansMurdered++;
   };
}; // void Master::killActiveOrphans3()



//====================================================




