//
// File:         Vampire.h
// Description:  Main module of Vampire.
// Created:      Mar 6, 2000, 14:30
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef VAMPIRE_H
//=================================================
#define VAMPIRE_H
#include <iostream.h>
#include <fstream.h> 
#include <signal.h>
#include "jargon.h"
#include "MasterDebugFlags.h"
#include "InferenceTree.h"
#include "ClauseNumIndex.h"
#include "ClauseQueue.h"
#include "clause.h"
#include "weighting.h"
#include "Tabulation.h"
#include "Master.h"
#include "ProblemProfile.h"   
#include "Signature.h" 
#include "AutoFunctorPrecedence.h"
extern HEADER_WEIGHT_LIST* HWList;    
extern HEADER_WEIGHTING GlobalHeaderWeighting;
extern Signature TheSignature; 
#if ((defined INTERCEPT_FORWARD_MATCHING_INDEXING) || (defined INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING))
 #include "Logging.h"
 #include "HellInterceptMacros.h" 
#endif     
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE
 #define DEBUG_NAMESPACE "Vampire"
#endif
#include "debugMacros.h"
//=================================================
class Vampire
{
 public:
  class InitStatic
  {
   public:
    InitStatic() 
    {
     CALL("InitStatic::constructor InitStatic()");
     if (!_count) { Vampire::initStatic(); _count++; };
    };
    ~InitStatic() {};
   private: 
    long _count;
  };
  class Options
  { 
   public:
    Options() { reset(); };
    void reset();
    ostream& output(ostream& str) const;
   public:

    bool incompleteAPriori; // = never say "Unprovable"

    // automatic mode
    bool autoStrat;
  
    // main algorithm
    
    long main_alg;

    // inference rules
    bool hyper_resolution;
    bool no_inferences_between_rules;
    bool static_splitting;
    bool dynamic_splitting;
    long splitting_style;
    bool splitting_with_naming;
    long static_neg_eq_splitting;
    bool paramodulation;
    long sim_back_sup;
    bool hyper_positive;
    long selection;
    long literal_comparison_mode;
    bool inherited_negative_selection;
     
    // strategy 
    bool forward_subsumption;
    bool fs_set_mode;
    bool fs_optimised_multiset_mode;
    bool fs_use_sig_filters;
    bool fs_old_style;
    long simplify_by_fs;
    long forward_demodulation;
    bool fd_on_splitting_branches;
    bool normalize_by_commutativity;
    long backward_demodulation;
    bool bd_on_splitting_branches;
    bool simplify_by_eq_res;
    bool backward_subsumption;
    bool bs_set_mode;
    bool bs_optimised_multiset_mode;
    long orphan_murder;
    bool unit_conflict;
    long lrs;
    ofstream* plot_usage_pace;
    ofstream* plot_weight_limit_changes;
    long elim_def;
    long header_rank_by_arity;
    bool header_rank_kinky;
    bool header_rank_check_positive_selection;

    // numeric options 
    float timeLimit;
    float memoryLimit;
    double allocation_buffer_price; 
    long maxActive;
    long maxPassive;
    long maxWeight;
    long optGrCompWeight;  
    long age_weight_ratio;
    float selected_pos_eq_penalty_coefficient;    
    long internal_tick_frequency;
    long first_time_check;   
    long first_mem_check;   
    
    // interface options    
    bool showOpt;
    bool showInput;
    bool showPrepro;
    bool showGen;
    bool showPassive;
    bool showActive; 
    bool showHyperRules;
    bool showDefinitions;
    bool showProof;
    bool showProfile;
    bool showWeightLimitChanges;  
    Tabulation* tabulation;
  };
 public:
  Vampire() {};
  ~Vampire() {};
  static void initSIGINTHandler() 
  { 
   #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
    signal(SIGINT,handleSIGINT); 
   #else
    //signal(SIGBREAK,handleSIGINT); 
	//signal(SIGINT,handleSIGINT);
   #endif
  };
  static void makeOptionsConsistent(); 
  static void terminate(const char* msg);
  static void abort();
  static bool inputClause(Clause* cl);
  static void definedHeader(const TERM& hd);
  static void prove();
  static void preproOnly();
  static ulong nextClauseNum();
  static bool releaseMemory(ulong size) { return _master.releaseMemory(size); }; 
  static void reportAllocationFailure(ulong size)
  {
   _master.reportAllocationFailure(size);
   reportTermination();
  };

 public: 
  static Options options;
 private:
  static void initStatic();
  static float availableTime() { return options.timeLimit - _timer.elapsed(); }; 
  static void preprocess();
  static void preproSimplify(ulong iterNum);
  static void preproEliminateDef(ulong iterNum);
  static void prepareMasterForPreproSimplification();
  static void prepareMasterForPreproDefElimination();
  static void initPredefinedStrategies();
  static void cancelEquality();
  static void manualStrat(Master::Options& opt);
  static void autoNNE(Master::Options& opt);
  static void autoHNE1(Master::Options& opt);              
  static void autoHNE2(Master::Options& opt);
  static void autoNEQ(Master::Options& opt);
  static void autoHEQ(Master::Options& opt);
  static void autoPEQ(Master::Options& opt);
  static void autoUEQ(Master::Options& opt);
  static void extractPreprocessed(); 
  static void preprocessedToHistory();
  static void search(float timeLimit,const Master::Options& strat);
  static void runAuto();
  static void load(const ClauseQueue& clauses);
  static void outputClause(const Clause* cl);
  static void outputHyperRule(const Clause* cl);
  static void outputDefinition(const Clause* cl);
  static void output(const ClauseQueue& clauses);
  static void outputInTPTPFormat(const ClauseQueue& clauses); 
  static void showWeightChange(long prevMW,long newMW); 
  static void showUsagePace(float usagePace);
  static void outputStatistics();
  static void tabulateStatistics();
  static void tabulatePreprocessingStatistics();
  static void tabulateOptions();
  static void handleSIGINT(int);
  #if (COMPILER == COMPILER_MS_VISUAL_C_PLUS_PLUS)
   static void raiseSIGTERM(int);
  #endif
  static void reportTermination();
  static void reportAbortion();
  static void outputProof();
  static void tabulateProof();
  static void assertHeaderRanks();
  static void setSignatureProperties(); 

 private: // predefined strategies
  static Master::Options _manualStrat;
  static Master::Options _autoNNE;
  static Master::Options _autoHNE1;
  static Master::Options _autoHNE2;
  static Master::Options _autoNEQ;
  static Master::Options _autoHEQ;
  static Master::Options _autoPEQ;
  static Master::Options _autoUEQ;
 private:
  static Timer _timer;
  static ulong _nextClauseNum;
  static Master _master; 
  static ClauseQueue _inputClauses;
  static ClauseQueue _preprocessedClauses;
  static ClauseQueue _preprocessingHistory;
  static InferenceTree _proofTree;
  static ClauseNumIndex _clauseNumIndex;  
  static ProblemProfile _problemProfile;
  static bool _definedHeader[HellConst::MaxNumOfDifferentHeaders]; 
  static ulong _numOfDefinedHeaders;  
  static bool _solvedByPreprocessing;    
 friend class InitStatic;                   
}; // class Vampire
 
inline void Vampire::Options::reset()
{
 
 incompleteAPriori = false;

 // automatic mode
 autoStrat = false;
  
 // main algorithm 
 
 main_alg = 0;

 // inference rules
 hyper_resolution = false;
 no_inferences_between_rules = false;
 static_splitting = false;
 dynamic_splitting = false;
 splitting_style = 0L;
 splitting_with_naming = false;
 static_neg_eq_splitting = 0L;
 paramodulation = false;
 sim_back_sup = 0;
 hyper_positive = false;
 selection = 1;
 literal_comparison_mode = 1;
 inherited_negative_selection = false;     

 // strategy 
 forward_subsumption = false;
 fs_set_mode = false;
 fs_optimised_multiset_mode = false;
 fs_use_sig_filters = false;
 fs_old_style = false;
 simplify_by_fs = 0;
 forward_demodulation = 0;
 fd_on_splitting_branches = false;
 normalize_by_commutativity = false;
 backward_demodulation = 0;
 bd_on_splitting_branches = false;
 simplify_by_eq_res = false;
 backward_subsumption = false;
 bs_set_mode = false;
 bs_optimised_multiset_mode = false;
 orphan_murder = 0;
 unit_conflict = false;
 lrs = 0;
 plot_usage_pace = (ofstream*)0;
 plot_weight_limit_changes = (ofstream*)0;
 elim_def = 0;
 header_rank_by_arity = 0;
 header_rank_kinky = false;
 header_rank_check_positive_selection = false;

 // numeric options 
 timeLimit = (float)LONG_MAX;
 memoryLimit = (float)LONG_MAX;
 allocation_buffer_price = 1.0; 
 maxActive = LONG_MAX;
 maxPassive = LONG_MAX;
 maxWeight = LONG_MAX;
 optGrCompWeight = LONG_MAX;
 age_weight_ratio = 0;    
 selected_pos_eq_penalty_coefficient = 1.0;
 internal_tick_frequency = LONG_MAX;
 first_time_check = LONG_MAX;   
 first_mem_check = LONG_MAX;    
    
  // interface options    
 showOpt = false;
 showInput = false;
 showPrepro = false;
 showGen = false;
 showPassive = false;
 showActive = false;
 showHyperRules = false;
 showDefinitions = false;
 showProof = false;
 showProfile = false;
 showWeightLimitChanges = false;
 tabulation = (Tabulation*)0;
}; // void Vampire::Options::reset() 

inline ostream& Vampire::Options::output(ostream& str) const
{
 str << "%================== Options: =======================\n";
 str << "% incompleteAPriori = " << incompleteAPriori << '\n';
 // automatic mode
 str << "% autoStrat = " << autoStrat << '\n';

 // main algorithm

 str << "% main_alg = " << main_alg << '\n';
  
 // inference rules
 str << "% hyper_resolution = " << hyper_resolution << '\n';
 str << "% no_inferences_between_rules = " << no_inferences_between_rules << '\n';
 str << "% static_splitting = " << static_splitting << '\n';
 str << "% splitting_style = " << splitting_style << '\n'; 
 str << "% splitting_with_naming = " << splitting_with_naming << '\n'; 
 str << "% dynamic_splitting = " << dynamic_splitting << '\n';
 str << "% static_neg_eq_splitting = " << static_neg_eq_splitting << '\n';
 str << "% paramodulation = " << paramodulation << '\n'; 
 str << "% sim_back_sup = " << sim_back_sup << '\n';  
 str << "% hyper_positive = " << hyper_positive << '\n';  
 str << "% selection = " << selection << '\n';
 str << "% literal_comparison_mode = " << literal_comparison_mode << '\n';
 str << "% inherited_negative_selection = " << inherited_negative_selection << '\n'; 
     
 // strategy 
 str << "% forward_subsumption = " << forward_subsumption << '\n'; 
 str << "% fs_set_mode = " << fs_set_mode << '\n'; 
 str << "% fs_optimised_multiset_mode = " << fs_optimised_multiset_mode << '\n'; 
 str << "% fs_use_sig_filters = " << fs_use_sig_filters << '\n';
 str << "% fs_old_style = " << fs_old_style << '\n';  
 str << "% simplify_by_fs = " << simplify_by_fs << '\n';
 str << "% forward_demodulation = " << forward_demodulation << '\n';
 str << "% fd_on_splitting_branches = " << fd_on_splitting_branches << '\n';
 str << "% normalize_by_commutativity = " << normalize_by_commutativity << '\n';
 str << "% backward_demodulation = " << backward_demodulation << '\n';
 str << "% bd_on_splitting_branches = " << bd_on_splitting_branches << '\n';
 str << "% simplify_by_eq_res = " << simplify_by_eq_res << '\n'; 
 str << "% backward_subsumption = " << backward_subsumption << '\n'; 
 str << "% bs_set_mode = " << bs_set_mode << "\n";
 str << "% bs_optimised_multiset_mode = " << bs_optimised_multiset_mode << "\n";
 str << "% orphan_murder = " << orphan_murder << '\n'; 
 str << "% unit_conflict = " << unit_conflict << '\n'; 
 str << "% lrs = " << lrs << '\n';
 str << "% header_rank_by_arity = " << header_rank_by_arity << '\n';
 str << "% header_rank_kinky = " << header_rank_kinky << '\n';
 str << "% header_rank_check_positive_selection " << header_rank_check_positive_selection << '\n';

 // numeric options 
 str << "% timeLimit = " << timeLimit  << '\n';
 str << "% memoryLimit = " << memoryLimit << '\n';
 str << "% allocation_buffer_price = " << allocation_buffer_price << '\n';
 str << "% maxActive = " << maxActive << '\n'; 
 str << "% maxPassive = " << maxPassive << '\n';  
 str << "% maxWeight = " << maxWeight << '\n';  
 str << "% optGrCompWeight = " << optGrCompWeight << '\n'; 
 str << "% age_weight_ratio = " << age_weight_ratio << '\n';
 str << "% selected_pos_eq_penalty_coefficient = " 
     << selected_pos_eq_penalty_coefficient << '\n';  
 str << "% internal_tick_frequency = " << internal_tick_frequency << '\n';  
 str << "% first_time_check = " << first_time_check << '\n';   
 str << "% first_mem_check = " << first_mem_check << '\n';  
    
  // interface options  
 str << "% tabulation = " << (bool)tabulation << '\n';
 
 str << "%========== End of options: ======================\n";
 str.flush();
 return str;
}; // ostream& Vampire::Options::output(ostream& str) const


inline void Vampire::makeOptionsConsistent()
{
 CALL("makeOptionsConsistent()");
 switch (options.selection)
  { 
   case 1006L:
   case 1007L:
   case 1008L:
   case 1009L:
   case -1009L:
    options.inherited_negative_selection = false; 
    cout << "% WARNING: bidirectional literal selection "
         << options.selection 
         << " switches inherited_negative_selection off!\n";
  };
 if ((options.selection < 0)
     && (options.literal_comparison_mode != 0)
     && (!options.header_rank_check_positive_selection))
  {
   cout << "% WARNING: positive selection is used without --header_rank_check_positive_selection,\n";
   cout << "%          the strategy is considered as incomplete.";
   options.incompleteAPriori = true;
  };
 
 if (options.splitting_style == 0)
  {
   cout << "% WARNING: --splitting_style 0 causes incompleteness\n";
   cout << "%          in the current version.";
   options.incompleteAPriori = true;
  };

 if (options.fs_set_mode || options.bs_set_mode)
  {
   options.incompleteAPriori = true;
  };
}; // void Vampire::makeOptionsConsistent()

static Vampire::InitStatic vampireInitStatic;

//=================================================
#endif
