// *************************************************************
//
//  Class Alist<C,D> for implementing association lists
//
//  14/05/2000, Dallas
//  04/06/2000, Manchester, length () added
//
// *************************************************************

#ifndef __alist__
#  define __alist__


// #include "memory.h"
#include "debug.h"
#include "pair.h"
#include "list.h"


// *******************Class Alist*******************************


template <class C,class D>
class Alist
{
 public:

  typedef Pair <C,D> AssocPair;
  typedef List <AssocPair*> AssocList;
  typedef AssocList* Mark;

  // constructors and destructors
  Alist ();
  ~Alist ();

  bool bound ( C, D& );
  bool rassoc ( D, C& );
  void bind ( C, D );      // binds C to D without checking that C is unbound
  void replace ( C, D );  // rebinds C to D, if C was bound, otherwise binds C to D
  Mark mark ();           // returns mark of the current state
  void restore ( Mark );  // restores the state to the marked one
  int length ();

 protected:  // structure
  AssocList* list;

  // for debugging 
  static const char* name ();
};  // class Alist


// ******************* Class Alist, definitions ****************************


template <class C,class D>
inline
const char* Alist<C,D>::name ()
{
  return "Alist";
}


template <class C,class D>
inline
Alist<C,D>::Alist ()
  :
  list ( AssocList::empty () )
{
} // Alist::Alist


template <class C,class D>
bool Alist<C,D>::bound ( C lft, D& rht )
{
  for ( AssocList* lst = list; ! lst->isEmpty (); lst = lst->tail () ) {
    AssocPair* head = lst->head ();
    if ( head->left () == lft ) {
      rht = head->right ();
      return true;
    }
  }

  // not found
  return false;
} // Alist::assoc


template <class C,class D>
bool Alist<C,D>::rassoc ( D rht, C& lft )
{
  for ( AssocList* lst = list; ! lst->isEmpty (); lst = lst->tail () ) {
    AssocPair* head = lst->head ();
    if ( head->right () == rht ) {
      lft = head->left ();
      return true;
    }
  }

  // not found
  return false;
} // Alist::rassoc


template <class C,class D>
void Alist<C,D>::bind ( C lft, D rht )
{
  list->push ( new AssocPair (lft, rht), list );
} // Alist::add


template <class C,class D>
Alist<C,D>::~Alist ()
{
  // destroy all pairs in the list
  for ( AssocList* lst = list; ! lst->isEmpty (); lst = lst->tail () ) 
    delete lst->head ();

  // destroy the list itself
  list->destroy ();
} // Alist::~Alist


template <class C,class D>
void Alist<C,D>::replace ( C lft, D rht )
{
  for ( AssocList* lst = list; ! lst->isEmpty (); lst = lst->tail () ) {
    AssocPair* head = lst->head ();
    if ( head->left () == lft ) {
      head->right ( rht );
      return;
    }
  }

  // not found
  add ( lft, rht );
} // Alist::replace


template <class C,class D>
inline
Alist<C,D>::Mark Alist<C,D>::mark ()
{
  return list;
}


template <class C,class D>
void Alist<C,D>::restore ( Mark mark )
{
  // destroy all pairs in the list up the mark
  while ( list != mark ) {
    ASSERT( debug_alist, ! list->isEmpty () );

    delete list->head ();
    AssocList* tail = list->tail ();
    delete list;

    list = tail;
  }
} // Alist::restore


template <class C,class D>
int Alist<C,D>::length ()
{
  return list->length ();
} // Alist::length


#endif // alist
