//
//  file atom.h 
//  defines class Atom
//
//  01/04/2001, Machester, made from term.h (to make the code cleaner)
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: destroy (), introduceInverse()
//  30/05/2001 Manchester, splitLargeArity/2
//  31/05/2001 Manchester, normalize/0, negativeEqualitySplit/1
//

#ifndef __atom__
#define __atom__

class FSymbol;

#include "typedefs.h"
#include "int.h"
#include "term.h"
#include "psymbol.h"


// ****************** class Atom, definition *************************


class Atom {
 public:
  
  typedef List <Atom*> List;

  // constructors
  Atom ( PSymbol*, Term::List* ); 

  // query the structure
  PSymbol* functor () const;
  bool isEquality () const;
  Term::List* args () const;
  int arity () const;

  // deallocation and copying
  void destroy ();
  Atom* copy () const;

  // statistics
  int countOccurrences ( const FSymbol* ) const;
  int countOccurrences ( const Var ) const;
  Var maxVar ();
  int weight () const;
  bool occurs ( const FSymbol* ) const;

  // miscellaneous
  void rename ( VarSubst& );
  bool equal ( Atom* ) const;
  Atom* apply ( Subst&, VarSubst&, int& lastVar ); // copies the argument
  Atom* replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs );
  void replace ( Var, Term* );  // actually reuses the memory allocated for the term
  bool isLinearAndFlat () const; // atom has the form P (x1,...,xn) where xi are pairwise different
  Comparison compare ( const Atom* ) const;  // comparison of two atoms, needed to normalize
  void apply ( Subst& );  // actually reuses the memory allocated for the term
  void introduceInverse ( int argNo,FSymbol* inv );
  void splitLargeArity ( const PSymbol* p, FSymbol::List* newFs );
  void splitLargeArity ( const FSymbol* f, FSymbol::List* newFs );
  void normalize ();
  Atom* negativeEqualitySplit ( int thresholdWeight );

/*
  bool isFlat (); // term has the form f (x1,...,xn) where xi are variables

  // deallocation and copying
  static Term::List* copy ( const Term::List* );

  // miscellaneous
  bool occurs ( Var );
  static bool unify ( IndexedTerm, IndexedTerm, Dsubst& );
  bool isLinear ( Term* rhs ); // in the definition this = rhs, every variable of this occurs
                               // at most once in rhs
*/

 private: 
  // structure
  PSymbol* _functor;
  Term::List* _args;

  void functor ( PSymbol* );
  void args ( Term::List* );

/*
  static int weight ( Term::List* );
  static bool occurs ( Term::List*, Functor* );
  static bool occurs ( Term::List*, Var );
  static bool occurs ( IndexedVar, IndexedTerm, Dsubst& );
  static Term::List* split ( Term::List*, Functor* tosplit, int maxArity, int iterations, int restArity, 
                           Functor* firstFun, Functor* restFun );
  static Term::List* apply ( Term::List*, Subst&, VarSubst&, int& lastVar ); // copies the list
  static Comparison compare ( const Term::List*, const Term::List* );  // comparison of two termlists
*/
}; // class Atom


ostream& operator << ( ostream&, Atom* );


// ****************** class Atom, implementation *************************


inline
Atom::Atom ( PSymbol* f, Term::List* args ) 
  :
  _functor( f ),
  _args ( args )
{
} // Atom::Atom


inline
Term::List* Atom::args () const
{
  return _args;
} // Atom::args


inline
void Atom::args ( Term::List* ts )
{
  _args = ts;
} // Atom::args


inline
void Atom::functor ( PSymbol* f )
{
  _functor = f;
} // Atom::functor


inline
PSymbol* Atom::functor () const
{
  return _functor;
} // Atom::functor


inline
int Atom::arity () const
{
  return _functor->arity ();
} // Atom::arity


inline
bool Atom::isEquality () const
{
  return _functor->isEquality ();
} // Atom::isEquality


#endif // __atom__
