//
//  file fsymbol.h 
//  defines class FSymbol
//
//  21/05/2001 Manchester, made from name_table.h and functor.h
//  30/05/2001 Manchester, noOfSymbols(), signature(), arity/1
//

#ifndef __fsymbol__
#define __fsymbol__


// ***************** class Fsymbol, definition ****************************


class ostream;


#include "list.h"


class FSymbol {
 public:
 
  typedef List <FSymbol*> List;

  static FSymbol* create ( const char* name, int arity );
  static FSymbol* createNew ( const char* prefix, int arity );  // new function with prefix
  int arity () const;
  void arity ( int ar ) { _arity = ar; };
  const char* name () const;
  static FSymbol** signature (); // return a new array of all symbols
  static int noOfSymbols () { return _noOfSymbols; }
  static void arities ( int max, int* ars );
 
 private:
 
  FSymbol ( const char* name, int arity );
  const char* _name;
  int _arity;
  int _number;

  static int _noOfSymbols;

  // hash table for names
  static const int _noOfBuckets;
  static List * _buckets [];
  static int hash ( const char* );  // hashing function

  // to initialize name table
  static FSymbol _dummy; 
  FSymbol ();

 friend ostream& operator << ( ostream&, const FSymbol* );
}; // class FSymbol


// ***************** class FSymbol, implementation ****************************


inline
FSymbol::FSymbol ( const char* name, int arity )
  :
  _name ( name ),
  _arity ( arity ),
  _number (_noOfSymbols++)
{
}


inline
const char* FSymbol::name () const
{
  return _name;
}


inline
int FSymbol::arity () const
{
  return _arity;
}



#endif // __fsymbol__
