//
//  file literal.h 
//  defines class Literal
//
//  03/06/2000, Manchester 
//  04/06/2000, Manchester, atom () 
//  09/06/2000, Manchester, complementary() 
//  10/06/2000, Manchester, sign()
//  11/06/2000, Manchester, apply()
//  29/03/2001, Machester, Functor changed to Functor* 
//  31/03/2001, Manchester, _weight and weight () added, Literal(,,)
//  15/04/2001, Manchester, sort, copy moved from clause.h
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: destroy (), introduceInverse()
//  30/05/2001 Manchester, splitLargeArity/2
//  31/05/2001 Manchester, normalize/0, negativeEqualitySplit/1
//


#ifndef __literal__
#define __literal__


#include "typedefs.h"
#include "atom.h"


// ******************* class Literal, definition *********************


class Literal {
 public:

  typedef List <Literal*> List;

  // constructor
  Literal ( bool sign, Atom* atom );
  Literal ( bool sign, Atom* atom, int weight );

  // query/change the structure
  bool positive () const;
  bool negative () const;
  bool sign () const;
  Atom* atom () const;
  void atom ( Atom* );
  void destroy ();

  //miscellaneous
  int countOccurrences ( const FSymbol* ) const;
  int countOccurrences ( Var ) const;
  bool occurs ( const FSymbol* ) const;
  bool complementary ( Literal* ); // not for equality literals
  Var maxVar ();
  Literal* apply ( Subst&, VarSubst&, int& lastVar ); // copies the argument
  Literal* copy () const; 
  static List* copy ( const List* );
  static List* sort ( List* );
  int weight (); // weight in the number of symbols
  Comparison compare ( const Literal* ) const;  // comparison of two literals, needed to normalize
  void introduceInverse ( int argNo,FSymbol* inv );
  void splitLargeArity (const PSymbol* p, FSymbol::List* newFs );
  void splitLargeArity (const FSymbol* p, FSymbol::List* newFs );
  void normalize ();
  Literal* negativeEqualitySplit ( int thresholdWeight );

 private:
  
  bool _sign;
  Atom* _atom;
  int _weight;  // -1 is not computed before

  static List* sort ( List* toSort, List* appendTo );
  static void split ( Literal*, List*, // for sorting
                      List*& less, List*& equal, List*& greater );
}; // class Literal


ostream& operator << ( ostream&, const Literal* );
ostream& operator << ( ostream&, Literal* );


// ******************* class Literal, implementation *********************


inline 
Literal::Literal ( bool sign, Atom* atom )
  :
  _sign ( sign ),
  _atom ( atom ),
  _weight ( -1 )
{
} // Literal::Literal


inline 
Literal::Literal ( bool sign, Atom* atom, int weight )
  :
  _sign ( sign ),
  _atom ( atom ),
  _weight ( weight )
{
} // Literal::Literal


inline
bool Literal::sign () const
{
  return _sign;
} // Literal::positive


inline
bool Literal::positive () const
{
  return _sign;
} // Literal::positive


inline
bool Literal::negative () const
{
  return ! _sign;
} // Literal::negative


inline
Atom* Literal::atom () const
{
  return _atom;
} // Literal::atom
 

inline
void Literal::atom ( Atom* t )
{
  _atom = t;
  _weight = -1;
} // Literal::atom
 

inline
int Literal::countOccurrences ( const FSymbol* f ) const
{
  return atom()->countOccurrences ( f );
} // Literal::countOccurrences


inline
bool Literal::occurs ( const FSymbol* f ) const
{
  return atom()->occurs ( f );
} // Literal::occurs


inline
int Literal::countOccurrences ( Var v ) const
{
  return atom()->countOccurrences ( v );
} // Literal::countOccurrences


inline 
Var Literal::maxVar ()
{
  return atom()->maxVar ();
} // Literal::maxVar


inline 
bool Literal::complementary ( Literal* l )
{
  return _sign != l->_sign &&
         ( ! atom()->isEquality() ) &&
         atom()->equal( l->atom() );
} // Literal::complementary


inline
Literal* Literal::apply ( Subst& s, VarSubst& vs, int& lastVar ) // copies the argument
{
  return new Literal ( sign(), atom()->apply(s,vs,lastVar) );
} // Literal::apply


inline
Literal* Literal::copy () const
{
  return new Literal ( _sign, _atom->copy(), _weight );
} // Literal::copy


#endif
