//
//  file options.cpp
//  implements Vampire options
//
//  11/06/2000, Manchester
//  11/12/2000, Manchester, subsumption modes added, defaults changed to optimized subsumption
//  31/12/2000, Manchester, splitting-related options changed
//  03/01/2001, Manchester, demodulationOnSplittingBranches added
//  06/06/2001 Manchester, completely rewritten
//


#include <iostream.h>

#include "assert.h"
#include "debug.h"
#include "options.h"


// 05/06/2001
// makes default options
Options::Options ()
  :
  _testId (0),     
  _tabFile (0)
{
  setDefault ();
} // Options::Options


// 05/06/2001
bool Options::validSelection ( int sel )
{
  switch (sel) {
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 1002:
    case 1003:
    case 1004:
    case 1005:
    case 1006:
    case 1007:
    case 1008:
    case 1009:
    case -2:
    case -3:
    case -4:
    case -5:
    case -6:
    case -7:
    case -1005:
    case -1009:
      return true;

    default:
      return false;
  }
} // Options::validSelection


// 06/06/2001
bool Options::completeSelection ( int sel )
{
  switch (sel) {
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case -2:
    case -3:
    case -4:
    case -5:
    case -6:
    case -7:
      return true;

    default:
      return false;
  }
} // Options::completeSelection


// 06/06/2001
bool Options::bidirectionalSelection ( int sel )
{
  switch (sel) {
    case 1006:
    case 1007:
    case 1008:
    case 1009:
    case -1009:
      return true;

    default:
      return false;
  }
} // Options::bidirectionalSelection


// 05/06/2001
void Options::selection (int newVal)
{
  ASSERT ( debug_options, validSelection (newVal) );
  
  _selection = newVal;
} // Options::selection


// 06/06/2001
void Options::awr (int newVal)
{
  ASSERT ( debug_options, newVal > 0 );

  _awr = newVal;
} // Options::awr


// 06/06/2001
void Options::memoryLimit (int newVal)
{
  ASSERT ( debug_options, newVal >= 0 );

  _memoryLimit = newVal;
} // Options::memoryLimit


// 06/06/2001
void Options::timeLimit (int seconds)
{
  ASSERT ( debug_options, seconds >= 0 );

  _timeLimit = seconds;
} // Options::timeLimit


// 06/06/2001
void Options::defElimination (int rounds)
{
  ASSERT ( debug_options, rounds >= 0 );

  _defElimination = rounds;
} // Options::_defElimination


// 06/06/2001
void Options::groundSplitThreshold (int gst)
{
  ASSERT ( debug_options, gst >= 0 );

  _groundSplitThreshold = gst;
} // Options::groundSplitThreshold


// 06/06/2001
void Options::setDefault ()
{
  mainLoop (LRS);
  selection (1);
    inheritedSelection (false);
  precedence (ARITY);
  literalComparison (RANK);
  awr (5);
  //memoryLimit (128);
  //timeLimit (60); 
  maxWeight (0);
  maxActive (0);
  maxPassive (0);
  hyper (HNONE);
    unitConflict (true);
  splitting (NONE);
    staticOnly (false);
    groundSplitThreshold (3);
    naming (true);
    branchFD (true);
    branchBD (true);
  // simplifications
  forwardSubsumption (MULTISET);
    fsSigFilters (false);
    fsNonAdaptive (false);
  backwardSubsumption (MULTISET);
  orphanMurder (false);
  forwardSubsumptionResolution (true);
  // equality options
  paramodulation (true);
  simplifyByEqRes (true);
  forwardDemodulation (ALL);
  backwardDemodulation (ALL);
  commutativity (false);
  // memory management (lrs only)
  allocationBufferSize (10);
  allocationBufferPrice (2);
  internalTickFrequency (100);
  firstTimeCheck (5);
  firstMemCheck (0);
  // preprocessing
  defElimination (0); 
  // info options
  testId (0);
  //tabFile (0);
  showOpt (false);
  showInput (false);
  showPrepro (false);
  showGenerated (false);
  showPassive (false);
  showActive (false);
  showDefs (false);
  showWeightLimitChanges (false);
  showProfile (false);
  proofOutput (true);
} // Options::setDefault


// 06/06/2001
void Options::output ( ostream& str )
{
  switch ( mainLoop() ) {
    case OTTER:
      str << " --main_alg 0 --lrs 0";
      break;
    case DISCOUNT:
      str << " --main_alg 3 --lrs 0";
      break;
    case LRS:
      str << " --main_alg 0 --lrs 1";
      break;
  }

  str << " --selection " << selection ();
  if ( bidirectionalSelection(selection()) ) {
    str << " --no_inferences_between_rules";
  }
  if ( inheritedSelection() )
    str << " --inherited_negative_selection";

  str << " --age_weight_ratio " << awr();

  if ( timeLimit() ) {
    str << " -t " << timeLimit ();
  }
  if ( memoryLimit() ) {
    str << " -m " << memoryLimit () * 1000;
  }

  switch ( precedence()) {
    case PNONE:
      break;
    case OCCURRENCE:
      str << " --header_rank_by_arity 0";
      break;
    case ARITY:
      str << " --header_rank_by_arity 1";
      break;
    case REVERSE_ARITY:
      str << " --header_rank_by_arity -1";
      break;
    case KINKY:
      str << " --header_rank_kinky";
      break;
  };

  str << " --literal_comparison_mode ";
  switch ( literalComparison() ) {
    case NORMAL:
      str << 0;
      break;
    case RANK:
      str << 1;
      break;
    case PREDICATE:
      str << 2;
      break;
  }

  switch ( hyper() ) {
    case HPOS:
      str << " --hyper_positive";
    case HNEG:
      str << " --hyper";
      if ( unitConflict() ) {
        str << " --unit_conflict";
      }
      break;

    case HNONE:
      break;
  }

  if ( splitting() != NONE ) {
    str << " --static_splitting";
    if ( ! staticOnly() ) {
      str << " --dynamic_splitting";
    }

    str << " --splitting_style ";
    switch ( splitting() ) {
      case NONE: // just to avoid warnings
      case OLD:
        str << 0;
        break;
      case BLOCKING:
        str << 2;
        break;
      case PARALLEL:
        str << 1;
        break;
    }
    if ( naming() ) 
      str << " --splitting_with_naming";
    if ( groundSplitThreshold() ) 
      str << " --opt_gr_comp_weight " << groundSplitThreshold();
    if ( branchFD() ) 
      str << " --fd_on_splitting_branches";
    if ( branchBD() ) 
      str << " --bd_on_splitting_branches";
  }

  if ( paramodulation() ) 
    str << " --paramodulation";
  if ( simplifyByEqRes() ) 
    str << " --simplify_by_eq_res";

  switch ( forwardDemodulation() ) {
    case ALL:
      str << " --forward_demodulation 4";
      break;
    case PREORDERED:
      str << " --forward_demodulation 1";
      break;
    case OFF:
      break;
  }

  switch ( backwardDemodulation() ) {
    case ALL:
      str << " --backward_demodulation 4";
      break;
    case PREORDERED:
      str << " --backward_demodulation 1";
      break;
    case OFF:
      break;
  }

  if ( commutativity() )
    str << " --normalize_by_commutativity";

  if ( forwardSubsumptionResolution() ) 
    str << " --simplify_by_fs 1";

  if ( forwardSubsumption() != SNONE ) {
    switch ( forwardSubsumption() ) {
    case SNONE: // just to avoid warnings
      case SET:
        str << " --forward_subsumption --fs_set_mode";
        break;
      case MULTISET:
        str << " --forward_subsumption";
        break;
      case OPTIMIZED:
        str << " --forward_subsumption --fs_optimised_multiset_mode";
        break;
    }

    if ( fsSigFilters() ) 
      str << " --fs_use_sig_filters";
    if ( fsNonAdaptive() ) 
      str << " --fs_old_style";
  }

  switch ( backwardSubsumption() ) {
    case SNONE:
      break;
    case SET:
      str << " --backward_subsumption --bs_set_mode";
      break;
    case MULTISET:
      str << " --backward_subsumption";
      break;
    case OPTIMIZED:
      str << " --backward_subsumption --bs_optimised_multiset_mode";
      break;
  }

  if ( orphanMurder() ) 
    str << " --orphan_murder 1";

  // memory management (lrs only)
  if ( mainLoop() == LRS || mainLoop() == OTTER ) {
    str << " --allocation_buffer_size " << allocationBufferSize()
        << " --allocation_buffer_price " << allocationBufferPrice ()
        << " --internal_tick_frequency " << internalTickFrequency () 
        << " --first_time_check " << firstTimeCheck ();

    if ( firstMemCheck () )
      str << " --first_mem_check " << firstMemCheck ();
  }

  if ( defElimination() != 0 ) 
    str << " --elim_def " << defElimination ();

  if ( proofOutput() ) 
    str << " -p";
  if ( testId() ) 
    str << " -T " << testId ();
  if ( tabFile() ) 
    str << " --tab " << tabFile ();
  if ( showOpt() ) 
    str << " --show_opt";
  if ( showInput() )
    str << " --show_input";
  if ( showPrepro() )
    str << " --show_prepro";
  if ( showGenerated() ) 
    str << " --show_gen";
  if ( showPassive() ) 
    str << " --show_passive";
  if ( showActive() ) 
    str << " --show_active";
  if ( showDefs() )
    str << " --show_definitions";
  if ( showWeightLimitChanges() )
    str << " --show_weight_limit_changes";
  if ( showProfile() ) 
    str << " --show_profile";
  if ( maxWeight() ) 
  str << " -w " << maxWeight ();
  if ( maxActive () )
    str << " --max_active " << maxActive();
  if ( maxPassive() )
    str << " --max_passive " << maxPassive();
} // Options::output


// 06/06/2001
ostream& operator << ( ostream& str , Options& opt )
{
  opt.output (str);
  return str;
} // operator << ( ostream& str , Options& opt )
  

// 06/06/2001
void Options::cancelEquality ()
{
  paramodulation (false);
  simplifyByEqRes (false);
  forwardDemodulation (OFF);
  backwardDemodulation (OFF);
  commutativity (false); 
  branchFD (false);
  branchBD (false);
} // Options::cancelEquality


// 06/06/2001
void Options::otter ( int sel, int ratio )
{
  //  setDefault ();
  mainLoop (OTTER);
  selection (sel);
  awr (ratio);
} // Options::otter


// 06/06/2001
void Options::discount ( int sel, int ratio )
{
  // setDefault ();
  mainLoop (DISCOUNT);
  selection (sel);
  awr (ratio);
} // Options::otter


// 06/06/2001
void Options::lrs ( int sel, int ratio )
{
  // setDefault ();
  mainLoop (LRS);
  selection (sel);
  awr (ratio);
} // Options::otter


// 07/06/2001
void Options::ueq ( MainLoop loop, int ratio )
{
  // setDefault ();
  mainLoop (loop);
  selection (1);
  awr (ratio);
} // Options::ueq


// 07/06/2001
void Options::fullSplitting ()
{
  splitting (PARALLEL);
  staticOnly (false);
  groundSplitThreshold (3);
  naming (true);
  branchFD (true);
  branchBD (true);
} // Options::fullSplitting


void Options::str001 ()
{
  lrs (4,1);
  forwardDemodulation (PREORDERED);
  backwardDemodulation (OFF);
} 


void Options::str002 ()
{
  ueq (LRS, 20);

  orphanMurder (true);
  forwardSubsumptionResolution (true);
} 


void Options::str003 ()
{
  ueq (LRS, 26);
  backwardDemodulation (OFF);
  forwardSubsumptionResolution (true);
} 


void Options::str004 ()
{
  ueq (LRS, 13);
  backwardDemodulation (PREORDERED);
} 


void Options::str009 ()
{
  discount (-1005, 8);
  literalComparison (NORMAL);
  precedence (OCCURRENCE);
  backwardSubsumption (SNONE);
} 


void Options::str011 ()
{
  lrs (4, 3);
  inheritedSelection (true);
  forwardSubsumption (OPTIMIZED);
  backwardSubsumption (SET);
  literalComparison (NORMAL);
  precedence (OCCURRENCE);
} 


void Options::str010 ()
{
  discount (6, 2);

  forwardSubsumption (OPTIMIZED);
  literalComparison (RANK);
  orphanMurder (true);
  precedence (REVERSE_ARITY);
} 


void Options::str012 ()
{
  lrs (4, 4);

  forwardSubsumption (SET);
  backwardSubsumption (SNONE);
  literalComparison (PREDICATE);
  forwardSubsumptionResolution (false);
  precedence (KINKY);
} 


void Options::str013 ()
{
  lrs (4, 26);
  inheritedSelection (true);
  literalComparison (NORMAL);
  fullSplitting ();
  splitting (OLD);
  commutativity (true);
  orphanMurder (true);
} 


void Options::str016 ()
{
  discount (7, 22);
  forwardSubsumption (OPTIMIZED);
  literalComparison (PREDICATE);
  precedence (KINKY);
} 


void Options::str017 ()
{
  discount (-2, 13);
  forwardSubsumption (OPTIMIZED);
  backwardSubsumption (SNONE);
  literalComparison (PREDICATE);
  forwardSubsumptionResolution (false);
  precedence (KINKY);
} 


void Options::str014 ()
{
  lrs (1, 23);

  forwardSubsumption (OPTIMIZED);
  backwardSubsumption (SNONE);
  literalComparison (PREDICATE);
  backwardDemodulation (PREORDERED);
  forwardSubsumptionResolution (false);
  orphanMurder (true);
  precedence (KINKY);
} 


void Options::str015 ()
{
  lrs (-6,1);

  backwardSubsumption (SNONE);
  literalComparison (RANK);
  backwardDemodulation (PREORDERED);
  orphanMurder (true);
  precedence (KINKY);
} 


void Options::str007 ()
{
  lrs (1, 10);
  inheritedSelection (true);
  forwardSubsumption (SET);
  literalComparison (PREDICATE);
  forwardSubsumptionResolution (false);
  precedence (OCCURRENCE);
} 


void Options::str008 ()
{
  discount (-4, 26);

  inheritedSelection (true);
  forwardSubsumption (OPTIMIZED);
  backwardSubsumption (OPTIMIZED);
  literalComparison (PREDICATE);
  backwardDemodulation (PREORDERED);
  precedence (KINKY);
} 


void Options::str005 ()
{
  lrs (1004, 7);

  forwardSubsumption (SET);
  orphanMurder (true);
} 


void Options::str006 ()
{
  lrs (5, 35);
  backwardSubsumption (SNONE);
  orphanMurder (true);
} 


void Options::str018 ()
{
  lrs (1009, 28);
  forwardSubsumption (OPTIMIZED);
  forwardSubsumptionResolution (false);
  orphanMurder (true);
  precedence (KINKY);
} 


void Options::str023 ()
{
  discount (4, 16);
  
  forwardSubsumption (SET);
  backwardSubsumption (OPTIMIZED);
  literalComparison (NORMAL);
  forwardSubsumptionResolution (false);
  orphanMurder (false);
} 


void Options::str022 ()
{
  discount (7, 18);
  inheritedSelection (true);
  backwardSubsumption (SNONE);
  literalComparison (NORMAL);
} 


void Options::str024 ()
{
  discount (7,6);

  forwardSubsumption (SET);
  backwardSubsumption (OPTIMIZED);
  literalComparison (PREDICATE);
  precedence (OCCURRENCE);
} 


void Options::str025 ()
{
  otter (4, 2);
  inheritedSelection (true);
  backwardSubsumption (SNONE);
  literalComparison (NORMAL);
} 


void Options::str019 ()
{
  discount (-7, 3);

  inheritedSelection (true);
  forwardSubsumption (SET);
  backwardSubsumption (SET);
  literalComparison (PREDICATE);
  precedence (KINKY);
  orphanMurder (true);
  forwardSubsumptionResolution (false);  
} 


void Options::str020 ()
{
  lrs (4, 4);
  inheritedSelection (true);
  forwardSubsumption (SET);
  backwardSubsumption (SNONE);
  literalComparison (PREDICATE);
  forwardSubsumptionResolution (false);  
} 


void Options::str021 ()
{
  lrs (4, 7);
  inheritedSelection (true);
  literalComparison (PREDICATE);
} 


void Options::str026 ()
{
  discount (7, 6);
  literalComparison (PREDICATE);
  forwardSubsumption (OPTIMIZED);
  literalComparison (PREDICATE);
  precedence (KINKY);
} 


void Options::str027 ()
{
  lrs (1007, 3);
  literalComparison (RANK);
} 


void Options::str028 ()
{
  lrs (5, 8);
  forwardSubsumption (OPTIMIZED);
  backwardSubsumption (SET);
  literalComparison (RANK);
  forwardSubsumptionResolution (true);
} 


void Options::str029 ()
{
  lrs (7, 1);
  inheritedSelection (true);
  forwardSubsumption (SET);
  backwardSubsumption (SNONE);
  literalComparison (NORMAL);
  forwardSubsumptionResolution (false);  
  precedence (OCCURRENCE);
} 


void Options::str030 ()
{
  lrs (-3, 6);
  inheritedSelection (true);
  forwardSubsumption (OPTIMIZED);
  literalComparison (NORMAL);
  backwardDemodulation (OFF);
  orphanMurder (true);
} 


void Options::str031 ()
{
  discount (1003, 8);
  inheritedSelection (true);
  forwardSubsumption (SET);
  literalComparison (NORMAL);
  backwardDemodulation (PREORDERED);
  forwardSubsumptionResolution (false);
} 


void Options::str034 ()
{
  discount (1,16);
  forwardSubsumption (SET);
  backwardSubsumption (OPTIMIZED);
  literalComparison(NORMAL);
  orphanMurder (true);
  fullSplitting ();
  orphanMurder (true);
  precedence (OCCURRENCE);
} 


void Options::str035 ()
{
  lrs (1,1);
  literalComparison(PREDICATE);
  backwardSubsumption (OPTIMIZED);
  orphanMurder (true);
} 


void Options::str033 ()
{
  discount (1,6);
  forwardSubsumption (SET);

  literalComparison(PREDICATE);
  forwardSubsumptionResolution (false);
  precedence (KINKY);
} 


void Options::str032 ()
{
  discount (-5,14);

  inheritedSelection (true);
  literalComparison(PREDICATE);
} 


void Options::str037 ()
{
  lrs (1007,12);
  forwardSubsumption (OPTIMIZED);
  backwardSubsumption (SET);

  literalComparison(PREDICATE);
  backwardSubsumption (OPTIMIZED);
  orphanMurder (true);
} 


void Options::str036 ()
{
  lrs (2,5);

  fullSplitting ();
  splitting (OLD);
  precedence (KINKY);
} 


/*
    case PNONE:
      break;
    case OCCURRENCE:
      str << " --header_rank_by_arity 0";
    case ARITY:
      str << " --header_rank_by_arity 1";
    case REVERSE_ARITY:
      str << " --header_rank_by_arity -1";
    case KINKY:
      str << " --header_rank_kinky";
  };

  switch ( literalComparison() ) {
    case NORMAL:
      str << 0;
    case RANK:
      str << 1;
    case PREDICATE:
      str << 2;
*/
