//
//  file options.h
//  defines Vampire options
//
//  11/06/2000, Manchester
//  11/12/2000, Manchester, subsumption-related options changed
//  31/12/2000, Manchester, splitting-related options changed
//  31/12/2000, Manchester, inherited_negative_selection option added
//  03/01/2001, Manchester, demodulationOnSplittingBranches added
//  06/04/2001 Manchester, rewritten using options for Vampire 2.0
//

#ifndef __options__
#define __options__


class ostream;


class Options {
 public:

  // possible values for various options
  enum MainLoop {
    OTTER,
    DISCOUNT,
    LRS
  };

  enum Hyper {
    HPOS,
    HNEG,
    HNONE
  };

  enum Splitting {
    NONE,
    OLD,
    BLOCKING,
    PARALLEL
  };

  enum Subsumption {
    SNONE,
    SET,
    MULTISET,
    OPTIMIZED
  };

  enum Precedence {
    PNONE,
    OCCURRENCE,
    ARITY,
    REVERSE_ARITY,
    KINKY
  };

  enum Demodulation {
    ALL,
    PREORDERED,
    OFF
  };

  enum LitComp {
    NORMAL,
    RANK,
    PREDICATE
  };

  Options ();

  // all major main loops/calculi
  void otter ( int selection, int ratio );
  void discount ( int selection, int ratio );
  void lrs ( int selection, int ratio );
  void ueq ( MainLoop, int ratio );
  // switching to non-equality mode
  void cancelEquality ();
  void fullSplitting ();

  void output (ostream&);

  // publicly accessible options
  void tabFile (char* tf) { _tabFile = tf; }
  void testId (char* tid) { _testId = tid; }
  void memoryLimit (int newVal);
  void timeLimit (int newVal);
  void precedence (Precedence p) { _precedence = p; }
  void literalComparison (LitComp lc) { _literalComparison = lc; }
  void awr (int newVal);

  // sporadic options (from random runs)
  void str001 ();
  void str002 ();
  void str003 ();
  void str004 ();
  void str005 ();
  void str006 ();
  void str007 ();
  void str008 ();
  void str009 ();
  void str010 ();
  void str011 ();
  void str012 ();
  void str013 ();
  void str014 ();
  void str015 ();
  void str016 ();
  void str017 ();
  void str018 ();
  void str019 ();
  void str020 ();
  void str021 ();
  void str022 ();
  void str023 ();
  void str024 ();
  void str025 ();
  void str026 ();
  void str027 ();
  void str028 ();
  void str029 ();
  void str030 ();
  void str031 ();
  void str032 ();
  void str033 ();
  void str034 ();
  void str035 ();
  void str036 ();
  void str037 ();

  void reset () { setDefault(); }

 private:
  // get/set options
  MainLoop mainLoop () const { return _mainLoop; }
  void mainLoop ( MainLoop ml ) { _mainLoop = ml; }
  int selection () const { return _selection; }
  void selection (int newVal);
    bool inheritedSelection () const { return _inheritedSelection; }
    void inheritedSelection ( bool ins ) { _inheritedSelection = ins; }
  Precedence precedence () const { return _precedence; }
  LitComp literalComparison () const { return _literalComparison; }
  int awr () const { return _awr; }
  int memoryLimit () const { return _memoryLimit; }   // in MBytes
  int timeLimit () const { return _timeLimit; }   // in seconds
  int maxWeight () const { return _maxWeight; }
  void maxWeight (int mw) { _maxWeight = mw; }
  int maxActive () const { return _maxActive; }
  void maxActive (int mac) { _maxActive = mac; }
  int maxPassive () const { return _maxPassive; }
  void maxPassive (int mp) { _maxPassive = mp; }
  Hyper hyper () const { return _hyper; }
  void hyper (Hyper h) { _hyper = h; }
    bool unitConflict () const { return _unitConflict; }
    void unitConflict ( bool uc ) { _unitConflict = uc; }
  Splitting splitting () const { return _splitting; }
  void splitting (Splitting spl) { _splitting = spl; }
    bool staticOnly () const { return _staticOnly; }
    void staticOnly (bool so) { _staticOnly = so; }
    int groundSplitThreshold () const { return _groundSplitThreshold; }
    void groundSplitThreshold (int gst);
    bool naming () const { return _naming; }
    void naming ( bool nm ) { _naming = nm; }
    bool branchFD  () const { return _branchFD; }
    void branchFD  ( bool bfd ) { _branchFD = bfd; }
    bool branchBD () const { return _branchBD; }
    void branchBD  ( bool bbd ) { _branchBD = bbd; }
  
  // simplifications
  Subsumption forwardSubsumption () const { return _forwardSubsumption; };
  void forwardSubsumption ( Subsumption fs ) { _forwardSubsumption = fs; };
    bool fsSigFilters () const { return _fsSigFilters; }
    void fsSigFilters (bool sf) { _fsSigFilters = sf; }
    bool fsNonAdaptive () const { return _fsNonAdaptive; }
    void fsNonAdaptive (bool na) { _fsNonAdaptive = na; }
  Subsumption backwardSubsumption () const { return _backwardSubsumption; };
  void backwardSubsumption ( Subsumption bs ) { _backwardSubsumption = bs; };
  bool orphanMurder () const { return _orphanMurder; }
  void orphanMurder (bool om) { _orphanMurder = om; }
  bool forwardSubsumptionResolution () const { return _forwardSubsumptionResolution; }
  void forwardSubsumptionResolution (bool fsr) { _forwardSubsumptionResolution = fsr; }

  // equality options
  bool paramodulation () const { return _paramodulation; }
  void paramodulation (bool par) { _paramodulation = par; }
  bool simplifyByEqRes () { return _simplifyByEqRes; }
  void simplifyByEqRes (bool simp) { _simplifyByEqRes = simp; }
  Demodulation forwardDemodulation () const { return _forwardDemodulation; }
  void forwardDemodulation (Demodulation dem) { _forwardDemodulation = dem; }
  Demodulation backwardDemodulation () const { return _backwardDemodulation; }
  void backwardDemodulation (Demodulation dem) { _backwardDemodulation = dem; }
  bool commutativity () const { return _commutativity; }
  void commutativity (bool com) { _commutativity = com; }
  
  // memory management (lrs only)
  int allocationBufferSize () const { return _allocationBufferSize; }
  void allocationBufferSize (int percent) { _allocationBufferSize = percent; }
  int allocationBufferPrice () const { return _allocationBufferPrice; }
  void allocationBufferPrice (int ratio) { _allocationBufferPrice = ratio; }
  int internalTickFrequency () const { return _internalTickFrequency; }
  void internalTickFrequency (int ticks) { _internalTickFrequency = ticks; }
  int firstTimeCheck () const { return _firstTimeCheck; }
  void firstTimeCheck (int percent) { _firstTimeCheck = percent; }
  int firstMemCheck () const { return _firstMemCheck; }
  void firstMemCheck (int percent) { _firstMemCheck = percent; }

  // preprocessing
  int defElimination () const { return _defElimination; }
  void defElimination (int rounds);

  // info options
  char* testId () const { return _testId; }   
  char* tabFile () const { return _tabFile; } 
  bool showOpt () const { return _showOpt; }
  void showOpt (bool so) { _showOpt = so; }
  bool showInput () const { return _showInput; }
  void showInput (bool si) { _showInput = si; }
  bool showPrepro () const { return _showPrepro; }
  void showPrepro (bool pre) { _showPrepro = pre; }
  bool showGenerated () const { return _showGenerated; }
  void showGenerated (bool showgen) { _showGenerated = showgen; }
  bool showPassive () const { return _showPassive; }
  void showPassive ( bool sp ) { _showPassive = sp; }
  bool showActive () const { return _showActive; }
  void showActive (bool sa) { _showActive = sa; }
  bool showDefs () const { return _showDefs; }
  void showDefs (bool sd) { _showDefs = sd; }
  bool showWeightLimitChanges () const { return _showWeightLimitChanges; }
  void showWeightLimitChanges (bool swlc) { _showWeightLimitChanges = swlc; }
  bool showProfile () const { return _showProfile; }
  void showProfile (bool sp) { _showProfile = sp; }
  bool proofOutput () const { return _proofOutput; }
  void proofOutput (bool po) { _proofOutput = po; }

  // structure
  MainLoop _mainLoop;
  int _selection;
    bool _inheritedSelection;
  Precedence _precedence;
  LitComp _literalComparison;
  int _awr;
  int _memoryLimit;
  int _timeLimit;
  int _maxWeight;
  int _maxActive;
  int _maxPassive;
  Hyper _hyper;
    bool _unitConflict;
  Splitting _splitting;
    bool _staticOnly;
    int _groundSplitThreshold;
    bool _naming;
    bool _branchFD;
    bool _branchBD;
  // simplifications
  Subsumption _forwardSubsumption;
    bool _fsSigFilters;
    bool _fsNonAdaptive;
    bool _commutativity;
  Subsumption _backwardSubsumption;
  bool _orphanMurder;
  bool _forwardSubsumptionResolution;
  // equality options
  bool _paramodulation;
  bool _simplifyByEqRes;
  Demodulation _forwardDemodulation;
  Demodulation _backwardDemodulation;
  // memory management (lrs only)
  int _allocationBufferSize;
  int _allocationBufferPrice;
  int _internalTickFrequency;
  int _firstTimeCheck;
  int _firstMemCheck;
  // preprocessing
  int _defElimination;
  // info options
  char*_testId;   // the two strings _testId and _tabFile are not owned by Options
  char*_tabFile; 
  bool _showOpt;
  bool _showInput;
  bool _showPrepro;
  bool _showGenerated;
  bool _showPassive;
  bool _showActive;
  bool _showDefs;
  bool _showWeightLimitChanges;
  bool _showProfile;
  bool _proofOutput;

  void setDefault ();
  static bool validSelection ( int sel );
  static bool completeSelection ( int sel );
  static bool bidirectionalSelection ( int sel );
};


ostream& operator << ( ostream&, Options& );
 

#endif

