//
//  file property.cpp (syntactic properties of problems)
//  06/06/2001 Manchester
//


#include <string.h>


#include "iostream.h"
#include "property.h"
#include "problem.h"
#include "psymbol.h"


// 06/06/2001 Manchester (should be rewritten later for efficiency
Property::Property ( Problem& p )
{
  bool isPropositional = p.isPropositional();
  bool hasEquality = p.hasEquality ();
  bool hasFunctionSymbols = p.hasFunctionSymbols ();
  bool hasNonEquality = p.hasNonEquality();
  bool isHorn = p.isHorn ();
  bool isUnit = p.isUnit ();

  if ( isPropositional ) {
    category (PROP);
  }
  else if ( isUnit ) { // non-propositional but unit
    category ( hasNonEquality ? EUEQ : UEQ );
  }
  else if ( hasEquality ) { // non-propositional, non-unit, equality
    if ( hasNonEquality ) { // non-propositional, non-unit, both equality and non-equality
      category ( isHorn ? HEQ : NEQ );
    }
    else { // non-propositional, non-unit, equality only
      category (PEQ);
    }
  }
  else { // non-propositional, non-unit, non-equality
    if ( hasFunctionSymbols ) { // non-unit, non-equality, function symbols
      category ( isHorn ? HNE : NNE );
    }
    else { // non-propositional, non-unit, non-equality, no function symbols
      category (EPR);
    }
  }

  _equalityInGoal = p.hasEqualityInGoal();
  _goalWeight = p.goalWeight ();
} // Property::Property


void Property::toString ( char* str )
{
  int position = 0;

  switch ( category() ) {
    case NEQ:
      str[position] = 'N';
      break;
    case HEQ:
      str[position] = 'H';
      break;
    case PEQ:
      str[position] = 'P';
      break;
    case UEQ:
      str[position] = 'U';
      break;
    case EUEQ:
      str[position] = 'u';
      break;
    case HNE:
      str[position] = 'h';
      break;
    case NNE:
      str[position] = 'n';
      break;
    case EPR:
      str[position] = 'e';
      break;
    case PROP:
      str[position] = 'p';
      break;
  }
  position ++;

  // computing arities of predicate symbols
  const int max = 4; // maximal interesting arity 
  int arities [max+1];
  for ( int i = 0; i <= max; i++ )
    arities [i] = 0;
  PSymbol::arities ( max, arities );
  {for ( int i = 0; i <= max; i++ ) {
    if ( arities [i] <= 9 ) {
      Int::toString (arities [i], str + position);
    }
    else if ( arities [i] <= 35 ) {
      char c = 'A' + arities [i];
      str[position] = c;
    }
    else {
      str[position] = 'z';
    }
    position++;      
  }}

  // computing arities of function symbols
  {for ( int i = 0; i <= max; i++ )
    arities [i] = 0;}
  FSymbol::arities ( max, arities );
  {for ( int i = 0; i <= max; i++ ) {
    if ( arities [i] <= 9 ) {
      Int::toString (arities [i], str + position);
    }
    else if ( arities [i] <= 35 ) {
      char c = 'A' + arities [i];
      str[position] = c;
    }
    else {
      str[position] = 'z';
    }
    position++;      
  }}

  char wht [10];
  wht[0] = '_';
  Int::toString (_goalWeight, wht+1);
  strcpy (str+position, wht);
  position += strlen (wht);

  str [position++] = _equalityInGoal ? 'y' : 'n';
  str [position] = 0;
} // Property::toString


ostream& operator << ( ostream& str, Property& p )
{
  char pstring [100]; // 100 is too much
  p.toString (pstring);

  str << pstring;

  return str;
} // operator << ( ostream& str, Property& p )


bool Property::hasEquality () const
{
  switch ( category () ) {
    case NEQ:
    case HEQ:
    case PEQ:
    case UEQ:
    case EUEQ:
      return true;

    case HNE:
    case NNE:
    case EPR:
    case PROP:
      default:
      return false;
  }
} // Property::hasEquality ()
