//
//  file psymbol.h 
//  defines class PSymbol
//
//  21/05/2001, Manchester, made from name_table.h and functor.h
//  27/05/2001 Manchester, nth
//  30/05/2001 Manchester, createNew/2, arity/1
//

#ifndef __psymbol__
#define __psymbol__


// ***************** class Psymbol, definition ****************************


class ostream;
class PSymbol;


#include "list.h"


class PSymbol {
 public:
 
  typedef List <PSymbol*> List;

  // query structure
  int arity () const;
  void arity ( int ar ) { _arity = ar; };
  const char* name () const;
  const int number () const { return _number; }

  // miscellaneous
  static PSymbol* create ( const char* name, int arity );
  static PSymbol* createNew ( const char* prefix, int arity );  // new predicate with prefix
  bool isEquality () const;
  static PSymbol* equality();
  static PSymbol** signature (); // return a new array of all symbols
  static int noOfSymbols () { return _noOfSymbols; }
  static PSymbol* nth (int n);  // symbol with the number n
  static void arities ( int max, int* ars );
 
 private:
 
  PSymbol ( const char* name, int arity );
  const char* _name;
  int _arity;
  int _number;

  static PSymbol _equality;
  static int _noOfSymbols;

  // hash table for names
  static const int _noOfBuckets;
  static List * _buckets [];
  static int hash ( const char* );  // hashing function

  // to initialize name table
  static PSymbol _dummy; 
  PSymbol ();

 friend ostream& operator << ( ostream&, const PSymbol* );
}; // class PSymbol



// ***************** class PSymbol, implementation ****************************



inline
PSymbol::PSymbol ( const char* name, int arity )
  :
  _name ( name ),
  _arity ( arity ),
  _number (_noOfSymbols++)
{
}


inline
const char* PSymbol::name () const
{
  return _name;
}


inline
int PSymbol::arity () const
{
  return _arity;
}


inline
PSymbol* PSymbol::equality()
{
  return & _equality;
} // Psymbol::equality


#endif // __psymbol__
